/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/executable/main.cpp

  \brief A structure to hold the set of attributes.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// TerraLib includes
#include <terralib/common.h>
#include <terralib/raster/Raster.h>
#include <terralib/geometry/Geometry.h>
#include <terralib_buildconfig.h>

// Boost Includes
#include <boost/algorithm/string.hpp>
#include <boost/bimap.hpp>
#include <boost/filesystem/convenience.hpp>
#include <boost/shared_ptr.hpp>

// STL includes
#include <stdlib.h>
#include <chrono>  // NOLINT
#include <cstdlib>
#include <ctime>
#include <iostream>
#include <map>
#include <memory>
#include <set>
#include <string>
#include <utility>  // REMOVE IT
#include <vector>

// GeoDMA Includes
#include "BuildConfig.hpp"
#include "geodmalib/common/commandLineProgBar.hpp"
#include "geodmalib/common/commandLineProgViewer.hpp"
#include "geodmalib/common/fileSystem.hpp"
#include "geodmalib/common/observerProgressBar.hpp"
#include "geodmalib/common/utils.hpp"
#include "geodmalib/datamanager/dataManager.hpp"
#include "geodmalib/featureextraction/attribute.hpp"
#include "geodmalib/featureextraction/featuresSummary.hpp"
#include "geodmalib/featureextraction/landscapeFeatures.hpp"
#include "geodmalib/featureextraction/spatialFeatures.hpp"
#include "geodmalib/featureextraction/spectralFeatures.hpp"

enum GeoDMAMode {
    SpectralAndSpatial = 0,
    Spectral = 1,
    Spatial = 2,
    Landscape = 3
};

void usage() {
    std::ostringstream output;
    output << std::endl;
    output << "GeoDMA " << GEODMA_VERSION_STRING << ". (" << __DATE__ << " " << __TIME__ << ")";
    output << std::endl;
    output << std::endl;

    output << "Usage: geodma [option 1] [argument 1] [option 2] [argument 2] ..." << std::endl;
    output << std::endl;

    output << "Options: " << std::endl;
    output << "  --help \t Print short help message with usage instructions and exit." << std::endl;
    output << std::endl;

    output << "  -k, -K \t Receive as argument the kind of attributes to be computed. The options";
    output << " are:" << std::endl;
    output << "\t\t    -> spectralAndSpatial (default) - Computes spectral and spatial attributes.";
    output << std::endl;
    output << "\t\t    -> spectral - Computes spectral attributes." << std::endl;
    output << "\t\t    -> spatial - Computes spatial attributes." << std::endl;
    output << "\t\t    -> landscape - Computes landscape ecology based metrics." << std::endl;
    output << std::endl;

    output << " To compute Spectral and Spatial attributes:" << std::endl;
    output << "  -r, -R \t Receive a Raster file as argument." << std::endl;
    output << "  -v, -V \t Receive a Shape file as argument." << std::endl;
    output << "  -o, -O \t Receive the CSV output file as argument." << std::endl;
    output << "  -s, -S \t Receive the Shape File output file as argument." << std::endl;
    output << std::endl;

    output << " To compute Landscape Ecology based metrics:" << std::endl;
    output << "  -c, -C \t Receive a Shape file corresponding to the coarser (target) layer as ";
    output << "argument." << std::endl;
    output << "  -f, -F \t Receive a Shape file corresponding to the finner (classified) layer as ";
    output << "argument." << std::endl;
    output << "  -o, -O \t Receive the CSV output file as argument." << std::endl;
    output << "  -s, -S \t Receive the Shape File output file as argument." << std::endl;
    output << "  -column \t Receive the name of the column that stores the classes of the objects.";
    output << std::endl;

    std::cout << output.str();
}

int main(int argc, char* argv[]) {
    std::string pathRaster, pathShp, pathCSV, pathOutputShp, pathCoarserLayer, pathFinerLayer;
    std::string columnClasses;
    int mode = GeoDMAMode::SpectralAndSpatial;

    GEODMA_HOME = std::getenv(GEODMA_DIR_VAR_NAME);
    if (GEODMA_HOME) {
        gdma::common::SetEnvVariable(TERRALIB_DIR_VAR_NAME, GEODMA_HOME);
    }

    if (argc > 1) {
        for (int i = 1; i < argc; i++) {
            if (boost::iequals(argv[i], "-k")) {
                if (boost::iequals(argv[i + 1], "spectralAndSpatial")) {
                    mode = GeoDMAMode::SpectralAndSpatial;
                } else if (boost::iequals(argv[i + 1], "spectral")) {
                    mode = GeoDMAMode::Spectral;
                } else if (boost::iequals(argv[i + 1], "spatial")) {
                    mode = GeoDMAMode::Spatial;
                } else if (boost::iequals(argv[i + 1], "landscape")) {
                    mode = GeoDMAMode::Landscape;
                }
                i++;
            } else if (boost::iequals(argv[i], "-r")) {
                pathRaster = argv[i + 1];
                i++;
            } else if (boost::iequals(argv[i], "-v")) {
                pathShp = argv[i + 1];
                i++;
            } else if (boost::iequals(argv[i], "-o")) {
                pathCSV = argv[i + 1];
                i++;
            } else if (boost::iequals(argv[i], "-s")) {
                pathOutputShp = argv[i + 1];
                i++;
            } else if (boost::iequals(argv[i], "-c")) {
                pathCoarserLayer = argv[i + 1];
                i++;
            } else if (boost::iequals(argv[i], "-f")) {
                pathFinerLayer = argv[i + 1];
                i++;
            } else if (boost::iequals(argv[i], "-column")) {
                columnClasses = argv[i + 1];
                i++;
            } else if (std::strcmp(argv[i], "--help") == 0) {
                usage();

                return EXIT_SUCCESS;
            } else {
                std::cout << std::endl << "ERROR: Invalid Option '" << argv[i]
                    << "'. See usage instructions bellow." << std::endl;

                usage();

                return EXIT_FAILURE;
            }
        }
    } else {
        std::cout << std::endl <<
            "WARNING: No files passed as parameter. The default test files will be used."
            << std::endl;
        std::cout << std::endl;
    }

    std::cout << std::endl;

    switch (mode) {
    case GeoDMAMode::SpectralAndSpatial:
    case GeoDMAMode::Spectral:
        if (pathShp.empty())
            pathShp = gdma::common::FindInGeoDMAPath("/data/segmentation_example.shp");
        if (pathRaster.empty())
            pathRaster = gdma::common::FindInGeoDMAPath(
                "/data/GeoReferencedImagesRegisterExample.tif");
        if (pathOutputShp.empty())
            pathOutputShp = "./" + boost::filesystem::basename(pathShp) + "_ATTR.shp";

        std::cout << "Input Raster file: " << pathRaster << std::endl;
        std::cout << "Input Shape File: " << pathShp << std::endl;
        break;
    case GeoDMAMode::Spatial:
        if (pathShp.empty())
            pathShp = gdma::common::FindInGeoDMAPath("/data/segmentation_example.shp");

        if (pathOutputShp.empty())
            pathOutputShp = "./" + boost::filesystem::basename(pathShp) + "_ATTR.shp";

        std::cout << "Input Shape File: " << pathShp << std::endl;
        break;
    case GeoDMAMode::Landscape:
        if (pathCoarserLayer.empty())
            pathCoarserLayer = gdma::common::FindInGeoDMAPath(
                "/data/tests/terraclass_tests/terraclass_cells_tests.shp");
        if (pathFinerLayer.empty())
            pathFinerLayer = gdma::common::FindInGeoDMAPath(
                "/data/tests/terraclass_tests/terraclass_tests.shp");
        if (pathOutputShp.empty())
            pathOutputShp = "./" + boost::filesystem::basename(pathCoarserLayer) + "_ATTR.shp";
        if (columnClasses.empty())
            columnClasses = "tcclasse";

        std::cout << "Input Coarser Layer: " << pathCoarserLayer << std::endl;
        std::cout << "Input Finer File: " << pathFinerLayer << std::endl;
        std::cout << "Classes column: \"" << columnClasses << "\"" << std::endl;
        break;
    default:
        break;
    }

    if (pathCSV.empty())
        pathCSV = "./attributes.csv";

    std::cout << "CSV output file: " << pathCSV << std::endl;
    std::cout << "Output Shape File: " << pathOutputShp << std::endl;
    std::cout << std::endl;

    try {
        TerraLib::getInstance().initialize();
        // Loads Data Access TerraLib Drivers
        if (!gdma::common::loadTerraLibDrivers())
            return EXIT_FAILURE;
    } catch (te::common::Exception &e) {
        std::cerr << "ERROR: " << e.what() << std::endl;
        return EXIT_FAILURE;
    }

    {
        std::unique_ptr<te::rst::Raster> inputRaster;
        std::map<std::string, te::gm::Geometry*> gm_ctner;
        std::vector<unsigned int> bands;
        std::map<std::string, gdma::fte::FeaturesSummary> m_attr_container;

        // Polygons Layer -- Landscape Features
        boost::bimap<std::string, std::size_t> pols_positions;
        std::vector<te::gm::Geometry*> pols_container;
        std::vector<std::string> pol_classes;
        std::set<std::string> class_values;

        // cells Layer  -- Landscape Features
        boost::bimap<std::string, std::size_t> cells_positions;
        std::vector<te::gm::Geometry*> cells_container;

        std::string finerDataSourceId, coarserDataSourceId;

        std::unique_ptr<gdma::common::CommandLineProgViewer> prog(
            new gdma::common::CommandLineProgViewer(100));
        int id = te::common::ProgressManager::getInstance().addViewer(prog.get());

        switch (mode) {
        case GeoDMAMode::SpectralAndSpatial:
            // Opens the input raster
            inputRaster.reset(gdma::dm::openRaster(pathRaster));

            // Initializing vector of bands (Now, all bands will be used. Next versions will allow
            // to the user
            // to choose what band will be used).
            for (int i = 0; i < inputRaster->getNumberOfBands(); i++) {
                bands.push_back(i);
            }

            try {
                // Opens the input shapefile and retrieves its geometries
                std::string dataSourceId = gdma::dm::openShapefile(pathShp);
                gdma::dm::getGeometriesFromDataSource(dataSourceId, gm_ctner);

                // Computing Attributes
                gdma::fte::SpectralFeatures::getFeatures(inputRaster.get(), gm_ctner, bands,
                    m_attr_container);
                gdma::fte::SpatialFeatures::getFeatures(gm_ctner, m_attr_container, true, -1000 );

                // Saving attributes
                gdma::dm::saveAttributesToCSV(m_attr_container, pathShp,
                    boost::filesystem::basename(pathShp), pathCSV);
                gdma::dm::saveToOGRDataSource(pathOutputShp, boost::filesystem::basename(pathShp),
                    pathShp, m_attr_container);
            }
            catch (te::common::Exception &e) {
                std::cerr << "ERROR: " << e.what() << std::endl;
                return EXIT_FAILURE;
            }
            break;

        case GeoDMAMode::Spectral:
            // Opens the input raster
            inputRaster.reset(gdma::dm::openRaster(pathRaster));

            // Initializing vector of bands (Now, all bands will be used. Next versions will allow
            // to the user
            // to choose what band will be used).
            for (int i = 0; i < inputRaster->getNumberOfBands(); i++) {
                bands.push_back(i);
            }

            try {
                // Opens the input shapefile and retrieves its geometries
                std::string dataSourceId = gdma::dm::openShapefile(pathShp);
                gdma::dm::getGeometriesFromDataSource(dataSourceId, gm_ctner);

                // Computing Attributes
                gdma::fte::SpectralFeatures::getFeatures(inputRaster.get(), gm_ctner, bands,
                    m_attr_container);

                // Saving attributes
                gdma::dm::saveAttributesToCSV(m_attr_container, pathShp,
                    boost::filesystem::basename(pathShp), pathCSV);
                gdma::dm::saveToOGRDataSource(pathOutputShp, boost::filesystem::basename(pathShp),
                    pathShp, m_attr_container);
            }
            catch (te::common::Exception &e) {
                std::cerr << "ERROR: " << e.what() << std::endl;
                return EXIT_FAILURE;
            }
            break;

        case GeoDMAMode::Spatial:
            try {
                // Opens the input shapefile and retrieves its geometries
                std::string dataSourceId = gdma::dm::openShapefile(pathShp);
                gdma::dm::getGeometriesFromDataSource(dataSourceId, gm_ctner);

                gdma::fte::SpatialFeatures::getFeatures(gm_ctner, m_attr_container, true, -1000 );

                // Saving attributes
                gdma::dm::saveAttributesToCSV(m_attr_container, pathShp,
                    boost::filesystem::basename(pathShp), pathCSV);
                gdma::dm::saveToOGRDataSource(pathOutputShp, boost::filesystem::basename(pathShp),
                    pathShp, m_attr_container);
            } catch (te::common::Exception &e) {
                std::cerr << "ERROR: " << e.what() << std::endl;
                return EXIT_FAILURE;
            }
            break;
        case GeoDMAMode::Landscape:
            // Opens the input shapefile and retrieves its geometries
            finerDataSourceId = gdma::dm::openShapefile(pathFinerLayer);
            gdma::dm::getGeometriesFromDataSource(finerDataSourceId, pols_positions, pols_container,
                columnClasses, &pol_classes, &class_values);

            coarserDataSourceId = gdma::dm::openShapefile(pathCoarserLayer);
            gdma::dm::getGeometriesFromDataSource(coarserDataSourceId, cells_positions,
                cells_container);
            try {
                gdma::fte::LandscapeFeatures ldscFeat(cells_container, cells_positions,
                    pols_container, pols_positions, pol_classes, class_values);

                ldscFeat.getFeatures(m_attr_container);

                std::map<std::string, int> mapClassNames = ldscFeat.getMapClassNames();

                // Saving attributes
                gdma::dm::saveAttributesToCSV(m_attr_container, pathCoarserLayer,
                    boost::filesystem::basename(pathCoarserLayer), pathCSV);
                gdma::dm::saveToOGRDataSource(pathOutputShp,
                    boost::filesystem::basename(pathCoarserLayer), pathCoarserLayer,
                    m_attr_container);

                std::ostringstream outLegend;
                outLegend << "WARNING: Due to limitations in the size of the column names in ";
                outLegend << std::endl;
                outLegend << "shape files, the names of the classes saved in this kind of file ";
                outLegend << std::endl;
                outLegend << "were mapped to numbers." << std::endl;
                outLegend << "This map was saved in the file \"legend.txt\" and is shown bellow:";
                outLegend << std::endl;

                for (const auto mapValue : mapClassNames) {
                    outLegend << "\t" << mapValue.first << " - " << mapValue.second << std::endl;
                }

                outLegend << "This means that IN THE SHAPE FILE, the attribute class area (CA) ";
                outLegend << std::endl;
                outLegend << "of the class " << mapClassNames.begin()->first << ", for example, ";
                outLegend << "will have the name \"CA_" << mapClassNames.begin()->second << "\".";

                std::cout << std::endl << outLegend.str() << std::endl;

                std::ofstream legendFile;
                legendFile.open("./legend.txt", std::ios::out);
                legendFile << outLegend.str();
                legendFile.close();
            } catch (te::common::Exception &e) {
                std::cerr << "ERROR: " << e.what() << std::endl;
                return EXIT_FAILURE;
            }

            break;
        default:
            break;
        }

        te::common::FreeContents(gm_ctner);
        te::common::FreeContents(pols_container);
    }

    gdma::common::unloadTerraLibDrivers();
    TerraLib::getInstance().finalize();

    std::cout << std::endl << std::endl << "Done!" << std::endl;
    return EXIT_SUCCESS;
}
