/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/geodma_tvplugin/featureExtractionController.cpp

  \brief This file contains ...

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// TerraLib includes
#include <terralib/dataaccess.h>
#include <terralib/dataaccess/utils/Utils.h>
#include <terralib/datatype/Property.h>
#include <terralib/geometry/GeometryProperty.h>
#include <terralib/maptools/AbstractLayer.h>
#include <terralib/qt/widgets/layer/utils/DataSet2Layer.h>

// QtIncludes
#include <QStringList>

// Boost Includes
#include <boost/filesystem.hpp>

// STL includes
#include <list>
#include <string>

// GeoDMA Includes
#include "featureExtractionController.hpp"
#include "geodmalib/datamanager/dataManager.hpp"

gdma::tvplg::FeatureExtractionController::FeatureExtractionController() :
    m_layers(),
    m_rasterLayersNames(),
    m_vectorLayersNames(),
    m_saveCSV(false),
    m_updateDataSource(true),
    m_noDataValue(-1000) {
}


gdma::tvplg::FeatureExtractionController::~FeatureExtractionController() {
}

void gdma::tvplg::FeatureExtractionController::setLayersNames(
    std::list<te::map::AbstractLayerPtr> &layers) {
    m_layers = layers;

    for (auto &layer : m_layers) {
        te::da::DataSetTypePtr dsType = layer.get()->getSchema();

        if (dsType->hasRaster())
            m_rasterLayersNames.push_back(layer);

        if (dsType->hasGeom()) {
            te::gm::GeometryProperty* geomProp = te::da::GetFirstGeomProperty(dsType.get());

            if (geomProp->getGeometryType() == te::gm::PolygonType ||
                geomProp->getGeometryType() == te::gm::MultiPolygonType) {
                m_vectorLayersNames.push_back(layer);
            }
        }
    }
}

te::map::AbstractLayerPtr gdma::tvplg::FeatureExtractionController::getOutputLayer() {
    return m_outputLayer;
}

std::list<te::map::AbstractLayerPtr>
    &gdma::tvplg::FeatureExtractionController::getRasterLayersNames() {
    return m_rasterLayersNames;
}

std::list<te::map::AbstractLayerPtr>
    &gdma::tvplg::FeatureExtractionController::getVectorLayersNames() {
    return m_vectorLayersNames;
}

void gdma::tvplg::FeatureExtractionController::setSelectedDataLayer(std::string layerID) {
    for (auto &layer : m_layers) {
        if (layerID == layer->getId().c_str())
            m_layerData = layer;
    }
}

void gdma::tvplg::FeatureExtractionController::setSelectedTargetLayer(std::string layerID) {
    for (auto &layer : m_layers) {
        if (layerID == layer->getId().c_str())
            m_layerTarget = layer;
    }
}

void gdma::tvplg::FeatureExtractionController::setOutputLayerName(std::string outputLayerName) {
    m_outputLayerName = outputLayerName;
}

void gdma::tvplg::FeatureExtractionController::setOutputCSV(std::string outputCSV) {
    m_outputCSV = outputCSV;
    m_saveCSV = true;
}

void gdma::tvplg::FeatureExtractionController::setNoDataValue(double value) {
    m_noDataValue = value;
}

double gdma::tvplg::FeatureExtractionController::getNoDataValue() {
    return m_noDataValue;
}

void gdma::tvplg::FeatureExtractionController::createOutputLayer() {
    te::da::DataSourceInfoPtr dsInfo = gdma::dm::openDataSource(m_outputDataSet.string());
    te::da::DataSourcePtr outDataSource = te::da::GetDataSource(dsInfo->getId());
    te::qt::widgets::DataSet2Layer converter(dsInfo->getId());
    te::da::DataSetTypePtr dt =
        outDataSource->getDataSetType(boost::filesystem::basename(m_outputDataSet));
    m_outputLayer = converter(dt);
    m_outputLayer->setTitle(m_outputLayerName);
}

void gdma::tvplg::FeatureExtractionController::setOutputDataset(
    boost::filesystem::path outputDataset, bool saveSHP) {
    m_outputDataSet = outputDataset;
    m_saveSHP = saveSHP;
}

bool gdma::tvplg::FeatureExtractionController::getUpdateDataSource() {
    return m_updateDataSource;
}

void gdma::tvplg::FeatureExtractionController::setUpdateDataSource(bool update) {
    m_updateDataSource = update;
}
