/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/geodma_tvplugin/geodmaPluginDialog.cpp

  \brief

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// TerraLib Includes
#include <terralib/common/progress/ProgressManager.h>
#include <terralib/core/translator/Translator.h>
#include <terralib/dataaccess/dataset/DataSet.h>
#include <terralib/dataaccess/dataset/DataSetType.h>
#include <terralib/datatype/Property.h>
#include <terralib/geometry/GeometryProperty.h>
#include <terralib/qt/widgets/datasource/selector/DataSourceSelectorDialog.h>
#include <terralib/qt/widgets/progress/ProgressViewerDialog.h>

// Qt Includes
#include <QDialog>
#include <QFileDialog>
#include <QIntValidator>
#include <QMessageBox>
#include <QStringList>
#include <QStringListModel>

// Boost Includes
#include <boost/filesystem.hpp>
#include <boost/lexical_cast.hpp>

// STL Includes
#include <limits>
#include <list>
#include <map>
#include <memory>
#include <string>
#include <vector>

// GeoDMA Includes
#include "geodmalib/common/messageMediator.hpp"
#include "geodmalib/guicomponents/dialogMessageObserver.hpp"
#include "geodma_tvplugin/featureExtractionDialog.hpp"
#include "geodma_tvplugin/featureExtractionController.hpp"
#include "geodma_tvplugin/landscapeFeaturesController.hpp"
#include "geodma_tvplugin/spectralAndSpatialFeaturesController.hpp"
#include "ui_FeatureExtractionDialogForm.h"


gdma::tvplg::FeatureExtractionDialog::FeatureExtractionDialog(QWidget* parent, Qt::WindowFlags f) :
    QDialog(parent, f),
    m_ui(new Ui::FeatureExtractionDialogForm) {
    // add controls
    m_ui->setupUi(this);

    // add icons
    this->setWindowIcon(QIcon::fromTheme("icon_geodma"));
    m_ui->m_gdmaImgLabel->setPixmap(QIcon::fromTheme("icon_geodma_name").pixmap(112, 65));
    m_ui->m_featExtImgLabel->setPixmap(QIcon::fromTheme("raster-vector-hint").pixmap(112, 48));
    m_ui->m_targetDatasourceToolButton->setIcon(QIcon::fromTheme("datasource"));

    m_dialogTitle = tr("GeoDMA Feature Extraction");

    int spTabIndex = m_ui->m_kindOfFeaturesTab->indexOf(m_ui->m_spectralAndSpatialTab);
    int ldscTabIndex = m_ui->m_kindOfFeaturesTab->indexOf(m_ui->m_landscapeTab);
    m_controllers[spTabIndex] = std::shared_ptr<gdma::tvplg::FeatureExtractionController>(
        new gdma::tvplg::SpectralAndSpatialFeaturesController());
    m_controllers[ldscTabIndex] = std::shared_ptr<gdma::tvplg::FeatureExtractionController>(
        new gdma::tvplg::LandscapeFeaturesController());

    m_ui->m_kindOfFeaturesTab->setCurrentWidget(m_ui->m_spectralAndSpatialTab);

    m_ui->m_outNoDataLineEdit->setValidator(new QIntValidator());
    m_ui->m_outNoDataLineEdit->setText(
        boost::lexical_cast<std::string>(m_controllers[ldscTabIndex]->getNoDataValue()).c_str());

    m_ui->m_helpPushButton->setNameSpace("dpi.inpe.br.geodma");
    m_ui->m_helpPushButton->setPageReference(
        "geodma/feature_extraction/geodma_feature_extraction.html");

    connect(m_ui->m_kindOfFeaturesTab, SIGNAL(currentChanged(int)), this,
         SLOT(onKindOfFeatTabChanged(int)));

    // Tab Spatial and Spectral Features
    connect(m_ui->m_inRasterComboBox, SIGNAL(currentIndexChanged(int)), this,
        SLOT(onRasterComboBoxSelected(int)));
    connect(m_ui->m_inVectorComboBox, SIGNAL(currentIndexChanged(int)), this,
        SLOT(onVectorComboBoxSelected(int)));
    connect(m_ui->m_bandsListWidget, SIGNAL(itemSelectionChanged()), this,
        SLOT(onBandsSelectedChanged()));
    connect(m_ui->m_spectral_box, SIGNAL(toggled(bool)), this,
        SLOT(onKindCheckBoxChanged()));
    connect(m_ui->m_spatial_box, SIGNAL(toggled(bool)), this,
        SLOT(onKindCheckBoxChanged()));


    // Tab Landscape Based Features
    connect(m_ui->m_inTargetVectorComboBox, SIGNAL(currentIndexChanged(int)), this,
        SLOT(onVectorComboBoxSelected(int)));
    connect(m_ui->m_inClassVectorComboBox, SIGNAL(currentIndexChanged(int)), this,
        SLOT(onClassVectorComboBoxChanged(int)));
    connect(m_ui->m_inColumnComboBox, SIGNAL(currentIndexChanged(int)), this,
        SLOT(onColumnComboBoxChanged(int)));

    // Common widgets
    connect(m_ui->m_targetFileToolButton, SIGNAL(pressed()), this,
        SLOT(onSHPTargetFileToolButtonPressed()));
    connect(m_ui->m_targetCSVFileToolButton, SIGNAL(pressed()), this,
        SLOT(onCSVTargetFileToolButtonPressed()));
    connect(m_ui->m_targetDatasourceToolButton, SIGNAL(pressed()), this,
        SLOT(onTargetDatasourceToolButtonPressed()));
    connect(m_ui->m_outNoDataLineEdit, SIGNAL(editingFinished()), this,
        SLOT(onNoDataLineEditChanged()));
    connect(m_ui->m_newLayerNameLineEdit, SIGNAL(editingFinished()), this,
        SLOT(onNewLayerNameLineEditEdited()));

    connect(m_ui->m_okPushButton, SIGNAL(clicked()), this, SLOT(onOkPushButtonClicked()));
    connect(m_ui->m_cancelPushButton, SIGNAL(clicked()), this, SLOT(onCancelPushButtonClicked()));
}

gdma::tvplg::FeatureExtractionDialog::~FeatureExtractionDialog() {}

void gdma::tvplg::FeatureExtractionDialog::setLayers(std::list<te::map::AbstractLayerPtr> layers) {
    double hasRaster = false;

    for (auto &controller : m_controllers)
        controller.second->setLayersNames(layers);

    int spTabIndex = m_ui->m_kindOfFeaturesTab->indexOf(m_ui->m_spectralAndSpatialTab);
    std::list<te::map::AbstractLayerPtr> rasterLayers =
        m_controllers[spTabIndex]->getRasterLayersNames();
    std::list<te::map::AbstractLayerPtr> vectorLayers =
        m_controllers[spTabIndex]->getVectorLayersNames();

    for (auto &rstLayer : rasterLayers)
        m_ui->m_inRasterComboBox->addItem(QString(rstLayer.get()->getTitle().c_str()),
            QVariant(rstLayer.get()->getId().c_str()));

    for (auto &vecLayer : vectorLayers) {
        // Spatial and Spectral Tab
        m_ui->m_inVectorComboBox->addItem(QString(vecLayer.get()->getTitle().c_str()),
            QVariant(vecLayer.get()->getId().c_str()));

        // Landscape Features Tab
        m_ui->m_inClassVectorComboBox->addItem(QString(vecLayer.get()->getTitle().c_str()),
            QVariant(vecLayer.get()->getId().c_str()));
        m_ui->m_inTargetVectorComboBox->addItem(QString(vecLayer.get()->getTitle().c_str()),
            QVariant(vecLayer.get()->getId().c_str()));
    }

    hasRaster = !rasterLayers.empty();

    m_ui->m_spectral_box->setChecked(hasRaster);
    if (!hasRaster) {
        m_ui->m_spectral_box->setDisabled(true);
        ((gdma::tvplg::SpectralAndSpatialFeaturesController*)m_controllers[spTabIndex].get())
            ->setComputeSpectral(false);
    }

    m_ui->m_spatial_box->setChecked(true);
}

te::map::AbstractLayerPtr gdma::tvplg::FeatureExtractionDialog::getOutputLayer() {
    return m_controllers[m_ui->m_kindOfFeaturesTab->currentIndex()]->getOutputLayer();
}

void gdma::tvplg::FeatureExtractionDialog::onBandsSelectedChanged() {
    std::vector<unsigned int> vecBands;

    for (auto &item : m_ui->m_bandsListWidget->selectedItems())
        vecBands.push_back(item->text().toInt());

    int spTabIndex = m_ui->m_kindOfFeaturesTab->indexOf(m_ui->m_spectralAndSpatialTab);
    ((gdma::tvplg::SpectralAndSpatialFeaturesController*)
        m_controllers[spTabIndex].get())->setInputBands(vecBands);
}

void gdma::tvplg::FeatureExtractionDialog::onKindCheckBoxChanged() {
    QCheckBox* sender = reinterpret_cast<QCheckBox*>(QObject::sender());
    bool checked = sender->isChecked();

    int spTabIndex = m_ui->m_kindOfFeaturesTab->indexOf(m_ui->m_spectralAndSpatialTab);

    if (sender == m_ui->m_spectral_box)
        ((gdma::tvplg::SpectralAndSpatialFeaturesController*)
            m_controllers[spTabIndex].get())->setComputeSpectral(checked);
    else if (sender == m_ui->m_spatial_box)
        ((gdma::tvplg::SpectralAndSpatialFeaturesController*)
            m_controllers[spTabIndex].get())->setComputeSpatial(checked);
}

void gdma::tvplg::FeatureExtractionDialog::onKindOfFeatTabChanged(int) {
    // Restarting controller
    if (m_ui->m_kindOfFeaturesTab->currentWidget() == m_ui->m_spectralAndSpatialTab) {
        m_ui->m_outNoDataLineEdit->setDisabled(true);
    } else if (m_ui->m_kindOfFeaturesTab->currentWidget() == m_ui->m_landscapeTab) {
        m_ui->m_outNoDataLineEdit->setEnabled(true);
    }
}

void gdma::tvplg::FeatureExtractionDialog::onRasterComboBoxSelected(int index) {
    std::string layerID = reinterpret_cast<QComboBox*>(QObject::sender())->itemData(
        index, Qt::UserRole).toString().toUtf8().data();

    int spTabIndex = m_ui->m_kindOfFeaturesTab->indexOf(m_ui->m_spectralAndSpatialTab);
    m_controllers[spTabIndex]->setSelectedDataLayer(layerID);

    m_ui->m_bandsListWidget->clear();

    std::size_t n_bands;

    ((gdma::tvplg::SpectralAndSpatialFeaturesController*)
        m_controllers[spTabIndex].get())->getRasterProperties(n_bands);

    for (std::size_t b = 0; b < n_bands; ++b)
        m_ui->m_bandsListWidget->addItem(boost::lexical_cast<std::string>(b).c_str());

    m_ui->m_bandsListWidget->selectAll();
}

void gdma::tvplg::FeatureExtractionDialog::onVectorComboBoxSelected(int index) {
    std::string layerID = reinterpret_cast<QComboBox*>(QObject::sender())->itemData(
        index, Qt::UserRole).toString().toUtf8().data();

    for (auto &controller : m_controllers)
        controller.second->setSelectedTargetLayer(layerID);
}

void gdma::tvplg::FeatureExtractionDialog::onClassVectorComboBoxChanged(int index) {
    std::string layerID = reinterpret_cast<QComboBox*>(QObject::sender())->itemData(
        index, Qt::UserRole).toString().toUtf8().data();

    int ldscTabIndex = m_ui->m_kindOfFeaturesTab->indexOf(m_ui->m_landscapeTab);
    m_controllers[ldscTabIndex]->setSelectedDataLayer(layerID);

    m_ui->m_inColumnComboBox->clear();

    std::vector<std::string> propNames;
    ((gdma::tvplg::LandscapeFeaturesController*)
        m_controllers[ldscTabIndex].get())->getVectorProperties(propNames);

    for (auto &prop : propNames)
        m_ui->m_inColumnComboBox->addItem(prop.c_str());
}

void gdma::tvplg::FeatureExtractionDialog::onColumnComboBoxChanged(int index) {
    std::string className = reinterpret_cast<QComboBox*>(QObject::sender())->itemText(
        index).toUtf8().data();

    int ldscTabIndex = m_ui->m_kindOfFeaturesTab->indexOf(m_ui->m_landscapeTab);
    ((gdma::tvplg::LandscapeFeaturesController*)
        m_controllers[ldscTabIndex].get())->setSelectedClassName(className);
}

void gdma::tvplg::FeatureExtractionDialog::onNoDataLineEditChanged() {
    double noDataValue = reinterpret_cast<QLineEdit*>(QObject::sender())->text().toDouble();

    for (auto &controller : m_controllers)
        controller.second->setNoDataValue(noDataValue);
}

void gdma::tvplg::FeatureExtractionDialog::onSHPTargetFileToolButtonPressed() {
    m_ui->m_newLayerNameLineEdit->clear();
    m_ui->m_repositoryLineEdit->clear();

    QFileInfo fileInfo;
    QString fileName;

    do {
        fileName = QFileDialog::getSaveFileName(this, tr("Save as..."),
            QString(), tr("ESRI Shapefile (*.shp *.SHP);;"), 0, QFileDialog::DontConfirmOverwrite);

        fileInfo = QFileInfo(fileName);

        if (fileName.isEmpty())
            return;

        if (fileInfo.exists()) {
            QString message = QString("\"") + fileName +
                QString("\" already exists.\nSelect another file!");
            gdma::common::MessageMediator::getInstance().sendWarningMsg(
                TE_TR(message.toStdString().c_str()));
        }
    } while (fileInfo.exists());

    boost::filesystem::path outfile(fileName.toLatin1().data());
    std::string outputLayerName = outfile.stem().string();
    m_ui->m_newLayerNameLineEdit->setText(outputLayerName.c_str());
    // boost::filesystem::path outputFileStr = outfile.string();
    m_ui->m_repositoryLineEdit->setText(fileName);

    for (int tab = 0; tab < m_ui->m_kindOfFeaturesTab->count(); tab++) {
        m_controllers[tab]->setOutputDataset(outfile, true);
        m_controllers[tab]->setOutputLayerName(outputLayerName);
    }
}

void gdma::tvplg::FeatureExtractionDialog::onTargetDatasourceToolButtonPressed() {
    m_ui->m_newLayerNameLineEdit->clear();
    m_ui->m_newLayerNameLineEdit->setEnabled(true);

    te::qt::widgets::DataSourceSelectorDialog dlg(this);
    dlg.exec();

    std::list<te::da::DataSourceInfoPtr> dsPtrList = dlg.getSelecteds();

    if (dsPtrList.empty())
        return;

    std::list<te::da::DataSourceInfoPtr>::iterator it = dsPtrList.begin();
    m_ui->m_repositoryLineEdit->setText(QString(it->get()->getTitle().c_str()));

    // @TODO(Raian): Finish HERE -- see attributeFill example
}

void gdma::tvplg::FeatureExtractionDialog::onCSVTargetFileToolButtonPressed() {
    m_ui->m_outCSVLineEdit->clear();

    QString fileName = QFileDialog::getSaveFileName(this, tr("Save as..."),
        QString(), tr("CSV File (*.csv *.CSV);;"), 0);

    boost::filesystem::path outfile(fileName.toUtf8().data());
    std::string aux = outfile.leaf().string();
    aux = outfile.string();
    m_ui->m_outCSVLineEdit->setText(aux.c_str());
}

void gdma::tvplg::FeatureExtractionDialog::onOkPushButtonClicked() {
    if (m_ui->m_kindOfFeaturesTab->currentWidget() == m_ui->m_spectralAndSpatialTab) {
        if (m_ui->m_spectral_box->isChecked() && m_ui->m_inRasterComboBox->count() == 0) {
            QMessageBox::warning(this, m_dialogTitle, tr("Select an input Raster Layer."));
            return;
        }

        if (m_ui->m_inVectorComboBox->count() == 0) {
            QMessageBox::warning(this, m_dialogTitle, tr("Select an input Vector Layer."));
            return;
        }
    } else if (m_ui->m_kindOfFeaturesTab->currentWidget() == m_ui->m_landscapeTab) {
        if (m_ui->m_inClassVectorComboBox->count() == 0) {
            QMessageBox::warning(this, m_dialogTitle,
                tr("Select an input Classified Vector Layer."));
            return;
        }

        if (m_ui->m_inTargetVectorComboBox->count() == 0) {
            QMessageBox::warning(this, m_dialogTitle,
                tr("Select an input Target Vector Layer."));
            return;
        }
    }

    if (m_ui->m_repositoryLineEdit->text().isEmpty()) {
        QMessageBox::warning(this, m_dialogTitle, tr("Define a repository for the result."));
        return;
    }

    if (m_ui->m_newLayerNameLineEdit->text().isEmpty()) {
        QMessageBox::warning(this, m_dialogTitle, tr("Define a name for the resulting layer."));
        return;
    }

    if (!m_ui->m_outCSVLineEdit->text().isEmpty()) {
        for (int tab = 0; tab < m_ui->m_kindOfFeaturesTab->count(); tab++)
            m_controllers[tab]->setOutputCSV(m_ui->m_outCSVLineEdit->text().toUtf8().data());
    }

    // progress
    te::qt::widgets::ProgressViewerDialog v(this);
    int progId = te::common::ProgressManager::getInstance().addViewer(&v);

    // Messages
    gdma::guicmp::DialogMessageObserver msgObs(this, m_dialogTitle.toStdString());
    int msgObsId = gdma::common::MessageMediator::getInstance().addObserver(&msgObs);

    this->setCursor(Qt::WaitCursor);
    try {
        bool returnValue = m_controllers[m_ui->m_kindOfFeaturesTab->currentIndex()]->computeFeatures();
        if( returnValue == false )
        {
          this->setCursor(Qt::ArrowCursor);
          QMessageBox::critical(this, m_dialogTitle, tr("Unknown error found!"));
          reject();
          return;
        }
    } catch(te::common::Exception &e) {
        this->setCursor(Qt::ArrowCursor);
        QMessageBox::critical(this, m_dialogTitle, e.what());
        reject();
        return;
    } catch (std::exception &e) {
        this->setCursor(Qt::ArrowCursor);
        QMessageBox::critical(this, m_dialogTitle, e.what());
        reject();
        return;
    } catch (...) {
        this->setCursor(Qt::ArrowCursor);
        QMessageBox::critical(this, m_dialogTitle, tr("Unknown error found!"));
        reject();
        return;
    }

    QMessageBox::information(this, m_dialogTitle, tr("Done!"));

    te::common::ProgressManager::getInstance().removeViewer(progId);
    gdma::common::MessageMediator::getInstance().removeObserver(msgObsId);

    this->setCursor(Qt::ArrowCursor);
    accept();
}

void gdma::tvplg::FeatureExtractionDialog::onCancelPushButtonClicked() {
    reject();
}

void gdma::tvplg::FeatureExtractionDialog::onNewLayerNameLineEditEdited() {
    std::string layerName = reinterpret_cast<QLineEdit*>(QObject::sender())->text().toStdString();
    for (auto &controller : m_controllers)
        controller.second->setOutputLayerName(layerName);
}
