/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/geodma_tvplugin/landscapeFeaturesController.cpp

  \brief This file contains ...

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// TerraLib Includes
#include <terralib/common/STLUtils.h>
#include <terralib/core/translator/Translator.h>
#include <terralib/dataaccess/dataset/DataSetType.h>
#include <terralib/datatype/Property.h>

// QtIncludes
#include <QStringList>

// Boost Includes
#include <boost/bimap.hpp>
#include <boost/filesystem.hpp>
#include <boost/filesystem/fstream.hpp>

// STL Includes
#include <locale>
#include <map>
#include <memory>
#include <set>
#include <string>
#include <vector>

// GeoDMA Includes
#include "geodmalib/common/messageMediator.hpp"
#include "geodmalib/datamanager/dataManager.hpp"
#include "geodmalib/featureextraction/landscapeFeatures.hpp"
#include "geodma_tvplugin/featureExtractionController.hpp"
#include "geodma_tvplugin/landscapeFeaturesController.hpp"


gdma::tvplg::LandscapeFeaturesController::LandscapeFeaturesController() {
    m_noDataValue = -1000;
}


gdma::tvplg::LandscapeFeaturesController::~LandscapeFeaturesController() {
}

void gdma::tvplg::LandscapeFeaturesController::setSelectedClassName(std::string className) {
    this->m_className = className;
}

void gdma::tvplg::LandscapeFeaturesController::getVectorProperties(
    std::vector<std::string> &propNames) {
    std::unique_ptr<te::da::DataSetType> dsType = m_layerData->getSchema();
    std::vector<te::dt::Property*> propVector = dsType->getProperties();
    te::da::PrimaryKey* pk = dsType->getPrimaryKey();
    std::vector<te::dt::Property*> pkVector = pk->getProperties();

    for (auto &prop : propVector) {
        if (prop->getType() == te::dt::STRING_TYPE && prop->getName().compare("id") != 0) {
            for (auto &key : pkVector) {
                if (key->getName().compare(prop->getName()) != 0)
                    propNames.push_back(prop->getName());
            }
        }
    }
}

bool gdma::tvplg::LandscapeFeaturesController::computeFeatures() {
    // Polygons Layer -- Landscape Features
    boost::bimap<std::string, std::size_t> pols_positions;
    std::vector<te::gm::Geometry*> pols_container;
    std::vector<std::string> pol_classes;
    std::set<std::string> class_values;

    gdma::dm::getGeometriesFromDataSource(m_layerData->getDataSourceId(), pols_positions,
        pols_container, m_className, &pol_classes, &class_values);

    // cells Layer  -- Landscape Features
    boost::bimap<std::string, std::size_t> cells_positions;
    std::vector<te::gm::Geometry*> cells_container;

    gdma::dm::getGeometriesFromDataSource(m_layerTarget->getDataSourceId(), cells_positions,
        cells_container);

    gdma::fte::LandscapeFeatures ldscFeat(cells_container, cells_positions, pols_container,
        pols_positions, pol_classes, class_values, m_noDataValue);

    std::map<std::string, gdma::fte::FeaturesSummary> attr_container;

    ldscFeat.getFeatures(attr_container);

    if (m_saveSHP) {
        if (!gdma::dm::saveToOGRDataSource(m_outputDataSet.string(),
                m_layerTarget->getDataSetName(), m_layerTarget->getDataSourceId(), attr_container))
            return false;
    }

    if (m_saveCSV) {
        if (!gdma::dm::saveAttributesToCSV(attr_container, m_layerTarget->getDataSourceId(),
            m_layerTarget->getDataSetName(), m_outputCSV))
            return false;
    }

    if (m_saveSHP) {
        std::map<std::string, int> mapClassNames = ldscFeat.getMapClassNames();

        boost::filesystem::path legfileName(boost::filesystem::basename(m_outputDataSet) + "_leg.txt");
        boost::filesystem::path legFile = m_outputDataSet.parent_path() / legfileName;
        std::string legFileStr = QString::fromLatin1(legFile.string().c_str()).toStdString();

        std::ostringstream outLegend;
        outLegend << "Due to limitations in the size of the column names in ";
        outLegend << std::endl;
        outLegend << "shape files, the names of the classes saved in this kind of file ";
        outLegend << std::endl;
        outLegend << "were mapped to numbers." << std::endl;
        outLegend << "This map was saved in the file \"" << legFileStr << "\" and is shown bellow:";
        outLegend << std::endl;

        std::ostringstream legend;
        for (const auto mapValue : mapClassNames) {
            outLegend << "\t" << mapValue.first << " - " << mapValue.second << std::endl;
            legend << mapValue.first << " - " << mapValue.second << std::endl;
        }

        outLegend << "This means that IN THE SHAPE FILE, the attribute class area (CA) ";
        outLegend << std::endl;
        outLegend << "of the class " << mapClassNames.begin()->first << ", for example, ";
        outLegend << "will have the name \"CA_" << mapClassNames.begin()->second << "\".";

        gdma::common::MessageMediator::getInstance().sendInformationMsg(TE_TR(outLegend.str()));

        boost::filesystem::ofstream legendFile;
        legendFile.open(legFile);
        legendFile << legend.str();
        legendFile.close();
    }

    // @TODO(Raian): Verify if it is really necessary
    te::common::FreeContents(pols_container);
    te::common::FreeContents(cells_container);

    createOutputLayer();

    return true;
}
