/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/geodma_tvplugin/plugin.cpp

  \brief

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// TerraLib Includes
#include <terralib/core/plugin/CppPlugin.h>
#include <terralib/core/translator/Translator.h>
#include <terralib/qt/af/ApplicationController.h>
#include <terralib/qt/af/Utils.h>
#include <terralib/qt/widgets/help/HelpManager.h>
// #include <terralib/qt/widgets/help/AssistantHelpManagerImpl.h>

// Boost Includes
#include <boost/filesystem.hpp>

// STL Includes
#include <string>

// GeoDMA includes
#include "geodma_tvplugin/classification/c50ClassificationAction.hpp"
#include "geodma_tvplugin/featureExtractionAction.hpp"
#include "geodma_tvplugin/helpAction.hpp"
#include "geodma_tvplugin/plugin.hpp"
#include "geodma_tvplugin/sampleSelectionAction.hpp"
#include "geodmalib/common/fileSystem.hpp"


gdma::tvplg::Plugin::Plugin(const te::core::PluginInfo& pluginInfo) :
    te::core::CppPlugin(pluginInfo),
    m_GeoDMAMenu(0),
    m_classificationMenu(0) {
}


gdma::tvplg::Plugin::~Plugin() { }

void gdma::tvplg::Plugin::startup() {
    if (m_initialized)
        return;

    // add plugin menu
    m_GeoDMAMenu = te::qt::af::AppCtrlSingleton::getInstance().getMenu("GeoDMA");
    m_classificationMenu = new QMenu(m_GeoDMAMenu);
    m_classificationMenu->setTitle(TE_TR("Classification"));
    m_classificationMenu->setIcon(QIcon::fromTheme("classifier"));

    registerActions();

    // add pop up menu
    m_classifPopupAction = new QAction(m_classificationMenu);
    m_classifPopupAction->setText(TE_TR("Classification"));

    te::qt::af::AddActionToCustomToolbars(&te::qt::af::AppCtrlSingleton::getInstance(),
        m_featureExtractionAction->getAction());
    te::qt::af::AddActionToCustomToolbars(&te::qt::af::AppCtrlSingleton::getInstance(),
        m_samplesSelectionAction->getAction());
    te::qt::af::AddActionToCustomToolbars(&te::qt::af::AppCtrlSingleton::getInstance(),
        m_c50ClassificationAction->getAction());
    te::qt::af::AddActionToCustomToolbars(&te::qt::af::AppCtrlSingleton::getInstance(),
        m_kmeansClassificationAction->getAction());
    te::qt::af::AddActionToCustomToolbars(&te::qt::af::AppCtrlSingleton::getInstance(),
        m_helpAction->getAction());

    /*te::qt::widgets::AssistantHelpManagerImpl* impl =
        new te::qt::widgets::AssistantHelpManagerImpl(
        "C:/Users/Raian/Documents/INPE/Projeto_BNDES/source_code/terralib/built/share/terraview/help/geodma/geodma_help.qhc", this);
    te::qt::widgets::HelpManager::getInstance().setImpl(impl);*/

    /*auto i = static_cast<te::qt::widgets::AssistantHelpManagerImpl*>(te::qt::widgets::HelpManager::getInstance().getImpl());
    i->startAssistant();*/
    // te::qt::widgets::HelpManager::getInstance().appendDoc("C:/Users/Raian/Documents/INPE/Projeto_BNDES/source_code/terralib/built/share/terraview/help/geodma/geodma_help.qhc");

    m_initialized = true;

    te::qt::af::AppCtrlSingleton::getInstance().addListener(this, te::qt::af::SENDER);
}

void gdma::tvplg::Plugin::shutdown() {
    if (!m_initialized)
        return;

    // Remove Menus
    delete m_classifPopupAction;
    m_classifPopupAction = 0;
    delete m_classificationMenu;
    m_classificationMenu = 0;
    delete m_GeoDMAMenu;
    m_GeoDMAMenu = 0;

    // Unregister Actions
    unRegisterActions();

    m_initialized = false;

    te::qt::af::AppCtrlSingleton::getInstance().removeListener(this);
}

void gdma::tvplg::Plugin::registerActions() {
    QAction* pluginsSeparator =
        te::qt::af::AppCtrlSingleton::getInstance().findAction("ManagePluginsSeparator");

    m_featureExtractionAction = new gdma::tvplg::FeatureExtractionAction(m_GeoDMAMenu);
    m_samplesSelectionAction = new gdma::tvplg::SampleSelectionAction(m_GeoDMAMenu);
    m_GeoDMAMenu->insertMenu(pluginsSeparator, m_classificationMenu);
    m_c50ClassificationAction = new gdma::tvplg::C50ClassificationAction(m_classificationMenu);
    m_kmeansClassificationAction = new gdma::tvplg::KMeansClassificationAction(m_classificationMenu);
    m_helpAction = new gdma::tvplg::HelpAction(m_GeoDMAMenu);

    connect(m_featureExtractionAction, SIGNAL(triggered(te::qt::af::evt::Event*)),
        SIGNAL(triggered(te::qt::af::evt::Event*)));
    connect(m_samplesSelectionAction, SIGNAL(triggered(te::qt::af::evt::Event*)),
        SIGNAL(triggered(te::qt::af::evt::Event*)));
    connect(m_c50ClassificationAction, SIGNAL(triggered(te::qt::af::evt::Event*)),
        SIGNAL(triggered(te::qt::af::evt::Event*)));
    connect(m_kmeansClassificationAction, SIGNAL(triggered(te::qt::af::evt::Event*)),
        SIGNAL(triggered(te::qt::af::evt::Event*)));
    connect(m_helpAction, SIGNAL(triggered(te::qt::af::evt::Event*)),
        SIGNAL(triggered(te::qt::af::evt::Event*)));
}

void gdma::tvplg::Plugin::unRegisterActions() {
    delete m_featureExtractionAction;
    m_featureExtractionAction = 0;
    delete m_samplesSelectionAction;
    m_samplesSelectionAction = 0;
    delete m_c50ClassificationAction;
    m_c50ClassificationAction = 0;
    delete m_kmeansClassificationAction;
    m_kmeansClassificationAction = 0;
    delete m_helpAction;
    m_helpAction = 0;
}

TERRALIB_PLUGIN_CALL_BACK_IMPL(gdma::tvplg::Plugin)
