/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/geodma_tvplugin/sampleSelectionDialog.cpp

  \brief This file contains the SampleSelectionDialog implementation.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// TerraLib Includes
#include <terralib/common/progress/ProgressManager.h>
#include <terralib/qt/widgets/canvas/MapDisplay.h>
#include <terralib/qt/widgets/progress/ProgressViewerDialog.h>

// Qt Includes
#include <QDialog>
#include <QGridLayout>
#include <QMessageBox>

// STL Includes
#include <list>
#include <memory>
#include <string>

// GeoDMA Includes
#include "geodmalib/common/messageMediator.hpp"
#include "geodmalib/guicomponents/dialogMessageObserver.hpp"
#include "geodma_tvplugin/sampleSelectionDialog.hpp"
#include "ui_SampleSelectionDialogForm.h"


gdma::tvplg::SampleSelectionDialog::SampleSelectionDialog(QWidget* parent, Qt::WindowFlags f) :
    QDialog(parent, f),
    m_ui(new Ui::SampleSelectionDialogForm) {
    m_ui->setupUi(this);

    m_dialogTitle = "GeoDMA Sample Selection";

    // Message Dialog
    gdma::guicmp::DialogMessageObserver* msgObs =
        new gdma::guicmp::DialogMessageObserver(this, m_dialogTitle);
    m_msgObsId = gdma::common::MessageMediator::getInstance().addObserver(msgObs);

    // progress
    m_progViewer.reset(new te::qt::widgets::ProgressViewerDialog(this));
    m_progId = te::common::ProgressManager::getInstance().addViewer(m_progViewer.get());

    QGridLayout* layout = new QGridLayout(m_ui->m_widget);
    m_sampleSelWidget.reset(new gdma::guicmp::SampleSelectionWidget(m_ui->m_widget));
    layout->addWidget(m_sampleSelWidget.get(), 0, 0);
    layout->setContentsMargins(0, 0, 0, 0);

    // add icons
    this->setWindowIcon(QIcon::fromTheme("icon_geodma"));
    m_ui->m_gdmaImgLabel->setPixmap(QIcon::fromTheme("icon_geodma_name").pixmap(112, 65));
    m_ui->m_sampSelImgLabel->setPixmap(QIcon::fromTheme("sa-samplepoint-hint").pixmap(112, 48));

    m_ui->m_helpPushButton->setNameSpace("dpi.inpe.br.geodma");
    m_ui->m_helpPushButton->setPageReference(
        "geodma/sample_selection/geodma_sample_selection.html");

    connect(m_ui->m_okPushButton, SIGNAL(released()), this, SLOT(onOkPushButtonClicked()));
    connect(m_ui->m_okPushButton, SIGNAL(released()), this, SLOT(close()));

    connect(m_ui->m_cancelPushButton, SIGNAL(released()), this, SLOT(onCancelPushButtonClicked()));
    connect(m_ui->m_cancelPushButton, SIGNAL(released()), this, SLOT(close()));
}


gdma::tvplg::SampleSelectionDialog::~SampleSelectionDialog() {
    gdma::common::MessageMediator::getInstance().removeObserver(m_msgObsId);
    te::common::ProgressManager::getInstance().removeViewer(m_progId);
}

void gdma::tvplg::SampleSelectionDialog::setLayers(std::list<te::map::AbstractLayerPtr> layers) {
    m_sampleSelWidget->setLayers(layers);
}

void gdma::tvplg::SampleSelectionDialog::setMapDisplay(te::qt::widgets::MapDisplay* mapDisplay) {
    m_sampleSelWidget->setMapDisplay(mapDisplay);
}

void gdma::tvplg::SampleSelectionDialog::setActionGroup(QActionGroup* actionGroup) {
    m_sampleSelWidget->setActionGroup(actionGroup);
}

void gdma::tvplg::SampleSelectionDialog::clearCanvas() {
    m_sampleSelWidget->clearCanvas();
}

void gdma::tvplg::SampleSelectionDialog::closeEvent(QCloseEvent* e) {
    m_sampleSelWidget->disableTools();
    this->clearCanvas();

    emit closeTool();
}

void gdma::tvplg::SampleSelectionDialog::drawSamples() {
    m_sampleSelWidget->drawSamples();
}

// ------------- Slots (Public)
void gdma::tvplg::SampleSelectionDialog::onOkPushButtonClicked() {
    if (m_sampleSelWidget->hasUnsavedSamples()) {
        m_sampleSelWidget->saveSamples();
    }

    m_sampleSelWidget->saveSamplesInCSV();
}

void gdma::tvplg::SampleSelectionDialog::onCancelPushButtonClicked() {
    if (m_sampleSelWidget->hasUnsavedSamples()) {
        QMessageBox::StandardButton reply;
        reply = QMessageBox::question(this, "Unsaved Samples",
            "You have unsaved samples. Do you want to save them?",
            QMessageBox::Yes | QMessageBox::No);

        if (reply == QMessageBox::Yes) {
            m_sampleSelWidget->saveSamples();
            m_sampleSelWidget->saveSamplesInCSV();
        }
    }
}
