/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/geodma_tvplugin/spectralAndSpatialFeaturesController.cpp

  \brief This file contains ...

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// TerraLib includes
#include <terralib/dataaccess/utils/Utils.h>
#include <terralib/maptools/AbstractLayer.h>
#include <terralib/maptools/Utils.h>
#include <terralib/raster/Band.h>
#include <terralib/raster/BandProperty.h>
#include <terralib/raster/Grid.h>
#include <terralib/raster/Raster.h>
#include <terralib/raster/RasterProperty.h>

// QtIncludes
#include <QStringList>

#include <memory>
#include <string>
#include <vector>
#include <map>

// GeoDMA Includes
#include "geodmalib/datamanager/dataManager.hpp"
#include "geodmalib/featureextraction/featuresSummary.hpp"
#include "geodmalib/featureextraction/spectralFeatures.hpp"
#include "geodmalib/featureextraction/spatialFeatures.hpp"
#include "geodma_tvplugin/featureExtractionController.hpp"
#include "geodma_tvplugin/spectralAndSpatialFeaturesController.hpp"

gdma::tvplg::SpectralAndSpatialFeaturesController::SpectralAndSpatialFeaturesController() {
    m_computeSpectral = true;
    m_computeSpatial = true;
}


gdma::tvplg::SpectralAndSpatialFeaturesController::~SpectralAndSpatialFeaturesController() {
}

void gdma::tvplg::SpectralAndSpatialFeaturesController::getRasterProperties(std::size_t &n_bands) {
    std::unique_ptr<te::rst::Raster> raster(te::map::GetRaster(m_layerData.get()));
    n_bands = raster->getNumberOfBands();
}

void gdma::tvplg::SpectralAndSpatialFeaturesController::setInputBands(
    std::vector<unsigned int> bands) {
    m_inputBands = bands;
}

void gdma::tvplg::SpectralAndSpatialFeaturesController::setComputeSpectral(bool compute) {
    m_computeSpectral = compute;
}

void gdma::tvplg::SpectralAndSpatialFeaturesController::setComputeSpatial(bool compute) {
    m_computeSpatial = compute;
}

bool gdma::tvplg::SpectralAndSpatialFeaturesController::computeFeatures() {
    std::map<std::string, te::gm::Geometry*> gm_container;
    std::vector<unsigned int> bands;
    std::map<std::string, gdma::fte::FeaturesSummary> attr_container;

    if (!gdma::dm::getGeometriesFromDataSource(m_layerTarget->getDataSourceId(), gm_container))
        return false;

    if (m_computeSpectral) {
        std::unique_ptr<te::rst::Raster> inputRaster(te::map::GetRaster(m_layerData.get()));

        if (!gdma::fte::SpectralFeatures::getFeatures(inputRaster.get(), gm_container, m_inputBands,
            attr_container, true, m_noDataValue))
            return false;
    }

    if (m_computeSpatial) {
        if (!gdma::fte::SpatialFeatures::getFeatures(gm_container, attr_container, true, m_noDataValue))
            return false;
    }

    if (m_saveSHP) {
        // @TODO(Raian): Mudar os saves para usar LAYER, assim como os loads
        if (!gdma::dm::saveToOGRDataSource(m_outputDataSet.string(),
            m_layerTarget->getDataSetName(), m_layerTarget->getDataSourceId(), attr_container))
            return false;
    }

    if (m_saveCSV) {
        if (!gdma::dm::saveAttributesToCSV(attr_container, m_layerTarget->getDataSourceId(),
            m_layerTarget->getDataSetName(), m_outputCSV))
            return false;
    }

    te::common::FreeContents(gm_container);

    createOutputLayer();

    return true;
}
