/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/classification/classDescription.hpp

  \brief This file contains the ClassDescription class definition.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

#ifndef GEODMA_LIB_CLFC_INTERNAL_CLASSDESCRIPTION_HPP_
#define GEODMA_LIB_CLFC_INTERNAL_CLASSDESCRIPTION_HPP_

// TerraLib Includes
#include <terralib/dataaccess/dataset/DataSetType.h>
#include <terralib/dataaccess/dataset/ObjectId.h>
#include <terralib/dataaccess/dataset/ObjectIdSet.h>
#include <terralib/geometry.h>

// STL Includes
#include <map>
#include <set>
#include <string>

// GeoDMA Includes
#include "geodmalib/classification/config.hpp"

namespace gdma{
namespace cl {
class GDMACLEXPORT ClassDescription {
 public:
    ClassDescription();

    /*!
      \brief Default Constructor
      \param[in] classLabel The label associated with this class.
      \param[in] classColor The hexa color associated with this class. Default value is white
                            (#ffffff).
    */
    explicit ClassDescription(std::string classLabel, const te::da::DataSetType* schema,
        std::string classColor = "#ffffff");

    ClassDescription(const ClassDescription& other);

    /*!
      \brief Default Destructor;
    */
    ~ClassDescription();

    /*!
      \brief Assignment operator.
    */
    ClassDescription& operator=(const ClassDescription& rhs);

    /*!
      \brief Set the label associated with the current class.
      \param[in] newLabel The label associated with the current class.
    */
    void setLabel(std::string newLabel);

    /*!
      \brief Retrieve the label of the current class.
      \return The label associated with this class.
    */
    std::string getLabel();

    /*!
      \brief Set the color associated with the current class.
      \param[in] newColor The new color associated with the current class.
    */
    void setColor(std::string newColor);

    /*!
      \brief Retrieve the color associated with the current class.
      \return The label associated with this class.
    */
    std::string getColor();

    /*!
      \brief Retrieve the number of metadata fields associated with the current class
      \return The size of the container \ref m_metadata.
    */
    int getMetadataSize();

    /*!
      \brief Add a new metadata.
      \param[in] name The name (id) of the metadata being inserted.
      \param[in] description The value or description of the metadata being inserted.
      \note: The \ref name must be UNIQUE. If you try to insert a new metadata with a duplicated
             name, the \ref description will overwrite the existing description, i. e., the
             existing metadata will be UPDATED.
    */
    void addMetadata(std::string name, std::string description);

    /*!
      \brief Add a new metadata.
      \param[in] name The name (id) of the metadata being inserted.
      \param[in] description The value or description of the metadata being inserted.
    */
    bool removeMetadata(std::string name);

    /*!
      \brief Retrieve the value (description) of a metadata given it name (id).
      \param[in] name The name (id) of the metadata to be retrieved.
    */
    std::string getMetadataValue(std::string name);

    /*!
      \brief Retrieve a reference for the whole metadata container.
      \return A map<key, value> containing all the metadata.
    */
    const std::map<std::string, std::string>& getAllMetadata();

    /*!
      \brief Add a new sample.
      \param[in] oid The object Id (\ref te::da::ObjectId) of the sample to be added.
    */
    void addSample(te::da::ObjectId* oid);

    /*!
      \brief Remove a sample from the set.
      \param[in] oid The object Id (\ref te::da::ObjectId) of the sample to be removed.
    */
    void removeSample(te::da::ObjectId* oid);

    /*!
      \brief Get a sample from the set given its value as string.
      \param[in] id The object Id value in the string form
                 (\ref te::da::ObjectId::getValueAsString()) of the sample.
      \note When calling this method it WILL NOT take the ownership of the pointers. This means
            that the caller MUST NOT delete the pointers or it will lead to an inconsistent state.
    */
    te::da::ObjectId* getSample(std::string id);

    /*!
      \brief Remove a sample from the set given its value as string.
      \param[in] id The object Id value in the string form
                 (\ref te::da::ObjectId::getValueAsString()) of the sample.
    */
    void removeSample(std::string id);

    /*!
      \brief Verify if the sample set contains a given sample.
      \param[in] oid The object Id (\ref te::da::ObjectId) of the sample.
    */
    bool containsId(te::da::ObjectId* oid);

    /*!
      \brief Verify if the sample set contains a given sample given its value as string.
      \param[in] id The object Id value in the string form
                 (\ref te::da::ObjectId::getValueAsString()) of the sample.
    */
    bool containsId(std::string id);

    /*!
      \brief Retrieves the total number of samples.
    */
    int getNumberOfSamples();

    /*!
      \brief Gets the ObjectIdSet corresponding to the calss samples.
    */
    const te::da::ObjectIdSet* getSampleSet();

    /*!
      \brief Gets the ids of the samples as string.
    */
    const std::set<std::string>& getSampleSetAsString();

 private:
    std::string m_label;  //!< Class label.
    std::string m_color;  //!< The hexa color associated with this Class.
    std::map<std::string, std::string> m_metadata;  //!< Metadata description for the class.
    te::da::ObjectIdSet* m_sampleSet;  //!< The map of acquired Samples.
    std::set<std::string> m_sampleSetString;  //!< The sample set ids as strings.
};
}  // namespace cl
}  // namespace gdma

#endif  // GEODMA_LIB_CLFC_INTERNAL_CLASSDESCRIPTION_HPP_
