/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/classification/sampleSet.hpp

  \brief This file contains the SampleSet class definition.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

#ifndef GEODMA_LIB_CLFC_INTERNAL_CLASSDESCRIPTIONSET_HPP_
#define GEODMA_LIB_CLFC_INTERNAL_CLASSDESCRIPTIONSET_HPP_

// STL Includes
#include <map>
#include <memory>
#include <string>

// GeoDMA Includes
#include "geodmalib/classification/classDescription.hpp"
#include "geodmalib/classification/config.hpp"

namespace gdma{
namespace cl {
typedef std::map<std::string, ClassDescription> ClassDescriptionMap;
class GDMACLEXPORT ClassDescriptionSet {
 public:
    /*!
      \brief Default Constructor
    */
    ClassDescriptionSet();

    /*!
      \brief Copy Constructor
    */
    ClassDescriptionSet(const ClassDescriptionSet& other);

    /*!
      \brief Assignment Operator.
    */
    ClassDescriptionSet& operator=(const ClassDescriptionSet& rhs);

    /*!
      \brief Default Destructor;
    */
    virtual ~ClassDescriptionSet();

    /*!
      \brief Add a class description in the structure.
      \param[in] clazz A reference to the class structure being added.
      \sa gdma::cl::ClassDescription
      \return Return true if the class was added, false if the label already exists in the
              structure.
    */
    bool addClass(ClassDescription& clazz);

    /*!
      \brief Remove a class description in the structure.
      \param[in] label The label of the class to remove.
      \sa gdma::cl::ClassDescription
      \return Return true if the class was corectly removed, false if the label could not be found
              in the structure.
    */
    bool removeClass(std::string label);

    /*!
      \brief Gets a class description in the structure.
      \param[in] label The label of the class to retrieved.
      \sa gdma::cl::ClassDescription
      \return Returns a pointer to a given class description.
      \note When calling this method the caller WILL NOT take the ownership of the pointers.
            This means that the caller MUST NOT delete the pointers or it will lead to an
            inconsistent state.
    */
    ClassDescription* getClass(std::string label);

    /*!
      \brief Set a label of a class in the structure.
      \param[in] clazz A reference to the class structure being seted.
      \sa gdma::cl::ClassDescription
      \param[in] newLabel A string containing the new label of the class.
      \return Return true if the the opperation was successful performed, false if the class does
              not exist in the structure.
      \note The object passed as parameter will be destructed and a copy of it will be stored in
            the position of the new label in the map. Thus, if you need to manipulate it after
            changing the label, you will need to retrieve the object again through
            \ref getClass.
    */
    bool setClassLabel(ClassDescription& clazz, std::string newLabel);

    /*!
      \brief Gets the total number of classes in the structure.
    */
    int getNumberOfClasses();

    /*
      \brief Get the number of samples of the class passed as parameter.
      \param[in] classLabel The label of the class.
      \return The number of samples contained in the class \ref classLabel.
    */
    int getNumberOfSamples(std::string classLabel);

    /*!
      \brief The begin of the iterator of the classes in \ref m_classSet container.
      \return An iterator to the objects in the \ref m_classSet container.
    */
    ClassDescriptionMap::iterator begin();

    /*!
      \brief The end of the iterator of the classes in \ref m_classSet container.
      \return An iterator to the objects in the \ref m_classSet container.
    */
    ClassDescriptionMap::iterator end();

    /*!
      \brief Clears the classes set.
    */
    void clear();

    /*!
      \brief Gets the total number of samples, taking into account all the classes in the container.
      \return The total number of samples.
    */
    int getTotalNumberOfSamples();

    /*!
      \brief Save the samples in the disk data source.
      \return Returns true if the operation was performed successfully, false otherwise.
    */
    bool saveSamplesInDataSource(std::string dataSourceId, std::string columnName);

    /*!
      \brief Load the samples from a disk data source.
    */
    bool loadSamplesFromDataSource(std::string dataSourceId, std::string columnName);

 private:
    ClassDescriptionMap m_classSet;
};
}  // namespace cl
}  // namespace gdma

#endif  // GEODMA_LIB_CLFC_INTERNAL_CLASSDESCRIPTIONSET_HPP_
