/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/classification/c50DecisionTree.hpp

  \brief This file contains the C50DecisionTree class definition.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// C50 Includes
extern "C" {
    // @TODO(Raian): Fix this in the library. Find a way to be equal in linux (installed) and
    // windows (inside the dev environment). Move another files to other directory, leaving only
    // this file visible for the library user.
#include <c50lib/c50_LibEntry.h>
}

// TerraLib Includes
#include <terralib/common/progress/TaskProgress.h>
#include <terralib/core/translator/Translator.h>
#include <terralib/dataaccess/dataset/DataSet.h>
#include <terralib/dataaccess/utils/Utils.h>

// Boost Filesystem
#include <boost/filesystem.hpp>

// STL Includes
#include <chrono>
#include <map>
#include <set>
#include <string>
#include <strstream>
#include <vector>

// GeoDMA Includes
#include "BuildConfig.hpp"
#include "geodmalib/common/fileSystem.hpp"
#include "geodmalib/common/messageMediator.hpp"
#include "geodmalib/datamanager/dataManager.hpp"
#include "geodmalib/classification/decision_trees/c50DecisionTree.hpp"
#include "geodmalib/classification/decision_trees/c50Parameters.hpp"

gdma::cl::C50DecisionTree::C50DecisionTree() : m_parameters() {
    m_namesFileContents.str(std::string());
    m_dataFileContents.str(std::string());
    m_costsFileContents.str(std::string());

    m_inputDataSet.reset();
    m_inputDSType.reset();
    m_inputDataSourceId = "";
    m_trainFeature = "trainClass";
    m_loadedFiles = false;
    m_noDataValue = -1000;
}

gdma::cl::C50DecisionTree::C50DecisionTree(const te::da::DataSetPtr& inputDataSet,
    const te::da::DataSetTypePtr inputDSType, const std::vector<std::string>& featuresToUse,
    gdma::cl::C50Parameters parameters, std::string trainFeature) {
    // @TODO(Raian): Review this contructor. Is it really necessary? Or it is better to make only
    // a default contructor?
    m_namesFileContents.str(std::string());
    m_dataFileContents.str(std::string());
    m_costsFileContents.str(std::string());

    m_inputDataSet = inputDataSet;
    m_inputDSType = inputDSType;
    m_trainFeature = trainFeature;
    m_parameters = parameters;
    m_featuresToUse = featuresToUse;

    std::vector<std::string> idNames;
    te::da::GetOIDPropertyNames(m_inputDSType.get(), idNames);
    m_idFeature = idNames[0];
    m_loadedFiles = false;
    m_noDataValue = -1000;
}

gdma::cl::C50DecisionTree::~C50DecisionTree() {
    m_namesFileContents.clear();
    m_dataFileContents.clear();
    m_costsFileContents.clear();

    m_inputDataSet.reset();
    m_inputDSType.reset();

    m_featuresValues.clear();
    m_classValues.clear();
    m_trainFeature.clear();
    m_idFeature.clear();
    m_outputFiles.clear();

    m_parameters.reset();
}

// @TODO(Raian): Refactor this method and class to use only DataSourceId. Take dataSet and dsType
// from it. Is it necessary to keep dataset and dstype? Put here the DataSet Name as parameter? It
// can be taken from the layer.
void gdma::cl::C50DecisionTree::setInputDataSet(const std::string inputDataSourceId,
    const std::string inputDataSetName) {
    m_inputDataSourceId = inputDataSourceId;

    te::da::DataSourcePtr dataSource = te::da::GetDataSource(inputDataSourceId, true);
    m_inputDataSet = dataSource->getDataSet(inputDataSetName);
    m_inputDSType = dataSource->getDataSetType(inputDataSetName);

    std::vector<std::string> idNames;
    te::da::GetOIDPropertyNames(m_inputDSType.get(), idNames);
    m_idFeature = idNames[0];
}

void gdma::cl::C50DecisionTree::setFeaturesToUse(const std::vector<std::string>& featuresToUse) {
    m_featuresToUse = featuresToUse;
}

std::vector<std::string> gdma::cl::C50DecisionTree::getFeaturesToUse() const {
    return m_featuresToUse;
}

void gdma::cl::C50DecisionTree::setParameters(const gdma::cl::C50Parameters& parameters) {
    m_parameters = parameters;
}

gdma::cl::C50Parameters& gdma::cl::C50DecisionTree::getParameters() {
    return m_parameters;
}

void gdma::cl::C50DecisionTree::setTrainFeature(const std::string featureName) {
    m_trainFeature = featureName;
}

std::string gdma::cl::C50DecisionTree::getTrainFeature() const {
    return m_trainFeature;
}

void gdma::cl::C50DecisionTree::setOutputClassFeatureName(const std::string classFeatName) {
    m_outputClassFeature = classFeatName;
}

std::string gdma::cl::C50DecisionTree::getOutputClassFeatureName() const {
    return m_outputClassFeature;
}

std::string gdma::cl::C50DecisionTree::getOutputTextFile() {
    return m_outputFiles["outputText"];
}

std::string gdma::cl::C50DecisionTree::getOutputTreeFile() {
    return m_outputFiles["tree"];
}

std::string gdma::cl::C50DecisionTree::getOutputRulesFile() {
    return m_outputFiles["rules"];
}

void gdma::cl::C50DecisionTree::setOutputFiles(std::map<std::string, std::string> newFiles) {
    m_outputFiles.clear();
    m_outputFiles = newFiles;
}

bool gdma::cl::C50DecisionTree::loadExternalFiles(std::string pathToFiles) {
    boost::filesystem::path pathFile(pathToFiles);
    pathToFiles = pathFile.replace_extension(".tree").string();
    std::string fileContents;

    if (gdma::common::loadTextFileToString(pathToFiles, fileContents)) {
        m_outputFiles["tree"] = fileContents.c_str();
        m_outputFiles["rules"] = "";
    } else {
        pathToFiles = pathFile.replace_extension(".rules").string();
        fileContents.clear();

        if (gdma::common::loadTextFileToString(pathToFiles, fileContents)) {
            m_outputFiles["rules"] = fileContents.c_str();
            m_outputFiles["tree"] = "";
        } else {
            gdma::common::MessageMediator::getInstance().sendErrorMsg(
                TE_TR("Problem when loading .tree or .rules file."));
            return false;
        }
    }

    pathToFiles = pathFile.replace_extension(".names").string();
    fileContents.clear();

    if (gdma::common::loadTextFileToString(pathToFiles, fileContents)) {
        m_namesFileContents.clear();
        m_namesFileContents.str(std::string());
        m_namesFileContents << fileContents.c_str();
    } else {
        std::string msg = "Problem when loading \"" + pathToFiles + "\" file.";
        gdma::common::MessageMediator::getInstance().sendErrorMsg(TE_TR(msg));
        return false;
    }

    pathToFiles = pathFile.replace_extension(".txt").string();
    fileContents.clear();

    if (gdma::common::loadTextFileToString(pathToFiles, fileContents))
        m_outputFiles["outputText"] = fileContents.c_str();

    parseNamesStream();

    m_loadedFiles = true;

    return true;
}

void gdma::cl::C50DecisionTree::saveOutputFiles(boost::filesystem::path pathOutputFiles) {
    pathOutputFiles = pathOutputFiles.replace_extension(".txt");

    if (m_outputFiles["outputText"].compare("") != 0)
        gdma::common::saveStringToTextFile(pathOutputFiles, m_outputFiles["outputText"]);

    pathOutputFiles = pathOutputFiles.replace_extension(".names");
    if (m_namesFileContents.str().compare("") != 0)
        gdma::common::saveStringToTextFile(pathOutputFiles, m_namesFileContents.str());

    pathOutputFiles = pathOutputFiles.replace_extension(".tree");
    if (m_outputFiles["tree"].compare("") != 0)
        gdma::common::saveStringToTextFile(pathOutputFiles, m_outputFiles["tree"]);

    pathOutputFiles = pathOutputFiles.replace_extension(".rules");
    if (m_outputFiles["rules"].compare("") != 0)
        gdma::common::saveStringToTextFile(pathOutputFiles, m_outputFiles["rules"]);
}

void gdma::cl::C50DecisionTree::train() {
    generateDataFile(m_featuresToUse);
    generateNamesFile(m_featuresToUse);
    generateCostsFile();

    te::common::TaskProgress task(TE_TR("Training C5.0 Decision Tree"));
    // task.setTotalSteps(0);

    char* namesFile = m_namesFileContents.str().empty()
        ? 0 : strdup(m_namesFileContents.str().c_str()); //NOLINT
    char* dataFile = m_dataFileContents.str().empty()
        ? 0 : strdup(m_dataFileContents.str().c_str()); //NOLINT
    char* costsFile = m_costsFileContents.str().empty()
        ? 0 : strdup(m_costsFileContents.str().c_str()); //NOLINT

    char* tree = 0;
    char* rulesFile = 0;
    char* output = 0;

    m_parameters.m_seed = std::time(0) & 07777;

    int *intSubset = reinterpret_cast<int*>(malloc(sizeof(int)));
    intSubset[0] = static_cast<int>(m_parameters.m_subset);
    int *rules_in = reinterpret_cast<int*>(malloc(sizeof(int)));
    rules_in[0] = static_cast<int>(m_parameters.m_rules);
    int *winnow = reinterpret_cast<int*>(malloc(sizeof(int)));
    winnow[0] = static_cast<int>(m_parameters.m_winnow);
    int *noGlobalPruning = reinterpret_cast<int*>(malloc(sizeof(int)));
    noGlobalPruning[0] = static_cast<int>(m_parameters.m_noGlobalPruning);
    int *fuzzyThreshold = reinterpret_cast<int*>(malloc(sizeof(int)));
    fuzzyThreshold[0] = static_cast<int>(m_parameters.m_fuzzyThreshold);
    int *earlyStopping = reinterpret_cast<int*>(malloc(sizeof(int)));
    earlyStopping[0] = static_cast<int>(m_parameters.m_earlyStopping);
    int *utility = reinterpret_cast<int*>(malloc(sizeof(int)));
    utility[0] = m_parameters.m_utility;
    int *trials = reinterpret_cast<int*>(malloc(sizeof(int)));
    trials[0] = m_parameters.m_trials;
    double *trainSamples = reinterpret_cast<double*>(malloc(sizeof(double)));
    trainSamples[0] = m_parameters.m_trainSamples;
    double *confidenceFactor = reinterpret_cast<double*>(malloc(sizeof(double)));
    confidenceFactor[0] = m_parameters.m_confidenceFactor;
    int *seed = reinterpret_cast<int*>(malloc(sizeof(int)));
    seed[0] = m_parameters.m_seed;
    int *minCases = reinterpret_cast<int*>(malloc(sizeof(int)));
    minCases[0] = m_parameters.m_minCases;

    c50(&namesFile, &dataFile, &costsFile, intSubset, rules_in, utility,
        trials, winnow, trainSamples, seed, noGlobalPruning,
        confidenceFactor, minCases, fuzzyThreshold, //NOLINT
        earlyStopping, &tree, &rulesFile, &output); //NOLINT

    m_outputFiles.clear();

    if( tree ) {
        m_outputFiles["tree"] = std::string(tree);
        free( tree );
    }

    if( rulesFile ) {
        m_outputFiles["rules"] = std::string(rulesFile);
        free( rulesFile );
    }

    if( output ) {
        m_outputFiles["outputText"] = std::string(output);
        free( output );
    }

    if (namesFile) free(namesFile);
    if (dataFile) free(dataFile);
    if (costsFile) free(costsFile);
    if (intSubset) free(intSubset);
    if (rules_in) free(rules_in);
    if (winnow) free(winnow);
    if (noGlobalPruning) free(noGlobalPruning);
    if (fuzzyThreshold) free(fuzzyThreshold);
    if (earlyStopping) free(earlyStopping);
    if (utility) free(utility);
    if (trials) free(trials);
    if (trainSamples) free(trainSamples);
    if (confidenceFactor) free(confidenceFactor);
    if (seed) free(seed);
    if (minCases) free(minCases);
}

void gdma::cl::C50DecisionTree::predict() {
    if (m_loadedFiles) {
        generateDataFile(m_featuresToUse);
        generateCostsFile();
    } else {
        if ((m_outputFiles["tree"].empty() || m_outputFiles["tree"].compare("") == 0) &&
            (m_outputFiles["rules"].empty() || m_outputFiles["rules"].compare("") == 0))
            throw te::common::Exception(TE_TR("Classification model not found!"));
    }

    if (m_namesFileContents.str().compare("") == 0) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(TE_TR("Names file not found!"));
        return;
    }

    te::common::TaskProgress task(TE_TR("C5.0 - Classifying"));
    // task.setTotalSteps(0);

    char* namesFile = m_namesFileContents.str().empty()
        ? 0 : strdup( m_namesFileContents.str().c_str() ); //NOLINT
    char* dataFile = m_dataFileContents.str().empty()
        ? 0 : strdup( m_dataFileContents.str().c_str() ); //NOLINT
    char* costsFile = m_costsFileContents.str().empty()
        ? 0 : strdup( m_costsFileContents.str().c_str() ); //NOLINT

    char* output = 0;
    char* tree = m_outputFiles["tree"].empty() ? strdup("") : strdup(m_outputFiles["tree"].c_str());
    char* rules = m_outputFiles["rules"].empty() ? strdup("") : strdup(m_outputFiles["rules"].c_str());
    int *predv = reinterpret_cast<int*>(malloc( sizeof( int ) * m_inputDataSet->size() ));
    double *confidencev = reinterpret_cast<double*>(malloc( sizeof( double ) * m_classValues.size() * m_inputDataSet->size() ));

    int *trials = reinterpret_cast<int*>(malloc(sizeof(int)));
    trials[ 0 ] = m_parameters.m_trials;

    predictions(&dataFile, &namesFile, &tree, &rules, &costsFile,
        predv, confidencev, trials, &output);

    fillClassSummary(predv, m_outputClassFeature);

    if( trials ) free( trials );
    if( confidencev ) free( confidencev );
    if( predv ) free( predv );
    if( tree ) free( tree );
    if( rules ) free( rules );
    if( namesFile ) free( namesFile );
    if( dataFile ) free( dataFile );
    if( costsFile ) free( costsFile );
    if( output ) free( output );
}

// @TODO(Raian): Review this method
bool gdma::cl::C50DecisionTree::saveClassificationResults(const std::string outputDataSourcePath,
    const std::string pathToCSV) {
    try {
        te::da::DataSourcePtr inputDS = te::da::GetDataSource(m_inputDataSourceId);
        std::string inputSHPPath = inputDS->getConnectionInfo().host()
            + inputDS->getConnectionInfo().path();
        if (outputDataSourcePath.compare(inputSHPPath) == 0) {
            gdma::dm::updateDataSource(outputDataSourcePath, m_predictions);
        } else {
            if (!boost::filesystem::exists(outputDataSourcePath))
                gdma::dm::saveToOGRDataSource(outputDataSourcePath, m_inputDSType->getName(),
                    m_inputDataSourceId, m_predictions);
            else
                throw te::common::Exception(TE_TR("File already exists. You can not overwrite it!"));
        }

        // @TODO(Raian): When try to use the method saveDataSourceToCSV, it can not find
        // the datasource. It probably is because of a lack of pattern whe dealing with data
        // source ids. Refatorar metodos de salvamento para que funcionem de maneira uniforme.
        if (pathToCSV.compare("") != 0)
            gdma::dm::saveAttributesToCSV(m_predictions, m_inputDataSourceId,
                m_inputDSType->getName(), pathToCSV);
    } catch (te::common::Exception& e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(TE_TR(e.what()));
        return false;
    } catch (std::exception& e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(TE_TR(e.what()));
        return false;
    } catch (...) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(TE_TR("Unknown Exception!"));
        return false;
    }
    return true;
}

void gdma::cl::C50DecisionTree::setNoDataValue(int noDataValue) {
    m_noDataValue = noDataValue;
}

int gdma::cl::C50DecisionTree::getNoDataValue() {
    return m_noDataValue;
}

// ---------------------------------------
// Private Members
// ---------------------------------------

void gdma::cl::C50DecisionTree::generateNamesFile(const std::vector<std::string>& featuresToUse) {
    m_namesFileContents.str(std::string());

    auto now = std::chrono::system_clock::now();
    auto now_c = std::chrono::system_clock::to_time_t(now);

    m_namesFileContents << "| Generated using GeoDMA " << GEODMA_VERSION_STRING << std::endl;
    m_namesFileContents << "| on " << std::put_time(std::localtime(&now_c), "%c") << std::endl;
    m_namesFileContents << "|" << std::endl << std::endl;

    m_namesFileContents << m_trainFeature << "." << std::endl;

    m_namesFileContents << std::endl << m_trainFeature << ": ";

    const char* separator = "";
    for (auto &classVal : m_classValues) {
        m_namesFileContents << separator << classVal;
        separator = ",";
    }

    m_namesFileContents << "." << std::endl;

    for (auto &featureName : featuresToUse) {
        m_namesFileContents << featureName << ": ";

        separator = "";
        for (auto &featValue : m_featuresValues[featureName]) {
            m_namesFileContents << separator << featValue;
            separator = ",";
        }

        m_namesFileContents << "." << std::endl;
    }
}

void gdma::cl::C50DecisionTree::generateDataFile(const std::vector<std::string>& featuresToUse) {
    te::common::TaskProgress task(TE_TR("Preparing Data"));
    task.setTotalSteps(m_inputDataSet->size() * featuresToUse.size());
    task.useTimer(true);

    m_dataFileContents.str(std::string());

    bool training = !m_loadedFiles;

    std::set<std::string> classValues;
    std::size_t classPropertyPos;
    if (training)
        classPropertyPos = te::da::GetPropertyPos(m_inputDataSet.get(), m_trainFeature);

    m_inputDataSet->moveBeforeFirst();
    while (m_inputDataSet->moveNext()) {
        if (training) {
            std::string classValue = m_inputDataSet->getValue(classPropertyPos)->toString();
            switch (m_inputDataSet->getPropertyDataType(classPropertyPos)) {
            case te::dt::STRING_TYPE:
                if (classValue.compare("") != 0 && classValue.compare(" ") != 0)
                    classValues.insert(classValue);
                break;
            case te::dt::BOOLEAN_TYPE:
                classValues.insert("true, false");
                break;
            default:
                classValues.insert("continuous");
                break;
            }
            if (classValue.compare("") != 0 && classValue.compare(" ") != 0)
                m_dataFileContents << classValue << ",";
            else
                m_dataFileContents << "?" << ",";
        } else {
            m_dataFileContents << "?" << ",";
        }

        const char* separator = "";
        for (auto &featureName : featuresToUse) {
            std::string stringFeatValue = m_inputDataSet->getValue(featureName)->toString();
            std::size_t featurePos = m_inputDSType->getPropertyPosition(featureName);
            switch (m_inputDataSet->getPropertyDataType(featurePos)) {
            case te::dt::STRING_TYPE:
                m_featuresValues[featureName].insert(stringFeatValue);
                break;
            case te::dt::BOOLEAN_TYPE:
                m_featuresValues[featureName].insert("true, false");
                break;
            default:
                m_featuresValues[featureName].insert("continuous");

                double doubleVal = m_inputDataSet->getDouble(featureName);
                if (std::isinf(doubleVal) || std::isnan(doubleVal) || static_cast<int>(doubleVal) == m_noDataValue)
                    stringFeatValue = "N/A";
                break;
            }
            m_dataFileContents << separator << stringFeatValue;
            separator = ",";
            task.pulse();
        }

        m_dataFileContents << std::endl;
    }

    if (training) {
        m_classValues.clear();
        for (auto &classVal : classValues)
            m_classValues.push_back(classVal);
    }
}

void gdma::cl::C50DecisionTree::generateCostsFile() {
    // @TODO(Raian): Implement this method
    m_costsFileContents.str(std::string());
    m_costsFileContents << "" << std::endl;
}

void gdma::cl::C50DecisionTree::fillClassSummary(int* predictions, std::string classColumn) {
    std::size_t idPos = te::da::GetPropertyPos(m_inputDSType.get(), m_idFeature);
    m_inputDataSet->moveBeforeFirst();

    m_predictions.clear();
    std::size_t count = 0;
    while (m_inputDataSet->moveNext()) {
        std::string id = m_inputDataSet->getValue(idPos)->toString();
        gdma::fte::FeaturesSummary fteClass;

        fteClass.addFeature(classColumn, classColumn, m_classValues[predictions[count] - 1]);
        m_predictions[id] = fteClass;
        count++;
    }
}

bool gdma::cl::C50DecisionTree::parseNamesStream() {
    te::common::TaskProgress task(TE_TR("Reading Names File"));
    task.setTotalSteps(0);
    task.useTimer(true);

    std::string line;

    // Read file reader, consuming the comments.
    std::getline(m_namesFileContents, line);
    while (line.compare("") == 0 || line.at(0) == '|') {
        std::getline(m_namesFileContents, line);
        task.pulse();
    }

    // Consume the class name
    std::stringstream buffer(line);
    std::getline(buffer, m_trainFeature, '.');

    int count;
    m_featuresToUse.clear();
    m_classValues.clear();
    while (std::getline(m_namesFileContents, line)) {
        task.pulse();
        if (line.compare("") != 0 && line.at(0) != '|') {
            std::string featName;
            std::string token;
            buffer.clear();
            buffer.str(line);
            // Consume the feature name
            std::getline(buffer, featName, ':');

            std::vector<std::string> values;

            // Consume the feature values
            while (std::getline(buffer, token, ',')) {
                std::string::iterator end_pos = std::remove(token.begin(), token.end(), ' ');
                token.erase(end_pos, token.end());
                end_pos = std::remove(token.begin(), token.end(), '.');
                token.erase(end_pos, token.end());

                values.push_back(token);
                task.pulse();
            }

            // Verify if the feature is class, ID or a simple feature.
            if (featName.compare(m_trainFeature.c_str()) == 0) {
                for (auto &val : values)
                    m_classValues.push_back(val);
            } else if (featName.compare("ID") == 0) {
                m_idFeature = values[0];
            } else {
                if (m_inputDSType->getProperty(featName)) {
                    m_featuresToUse.push_back(featName);
                } else {
                    std::string msg = "Column \"" + featName + "\" found in the \""
                        + ".names\" file does not exists in the dataset.";
                    throw te::common::Exception(TE_TR(msg));
                }
            }
        }
    }

    return true;
}
