/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/classification/c50DecisionTree.hpp

  \brief This file contains the C50DecisionTree class definition. 

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

#ifndef GEODMA_LIB_CLFC_INTERNAL_C50DECISIONTREE_HPP_
#define GEODMA_LIB_CLFC_INTERNAL_C50DECISIONTREE_HPP_

// TerraLib Includes
#include <terralib/dataaccess/dataset/DataSet.h>

// Boost Filesystem
#include <boost/filesystem.hpp>

// STL Includes
#include <map>
#include <set>
#include <string>
#include <strstream>
#include <vector>

// GeoDMA Includes
#include "geodmalib/classification/config.hpp"
#include "geodmalib/classification/decision_trees/c50Parameters.hpp"
#include "geodmalib/featureextraction/featuresSummary.hpp"

namespace gdma{
namespace cl {
class GDMACLEXPORT C50DecisionTree {
 public:
    /*!
      \brief Default constructor
    */
    C50DecisionTree();

    /*!
      \brief Regular constructor.
      \param[in] inputDataSet The input Data Set.
      \param[in] featuresToUse A vector of strings containing the names of the features to be used.
    */
     C50DecisionTree(const te::da::DataSetPtr& inputDataSet, const te::da::DataSetTypePtr inputDSType,
         const std::vector<std::string>& featuresToUse, gdma::cl::C50Parameters parameters,
         std::string trainFeature = "trainClass");

    /*!
      \brief Default Destructor.
    */
    ~C50DecisionTree();

    /*!
      \brief Set the input dataset that will be used for training or classification.
      \param[in] inputDataSourceId The ID of the datasource that contains the input dataset.
      \param[in] inputDataSetName The name of the inputDataset.
    */
    void setInputDataSet(const std::string inputDataSourceId, const std::string inputDataSetName);

    /*!
      \brief Set the features to be used for training or classification.
      \param[in] featuresToUse Vector of strings containing the names of the features to be used.
    */
    void setFeaturesToUse(const std::vector<std::string>& featuresToUse);

    /*!
      \brief Returns the features being used for training or classification.
      \return Vector of strings containing the names of the features being used.
    */
    std::vector<std::string> getFeaturesToUse() const;

    /*!
      \brief Sets the parameters of C5.0.
      \param[in] parameters A \ref gdma::cl::C50Parameters objetct containing the parameters for
                            the C5.0 algorithm.
    */
    void setParameters(const gdma::cl::C50Parameters& parameters);

    /*!
      \brief Gets the parameters of C5.0.
      \return A \ref gdma::cl::C50Parameters objetct containing the parameters for the C5.0 algorithm.
    */
    gdma::cl::C50Parameters& getParameters();

    /*!
      \brief Sets the name of the training features, i. e., the feature containing the names of the
             classes of interest.
      \param[in] featureName A string containing the name of the training feature.
    */
    void setTrainFeature(const std::string featureName);

    /*!
      \brief Gets the name of the training features, i. e., the feature containing the names of the
             classes of interest.
      \return A string containing the name of the training feature.
    */
    std::string getTrainFeature() const;

    /*!
      \brief Sets the name of the output feature, i. e., the name of the column where the output
             classification will be saved.
      \param[in] classFeatName A string containing the name of output feature where the
                               classification will be saved.
    */
    void setOutputClassFeatureName(const std::string classFeatName);

    /*!
      \brief Gets the name of the output feature, i. e., the name of the column where the output
             classification will be saved.
      \return A string containing the name of output feature where the classification will be saved.
    */
    std::string getOutputClassFeatureName() const;

    /*!
      \brief Gets the path for the output text file where the results of the training process will
             be saved.
      \return A string containing the path for the output file where the results of the training
              will be saved.
    */
    std::string getOutputTextFile();

    /*!
      \brief Gets the path for the output tree file where decision tree produced in the training
             process will be saved.
      \return A string containing the path for the output tree file where decision tree produced in
              the training process will be saved.
    */
    std::string getOutputTreeFile();

    /*!
      \brief Gets the path for the output rules file where decision tree produced in the training
             process will be saved in the form of rules set.
      \return A string containing path for the output rules file where decision tree produced in
              the training process will be saved in the form of rules set.
    */
    std::string getOutputRulesFile();

    /*!
      \brief Set the output files content.
      \param[in] newFiles A map <string, string> containing the output files contents in the
                          following structure: <"name", "file content">, where "name" must be one
                          of the following "outputText", "tree" or "rules".
    */
    void setOutputFiles(std::map<std::string, std::string> newFiles);

    /*!
      \brief Load the output files from an external source.
      \param[in] pathToFiles A string containing the path to the files to be loaded.
    */
    bool loadExternalFiles(std::string pathToFiles);

    /*!
      \brief Save the output files in an external source.
      \param[in] pathOutputFiles A boost::filesystem::path containing the path where the files will
                                 be saved.
    */
    void saveOutputFiles(boost::filesystem::path pathOutputFiles);

    /*!
      \brief Train the Decision Tree.
    */
    void train();

    /*!
      \brief Make a prediction with a generated Decision Tree.
    */
    void predict();

    /*!
      \brief Save the classification results in an output data source and/or in an output CSV.
      \param[in] outputDataSourceId A string containing the ID or path of the output data source.
      \param[in] pathToCSV A string containing the path to the output CSV.
    */
    bool saveClassificationResults(const std::string outputDataSourceId,
        const std::string pathToCSV = "");

    /*!
      \brief Set the "no data" value contained in the input data, i. e., the value to be ignored.
      \param[in] noDataValue The "no data" value.
    */
    void setNoDataValue(int noDataValue);

    /*!
      \brief Gets the "no data" value contained in the input data, i. e., the value being ignored.
    */
    int getNoDataValue();

 private:
    /*!
      \brief Generate the .names file required by the C5.0 Algorithm as a stream.
      \param[in] featuresToUse A vector of strings containing the names of the features in the dataset
                               that will be used for the training or classification process.
    */
    void generateNamesFile(const std::vector<std::string>& featuresToUse);

    /*!
      \brief Generate the .data file required by the C5.0 Algorithm as a stream.
      \param[in] featuresToUse A vector of strings containing the names of the features in the dataset
                               that will be used for the training or classification process.
    */
    void generateDataFile(const std::vector<std::string>& featuresToUse);

    /*!
      \brief Generate the .costs file required by the C5.0 Algorithm as a stream.
      \note This method is not implemented yet. It is currently generating an empty stream.
    */
    void generateCostsFile();

    /*!
      \brief Generate a summary with the predictions generated by the Decision Tree algortithm.
      \param[in] predictions A pointer for a vector of int containing the predictions generated by the
                             Decision tree.
      \param[in] className The name of the output column to be generated in the output datasource.
    */
    void fillClassSummary(int* predictions, std::string classColumn);

    /*!
      \brief Parse a loaded .names stream retrieving the information needed to classify the data.
      \return true if the parse finished successfully, false otherwise
    */
    bool parseNamesStream();

 protected:
    std::stringstream m_namesFileContents;  //!< Stream containing the .names file contents
    std::stringstream m_dataFileContents;  //!< Stream containing the .data file contents
    std::stringstream m_costsFileContents;  //!< Stream containing the .costs file contents
    te::da::DataSetPtr m_inputDataSet;  // @TODO(Raian): Change this. Put here the dataset and datasource names
    te::da::DataSetTypePtr m_inputDSType;
    std::string m_inputDataSourceId;

    std::map<std::string, std::set<std::string>> m_featuresValues;
    std::vector<std::string>  m_featuresToUse;
    std::vector<std::string> m_classValues;
    std::string m_trainFeature;
    std::string m_idFeature;
    std::string m_outputClassFeature;
    int m_noDataValue;

    gdma::cl::C50Parameters m_parameters;
    std::map<std::string, gdma::fte::FeaturesSummary> m_predictions;

    std::map<std::string, std::string> m_outputFiles;
    bool m_loadedFiles;
};
}  // namespace cl
}  // namespace gdma
#endif  // GEODMA_LIB_CLFC_INTERNAL_C50DECISIONTREE_HPP_
