/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/classification/vectorDatasetInputAdaptor.hpp

  \brief This file contains the VectorDatasetInputAdaptor class definition. It is an adaptor to
         access Datasets of geometries.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

#ifndef GEODMA_LIB_CLFC_VECTORDATASETINPUTADAPTOR_HPP_
#define GEODMA_LIB_CLFC_VECTORDATASETINPUTADAPTOR_HPP_

// TerraLib Includes
#include <terralib/classification/Adaptors.h>
#include <terralib/dataaccess/dataset/DataSet.h>

// GeoDMA Includes
#include "geodmalib/classification/config.hpp"

namespace gdma {
namespace cl {
template<typename DataType>

class VectorDatasetInputAdaptor : te::cl::InputAdaptor<DataType> {
 public:
    explicit VectorDatasetInputAdaptor(const te::da::DataSetPtr inputDataSet) {
        if (!inputDataSet->isEmpty())
            m_inputDataSet = inputDataSet;
        else
            throw te::common::Exception("Input Data Set is Empty!");
    }

    ~VectorDatasetInputAdaptor() {}

    void getFeature(const unsigned int& elementIndex,
        const unsigned int& featureIndex, DataType& featureValue) const override {
        if (m_inputDataSet->move(elementIndex)) {
            if (featureIndex < m_inputDataSet->getNumProperties()) {
                switch (m_inputDataSet->getPropertyDataType(featureIndex)) {
                case te::dt::INT16_TYPE:
                    featureValue = m_inputDataSet->getInt16(featureIndex);
                    break;
                case te::dt::INT32_TYPE:
                    featureValue = m_inputDataSet->getInt32(featureIndex);
                    break;
                case te::dt::INT64_TYPE:
                    featureValue = m_inputDataSet->getInt64(featureIndex);
                    break;
                case te::dt::DOUBLE_TYPE:
                    featureValue = m_inputDataSet->getDouble(featureIndex);
                    break;
                /*case te::dt::STRING_TYPE:
                    featureValue = m_inputDataSet->getString(featureIndex);
                    break;
                case te::dt::BOOLEAN_TYPE:
                    featureValue = m_inputDataSet->getBool(featureIndex);
                    break;*/
                default:
                    throw te::common::Exception("Unexpected feature type!");
                }
            } else {
                throw te::common::Exception("Feature index \"" + std::to_string(featureIndex) +
                    "\" does not exist in Data Set!");
            }
        } else {
            throw te::common::Exception("Element index \"" + std::to_string(elementIndex) +
                "\" does not exist in Data Set!");
        }
    }

    unsigned int getElementsCount() const override {
        return m_inputDataSet->size();
    }

    unsigned int getFeaturesCount() const override {
        return m_inputDataSet->getNumProperties();
    }

 protected:
     te::da::DataSetPtr m_inputDataSet;
};
}  // namespace cl
}  // namespace gdma

#endif  // GEODMA_LIB_CLFC_VECTORDATASETINPUTADAPTOR_HPP_
