/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/common/abstractMessageObserver.hpp

  \brief An abstract Observer to receive message and show to the user.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

#ifndef GEODMA_LIB_COMMON_ABSTRACTMESSAGEOBSERVER_HPP_
#define GEODMA_LIB_COMMON_ABSTRACTMESSAGEOBSERVER_HPP_

// STL includes
#include <string>

// GeoDMA includes
#include "geodmalib/common/config.hpp"

namespace gdma {
namespace common {
/*!
  \class AbstractMessageObserver

  \brief An abstract observer that defines the interface to Message Observers.

  This interface is used to implement classes that receive messages from the MessageMediator and
  display them to the user.

  \ingroup common

  \sa MessageMediator, CommandLineMsgObserver, LogMessageObserver
*/
class AbstractMessageObserver {
 public:
     /*! \brief Default constructor. */
    AbstractMessageObserver() { }

    /*! \brief Default destructor. */
    virtual ~AbstractMessageObserver() { }

    /*!
      \brief Display an information message.
      \param[in] message The information message to be displayed.
    */
    virtual void displayInformationMsg(std::string msg) = 0;

    /*!
      \brief Display a warning message.
      \param[in] message The warning message to be displayed.
    */
    virtual void displayWarningMsg(std::string msg) = 0;

    /*!
      \brief Display an error message.
      \param[in] message The error message to be displayed.
    */
    virtual void displayErrorMsg(std::string msg) = 0;

    /*!
      \brief Display a debug message.
      \param[in] message The debug message to be displayed.
    */
    virtual void displayDebugMsg(std::string msg) = 0;

    /*!
      \brief Display a trace message.
      \param[in] message The trace message to be displayed.
    */
    virtual void displayTraceMsg(std::string msg) = 0;

    /*!
      \brief Display a fatal message.
      \param[in] message The fatal message to be displayed.
    */
    virtual void displayFatalMsg(std::string msg) = 0;
};
}  // namespace common
}  // namespace gdma

#endif  // GEODMA_LIB_COMMON_ABSTRACTMESSAGEOBSERVER_HPP_

