/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/common/commandLineStatusBar.cpp

  \brief Command line status bar.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// Boost Includes
#include <boost/chrono.hpp>

// STL includes
#include <cmath>
#include <string>
#include <iostream>
#include <sstream>

// GeoDMA Includes
#include "geodmalib/common/commandLineProgBar.hpp"
#include "geodmalib/common/utils.hpp"

// Default constructor
gdma::common::CommandLineProgBar::CommandLineProgBar(const std::string& label,
    double total_steps, int barWidth) {
    this->m_progress = 0.0;
    this->m_posBar = 0;
    this->m_barWidth = barWidth - label.size();
    this->m_totalSteps = total_steps;
    this->m_label = label;

    this->m_startTime = boost::chrono::high_resolution_clock::now();
    this->m_currentTime = boost::chrono::duration_cast<boost::chrono::seconds>(
        boost::chrono::high_resolution_clock::now() - this->m_startTime);

    this->setBar();
    std::cout << this->m_label << " " << this->m_bar << std::flush;
}

// Copy Constructor
gdma::common::CommandLineProgBar::CommandLineProgBar(const CommandLineProgBar &rhs) {
    this->m_progress = rhs.m_progress;
    this->m_posBar = rhs.m_posBar;
    this->m_barWidth = rhs.m_barWidth;
    this->m_totalSteps = rhs.m_totalSteps;
    this->m_label = rhs.m_label;
    this->m_startTime = rhs.m_startTime;
    this->m_currentTime = rhs.m_currentTime;
    this->m_bar = rhs.m_bar;
}

gdma::common::CommandLineProgBar::~CommandLineProgBar() {
    std::cout << std::endl;
}

void gdma::common::CommandLineProgBar::pulse() {
    this->m_progress += 1.0 / this->m_totalSteps;

    int prevPosBar = this->m_posBar;
    this->m_posBar = this->m_barWidth * this->m_progress;

    boost::chrono::duration<int> time =
        boost::chrono::duration_cast<boost::chrono::seconds>(
        boost::chrono::high_resolution_clock::now() - this->m_startTime);

    if (this->m_posBar > prevPosBar || time > this->m_currentTime) {
        this->m_currentTime = time;
        this->setBar();

        std::cout << this->m_label << " " << this->m_bar;
        std::cout.flush();
    }
}

void gdma::common::CommandLineProgBar::setBar() {
    std::ostringstream barStr;
    barStr << "[";

    for (int i = 0; i < this->m_barWidth; ++i) {
        if (i < this->m_posBar)
            barStr << "=";
        else if (i == this->m_posBar)
            barStr << ">";
        else
            barStr << " ";
    }

    barStr << "] " << std::round(this->m_progress * 100.0) << " % | Time: "
        << gdma::common::formatTime(this->m_currentTime.count()) << "\r";

    this->m_bar = barStr.str();
}
