/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/common/fileSystem.cpp

  \brief This file contains some File system related functions.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// Boost includes
#include <boost/filesystem.hpp>

// STL Includes
#include <iostream>
#include <string>
#include <sstream>

// GeoDMA Includes
#include "BuildConfig.hpp"
#include "geodmalib/common/config.hpp"
#include "geodmalib/common/fileSystem.hpp"

std::string gdma::common::FindInGeoDMAPath(std::string path) {
    // 1st: look in the neighborhood of the executable
    boost::filesystem::path gdma_path = boost::filesystem::current_path();

    boost::filesystem::path eval_path = gdma_path / path;

    if (boost::filesystem::exists(eval_path))
        return eval_path.string();

    // Check for one path above
    gdma_path /= "..";
    eval_path = gdma_path / path;

    if (boost::filesystem::exists(eval_path))
        return eval_path.string();

    // 2rd: look for an environment variable defined by macro GEODMA_DIR_VAR_NAME
    const char* te_env = getenv(GEODMA_DIR_VAR_NAME);

    if (te_env != 0) {
        gdma_path = te_env;

        eval_path = gdma_path / path;

        if (boost::filesystem::exists(eval_path))
            return eval_path.string();
    }

    // 3th: look into install prefix-path
    gdma_path = GEODMA_INSTALL_PREFIX_PATH;

    eval_path = gdma_path / path;

    if (boost::filesystem::exists(eval_path))
        return eval_path.string();

    // 4nd: look into the development environment path
    gdma_path = GEODMA_DEVEL_PATH;

    eval_path = gdma_path / path;

    if (boost::filesystem::exists(eval_path))
        return eval_path.string();

    std::string errMsg = std::string("File '") + path + "' does not exists!";
    throw std::runtime_error(errMsg.c_str());
}

void gdma::common::SetEnvVariable(char* varName, char* value) {
#ifdef WIN32
        _putenv_s(varName, value);
#else
        setenv(varName, value, true);
#endif
}

bool gdma::common::loadTextFileToString(std::string pathToFile,
    std::string& fileContents) {
    if (boost::filesystem::exists(boost::filesystem::path(pathToFile))) {
        std::ifstream file;
        fileContents.clear();
        file.open(pathToFile, std::ios::in);
        std::stringstream buffer;
        buffer << file.rdbuf();
        fileContents = buffer.str();
        file.close();
        return true;
    }

    return false;
}

bool gdma::common::saveStringToTextFile(const boost::filesystem::path pathToFile,
    const std::string& fileContents) {
    std::ofstream outputFile;
    outputFile.open(pathToFile.string(), std::ios::out);

    outputFile << fileContents;

    outputFile.close();
    return true;
}
