/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/common/logLineMsgObserver.hpp

  \brief A concrete observer that save messages in a log file.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// STL includes
#include <string>

// GeoDMA Includes
#include "geodmalib/common/abstractMessageObserver.hpp"
#include "geodmalib/common/config.hpp"

namespace gdma {
namespace common {
/*!
  \class LogLineMsgObserver

  \brief A concrete observer that save Messages in a log file, using TerraLib Log structure.

  This class implements the interface \ref AbstractMessageObserver and is used to save messages
  received from \ref MessageMediator in the command line.

  \ingroup common

  \sa MessageMediator, AbstractMessageObserver, CommandLineMsgObserver
*/
class GDMACOMMONEXPORT LogMessageObserver : public AbstractMessageObserver {
 public:
    /*!
      \brief Constructor.
      \param[in] pathToLog Path to the log file.
    */
    explicit LogMessageObserver(std::string pathToLog);

    /*!
      \brief Constructor for specific configurable logs.
      \param[in] logName The name of the log
      \param[in] pathToLog Path to the log file.
      \param[in] format The format string of the logger. Format string must be formatted like
                 "[%TimeStamp%] <%Severity%>: %Message%". \sa 
    */
    LogMessageObserver(std::string logName, std::string pathToLog, std::string format);

    /*!
      \brief Constructor for configurable logs.
      \param[in] logName The name of the log
      \param[in] includeTimeStamp Boolean value that defines if the time stamp will be included in
                 log messages or not.
      \param[in] includeSeverity Boolean value that defines if the severity will be included in the
                 log messages or not.
    */
    LogMessageObserver(std::string logName, std::string pathToLog, bool includeTimeStamp = true,
        bool includeSeverity = true);

    /*! \brief Default destructor. */
    ~LogMessageObserver();

    /*!
      \brief Display an information message.
      \param[in] message The information message to be displayed.
    */
    void displayInformationMsg(std::string msg) override;

    /*!
      \brief Display a warning message.
      \param[in] message The warning message to be displayed.
    */
    void displayWarningMsg(std::string msg) override;

    /*!
      \brief Display an error message.
      \param[in] message The error message to be displayed.
    */
    void displayErrorMsg(std::string msg) override;

    /*!
      \brief Display a debug message.
      \param[in] message The debug message to be displayed.
    */
    void displayDebugMsg(std::string msg) override;

    /*!
      \brief Display a trace message.
      \param[in] message The trace message to be displayed.
    */
    void displayTraceMsg(std::string msg) override;

    /*!
      \brief Display a fatal message.
      \param[in] message The fatal message to be displayed.
    */
    void displayFatalMsg(std::string msg) override;

 public:
    std::string m_logName;
    std::string m_logFormat;
    std::string m_logPath;

    bool m_includeTimeStamp;
    bool m_includeSeverity;
};
}  // namespace common
}  // namespace gdma

