/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/common/messageMediator.cpp

  \brief A singleton class used to manage messages.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// STL includes
#include <map>
#include <string>

// GeoDMA Includes
#include "geodmalib/common/abstractMessageObserver.hpp"
#include "geodmalib/common/messageMediator.hpp"


gdma::common::MessageMediator::MessageMediator() : m_obsCounter(0) {
}


gdma::common::MessageMediator::~MessageMediator() {
    clearAll();
}

int gdma::common::MessageMediator::addObserver(AbstractMessageObserver* obs) {
    int id = generateObserverId();

    m_observers.insert(std::map<int, AbstractMessageObserver*>::value_type(id, obs));
    return id;
}

void gdma::common::MessageMediator::removeObserver(int obsId) {
    std::map<int, AbstractMessageObserver*>::iterator it = m_observers.find(obsId);

    if (it != m_observers.end())
        m_observers.erase(it);
}

int gdma::common::MessageMediator::getNumberOfObservers() {
    return m_observers.size();
}

void gdma::common::MessageMediator::clearAll() {
    m_observers.clear();
    // @TODO(Raian): Remove it
    m_obsCounter = 0;
}

void gdma::common::MessageMediator::sendInformationMsg(std::string message) {
    for (auto &elem : m_observers)
        elem.second->displayInformationMsg(message);
}

void gdma::common::MessageMediator::sendWarningMsg(std::string message) {
    for (auto &elem : m_observers)
        elem.second->displayWarningMsg(message);
}

void gdma::common::MessageMediator::sendErrorMsg(std::string message) {
    for (auto &elem : m_observers)
        elem.second->displayErrorMsg(message);
}

void gdma::common::MessageMediator::sendDebugMsg(std::string message) {
    for (auto &elem : m_observers)
        elem.second->displayDebugMsg(message);
}

void gdma::common::MessageMediator::sendTraceMsg(std::string message) {
    for (auto &elem : m_observers)
        elem.second->displayTraceMsg(message);
}

void gdma::common::MessageMediator::sendFatalMsg(std::string message) {
    for (auto &elem : m_observers)
        elem.second->displayFatalMsg(message);
}

// --------------------------------------------------------------------------------------
// Protected Members
int gdma::common::MessageMediator::generateObserverId() {
    return m_obsCounter++;
}
