/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/common/messageMediator.hpp

  \brief A singleton class used to manage messages.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

#ifndef GEODMA_LIB_COMMON_MESSAGEMEDIATOR_HPP_
#define GEODMA_LIB_COMMON_MESSAGEMEDIATOR_HPP_

// TerraLib Includes
#include <terralib/common/Singleton.h>
#include <terralib/common/ThreadingPolicies.h>

// STL Includes
#include <string>
#include <map>

// GeoDMA Includes
#include "geodmalib/common/config.hpp"

namespace gdma {
namespace common {

class AbstractMessageObserver;
/*!
  \class ProgressManager

  \brief A singleton class used to manage messages, warning and errors, and the created Observers
  (Command Line, Dialog, log, etc.)

  This singleton is used to send messages to Observers and store all message viewers created.
  If one or more Observers are attached, all instances will be used to display the messages.

  \ingroup common

  \sa AbstractMessageObserver, CommandLineMsgObserver, LogMessageObserver
*/
class GDMACOMMONEXPORT MessageMediator : public te::common::ObjectLevelLockable<MessageMediator,
                                                    boost::recursive_mutex,
                                                    boost::lock_guard<boost::recursive_mutex>,
                                                    boost::lock_guard<boost::recursive_mutex> >,
                    public te::common::Singleton<MessageMediator> {
    friend class te::common::Singleton<MessageMediator>;
 public:
    /*!
      \brief Attach a progress observer
      \param[in] apv Progress viewer instance
      \return Progress viewer identifier

      \note The Manager will NOT take the ownership of the Observers (obs).
    */
    int addObserver(AbstractMessageObserver* obs);

    /*!
      \brief Dettach a progress viewer
      \param[in] apv Progress viewer instance
      \return Progress viewer identifier
      \note
    */
    void removeObserver(int obsId);

    /*!
      \brief Retrieve the number of observers registered in the Mediator.
      \return The total number of observers added to the mediator.
    */
    int getNumberOfObservers();

    /*!
      \brief Removes references for all observers.
    */
    void clearAll();

    /*!
      \brief Send an information message to the observers
      \param[in] message The information message to be sent
    */
    void sendInformationMsg(std::string message);

    /*!
      \brief Send a warning message to the observers
      \param[in] message The warning message to be sent
    */
    void sendWarningMsg(std::string message);

    /*!
      \brief Send an error message to the observers
      \param[in] message The error message to be sent
    */
    void sendErrorMsg(std::string message);

    /*!
      \brief Send a debug message to the observers
      \param[in] message The debug message to be sent
    */
    void sendDebugMsg(std::string message);

    /*!
      \brief Send a trace message to the observers
      \param[in] message The trace message to be sent
    */
    void sendTraceMsg(std::string message);

    /*!
      \brief Send a fatal message to the observers
      \param[in] message The fatal message to be sent
    */
    void sendFatalMsg(std::string message);

 protected:
     /*! \brief Default constructor. */
    MessageMediator();

    /*! \brief Default destructor. */
    ~MessageMediator();

    /*!
      \brief Used to generate a new observer id (use internal counter).

      \return Observer identification as integer value.
    */
    int generateObserverId();

 protected:
    int m_obsCounter;                                //!< Counter used to generate the observers id.
    std::map<int, AbstractMessageObserver*> m_observers;  //!< Container with observers
};
}  // namespace common
}  // namespace gdma

#endif  // GEODMA_LIB_COMMON_MESSAGEMEDIATOR_HPP_
