/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/common/utils.cpp

  \brief This file contains some general utility functions.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// TerraLib Includes
#include <terralib/core/plugin.h>
#include <terralib/core/utils/Platform.h>
#include <terralib/xml.h>  // REMOVE IT
#include <terralib/common/Exception.h>  // REMOVE
#include <terralib/dataaccess/datasource/DataSourceCapabilities.h>
#include <terralib/dataaccess/datasource/DataSourceFactory.h>
#include <terralib/dataaccess/datasource/DataSource.h>
#include <terralib/common/StringUtils.h>

#include <boost/filesystem/convenience.hpp>  // Remove it
#include <boost/format.hpp>

// STL Includes
#include <ctime>
#include <iomanip>
#include <iostream>
#include <memory>
#include <map>
#include <sstream>
#include <string>
#include <vector>

// GeoDMA Includes
#include "geodmalib/common/messageMediator.hpp"
#include "geodmalib/common/utils.hpp"

// Load TerraLib Modules
bool gdma::common::loadTerraLibDrivers() {
    std::string plugins_path = te::core::FindInTerraLibPath("share/terralib/plugins");

    try {
        te::core::PluginInfo info;

        info = te::core::JSONPluginInfoSerializer(plugins_path + "/te.da.ogr.teplg.json");
        te::core::PluginManager::instance().insert(info);
        te::core::PluginManager::instance().load(info.name);

        info = te::core::JSONPluginInfoSerializer(plugins_path + "/te.da.gdal.teplg.json");
        te::core::PluginManager::instance().insert(info);
        te::core::PluginManager::instance().load(info.name);
    }
    catch (const te::common::Exception& e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(e.what());
        return false;
    }
    catch (const std::exception& e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(e.what());
        return false;
    }
    catch (...) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg("Unknown Error!");
        return false;
    }

    return true;
}

// Load TerraLib Modules
void gdma::common::unloadTerraLibDrivers() {
    te::core::PluginManager::instance().clear();
}

std::string gdma::common::formatTime(int seconds) {
    int hours = seconds / 3600;
    seconds -= hours * 3600;
    int minutes = seconds / 60;
    seconds -= minutes * 60;

    std::ostringstream output;
    output << std::setfill('0') << std::setw(2) << hours;
    output << ":" << std::setw(2) << minutes;
    output << ":" << std::setw(2) << seconds;
    return output.str();
}

void gdma::common::getTerralibRasterExtensions( std::vector< std::string >& rasterExtensions,
  const bool withCreationSupport )
{
  rasterExtensions.clear();

  if( te::da::DataSourceFactory::find( "GDAL" ) )
  {
    std::unique_ptr< te::da::DataSource > dsPtr =
      te::da::DataSourceFactory::make( "GDAL", "file://" );

    if( dsPtr.get() )
    {
      std::map< std::string, std::string > specCap =
        dsPtr->getCapabilities().getSpecificCapabilities();

      const std::string extensionName = withCreationSupport ?
        std::string( "SUPPORTED_RASTER_CREATION_EXTENSIONS" ) :
        std::string( "SUPPORTED_RASTER_EXTENSIONS" );

      if( specCap.find( extensionName ) != specCap.end() )
      {
        std::string fileExtensions = specCap[ extensionName ];
        std::string extFilter;
        std::string uCaseToken;
        std::string lCaseToken;

        std::vector< std::string > tokens;
        te::common::Tokenize( fileExtensions, rasterExtensions, ";" );

        for( std::size_t rasterExtensionsIdx = 0 ; rasterExtensionsIdx <
          rasterExtensions.size() ; ++rasterExtensionsIdx )
        {
          rasterExtensions[ rasterExtensionsIdx ] =
            te::common::Convert2UCase( rasterExtensions[ rasterExtensionsIdx ] );
        }
      }
    }
  }
}

std::map<std::string, int> gdma::common::generateRandomColor() {
    std::map<std::string, int> color;
    int seed = static_cast<int>(std::time(0));
    std::srand(seed);

    color["red"] = std::rand() % 255;
    color["green"] = std::rand() % 255;
    color["blue"] = std::rand() % 255;
    color["alpha"] = std::rand() % 255;

    return color;
}

std::string gdma::common::generateRandomHexColor() {
    std::vector<char> chars = { '0', '1', '2', '3',
                                '4', '5', '6', '7',
                                '8', '9', 'a', 'b',
                                'c', 'd', 'e', 'f' };

    std::string color = "#";
    for (int i = 0; i < 6; i++) {
        color.push_back(chars[std::rand() % chars.size()]);
    }

    return color;
}
void gdma::common::subStrReplace( const std::string& oldSubStr, const std::string& newSubStr,
  std::string& otherStr )
{
  std::size_t pos = std::string::npos;

  while( ( pos = otherStr.find( oldSubStr ) ) != std::string::npos )
  {
    otherStr.replace( pos, oldSubStr.size(), newSubStr );
  }
}
