/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/common/utils.hpp

  \brief This file contains some general utility functions.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

#ifndef GEODMA_LIB_COMMON_UTILS_HPP_
#define GEODMA_LIB_COMMON_UTILS_HPP_

// STL Includes
#include <map>
#include <string>
#include <vector>

// GeoDMA Includes
#include "geodmalib/common/config.hpp"

namespace gdma {
namespace common {
/*! \brief Load TerraLib Modules needed */
GDMACOMMONEXPORT bool loadTerraLibDrivers();

/*! \brief Unload loaded TerraLib Modules*/
GDMACOMMONEXPORT void unloadTerraLibDrivers();

/*!
  \brief Given a time in seconds, formats it in "hh:mm:ss"
  \param seconds The time to be formated, in seconds.
  \return An string with the time formated in "hh:mm::ss".
*/
GDMACOMMONEXPORT std::string formatTime(int seconds);

/*!
  \brief Returns the Terralib raster supported extensions (upper case).
  \param rasterExtensions A vector of supported extensions (upper case).
  \param withCreationSupport If true, only the extensions with creation support will be returned.
*/
GDMACOMMONEXPORT void getTerralibRasterExtensions(
  std::vector< std::string >& rasterExtensions,
  const bool withCreationSupport );

/*!
  \brief Generates a random color in the format RGBA.
  \return Returns a map<std::string, int> containing the components "red", "green", "blue" and
          "alpha" of the color.
*/
GDMACOMMONEXPORT std::map<std::string, int> generateRandomColor();

/*!
  \brief Generates a random color in the HEXADECIMAL format.
  \return Returns a string in the format "#XXXXXX" containing the hexadecimal color.
*/
GDMACOMMONEXPORT std::string generateRandomHexColor();

/*!
  \brief Replace all occurrences of subStr by newSubStr within otherStr.
  \param oldSubStr Old Sub-string to be replaced
  \param newSubStr New Sub-string
  \param otherStr String where the substitutions will occur
*/
GDMACOMMONEXPORT void subStrReplace(const std::string& oldSubStr, const std::string& newSubStr,
  std::string& otherStr );

}  // end namespace common
}  // end namespace gdma

#endif  // GEODMA_LIB_COMMON_UTILS_HPP_
