/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/dataManager/dataManager.cpp

  \brief This file implements functions to manage data.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// TerraLib Includes
#include <terralib/dataaccess.h>
#include <terralib/raster.h>
#include <terralib/common/Exception.h>
#include <terralib/common/Module.h>
#include <terralib/common/progress/TaskProgress.h>
#include <terralib/memory.h>
#include <terralib/memory/DataSource.h>
#include <terralib/datatype/SimpleData.h>

// Boost Includes
#include <boost/bimap.hpp>
#include <boost/filesystem.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/uuid/random_generator.hpp>
#include <boost/uuid/uuid_io.hpp>

// STL Includes
#include <fstream>
#include <iostream>
#include <map>
#include <memory>
#include <regex>
#include <set>
#include <string>
#include <vector>
#include <utility>

// GeoDMA Includes
#include "geodmalib/common/commandLineProgBar.hpp"
#include "geodmalib/common/messageMediator.hpp"
#include "geodmalib/datamanager/dataManager.hpp"
#include "geodmalib/featureextraction/attribute.hpp"
#include "geodmalib/featureextraction/featuresSummary.hpp"

// Opens the input raster file
te::rst::Raster* gdma::dm::openRaster(std::string pathRaster) {
    std::map<std::string, std::string> rasterInfo;
    rasterInfo["URI"] = pathRaster;

    te::rst::Raster* inputRaster = te::rst::RasterFactory::open(rasterInfo);

    return inputRaster;
}

std::string gdma::dm::openShapefile(std::string pathShp) {
    if (std::strcmp(boost::filesystem::extension(pathShp).c_str(), ".shp") != 0) {
        std::ostringstream output;
        output << "ERROR: the file \"" << pathShp << "\" is not a shape file!" << std::endl;
        throw te::common::Exception(TE_TR(output.str()));
    }

    boost::filesystem::path uri(pathShp);
    const std::string connInfo("file://" + pathShp);

    // let's include the new datasource in the managers
    boost::uuids::basic_random_generator<boost::mt19937> gen;
    boost::uuids::uuid u = gen();
    std::string id = boost::uuids::to_string(u);

    te::da::DataSourceInfoPtr ds(new te::da::DataSourceInfo);

    ds->setConnInfo(connInfo);
    ds->setTitle(uri.stem().string());
    ds->setAccessDriver("OGR");
    ds->setType("OGR");
    ds->setDescription(uri.string());
    ds->setId(id);

    te::da::DataSourceInfoManager::getInstance().add(ds);

    return ds->getId();
}

// @TODO(RAIAN): Extends it to work either with Databases
// @TODO(RAIAN): Change the parameter, to receive the URI instead of the path
te::da::DataSourceInfoPtr gdma::dm::openDataSource(std::string pathShp) {
    if (std::strcmp(boost::filesystem::extension(pathShp).c_str(), ".shp") != 0) {
        std::ostringstream output;
        output << "ERROR: the file \"" << pathShp << "\" is not a shape file!" << std::endl;
        throw te::common::Exception(TE_TR(output.str()));
    }

    boost::filesystem::path uri(pathShp);
    const std::string connInfo("file://" + pathShp);

    // let's include the new datasource in the managers
    boost::uuids::basic_random_generator<boost::mt19937> gen;
    boost::uuids::uuid u = gen();
    std::string id = boost::uuids::to_string(u);

    te::da::DataSourceInfoPtr ds(new te::da::DataSourceInfo);

    ds->setConnInfo(connInfo);
    ds->setTitle(uri.stem().string());
    ds->setAccessDriver("OGR");
    ds->setType("OGR");
    ds->setDescription(uri.string());
    ds->setId(id);

    te::da::DataSourceInfoManager::getInstance().add(ds);

    return ds;
}

// Opens the input shape-file
bool gdma::dm::getGeometriesFromDataSource(std::string dataSourceId,
    std::map<std::string, te::gm::Geometry*> &gm_container) {

    te::da::DataSourcePtr m_dataSource;

    try {
        m_dataSource = te::da::GetDataSource(dataSourceId, true);
    } catch (te::common::Exception &e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(e.what());
        return false;
    } catch (std::exception &e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(e.what());
        return false;
    } catch (...) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg("Unknown Exception!");
        return false;
    }

    if (m_dataSource.get() == 0)
        return false;

    te::da::DataSourceTransactorPtr transactor = m_dataSource->getTransactor();

    std::vector<std::string> dataSetsNames = transactor->getDataSetNames();

    // Transverse the vector with the names of the dada sets and retrieve each one, inserting
    // each geometry of the dataset inside a vector.
    for (int dataSetsNamesIndex = 0; dataSetsNamesIndex <
        dataSetsNames.size(); dataSetsNamesIndex++) {
        te::da::DataSetPtr dataset = transactor->getDataSet(
            dataSetsNames[dataSetsNamesIndex]);

        te::common::TaskProgress task(TE_TR("Loading Data"));
        task.setTotalSteps(dataset->size());
        task.useTimer(true);

        // Trying to get ObjectId
        te::da::DataSetTypePtr d_type = transactor->getDataSetType(
            dataSetsNames[dataSetsNamesIndex]);
        std::vector<std::string> pnames;
        te::da::GetOIDPropertyNames(d_type.get(), pnames);

        if (!dataset->isEmpty()) {
            dataset->moveBeforeFirst();
            while (dataset->moveNext()) {
                std::size_t geomPos = te::da::GetFirstPropertyPos(dataset.get(),
                    te::dt::GEOMETRY_TYPE);

                std::string key = "";
                for (int posNames = 0; posNames < pnames.size(); posNames++) {
                    if (posNames < pnames.size() - 1)
                        key = std::string(key) + dataset->getString(pnames[posNames]) +
                            std::string("|");
                    else
                        key = std::string(key) + dataset->getString(pnames[posNames]);
                }
                // std::string key = dataset->getString(pnames[0]);
                te::gm::Geometry* geometry = dataset->getGeometry(geomPos).release();
                gm_container.insert(
                    std::pair<std::string, te::gm::Geometry*>(key, geometry));

                if (task.isActive() == false)
                    throw te::common::Exception(TE_TR("Operation canceled!"));
            }
        } else {
            std::ostringstream output;
            output << "Dataset " << dataSetsNames[dataSetsNamesIndex] << "is empty and can not ";
            output << "be loaded." << std::endl;
            throw te::common::Exception(TE_TR(output.str()));
        }
    }

  return true;
}

// Opens the input shape - file
bool gdma::dm::getGeometriesFromDataSource(std::string dataSourceId,
boost::bimap<std::string, std::size_t> &positions, std::vector<te::gm::Geometry*> &geom_container,
    std::string prop_class, std::vector<std::string>* prop_vec,
    std::set<std::string>* class_values) {

    te::da::DataSourcePtr m_dataSource;

    try {
        m_dataSource = te::da::GetDataSource(dataSourceId, true);
    } catch (te::common::Exception &e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(e.what());
        return false;
    } catch (std::exception &e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(e.what());
        return false;
    } catch (...) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg("Unknown Exception!");
        return false;
    }

    if (m_dataSource.get() == 0)
        return false;

    te::da::DataSourceTransactorPtr transactor = m_dataSource->getTransactor();

    std::vector<std::string> dataSetsNames = transactor->getDataSetNames();

    if (geom_container.size() > 0) {
        te::common::FreeContents(geom_container);
        geom_container.resize(0);
    }

    if (prop_vec != 0 && prop_vec->size() > 0)
        prop_vec = new std::vector<std::string>();

    if (class_values != 0 && class_values->size() > 0)
        class_values = new std::set<std::string>();

    bool loadProperty = true;
    if (prop_class.compare("") == 0)
        loadProperty = false;

    // Transverse the vector with the names of the dada sets and retrieve each one, inserting
    // each geometry of the dataset inside a vector.
    std::size_t count = 0;
    for (int dataSetsNamesIndex = 0; dataSetsNamesIndex <
        dataSetsNames.size(); dataSetsNamesIndex++) {
        te::da::DataSetTypePtr dataSetTypePtr =
            transactor->getDataSetType(dataSetsNames[dataSetsNamesIndex]);
        bool propertyExists = gdma::dm::verifyIfColumnExists(dataSetTypePtr, prop_class);

        if (!loadProperty || propertyExists) {
            te::da::DataSetPtr dataset = transactor->getDataSet(
                dataSetsNames[dataSetsNamesIndex]);

            te::common::TaskProgress task(TE_TR("Loading Data"));
            task.setTotalSteps(dataset->size());
            task.useTimer(true);

            // Trying to get ObjectId
            te::da::DataSetTypePtr d_type = transactor->getDataSetType(
                dataSetsNames[dataSetsNamesIndex]);
            std::vector<std::string> pnames;
            te::da::GetOIDPropertyNames(d_type.get(), pnames);

            if (!dataset->isEmpty()) {
                dataset->moveBeforeFirst();
                while (dataset->moveNext()) {
                    std::size_t geomPos = te::da::GetFirstPropertyPos(dataset.get(),
                        te::dt::GEOMETRY_TYPE);

                    std::string key = "";
                    for (int posNames = 0; posNames < pnames.size(); posNames++) {
                        if (posNames < pnames.size() - 1)
                            key = std::string(key) + dataset->getString(pnames[posNames]) +
                            std::string("|");
                        else
                            key = std::string(key) + dataset->getString(pnames[posNames]);
                    }

                    te::gm::Geometry* geometry = dataset->getGeometry(geomPos).release();
                    positions.insert(boost::bimap<std::string, std::size_t>::value_type(key, count));
                    geom_container.push_back(geometry);
                    if (loadProperty) {
                        std::string class_val = dataset->getValue(prop_class)->toString();
                        prop_vec->push_back(class_val);
                        class_values->insert(class_val);
                    }

                    count++;

                    if (task.isActive() == false)
                        throw te::common::Exception(TE_TR("Operation canceled!"));

                    task.pulse();
                }
                break;
            } else {
                std::ostringstream output;
                output << "Dataset " << dataSetsNames[dataSetsNamesIndex] << "is empty and can not ";
                output << "be loaded." << std::endl;
                throw te::common::Exception(TE_TR(output.str()));
            }
        } else {
            std::stringstream msg;
            msg << "Column \"" << prop_class << "\" does not exists in the ";
            msg << "data set!";
            throw te::common::Exception(TE_TR(msg.str()));
        }
    }
}

std::vector<te::gm::Geometry*> gdma::dm::getGeometriesFromDataset(te::da::DataSet* dataset,
    int blockSize, int currPosition) {
    std::vector<te::gm::Geometry*> gm_container;
    if (!dataset->isEmpty()) {
        dataset->moveBeforeFirst();
        while (dataset->moveNext()) {
            std::size_t geomPos = te::da::GetFirstPropertyPos(dataset, te::dt::GEOMETRY_TYPE);

            gm_container.push_back(dataset->getGeometry(geomPos).release());
        }
    } else {
        std::ostringstream output;
        output << "Dataset is empty and can not be loaded." << std::endl;
        throw te::common::Exception(TE_TR(output.str()));
    }
    return gm_container;
}

bool gdma::dm::diskVectorFileCopy( const std::string& inputFileName,
  const std::string& outputFileName ) {
  if( inputFileName.empty() ) return false;
  if( outputFileName.empty() ) return false;

  // input data

  const std::string inputDataSetName = boost::filesystem::path( inputFileName ).stem().string();
  const std::string outputDataSetName = boost::filesystem::path( outputFileName ).stem().string();

  std::unique_ptr< te::da::DataSource > inputDataSourcePtr( te::da::DataSourceFactory::make(
    "OGR", "file://" + inputFileName ) );
  if( inputDataSourcePtr.get() == 0 )
  {
    return false;
  }

  inputDataSourcePtr->open();

  if( ! inputDataSourcePtr->isValid() )
  {
    return false;
  }

  if( ! inputDataSourcePtr->dataSetExists( inputDataSetName ) )
  {
    return false;
  }

  std::unique_ptr<te::da::DataSetType> inputDataSetTypePtr(new te::da::DataSetType(
      *inputDataSourcePtr->getDataSetType( inputDataSetName ).get()));

  // Creating output data source (new SHP)

  te::da::DataSourcePtr outputDataSourcePtr = te::da::DataSourceFactory::make("OGR",
      "file://" + outputFileName );

  if (outputDataSourcePtr.get() == 0)
  {
    return false;
  }

  outputDataSourcePtr->open();

  if( ! outputDataSourcePtr->isValid() )
  {
    return false;
  }

  // Creates the dataset converter

  te::da::DataSetTypeConverter datasetConverter( inputDataSetTypePtr.get(),
    outputDataSourcePtr->getCapabilities() );

  std::unique_ptr< te::da::DataSetType > outputDataSetTypePtr( new te::da::DataSetType( *datasetConverter.getResult() ) );

  // Get the transactor and creates the new dataset

  te::da::DataSourceTransactorPtr outputTransactorPtr;

  try {
    outputTransactorPtr = outputDataSourcePtr->getTransactor();

    outputTransactorPtr->begin();

    outputTransactorPtr->createDataSet( outputDataSetTypePtr.get(),
        std::map<std::string, std::string>());

    std::unique_ptr<te::da::DataSet> inputDatasetPtr(
        inputDataSourcePtr->getDataSet( inputDataSetName ) );

    std::unique_ptr<te::da::DataSetAdapter> datasetAdapter(te::da::CreateAdapter(
        inputDatasetPtr.get(), &datasetConverter, false ) );

    // Adding new dataset to output datasource

    if (! inputDatasetPtr->moveBeforeFirst() )
    {
      return false;
    }

    outputTransactorPtr->add( outputDataSetName, datasetAdapter.get(),
        std::map< std::string, std::string >(), 0);

    outputTransactorPtr->commit();
  } catch (...) {
    return false;
  }

  outputDataSourcePtr->close();

  return true;
}

bool gdma::dm::updateDataSource(std::string dataSourceId,
    std::map<std::string, gdma::fte::FeaturesSummary> &attr_container) {
    // Get the DataSource
    te::da::DataSourcePtr dataSource;

    try {
        dataSource = te::da::GetDataSource(dataSourceId, true);
    } catch (te::common::Exception &e) {
        te::da::DataSourceInfoPtr dsInfo = openDataSource(dataSourceId);
        dataSource = te::da::GetDataSource(dsInfo->getId(), true);
    } catch (std::exception &e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(e.what());
        return false;
    } catch (...) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg("Unknown Exception!");
        return false;
    }

    if (dataSource.get() == 0)
        throw te::common::Exception(TE_TR("Data source could not be opened!"));

    te::da::DataSourceTransactorPtr transactor = dataSource->getTransactor();

    std::vector<std::string> dataSetsNames = transactor->getDataSetNames();
    std::string dataSetName = dataSetsNames[0];

    te::da::DataSetPtr dataSet = dataSource->getDataSet(dataSetName);

    te::common::TaskProgress task(TE_TR("Saving SHP"));
    task.setTotalSteps(attr_container.size());
    task.useTimer(true);

    // Create the new properties for the attributes being saved
    std::vector<gdma::fte::Attribute> features = attr_container.begin()->second.getFeatures();

    for (int pos = 0; pos < features.size(); pos++) {
        std::string columnName = features[0].getShortName();
        if (!transactor->propertyExists(dataSetName, columnName)) {
            std::regex validator("[A-Za-z0-9_]+");
            if (!std::regex_match(columnName, validator)) {
                std::stringstream msg;
                msg << "Column Name \"" << columnName << "\" contains special characters!";
                throw te::common::Exception(msg.str());
            }

            transactor->addProperty(dataSetName, new te::dt::SimpleProperty(
                features[0].getShortName(), features[0].getValueType(), true));
        }
    }

    te::da::DataSetTypePtr dataSetTypePtr = dataSource->getDataSetType(dataSetName);
    std::unique_ptr<te::mem::DataSet> memDataSetPtr(new te::mem::DataSet(dataSetTypePtr.get()));

    //// Get primary keys and their positions
    te::da::PrimaryKey* primKey = dataSetTypePtr->getPrimaryKey();
    std::vector<te::dt::Property*> primKeyProps = primKey->getProperties();
    std::size_t numPropsPk = primKeyProps.size();
    std::vector<std::size_t> idsPk;

    for (std::size_t posPk = 0; posPk < numPropsPk; posPk++) {
        std::string propName = primKeyProps[posPk]->getName();
        idsPk.push_back(dataSetTypePtr->getPropertyPosition(propName));
    }

    // get the new properties position
    std::vector<std::set<int>> fields;

    dataSet->moveBeforeFirst();
    while (dataSet->moveNext()) {
        te::mem::DataSetItem* dsItem = new te::mem::DataSetItem(memDataSetPtr.get());
        std::set<int> attribPos;

        std::string key = "";
        for (unsigned int posPk = 0; posPk < primKeyProps.size(); posPk++) {
            std::string pkName = primKeyProps[posPk]->getName();

            if (posPk < primKeyProps.size() - 1)
                key = std::string(key) + dataSet->getAsString(primKeyProps[posPk]->getName()) +
                    std::string("|");
            else
                key = std::string(key) + dataSet->getAsString(primKeyProps[posPk]->getName());

            dsItem->setValue(pkName, dataSet->getValue(pkName).release());
        }

        std::map<std::string, gdma::fte::FeaturesSummary>::iterator itAttr =
            attr_container.find(key);

        if (itAttr != attr_container.end()) {
            std::vector<gdma::fte::Attribute> features = itAttr->second.getFeatures();

            for (unsigned int pos = 0; pos < features.size(); pos++) {
                attribPos.insert(dataSetTypePtr->getPropertyPosition(
                    features[pos].getShortName()));

                switch (features[pos].getValueType()) {
                case te::dt::STRING_TYPE:
                    dsItem->setString(features[pos].getShortName(),
                        features[pos].getString());
                    break;
                case te::dt::INT16_TYPE:
                    dsItem->setInt16(features[pos].getShortName(),
                        features[pos].getInt16());
                    break;
                case te::dt::INT32_TYPE:
                    dsItem->setInt32(features[pos].getShortName(),
                        features[pos].getInt32());
                    break;
                case te::dt::INT64_TYPE:
                    dsItem->setInt64(features[pos].getShortName(),
                        features[pos].getInt64());
                    break;
                case te::dt::FLOAT_TYPE:
                    dsItem->setFloat(features[pos].getShortName(),
                        features[pos].getFloat());
                    break;
                case te::dt::DOUBLE_TYPE:
                    dsItem->setDouble(features[pos].getShortName(),
                        features[pos].getDouble());
                    break;
                default:
                    break;
                }
            }
            memDataSetPtr->add(dsItem);
            fields.push_back(attribPos);
        } else {
            delete dsItem;
            dsItem = 0;
        }

        if (task.isActive() == false)
            throw te::common::Exception(TE_TR("Operation canceled!"));
        task.pulse();
    }

    try {
        transactor->update(dataSetName, memDataSetPtr.get(), fields, idsPk);
    } catch (const te::common::Exception& e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(e.what());
        transactor->rollBack();
        return false;
    } catch (const std::exception& e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(e.what());
        transactor->rollBack();
        return false;
    } catch (...) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg("Unknown Error!");
        transactor->rollBack();
        return false;
    }

    transactor->commit();

    return true;
}

bool gdma::dm::saveToOGRDataSource(std::string newDataSource, std::string oldDataSetName,
    std::string oldDataSourceName,
    std::map<std::string, gdma::fte::FeaturesSummary> &attr_container) {
    // Find the DataSource
    te::da::DataSourcePtr m_inputDataSource =
        te::da::DataSourceManager::getInstance().find(oldDataSourceName);
    if (m_inputDataSource.get() == 0)
        return false;

    // Retrieve the datasetType
    te::da::DataSetTypePtr dataSetTypePtr = m_inputDataSource->getDataSetType(oldDataSetName);

    te::da::DataSetPtr m_dataSet = m_inputDataSource->getDataSet(oldDataSetName);

    te::common::TaskProgress task(TE_TR("Saving SHP"));
    task.setTotalSteps(attr_container.size());
    task.useTimer(true);

    te::mem::DataSetPtr memDataSetPtr(new te::mem::DataSet(dataSetTypePtr.get()));

    memDataSetPtr->moveBeforeFirst();
    memDataSetPtr->copy(*m_dataSet, m_dataSet->size());

    // Create the new properties for the attributes being saved
    std::map<std::string, gdma::fte::FeaturesSummary>::iterator itAttrs = attr_container.begin();
    std::vector<gdma::fte::Attribute> features = itAttrs->second.getFeatures();

    for (int i = 0; i < features.size(); i++) {
        dataSetTypePtr->add(new te::dt::SimpleProperty(
            features[i].getShortName(), features[i].getValueType(), true));
        memDataSetPtr->add(features[i].getShortName(), features[i].getValueType());
    }

    std::vector<std::string> pnames;
    te::da::GetOIDPropertyNames(dataSetTypePtr.get(), pnames);

    memDataSetPtr->moveBeforeFirst();
    while (memDataSetPtr->moveNext()) {
        std::string key = "";
        for (int posNames = 0; posNames < pnames.size(); posNames++) {
            if (posNames < pnames.size() - 1)
                key = std::string(key) + memDataSetPtr->getValue(
                    dataSetTypePtr->getPropertyPosition(pnames[posNames]))->toString() +
                std::string("|");
            else
                key = std::string(key) + memDataSetPtr->getValue(
                    dataSetTypePtr->getPropertyPosition(pnames[posNames]))->toString();
        }

        std::vector<gdma::fte::Attribute> features = attr_container[key].getFeatures();
        for (unsigned int posFt = 0; posFt < features.size(); posFt++) {
            switch (features[posFt].getValueType()) {
            case te::dt::STRING_TYPE:
                memDataSetPtr->setString(features[posFt].getShortName(),
                    features[posFt].getString());
                break;
            case te::dt::INT16_TYPE:
                memDataSetPtr->setInt16(features[posFt].getShortName(),
                    features[posFt].getInt16());
                break;
            case te::dt::INT32_TYPE:
                memDataSetPtr->setInt32(features[posFt].getShortName(),
                    features[posFt].getInt32());
                break;
            case te::dt::INT64_TYPE:
                memDataSetPtr->setInt64(features[posFt].getShortName(),
                    features[posFt].getInt64());
                break;
            case te::dt::FLOAT_TYPE:
                memDataSetPtr->setFloat(features[posFt].getShortName(),
                    features[posFt].getFloat());
                break;
            case te::dt::DOUBLE_TYPE:
                memDataSetPtr->setDouble(features[posFt].getShortName(),
                    features[posFt].getDouble());
                break;
            default:
                break;
            }
        }

        if (task.isActive() == false)
            throw te::common::Exception(TE_TR("Operation canceled!"));

        task.pulse();
    }

    // Create Output DataSource
    te::da::DataSourcePtr dsOGR(
        te::da::DataSourceFactory::make("OGR", "file://" + newDataSource));
    dsOGR->open();
    if (dsOGR.get() == 0)
        return false;

    memDataSetPtr->moveBeforeFirst();

    try {
        dsOGR->createDataSet(dataSetTypePtr.get(),
            std::map<std::string, std::string>());
        dsOGR->add(boost::filesystem::basename(newDataSource), memDataSetPtr.get(),
            std::map<std::string, std::string>());
        dsOGR->close();
    } catch (const te::common::Exception& e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(e.what());
        return false;
    }

    return true;
}

bool gdma::dm::saveAttributesToCSV(
    std::map<std::string, gdma::fte::FeaturesSummary> &attr_container,
    std::string dataSourceName, std::string dataSetName, std::string pathFile) {

    std::ofstream attrFile;
    attrFile.open(pathFile.c_str(), std::ios::out);

    te::da::DataSourcePtr m_dataSourcePtr =
        te::da::DataSourceManager::getInstance().find(dataSourceName);
    te::da::DataSourceTransactorPtr transactor = m_dataSourcePtr->getTransactor();
    te::da::DataSetPtr dataset = transactor->getDataSet(dataSetName);

    std::vector<std::string> pnames;
    te::da::GetOIDPropertyNames(transactor->getDataSetType(dataSetName).get(), pnames);

    te::common::TaskProgress task(TE_TR("Saving CSV"));
    task.setTotalSteps(attr_container.size());
    task.useTimer(true);

    for (int posNames = 0; posNames < pnames.size(); posNames++) {
        if (posNames < pnames.size() - 1)
            attrFile << pnames[posNames] << "|";
        else
            attrFile << pnames[posNames] << ";";
    }

    std::map<std::string, gdma::fte::FeaturesSummary>::iterator itAttrs = attr_container.begin();
    std::vector<gdma::fte::Attribute> features = itAttrs->second.getFeatures();
    for (int i = 0; i < features.size(); i++) {
        attrFile << features[i].getName().c_str() << ";";
    }
    attrFile << std::endl;

    std::string key;

    while (itAttrs != attr_container.end()) {
        key = itAttrs->first;
        attrFile << key << ";";

        std::vector<gdma::fte::Attribute> features = itAttrs->second.getFeatures();

        for (int pos = 0; pos < features.size(); pos++) {
            switch (features[pos].getValueType()) {
            case te::dt::STRING_TYPE:
                attrFile << features[pos].getString() << ";";
                break;
            case te::dt::INT16_TYPE:
                attrFile << features[pos].getInt16() << ";";
                break;
            case te::dt::INT32_TYPE:
                attrFile << features[pos].getInt32() << ";";
                break;
            case te::dt::INT64_TYPE:
                attrFile << features[pos].getInt64() << ";";
                break;
            case te::dt::FLOAT_TYPE:
                attrFile << features[pos].getFloat() << ";";
                break;
            case te::dt::DOUBLE_TYPE:
                attrFile << features[pos].getDouble() << ";";
                break;
            default:
                break;
            }
        }
        attrFile << std::endl;

        if (task.isActive() == false)
            throw te::common::Exception(TE_TR("Operation canceled!"));

        task.pulse();
        itAttrs++;
    }
    attrFile.close();
    return true;
}

// Save a vector of polygons in a shapefile
bool gdma::dm::saveShpFromVector(std::vector<te::gm::Geometry*> gmVec,
    std::string pathShp) {
    te::mem::DataSetPtr memDataSetPtr;

    te::da::DataSetTypePtr dataSetTypePtr(new te::da::DataSetType(pathShp));

    te::gm::GeometryProperty* propPtr = new te::gm::GeometryProperty("geom", 0,
        gmVec[0]->getGeomTypeId(), true);
    propPtr->setSRID(gmVec[0]->getSRID());
    dataSetTypePtr->add(propPtr);

    memDataSetPtr.reset(new te::mem::DataSet(dataSetTypePtr.get()));

    for (unsigned int polygonsIdx = 0; polygonsIdx < gmVec.size(); polygonsIdx++) {
        te::mem::DataSetItem* dsItemPtr = new te::mem::DataSetItem(memDataSetPtr.get());
        dsItemPtr->setGeometry(0, (te::gm::Polygon*)gmVec[polygonsIdx]);

        memDataSetPtr->add(dsItemPtr);
    }

    te::da::DataSourcePtr dsOGR(te::da::DataSourceFactory::make("OGR", "file://" + pathShp));
    dsOGR->open();

    memDataSetPtr->moveBeforeFirst();
    try {
        dsOGR->createDataSet(dataSetTypePtr.get(), std::map<std::string, std::string>());
        dsOGR->add(boost::filesystem::basename(pathShp), memDataSetPtr.get(),
            std::map<std::string, std::string>());
        dsOGR->close();
    }
    catch (const te::common::Exception& e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(e.what());
        return false;
    }

    return true;
}

bool gdma::dm::createColumnOGRDataSource(std::string dataSourceId, std::string columnName,
    int dataType) {
    std::regex validator("[A-Za-z0-9_]+");
    if (!std::regex_match(columnName, validator)) {
        std::stringstream msg;
        msg << "Column Name \"" << columnName << "\" contains special characters!";
        throw te::common::Exception(msg.str());
    }

    te::da::DataSourcePtr dataSource;

    try {
        dataSource = te::da::GetDataSource(dataSourceId, true);
    } catch (te::common::Exception &e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(e.what());
        return false;
    } catch (std::exception &e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(e.what());
        return false;
    } catch (...) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg("Unknown Exception!");
        return false;
    }

    if (dataSource.get() == 0)
        return false;

    te::da::DataSourceTransactorPtr transactor = dataSource->getTransactor();

    std::vector<std::string> dataSetsNames = transactor->getDataSetNames();
    std::string dataSetName = dataSetsNames[0];

    te::da::DataSetPtr dataSet = dataSource->getDataSet(dataSetName);

    te::common::TaskProgress task(TE_TR("Creating column"));
    task.setTotalSteps(dataSet->size());
    task.useTimer(true);

    switch (dataType) {
    case te::dt::INT16_TYPE:
        dataType = te::dt::INT32_TYPE;
        break;
    case te::dt::FLOAT_TYPE:
        dataType = te::dt::DOUBLE_TYPE;
        break;
    default:
        break;
    }

    te::da::DataSetTypePtr dataSetTypePtr = dataSource->getDataSetType(dataSetName);
    bool columnExists = gdma::dm::verifyIfColumnExists(dataSetTypePtr, columnName);
    if (columnExists) {
        std::stringstream msg;
        msg << "Column \"" << columnName << "\" can not be created. It already exists in the ";
        msg << "data set!";
        throw te::common::Exception(TE_TR(msg.str()));
    }

    // Creating the new property
    transactor->addProperty(dataSetName, new te::dt::SimpleProperty(
        columnName, dataType, true));

    dataSetTypePtr = dataSource->getDataSetType(dataSetName);
    std::unique_ptr<te::mem::DataSet> memDataSetPtr(new te::mem::DataSet(dataSetTypePtr.get()));

    //// Get primary keys and their positions
    te::da::PrimaryKey* primKey = dataSetTypePtr->getPrimaryKey();
    std::vector<te::dt::Property*> primKeyProps = primKey->getProperties();
    std::size_t numPropsPk = primKeyProps.size();
    std::vector<std::size_t> idsPk;

    for (std::size_t posPk = 0; posPk < numPropsPk; posPk++) {
        std::string propName = primKeyProps[posPk]->getName();
        idsPk.push_back(dataSetTypePtr->getPropertyPosition(propName));
    }

    // get the new properties position
    std::vector<std::set<int>> fields;
    int count = 0;

    dataSet->moveBeforeFirst();
    while (dataSet->moveNext()) {
        te::mem::DataSetItem* dsItem = new te::mem::DataSetItem(memDataSetPtr.get());
        std::set<int> attribPos;

        for (unsigned int posPk = 0; posPk < primKeyProps.size(); posPk++) {
            std::string pkName = primKeyProps[posPk]->getName();
            dsItem->setValue(pkName, dataSet->getValue(pkName).release());
        }

        attribPos.insert(dataSetTypePtr->getPropertyPosition(columnName));
        count++;

        switch (dataType) {
        case te::dt::STRING_TYPE:
            dsItem->setString(columnName, "");
            break;
        case te::dt::INT32_TYPE:
            dsItem->setInt32(columnName, 0);
            break;
        case te::dt::INT64_TYPE:
            dsItem->setInt64(columnName, 0);
            break;
        case te::dt::DOUBLE_TYPE:
            dsItem->setDouble(columnName, 0);
            break;
        default:
            throw(te::common::Exception(TE_TR("Unsupported data type.")));
        }

        memDataSetPtr->add(dsItem);
        fields.push_back(attribPos);

        if (task.isActive() == false)
            throw te::common::Exception(TE_TR("Operation canceled!"));

        task.pulse();
    }

    try {
        transactor->update(dataSetName, memDataSetPtr.get(), fields, idsPk);
    } catch (const te::common::Exception& e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(e.what());
        transactor->rollBack();
        return false;
    } catch (const std::exception& e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(e.what());
        transactor->rollBack();
        return false;
    } catch (...) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg("Unknown Error!");
        transactor->rollBack();
        return false;
    }

    transactor->commit();

    return true;
}

bool gdma::dm::verifyIfColumnExists(te::da::DataSetTypePtr dsType, std::string columnName) {
    std::vector<te::dt::Property*> propVector =
        dsType->getProperties();

    for (auto &prop : propVector) {
        if (prop->getName().compare(columnName) == 0)
            return true;
    }
    return false;
}

bool gdma::dm::saveDataSourceToCSV(std::string dataSourceId, std::string pathToCSV,
    std::string dataSetName) {
    te::da::DataSourcePtr dataSource;

    try {
        dataSource = te::da::GetDataSource(dataSourceId, true);
    } catch (te::common::Exception &e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(e.what());
        return false;
    } catch (std::exception &e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(e.what());
        return false;
    } catch (...) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg("Unknown Exception!");
        return false;
    }

    if (dataSource.get() == 0)
        return false;

    te::da::DataSourceTransactorPtr transactor = dataSource->getTransactor();
    if (dataSetName.compare("") == 0)
        dataSetName = transactor->getDataSetNames()[0];

    te::da::DataSetPtr dataset = transactor->getDataSet(dataSetName);
    te::da::DataSetTypePtr dsType = dataSource->getDataSetType(dataSetName);

    te::common::TaskProgress task(TE_TR("Saving CSV"));
    task.setTotalSteps(dataset->size());
    task.useTimer(true);

    std::ofstream attrFile;
    attrFile.open(pathToCSV.c_str(), std::ios::out);

    std::vector<te::dt::Property*> properties = dsType->getProperties();

    for (int pos = 0; pos < properties.size(); pos++) {
        if (properties[pos]->getType() != te::dt::GEOMETRY_TYPE) {
            if (pos < properties.size() - 1)
                attrFile << properties[pos]->getName() << ";";
            else
                attrFile << properties[pos]->getName();
        }
    }
    attrFile << std::endl;

    dataset->moveBeforeFirst();
    while (dataset->moveNext()) {
        for (auto &prop : properties) {
            switch (prop->getType()) {
                case te::dt::STRING_TYPE:
                    attrFile << dataset->getString(prop->getName()) << ";";
                    break;
                case te::dt::INT16_TYPE:
                    attrFile << dataset->getInt16(prop->getName()) << ";";
                    break;
                case te::dt::INT32_TYPE:
                    attrFile << dataset->getInt32(prop->getName()) << ";";
                    break;
                case te::dt::INT64_TYPE:
                    attrFile << dataset->getInt64(prop->getName()) << ";";
                    break;
                case te::dt::FLOAT_TYPE:
                    attrFile << dataset->getFloat(prop->getName()) << ";";
                    break;
                case te::dt::DOUBLE_TYPE:
                    attrFile << dataset->getDouble(prop->getName()) << ";";
                    break;
                default:
                    break;
            }
        }

        attrFile << std::endl;

        if (task.isActive() == false)
            throw te::common::Exception(TE_TR("Operation canceled!"));

        task.pulse();
    }

    attrFile.close();
    return true;
}
