/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/dataManager/dataManager.hpp

  \brief Creates functions to manage data.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

#ifndef GEODMA_LIB_DATAMANAGER_DATAMANAGER_HPP_
#define GEODMA_LIB_DATAMANAGER_DATAMANAGER_HPP_

// TerraLib Includes
#include <terralib/dataaccess/datasource_h.h>
#include <terralib/maptools/AbstractLayer.h>

// Boost Includes
#include <boost/bimap.hpp>
#include <boost/shared_ptr.hpp>

// STL Includes
#include <map>
#include <set>
#include <string>
#include <vector>

// GeoDMA
#include "geodmalib/datamanager/config.hpp"
#include "geodmalib/featureextraction/featuresSummary.hpp"

namespace gdma {
namespace dm {
/*! \brief Opens the input raster file */
GDMADMEXPORT te::rst::Raster* openRaster(std::string pathRaster);

GDMADMEXPORT std::string openShapefile(std::string pathShp);

GDMADMEXPORT te::da::DataSourceInfoPtr openDataSource(std::string pathShp);

/*! \brief Opens the input shape-file */
GDMADMEXPORT bool getGeometriesFromDataSource(std::string dataSourceId,
    std::map<std::string, te::gm::Geometry*> &gm_container);

/*! 
  \brief Opens the input shape-file 
  \param[in] dataSourceId Id of the Data Source where geometries are stored.
  \param[out] positions A reference to an associative container (std::map<std::string, std::size_t)
                        that will link the id of the geometries to its positions in the vector of
                        geometries and in the vector of properties.
  \param[out] geom_container A reference to a vector of geometries. It will contain the geometries
                             loaded from the shape file.
  \param[in] prop_class An string containing the name of the column that corresponds to the class
                        of the polygons. This parameter is optional, and if it is not passed, only
                        the geometries will be loaded.
  \param[out] prop_vec A pointer to a vector of strings. It will contain the values of the property
                       loaded for each geometry.
  \param[out] class_values A set of strings. It will contain the existing values of the classes,
                           without duplicates.
*/
GDMADMEXPORT bool getGeometriesFromDataSource(std::string dataSourceId,
    boost::bimap<std::string, std::size_t> &positions, std::vector<te::gm::Geometry*> &geom_container,
    std::string prop_class = "", std::vector<std::string>* prop_vec = 0,
    std::set<std::string>* class_values = 0);

GDMADMEXPORT std::vector<te::gm::Geometry*> getGeometriesFromDataset(te::da::DataSet* dataset,
    int blockSize, int currPosition);

/*! 
  \brief Save a csv file with the attributes calculated
  \param attr_container[in] The container containing the new attributes to be saved.
  \param dataSourceName[in] Name of the data source containing the geometries used to compute the
                            given attributes.
  \param dataSetName[in] Name of the data set containing the geometries used to compute the
                         given attributes.
  \param pathFile[in] Path to the CSV file being created.
  \return true if ok, false on errors.
*/
GDMADMEXPORT bool saveAttributesToCSV(
    std::map<std::string, gdma::fte::FeaturesSummary> &attr_container,
    std::string dataSourceName, std::string dataSetName, std::string pathFile);

/*!
  \brief Create a disk vector filer copy.
  \param inputFileName[in] Full path of an existent existent vector disk file.
  \param outputFileName[in] Full path of the new vector disk file copy.
  \return true if ok, false on errors.
*/
GDMADMEXPORT bool diskVectorFileCopy( const std::string& inputFileName,
  const std::string& outputFileName );

/*!
  \brief Update an OGR data source from a container of attributes.
  \param dataSourceId[in] The ID of the data source being updated.
  \param attr_container[in] The container containing the new attributes to be saved.
  \return true if ok, false on errors.
*/
GDMADMEXPORT bool updateDataSource(std::string dataSourceId,
    std::map<std::string, gdma::fte::FeaturesSummary> &attr_container);

/*!
  \brief Save attributes to a new OGR Data Source. It clones all the columns of a given data source
         and add the new attributes.
  \param newDataSource[in] The path to the new data source being created.
  \param oldDataSetName[in] The name of the data set that will be cloned.
  \param oldDataSourceName[in] The name of the data source that will be cloned.
  \param attr_container[in] The container containing the new attributes to be saved.
  \return true if ok, false on errors.
*/
GDMADMEXPORT bool saveToOGRDataSource(std::string newDataSource, std::string oldDataSetName,
    std::string oldDataSourceName,
    std::map<std::string, gdma::fte::FeaturesSummary> &attr_container);

/*!
  \brief Creates a shape file from a vector of geometries.
  \param gmVec[in] The vector of pointers to the geometries to save in the shape file.
  \param pathShp[in] The path to the shape file being created.
  \return true if ok, false on errors.
*/
GDMADMEXPORT bool saveShpFromVector(std::vector<te::gm::Geometry*> gmVec, std::string pathShp);

/*!
  \brief Create a new column in the Data source and fill it with a default value, which depends on
         the data type. Default value for strings is an empty string, for numeric types, it is 0
         (zero).
  \param dataSourceId[in] The Id of the data source being updated.
  \param columnName[in] The name of the column being created.
  \return true if ok, false on errors.
*/
GDMADMEXPORT bool createColumnOGRDataSource(std::string dataSourceId, std::string columnName,
    int dataType);

/*!
  \brief Verify if a given column exists in a data set.
  \param dsType[in] The data set type of the data set being verifyed.
  \param columnName[in] The name of the column being searched.
  \return true if the column exists, false otherwise.
*/
GDMADMEXPORT bool verifyIfColumnExists(te::da::DataSetTypePtr dsType, std::string columnName);

/*!
  \brief Generate a CSV file from a given Data Source.
  \param dataSourceId[in] The Id of the data source being updated.
  \param pathToCSV[in] String containing the path to the CSV file being created.
  \param dataSetName[in] The name of the data set. If it is not given, the firts data set of the
                         data source will be used.
  \return true if ok, false on errors.
*/
GDMADMEXPORT bool saveDataSourceToCSV(std::string dataSourceId, std::string pathToCSV,
    std::string dataSetName = "");
}  // end namespace dm
}  // end namespace gdma

#endif  // GEODMA_LIB_DATAMANAGER_DATAMANAGER_HPP_
