/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/datamanager/datasetIterator.hpp

  \brief This file contains the DatasetIterator class definition. It iterates sequentially over
         all the positions of a given dataset.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// TerraLib Includes
#include <terralib/core/translator/Translator.h>

// STL Includes
#include <string>
#include <vector>

// GeoDMA Includes
#include "geodmalib/datamanager/datasetIterator.hpp"


gdma::dm::DatasetIterator::DatasetIterator(const te::da::DataSetPtr dataset) {
    m_dataset = dataset;
    m_current = 0;
    m_endPos = dataset->size();
}


gdma::dm::DatasetIterator::~DatasetIterator() {
}

unsigned int gdma::dm::DatasetIterator::currentPosition() {
    return m_current;
}

gdma::dm::DatasetIterator gdma::dm::DatasetIterator::begin(const te::da::DataSetPtr dataset) {
    return gdma::dm::DatasetIterator(dataset);
}

gdma::dm::DatasetIterator gdma::dm::DatasetIterator::end(const te::da::DataSetPtr dataset) {
    gdma::dm::DatasetIterator it(dataset);
    it.m_current = dataset->size();
    return it;
}

bool gdma::dm::DatasetIterator::operator!=(const DatasetIterator &rhs) const {
    return this->m_current != rhs.m_current;
}

bool gdma::dm::DatasetIterator::operator==(const DatasetIterator &rhs) const {
    return this->m_current == rhs.m_current;
}

void gdma::dm::DatasetIterator::operator++() {
    m_current++;
}

double gdma::dm::DatasetIterator::operator[](const unsigned int pos) const {
    if (!m_dataset->move(m_current))
        throw te::common::Exception(TE_TR("Position Inexistent in the DataSet!"));

    return m_dataset->getDouble(pos);
}

double gdma::dm::DatasetIterator::operator[](const std::string featName) const {
    if (!m_dataset->move(m_current))
        throw te::common::Exception(TE_TR("Position Inexistent in the DataSet!"));

    return m_dataset->getDouble(featName);
}

std::string gdma::dm::DatasetIterator::operator()(const unsigned int pos) const {
    if (!m_dataset->move(m_current))
        throw te::common::Exception(TE_TR("Position Inexistent in the DataSet!"));

    return m_dataset->getAsString(pos);
}

std::string gdma::dm::DatasetIterator::operator()(const std::string featName) const {
    if (!m_dataset->move(m_current))
        throw te::common::Exception(TE_TR("Position Inexistent in the DataSet!"));

    return m_dataset->getAsString(featName);
}

gdma::dm::DatasetIterator gdma::dm::DatasetIterator::operator=(
    const DatasetIterator &rhs) {
    if (this != &rhs) {
        m_dataset = rhs.m_dataset;
        m_current = rhs.m_current;
        m_endPos = rhs.m_endPos;
    }

    return *this;
}

const std::vector<double> gdma::dm::DatasetIterator::operator*() const {
    if (!m_dataset->move(m_current))
        throw te::common::Exception(TE_TR("Position Inexistent in the DataSet!"));

    std::vector<double> values;
    for (unsigned int featPos = 0; featPos < m_dataset->getNumProperties(); featPos++)
        values.push_back(m_dataset->getDouble(featPos));

    return values;
}
