/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/datamanager/datasetIterator.hpp

  \brief This file contains the DatasetIterator class definition. It iterates sequentially over
         all the positions of a given dataset.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

#ifndef GEODMA_DM_DATASETITERATOR_HPP_
#define GEODMA_DM_DATASETITERATOR_HPP_

// TerraLib Includes
#include <terralib/dataaccess/dataset/DataSet.h>

// STL Includes
#include <string>
#include <vector>

// GeoDMA Includes
#include "geodmalib/datamanager/config.hpp"

namespace gdma {
namespace dm {
class GDMADMEXPORT DatasetIterator {
 public:
     /*!
         \brief Constructor
         \param[in] dataset The dataset to be iterated.
     */
     explicit DatasetIterator(const te::da::DataSetPtr dataset);

     /*! \brief Default Destructor */
     ~DatasetIterator();

     /*!
         \brief Retrieve the current position of the iterator in the dataset.
         \return Returns the current position of the iterator in the dataset.
     */
     unsigned int currentPosition();

     /*!
         \brief Creates an iterator refering to the first position to be iterated in the dataset.
         \param[in] dataset The dataset to be iterated.
         \return Returns the iterator.
     */
     static DatasetIterator begin(const te::da::DataSetPtr dataset);

     /*!
         \brief Creates an iterator refering to after the last position of the dataset.
         \param[in] dataset The dataset to be iterated.
         \return Returns the iterator.
     */
     static DatasetIterator end(const te::da::DataSetPtr dataset);

     /*!
         \brief Difference operator.
         \param[in] rhs The right-hand side object to be compared.
         \return Returns true if the iterators are in different positions, false otherwise.
     */
     bool operator!=(const DatasetIterator &rhs) const;

     /*!
         \brief Equal operator.
         \param[in] rhs The right-hand side object to be compared.
         \return Returns true if the iterators are in the same position, false otherwise.
     */
     bool operator==(const DatasetIterator &rhs) const;

     /*!
         \brief Increment position operator. It moves the iterator to the next position to be iterated.
     */
     void operator++();

     /*!
         \brief Returns the double value of the feature \ref pos in the current position of the
                iterator.
         \param[in] pos The position of the feature.
         \return The value of the feature as double.
         \note For string features, use \ref operator()
     */
     double operator[](const unsigned int pos) const;

     /*!
         \brief Returns the double value of the feature \ref featName in the current position of the
                iterator.
         \param[in] featName The string containing the name of the feature.
         \return The value of the feature as double.
         \note For string features, use \ref operator()
     */
     double operator[](const std::string featName) const;

     /*!
         \brief Returns the value of the feature \ref pos as string in the current position of the
                iterator.
         \param[in] pos The position of the feature.
         \return The value of the feature as string.
         \note For numeric features, use \ref operator[]
     */
     std::string operator()(const unsigned int pos) const;

     /*!
         \brief Returns the value of the feature \ref featName as string in the current position of
                the iterator.
         \param[in] featName The string containing the name of the feature.
         \return The value of the feature as string.
         \note For numeric features, use \ref operator[]
     */
     std::string operator()(const std::string featName) const;

     /*!
         \brief Assignment operator.
         \param[in] rhs The right-hand side object to be copyied.
         \return Returns a reference to the new object.
     */
     DatasetIterator operator=(const DatasetIterator &rhs);

     /*! 
         \brief Returns a vector of the values (features) in current position from iterator.
     */
     const std::vector<double> operator*() const;

 private:
     te::da::DataSetPtr m_dataset;                      //!< Pointer to the dataset to be iterated
     std::size_t m_current;                             //!< Current position of the iterator
     std::size_t m_endPos;  //!< The end position (position after the last position of the dataset).
};
}  // namespace dm
}  // namespace gdma

#endif  // GEODMA_DM_DATASETITERATOR_HPP_
