/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/datamanager/datasetRandomParcialIterator.hpp

  \brief This file contains the DatasetRandomParcialIterator class definition. It iterates over a
         percentage of the dataset items randomly predefined.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// TerraLib Includes
#include <terralib/core/translator/Translator.h>

// STL Includes
#include <iostream>
#include <memory>
#include <set>
#include <string>
#include <vector>

// GeoDMA Includes
#include "geodmalib/datamanager/datasetParcialIterator.hpp"


gdma::dm::DatasetParcialIterator::DatasetParcialIterator(const te::da::DataSetPtr dataset,
    const std::shared_ptr<std::vector<unsigned int>> positions) {
    m_dataset = dataset;
    m_positions = positions;

    if (m_positions->size() > m_dataset->size())
        throw te::common::Exception(TE_TR("Positions vector have more positions than the DataSet!"));

    m_current = 0;
    m_endPos = dataset->size();
}


gdma::dm::DatasetParcialIterator::~DatasetParcialIterator() {
}

unsigned int gdma::dm::DatasetParcialIterator::currentPosition() {
    if (m_current < m_positions->size())
        return (*m_positions.get())[m_current];
    else
        return m_endPos;
}

gdma::dm::DatasetParcialIterator gdma::dm::DatasetParcialIterator::begin(
    const te::da::DataSetPtr dataset, const std::shared_ptr<std::vector<unsigned int>> positions) {
    return gdma::dm::DatasetParcialIterator(dataset, positions);
}

gdma::dm::DatasetParcialIterator gdma::dm::DatasetParcialIterator::end(
    const te::da::DataSetPtr dataset, const std::shared_ptr<std::vector<unsigned int>> positions) {
    gdma::dm::DatasetParcialIterator it(dataset, positions);

    it.m_current = dataset->size();

    return it;
}

bool gdma::dm::DatasetParcialIterator::operator!=(const DatasetParcialIterator &rhs) const {
    return this->m_current != rhs.m_current;
}

bool gdma::dm::DatasetParcialIterator::operator==(const DatasetParcialIterator &rhs) const {
    return this->m_current == rhs.m_current;
}

void gdma::dm::DatasetParcialIterator::operator++() {
    if (m_current < (m_positions->size() - 1)) {
        m_current++;
    } else {
        m_current = m_endPos;
    }
}

double gdma::dm::DatasetParcialIterator::operator[](const unsigned int pos) const {
    if (!m_dataset->move((*m_positions.get())[m_current]))
        throw te::common::Exception(TE_TR("Position Inexistent in the DataSet!"));

    return m_dataset->getDouble(pos);
}

double gdma::dm::DatasetParcialIterator::operator[](const std::string featName) const {
    if (!m_dataset->move((*m_positions.get())[m_current]))
        throw te::common::Exception(TE_TR("Position Inexistent in the DataSet!"));

    return m_dataset->getDouble(featName);
}

std::string gdma::dm::DatasetParcialIterator::operator()(const unsigned int pos) const {
    if (!m_dataset->move((*m_positions.get())[m_current]))
        throw te::common::Exception(TE_TR("Position Inexistent in the DataSet!"));

    return m_dataset->getAsString(pos);
}

std::string gdma::dm::DatasetParcialIterator::operator()(const std::string featName) const {
    if (!m_dataset->move((*m_positions.get())[m_current]))
        throw te::common::Exception(TE_TR("Position Inexistent in the DataSet!"));

    return m_dataset->getAsString(featName);
}

gdma::dm::DatasetParcialIterator gdma::dm::DatasetParcialIterator::operator=(
    const DatasetParcialIterator &rhs) {
    if (this != &rhs) {
        m_positions = rhs.m_positions;
        m_dataset = rhs.m_dataset;
        m_current = rhs.m_current;
        m_endPos = rhs.m_endPos;
    }

    return *this;
}

const std::vector<double> gdma::dm::DatasetParcialIterator::operator*() const {
    if (!m_dataset->move((*m_positions.get())[m_current]))
        throw te::common::Exception(TE_TR("Position Inexistent in the DataSet!"));

    std::vector<double> values;
    for (unsigned int featPos = 0; featPos < m_dataset->getNumProperties(); featPos++)
        values.push_back(m_dataset->getDouble(featPos));

    return values;
}
