/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/featureextraction/attribute.hpp

  \brief A structure to hold the informations about features.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

#ifndef GEODMA_LIB_FEATUREEXTRACTION_ATTRIBUTE_HPP_
#define GEODMA_LIB_FEATUREEXTRACTION_ATTRIBUTE_HPP_

// TerraLib Includes
#include <terralib/datatype.h>

// STL Includes
#include <string>

// GeoDMA Includes
#include "geodmalib/featureextraction/config.hpp"

namespace gdma {
namespace fte {

class GDMAFTEEXPORT Attribute {
 private:
    std::string m_name;
    std::string m_shortName;
    int m_valueType;

    std::string m_stringValue;
    int16_t m_int16Value;
    int32_t m_int32Value;
    int64_t m_int64Value;
    float m_floatValue;
    double m_doubleValue;

 public:
    /*! \brief Default constructor. */
    Attribute();

    /*! 
      \brief Constructor to store string values.
      \param name Name of the Attribute.
      \param shortName Short name of the Attribute.
      \param stringValue Value to be stored.
    */
    Attribute(std::string name, std::string shortName, std::string stringValue);

    /*!
      \brief Constructor to store 16 bit int values.
      \param name Name of the Attribute.
      \param shortName Short name of the Attribute.
      \param int16Value Value to be stored.
    */
    Attribute(std::string name, std::string shortName, int16_t int16Value);

    /*!
      \brief Constructor to store 32 bit int values.
      \param name Name of the Attribute.
      \param shortName Short name of the Attribute.
      \param int32Value Value to be stored.
    */
    Attribute(std::string name, std::string shortName, int32_t int32Value);

    /*!
      \brief Constructor to store 64 bit int values.
      \param name Name of the Attribute.
      \param shortName Short name of the Attribute.
      \param int64Value Value to be stored.
    */
    Attribute(std::string name, std::string shortName, int64_t int64Value);

    /*!
      \brief Constructor to store float values.
      \param name Name of the Attribute.
      \param shortName Short name of the Attribute.
      \param floatValue Value to be stored.
    */
    Attribute(std::string name, std::string shortName, float floatValue);

    /*!
      \brief Constructor to store double precision float values.
      \param name Name of the Attribute.
      \param shortName Short name of the Attribute.
      \param doubleValue Value to be stored.
    */
    Attribute(std::string name, std::string shortName, double doubleValue);

    /*! 
      \brief Copy Constructor.
    */
    Attribute(const Attribute &rhs);

    /*! \brief Default destructor. */
    ~Attribute();

    /*!
      \brief Method that retrieve the attribute name.
    */
    std::string getName() const;

    /*!
      \brief Method that set the attribute name.
      \param name The new name for the attribute.
    */
    void setName(std::string name);

    /*!
      \brief Method that retrieve the short name of the attribute.
    */
    std::string getShortName() const;

    /*!
      \brief Method that set the short name of the attribute.
      \param name The new name for the attribute.
    */
    void setShortName(std::string name);

    /*!
      \brief Method that retrieve the type of the value stored.
      \note The returned value is a terralib basic data types (See terralib data type module).
    */
    int getValueType() const;

    /*!
      \brief Method that retrieve the attribute value as string.
    */
    std::string getString() const;

    /*!
      \brief Method that set the string attribute value.
      \param value The new value for the attribute.
    */
    bool setString(std::string value);

    /*!
      \brief Method that retrieve the attribute value as int 16 bit.
    */
    int16_t getInt16() const;

    /*!
      \brief Method that set the int 16 bit attribute value.
      \param value The new value for the attribute.
    */
    bool setInt16(int16_t value);

    /*!
      \brief Method that retrieve the attribute value as int 16 bit.
    */
    int32_t getInt32() const;

    /*!
      \brief Method that set the int 16 bit attribute value.
      \param value The new value for the attribute.
    */
    bool setInt32(int32_t value);

    /*!
      \brief Method that retrieve the attribute value as int 16 bit.
    */
    int64_t getInt64() const;

    /*!
      \brief Method that set the int 16 bit attribute value.
      \param value The new value for the attribute.
    */
    bool setInt64(int64_t value);

    /*!
      \brief Method that retrieve the attribute value as float.
    */
    float getFloat() const;

    /*!
      \brief Method that set the float attribute value.
      \param value The new value for the attribute.
    */
    bool setFloat(float value);

    /*!
      \brief Method that retrieve the attribute value as double.
    */
    double getDouble() const;

    /*!
      \brief Method that set the double attribute value.
      \param value The new value for the attribute.
    */
    bool setDouble(double value);

    /*!
      \brief Method that retrieve the any numeric attribute value converted as double (casted convertion).
    */
    double getAsDouble() const;

 private:
    std::string sendShortNameWarning(std::string shortName);
};
}  // end namespace fte
}  // end namespace gdma

#endif  // GEODMA_LIB_FEATUREEXTRACTION_ATTRIBUTE_HPP_
