/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/featureextraction/featuresSummary.hpp

  \brief A structure to hold the set of features.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

#ifndef GEODMA_LIB_FEATURESEXTRACTION_FEATURESSUMMARY_HPP_
#define GEODMA_LIB_FEATURESEXTRACTION_FEATURESSUMMARY_HPP_

// TerraLib Includes
#include <terralib/statistics/core/NumericStatisticalSummary.h>
#include <terralib/rp/Texture.h>

// STL Includes
#include <map>
#include <string>
#include <vector>

// GeoDMA Includes
#include "geodmalib/featureextraction/config.hpp"
#include "geodmalib/featureextraction/attribute.hpp"

namespace gdma {
namespace fte {
/*!
 \struct NumericStatisticalSummary

 \brief A structure to hold the set of values of the features.
 */
class GDMAFTEEXPORT FeaturesSummary {
 public:
    /*!
      \brief Constructor.
    */
    FeaturesSummary();

    /*!
      \brief Copy Constructor.
    */
    FeaturesSummary(const FeaturesSummary &rhs);

    /*! 
      \brief Destructor
    */
    ~FeaturesSummary();

    /*!
      \brief Assignment operator.
      \param rhs The right-hand-side copy used to copy from.
      \return A reference to the object.
    */
    FeaturesSummary& operator=(const FeaturesSummary &rhs);

    /*!
      \brief Clear the structure.
    */
    void clear();

    /*! 
      \brief ToString method. 
      \return A string with the features. 
    */
    std::string toString();

    /*!
      \brief Add a new string feature to the summary.
      \param name Name of the feature
      \param shortName Short name, used to save feature in Shape files (Max 10 characters)
      \param value Value to be stored
    */
    void addFeature(std::string name, std::string shortName, std::string value);

    /*!
      \brief Add a new 16 bit int feature to the summary.
      \param name Name of the feature
      \param shortName Short name, used to save feature in Shape files (Max 10 characters)
      \param value Value to be stored
    */
    void addFeature(std::string name, std::string shortName, int16_t value);

    /*!
      \brief Add a new 32 bit int feature to the summary.
      \param name Name of the feature
      \param shortName Short name, used to save feature in Shape files (Max 10 characters)
      \param value Value to b e stored
    */
    void addFeature(std::string name, std::string shortName, int32_t value);

    /*!
      \brief Add a new 64 bit int feature to the summary.
      \param name Name of the feature
      \param shortName Short name, used to save feature in Shape files (Max 10 characters)
      \param value Value to be stored
    */
    void addFeature(std::string name, std::string shortName, int64_t value);

    /*!
      \brief Add a new float feature to the summary.
      \param name Name of the feature
      \param shortName Short name, used to save feature in Shape files (Max 10 characters)
      \param value Value to be stored
    */
    void addFeature(std::string name, std::string shortName, float value);

    /*!
      \brief Add a new double feature to the summary.
      \param name Name of the feature
      \param shortName Short name, used to save feature in Shape files (Max 10 characters)
      \param value Value to be stored
    */
    void addFeature(std::string name, std::string shortName, double value);

    /*!
      \brief Add statistical features from a NumericStatisticalSummary.
      \param statSummary The NumericStatisticalSummary (see te::stat::NumericStatisticalSummary)
      \param band The number of the raster band from where the statistics were extracted.
      \param noDataValue Value used for no data. Default is -1
    */
    void addStatistics(te::stat::NumericStatisticalSummary statSummary, int band,
        double noDataValue = -1);

    /*!
      \brief Add texture features from a te::rp::Texture.
      \param textures The Texture object (see te::rp::Texture)
      \param band The number of the raster band from where the textures were computed.
      \param direction The direction of the GLCM offset. For example, if you shift it 1 pixel right
                       and 1 pixel left, use "SE" (South eastern). This parameter will only compose 
                       the name of the attribute. Default value is an empty string. 
                       \note Use maximum two characters to describe the direction. 
    */
    void addTextures(te::rp::Texture textures, int band, std::string direction = "");


    /*!
      \brief Add features from a std::map<std::string, double>. It is mainly used for landscape
             ecology metrics.
      \param map_features The map containing the features to be added, indexed by the class names.
      \param featureName The name of the feature to be stored.
    */
    void addFeaturesFromMap(std::map<std::string, double> &map_features, std::string featureName,
        std::map<std::string, int> &mapClassNames);

    /*!
      \brief Add features from a std::map<std::string, int>. It is mainly used for landscape
             ecology metrics.
      \param map_features The map containing the features to be added, indexed by the class names.
      \param featureName The name of the feature to be stored.
    */
    void addFeaturesFromMap(std::map<std::string, int> &map_features, std::string featureName,
        std::map<std::string, int> &mapClassNames);

    /*!
      \brief Returns a vector with all features stored.
    */
    const std::vector<gdma::fte::Attribute>& getFeatures() const;

 private:
    std::vector<gdma::fte::Attribute> m_features;
};
}  // end namespace fte
}  // end namespace gdma

#endif /* defined(GEODMA_LIB_FEATURESEXTRACTION_FEATURESSUMMARY_HPP_) */
