/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/featureextraction/intersectionCache.hpp

  \brief Class to store intersections between two layers, to make it faster to search.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

#ifndef GEODMA_LIB_FEATURESEXTRACTION_INTERSECTIONCACHE_HPP_
#define GEODMA_LIB_FEATURESEXTRACTION_INTERSECTIONCACHE_HPP_

// TerraLib Includes
#include <terralib/geometry/Geometry.h>
#include <terralib/sam/rtree/Index.h>  // R-tree

// Boost Includes
#include <boost/bimap.hpp>
#include <boost/numeric/ublas/matrix_sparse.hpp>

// STL Includes
#include <map>
#include <memory>
#include <set>
#include <string>
#include <vector>

// GeoDMA Includes
#include "geodmalib/featureextraction/config.hpp"

namespace gdma {
namespace fte {

typedef std::vector<boost::numeric::ublas::compressed_vector<te::gm::GeometryShrPtr>> IntersectionMatrix;

typedef boost::numeric::ublas::compressed_vector<te::gm::GeometryShrPtr> MatrixLine;

/*!
  \class IntersectionCache

  \brief A class to compute intersection between two sets of geometries and create a fast access
         cache.

  \ingroup featextraction
*/
class GDMAFTEEXPORT IntersectionCache {
 public:
    /*!
      \brief Constructor.
      \param[in] layer_1 First set of geometries. It must be a vector containing pointers to
                         geometries.
      \param[in] layer1_IDs Boost bimap linking the position of the geometries of layer_1 with
                            their corresponding ID.
      \param[in] layer_2 Second set of geometries. It must be a vector containing pointers to
                         geometries.
      \param[in] layer2_IDs Boost bimap linking the position of the geometries of layer_2
                            with their corresponding ID.
    */
    IntersectionCache(const std::vector<te::gm::Geometry*> &layer1,
        const boost::bimap<std::string, std::size_t> &layer1IDs,
        const std::vector<te::gm::Geometry*> &layer2,
        const boost::bimap<std::string, std::size_t> &layer2IDs);

    /*! \brief Default destructor. */
    ~IntersectionCache();

    /*!
      \brief Given a geometry position in Layer 1, returns all of its intersections.
      \param[in] geomPositionLayer1 The position of the geometry in the container layer_1.
      \param[out] matrixLine A sparse vector containing all intersections.
      \return Returns true if the intersection was successfully executed, or false if there are
              invalid geometries.
    */
    bool searchForIntersections(std::size_t geomPositionLayer1, MatrixLine& matrixLine);

    /*!
      \brief Given a geometry position in Layer 1, returns true if the intersection was sucessfully
             executed or false if there are invalid geometries.
      \param[in] geomPositionLayer1 The position of the geometry in the container layer_1.
      \return Returns true if the intersection was successfully executed, or false if there are
              invalid geometries.
    */
    bool isInvalid(std::size_t geomPositionLayer1);

 private:
    //NOLINT
    gdma::fte::IntersectionMatrix m_intersection_matrix;  //!< An sparse matrix to store intersections
    te::sam::rtree::Index<std::size_t, 8, 4> m_rtree;  //!< A R-Tree to compute intersections.
    std::set<std::size_t> m_invalidgeometriesLayer1;   //!< A set with the invalid geometries
};
}  // namespace fte
}  // namespace gdma


#endif /* #ifndef GEODMA_LIB_FEATURESEXTRACTION_INTERSECTIONCACHE_HPP_ */
