/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/featureextraction/landscapeFeatures.hpp

  \brief Functionalities to compute Landscape Ecology features.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

#ifndef GEODMA_LIB_FEATURESEXTRACTION_LANDSCAPEFEATURES_HPP_
#define GEODMA_LIB_FEATURESEXTRACTION_LANDSCAPEFEATURES_HPP_

// TerraLib Includes
#include <terralib/geometry.h>
#include <terralib/sam/rtree/Index.h>  // R-tree

// Boost Includes
#include <boost/bimap.hpp>

// STL Includes
#include <map>
#include <set>
#include <string>
#include <utility>
#include <vector>

// GeoDMA Includes
#include "geodmalib/featureextraction/config.hpp"
#include "geodmalib/featureextraction/featuresSummary.hpp"
#include "geodmalib/featureextraction/intersectionCache.hpp"

namespace gdma {
namespace fte {

typedef std::multimap<std::pair<std::string, std::string>,
        std::pair<double, te::gm::Polygon* >> IntersectionsContainer;

class GDMAFTEEXPORT LandscapeFeatures {
 public:
    /*! \brief Default constructor. */
    // LandscapeFeatures();

    /*!
      \brief Constructor to initialize the LandscapeFeatures with the vector of polygons.
      \param[in] polsCtnerCoarserL The container of landscape polygons that will be the target of the
                                   computed features.
      \param[in] positionsIdsCoarserL A boost::bimap structure relating the IDs of the polygons in the
                                      dataset to the corresponding position of the polygon in the \ref
                                      polsCtnerCoarserL.
      \param[in] polsCtnerFinerL The container of classified polygons from where the features will
                 be computed.
      \param[in] positionsIdsFinerL A boost::bimap structure relating the IDs of the polygons in the
                                    dataset to the corresponding position of the polygon in the \ref
                                    polsCtnerFinerL.
      \param[in] classesFinerL The container of strings containing the classes of the corresponding
                               polingons in \ref polsCtnerFinerL.
      \param[in] classValuesFinerL The set containing the names of the existing classes without
                                   duplicates.
    */
    LandscapeFeatures(const std::vector<te::gm::Geometry*> &polsCtnerCoarserL,
        const boost::bimap<std::string, std::size_t> &positionsIdsCoarserL,
        const std::vector<te::gm::Geometry*> &polsCtnerFinerL,
        const boost::bimap<std::string, std::size_t> &positionsIdsFinerL,
        const std::vector<std::string> &classesFinerL, const std::set<std::string> &classValuesFinerL,
        const double dummyValueDouble = -1);

    /*! \brief Default Destructor. */
    ~LandscapeFeatures();

    /*!
      \brief Method to compute the Landscape Ecology based features.
      \param[out] attr_container A reference to a std::map<std::string, gdma::fte::FeaturesSummary>,
                                 that will be the output container, where the computed features
      \exception te::srs::Exception if the Geometries (cells and polygons) are not Polygons or
                                    Multipolygons.
      \exception te::srs::Exception if the Polygons and the Cells are not in the same spatial
                                    reference system.
    */
    bool getFeatures(std::map<std::string, gdma::fte::FeaturesSummary> &attr_container);

    /*!
      \brief Method to compute the Class Area feature. This feature means the total area
             of a given class that inside a given object.
      \param[in] cellPos The position of the target geometry (cell or polygon) in the input vector
                         (see \ref LandscapeFeatures constructor) to compute the metrics.
      \param[out] classAreas A reference to a map<string, double>, indexed by the name of the
                             existing classes. This map will contain the value of the feature for
                             each class. 
    */
    void getClassArea(std::size_t cellPos, std::map<std::string, double> &classAreas);

    /*!
      \brief Method to compute the Percent Land feature. This feature corresponds to the sum of
             the areas of all patches of the corresponding class (class area), divided by total
             landscape area. Percent land is equals to the percentage the landscape comprised of the
             corresponding class.
      \param[in] cellPos The position of the target geometry (cell or polygon) in the input vector
                         (see \ref LandscapeFeatures constructor) to compute the metrics.
      \param[in] classAreas A reference to a map<string, double>, indexed by the name of the
                             existing classes. This map must contain the value of the class area
                             feature for each class. (See \ref getClassArea)
      \param[out] percent_lands A reference to a map<string, double>, indexed by the name of the
                                existing classes. This map will contain the value of the computed
                                feature for each class.
    */
    void getPercentLand(std::size_t cellPos, std::map<std::string, double> &class_areas,
        std::map<std::string, double> &percent_lands);

    /*!
      \brief Method to compute the Patch Density (PD) feature, which is equals the number of patches
             in the landscape, divided by total landscape area (\f$m^2\f$), multiplied by 10,000 and
             100 (to convert to 100 hectares). Note, PD does not include background patches or
             patches in the landscape border, if present. However, total landscape area (\f$A\f$)
             includes any internal background present. Thus, it is computed through:
             \f$PD = \frac{\sum_{j=1}^n a_j}{A} \times 100\f$
      \param[in] cellPos The position of the target geometry (cell or polygon) in the input vector
                         (see \ref LandscapeFeatures constructor) to compute the metrics.
      \param[out] patch_densities A reference to a map<string, double>, indexed by the name of the
                                  existing classes. This map will contain the value of the computed
                                  feature for each class.
    */
    void getPatchDensities(std::size_t cellPos, std::map<std::string, double> &patch_densities);

    /*!
      \brief Method to compute the Mean Patch Size (MPS) feature, which is equals to the sum of the
             are of all patches of the corresponding patch type, divided by the number of patches of
             the same type. Thus, it is computed through
             \f$MPS = \frac{\sum_{j=1}^n a_j}{n} 10^{-4}\f$
      \param[in] cellPos The position of the target geometry (cell or polygon) in the input vector
                         (see \ref LandscapeFeatures constructor) to compute the metrics.
      \param[in] numberOfPatches A reference to a map<string, int>, indexed by the name of the
                                 existing classes. This map must contain the number of patches of
                                 each class inside the target geometry.
                                 (See \ref getNumberOfPatches)
      \param[out] mean_patch_sizes A reference to a map<string, double>, indexed by the name of the
                                   existing classes. This map will contain the value of the computed
                                   feature for each class.
    */
    void getMeanPatchSizes(std::size_t cellPos, std::map<std::string, int> &numberOfPatches,
        std::map<std::string, double> &mean_patch_sizes);

    /*!
      \brief Method to compute the Patch Size Standard Deviation (PSSD) feature, which is the root
             mean squared error (deviation from the mean) in patch size. This is the population
             standard deviation, not the sample standard deviation. Thus, it is computed through:
             \f$PSSD = \sqrt{\frac{\sum_{j=1}^n \left(a_j - MPS \right)^2}{n}} 10^{-4}\f$
      \param[in] cellPos The position of the target geometry (cell or polygon) in the input vector
                         (see \ref LandscapeFeatures constructor) to compute the metrics.
      \param[in] numberOfPatches A reference to a map<string, int>, indexed by the name of the
                                 existing classes. This map must contain the number of patches of
                                 each class inside the target geometry.
                                 (See \ref getNumberOfPatches)
      \param[in] mean_patch_sizes A reference to a map<string, double>, indexed by the name of the
                                  existing classes. This map must contain the mean patch size
                                  feature of each class inside the target geometry.
                                  (See \ref getMeanPatchSizes)
      \param[out] patch_sizes_std A reference to a map<string, double>, indexed by the name of the
                                  existing classes. This map will contain the value of the computed
                                  feature for each class.
    */
    void getPatchSizesStd(std::size_t cellPos, std::map<std::string, double> &mean_patch_sizes,
       std::map<std::string, int> &numberOfPatches, std::map<std::string, double> &patch_sizes_std);

    /*!
      \brief Method to compute the Landscape Shape Index (LSI) feature, which is equals the sum of
             the landscape boundary and all edge segments (\f$m\f$) within the boundary. This sum
             involves the corresponding patch type (including borders), divided by the two times the
             square root of the total landscape area (\f$m^2\f$) multiplied by pi (\f$\pi\f$). Thus,
             it is computed through:
             \f$LSI = \frac{\sum_{j=1}^n e_j}{2\sqrt{\pi \times A}}\f$
      \param[in] cellPos The position of the target geometry (cell or polygon) in the input vector
                         (see \ref LandscapeFeatures constructor) to compute the metrics.
      \param[out] landscapeShapeIndexes A reference to a map<string, double>, indexed by the name of
                                        the existing classes. This map will contain the value of the
                                        computed feature for each class.
    */
    void getLandscapeShapeIndexes(std::size_t cellPos,
        std::map<std::string, double> &landscapeShapeIndexes);

    /*!
      \brief Method to compute the Mean Shape Index (MSI) feature, which is equals the sum of the
             patch perimeter (\f$m\f$) by divided two times the square root of patch area
             (\f$m^2\f$) multiplied by pi (\f$\pi\f$) for each patch of the corresponding patch
             type, divided by the the number of patches of the same patch type (class). Thus, it is
             computed through:
             \f$MSI = \frac{\sum_{j=1}^n \frac{p_j}{2 \times \sqrt{\pi \times a_j}}}{n}\f$
      \param[in] cellPos The position of the target geometry (cell or polygon) in the input vector
                         (see \ref LandscapeFeatures constructor) to compute the metrics.
      \param[in] numberOfPatches A reference to a map<string, int>, indexed by the name of the
                                 existing classes. This map must contain the number of patches of
                                 each class inside the target geometry.
                                 (See \ref getNumberOfPatches)
      \param[out] meanShapeIndexes A reference to a map<string, double>, indexed by the name of
                                   the existing classes. This map will contain the value of the
                                   computed feature for each class.
    */
    void getMeanShapeIndexes(std::size_t cellPos, std::map<std::string, int> &numberOfPatches,
        std::map<std::string, double> &meanShapeIndexes);

    /*!
      \brief Method to compute the Area-Weighted Mean Shape Index (MSI) feature, which is equals the
             sum of the landscape boundary and all edge segments (\f$m\f$) within the boundary. This
             sum involves the corresponding patch type (including borders), divided by the two times
             the square root of the total landscape area (\f$m^2\f$) multiplied by pi (\f$\pi\f$).
             This first term is multiplied by the area of the corresponding patch, divided by the
             sum of the areas of all patches of the same patch type (class). Thus, it is computed
             through:
             \f$AWMSI = \sum_{j=1}^n \left[ \frac{p_j}{2 \sqrt{\pi \times a_j}} \times 
             \frac{a_j}{\sum_{j=1}^n a_j} \right]\f$
      \param[in] cellPos The position of the target geometry (cell or polygon) in the input vector
                         (see \ref LandscapeFeatures constructor) to compute the metrics.
      \param[in] classAreas A reference to a map<string, double>, indexed by the name of the
                             existing classes. This map must contain the value of the class area
                             feature for each class. (See \ref getClassArea)
      \param[out] awMeanShapeIndexes A reference to a map<string, double>, indexed by the name of
                                     the existing classes. This map will contain the value of the
                                     computed feature for each class.
    */
    void getAWMeanShapeIndexes(std::size_t cellPos, std::map<std::string, double> &class_areas,
        std::map<std::string, double> &awMeanShapeIndexes);

    /*!
      \brief Method to compute the Mean Patch Fractal Dimension (MPFD) feature, that is computed
             through:
             \f$MPFD = \frac{\sum_{j=1}^{n} \frac{2 \times \ln{p_j}}{\ln{a_j}}}{N}\f$
      \param[in] cellPos The position of the target geometry (cell or polygon) in the input vector
                         (see \ref LandscapeFeatures constructor) to compute the metrics.
      \param[in] numberOfPatches A reference to a map<string, int>, indexed by the name of the
                                 existing classes. This map must contain the number of patches of
                                 each class inside the target geometry.
                                 (See \ref getNumberOfPatches)
      \param[out] mpfds A reference to a map<string, double>, indexed by the name of the existing
                        classes. This map will contain the value of the computed feature for each
                        class.
    */
    void getMeanPatchFractalDimensions(std::size_t cellPos,
        std::map<std::string, int> numberOfPatches, std::map<std::string, double> &mpfds);

    /*!
      \brief Method to compute the Mean Patch Fractal Dimension (MPFD) feature, that is computed
             through:
             \f$AWMPFD = \sum_{j = 1}^{n} [\frac{2 \times \ln{p_j}}{\ln{a_j}} \times \frac{a_j}
             {\sum_{j = 1}^{n} a_{ij}}]\f$
      \param[in] cellPos The position of the target geometry (cell or polygon) in the input vector
                         (see \ref LandscapeFeatures constructor) to compute the metrics.
      \param[in] classAreas A reference to a map<string, double>, indexed by the name of the
                            existing classes. This map must contain the value of the class area
                            feature for each class. (See \ref getClassArea)
      \param[out] awmpfds A reference to a map<string, double>, indexed by the name of the existing
                          classes. This map will contain the value of the computed feature for each
                          class.
    */
    void getAreaWeightedMeanPatchFractalDimensions(std::size_t cellPos,
        std::map<std::string, double> &class_areas, std::map<std::string, double> &awmpfds);

    /*!
      \brief Method to compute the Edge Density (ED) feature, which is equals the sum of the lengths
             (\f$m\f$) of all edge segments involving the corresponding patch type, divided by the
             total landscape area (\f$m^2\f$). Thus, it is computed through:
             \f$ED = \frac{\sum_{j=1}^m e_j}{A} 10^{4}\f$
      \param[in] cellPos The position of the target geometry (cell or polygon) in the input vector
                         (see \ref LandscapeFeatures constructor) to compute the metrics.
      \param[in] class_perimeters A reference to a map<string, double>, indexed by the name of the
                                  existing classes. This map must contain the value of the total
                                  edges feature for each class. (See \ref getTotalEdges)
      \param[out] edgeDensities A reference to a map<string, double>, indexed by the name of the
                                existing classes. This map will contain the value of the computed
                                feature for each class.
    */
    void getEdgeDensities(std::size_t cellPos, std::map<std::string, double> &class_perimeters,
        std::map<std::string, double> &edgeDensities);

    /*!
      \brief Method to compute the Total Edge (TE) feature, which is equals the total size of the
             edges of the all patches of the given patch type (class). Thus, it is computed through:
             $TE = \sum_{j=0}^n e_j$
      \param[in] cellPos The position of the target geometry (cell or polygon) in the input vector
                         (see \ref LandscapeFeatures constructor) to compute the metrics.
      \param[out] total_edges A reference to a map<string, double>, indexed by the name of the
                              existing classes. This map will contain the value of the computed
                              feature for each class.
    */
    void getTotalEdges(std::size_t cellPos, std::map<std::string, double> &total_edges);

    /*!
      \brief Method to compute the Mean Perimeter Area Ratio (MPAR), which is equals the sum of
             ratios between perimeters and areas, divided by the number of patches of the same type.
             Thus, it is computed through:
             \f$MPAR = \frac{ \sum_{j=1}^n \frac{p_j}{a_j}}{n}\f$
      \param[in] cellPos The position of the target geometry (cell or polygon) in the input vector
                         (see \ref LandscapeFeatures constructor) to compute the metrics.
      \param[in] numberOfPatches A reference to a map<string, int>, indexed by the name of the
                                 existing classes. This map must contain the number of patches of
                                 each class inside the target geometry.
                                 (See \ref getNumberOfPatches)
      \param[out] meanPerimAreaRatios A reference to a map<string, double>, indexed by the name of
                                      the existing classes. This map will contain the value of the
                                      computed feature for each class.
    */
    void getMeanPerimeterAreaRatios(std::size_t cellPos,
        std::map<std::string, int> &numberOfPatches,
        std::map<std::string, double> &meanPerimAreaRatios);

    /*!
      \brief Method to compute the Number of Patches (NP) feature, which is equals to the number of
             patches of a corresponding patch type (class) inside the target geometry.
      \param[in] cellPos The position of the target geometry (cell or polygon) in the input vector
                         (see \ref LandscapeFeatures constructor) to compute the metrics.
      \param[out] mean_patch_sizes A reference to a map<string, int>, indexed by the name of the
                                   existing classes. This map will contain the value of the computed
                                   feature for each class.
    */
    void getNumberOfPatches(std::size_t cellPos, std::map<std::string, int> &numbers_of_patches);

    /*!
      \brief Method to compute the Patch Size Coefficient of Variation (PSCOV) feature, which
             calculates the ratio between the features Patch Sizes Standard Deviation (PSSD - see
             \ref getPatchSizesStd) and Mean Patch Size (MPS - see \ref getMeanPatchSizes). Thus, it
             is computed through:
             \f$PSCOV = \frac{PSSD}{MPS} \times 100\f$
      \param[in] mean_patch_sizes A reference to a map<string, double>, indexed by the name of the
                                 existing classes. This map must contain the values of mean patch
                                 sizes (MPS) of each class inside the target geometry.
                                 (See \ref getMeanPatchSizes)
      \param[in] patch_sizes_std A reference to a map<string, double>, indexed by the name of the
                                 existing classes. This map must contain the values of patch sizes
                                 standard deviation feature (PSSD) of each class inside the target
                                 geometry. (See \ref getPatchSizesStd)
      \param[out] psCoeffVariation A reference to a map<string, double>, indexed by the name of
                                   the existing classes. This map will contain the value of the
                                   computed feature for each class.
    */
    void getPatchSizeCoeffVariation(std::map<std::string, double> &mean_patch_sizes,
        std::map<std::string, double> &patch_sizes_std,
        std::map<std::string, double> &psCoeffVariation);

    /*!
      \brief Method to compute the Interspersion and Juxtaposition Index (IJI) feature. It consists
             in the observed interspersion over the maximum possible interspersion for the given
             number of patch types. It only exists for \f$n > 3\f$. Thus, it is computed through:
             \f$IJI = \frac{-\sum_{j=1}^n (\frac{e_j}{\sum_{k=1}^n e_k}) \times
             \ln{(\frac{e_j}{\sum_{k=1}^n e_k})}}{\ln(n - 1)} \times 100\f$
      \param[in] cellPos The position of the target geometry (cell or polygon) in the input vector
                         (see \ref LandscapeFeatures constructor) to compute the metrics.
      \param[in] total_edges A reference to a map<string, double>, indexed by the name of the
                             existing classes. This map must contain the value of the total
                             edges feature for each class. (See \ref getTotalEdges)
      \param[in] patchRichness The value of the Patch Richness feature. (See \ref getPatchRichness).
      \param[out] ijIndexes A reference to a map<string, double>, indexed by the name of the
                            existing classes. This map will contain the value of the computed
                            feature for each class.
    */
    void getInterspersionAndJuxtapositionIndex(std::size_t cellPos,
        std::map<std::string, double> &total_edges, int patchRichness,
        std::map<std::string, double> &ijIndexes);

    /*!
      \brief Method to compute the Total Area of the Biggest Object (TABO) feature. It consists in
             the area (\f$ha\f$) of the biggest object that intersects the landscape area.
      \param[in] cellPos The position of the target geometry (cell or polygon) in the input vector
                         (see \ref LandscapeFeatures constructor) to compute the metrics.
      \param[out] tabo A reference to a map<string, double>, indexed by the name of the
                       existing classes. This map will contain the value of the computed
                       feature for each class.
    */
    void getTotalAreaBiggestObject(std::size_t cellPos, std::map<std::string, double> &tabo);

    /*!
      \brief Method to compute the Biggest Intersection Area (BIA) and the Total Area of the Object
             with Biggest Intersection Area (TAOBIA). BIA consists in the area (\f$ha\f$) of the
             biggest intersection with the landscape area. TAOBIA consists in the total area
             (\f$ha\f$) of the object that has the biggest area of intersection;
      \param[in] cellPos The position of the target geometry (cell or polygon) in the input vector
                         (see \ref LandscapeFeatures constructor) to compute the metrics.
      \param[out] bia A reference to a map<string, double>, indexed by the name of the
                       existing classes. This map will contain the value of the computed
                       feature BIA for each class.
      \param[out] taobia A reference to a map<string, double>, indexed by the name of the
                         existing classes. This map will contain the value of the computed
                         feature TAOBIA for each class.
    */
    void getBiggestIntersectionAreas(std::size_t cellPos, std::map<std::string, double> &bia,
        std::map<std::string, double> &taobia);

    /*!
      \brief Method to compute the Patch Richness (PR) feature, which is equals the number of
             different patch types (classes) present within the landscape boundary. \f$PR = m\f$
      \param[in] cellPos The position of the target geometry (cell or polygon) in the input vector
                         (see \ref LandscapeFeatures constructor) to compute the metrics.
      \return An integer containing the value of the computed feature.
    */
    int getPatchRichness(std::size_t cellPos);

    /*!
      \brief Method to compute the Patch Richness Density (PRD) feature, which is equals the number
             of different patch types present within the landscape boundary divided by total
             landscape area (\f$m^2\f$), multiplied by 10000 and 100 (to convert to 100 hectares).
             Note, total landscape area (\f$A\f$) includes any internal background present. Thus,
             it is computed through: \f$PRD = \frac{m}{A} \times 10000 \times 100\f$
      \param[in] cellPos The position of the target geometry (cell or polygon) in the input vector
                         (see \ref LandscapeFeatures constructor) to compute the metrics.
      \param[in] patchRichness The value of the Patch Richness feature. (See \ref getPatchRichness).
      \return A double containing the value of the computed feature.
    */
    double getPatchRichnessDensity(std::size_t cellPos, int &patchRichness);

    /*!
      \brief Method to compute the Shannon's Diversity Index (SHDI), which is equals to minus the
             sum, across all patch types, of the proportional abundance of each patch type
             multiplied by that proportion. Note, \f$P_i\f$, which is the proportion of the
             landscape occupied by patch type (class) \f$i\f$, is based on total landscape area
             (\f$A\f$) excluding any internal background present. It is computed through:
             \f$SHDI = -\sum_{i = 0}^{m} P_i \times \ln{P_i}\f$
      \param[in] cellPos The position of the target geometry (cell or polygon) in the input vector
                         (see \ref LandscapeFeatures constructor) to compute the metrics.
      \param[in] percent_land A reference to a map<string, double>, indexed by the name of the
                              existing classes. This map will contain the value of the proportion
                              of the target geometry area occupied by the each class. (See
                              \ref getPercentLand)
    */
    double getShannonsDiversityIndex(std::size_t cellPos,
        std::map<std::string, double> &percent_land);

    /*!
      \brief Method to compute the Simpson's Diversity Index (SIDI), which is equals to 1 minus the
             sum, across all patch types, of the proportional abundance of each patch type squared.
             Note, \f$P_i\f$, which is the proportion of the landscape occupied by patch type
             (class) \f$i\f$, is based on total landscape area (\f$A\f$) excluding any internal
             background present. It is computed through: \f$SIDI = 1 - \sum_{i = 0}^{m} P_i^2\f$
      \param[in] cellPos The position of the target geometry (cell or polygon) in the input vector
                         (see \ref LandscapeFeatures constructor) to compute the metrics.
      \param[in] percent_land A reference to a map<string, double>, indexed by the name of the
                              existing classes. This map will contain the value of the proportion
                              of the target geometry area occupied by the each class. (See
                              \ref getPercentLand)
    */
    double getSimpsonsDiversityIndex(std::size_t cellPos,
        std::map<std::string, double> &percent_land);

    /*!
      \brief Method to compute the Shannon's Evenness Index (SHEI), which is equals to minus the
             sum, across all patch types, of the proportional abundance of each patch type
             multiplied by that proportion, divided by the logarithm of the number of patch types.
             In other words, the observed Shannon's Diversity Index (SHDI) divided by the maximum
             Shannon's Diversity Index for that number of patch types. Note, \f$P_i\f$, which is the
             proportion of the landscape occupied by patch type (class) \f$i\f$, is based on total
             landscape area (\f$A\f$) excluding any internal background present. It is computed
             through:
             \f$SHEI = \frac{-\sum_{i = 0}^{m} P_i \times \ln{P_i}}{\ln(m)}\f$
      \param[in] shannonDivIndex A reference to a double that contains the value of the Shannon's
                                 Diversity Index (SHDI) for the current geometry.
                                 (See \ref getShannonsDiversityIndex)
      \param[in] patchRichness A reference to a int that contains the value of the Patch Richness
                               of the current geometry, i. e., the number of different classes
                               present within its boundary. (See \ref getPatchRichness)
    */
    double getShannonsEvennessIndex(double &shannonDivIndex, int &patchRichness);

    /*!
      \brief Method to compute the Simpson's Evenness Index, which is equals to 1 minus the sum,
             across all patch types, of the proportional abundance of each patch type squared,
             divided by 1 minus 1 divided by the number of patch types. In other words, the observe
             Simpson's Diversity Index (SIDI) divided by the maximum Simpson's Diversity Index for
             that number of patch types. Note, \f$P_i\f$, which is the proportion of the landscape
             occupied by patch type (class) \f$i\f$, is based on total landscape area (\f$A\f$)
             excluding any internal background present. It is computed through:
             \f$SIEI = \frac{1 - \sum_{i = 0}^{m} P_i^2}{1 - (\frac{1}{m})}\f$
      \param[in] simpsonDivIndex A reference to a double that contains the value of the Simpsons's
                                 Diversity Index (SIDI) for the current geometry.
                                 (See \ref getSimpsonsDiversityIndex)
      \param[in] patchRichness A reference to a int that contains the value of the Patch Richness
                               of the current geometry, i. e., the number of different classes
                               present within its boundary. (See \ref getPatchRichness)
    */
    double getSimpsonsEvennessIndex(double &simpsonDivIndex, int &patchRichness);

    const std::map<std::string, int> &getMapClassNames();

 protected:
    /*!
      \brief Method to compute the area and perimeter of a vector of geometries of the type Polygon.
      \param[in] geometryVector The vector of geometries to compute area and perimeter.
      \param[out] area The computed area.
      \param[out] perimeter The computed perimeter.
    */
    void computePatchAreaAndPerimeter(std::vector<te::gm::Geometry*> &geometryVector, double &area,
        double &perimeter);

    /*!
      \brief Method to compute the area of a vector of geometries of the type Polygon.
      \param[in] geometryVector The vector of geometries to compute the area.
      \param[out] area The computed area.
    */
    void computePolygonArea(std::vector<te::gm::Geometry*> &geometryVector, double &area);

    /*!
      \brief Method to compute the perimeter of a vector of geometries of the type Polygon.
      \param[in] geometryVector The vector of geometries to compute the perimeter.
      \param[out] perimeter The computed area.
    */
    void computePolygonPerimeter(std::vector<te::gm::Geometry*> &geometryVector, double &perimeter);

    /*!
      \brief Method to retrieve the dummy value that will be used for invalid geometries.
      \return The dummy value;
    */
    double getDummyValue();

    /*!
      \brief Method to set the dummy value that will be used for invalid geometries.
      \param[in] value The new dummy value.
    */
    void setDummyValue(double value);

 private:
    const std::vector<te::gm::Geometry*> *m_pols_ctner_finerL;
    const boost::bimap<std::string, std::size_t> *m_positionsIdsFinerL;
    const std::vector<std::string> *m_classes_finerL;
    const std::set<std::string> *m_class_values_finerL;
    std::map<std::string, int> m_class_map_codes;

    const boost::bimap<std::string, std::size_t> *m_positionsIdsCoarserL;
    const std::vector<te::gm::Geometry*> *m_pols_ctner_coarserL;

    double m_dummyValueDouble;
    int m_dummyValueInt;

    gdma::fte::IntersectionCache *m_intersectionCache;
};
}  // namespace fte
}  // namespace gdma

#endif  // GEODMA_LIB_FEATURESEXTRACTION_LANDSCAPEFEATURES_HPP_
