/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/featureextraction/spatialFeatures.hpp

  \brief Functionalities to compute segmentation-based spatial features for regions based on 
  geometries.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

#ifndef GEODMA_LIB_FEATUREEXTRACTION_SPATIALFEATURES_HPP_
#define GEODMA_LIB_FEATUREEXTRACTION_SPATIALFEATURES_HPP_

// TerraLib includes
#include <terralib/geometry.h>
#include <terralib/raster.h>
#include <terralib/common/progress/TaskProgress.h>

// BOOST - REMOVE IT
#include <boost/numeric/ublas/matrix.hpp>

// STL includes
#include <map>
#include <string>
#include <vector>
#include <mutex>
#include <thread>

// GeoDMA Includes
#include "geodmalib/featureextraction/config.hpp"
#include "geodmalib/featureextraction/featuresSummary.hpp"

namespace gdma {
namespace fte {
class GDMAFTEEXPORT SpatialFeatures {
 public:
    /*! \brief Default constructor. */
    SpatialFeatures();

    /*! \brief Default Destructor. */
    ~SpatialFeatures();

    /*!
      \brief Method that compute the Spatial Features.
      \param[in] inputRaster Input raster to be used to compute the features.
      \param[in] gm_container A Map container (std::map<std::string, te::gm::Geometry*>) of
                              geometries (Polygons or Multipolygons) containing the polygons that
                              will be used to compute the features, indexed by a string composed by
                              the primary key of the geometry in the source dataset. This container
                              can be generated by \ref gdma::dm::getGeometriesFromShp.
      \param[in] bands Vector with the numbers of the bands that will be used to compute the
                       features.
      \param[out] attr_container Reference to a Map container
                                 (std::map<std::string, gdma::fte::FeaturesSummary>), where
                                 the computed features will be stored.
      \param [in] useMultiThreads If true, multiple threads/cores will be used.
      \param[in] enableRasterCache Boolean that indicates if the raster will be cached or not.
                                   Setting this parameter to "true" can make huge rasters processing
                                   faster, reducing the number of accesses to the disk. Default
                                   value is "true".
      \return A boolean indicating if the execution was succefull of if it failed.
      \exception te::srs::Exception if the Geometry is not a Polygon or Multipolygon.
    */
    static bool getFeatures(const std::map<std::string, te::gm::Geometry*> gm_container,
        std::map<std::string, gdma::fte::FeaturesSummary> &attr_container,
        const bool useMultiThreads,
        const double dummyValueDouble );

    /*!
      \brief Computes the area of a polygon (feature POL_AREA) measured in the unit of measure of
             the current Spatial Reference System.
      \param[in] polygon A pointer to a polygon object.
      \return A double value corresponding to the polygon area.
    */
    static double computeArea(te::gm::Polygon* polygon);

    /*!
      \brief Computes the area of the bounding box of a polygon (feature POL_BBOX_AREA) measured in
             the unit of measure of the current Spatial Reference System.
      \param[in] pbox A pointer to an Envelope (bounding box of a polygon) object.
      \return A double value corresponding to the area of the bounding box of the polygon.
    */
    static double computePBoxArea(te::gm::Envelope const * const pbox);

    /*!
      \brief Computes the perimeter of a polygon (feature POL_PERIMETER) measured in
             the unit of measure of the current Spatial Reference System.
      \param[in] polygon A pointer to a polygon object.
      \return A double value corresponding to the polygon perimeter.
    */
    static double computePerimeter(te::gm::Polygon* polygon);

    /*!
      \brief Computes the fractal dimension of a polygon (Feature POL_FRACTALDIM)
      \param[in] area A double value corresponding to the area of the corresponding polygon.
      \param[in] perimeter A double value corresponding to the perimeter of the corresponding
                           polygon.
      \return A double value corresponding to the fractal dimension of the corresponding polygon.
    */
    static double computeFractalDim(double area, double perimeter);

    /*!
      \brief Computes the Height of the bounding box of a polygon (Feature POL_LENGTH) measured in
             the unit of measure of the current Spatial Reference System.
      \param[in] boundingBox A pointer to an Envelope (bounding box of a polygon) object.
      \return A double value corresponding to the height of the bounding box of the polygon.

    */
    static double computePBoxLength(te::gm::Envelope const * const boundingBox);

    /*!
      \brief Computes the Width of the bounding box of a polygon (Feature POL_WIDTH) measured in
             the unit of measure of the current Spatial Reference System.
      \param[in] boundingBox A pointer to an Envelope (bounding box of a polygon) object.
      \return A double value corresponding to the width of the bounding box of the polygon.

    */
    static double computePBoxWidth(te::gm::Envelope const * const boundingBox);

    /*!
      \brief Computes the ratio between the perimeter and the area of the polygon (Feature
             POL_PERIM_AREA_RATIO)
      \param[in] area A double value corresponding to the area of the corresponding polygon.
      \param[in] perimeter A double value corresponding to the perimeter of the corresponding
                           polygon.
      \return A double value corresponding to the ratio between the area and the perimeter of the
              corresponding polygon.
    */
    static double computePerimAreaRatio(double area, double perimeter);

    /*!
      \brief Computes the compacity of the polygon (Feature POL_COMPACITY).
      \param[in] area A double value corresponding to the area of the corresponding polygon.
      \param[in] perimAreaRatio A double value corresponding to the ratio between the perimeter and
                                the area of the corresponding polygon.
      \return A double value corresponding to the compacity of the corresponding polygon.
    */
    static double computeCompacity(double area, double perimAreaRatio);

    /*!
      \brief Computes the coefficients of the minimum ellipse that circunscribes the given polygon.
      \param[in] polygon A pointer to a polygon geometry.
      \param[out] coefs A vector of double values corresponding to the ellipse equation
                        coefficients in the form \f$coefs[0]x^2 + coefs[1]xy + coefs[2]y^2 +
                        coefs[3]x + coefs[4]y + coefs[5] = 0\f$.
      \param[out] offsetX The Offset applied in the polygon in the X axis to generate the ellipse
                          equation. This opperation is necessary to translate the polygon to the
                          positive quadrant. This value must be used as parameter to the method
                          \ref createEllipsePolygon to generate the ellipse geometry in the correct
                          position.
      \param[out] offsetY The Offset applied in the polygon in the Y axis to generate the ellipse
                          equation. This opperation is necessary to translate the polygon to the
                          positive quadrant. This value must be used as parameter to the method
                          \ref createEllipsePolygon to generate the ellipse geometry in the correct
                          position.

      \note Opperation performed by CGAL.
      \sa solveEllipse and createEllipsePolygon.
    */
    static bool fitEllipse(te::gm::Polygon const * const polygon, std::vector<double> &coefs, double &offsetX,
        double &offsetY);

    /*!
      \brief Computes the charachteristics of the minimum ellipse that circunscribes the polygon.
      \param[in] coefs A vector containing the coefficients of the Ellipse equation, as generated by
                       \ref fitEllipse method.
      \param[out] minor_radius The Minor Radius of the ellipse.
      \param[out] major_radius The Major Radius of the ellipse.
      \param[out] centerX The X coordinate of the ellipse center.
      \param[out] centerY The Y coordinate of the ellipse center.
      \param[out] angle The inclination angle of the Ellipse, i.e., the angle between the major axis
                        of the Ellipse and the X axis of the coordinate system.

      \sa fitEllipse and createEllipsePolygon.
    */
    static bool solveEllipse(std::vector<double> coefs, double &minor_radius, double &major_radius,
        double &centerX, double &centerY, double &angle);

    /*!
      \brief Generates a geometry (polygon) for the ellipse that circunscribes the polygon.
      \param[in] minor_radius The Minor Radius of the ellipse.
      \param[in] major_radius The Major Radius of the ellipse.
      \param[in] centerX The X coordinate of the ellipse center.
      \param[in] centerY The Y coordinate of the ellipse center.
      \param[in] angle The inclination angle of the Ellipse, i.e., the angle between the major axis
                       of the Ellipse and the X axis of the coordinate system.
      \param[in] srid The SRID of the geometry that will be generated.
      \param[in] offsetX The Offset applied in the polygon in the X axis by the method
                         \ref fitEllipse to generate the ellipse equation.
      \param[in] offsetY The Offset applied in the polygon in the Y axis by the method
                         \ref fitEllipse to generate the ellipse equation.
      \return A polygon corresponding to the minimum ellipse that circunscribes the polygon.
      
      \note The caller of this method will take the ownership of the returned Geometry.
      \sa fitEllipse and solveEllipse.
    */
    static te::gm::Polygon* createEllipsePolygon(double minor_radius, double major_radius,
        double centerX, double centerY, double angle, int srid = 0, double offsetX = 0,
        double offsetY = 0);

    /*!
      \brief Computes the Elliptic Fit of the polygon (Feature POL_ELLIPTIC_FIT). It is the ratio
             between the polygon's area and the area of its minimum circunscribing ellipse.
      \param[in] ellipseArea The area of the polygon's minimum circunscribing ellipse.
      \param[in] polygonArea The area of the given polygon.
    
      \return The ratio between the polygon area and the ellipse area.

      \note The ellipse area (A) can be computed by \f$ A = \pi * major_radius * minor_radius\f$.
            It is not necessary to generate the ellipse geometry to compute it.
      \sa fitEllipse and solveEllipse.
    */
    static double computeEllipticFit(double ellipseArea, double polygonArea);

    /*!
      \brief Computes the Gyration Radius feature (Feature POL_GYRATION_RATIUS). This feature is
             equals the average distance between each vertex of the polygon and it's centroid. The
             more similar to a circle is the object, the more likely the centroid will be inside it,
             and therefore this feature will be closer to 0.
      \param[in] polygon A pointer to the polygon.
      \param[out] gyrRadius The value of the Gyration Radius Attribute.
      \param[out] maxDistance The Maximum distance computed. This value is the radius of the
                              polygon, that will be used to compute the polygon circle (method
                              \ref computePolygonCircle).
    */
    static bool computeGyrationRadius(te::gm::Polygon* polygon, double &gyrRadius,
        double &maxDistance);

    /*!
      \brief Computes the ratio between the area of the minimum circle that circunscribes the
             polygon and the area of the polygon (feature POL_CIRCLE).
      \param[in] radius The radius of the polygon (maximum distance between the centroid and the
                        vertices of the polygon). It is computed by the method
                        \ref computeGyrationRadius.
      \param[in] polArea The area of the polygon.
      \return A double value corresponding to the ratio between the area of the minimum circle that
              circunscribes the polygon and the area of the polygon.
    */
    static double computePolygonCircle(double radius, double polArea);

    /*!
      \brief Computes the Polygon Shape Index feature (feature POL_SHAPE_INDEX). This feature
             corresponds to the ratio between the polygon perimeter and the squared root of the
             polygon area.
      \param[in] polPerimeter The polygon perimeter.
      \param[in] polArea The polygon area
      \returns A double value corresponding to the computed feature.
    */
    static double computePolShapeIndex(double polPerimeter, double polArea);

    /*!
      \brief Computes the Polygon Density (feature POL_DENSITY). This feature corresponds to the
             ratio between the polygon area and the polygon radius. 
      \param[in] polRadius The radius of the polygon (maximum distance between the centroid and the
                           vertices of the polygon). It is computed by the method
                           \ref computeGyrationRadius.
      \param[in] polArea The polygon area.
      \return A double value corresponding to the polygon density.
    */
    static double computePolygonDensity(double polRadius, double polArea);

    /*!
      \brief Computes the minimum rectangle outside the object.
      \param[in] polygon The current polygon.
      \return A polygon corresponding to the minimum enclosing rectangle.
      \note The caller of this method will take the ownership of the returned Geometry.
    */
    static te::gm::Polygon* fitRectangle(te::gm::Polygon* polygon);

    /*!
      \brief Computes the Rectangular fit (Feature POL_RECTANGULAR_FIT). This feature fits a minimum
             rectangle outside the object and calculates the ratio between its area and the area of
             this rectangle. The closer to 1 is this feature, the most similar to a rectangle. This
             rectangle can be generated by the method \ref fitRectangle.
       \param[in] polArea The area of the current polygon.
       \param[out] rectArea The area of the corresponding rectangle.
       \return A double value corresponding to the value of the computed feature.
       \sa fitRectangle
    */
    static double computeRectangularFit(double polArea, double rectArea);

    // Implement polygon contiguity
 protected :
    /*!
      \brief Thread entry that compute the Spatial Features.
      \param[in] geomsItPtr A pointer to Input geometries container iterator.
      \param[in] geomsItEnd A pointer to Input geometries container iterator end.
      \param[out] attr_container A pointer to a Map container
                                 (std::map<std::string, gdma::fte::FeaturesSummary>), where
                                 the computed features will be stored.
      \param dummyValueDouble A pointer to Attributes dummy value.
      \param mutex A pointer to a global mutex to syncronize threads
      \return A pointer to a boolean indicating if the execution was succefull of if it failed.
      \exception te::srs::Exception if the Geometry is not a Polygon or Multipolygon.
    */  
      static void getFeaturesThreadEntry(
          std::map<std::string, te::gm::Geometry*>::const_iterator * const geomsItPtr,
          std::map<std::string, te::gm::Geometry*>::const_iterator const * const geomsItEndPtr,
          std::map<std::string, gdma::fte::FeaturesSummary>* attrContainerPtr,
          double const * const dummyValueDoublePtr,
          bool * const returnValuePtr,
          std::mutex * const mutexPtr,
          te::common::TaskProgress * const taskProgressPtr );
};
}  // end namespace fte
}  // end namespace gdma

#endif  // GEODMA_LIB_FEATUREEXTRACTION_SPATIALFEATURES_HPP_
