/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/featureextraction/spectralFeatures.cpp

  \brief Functionalities to compute segmentation-based spectral features for regions based on 
  geometries.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

#include <time.h>  // This include is to calculate execution time (REMOVE IT)

// TerraLib includes
#include <terralib/common/progress/TaskProgress.h>
#include <terralib/geometry.h>
#include <terralib/memory/CachedRaster.h>
#include <terralib/raster.h>
#include <terralib/rp/RasterAttributes.h>
#include <terralib/rp/Texture.h>
#include <terralib/statistics/core/NumericStatisticalSummary.h>
// #include <terralib/vp/Utils.h>

// Boost Includes
#include <boost/chrono.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <boost/numeric/ublas/matrix.hpp>

// STL includes
#include <map>
#include <vector>
#include <utility>
#include <string>

// GeoDMA includes
#include "geodmalib/featureextraction/multiGeometryAlgorithms.hpp"
#include "geodmalib/featureextraction/featuresSummary.hpp"
#include "geodmalib/featureextraction/spectralFeatures.hpp"
#include "geodmalib/common/commandLineProgBar.hpp"

// Default constructor
gdma::fte::SpectralFeatures::SpectralFeatures() { }

// Default destructor
gdma::fte::SpectralFeatures::~SpectralFeatures() { }

bool gdma::fte::SpectralFeatures::getFeatures(const te::rst::Raster* inputRaster,
    const std::map<std::string, te::gm::Geometry*> gm_container,
    const std::vector<unsigned int> bands,
    std::map<std::string, gdma::fte::FeaturesSummary> &attr_container,
    const bool enableRasterCache, double noDataValue) {
    // Dealing with cache issues
    std::auto_ptr<te::rst::Raster> cachedRasterManager;
    te::rst::Raster const * cachedRasterPtr = inputRaster;

    // Enabling CachedRaster
    if (enableRasterCache) {
        cachedRasterManager.reset(new te::mem::CachedRaster(*inputRaster, 40, 0));
        cachedRasterPtr = cachedRasterManager.get();
    }

    if (gm_container.size() < 1) {
        throw te::common::Exception(TE_TR("Geometry container received is empty."));
    }

    te::common::TaskProgress task(TE_TR("Computing Spectral Features"));
    task.setTotalSteps((bands.size()*gm_container.size()));
    task.useTimer(true);

#if defined GDMA_PERFORMANCE_STAT
    boost::chrono::microseconds sumTimeGLCM = boost::chrono::microseconds{ 0 };
    boost::chrono::microseconds sumTimeStatistics = boost::chrono::microseconds{ 0 };
#endif

    std::map<std::string, std::vector<double> > means;
    for (int countBand = 0; countBand < bands.size(); countBand++) {
        // Retrieving the maximum and minimum GL values of the band.
        double minPixel, maxPixel;
        te::rst::GetDataTypeRanges(cachedRasterPtr->getBandDataType(countBand), minPixel,
            maxPixel);

        if ((maxPixel - minPixel) > 255) {
            minPixel = cachedRasterPtr->getBand(bands[countBand])->getMinValue(true).real();
            maxPixel = cachedRasterPtr->getBand(bands[countBand])->getMaxValue(true).real();
        }
        //

        std::map<std::string, te::gm::Geometry*>::const_iterator itGm = gm_container.begin();
        while (itGm != gm_container.end()) {
            if (attr_container.find(itGm->first) == attr_container.end()) {
                gdma::fte::FeaturesSummary featSummary;

                attr_container.insert(std::pair<std::string, gdma::fte::FeaturesSummary>(
                    itGm->first, featSummary));
            }

            std::vector<te::gm::Geometry*> auxPolVec;

            gdma::fte::Multi2SilgleFilterPolygonsRestrict(itGm->second, auxPolVec);

            // Verify the SRID
            if (itGm->second->getSRID() != cachedRasterPtr->getSRID()) {
                std::ostringstream output;

                output << "Geometry received have an Spatial Reference different from the "
                    "Input Raster received!" << std::endl;
                output << "\tSRID Raster: " << cachedRasterPtr->getSRID() << std::endl;
                output << "\tSRID Geometry: " << itGm->second->getSRID() << std::endl;

                throw te::common::Exception(TE_TR(output.str()));
                return false;
            }

            // for (int countPol = 0; countPol < auxPolVec.size(); countPol++) {
                te::gm::Polygon* m_polygon = static_cast<te::gm::Polygon*>(auxPolVec[0]);

#if defined GDMA_PERFORMANCE_STAT
                boost::chrono::high_resolution_clock::time_point tStart =
                    boost::chrono::high_resolution_clock::now();
#endif

                te::stat::NumericStatisticalSummary statistics;
                te::rp::RasterAttributes tl_RasterAttr(false);

                tl_RasterAttr.getStatisticsFromPolygon(*cachedRasterPtr,
                    bands[countBand], *m_polygon, statistics, noDataValue);

#if defined GDMA_PERFORMANCE_STAT
                boost::chrono::high_resolution_clock::time_point tEnd =
                    boost::chrono::high_resolution_clock::now();

                sumTimeStatistics +=
                    boost::chrono::duration_cast<boost::chrono::microseconds>(tEnd - tStart);

                tStart = boost::chrono::high_resolution_clock::now();
#endif

                boost::numeric::ublas::matrix<double> glcm_matrix =
                    tl_RasterAttr.getGLCM(*cachedRasterPtr, bands[countBand], 1, 1,
                        *m_polygon, minPixel, maxPixel);

                te::rp::Texture textures = tl_RasterAttr.getGLCMMetrics(glcm_matrix, noDataValue);

#if defined GDMA_PERFORMANCE_STAT
                tEnd = boost::chrono::high_resolution_clock::now();

                sumTimeGLCM +=
                    boost::chrono::duration_cast<boost::chrono::microseconds>(tEnd - tStart);
#endif

                attr_container[itGm->first].addStatistics(statistics, bands[countBand], noDataValue);
                attr_container[itGm->first].addTextures(textures, bands[countBand], "SE");


                means[itGm->first].push_back(statistics.m_mean);

                bool isNoData = false;
                if (countBand == bands.size() - 1) {
                    double sum_mean = 0;
                    for (int j = 0; j < bands.size(); j++) {
                        if (means[itGm->first][j] != noDataValue)
                            sum_mean += means[itGm->first][j];
                        else
                            isNoData = true;
                    }
                    for (int j = 0; j < bands.size(); j++) {
                        std::string attrName = "BAND_RATIO_BAND_" + std::to_string(j);
                        std::string attrShortName = "BRATIO_B" + std::to_string(j);
                        double bandRatio;
                        if (!isNoData)
                            bandRatio = means[itGm->first][j] / sum_mean;
                        else
                            bandRatio = noDataValue;
                        attr_container[itGm->first].addFeature(attrName, attrShortName, bandRatio);
                    }
                }
            // }
            // progressBar.pulse();
            if (task.isActive() == false)
                throw te::common::Exception(TE_TR("Operation canceled!"));

            task.pulse();
            itGm++;
        }
    }

#if defined GDMA_PERFORMANCE_STAT
    std::cout << "---------------------------------------" << std::endl;

    std::cout << "Average Time to compute Statistical Metrics: " <<
        (sumTimeStatistics / (bands.size()*gm_container.size())) << "." << std::endl;
    std::cout << "Total Time to compute Statistical Metrics: " << sumTimeStatistics
        << "." << std::endl;
    std::cout << "----" << std::endl;
    std::cout << "Average Time to compute GLCM Metrics: " <<
        (sumTimeGLCM / (bands.size()*gm_container.size())) << "." << std::endl;
    std::cout << "Total Time to compute GLCM Metrics: " << sumTimeGLCM << "." <<
        std::endl;
#endif
    return true;
}
