/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/featureextraction/spectralFeatures.cpp

  \brief Functionalities to compute segmentation-based spectral features for regions based on 
  geometries.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

#ifndef GEODMA_LIB_FEATURESEXTRACTION_SPECTRALFEATURES_HPP_
#define GEODMA_LIB_FEATURESEXTRACTION_SPECTRALFEATURES_HPP_

// TerraLib includes
#include <terralib/geometry.h>
#include <terralib/raster.h>

// STL includes
#include <map>
#include <string>
#include <vector>

// GeoDMA Includes
#include "geodmalib/featureextraction/config.hpp"
#include "geodmalib/featureextraction/featuresSummary.hpp"

namespace gdma {
namespace fte {

class GDMAFTEEXPORT SpectralFeatures {
 public:
    /*! \brief Default constructor. */
    SpectralFeatures();

    /*! \brief Default Destructor. */
    ~SpectralFeatures();

    /*! 
      \brief Method that compute the Spectral Features.
      \param[in] inputRaster A pointer to the Raster object that will be used to compute the
                             features.
      \param[in] gm_container A Map container (std::map<std::string, te::gm::Geometry*>) of
                              geometries (Polygons or Multipolygons) containing the polygons that
                              will be used to compute the features, indexed by a string composed by
                              the primary key of the geometry in the source dataset. This container
                              can be generated by \ref gdma::dm::getGeometriesFromShp.
      \param[in] bands Vector with the numbers of the bands that will be used to compute the
                       features.
      \param[out] attr_container Reference to a Map container
                                 (std::map<std::string, gdma::fte::FeaturesSummary>), where
                                 the computed features will be stored.
      \param[in] enableRasterCache Boolean that indicates if the raster will be cached or not.
                                   Setting this parameter to "true" can make huge rasters processing
                                   faster, reducing the number of accesses to the disk. Default
                                   value is "true".
      \param[in] nodataValue The value that will be used when the polygon does not intersects the
                             raster. Default value is 0.0.
      \return A boolean indicating if the execution was succefull of if it failed.
      \exception te::srs::Exception if the Geometry is not a Polygon or Multipolygon.
      \exception te::srs::Exception if the Geometry and the Raster are not in the same spatial
                                    reference system.
    */
    static bool getFeatures(const te::rst::Raster* inputRaster,
        const std::map<std::string, te::gm::Geometry*> gm_container,
        const std::vector<unsigned int> bands,
        std::map<std::string, gdma::fte::FeaturesSummary> &attr_container,
        const bool enableRasterCache = true, double noDataValue = -1000);
};
}  // end namespace fte
}  // end namespace gdma

#endif  // GEODMA_LIB_FEATURESEXTRACTION_SPECTRALFEATURES_HPP_
