/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/guicomponents/c50ClassificationController.hpp

  \brief This file contains the SampleSelectionWidget class definition.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// TerraLib Includes
#include <terralib/core/translator/Translator.h>
#include <terralib/dataaccess/datasource/DataSource.h>
#include <terralib/dataaccess/datasource/DataSourceInfo.h>
#include <terralib/dataaccess/dataset/DataSetType.h>
#include <terralib/dataaccess/dataset/PrimaryKey.h>
#include <terralib/dataaccess/utils/Utils.h>
#include <terralib/maptools/AbstractLayer.h>
#include <terralib/qt/widgets/layer/utils/DataSet2Layer.h>

// STL Includes
#include <fstream>
#include <list>
#include <map>
#include <memory>
#include <string>
#include <vector>

// GeoDMA Includes
#include "geodmalib/common/messageMediator.hpp"
#include "geodmalib/datamanager/dataManager.hpp"
#include "geodmalib/classification/decision_trees/c50DecisionTree.hpp"
#include "geodmalib/guicomponents/layerUtils.hpp"
#include "geodmalib/guicomponents/decision_trees/c50ClassificationController.hpp"

gdma::guicmp::C50ClassificationController::C50ClassificationController() :
    m_c50DecisionTree() {
    m_newLayerCreated = false;
}


gdma::guicmp::C50ClassificationController::~C50ClassificationController() {
}

void gdma::guicmp::C50ClassificationController::setCurrentLayer(te::map::AbstractLayerPtr layer) {
    m_currentLayer = layer;

    std::string dataSourceId = m_currentLayer->getDataSourceId();
    std::string dataSetName = m_currentLayer->getDataSetName();

    m_c50DecisionTree.setInputDataSet(dataSourceId, dataSetName);
}

te::map::AbstractLayerPtr gdma::guicmp::C50ClassificationController::getCurrentLayer() {
    return m_currentLayer;
}

void gdma::guicmp::C50ClassificationController::setLayerList(
    std::list<te::map::AbstractLayerPtr> layerList) {
    m_layerList = layerList;
}

std::list<te::map::AbstractLayerPtr> gdma::guicmp::C50ClassificationController::getLayerList() {
    return m_layerList;
}

void gdma::guicmp::C50ClassificationController::setLabelsColumn(std::string name) {
    m_c50DecisionTree.setTrainFeature(name);
}

std::string gdma::guicmp::C50ClassificationController::getLabelsColumn() {
    return m_c50DecisionTree.getTrainFeature();
}

void gdma::guicmp::C50ClassificationController::setAvailableFeatures() {
    std::unique_ptr<const te::map::LayerSchema> schema(m_currentLayer->getSchema());
    std::vector<te::dt::Property*> propVector = schema->getProperties();
    te::da::PrimaryKey* pk = schema->getPrimaryKey();
    std::vector<te::dt::Property*> pkVector = pk->getProperties();

    m_availableFeatures.clear();

    for (auto &prop : propVector) {
        if (prop->getType() != te::dt::GEOMETRY_TYPE && prop->getName().compare("id") != 0 &&
            prop->getName().compare("col") != 0 && prop->getName().compare("row") != 0 &&
            prop->getName().compare(m_c50DecisionTree.getTrainFeature()) != 0) {
            for (auto &key : pkVector) {
                if (key->getName().compare(prop->getName()) != 0)
                    m_availableFeatures.push_back(prop->getName());
            }
        }
    }
}

std::vector<std::string> gdma::guicmp::C50ClassificationController::getAvailableFeatures() {
    return m_availableFeatures;
}

void gdma::guicmp::C50ClassificationController::setTrials(int trials) {
    m_c50DecisionTree.getParameters().m_trials = trials;
}

int gdma::guicmp::C50ClassificationController::getTrials() {
    return m_c50DecisionTree.getParameters().m_trials;
}

void gdma::guicmp::C50ClassificationController::setGlobalPruning(bool globalPruning) {
    m_c50DecisionTree.getParameters().m_noGlobalPruning = !globalPruning;
}

bool gdma::guicmp::C50ClassificationController::getGlobalPruning() {
    return !m_c50DecisionTree.getParameters().m_noGlobalPruning;
}

void gdma::guicmp::C50ClassificationController::setMinCases(int minCases) {
    m_c50DecisionTree.getParameters().m_minCases = minCases;
}

int gdma::guicmp::C50ClassificationController::getMinCases() {
    return m_c50DecisionTree.getParameters().m_minCases;
}

void gdma::guicmp::C50ClassificationController::setTrainingSamples(double trainSamples) {
    m_c50DecisionTree.getParameters().m_trainSamples = trainSamples;
}

double gdma::guicmp::C50ClassificationController::getTrainingSamples() {
    return m_c50DecisionTree.getParameters().m_trainSamples;
}

void gdma::guicmp::C50ClassificationController::setWinnow(bool winnow) {
    m_c50DecisionTree.getParameters().m_winnow = winnow;
}

bool gdma::guicmp::C50ClassificationController::getWinnow() {
    return m_c50DecisionTree.getParameters().m_winnow;
}

void gdma::guicmp::C50ClassificationController::setRules(bool rules) {
    m_c50DecisionTree.getParameters().m_rules = rules;
}

bool gdma::guicmp::C50ClassificationController::getRules() {
    return m_c50DecisionTree.getParameters().m_rules;
}

void gdma::guicmp::C50ClassificationController::setFuzzyThresholds(bool fuzzy) {
    m_c50DecisionTree.getParameters().m_fuzzyThreshold = fuzzy;
}

bool gdma::guicmp::C50ClassificationController::getFuzzyThresholds() {
    return m_c50DecisionTree.getParameters().m_fuzzyThreshold;
}

void gdma::guicmp::C50ClassificationController::setEarlyStopping(bool earlyStopping) {
    m_c50DecisionTree.getParameters().m_earlyStopping = earlyStopping;
}

bool gdma::guicmp::C50ClassificationController::getEarlyStopping() {
    return m_c50DecisionTree.getParameters().m_earlyStopping;
}

void gdma::guicmp::C50ClassificationController::setSubset(bool subset) {
    m_c50DecisionTree.getParameters().m_subset = subset;
}

bool gdma::guicmp::C50ClassificationController::getSubset() {
    return m_c50DecisionTree.getParameters().m_subset;
}

void gdma::guicmp::C50ClassificationController::setConfidenceFactor(double confFactor) {
    m_c50DecisionTree.getParameters().m_confidenceFactor = confFactor;
}

double gdma::guicmp::C50ClassificationController::getConfidenceFactor() {
    return m_c50DecisionTree.getParameters().m_confidenceFactor;
}

void gdma::guicmp::C50ClassificationController::setFeaturesToUse(
    std::vector<std::string> featuresToUse) {
    m_c50DecisionTree.setFeaturesToUse(featuresToUse);
}

void gdma::guicmp::C50ClassificationController::trainTree() {
    m_c50DecisionTree.train();
}

std::string gdma::guicmp::C50ClassificationController::getOutputText() {
    return m_c50DecisionTree.getOutputTextFile();
}

std::string gdma::guicmp::C50ClassificationController::getTreeText() {
    return m_c50DecisionTree.getOutputTreeFile();
}

std::string gdma::guicmp::C50ClassificationController::getRulesText() {
    return m_c50DecisionTree.getOutputRulesFile();
}

void gdma::guicmp::C50ClassificationController::setOutputFilesPath(boost::filesystem::path path) {
    m_outputFilesPath = path;
}

boost::filesystem::path gdma::guicmp::C50ClassificationController::getOutputFilesPath() {
    return m_outputFilesPath;
}

void gdma::guicmp::C50ClassificationController::saveOutputFiles() {
    m_c50DecisionTree.saveOutputFiles(m_outputFilesPath);
}

bool gdma::guicmp::C50ClassificationController::loadExternalFiles(std::string pathToFiles) {
    try {
        m_c50DecisionTree.loadExternalFiles(pathToFiles);
    } catch (te::common::Exception &e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(TE_TR(e.what()));
        return false;
    } catch (std::exception &e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(TE_TR(e.what()));
        return false;
    } catch (...) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(TE_TR("Unknown Exception!"));
        return false;
    }
    return true;
}

void gdma::guicmp::C50ClassificationController::predict() {
    m_c50DecisionTree.predict();
}

void gdma::guicmp::C50ClassificationController::setOutputColumnName(std::string columnName) {
    m_c50DecisionTree.setOutputClassFeatureName(columnName);
}

std::string gdma::guicmp::C50ClassificationController::getOutputColumnName() {
    return m_c50DecisionTree.getOutputClassFeatureName();
}

bool gdma::guicmp::C50ClassificationController::savePredictionOutput(const std::string pathShp,
    const std::string layerName, const std::string pathCSV) {
    try {
        if (m_c50DecisionTree.saveClassificationResults(pathShp, pathCSV)) {
            bool containsLayer = false;
            for (auto &layer : m_layerList) {
                if (layer->getTitle().compare(layerName) == 0) {
                    containsLayer = true;
                    break;
                }
            }

            if (!containsLayer) {
                te::map::AbstractLayerPtr out_layer = gdma::guicmp::createTerraLibLayer(pathShp,
                    layerName);
                setCurrentLayer(out_layer);

                m_newLayerCreated = true;
            } else {
                m_currentLayer->setOutOfDate();
            }
            return true;
        }
        return false;
    } catch (te::common::Exception& e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(TE_TR(e.what()));
    } catch (std::exception& e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(TE_TR(e.what()));
    } catch (...) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(TE_TR("Unknown Exception!"));
    }
}

bool gdma::guicmp::C50ClassificationController::newLayerCreated() {
    return m_newLayerCreated;
}

std::string gdma::guicmp::C50ClassificationController::getCurrentDataSourcePath() {
    te::da::DataSourcePtr dataSource = te::da::GetDataSource(m_currentLayer->getDataSourceId());

#ifdef WIN32
    // @TODO(Raian): Verify here this behavior. What host() will return in linux systems?
    return dataSource->getConnectionInfo().host() + dataSource->getConnectionInfo().path();
#else
    return dataSource->getConnectionInfo().path();
#endif
}

void gdma::guicmp::C50ClassificationController::setNoDataValue(int value) {
    m_c50DecisionTree.setNoDataValue(value);
}

int gdma::guicmp::C50ClassificationController::getNoDataValue() {
    return m_c50DecisionTree.getNoDataValue();
}
