/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/geodma_tvplugin/c50ClassificationDialog.cpp

  \brief This file contains the C50ClassificationDialog implementation.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// TerraLib Includes
#include <terralib/common/progress/ProgressManager.h>
#include <terralib/qt/widgets/help/HelpManager.h>
#include <terralib/qt/widgets/progress/ProgressViewerDialog.h>

// Qt Includes
#include <QDialog>
#include <QGridLayout>
#include <QMessageBox>

// STL Includes
#include <list>
#include <memory>
#include <string>

// GeoDMA Includes
#include "geodmalib/common/messageMediator.hpp"
#include "geodmalib/guicomponents/decision_trees/c50ClassificationWidget.hpp"
#include "geodmalib/guicomponents/dialogMessageObserver.hpp"
#include "geodmalib/guicomponents/decision_trees/c50ClassificationDialog.hpp"
#include "ui_GeoDMADialogForm.h"


gdma::guicmp::C50ClassificationDialog::C50ClassificationDialog(QWidget* parent, Qt::WindowFlags f) :
    QDialog(parent, f),
    m_ui(new Ui::GeoDMADialogForm) {
    m_ui->setupUi(this);

    m_dialogTitle = "C5.0 Decision Trees";

    // Message Dialog
    gdma::guicmp::DialogMessageObserver* msgObs =
        new gdma::guicmp::DialogMessageObserver(this, m_dialogTitle);
    m_msgObsId = gdma::common::MessageMediator::getInstance().addObserver(msgObs);

    // progress
    m_progViewer.reset(new te::qt::widgets::ProgressViewerDialog(this));
    m_progId = te::common::ProgressManager::getInstance().addViewer(m_progViewer.get());

    QGridLayout* layout = new QGridLayout(m_ui->m_widget);
    m_c50ClassificationWidget.reset(new gdma::guicmp::C50ClassificationWidget(m_ui->m_widget));
    layout->addWidget(m_c50ClassificationWidget.get(), 0, 0);
    layout->setContentsMargins(0, 0, 0, 0);

    // add icons
    this->setWindowIcon(QIcon::fromTheme("icon_geodma"));
    this->setWindowTitle(tr(m_dialogTitle.c_str()));
    m_ui->m_gdmaImgLabel->setPixmap(QIcon::fromTheme("icon_geodma_name").pixmap(112, 65));
    m_ui->m_panelImgLabel->setPixmap(QIcon::fromTheme("decisionTree_Classification")
        .pixmap(112, 48));
    m_ui->m_panelTextLabel->setText(tr(m_dialogTitle.c_str()));

    m_resultsDlg.reset(new gdma::guicmp::C50ResultsDialog(this));

    connect(m_ui->m_runPushButton, SIGNAL(released()), this, SLOT(onRunPushButtonClicked()));

    connect(m_ui->m_closePushButton, SIGNAL(released()), this, SLOT(onClosePushButtonClicked()));
    connect(m_ui->m_closePushButton, SIGNAL(released()), this, SLOT(close()));
    connect(m_resultsDlg.get(), SIGNAL(resultsDialogClosed()), this,
        SLOT(onResultsDialogClosed()));
    connect(m_c50ClassificationWidget.get(), SIGNAL(predictionRan()), this, SLOT(onPredictionRan()));
    connect(m_c50ClassificationWidget.get(), SIGNAL(predictionRan()), this, SLOT(close()));
    connect(m_c50ClassificationWidget.get(), SIGNAL(trainingRan()), this,
        SLOT(onTrainingRunFinished()));

    m_ui->m_helpPushButton->setNameSpace("dpi.inpe.br.geodma");
    m_ui->m_helpPushButton->setPageReference(
        "geodma/classification/decision_trees/geodma_c50_decision_trees.html");
}


gdma::guicmp::C50ClassificationDialog::~C50ClassificationDialog() {
    gdma::common::MessageMediator::getInstance().removeObserver(m_msgObsId);
    te::common::ProgressManager::getInstance().removeViewer(m_progId);
}

void gdma::guicmp::C50ClassificationDialog::setLayers(std::list<te::map::AbstractLayerPtr> layers) {
    m_c50ClassificationWidget->setLayers(layers);
}

bool gdma::guicmp::C50ClassificationDialog::newLayerCreated() {
    return m_c50ClassificationWidget->getController()->newLayerCreated();
}

te::map::AbstractLayerPtr gdma::guicmp::C50ClassificationDialog::getOutputLayer() {
    return m_c50ClassificationWidget->getController()->getCurrentLayer();
}

// ------------- Slots (Public)
void gdma::guicmp::C50ClassificationDialog::onRunPushButtonClicked() {
    m_c50ClassificationWidget->run();
}

void gdma::guicmp::C50ClassificationDialog::onClosePushButtonClicked() {
    reject();
}

void gdma::guicmp::C50ClassificationDialog::onTrainingRunFinished() {
    m_resultsDlg->initialize(m_c50ClassificationWidget->getController());
    m_resultsDlg->show();
}

void gdma::guicmp::C50ClassificationDialog::onPredictionRan() {
    accept();
}

void gdma::guicmp::C50ClassificationDialog::onResultsDialogClosed() {
    m_c50ClassificationWidget->goToClassificationTab();
    m_c50ClassificationWidget->setInputTreeFileText();
}
