/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/guicomponents/c50ClassificationWidget.cpp

  \brief This file contains the C50ClassificationWidget class implementation.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// TerraLib Includes
#include <terralib/core/translator/Translator.h>
#include <terralib/dataaccess/dataset/DataSetType.h>
#include <terralib/dataaccess/dataset/PrimaryKey.h>
#include <terralib/dataaccess/utils/Utils.h>
#include <terralib/geometry/GeometryProperty.h>
#include <terralib/qt/widgets/utils/DoubleListWidget.h>

// QT Includes
#include <QFileDialog>
#include <QMessageBox>
#include <QString>

// STL Includes
#include <list>
// #include <map>
#include <memory>
#include <string>
#include <vector>

// GeoDMA Includes
#include "geodmalib/common/messageMediator.hpp"
#include "geodmalib/guicomponents/decision_trees/c50ClassificationWidget.hpp"
#include "geodmalib/guicomponents/decision_trees/c50ParametersDialog.hpp"
#include "geodmalib/guicomponents/layerUtils.hpp"
#include "ui_C50ClassificationWidgetForm.h"

gdma::guicmp::C50ClassificationWidget::C50ClassificationWidget(QWidget* parent, Qt::WindowFlags f) :
    QWidget(parent, f),
    m_ui(new Ui::C50ClassificationWidgetForm) {
    m_ui->setupUi(this);

    m_noColumnName = tr("--Select the column containing the labels--");

    m_advancedParamDlg.reset(new gdma::guicmp::C50ParametersDialog(this));

    // Setting the tab that is initial enabled.
    m_ui->m_tabWidget->setCurrentIndex(0);

    // Features Selection Widget
    m_featuresWidget.reset(new te::qt::widgets::DoubleListWidget(m_ui->m_inputFeaturesGroupBox));
    m_featuresWidget->setLeftLabel("Available Features");
    m_featuresWidget->setRightLabel("Used Features");
    QGridLayout* inputFeaturesLayout = new QGridLayout(m_ui->m_inputFeaturesGroupBox);
    inputFeaturesLayout->addWidget(m_featuresWidget.get());

    // Setting output widget
    QGridLayout* layout = new QGridLayout(m_ui->m_outClassifWidget);
    m_outputWidget.reset(new gdma::guicmp::OutputDataWidget(m_ui->m_outClassifWidget));
    layout->addWidget(m_outputWidget.get(), 0, 0);
    layout->setContentsMargins(0, 0, 0, 0);
    layout->setSizeConstraint(QLayout::SetMinimumSize);

    // Setting default No Data value
    m_ui->m_noDataLineEditTrain->setValidator(new QIntValidator());
    m_ui->m_noDataLineEditTrain->setText(std::to_string(m_controller.getNoDataValue()).c_str());
    m_ui->m_noDataLineEditClassif->setValidator(new QIntValidator());
    m_ui->m_noDataLineEditClassif->setText(std::to_string(m_controller.getNoDataValue()).c_str());

    // Training Tab
    connect(m_ui->m_inputLayerComboBox, SIGNAL(currentIndexChanged(int)), this,
        SLOT(onInputLayerComboBoxSelected(int)));
    connect(m_ui->m_labelColumnComboBox, SIGNAL(currentIndexChanged(int)), this,
        SLOT(onLabelColumnComboBoxSelected(int)));
    connect(m_ui->m_trialsSpinBox, SIGNAL(valueChanged(int)), this,
        SLOT(onTrialsSpinBoxChanged(int)));
    connect(m_ui->m_globPruningCheckBox, SIGNAL(stateChanged(int)), this,
        SLOT(onGlobalPruningCheckBoxStateChanged(int)));
    connect(m_ui->m_minCasesSpinBox, SIGNAL(valueChanged(int)), this,
        SLOT(onMinCasesSpinBoxChanged(int)));
    connect(m_ui->m_trainSampleSpinBox, SIGNAL(valueChanged(double)), this,
        SLOT(onTrainSampleSpinBoxChanged(double)));
    connect(m_ui->m_advancedParamToolButton, SIGNAL(released()), this,
        SLOT(onAdvancedParametersToolButtonClicked()));
    connect(m_ui->m_tabWidget, SIGNAL(tabBarClicked(int)), this,
        SLOT(onTabBarClicked(int)));
    connect(m_ui->m_noDataLineEditTrain, SIGNAL(valueChanged(int)), this,
        SLOT(onNoDataLineEditChanged(int)));

    // Classification Tab
    connect(m_ui->m_treeFileToolButton, SIGNAL(clicked()), this,
        SLOT(onInputTreeFileToolButtonClicked()));
    connect(m_ui->m_classifyInputLayerComboBox, SIGNAL(currentIndexChanged(int)), this,
        SLOT(onInputLayerComboBoxSelected(int)));
    connect(m_ui->m_noDataLineEditClassif, SIGNAL(valueChanged(int)), this,
        SLOT(onNoDataLineEditChanged(int)));
}


gdma::guicmp::C50ClassificationWidget::~C50ClassificationWidget() {
}

Ui::C50ClassificationWidgetForm* gdma::guicmp::C50ClassificationWidget::getForm() const {
    return m_ui.get();
}

void gdma::guicmp::C50ClassificationWidget::setLayers(std::list<te::map::AbstractLayerPtr> layers) {
    m_controller.setLayerList(layers);
    for (auto &layer : layers) {
        std::unique_ptr<te::da::DataSetType> dsType = layer.get()->getSchema();
        if (dsType->hasGeom()) {
            te::gm::GeometryProperty* geomProp = te::da::GetFirstGeomProperty(dsType.get());

            if (geomProp->getGeometryType() == te::gm::PolygonType ||
                geomProp->getGeometryType() == te::gm::MultiPolygonType) {
                m_vectorLayers.push_back(layer);
                m_ui->m_inputLayerComboBox->addItem(QString(layer.get()->getTitle().c_str()),
                    QVariant(layer.get()->getId().c_str()));
                m_ui->m_classifyInputLayerComboBox->addItem(QString(layer.get()->getTitle().c_str()),
                    QVariant(layer.get()->getId().c_str()));
            }
        }
    }
}

void gdma::guicmp::C50ClassificationWidget::setCurrentLayer(te::map::AbstractLayerPtr layer) {
    m_controller.setCurrentLayer(layer);
    m_outputWidget->setCurrentLayerName(layer->getDataSetName().c_str());
    m_ui->m_classifyInputLayerComboBox->setCurrentText(QString(layer.get()->getTitle().c_str()));
}

bool gdma::guicmp::C50ClassificationWidget::run() {
    if (m_ui->m_tabWidget->currentWidget() == m_ui->m_trainingTab) {
        std::vector<std::string> featuresToUse = m_featuresWidget->getOutputValues();

        if (m_controller.getLabelsColumn().compare(m_noColumnName.toUtf8().data()) == 0) {
            gdma::common::MessageMediator::getInstance().sendWarningMsg(
                TE_TR("Select the column containing the training classes!"));
            return false;
        }

        if (featuresToUse.size() == 0) {
            gdma::common::MessageMediator::getInstance().sendWarningMsg(
                TE_TR("Select at least one feature to be used to train the Decision Tree!"));
            return false;
        }

        m_controller.setFeaturesToUse(featuresToUse);

        m_controller.trainTree();
        emit trainingRan();
    } else {
        QString className = m_ui->m_classColumnLineEdit->text();

        if (className.isEmpty()) {
            std::stringstream msg;
            msg << "Name of class column is empty!";
            gdma::common::MessageMediator::getInstance().sendErrorMsg(msg.str());
            return false;
        }

        QRegExp validator("[A-Za-z0-9_]{1,10}");
        if (!validator.exactMatch(className)) {
            std::stringstream msg;
            msg << "Label \"" << className.toUtf8().data() <<
                "\" contains special characters or has more than 10 characters!";
            gdma::common::MessageMediator::getInstance().sendErrorMsg(msg.str());
            return false;
        } else {
            if (columnExistsInLayer(m_controller.getCurrentLayer(), className.toStdString())) {
                QMessageBox::StandardButton reply;
                reply = QMessageBox::question(this, "Column already exists",
                    "Column \"" + className + "\" already exists in the Data Set. Do you want to overwrite it?",
                    QMessageBox::Yes | QMessageBox::No);

                if (reply == QMessageBox::No)
                    return false;
            }

            m_controller.setOutputColumnName(className.toStdString());
        }

        try {
            m_controller.predict();
        } catch (te::common::Exception &e) {
            gdma::common::MessageMediator::getInstance().sendErrorMsg(TE_TR(e.what()));
            return false;
        } catch (std::exception &e) {
            gdma::common::MessageMediator::getInstance().sendErrorMsg(TE_TR(e.what()));
            return false;
        } catch (...) {
            gdma::common::MessageMediator::getInstance().sendErrorMsg(TE_TR("Unknown Exception!"));
            return false;
        }

        std::string outputLayerName = m_outputWidget->getCurrentLayerName();
        std::string outputShpPath = m_outputWidget->getCurrentSHPPath();
        std::string outputCSVPath = m_outputWidget->getCurrentCSVPath();

        if (m_controller.savePredictionOutput(outputShpPath, outputLayerName, outputCSVPath)) {
            std::stringstream msg;
            msg << "Classification was saved in the column \""
                << m_controller.getOutputColumnName() << "\" of the data source \""
                << outputShpPath << "\"!" << std::endl;
            gdma::common::MessageMediator::getInstance().sendInformationMsg(TE_TR(msg.str()));
            emit predictionRan();
        } else {
            return false;
        }
    }
    return true;
}

gdma::guicmp::C50ClassificationController* gdma::guicmp::C50ClassificationWidget::getController() {
    return &m_controller;
}

// ------------- Slots (Public)

void gdma::guicmp::C50ClassificationWidget::onInputLayerComboBoxSelected(int index) {
    std::string layerID = reinterpret_cast<QComboBox*>(QObject::sender())->itemData(
    index, Qt::UserRole).toString().toUtf8().data();

    for (auto &layer : m_vectorLayers) {
        if (layerID == layer->getId().c_str()) {
            setCurrentLayer(layer);
            m_outputWidget->setCurrentLayerName(layer->getTitle().c_str());
            m_outputWidget->setCurrentSHPPath(m_controller.getCurrentDataSourcePath());
        }
    }

    if (m_ui->m_tabWidget->currentWidget() == m_ui->m_trainingTab) {
        //// Fill the combo box containing the training column name.
        te::da::DataSetTypePtr dsType = m_controller.getCurrentLayer()->getSchema();
        std::vector<te::dt::Property*> propVector = dsType->getProperties();
        te::da::PrimaryKey* pk = dsType->getPrimaryKey();
        std::vector<te::dt::Property*> pkVector = pk->getProperties();
        m_ui->m_labelColumnComboBox->clear();
        m_ui->m_labelColumnComboBox->addItem(m_noColumnName, QVariant(m_noColumnName));


        for (auto &prop : propVector) {
            if (prop->getType() != te::dt::GEOMETRY_TYPE && prop->getName().compare("id") != 0 &&
                (prop->getType() == te::dt::STRING_TYPE || prop->getType() == te::dt::BOOLEAN_TYPE) &&
                prop->getName().compare("col") != 0 && prop->getName().compare("row") != 0) {
                for (auto &key : pkVector) {
                    if (key->getName().compare(prop->getName()) != 0)
                        m_ui->m_labelColumnComboBox->addItem(prop->getName().c_str(),
                            QVariant(prop->getName().c_str()));
                }
            }
        }
    }
}

void gdma::guicmp::C50ClassificationWidget::onLabelColumnComboBoxSelected(int index) {
    std::string columnName = reinterpret_cast<QComboBox*>(QObject::sender())->itemData(
        index, Qt::UserRole).toString().toUtf8().data();

    if (columnName.compare(m_noColumnName.toUtf8().data()) != 0)
        m_controller.setLabelsColumn(columnName);

    m_controller.setAvailableFeatures();

    m_featuresWidget->clearInputValues();
    m_featuresWidget->clearOutputValues();
    m_featuresWidget->setInputValues(m_controller.getAvailableFeatures());
}

void gdma::guicmp::C50ClassificationWidget::onTrialsSpinBoxChanged(int val) {
    m_controller.setTrials(val);
}

void gdma::guicmp::C50ClassificationWidget::onGlobalPruningCheckBoxStateChanged(int state) {
    bool globalPruning = reinterpret_cast<QCheckBox*> (QObject::sender())->isChecked();

    m_controller.setGlobalPruning(globalPruning);
}

void gdma::guicmp::C50ClassificationWidget::onMinCasesSpinBoxChanged(int val) {
    m_controller.setMinCases(val);
}

void gdma::guicmp::C50ClassificationWidget::onTrainSampleSpinBoxChanged(double val) {
    m_controller.setTrainingSamples(val);
}

void gdma::guicmp::C50ClassificationWidget::onAdvancedParametersToolButtonClicked() {
    m_advancedParamDlg->initialize(m_controller);

    m_advancedParamDlg->exec();
}

void gdma::guicmp::C50ClassificationWidget::onTabBarClicked(int tabIdx) {
    if (tabIdx != m_ui->m_tabWidget->indexOf(m_ui->m_classificationTab))
        return;

    if (m_controller.getTreeText().compare("") != 0 || m_controller.getRulesText().compare("") != 0)
        m_ui->m_treeFileLineEdit->setText(m_controller.getOutputFilesPath().string().c_str());
}

void gdma::guicmp::C50ClassificationWidget::onInputTreeFileToolButtonClicked() {
    m_ui->m_treeFileLineEdit->clear();

    QString fileName = QFileDialog::getSaveFileName(this, tr("Select Input File..."), QString(),
        tr("Input File (*.tree *.TREE *.rules *.RULES);;"), 0, QFileDialog::DontConfirmOverwrite);

    if (fileName.isEmpty())
        return;

    if (m_controller.loadExternalFiles(fileName.toStdString()))
        m_ui->m_treeFileLineEdit->setText(fileName);
}

void gdma::guicmp::C50ClassificationWidget::onNoDataLineEditChanged(int) {
    int value = reinterpret_cast<QLineEdit*>(QObject::sender())->text().toInt();

    m_controller.setNoDataValue(value);
}

void gdma::guicmp::C50ClassificationWidget::goToClassificationTab() {
    m_ui->m_tabWidget->setCurrentWidget(m_ui->m_classificationTab);
}

void gdma::guicmp::C50ClassificationWidget::setInputTreeFileText() {
    std::string treeFile = m_controller.getOutputFilesPath().string();
    if (treeFile.empty()) {
        treeFile = "Using last generated tree";
        QFont font = m_ui->m_treeFileLineEdit->font();
        font.setItalic(true);
        m_ui->m_treeFileLineEdit->setFont(font);
    } else {
        QFont font = m_ui->m_treeFileLineEdit->font();
        font.setItalic(false);
        m_ui->m_treeFileLineEdit->setFont(font);
    }
    m_ui->m_treeFileLineEdit->setText(
        QString::fromLatin1(treeFile.c_str()));
}
