/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/guicomponents/kmeans/kmeansClassificationController.cpp

  \brief This file contains the KmeansClassificationController class definition.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// TerraLib Includes
#include <terralib/classification/KMeans.h>
#include <terralib/common/Exception.h>
#include <terralib/common/progress/TaskProgress.h>
#include <terralib/dataaccess/dataset/DataSet.h>
#include <terralib/dataaccess/dataset/DataSetType.h>
#include <terralib/dataaccess/utils/Utils.h>
#include <terralib/maptools/AbstractLayer.h>

// Boost Includes
#include <boost/filesystem.hpp>

// STL Includes
#include <list>
#include <map>
#include <memory>
#include <string>
#include <vector>

// GeoDMA Includes
#include "BuildConfig.hpp"
#include "geodmalib/common/messageMediator.hpp"
#include "geodmalib/datamanager/datasetIterator.hpp"
#include "geodmalib/datamanager/dataManager.hpp"
#include "geodmalib/datamanager/datasetParcialIterator.hpp"
#include "geodmalib/datamanager/utils.hpp"
#include "geodmalib/guicomponents/kmeans/kmeansClassificationController.hpp"
#include "geodmalib/guicomponents/layerUtils.hpp"


gdma::guicmp::KmeansClassificationController::KmeansClassificationController() {
    m_parameters.reset();
    m_numGeometries = 0;
}


gdma::guicmp::KmeansClassificationController::~KmeansClassificationController() {
}

void gdma::guicmp::KmeansClassificationController::setLayerList(
    std::list<te::map::AbstractLayerPtr> layerList) {
    m_layerList = layerList;
}

std::list<te::map::AbstractLayerPtr> gdma::guicmp::KmeansClassificationController::getLayerList() {
    return m_layerList;
}

void gdma::guicmp::KmeansClassificationController::setCurrentLayer(
    te::map::AbstractLayerPtr layer) {
    m_currentLayer = layer;
    m_newLayerCreated = false;
}

te::map::AbstractLayerPtr gdma::guicmp::KmeansClassificationController::getCurrentLayer() {
    return m_currentLayer;
}

void gdma::guicmp::KmeansClassificationController::setNumberOfClusters(const unsigned numClusters) {
    m_parameters.m_K = numClusters;
}

unsigned int gdma::guicmp::KmeansClassificationController::getNumberOfClusters() const {
    return m_parameters.m_K;
}

void gdma::guicmp::KmeansClassificationController::setMaxIterations(const unsigned int maxIterations) {
    m_parameters.m_maxIterations = maxIterations;
}

unsigned int gdma::guicmp::KmeansClassificationController::getMaxIterations() const {
    return m_parameters.m_maxIterations;
}

void gdma::guicmp::KmeansClassificationController::setEpsilon(const double epsilon) {
    m_parameters.m_epsilon = epsilon;
}

double gdma::guicmp::KmeansClassificationController::getEpsilon() const {
    return m_parameters.m_epsilon;
}

void gdma::guicmp::KmeansClassificationController::setPercentTrainEntries(const double percent) {
    te::da::DataSourcePtr dataSource = te::da::GetDataSource(m_currentLayer->getDataSourceId(), true);
    te::da::DataSetPtr dataset = dataSource->getDataSet(m_currentLayer->getDataSetName());

    m_numGeometries = std::round((percent / 100) * dataset->size());
}

int gdma::guicmp::KmeansClassificationController::getNumberOfTrainEntries() const {
    return m_numGeometries;
}

void gdma::guicmp::KmeansClassificationController::setFeaturesToUse(
    const std::vector<std::string> featuresToUse) {
    te::da::DataSetTypePtr dsType = m_currentLayer->getSchema();

    m_featuresToUse.clear();
    for (auto &featName : featuresToUse)
        m_featuresToUse.push_back(dsType->getPropertyPosition(featName));
}

std::vector<unsigned int> gdma::guicmp::KmeansClassificationController::getFeaturesToUse() const {
    return m_featuresToUse;
}

std::map<std::string, gdma::fte::FeaturesSummary>
    gdma::guicmp::KmeansClassificationController::getClassificationResults() const {
    return m_classifResult;
}

std::string gdma::guicmp::KmeansClassificationController::getCurrentDataSourcePath() {
    te::da::DataSourcePtr dataSource = te::da::GetDataSource(m_currentLayer->getDataSourceId());

#ifdef WIN32
    // @TODO(Raian): Verify here this behavior. What host() will return in linux systems?
    return dataSource->getConnectionInfo().host() + dataSource->getConnectionInfo().path();
#else
    return dataSource->getConnectionInfo().path();
#endif
}


bool gdma::guicmp::KmeansClassificationController::run() {
    te::cl::KMeans<gdma::dm::DatasetParcialIterator, gdma::dm::DatasetIterator> classifier;

    te::da::DataSourcePtr dataSource = te::da::GetDataSource(m_currentLayer->getDataSourceId(), true);
    te::da::DataSetPtr dataset = dataSource->getDataSet(m_currentLayer->getDataSetName());

    std::shared_ptr<std::vector<unsigned int>> positionsTraining(new std::vector<unsigned int>());
    gdma::dm::generateRandomSequence(dataset, m_numGeometries, *positionsTraining.get());
    gdma::dm::DatasetParcialIterator itTrain =
        gdma::dm::DatasetParcialIterator::begin(dataset, positionsTraining);
    gdma::dm::DatasetParcialIterator itTrainEnd =
        gdma::dm::DatasetParcialIterator::end(dataset, positionsTraining);

    gdma::dm::DatasetIterator itClass = gdma::dm::DatasetIterator::begin(dataset);
    gdma::dm::DatasetIterator itClassEnd = gdma::dm::DatasetIterator::end(dataset);

    std::vector<unsigned int> classifResult;

    try {
        if (!classifier.initialize(m_parameters))
            return false;

        if (!classifier.train(itTrain, itTrainEnd, m_featuresToUse, std::vector<unsigned int>(), true))
            return false;

        if (!classifier.classify(itClass, itClassEnd, m_featuresToUse, classifResult, true))
            return false;

        return generateOutputMap(classifResult);
    } catch (te::common::Exception& e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(e.what());
        return false;
    } catch (std::exception& e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(e.what());
        return false;
    } catch (...) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg("Unknown Exception!");
        return false;
    }

    return true;
}

void gdma::guicmp::KmeansClassificationController::setOutputData(std::string outputDataSource,
    std::string outputLayerName, std::string outputCSV) {
    m_outputDataSource = outputDataSource;
    m_outputLayerName = outputLayerName;
    m_outputCSV = outputCSV;
}

void gdma::guicmp::KmeansClassificationController::setOutputColumnName(std::string columnName) {
    m_outColumnName = columnName;
}

bool gdma::guicmp::KmeansClassificationController::saveClassification() {
    try {
        te::da::DataSourcePtr inputDS = te::da::GetDataSource(m_currentLayer->getDataSourceId());
        std::string inputSHPPath = inputDS->getConnectionInfo().host()
            + inputDS->getConnectionInfo().path();
        if (m_outputDataSource.compare(inputSHPPath) == 0) {
            gdma::dm::updateDataSource(m_outputDataSource, m_classifResult);
        } else {
            if (!boost::filesystem::exists(m_outputDataSource)) {
                gdma::dm::saveToOGRDataSource(m_outputDataSource, m_currentLayer->getDataSetName(),
                    m_currentLayer->getDataSourceId(), m_classifResult);
            } else {
                gdma::common::MessageMediator::getInstance().sendWarningMsg(
                    "Output shape file already exists. You can not overwrite it!");
                return false;
            }
        }

        bool containsLayer = false;
        for (auto &layer : m_layerList) {
            if (layer->getTitle().compare(m_outputLayerName) == 0) {
                containsLayer = true;
                break;
            }
        }

        if (!containsLayer) {
            te::map::AbstractLayerPtr out_layer =
                gdma::guicmp::createTerraLibLayer(m_outputDataSource, m_outputLayerName);
            setCurrentLayer(out_layer);
            m_newLayerCreated = true;
        } else {
            m_currentLayer->setOutOfDate();
        }

        if (m_outputCSV.compare("") != 0)
            gdma::dm::saveAttributesToCSV(m_classifResult, m_currentLayer->getDataSourceId(),
                m_currentLayer->getDataSetName(), m_outputCSV);

        std::string outMsg = "Classification saved in the column \"" + m_outColumnName +
            "\" of the data source \"" + m_outputDataSource + "\"!";
        gdma::common::MessageMediator::getInstance().sendInformationMsg(outMsg);
    } catch (te::common::Exception& e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(e.what());
        return false;
    } catch (std::exception& e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(e.what());
        return false;
    } catch (...) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg("Unknown Exception!");
        return false;
    }

    return true;
}

bool gdma::guicmp::KmeansClassificationController::newLayerCreated() {
    return m_newLayerCreated;
}

// ---------------------------
// Private Members

bool gdma::guicmp::KmeansClassificationController::generateOutputMap(
    std::vector<unsigned int>& classifResult) {
    te::common::TaskProgress task(TE_TR("Preparing Data to save"));
    task.setTotalSteps(0);
    task.useTimer(true);

    te::da::DataSourcePtr dataSource = te::da::GetDataSource(m_currentLayer->getDataSourceId(), true);
    te::da::DataSetTypePtr dsType = m_currentLayer->getSchema();
    te::da::DataSetPtr dataset = dataSource->getDataSet(m_currentLayer->getDataSetName());

    std::vector<std::string> pnames;
    te::da::GetOIDPropertyNames(dsType.get(), pnames);

    for (int pos = 0; pos < classifResult.size(); pos++) {
        if (!dataset->move(pos))
            throw te::common::Exception("Position inexistent in dataset!");

        std::string key = dataset->getAsString(pnames[0]);

        gdma::fte::FeaturesSummary fteClass;
        std::string cluster = "cluster_" + std::to_string(classifResult[pos]);
        fteClass.addFeature(m_outColumnName, m_outColumnName, cluster);
        m_classifResult[key] = fteClass;

        if (task.isActive() == false)
            throw te::common::Exception(TE_TR("Operation canceled!"));

        task.pulse();
    }

    return true;
}
