/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/guicomponents/kmeans/kmeansClassificationController.hpp

  \brief This file contains the KmeansClassificationController class definition.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

#ifndef GEODMA_LIB_GUICMP_KMEANSLASSIFICATIONCONTROLLER_HPP_
#define GEODMA_LIB_GUICMP_KMEANSLASSIFICATIONCONTROLLER_HPP_

// TerraLib Includes
#include <terralib/classification/KMeans.h>
#include <terralib/maptools/AbstractLayer.h>

// STL Includes
#include <list>
#include <map>
#include <string>
#include <vector>

// GeoDMA Includes
#include "geodmalib/datamanager/datasetIterator.hpp"
#include "geodmalib/datamanager/datasetParcialIterator.hpp"
#include "geodmalib/guicomponents/config.hpp"
#include "geodmalib/featureextraction/featuresSummary.hpp"

namespace gdma {
namespace guicmp {
typedef te::cl::KMeans<gdma::dm::DatasetParcialIterator, gdma::dm::DatasetIterator>::Parameters
    KmeansParameters;

class GDMAGUICMPEXPORT KmeansClassificationController {
 public:
     KmeansClassificationController();

     ~KmeansClassificationController();

     void setLayerList(std::list<te::map::AbstractLayerPtr> layerList);

     std::list<te::map::AbstractLayerPtr> getLayerList();

     void setCurrentLayer(te::map::AbstractLayerPtr layer);

     te::map::AbstractLayerPtr getCurrentLayer();

     void setNumberOfClusters(const unsigned int numClusters);

     unsigned int getNumberOfClusters() const;

     void setMaxIterations(const unsigned int maxIterations);

     unsigned int getMaxIterations() const;

     void setEpsilon(const double epsilon);

     double getEpsilon() const;

     void setPercentTrainEntries(const double percent);

     int getNumberOfTrainEntries() const;

     void setFeaturesToUse(const std::vector<std::string> featuresToUse);

     std::vector<unsigned int> getFeaturesToUse() const;

     std::map<std::string, gdma::fte::FeaturesSummary> getClassificationResults() const;

     std::string getCurrentDataSourcePath();

     bool run();

     void setOutputData(std::string outputDataSource, std::string outputLayerName,
         std::string outputCSV = "");

     void setOutputColumnName(std::string columnName);

     bool saveClassification();

     bool newLayerCreated();

 private:
     bool generateOutputMap(std::vector<unsigned int>& classifResult);

 private:
     std::list<te::map::AbstractLayerPtr> m_layerList;  //!< List of all layers of the application
     te::map::AbstractLayerPtr m_currentLayer;      //!< The input layer
     KmeansParameters m_parameters;                 //!< The classifier parameters
     int m_numGeometries;                           //!< The number of entries to compute the means
     std::vector<unsigned int> m_featuresToUse;     //!< The indices of the features to be used
     std::map<std::string, gdma::fte::FeaturesSummary> m_classifResult;  //!< The container that will be filled with the classification results

     std::string m_outColumnName;                  //!< Name of the column to save the output data.
     std::string m_outputDataSource;             //!< The data source to save the clusters.
     std::string m_outputCSV;                    //!< The output CSV to save the data.
     std::string m_outputLayerName;              //!< The output layer to save the data;

     bool m_newLayerCreated;                       //!< Will be set to true if a new layer is created.
};
}  // namespace guicmp
}  // namespace gdma
#endif  // GEODMA_LIB_GUICMP_KMEANSLASSIFICATIONCONTROLLER_HPP_
