/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/guicomponents/kmeansClassificationDialog.cpp

  \brief This file contains the KMeansClassificationDialog class implementation.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// TerraLib Includes
#include <terralib/common/progress/ProgressManager.h>
#include <terralib/maptools/AbstractLayer.h>
#include <terralib/qt/widgets/help/HelpManager.h>
#include <terralib/qt/widgets/progress/ProgressViewerDialog.h>

// Qt Includes
#include <QDialog>

// STL Includes
#include <list>
#include <memory>
#include <string>

// GeoDMA Includes
#include "geodmalib/common/messageMediator.hpp"
#include "geodmalib/guicomponents/kmeans/kmeansClassificationDialog.hpp"
#include "geodmalib/guicomponents/dialogMessageObserver.hpp"
#include "ui_GeoDMADialogForm.h"


gdma::guicmp::KMeansClassificationDialog::KMeansClassificationDialog(QWidget* parent,
    Qt::WindowFlags f) : QDialog(parent, f),
    m_ui(new Ui::GeoDMADialogForm) {
    m_ui->setupUi(this);

    m_dialogTitle = "K-Means Clustering";

    // Message Dialog
    gdma::guicmp::DialogMessageObserver* msgObs =
        new gdma::guicmp::DialogMessageObserver(this, m_dialogTitle);
    m_msgObsId = gdma::common::MessageMediator::getInstance().addObserver(msgObs);

    // progress
    m_progViewer.reset(new te::qt::widgets::ProgressViewerDialog(this));
    m_progId = te::common::ProgressManager::getInstance().addViewer(m_progViewer.get());

    QGridLayout* layout = new QGridLayout(m_ui->m_widget);
    m_kmeansClassificationWidget.reset(new gdma::guicmp::KMeansClassificationWidget(m_ui->m_widget));
    layout->addWidget(m_kmeansClassificationWidget.get(), 0, 0);
    layout->setContentsMargins(0, 0, 0, 0);

    // add icons
    this->setWindowIcon(QIcon::fromTheme("icon_geodma"));
    this->setWindowTitle(tr(m_dialogTitle.c_str()));
    m_ui->m_gdmaImgLabel->setPixmap(QIcon::fromTheme("icon_geodma_name").pixmap(112, 65));
    m_ui->m_panelImgLabel->setPixmap(QIcon::fromTheme("kmeans_cluster").pixmap(112, 48));
    m_ui->m_panelTextLabel->setText(tr(m_dialogTitle.c_str()));

    connect(m_ui->m_closePushButton, SIGNAL(released()), this, SLOT(onClosePushButtonClicked()));
    connect(m_ui->m_closePushButton, SIGNAL(released()), this, SLOT(close()));
    connect(m_ui->m_runPushButton, SIGNAL(released()), this, SLOT(onRunPushButtonClicked()));

    m_ui->m_helpPushButton->setNameSpace("dpi.inpe.br.geodma");
    m_ui->m_helpPushButton->setPageReference(
        "geodma/classification/kmeans/kmeans_clustering.html");
}


gdma::guicmp::KMeansClassificationDialog::~KMeansClassificationDialog() {
    gdma::common::MessageMediator::getInstance().removeObserver(m_msgObsId);
    te::common::ProgressManager::getInstance().removeViewer(m_progId);
}

void gdma::guicmp::KMeansClassificationDialog::setLayers(std::list<te::map::AbstractLayerPtr> layers) {
    m_kmeansClassificationWidget->setLayers(layers);
}

bool gdma::guicmp::KMeansClassificationDialog::newLayerCreated() {
    return m_kmeansClassificationWidget->newLayerCreated();
}

te::map::AbstractLayerPtr gdma::guicmp::KMeansClassificationDialog::getCurrentLayer() {
    return m_kmeansClassificationWidget->getCurrentLayer();
}

// ------------- Slots (Public)
void gdma::guicmp::KMeansClassificationDialog::onRunPushButtonClicked() {
    if (m_kmeansClassificationWidget->run())
        accept();
}

void gdma::guicmp::KMeansClassificationDialog::onClosePushButtonClicked() {
    reject();
}
