/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/guicomponents/kmeansClassificationWidget.cpp

  \brief This file contains the KMeansClassificationWidget class implementation.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// TerraLib Includes
#include <terralib/dataaccess/utils/Utils.h>
#include <terralib/geometry/GeometryProperty.h>
#include <terralib/maptools/AbstractLayer.h>
#include <terralib/qt/widgets/utils/DoubleListWidget.h>

// Qt Includes
#include <QWidget>
#include <QMessageBox>
#include <QString>

// STL Includes
#include <list>
#include <memory>
#include <string>
#include <vector>

// GeoDMA Includes
#include "geodmalib/common/messageMediator.hpp"
#include "geodmalib/guicomponents/kmeans/kmeansClassificationWidget.hpp"
#include "geodmalib/guicomponents/layerUtils.hpp"
#include "geodmalib/guicomponents/outputDataWidget.hpp"
#include "ui_KMeansClassificationWidgetForm.h"


gdma::guicmp::KMeansClassificationWidget::KMeansClassificationWidget(QWidget* parent,
    Qt::WindowFlags f) :
    QWidget(parent, f),
    m_ui(new Ui::KMeansClassificationWidgetForm) {
    m_ui->setupUi(this);

    // Features Selection Widget
    m_featuresWidget.reset(new te::qt::widgets::DoubleListWidget(m_ui->m_inputFeatGroupBox));
    m_featuresWidget->setLeftLabel("Available Features");
    m_featuresWidget->setRightLabel("Used Features");
    QGridLayout* inputFeaturesLayout = new QGridLayout(m_ui->m_inputFeatGroupBox);
    inputFeaturesLayout->addWidget(m_featuresWidget.get());

    m_outputWidget.reset(new gdma::guicmp::OutputDataWidget(m_ui->m_outputGroupBox));
    QGridLayout* outputLayout = new QGridLayout(m_ui->m_outputGroupBox);
    outputLayout->addWidget(m_outputWidget.get(), 0, 0);

    connect(m_ui->m_inputLayerComboBox, SIGNAL(currentIndexChanged(int)), this,
        SLOT(onInputLayerComboBoxSelected(int)));
}


gdma::guicmp::KMeansClassificationWidget::~KMeansClassificationWidget() {
}

void gdma::guicmp::KMeansClassificationWidget::setLayers(std::list<te::map::AbstractLayerPtr> layers) {
    m_controller.setLayerList(layers);
    for (auto &layer : layers) {
        std::unique_ptr<te::da::DataSetType> dsType = layer.get()->getSchema();
        if (dsType->hasGeom()) {
            te::gm::GeometryProperty* geomProp = te::da::GetFirstGeomProperty(dsType.get());

            if (geomProp->getGeometryType() == te::gm::PolygonType ||
                geomProp->getGeometryType() == te::gm::MultiPolygonType) {
                m_vectorLayers.push_back(layer);
                m_ui->m_inputLayerComboBox->addItem(QString(layer.get()->getTitle().c_str()),
                    QVariant(layer.get()->getId().c_str()));
            }
        }
    }
}

void gdma::guicmp::KMeansClassificationWidget::setCurrentLayer(te::map::AbstractLayerPtr layer) {
    m_controller.setCurrentLayer(layer);
    m_outputWidget->setCurrentLayerName(layer->getDataSetName().c_str());
}

te::map::AbstractLayerPtr gdma::guicmp::KMeansClassificationWidget::getCurrentLayer() {
    return m_controller.getCurrentLayer();
}

bool gdma::guicmp::KMeansClassificationWidget::run() {
    m_controller.setFeaturesToUse(m_featuresWidget->getOutputValues());
    m_controller.setMaxIterations(m_ui->m_maxIterSpinBox->value());
    m_controller.setNumberOfClusters(m_ui->m_numClustersSpinBox->value());
    m_controller.setPercentTrainEntries(m_ui->m_inputPointsSpinBox->value());

    m_controller.setOutputData(m_outputWidget->getCurrentSHPPath(),
        m_outputWidget->getCurrentLayerName(), m_outputWidget->getCurrentCSVPath());

    QString className = m_ui->m_outColumnLineEdit->text();

    if (className.isEmpty()) {
        std::stringstream msg;
        msg << "Name of class column is empty!";
        gdma::common::MessageMediator::getInstance().sendErrorMsg(msg.str());
        return false;
    }

    QRegExp validator("[A-Za-z0-9_]{1,10}");
    if (!validator.exactMatch(className)) {
        std::stringstream msg;
        msg << "Label \"" << className.toUtf8().data() <<
            "\" contains special characters or has more than 10 characters!";
        gdma::common::MessageMediator::getInstance().sendErrorMsg(msg.str());
        return false;
    } else {
        if (columnExistsInLayer(m_controller.getCurrentLayer(), className.toStdString())) {
            QMessageBox::StandardButton reply;
            reply = QMessageBox::question(this, "Column already exists",
                "Column \"" + className + "\" already exists in the Data Set. Do you want to overwrite it?",
                QMessageBox::Yes | QMessageBox::No);

            if (reply == QMessageBox::No)
                return false;
        }

        m_controller.setOutputColumnName(className.toStdString());
    }

    return m_controller.run() && m_controller.saveClassification();
}

bool gdma::guicmp::KMeansClassificationWidget::newLayerCreated() {
    return m_controller.newLayerCreated();
}

// ------------- Slots (Public)

void gdma::guicmp::KMeansClassificationWidget::onInputLayerComboBoxSelected(int index) {
    std::string layerID = reinterpret_cast<QComboBox*>(QObject::sender())->itemData(
        index, Qt::UserRole).toString().toUtf8().data();

    for (auto &layer : m_vectorLayers) {
        if (layerID == layer->getId().c_str()) {
            setCurrentLayer(layer);
            m_outputWidget->setCurrentLayerName(layer->getDataSetName().c_str());
            m_outputWidget->setCurrentSHPPath(m_controller.getCurrentDataSourcePath());
        }
    }

    //// Fill the combo box containing the training column name.
    te::da::DataSetTypePtr dsType = m_controller.getCurrentLayer()->getSchema();
    std::vector<te::dt::Property*> propVector = dsType->getProperties();
    te::da::PrimaryKey* pk = dsType->getPrimaryKey();
    std::vector<te::dt::Property*> pkVector = pk->getProperties();

    m_featuresWidget->clearInputValues();
    m_featuresWidget->clearOutputValues();
    std::vector<std::string> availableFeatures;

    for (auto &prop : propVector) {
        if (prop->getType() != te::dt::GEOMETRY_TYPE && prop->getName().compare("id") != 0 &&
            prop->getType() != te::dt::STRING_TYPE && prop->getType() != te::dt::BOOLEAN_TYPE &&
            prop->getName().compare("col") != 0 && prop->getName().compare("row") != 0) {
            for (auto &key : pkVector) {
                if (key->getName().compare(prop->getName()) != 0)
                    availableFeatures.push_back(prop->getName().c_str());
            }
        }
    }

    m_featuresWidget->setInputValues(availableFeatures);
}
