/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/guicomponents/multiSelectionTool.cpp

  \brief This file contains the MultiSelectionTool class implementation.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// TerraLib Includes
#include <terralib/dataaccess/dataset/DataSet.h>
#include <terralib/dataaccess/dataset/DataSetType.h>
#include <terralib/dataaccess/dataset/ObjectIdSet.h>
#include <terralib/dataaccess/utils/Utils.h>
#include <terralib/geometry/Coord2D.h>
#include <terralib/geometry/Geometry.h>
#include <terralib/geometry/GeometryProperty.h>
#include <terralib/geometry/Point.h>
#include <terralib/geometry/Utils.h>
#include <terralib/qt/widgets/canvas/MapDisplay.h>
#include <terralib/maptools/AbstractLayer.h>
#include <terralib/qt/widgets/tools/RubberBand.h>
#include <terralib/qt/widgets/utils/ScopedCursor.h>

// Qt Includes
#include <QCursor>
#include <QtCore/QString>
#include <QKeyEvent>
#include <QMessageBox>
#include <QMouseEvent>

// STL Includes
#include <exception>
#include <list>
#include <memory>
#include <string>
#include <vector>

// GeoDMA Includes
#include "geodmalib/common/messageMediator.hpp"
#include "geodmalib/guicomponents/multiSelectionTool.hpp"


gdma::guicmp::MultiSelectionTool::MultiSelectionTool(te::qt::widgets::MapDisplay* display,
    const QCursor& cursor, const std::list<te::map::AbstractLayerPtr>& layers, QObject* parent)
    : te::qt::widgets::RubberBand(display, parent),
    m_layers(layers),
    m_selectionStarted(false),
    m_keepPreviousSelection(true),
    m_selectionByPointing(false) {
}


gdma::guicmp::MultiSelectionTool::~MultiSelectionTool() {
}

bool gdma::guicmp::MultiSelectionTool::mousePressEvent(QMouseEvent* e)
{
    if (e->button() != Qt::LeftButton)
        return false;

    m_selectionStarted = true;
    m_rect = QRectF();

    return te::qt::widgets::RubberBand::mousePressEvent(e);
}

bool gdma::guicmp::MultiSelectionTool::mouseMoveEvent(QMouseEvent* e)
{
    if (!m_selectionStarted)
        return false;

    return te::qt::widgets::RubberBand::mouseMoveEvent(e);
}

bool gdma::guicmp::MultiSelectionTool::mouseReleaseEvent(QMouseEvent* e) {
    te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);

    m_selectionStarted = false;

    if (e->button() != Qt::LeftButton)
        return false;

    te::qt::widgets::RubberBand::mouseReleaseEvent(e);

    Qt::KeyboardModifiers keys = e->modifiers();

    m_selectionByPointing = false;

    if (m_rect.isNull()) {
        m_selectionByPointing = true;
        QPointF pixelOffset(4.0, 4.0);
#if QT_VERSION >= 0x050000
        m_rect = QRectF(e->localPos() - pixelOffset, e->localPos() + pixelOffset);
#else
        m_rect = QRectF(e->posF() - pixelOffset, e->posF() + pixelOffset);
#endif
    }

    // Converts rect boundary to world coordinates
    QPointF ll(m_rect.left(), m_rect.bottom());
    QPointF ur(m_rect.right(), m_rect.top());
    ll = m_display->transform(ll);
    ur = m_display->transform(ur);

    te::gm::Envelope envelope(ll.x(), ll.y(), ur.x(), ur.y());

    // Select the layers
    std::list<te::map::AbstractLayerPtr>::iterator it;
    for (it = m_layers.begin(); it != m_layers.end(); ++it)
        executeSelection(*it, envelope);

    return true;
}

void gdma::guicmp::MultiSelectionTool::setLayers(
    const std::list<te::map::AbstractLayerPtr>& layers) {
    m_layers = layers;
}

void gdma::guicmp::MultiSelectionTool::executeSelection(const te::map::AbstractLayerPtr& layer,
    const te::gm::Envelope& e) {
    if (layer->getVisibility() != te::map::VISIBLE || !layer->isValid() ||
        layer->getType() == "FOLDERLAYER")
        return;

    te::da::DataSetTypePtr dsType = layer->getSchema();

    if (!dsType->hasGeom())
        return;

    te::gm::Envelope reprojectedEnvelope(e);

    if ((layer->getSRID() != TE_UNKNOWN_SRS) && (m_display->getSRID() != TE_UNKNOWN_SRS) &&
        (layer->getSRID() != m_display->getSRID()))
        reprojectedEnvelope.transform(m_display->getSRID(), layer->getSRID());

    // Clear previous selection?
    if (!m_keepPreviousSelection) {
        layer->clearSelected();
        emit layerSelectedObjectsChanged(layer);
    }

    if (!reprojectedEnvelope.intersects(layer->getExtent()))
        return;

    try {
        // Gets the layer schema
        std::unique_ptr<const te::map::LayerSchema> schema(layer->getSchema());
        if (!schema->hasGeom())
            return;

        te::gm::GeometryProperty* gp = te::da::GetFirstGeomProperty(schema.get());

        // Gets the dataset
        te::da::DataSetPtr dataset = layer->getData(gp->getName(), &reprojectedEnvelope, te::gm::INTERSECTS);
        assert(dataset.get());

        // Let's generate the oids
        te::da::ObjectIdSet* oids = 0;
        te::da::GetEmptyOIDSet(schema.get(), oids);
        assert(oids);

        std::vector<std::string> pnames;
        te::da::GetOIDPropertyNames(schema.get(), pnames);

        // Generates a geometry from the given extent. It will be used to refine the results
        te::gm::GeometryShrPtr geometryFromEnvelope(
            te::gm::GetGeomFromEnvelope(&reprojectedEnvelope, layer->getSRID()));

        if (m_selectionByPointing == false) {
            while (dataset->moveNext()) {
                te::gm::GeometryShrPtr g(dataset->getGeometry(gp->getName()));

                if (g.get() == 0)
                    continue;

                if (g->getSRID() == TE_UNKNOWN_SRS)
                    g->setSRID(layer->getSRID());

                if (!g->intersects(geometryFromEnvelope.get()))
                    continue;

                // Feature found!
                oids->add(te::da::GenerateOID(dataset.get(), pnames));
            }

            oids->setExpressionByIntersection(gp->getName(), reprojectedEnvelope, layer->getSRID());
        }
        else
        {
            // The restriction point. It will be used to refine the results
            te::gm::Coord2D center = reprojectedEnvelope.getCenter();
            te::gm::Point point(center.x, center.y, layer->getSRID());

            while (dataset->moveNext())
            {
                te::gm::GeometryShrPtr g(dataset->getGeometry(gp->getName()));

                if (g.get() == 0)
                    continue;

                if (g->getSRID() == TE_UNKNOWN_SRS)
                    g->setSRID(layer->getSRID());

                if (g->contains(&point) || g->crosses(geometryFromEnvelope.get()) ||
                    geometryFromEnvelope->contains(g.get())) {
                    // Feature found!
                    oids->add(te::da::GenerateOID(dataset.get(), pnames));
                }
            }

            oids->setExpressionByInClause();
        }

        assert(oids);

        // Adjusts the layer selection

        const te::da::ObjectIdSet* currentOids = layer->getSelected();
        if (currentOids == 0 || !m_keepPreviousSelection) {
            layer->select(oids);
            emit layerSelectedObjectsChanged(layer);
            return;
        }

        te::da::ObjectIdSet* finalSelected = currentOids->clone();
        finalSelected->symDifference(oids);

        delete oids;

        layer->clearSelected();
        layer->select(finalSelected);
        emit layerSelectedObjectsChanged(layer);
    } catch (te::common::Exception& e) {
        std::stringstream msg;
        msg << "The selection cannot be retrieved from the layer. Details: " << e.what();
        gdma::common::MessageMediator::getInstance().sendErrorMsg(msg.str());
    } catch (std::exception& e) {
        std::stringstream msg;
        msg << "The selection cannot be retrieved from the layer. Details: " << e.what();
        gdma::common::MessageMediator::getInstance().sendErrorMsg(msg.str());
    } catch (...) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg("Unknown Exception!");
    }
}

