/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/guicomponents/outputDataWidget.hpp

  \brief This file contains the OutputDataWidget class implementation.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// TerraLib Includes
#include <terralib/core/translator/Translator.h>

// QT Includes
#include <QFileDialog>
#include <QFileInfo>
#include <QString>
#include <QWidget>

// Boost Includes
#include <boost/filesystem.hpp>

// STL Includes
#include <string>

// GeoDMA Includes
#include "geodmalib/common/messageMediator.hpp"
#include "geodmalib/guicomponents/outputDataWidget.hpp"
#include "ui_outputDataWidgetForm.h"

gdma::guicmp::OutputDataWidget::OutputDataWidget(QWidget* parent, Qt::WindowFlags f)
    : QWidget(parent, f),
    m_ui(new Ui::OutputDataWidgetForm) {
    m_ui->setupUi(this);

    connect(m_ui->m_targetSHPFileToolButton, SIGNAL(pressed()), this,
        SLOT(onSHPTargetFileToolButtonPressed()));
    connect(m_ui->m_targetCSVFileToolButton, SIGNAL(pressed()), this,
        SLOT(onCSVTargetFileToolButtonPressed()));
    connect(m_ui->m_outLayerNameLineEdit, SIGNAL(textEdited(QString)), this,
        SLOT(onOutLayerNameLineEditChanged(QString)));
    m_ui->m_targetDatasourceToolButton->setIcon(QIcon::fromTheme("datasource"));
}


gdma::guicmp::OutputDataWidget::~OutputDataWidget() {
}

std::string gdma::guicmp::OutputDataWidget::getCurrentLayerName() {
    return m_currentLayerName.toStdString();
}

void gdma::guicmp::OutputDataWidget::setCurrentLayerName(std::string layerName) {
    m_currentLayerName = layerName.c_str();
    m_ui->m_outLayerNameLineEdit->setText(layerName.c_str());
}

std::string gdma::guicmp::OutputDataWidget::getCurrentSHPPath() {
    return m_currentSHPPath.toStdString();
}

void gdma::guicmp::OutputDataWidget::setCurrentSHPPath(std::string shpPath) {
    m_currentSHPPath = shpPath.c_str();
    m_ui->m_repositoryLineEdit->setText(m_currentSHPPath);
}

std::string gdma::guicmp::OutputDataWidget::getCurrentCSVPath() {
    return m_currentCSVPath.toStdString();
}

void gdma::guicmp::OutputDataWidget::setCurrentCSVPath(std::string csvPath) {
    m_currentCSVPath = csvPath.c_str();
    m_ui->m_outCSVLineEdit->setText(m_currentCSVPath);
}

// @TODO(Raian): Implement the behavior for the database button.

// ------------- Slots (Public)

void gdma::guicmp::OutputDataWidget::onSHPTargetFileToolButtonPressed() {
    m_ui->m_outLayerNameLineEdit->clear();
    m_ui->m_repositoryLineEdit->clear();

    QString fileName = QFileDialog::getSaveFileName(this, tr("Save as..."), QString(),
            tr("ESRI Shapefile (*.shp *.SHP);;"), 0);

    if (fileName.isEmpty())
        return;

    boost::filesystem::path outfile(fileName.toUtf8().data());
    std::string outputLayerName = outfile.leaf().string();
    m_ui->m_outLayerNameLineEdit->setText(outputLayerName.c_str());
    m_currentLayerName = outputLayerName.c_str();

    std::string outputFileStr = outfile.string();
    m_ui->m_repositoryLineEdit->setText(outputFileStr.c_str());
    m_currentSHPPath = outputFileStr.c_str();
}

void gdma::guicmp::OutputDataWidget::onCSVTargetFileToolButtonPressed() {
    m_ui->m_outCSVLineEdit->clear();

    QString fileName = QFileDialog::getSaveFileName(this, tr("Save as..."), QString(),
        tr("CSV File (*.csv *.CSV);;"), 0);

    if (fileName.isEmpty())
        return;

    boost::filesystem::path outfile(fileName.toUtf8().data());
    std::string outputFileStr = outfile.string();
    m_ui->m_outCSVLineEdit->setText(outputFileStr.c_str());
    m_currentCSVPath = outputFileStr.c_str();
}

void gdma::guicmp::OutputDataWidget::onOutLayerNameLineEditChanged(QString text) {
    m_currentLayerName = reinterpret_cast<QLineEdit*>(QObject::sender())->text();
}
