/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/guicomponents/sampleSelectionController.cpp

  \brief This file contains the SampleSelectionWidget class definition.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// TerraLib Includes
#include <terralib/dataaccess/dataset/DataSet.h>
#include <terralib/dataaccess/dataset/DataSetType.h>
#include <terralib/dataaccess/dataset/ObjectId.h>
#include <terralib/dataaccess/dataset/ObjectIdSet.h>
#include <terralib/dataaccess/utils/Utils.h>
#include <terralib/maptools/AbstractLayer.h>

// STL includes
#include <map>
#include <memory>
#include <set>
#include <string>
#include <utility>
#include <vector>

// GeoDMA Includes
#include "geodmalib/common/messageMediator.hpp"
#include "geodmalib/datamanager/dataManager.hpp"
#include "geodmalib/guicomponents/config.hpp"
#include "geodmalib/guicomponents/sampleSelectionController.hpp"


gdma::guicmp::SampleSelectionController::SampleSelectionController() :
    m_classesSet() {
    m_unsavedSamples = false;
}


gdma::guicmp::SampleSelectionController::~SampleSelectionController() {
}

void gdma::guicmp::SampleSelectionController::setCurrentLayer(te::map::AbstractLayerPtr layer) {
    m_classesSet.clear();
    m_currentLayer = layer;
}

te::map::AbstractLayerPtr gdma::guicmp::SampleSelectionController::getCurrentLayer() {
    return m_currentLayer;
}

void gdma::guicmp::SampleSelectionController::clearLayerSelection() {
    if (m_currentLayer.get())
        m_currentLayer->clearSelected();
}

void gdma::guicmp::SampleSelectionController::selectSamplesInLayer(
    std::multimap<std::string, std::string> selectedIDs) {
    clearLayerSelection();

    te::da::ObjectIdSet* selectedObjs;
    te::da::GetEmptyOIDSet(m_currentLayer->getSchema().get(), selectedObjs);

    for (auto &id : selectedIDs) {
        gdma::cl::ClassDescription* clazz = m_classesSet.getClass(id.first);
        selectedObjs->add(clazz->getSample(id.second)->clone());
    }

    if (selectedObjs->size() > 0)
        m_currentLayer->select(selectedObjs);
}

bool gdma::guicmp::SampleSelectionController::createSampleColumn(std::string columnName) {
    if (columnName.length() > 10) {
        std::stringstream msg;
        msg << "Column name must have maximum 10 characters." << std::endl;
        msg << "Enter another name!" << std::endl;

        gdma::common::MessageMediator::getInstance().sendErrorMsg(msg.str().c_str());
        return false;
    }

    try {
        if (gdma::dm::createColumnOGRDataSource(m_currentLayer->getDataSourceId(), columnName,
            te::dt::STRING_TYPE)) {
            m_currentLayer->setOutOfDate();
            return true;
        } else {
            return false;
        }
    } catch (te::common::Exception& e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(e.what());
        return false;
    } catch (std::exception& e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(e.what());
        return false;
    } catch (...) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(TE_TR("Unknown Exception!"));
        return false;
    }
}

void gdma::guicmp::SampleSelectionController::clearClassesSet() {
    m_unsavedSamples = true;
    m_classesSet.clear();
}

void gdma::guicmp::SampleSelectionController::setSamplesColumnName(std::string columnName) {
    m_samplesColumnName = columnName;
}

std::string gdma::guicmp::SampleSelectionController::getSamplesColumnName() {
    return m_samplesColumnName;
}

void gdma::guicmp::SampleSelectionController::addNewClass(std::string label, std::string color) {
    std::unique_ptr<const te::map::LayerSchema> schema(
        m_currentLayer->getSchema().release());
    gdma::cl::ClassDescription clazz(label, schema.get(), color);
    m_classesSet.addClass(clazz);
}

void gdma::guicmp::SampleSelectionController::removeClass(std::string label) {
    if (!m_classesSet.removeClass(label)) {
        std::stringstream msg;
        msg << "Class \"" << label << "\" does not exist!";
        gdma::common::MessageMediator::getInstance().sendErrorMsg(msg.str());
    }
    m_unsavedSamples = true;
}

bool gdma::guicmp::SampleSelectionController::setClassColor(std::string classLabel,
    std::string newColor) {
    gdma::cl::ClassDescription* clazz = m_classesSet.getClass(classLabel);
    if (clazz == 0) {
        std::stringstream msg;
        msg << "Class \"" << classLabel << "\" does not exist!";
        gdma::common::MessageMediator::getInstance().sendErrorMsg(msg.str());
        return false;
    } else {
        clazz->setColor(newColor);
        return true;
    }
}

std::string gdma::guicmp::SampleSelectionController::getClassColor(std::string classLabel) {
    gdma::cl::ClassDescription* clazz = m_classesSet.getClass(classLabel);
    if (clazz == 0) {
        std::stringstream msg;
        msg << "Class \"" << classLabel << "\" does not exist!";
        gdma::common::MessageMediator::getInstance().sendErrorMsg(msg.str());
        return "";
    } else {
        return clazz->getColor();
    }
}

bool gdma::guicmp::SampleSelectionController::setClassLabel(std::string classLabel,
    std::string newLabel) {
    gdma::cl::ClassDescription* clazz = m_classesSet.getClass(classLabel);

    if (clazz == 0) {
        std::stringstream msg;
        msg << "Class \"" << classLabel << "\" does not exist!";
        gdma::common::MessageMediator::getInstance().sendErrorMsg(msg.str());
        return false;
    } else {
        m_classesSet.setClassLabel(*clazz, newLabel);
        m_unsavedSamples = true;
        return true;
    }
}

int gdma::guicmp::SampleSelectionController::getNumberOfClasses() {
    return m_classesSet.getNumberOfClasses();
}

const std::set<std::string> gdma::guicmp::SampleSelectionController::addSelectedSamples(
    std::string classLabel) {
    gdma::cl::ClassDescription* clazz = m_classesSet.getClass(classLabel);
    std::set<std::string> addedIds;

    if (clazz == 0) {
        std::stringstream msg;
        msg << "Class \"" << classLabel << "\" does not exist!";
        gdma::common::MessageMediator::getInstance().sendErrorMsg(msg.str());
    } else {
        // Get Selected polygons
        const te::da::ObjectIdSet* selectedIds = m_currentLayer->getSelected();

        if (selectedIds && selectedIds->size() > 0) {
            for (auto &oid : *selectedIds) {
                clazz->addSample(oid);
                addedIds.insert(oid->getValueAsString());

                for (auto &cl : m_classesSet) {
                    if (cl.second.containsId(oid) && cl.first.compare(clazz->getLabel()) != 0)
                        cl.second.removeSample(oid);
                }
            }
        }
    }

    m_unsavedSamples = true;
    clearLayerSelection();

    return addedIds;
}

int gdma::guicmp::SampleSelectionController::getNumberOfSamples(std::string classLabel) {
    return m_classesSet.getNumberOfSamples(classLabel);
}

int gdma::guicmp::SampleSelectionController::getTotalNumberOfSamples() {
    return m_classesSet.getTotalNumberOfSamples();
}

const std::set<std::string>& gdma::guicmp::SampleSelectionController::getSampleIds(
    std::string classLabel) {
    return m_classesSet.getClass(classLabel)->getSampleSetAsString();
}

void gdma::guicmp::SampleSelectionController::removeSample(std::string classLabel, std::string oid) {
    gdma::cl::ClassDescription* clazz = m_classesSet.getClass(classLabel);

    if (m_classesSet.getClass("") == 0)
        addNewClass("", "");

    gdma::cl::ClassDescription* emptyClass = m_classesSet.getClass("");
    emptyClass->addSample(clazz->getSample(oid)->clone());

    m_unsavedSamples = true;
    clazz->removeSample(oid);
}

std::multimap<std::string, te::gm::GeometryShrPtr> gdma::guicmp::SampleSelectionController::getSampleGeometries() {
    std::multimap<std::string, te::gm::GeometryShrPtr> mapGeometries;
    for (auto &clazz : m_classesSet) {
        if (clazz.second.getNumberOfSamples() > 0 && clazz.first.compare("") != 0) {
            te::da::DataSetPtr geometriesDS = m_currentLayer->getData(clazz.second.getSampleSet());

            geometriesDS->moveBeforeFirst();
            while (geometriesDS->moveNext()) {
                std::size_t geomPos = te::da::GetFirstPropertyPos(geometriesDS.get(),
                    te::dt::GEOMETRY_TYPE);
                te::gm::GeometryShrPtr geometry(geometriesDS->getGeometry(geomPos).release());
                std::pair<std::string, te::gm::GeometryShrPtr> geometryPair(clazz.first,
                    geometry);
                mapGeometries.insert(geometryPair);
            }
        }
    }
    return mapGeometries;
}

std::vector <te::gm::GeometryShrPtr> gdma::guicmp::SampleSelectionController::getLayerSelectedGeometries() {
    const te::da::ObjectIdSet* selectedIds = m_currentLayer->getSelected();
    std::vector <te::gm::GeometryShrPtr> vecGeoms;

    if (selectedIds && (selectedIds->size() > 0)) {
        te::da::DataSetPtr geometriesDS = m_currentLayer->getData(selectedIds);

        geometriesDS->moveBeforeFirst();
        while (geometriesDS->moveNext()) {
            std::size_t geomPos = te::da::GetFirstPropertyPos(geometriesDS.get(),
                te::dt::GEOMETRY_TYPE);
            te::gm::GeometryShrPtr geometry(geometriesDS->getGeometry(geomPos).release());
            vecGeoms.push_back(geometry);
        }
    }

    return vecGeoms;
}

std::vector <std::string> gdma::guicmp::SampleSelectionController::getLayerSelectedIds() {
    const te::da::ObjectIdSet* selectedIds = m_currentLayer->getSelected();
    std::vector <std::string> selectedIdsStrings;

    if (selectedIds && (selectedIds->size() > 0)) {
        for (auto &oid : *selectedIds) {
            selectedIdsStrings.push_back(oid->getValueAsString());
        }
    }

    return selectedIdsStrings;
}

bool gdma::guicmp::SampleSelectionController::saveSamplesInDataSource() {
    try {
        if (m_classesSet.saveSamplesInDataSource(m_currentLayer->getDataSourceId(), m_samplesColumnName)) {
            m_currentLayer->setOutOfDate();
            m_unsavedSamples = false;
            if (m_classesSet.getClass("") != 0)
                removeClass("");
            return true;
        } else {
            return false;
        }
    } catch (te::common::Exception& e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(e.what());
        return false;
    } catch (std::exception& e) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(e.what());
        return false;
    } catch (...) {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(TE_TR("Unknown Exception!"));
        return false;
    }
}

bool gdma::guicmp::SampleSelectionController::saveSamplesInCSV(std::string pathToCSV) {
    if (m_unsavedSamples) {
        if (!saveSamplesInDataSource())
            return false;
    }

    return gdma::dm::saveDataSourceToCSV(m_currentLayer->getDataSourceId(), pathToCSV,
        m_currentLayer->getDataSetName());
}

bool gdma::guicmp::SampleSelectionController::loadSamplesFromDataSource() {
    m_unsavedSamples = false;
    return m_classesSet.loadSamplesFromDataSource(m_currentLayer->getDataSourceId(),
        m_samplesColumnName);
}

bool gdma::guicmp::SampleSelectionController::hasUnsavedSamples() {
    return m_unsavedSamples;
}

gdma::cl::ClassDescriptionMap::iterator gdma::guicmp::SampleSelectionController::begin() {
    return m_classesSet.begin();
}

gdma::cl::ClassDescriptionMap::iterator gdma::guicmp::SampleSelectionController::end() {
    return m_classesSet.end();
}
