/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/guicomponents/sampleSelectionController.hpp

  \brief This file contains the SampleSelectionWidget class definition.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

#ifndef GEODMA_LIB_GUICMP_SAMPLESELECTIONCONTROLLER_HPP_
#define GEODMA_LIB_GUICMP_SAMPLESELECTIONCONTROLLER_HPP_

// TerraLib Includes
#include <terralib/maptools/AbstractLayer.h>

// STL includes
#include <map>
#include <set>
#include <string>
#include <vector>

// GeoDMA Includes
#include "geodmalib/guicomponents/config.hpp"
#include "geodmalib/classification/classDescriptionSet.hpp"

namespace gdma {
namespace guicmp {
class GDMAGUICMPEXPORT SampleSelectionController {
 public:
    /*
      \brief Default Constructor.
    */
    SampleSelectionController();

    /*
      \brief Default Destructor.
    */
    ~SampleSelectionController();

    /*
      \brief Set the current layer being used.
      \param[in] layer The new layer.
    */
    void setCurrentLayer(te::map::AbstractLayerPtr layer);

    /*
      \brief Get the current layer being used.
    */
    te::map::AbstractLayerPtr getCurrentLayer();

    /*
      \brief Clear the layer selection.
    */
    void clearLayerSelection();

    /*
      \brief Select geometries in the layer.
      \param[in] selectedIDs A multimap<std::string, std::string> indexed by the name of the
                             classes, containing the IDs of each object to be selected. The IDs
                             must be in the string format.
    */
    void selectSamplesInLayer(std::multimap<std::string, std::string> selectedIDs);

    /*
      \brief Create the column to store the selected samples.
      \param[in] columnName The name of the column being created.
    */
    bool createSampleColumn(std::string columnName);

    /*
      \brief Set the name of the column used to store the label of the samples.
      \param[in] columnName The name of the column.
    */
    void setSamplesColumnName(std::string columnName);

    /*
      \brief Get the name of the column used to store the label of the samples.
      \return The name of the column.
    */
    std::string getSamplesColumnName();

    /*
      \brief Add a new class to the class set \ref m_classesSet.
      \param[in] label The label of the new class.
      \param[in] color The color of the new class.
    */
    void addNewClass(std::string label, std::string color);

    /*
      \brief Remove a class from the class set \ref m_classesSet.
      \param[in] label The label of the class being removed.
    */
    void removeClass(std::string label);

    /*
      \brief Set the color of a given class.
      \param[in] classlabel The label of the class being set.
      \return Return true if the color was successfully set, false if the class does not exist.
    */
    bool setClassColor(std::string classLabel, std::string newColor);

    /*
      \brief Get the color of a given class.
      \param[in] classlabel The label of the class being retrieved.
      \return Return a string containig the color of the class, if it exists. If the class does not
              exist, return an empty string.
    */
    std::string getClassColor(std::string classLabel);

    bool setClassLabel(std::string classLabel, std::string newLabel);

    /*
      \brief Clear the class set \ref m_classesSet.
    */
    void clearClassesSet();

    /*
      \brief Get the size of the class set \ref m_classesSet.
      \return The size of the class set \ref m_classesSet.
    */
    int getNumberOfClasses();

    /*
      \brief Add selected samples in the class passed as parameter.
      \param[in] classLabel The label of the class to add the samples.
      \return A set containing the added samples.
    */
    const std::set<std::string> addSelectedSamples(std::string classLabel);

    /*
      \brief Get the number of samples of the class passed as parameter.
      \param[in] classLabel The label of the class.
      \return The number of samples contained in the class \ref classLabel.
    */
    int getNumberOfSamples(std::string classLabel);

    /*
      \brief Get the total number of samples of all the classes.
      \return The total number of samples contained in \ref m_classesSet.
    */
    int getTotalNumberOfSamples();

    /*
      \brief Get the ids of the samples in the class passed as parameter.
      \param[in] classLabel The label of the class.
      \return A set containing the ids of the samples as string.
    */
    const std::set<std::string>& getSampleIds(std::string classLabel);

    /*
      \brief Remove samples from a given class.
      \param[in] classLabel The label of the class.
      \param[in] oid The string containing the object id to be removed.
    */
    void removeSample(std::string classLabel, std::string oid);

    /*
      \brief Get the geometry objects selected as samples in all the classes.
      \return Returns a multimap indexed by the class name containing the shared pointers with
              the selected geometries of each class.
    */
    std::multimap<std::string, te::gm::GeometryShrPtr> getSampleGeometries();

    /*
      \brief Get the geometry objects selected in the layer.
      \return Returns a vector containing the shared pointers with the selected geometries.
    */
    std::vector<te::gm::GeometryShrPtr> getLayerSelectedGeometries();

    /*
      \brief Get the Ids as strings of geometry objects selected in the layer.
      \return Returns a vector containing strings corresponding to the IDs of the selected objects.
    */
    std::vector<std::string> getLayerSelectedIds();

    /*
      \brief Save the samples in the data source.
      \return Returns true if the operation was performed successfully, false otherwise.
    */
    bool saveSamplesInDataSource();

    bool saveSamplesInCSV(std::string pathToCSV);

    /*
      \brief Load existing samples from data source.
      \return Returns true if the operation was performed successfully, false otherwise.
    */
    bool loadSamplesFromDataSource();

    /*
      \brief Verify if there is unsaved samples in the layer.
      \return Returns the value of \ref m_unsavedSamples. It is true if there is samples that were
              not saved in the layer data source, false otherwise.
    */
    bool hasUnsavedSamples();

    gdma::cl::ClassDescriptionMap::iterator begin();

    gdma::cl::ClassDescriptionMap::iterator end();

 private:
    te::map::AbstractLayerPtr m_currentLayer;  //!< Current selected layer
    gdma::cl::ClassDescriptionSet m_classesSet;  //!< Current selected samples
    std::string m_samplesColumnName;  //!< The name of the column to save the samples.
    bool m_unsavedSamples;  //!< Flag used to control state of saved and unsaved samples.
};
}  // namespace guicmp
}  // namespace gdma

#endif  // GEODMA_LIB_GUICMP_SAMPLESELECTIONCONTROLLER_HPP_
