/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/guicomponents/sampleSelectionWidget.cpp

  \brief This file contains the SampleSelectionWidget class implementation.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// TerraLib Includes
#include <terralib/color/RGBAColor.h>
#include <terralib/core/translator/Translator.h>
#include <terralib/dataaccess/dataset/DataSet.h>
#include <terralib/dataaccess/dataset/DataSetType.h>
#include <terralib/dataaccess/dataset/ObjectId.h>
#include <terralib/dataaccess/dataset/ObjectIdSet.h>
#include <terralib/dataaccess/dataset/PrimaryKey.h>
#include <terralib/datatype/Property.h>
#include <terralib/geometry/GeometryProperty.h>
#include <terralib/dataaccess/utils/Utils.h>
#include <terralib/maptools/AbstractLayer.h>
#include <terralib/qt/widgets/canvas/Canvas.h>
#include <terralib/qt/widgets/canvas/MapDisplay.h>
#include <terralib/qt/widgets/tools/PolygonAcquire.h>
#include <terralib/qt/widgets/tools/RubberBand.h>
#include <terralib/qt/widgets/utils/ColorPickerToolButton.h>

// QT Includes
#include <QAbstractItemModel>
#include <QFileDialog>
#include <QInputDialog>
#include <QMenu>
#include <QMessageBox>
#include <QWidget>

// Boost Includes
#include <boost/filesystem.hpp>

// STL Includes
#include <list>
#include <map>
#include <memory>
#include <set>
#include <string>
#include <utility>
#include <vector>

// GeoDMA Includes
#include "geodmalib/classification/classDescription.hpp"
#include "geodmalib/classification/classDescriptionSet.hpp"
#include "geodmalib/common/utils.hpp"
#include "geodmalib/common/messageMediator.hpp"
#include "geodmalib/guicomponents/multiSelectionTool.hpp"
#include "geodmalib/guicomponents/sampleSelectionController.hpp"
#include "geodmalib/guicomponents/sampleSelectionWidget.hpp"
#include "ui_SampleSelectionWidgetForm.h"

#define CLASS_TREE_ITEM 0
#define SAMPLE_TREE_ITEM 1

gdma::guicmp::SampleSelectionWidget::SampleSelectionWidget(QWidget* parent, Qt::WindowFlags f) :
    QWidget(parent, f),
    m_ui(new Ui::SampleSelectionWidgetForm),
    m_actionGroup(0),
    m_saveToCSV(false) {
    m_ui->setupUi(this);

    m_noColumnName = TE_TR("--Select or create column--");
    m_newColumn = false;

    // Icons
    m_ui->m_addClassToolButton->setIcon(QIcon::fromTheme("list-add"));
    m_ui->m_removeEntryToolButton->setIcon(QIcon::fromTheme("remove_entry"));
    m_ui->m_selectSamplesToolButton->setIcon(QIcon::fromTheme("pointer-selection"));
    m_ui->m_addSamplesToolButton->setIcon(QIcon::fromTheme("list-add"));
    m_ui->m_addColumnToolButton->setIcon(QIcon::fromTheme("list-add"));
    m_ui->m_saveSamplesToolButton->setIcon(QIcon::fromTheme("save_samples"));

    // Setting color for color picker
    std::map<std::string, int> color = gdma::common::generateRandomColor();

    // Color Picker
    m_colorPicker = new te::qt::widgets::ColorPickerToolButton(this);
    m_colorPicker->setFixedSize(70, 24);
    m_colorPicker->setColor(QColor(color["red"], color["green"], color["blue"], 255));
    m_colorPicker->setWindowIcon(QIcon::fromTheme("icon_geodma"));

    // build form
    QGridLayout* colorPickerLayout = new QGridLayout(m_ui->m_colorWidget);
    colorPickerLayout->setContentsMargins(0, 0, 0, 0);
    colorPickerLayout->addWidget(m_colorPicker);

    QRegExpValidator* classNameValidator = new QRegExpValidator(QRegExp("[A-Za-z0-9_]+"));
    m_ui->m_classLabelLineEdit->setValidator(classNameValidator);

    m_ui->m_classesTreeWidget->setContextMenuPolicy(Qt::CustomContextMenu);
    m_ui->m_classesTreeWidget->setColumnCount(2);
    m_classColumnWidth = 160;
    QStringList ColumnNames;
    ColumnNames << "Classes" << "";
    m_ui->m_classesTreeWidget->setHeaderLabels(ColumnNames);
    m_ui->m_classesTreeWidget->setColumnWidth(0, m_classColumnWidth);

    connect(m_ui->m_inputLayerComboBox, SIGNAL(currentIndexChanged(int)), this,
        SLOT(onInputLayerComboBoxSelected(int)));
    connect(m_ui->m_samplesColumnComboBox, SIGNAL(currentIndexChanged(int)), this,
        SLOT(onSamplesColumnComboBoxSelected(int)));
    connect(m_ui->m_addClassToolButton, SIGNAL(clicked()), this,
        SLOT(onAddClassToolButtonClicked()));
    connect(m_ui->m_classLabelLineEdit, SIGNAL(returnPressed()), this,
        SLOT(onAddClassToolButtonClicked()));
    connect(m_ui->m_selectSamplesToolButton, SIGNAL(toggled(bool)), this,
        SLOT(onSelectSamplesToolButtonToggled(bool)));
    connect(m_ui->m_addSamplesToolButton, SIGNAL(clicked()), this,
        SLOT(onAddSamplesToolButtonClicked()));
    connect(m_ui->m_removeEntryToolButton, SIGNAL(clicked()), this,
        SLOT(onRemoveEntryToolButtonClicked()));
    connect(m_ui->m_addColumnToolButton, SIGNAL(clicked()), this,
        SLOT(onAddColumnToolButtonClicked()));
    connect(m_ui->m_classesTreeWidget, SIGNAL(customContextMenuRequested(QPoint)), this,
        SLOT(onCustomContextMenuRequested(QPoint)));

    connect(m_ui->m_classesTreeWidget, SIGNAL(itemClicked(QTreeWidgetItem*, int)),
        this, SLOT(onSampleTreeSelectionChanged(QTreeWidgetItem*, int)));

    connect(m_ui->m_saveSamplesToolButton, SIGNAL(clicked()), this,
        SLOT(onSaveSamplesToolButtonClicked()));

    connect(m_ui->m_csvOutputToolButton, SIGNAL(clicked()), this,
        SLOT(onCSVOutputToolButtonClicked()));

    // connect(m_ui->m_opacityToolButton, SIGNAL(clicked()), this, SLOT(onOpacityToolButtonClicked()));
}


gdma::guicmp::SampleSelectionWidget::~SampleSelectionWidget() {
}

Ui::SampleSelectionWidgetForm* gdma::guicmp::SampleSelectionWidget::getForm() const {
    return m_ui.get();
}

void gdma::guicmp::SampleSelectionWidget::setMapDisplay(te::qt::widgets::MapDisplay* mapDisplay) {
    m_mapDisplay = mapDisplay;
}

void gdma::guicmp::SampleSelectionWidget::setLayers(std::list<te::map::AbstractLayerPtr> layers) {
    m_layerList = layers;
    for (auto &layer : layers) {
        std::unique_ptr<te::da::DataSetType> dsType = layer.get()->getSchema();
        if (dsType->hasGeom()) {
            te::gm::GeometryProperty* geomProp = te::da::GetFirstGeomProperty(dsType.get());

            if (geomProp->getGeometryType() == te::gm::PolygonType ||
                geomProp->getGeometryType() == te::gm::MultiPolygonType) {
                m_vectorLayers.push_back(layer);
                m_ui->m_inputLayerComboBox->addItem(QString(layer.get()->getTitle().c_str()),
                    QVariant(layer.get()->getId().c_str()));
            }
        }
    }
}

void gdma::guicmp::SampleSelectionWidget::setCurrentLayer(te::map::AbstractLayerPtr layer) {
    m_controller.setCurrentLayer(layer);

    m_mapDisplay->setLayerList(m_layerList);

    te::gm::Envelope ext = m_mapDisplay->getExtent();
    m_mapDisplay->setExtent(ext, true);
}

void gdma::guicmp::SampleSelectionWidget::setActionGroup(QActionGroup* actionGroup) {
    m_actionGroup = actionGroup;
    enableSelectionAction();
}

void gdma::guicmp::SampleSelectionWidget::enableSelectionAction() {
    m_selectionAction = new QAction(this);
    m_selectionAction->setIcon(QIcon::fromTheme("pointer-selection"));
    m_selectionAction->setCheckable(true);
    m_selectionAction->setEnabled(true);

    m_ui->m_selectSamplesToolButton->setDefaultAction(m_selectionAction);
    m_ui->m_selectSamplesToolButton->setVisible(true);

    m_actionGroup->addAction(m_ui->m_selectSamplesToolButton->defaultAction());
}

void gdma::guicmp::SampleSelectionWidget::disableTools() {
    if (m_selectionAction != 0 && m_selectionAction->isChecked()) {
        m_selectionAction->setChecked(false);
        m_mapDisplay->setCursor(Qt::ArrowCursor);
        m_mapDisplay->setCurrentTool(0);
    }
}

void gdma::guicmp::SampleSelectionWidget::clearCanvas() {
    QPixmap* draft = m_mapDisplay->getDraftPixmap();
    te::qt::widgets::Canvas canvasInstance(draft);
    canvasInstance.clear();
    m_mapDisplay->refresh();
}

void gdma::guicmp::SampleSelectionWidget::drawSamples() {
    QPixmap* draft = m_mapDisplay->getDraftPixmap();
    draft->fill(Qt::transparent);

    const te::gm::Envelope& mapExt = m_mapDisplay->getExtent();

    te::qt::widgets::Canvas canvasInstance(draft);
    canvasInstance.setWindow(mapExt.m_llx, mapExt.m_lly, mapExt.m_urx, mapExt.m_ury);

    std::multimap<std::string, te::gm::GeometryShrPtr> sampleGeometries =
        m_controller.getSampleGeometries();

    for (auto &elem : sampleGeometries) {
        te::color::RGBAColor rgbaCl(m_controller.getClassColor(elem.first));

        canvasInstance.setPolygonContourWidth(2);
        canvasInstance.setPolygonContourColor(rgbaCl);
        canvasInstance.setPolygonFillColor(te::color::RGBAColor(rgbaCl.getRed(),
            rgbaCl.getGreen(), rgbaCl.getBlue(), 80));

        canvasInstance.draw(elem.second.get());
    }

    QList<QTreeWidgetItem*> selectedItems = m_ui->m_classesTreeWidget->selectedItems();

    if (selectedItems.count() < 1) {
        m_mapDisplay->repaint();
        return;
    }

    QTreeWidgetItem* item = selectedItems[0];

    if (item->type() != CLASS_TREE_ITEM) {
        item = item->parent();
    }

    std::string classLabel = item->text(0).toUtf8().data();
    te::color::RGBAColor rgbaSel(m_controller.getClassColor(classLabel));

    canvasInstance.setPolygonContourWidth(2);
    canvasInstance.setPolygonContourColor(te::color::RGBAColor(0, 0, 0, 255));
    canvasInstance.setPolygonFillColor(te::color::RGBAColor(rgbaSel.getRed(),
        rgbaSel.getGreen(), rgbaSel.getBlue(), 255));

    std::vector<te::gm::GeometryShrPtr> vecGeoms = m_controller.getLayerSelectedGeometries();

    for (auto &geom : vecGeoms) {
        canvasInstance.draw(geom.get());
    }

    // Draw Selected Objects
    m_mapDisplay->repaint();
}

bool gdma::guicmp::SampleSelectionWidget::hasUnsavedSamples() {
    return m_controller.hasUnsavedSamples();
}

bool gdma::guicmp::SampleSelectionWidget::saveSamples() {
    std::string columnName = m_ui->m_samplesColumnComboBox->currentData()
        .toString().toUtf8().data();

    if (columnName.compare(m_noColumnName.toUtf8().data()) == 0) {
        gdma::common::MessageMediator::getInstance().sendWarningMsg(
            TE_TR("Select a column to save the samples!"));
        return false;
    }

    if (m_controller.saveSamplesInDataSource()) {
        return true;
    } else {
        gdma::common::MessageMediator::getInstance().sendErrorMsg(
            TE_TR("Can not save samples in Data Source. Unknown error found!"));
        return false;
    }
}

bool gdma::guicmp::SampleSelectionWidget::loadSamples() {
    if (!m_controller.loadSamplesFromDataSource())
        return false;

    if (m_ui->m_classesTreeWidget->topLevelItemCount() > 0)
        m_ui->m_classesTreeWidget->clear();

    for (auto &clazz : m_controller) {
        std::set<std::string> sampleIds = clazz.second.getSampleSetAsString();

        // Update Classes Tree
        QPixmap pix(16, 16);
        pix.fill(QColor(clazz.second.getColor().c_str()));
        QIcon icon(pix);

        QTreeWidgetItem* itemClass = new QTreeWidgetItem(m_ui->m_classesTreeWidget,
            CLASS_TREE_ITEM);
        itemClass->setExpanded(true);
        itemClass->setText(0, clazz.first.c_str());
        itemClass->setIcon(0, icon);

        for (auto &id : sampleIds) {
            QString fullName;
            fullName.append("Sample ID: ");
            fullName.append(id.c_str());

            QString data(id.c_str());

            QTreeWidgetItem* subItemSample = new QTreeWidgetItem(itemClass, SAMPLE_TREE_ITEM);
            subItemSample->setIcon(0, QIcon::fromTheme("file-vector"));
            subItemSample->setText(0, fullName);
            subItemSample->setData(0, Qt::UserRole, data);
            itemClass->addChild(subItemSample);
        }

        QString labelNumSamples;
        labelNumSamples = QString::number(itemClass->childCount()) + " samples";
        itemClass->setText(1, labelNumSamples);
    }

    drawSamples();
}

void gdma::guicmp::SampleSelectionWidget::getSelectedSamplesFromTreeWidget(
    QList<QTreeWidgetItem*> selectedItems, std::multimap<std::string, std::string> &selectedStrs) {
    for (auto &item : selectedItems) {
        if (item->type() == SAMPLE_TREE_ITEM) {
            std::string classLabel = item->parent()->text(0).toUtf8().data();
            std::string oid = item->data(0, Qt::UserRole).toString().toUtf8().data();

            selectedStrs.insert(std::pair<std::string, std::string>(classLabel, oid));
        }
    }
}

bool gdma::guicmp::SampleSelectionWidget::saveSamplesInCSV() {
    if (m_saveToCSV && !m_ui->m_csvOutputLineEdit->text().isEmpty())
        return m_controller.saveSamplesInCSV(m_ui->m_csvOutputLineEdit->text().toUtf8().data());
    return false;
}

// ------------- Slots (Public)

void gdma::guicmp::SampleSelectionWidget::onInputLayerComboBoxSelected(int index) {
    std::string layerID = reinterpret_cast<QComboBox*>(QObject::sender())->itemData(
        index, Qt::UserRole).toString().toUtf8().data();

    m_controller.clearClassesSet();
    m_controller.clearLayerSelection();
    if (m_ui->m_classesTreeWidget->topLevelItemCount() > 0) {
        m_ui->m_classesTreeWidget->clear();
    }

    for (auto &layer : m_vectorLayers) {
        if (layerID == layer->getId().c_str())
            setCurrentLayer(layer);
    }

    drawSamples();

    // Fill the combo box containing the training column name.
    te::da::DataSetTypePtr dsType = m_controller.getCurrentLayer()->getSchema();
    std::vector<te::dt::Property*> propVector = dsType->getProperties();
    te::da::PrimaryKey* pk = dsType->getPrimaryKey();
    std::vector<te::dt::Property*> pkVector = pk->getProperties();
    m_ui->m_samplesColumnComboBox->clear();
    m_ui->m_samplesColumnComboBox->addItem(m_noColumnName, QVariant(m_noColumnName));


    for (auto &prop : propVector) {
        if (prop->getType() == te::dt::STRING_TYPE && prop->getName().compare("id") != 0) {
            for (auto &key : pkVector) {
                if (key->getName().compare(prop->getName()) != 0)
                    m_ui->m_samplesColumnComboBox->addItem(prop->getName().c_str(),
                        QVariant(prop->getName().c_str()));
            }
        }
    }
}

void gdma::guicmp::SampleSelectionWidget::onSamplesColumnComboBoxSelected(int index) {
    std::string columnName = reinterpret_cast<QComboBox*>(QObject::sender())->itemData(
        index, Qt::UserRole).toString().toUtf8().data();

    if (columnName.compare(m_noColumnName.toUtf8().data()) != 0 &&
        columnName.compare("") != 0) {
        m_controller.setSamplesColumnName(columnName);

        if (!m_newColumn) {
            QMessageBox::StandardButton reply;
            reply = QMessageBox::question(this, "Load Samples",
                "Do you want to load existing samples from this column?",
                QMessageBox::Yes | QMessageBox::No);

            if (reply == QMessageBox::Yes)
                loadSamples();
        }
    }
    m_newColumn = false;
}

void gdma::guicmp::SampleSelectionWidget::onAddClassToolButtonClicked() {
    if (m_ui->m_classLabelLineEdit->text().isEmpty()) {
        gdma::common::MessageMediator::getInstance().sendWarningMsg(TE_TR("Class Label is empty!"));
        return;
    }

    // get Class info
    QString label = m_ui->m_classLabelLineEdit->text();
    QString color = m_colorPicker->getColor().name();

    QList<QTreeWidgetItem*> existingItems =
        m_ui->m_classesTreeWidget->findItems(label, Qt::MatchExactly | Qt::MatchRecursive,
        0);

    if (existingItems.count() > 0) {
        gdma::common::MessageMediator::getInstance().sendWarningMsg(
            TE_TR("This label already exists!"));
        return;
    }

    // Update Classes Tree
    QPixmap pix(16, 16);
    pix.fill(m_colorPicker->getColor());
    QIcon icon(pix);

    QTreeWidgetItem* item = new QTreeWidgetItem(m_ui->m_classesTreeWidget, CLASS_TREE_ITEM);
    item->setText(0, label);
    item->setIcon(0, icon);
    QString labelNumSamples;
    labelNumSamples = QString::number(item->childCount()) + " samples";
    item->setText(1, labelNumSamples);

    m_controller.addNewClass(label.toUtf8().data(), color.toUtf8().data());

    m_ui->m_classesTreeWidget->addTopLevelItem(item);
    m_ui->m_classesTreeWidget->clearSelection();
    m_ui->m_classesTreeWidget->setItemSelected(item, true);
    m_ui->m_classLabelLineEdit->clear();
}

void gdma::guicmp::SampleSelectionWidget::onSelectSamplesToolButtonToggled(bool checked) {
    if (!checked) {
        // @TODO(Raian): Verify this
        /*m_ui->m_selectSamplesToolButton->setAutoExclusive(false);
        m_ui->m_selectSamplesToolButton->setChecked(false);
        m_mapDisplay->setCursor(Qt::ArrowCursor);
        m_mapDisplay->setCurrentTool(0);
        m_ui->m_selectSamplesToolButton->setAutoExclusive(true);*/
        return;
    }

    if (!m_actionGroup->actions().contains(m_selectionAction))
        m_actionGroup->addAction(m_ui->m_selectSamplesToolButton->defaultAction());

    gdma::guicmp::MultiSelectionTool* selectionTool =
        new gdma::guicmp::MultiSelectionTool(m_mapDisplay, Qt::ArrowCursor, m_vectorLayers, this);

    m_mapDisplay->setCurrentTool(selectionTool);
    m_controller.clearLayerSelection();
    m_mapDisplay->refresh();

    connect(selectionTool, SIGNAL(layerSelectedObjectsChanged(const te::map::AbstractLayerPtr&)),
        this, SLOT(onLayerSelectedObjectsChanged(const te::map::AbstractLayerPtr&)));
}

void gdma::guicmp::SampleSelectionWidget::onAddSamplesToolButtonClicked() {
    if (!m_selectionAction->isChecked()) {
        gdma::common::MessageMediator::getInstance().sendWarningMsg(
            TE_TR("Selection Tool is not activated."));
        return;
    }

    if (m_ui->m_inputLayerComboBox->count() == 0) {
        gdma::common::MessageMediator::getInstance().sendWarningMsg(
            TE_TR("Select an input Vector Layer."));
        return;
    }

    QList<QTreeWidgetItem*> selectedTreeItems = m_ui->m_classesTreeWidget->selectedItems();
    if (selectedTreeItems.size() == 0) {
        gdma::common::MessageMediator::getInstance().sendWarningMsg(TE_TR("Select a class first!"));
        return;
    }

    QTreeWidgetItem* itemClass = selectedTreeItems[0];
    if (itemClass->type() != CLASS_TREE_ITEM) {
        gdma::common::MessageMediator::getInstance().sendWarningMsg(TE_TR("Select a class first!"));
        return;
    }

    std::set<std::string> selectedIds =
        m_controller.addSelectedSamples(itemClass->text(0).toUtf8().data());

    for (auto &id : selectedIds) {
        QString fullName;
        fullName.append("Sample ID: ");
        fullName.append(id.c_str());

        QString data(id.c_str());

        QList<QTreeWidgetItem*> existingItems =
            m_ui->m_classesTreeWidget->findItems(fullName, Qt::MatchExactly|Qt::MatchRecursive,
                0);

        if (existingItems.count() == 0) {
            QTreeWidgetItem* subItemSample = new QTreeWidgetItem(itemClass, SAMPLE_TREE_ITEM);
            QFont font = subItemSample->font(0);
            font.setItalic(true);
            font.setBold(true);
            subItemSample->setIcon(0, QIcon::fromTheme("file-vector"));
            subItemSample->setText(0, fullName);
            subItemSample->setFont(0, font);
            subItemSample->setData(0, Qt::UserRole, data);
            itemClass->addChild(subItemSample);
            itemClass->setFont(0, font);
        } else {
            for (auto &item : existingItems) {
                if (QString::compare(item->parent()->text(0), itemClass->text(0)) != 0) {
                    item->parent()->removeChild(item);
                    itemClass->addChild(item);
                    QFont font = item->font(0);
                    font.setItalic(true);
                    font.setBold(true);
                    item->setFont(0, font);
                    m_ui->m_classesTreeWidget->repaint();
                }
            }
        }
    }

    QTreeWidgetItemIterator it(m_ui->m_classesTreeWidget);

    QString labelNumSamples;
    while (*it) {
        if ((*it)->type() == CLASS_TREE_ITEM){
            labelNumSamples = QString::number((*it)->childCount()) + " samples";
            (*it)->setText(1, labelNumSamples);
        }
        it++;
    }

    itemClass->setExpanded(true);
    drawSamples();
}

void gdma::guicmp::SampleSelectionWidget::onRemoveEntryToolButtonClicked() {
    QList<QTreeWidgetItem*> selectedItems = m_ui->m_classesTreeWidget->selectedItems();

    if (selectedItems.count() == 0) {
        gdma::common::MessageMediator::getInstance()
            .sendWarningMsg(TE_TR("Select items to remove."));
        return;
    }

    for (auto &item : selectedItems) {
        if (item->type() == CLASS_TREE_ITEM) {
            m_controller.removeClass(item->text(0).toUtf8().data());
            if (item->childCount() > 0) {
                QList<QTreeWidgetItem*> childItems = item->takeChildren();
                for (auto &child : childItems) {
                    if (child) {
                        delete child;
                        child = 0;
                    }
                }
            }
            if (item) {
                delete item;
                item = 0;
            }
        } else if(item->type() == SAMPLE_TREE_ITEM) {
            std::string classLabel = item->parent()->text(0).toUtf8().data();
            std::string oid = item->data(0, Qt::UserRole).toString().toUtf8().data();
            m_controller.removeSample(classLabel, oid);

            QTreeWidgetItem* itemClass = item->parent();

            delete item;
            item = 0;

            QFont font = itemClass->font(0);
            font.setItalic(true);
            font.setBold(true);
            itemClass->setFont(0, font);
        }
    }

    QTreeWidgetItemIterator it(m_ui->m_classesTreeWidget);

    QString labelNumSamples;
    while (*it) {
        if ((*it)->type() == CLASS_TREE_ITEM){
            labelNumSamples = QString::number((*it)->childCount()) + " samples";
            (*it)->setText(1, labelNumSamples);
        }
        it++;
    }

    drawSamples();
}

void gdma::guicmp::SampleSelectionWidget::onCustomContextMenuRequested(QPoint pos) {
    QTreeWidgetItem* item = m_ui->m_classesTreeWidget->itemAt(pos);
    std::stringstream msg;

    if (item && item->type() == CLASS_TREE_ITEM) {
        QMenu* popUpMenu = new QMenu(this);

        QMenu* labelMenu = new QMenu(popUpMenu);
        QAction* changeLabelAction = new QAction(tr("Change Label"), popUpMenu);
        changeLabelAction->setIcon(QIcon::fromTheme("changeLabel"));
        labelMenu->addAction(changeLabelAction);
        popUpMenu->addAction(changeLabelAction);

        QMenu* colorMenu = new QMenu(popUpMenu);
        QAction* changeColorAction = new QAction(tr("Change Color"), popUpMenu);
        changeColorAction->setIcon(QIcon::fromTheme("changeColor"));
        colorMenu->addAction(changeColorAction);
        popUpMenu->addAction(changeColorAction);

        popUpMenu->popup(m_ui->m_classesTreeWidget->viewport()->mapToGlobal(pos));
        connect(colorMenu, SIGNAL(triggered(QAction*)), this,
            SLOT(onChangeColorClicked()));
        connect(labelMenu, SIGNAL(triggered(QAction*)), this,
            SLOT(onChangeLabelClicked()));
    }
}

void gdma::guicmp::SampleSelectionWidget::onLayerSelectedObjectsChanged(
    const te::map::AbstractLayerPtr& layer) {
    std::vector<std::string> selectedIds = m_controller.getLayerSelectedIds();

    QTreeWidgetItemIterator it(m_ui->m_classesTreeWidget);

    while (*it) {
        if ((*it)->type() != CLASS_TREE_ITEM)
            (*it)->setSelected(false);
        it++;
    }

    for (auto &id : selectedIds) {
        QString fullName;
        fullName.append("Sample ID: ");
        fullName.append(id.c_str());

        QList<QTreeWidgetItem*> itemsToSelect = m_ui->m_classesTreeWidget->findItems(fullName,
            Qt::MatchExactly | Qt::MatchRecursive, 0);
        for (auto &item : itemsToSelect) {
            m_ui->m_classesTreeWidget->setItemSelected(item, true);
        }
    }

    drawSamples();
}

void gdma::guicmp::SampleSelectionWidget::onAddColumnToolButtonClicked() {
    if (m_ui->m_inputLayerComboBox->count() == 0) {
        gdma::common::MessageMediator::getInstance().sendWarningMsg(
            TE_TR("Select an input Vector Layer."));
        m_ui->m_selectSamplesToolButton->setChecked(false);
        return;
    }

    QInputDialog inputDlg(this);
    inputDlg.setWindowTitle(tr("Label Column Name"));
    inputDlg.setWindowIcon(QIcon::fromTheme("icon_geodma"));
    inputDlg.setLabelText(tr("Enter name of the new column:"));
    inputDlg.setTextValue("trainClass");
    inputDlg.resize(250, inputDlg.height());

    QString columnName;
    bool validColumn = true;

    do {
        if (inputDlg.exec() != QDialog::Accepted)
            return;

        columnName = inputDlg.textValue();
        validColumn = true;

        if (columnName.compare("") == 0) {
            validColumn = false;
            gdma::common::MessageMediator::getInstance().sendErrorMsg("Enter a name for the column!");
        } else {
            try {
                validColumn = m_controller.createSampleColumn(columnName.toUtf8().data());
            } catch (te::common::Exception& e) {
                validColumn = false;
                gdma::common::MessageMediator::getInstance().sendErrorMsg(e.what());
            }
        }
    } while (!validColumn);

    m_newColumn = true;

    m_ui->m_samplesColumnComboBox->addItem(columnName);
    int idxNewColumn = m_ui->m_samplesColumnComboBox->findText(columnName);
    m_ui->m_samplesColumnComboBox->setCurrentIndex(idxNewColumn);
    m_controller.setSamplesColumnName(columnName.toUtf8().data());
}

void gdma::guicmp::SampleSelectionWidget::onChangeColorClicked() {
    QTreeWidgetItem* item = m_ui->m_classesTreeWidget->selectedItems()[0];

    if (item->type() != CLASS_TREE_ITEM)
        return;

    std::string classLabel = item->text(0).toUtf8().data();
    std::string classColor = m_controller.getClassColor(classLabel);

    QColorDialog colorDlg(this);
    colorDlg.setCurrentColor(QColor(classColor.c_str()));
    colorDlg.setWindowIcon(QIcon::fromTheme("icon_geodma"));
    colorDlg.setWindowTitle(tr("Select color"));

    if (colorDlg.exec() != QDialog::Accepted)
        return;

    QColor newColor = colorDlg.selectedColor();

    // Update Classes Tree
    QPixmap pix(16, 16);
    pix.fill(newColor);
    QIcon icon(pix);

    item->setIcon(0, icon);
    m_ui->m_classesTreeWidget->repaint();

    m_controller.setClassColor(classLabel, newColor.name().toUtf8().data());
    drawSamples();
}

void gdma::guicmp::SampleSelectionWidget::onChangeLabelClicked() {
    if (m_ui->m_classesTreeWidget->selectedItems().size() == 0)
        return;

    QTreeWidgetItem* item = m_ui->m_classesTreeWidget->selectedItems()[0];
    std::string classLabel = item->text(0).toUtf8().data();

    if (item->type() != CLASS_TREE_ITEM)
        return;

    QInputDialog inputDlg(this);
    inputDlg.setWindowTitle(tr("Class Label"));
    inputDlg.setWindowIcon(QIcon::fromTheme("icon_geodma"));
    inputDlg.setLabelText(tr("New Label:"));
    inputDlg.setTextValue(classLabel.c_str());

    bool validLabel = true;
    QString newLabel;

    do {
        if (inputDlg.exec() != QDialog::Accepted)
            return;

        newLabel = inputDlg.textValue();

        QRegExp validator("[A-Za-z0-9_]+");
        if (!validator.exactMatch(newLabel)) {
            validLabel = false;
            std::stringstream msg;
            msg << "Label \"" << newLabel.toUtf8().data() << "\" contains special characters!";
            gdma::common::MessageMediator::getInstance().sendErrorMsg(msg.str());
        } else {
            validLabel = true;
        }
    } while (!validLabel);

    item->setText(0, newLabel);
    QFont font = item->font(0);
    font.setItalic(true);
    font.setBold(true);
    item->setFont(0, font);
    for (int i = 0; i < item->childCount(); i++) {
        QTreeWidgetItem *child = item->child(i);
        child->setFont(0, font);
    }

    m_ui->m_classesTreeWidget->repaint();

    m_controller.setClassLabel(classLabel, newLabel.toUtf8().data());
}

void gdma::guicmp::SampleSelectionWidget::onSampleTreeSelectionChanged(QTreeWidgetItem * item,
    int column) {
    QList<QTreeWidgetItem*> selectedItems = m_ui->m_classesTreeWidget->selectedItems();

    if (selectedItems.size() == 0)
        return;

    std::multimap<std::string, std::string> selected;
    getSelectedSamplesFromTreeWidget(selectedItems, selected);
    m_controller.selectSamplesInLayer(selected);

    drawSamples();
}

void gdma::guicmp::SampleSelectionWidget::onSaveSamplesToolButtonClicked() {
    if (!saveSamples())
        return;

    QTreeWidgetItemIterator it(m_ui->m_classesTreeWidget);

    while (*it) {
        QFont font = (*it)->font(0);
        font.setItalic(false);
        font.setBold(false);
        (*it)->setFont(0, font);
        it++;
    }
}

void gdma::guicmp::SampleSelectionWidget::onCSVOutputToolButtonClicked() {
    m_ui->m_csvOutputLineEdit->clear();

    QString fileName = QFileDialog::getSaveFileName(this, tr("Save as..."),
        QString(), tr("CSV File (*.csv *.CSV);;"), 0);

    boost::filesystem::path outfile(fileName.toUtf8().data());
    std::string aux = outfile.leaf().string();
    aux = outfile.string();
    m_ui->m_csvOutputLineEdit->setText(aux.c_str());
    m_saveToCSV = true;
}

// void gdma::guicmp::SampleSelectionWidget::onOpacityToolButtonClicked() {
//    m_ui->m_opacityToolButton->setPopupMode(QToolButton::InstantPopup);
//
//    QWidget *sliderPopUp = new QWidget(this);
//
//    QSlider *slider= new QSlider(Qt::Vertical, sliderPopUp);
//    slider->setRange(0, 255);
//    slider->setValue(80);
//
//    QLabel* label = new QLabel(sliderPopUp);
//    label->setAlignment(Qt::AlignCenter);
//    label->setNum(80);
//    label->setMinimumWidth(label->sizeHint().width());
//
//    typedef void(QLabel::*IntSlot)(int);
//    connect(slider, &QAbstractSlider::valueChanged, label, static_cast<IntSlot>(&QLabel::setNum));
//
//    QBoxLayout *popupLayout = new QVBoxLayout(sliderPopUp);
//    popupLayout->setMargin(2);
//    popupLayout->addWidget(slider);
//    popupLayout->addWidget(label);
//
//    QWidgetAction *action = new QWidgetAction(m_ui->m_opacityToolButton);
//    action->setDefaultWidget(sliderPopUp);
//
//    // m_ui->m_opacityToolButton->addAction(action);
//
//    QMenu* menu = new QMenu(this);
//    menu->addAction(action);
//    m_ui->m_opacityToolButton->setMenu(menu);
//    // menu->popup(m_ui->m_opacityToolButton->pos());
//    // m_ui->m_opacityToolButton->addAction(action);
// }
