/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/guicomponents/sampleSelectionWidget.hpp

  \brief This file contains the SampleSelectionWidget class definition.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

#ifndef GEODMA_LIB_GUICMP_SAMPLESELECTIONWIDGET_HPP_
#define GEODMA_LIB_GUICMP_SAMPLESELECTIONWIDGET_HPP_

// TerraLib Includes
#include <terralib/maptools/AbstractLayer.h>
#include <terralib/qt/widgets/utils/ColorPickerToolButton.h>
#include <terralib/qt/widgets/canvas/MapDisplay.h>
#include <terralib/qt/widgets/rp/RpToolsWidget.h>

// Qt Includes
#include <QWidget>
#include <QTreeWidget>

// STL Includes
#include <list>
#include <map>
#include <memory>
#include <string>

// GeoDMA Includes
#include "geodmalib/classification/classDescriptionSet.hpp"
#include "geodmalib/guicomponents/config.hpp"
#include "geodmalib/guicomponents/sampleSelectionController.hpp"

namespace Ui { class SampleSelectionWidgetForm; }

namespace gdma {
namespace guicmp {
class GDMAGUICMPEXPORT SampleSelectionWidget : public QWidget {
    Q_OBJECT

 public:
    /*!
      \brief Default Constructor
    */
    explicit SampleSelectionWidget(QWidget* parent = 0, Qt::WindowFlags f = 0);

    /*!
      \brief Default Destructor
    */
    ~SampleSelectionWidget();

    /*!
      \brief Used to get access to the Widget Components
    */
    Ui::SampleSelectionWidgetForm* getForm() const;

    /*!
      \brief Sets the internal pointer to the Application Map Display.
      \param[in] A pointer to the Application Map Display.
    */
    void setMapDisplay(te::qt::widgets::MapDisplay* mapDisplay);

    void setLayers(std::list<te::map::AbstractLayerPtr> layers);

    void setCurrentLayer(te::map::AbstractLayerPtr layer);

    /*!
      \brief This method is used to configure the actions to interact with the map display
      \param[in] The action tools necessary to the interactions.
    */
    void setActionGroup(QActionGroup* actionGroup);

    void enableSelectionAction();

    void disableTools();

    void clearCanvas();

    // protected:
    void drawSamples();

    bool hasUnsavedSamples();

    bool saveSamples();

    bool loadSamples();

    void getSelectedSamplesFromTreeWidget(QList<QTreeWidgetItem*> selectedItems,
        std::multimap<std::string, std::string> &selectedStrs);

    bool saveSamplesInCSV();

 public slots: //NOLINT
    void onInputLayerComboBoxSelected(int index);
    void onSamplesColumnComboBoxSelected(int index);
    void onAddClassToolButtonClicked();
    void onSelectSamplesToolButtonToggled(bool checked);
    void onAddSamplesToolButtonClicked();
    void onRemoveEntryToolButtonClicked();
    void onCustomContextMenuRequested(QPoint pos);

    void onLayerSelectedObjectsChanged(const te::map::AbstractLayerPtr& layer);
    void onAddColumnToolButtonClicked();

    void onChangeColorClicked();
    void onChangeLabelClicked();

    void onSampleTreeSelectionChanged(QTreeWidgetItem * item, int column);
    void onSaveSamplesToolButtonClicked();
    void onCSVOutputToolButtonClicked();

 private:
    std::unique_ptr<Ui::SampleSelectionWidgetForm> m_ui;  //!< User interface
    te::qt::widgets::ColorPickerToolButton* m_colorPicker;  //!< Button for color selection
    te::qt::widgets::MapDisplay* m_mapDisplay;  //!< Map Display from TerraView

    QActionGroup* m_actionGroup;
    QAction* m_selectionAction;

    std::list<te::map::AbstractLayerPtr> m_vectorLayers;  //!< All vector layers of the application
    std::list<te::map::AbstractLayerPtr> m_layerList;  //!< List containing all layers

    QString m_noColumnName;  //!< String used for empty selection in column name combo box.
    bool m_newColumn;  //!< Flag to control when the column was just created or already exists.
    bool m_saveToCSV;  //< Flag to control when save CSV file with samples.
    int m_classColumnWidth;

    gdma::guicmp::SampleSelectionController m_controller;
};
}  // namespace guicmp
}  // namespace gdma

#endif  // GEODMA_LIB_GUICMP_SAMPLESELECTIONWIDGET_HPP_
