#include "semanticAnalysisWidget.hpp"
#include "../../semanticanalysis/projectIOFactory.hpp"
#include "ui_semanticAnalysisWidgetForm.h"
#include "../../semanticanalysis/project.hpp"

#include <QMessageBox>
#include <QCloseEvent>
#include <QFileDialog>
#include <QString>
#include <QProgressBar>

namespace gdma {
namespace guicmp {
SemanticAnalysisWidget::SemanticAnalysisWidget(QWidget* parent, Qt::WindowFlags f)
  : QWidget(parent, f), m_timer( this )
{
  m_executedTasksNumber = 0;
  m_totalTasksNumber = 0;

  m_uiPtr.reset(new Ui::SemanticAnalysisWidgetForm);
  m_uiPtr->setupUi(this);

  m_uiPtr->m_statusLineEdit->setText( "INACTIVE" );

  connect(m_uiPtr->m_newPushButton, SIGNAL(clicked()), this, SLOT(onNewPushButtonClicked()));
  connect(m_uiPtr->m_openPushButton, SIGNAL(clicked()), this, SLOT(onOpenPushButtonClicked()));
  connect(m_uiPtr->m_savePushButton, SIGNAL(clicked()), this, SLOT(onSavePushButtonClicked()));
  connect(m_uiPtr->m_startPushButton, SIGNAL(clicked()), this, SLOT(onStartPushButtonClicked()));
  connect(m_uiPtr->m_stopPushButton, SIGNAL(clicked()), this, SLOT(onStopPushButtonClicked()));
  connect(m_uiPtr->m_stepPushButton, SIGNAL(clicked()), this, SLOT(onStepPushButtonClicked()));
  connect(m_uiPtr->m_contextsPushButton, SIGNAL(clicked()), this, SLOT(onContextsPushButtonClicked()));
  connect(m_uiPtr->m_analysisPushButton, SIGNAL(clicked()), this, SLOT(onAnalysisPushButtonClicked()));
  connect(&m_timer, SIGNAL(timeout()), this, SLOT(onTimerTimeOut()));
}

SemanticAnalysisWidget::~SemanticAnalysisWidget()
{
}

SemanticAnalysisWidget::ExecuterStatus SemanticAnalysisWidget::getExecuterStatus() const
{
  return ( m_executer.getExecuterStatus() == gdma::sa::Executer::ACTIVE_EXECUTER )
    ? ACTIVE_EXECUTER : INACTIVE_EXECUTER;
}

void SemanticAnalysisWidget::onNewPushButtonClicked()
{
  if( m_executer.getExecuterStatus() == gdma::sa::Executer::ACTIVE_EXECUTER )
  {
    QMessageBox::critical( this, "GeoDMA", "Analysis still active" );
  }
  else
  {
    QMessageBox::critical( this, "GeoDMA", "Not implemented yet" );
  }
}

void SemanticAnalysisWidget::onOpenPushButtonClicked()
{
  if( m_executer.getExecuterStatus() == gdma::sa::Executer::ACTIVE_EXECUTER )
  {
    QMessageBox::critical( this, "GeoDMA", "Analysis still active" );
  }
  else
  {
    QString fileName = QFileDialog::getOpenFileName(this, tr("Open File"),
      ".", tr("Project files (*.txt)"));

    if( !fileName.isEmpty() )
    {
      if( ! m_executer.finalize() )
      {
        QMessageBox::critical( this, "GeoDMA", "Analysis still active" );

        return;
      }

      m_project.reset();

      std::string errorMessage;
      if( ! gdma::sa::ProjectIOFactory::load( fileName.toStdString(),
        m_project, errorMessage ) )
      {
        QMessageBox::critical( this, "GeoDMA", ( std::string( "Project load error" )
          + ( errorMessage.empty() ? std::string() :
          std::string(" [") + errorMessage + "]" ) ).c_str() );
      }
      else
      {
        updateAllAnalysisDraw( true );
      }
    }
  }
}

void SemanticAnalysisWidget::onSavePushButtonClicked()
{
  if( m_executer.getExecuterStatus() == gdma::sa::Executer::ACTIVE_EXECUTER )
  {
    QMessageBox::critical( this, "GeoDMA", "Analysis still active" );
  }
  else
  {
    QMessageBox::critical( this, "GeoDMA", "Not implemented yet" );
  }
}

void SemanticAnalysisWidget::onStartPushButtonClicked()
{
  if( m_executer.getExecuterStatus() == gdma::sa::Executer::ACTIVE_EXECUTER )
  {
    QMessageBox::critical( this, "GeoDMA", "Analysis still active" );
  }
  else
  {
    setEnabled( false );
    m_uiPtr->m_stopPushButton->setEnabled( true );

    if( !m_executer.initialize( m_project, gdma::sa::Executer::CONCURRENT ) )
    {
      setEnabled( true );
      QMessageBox::critical( this, "GeoDMA", "Analysis initialization error" );
    }
    else
    {
      m_executer.setCallBack( *this );
      m_executedTasksNumber = 0;
      m_totalTasksNumber = m_executer.getTasksCount();
      m_executer.startExecution();
      m_timer.start( 1000 );
    }
  }
}

void SemanticAnalysisWidget::onStopPushButtonClicked()
{
  m_executer.stopExecution();
  m_timer.stop();
  setEnabled( true );
}

void SemanticAnalysisWidget::onStepPushButtonClicked()
{
  if( m_executer.getExecuterStatus() == gdma::sa::Executer::ACTIVE_EXECUTER )
  {
    QMessageBox::critical( this, "GeoDMA", "Analysis still active" );
  }
  else
  {
    setEnabled( false );

    if( !m_executer.isInitialized() )
    {
      if( !m_executer.initialize( m_project, gdma::sa::Executer::SINGLE ) )
      {
        setEnabled( true );
        QMessageBox::critical( this, "GeoDMA", "Analysis initialization error" );
        return;
      }

      m_executedTasksNumber = 0;
      m_totalTasksNumber = m_executer.getTasksCount();

      m_executer.setCallBack( *this );
    }

    m_executer.startExecution();

    updateAllAnalysisDraw( false );

    setEnabled( true );
  }
}

void SemanticAnalysisWidget::onContextsPushButtonClicked()
{
  if( m_executer.getExecuterStatus() == gdma::sa::Executer::ACTIVE_EXECUTER )
  {
    QMessageBox::critical( this, "GeoDMA", "Analysis still active" );
  }
  else
  {
    QMessageBox::critical( this, "GeoDMA", "Not implemented yet" );
  }
}

void SemanticAnalysisWidget::onAnalysisPushButtonClicked()
{
  if( m_executer.getExecuterStatus() == gdma::sa::Executer::ACTIVE_EXECUTER )
  {
    QMessageBox::critical( this, "GeoDMA", "Analysis still active" );
  }
  else
  {
    QMessageBox::critical( this, "GeoDMA", "Not implemented yet" );
  }
}

void SemanticAnalysisWidget::closeEvent(QCloseEvent* eventPtr )
{
  if( m_executer.getExecuterStatus() == gdma::sa::Executer::ACTIVE_EXECUTER )
  {
    QMessageBox::critical( this, "GeoDMA", "Analysis still active" );
    eventPtr->ignore();
  }
  else
  {
    eventPtr->accept();
  }
}

void SemanticAnalysisWidget::verticeTaskStarted( const std::string& analysisName,
  const std::string& graphName, const std::string& verticeName )
{
}

void SemanticAnalysisWidget::verticeTaskExecuted( const std::string& analysisName,
  const std::string& graphName, const std::string& verticeName,
  const gdma::tp::TaskExecutionResult::BasicResult& basicResut )
{
  std::unique_lock<std::mutex> lock( m_mutex );

  if( basicResut == gdma::tp::TaskExecutionResult::SUCCESS )
  {
    ++m_executedTasksNumber;
  }
  else
  {
    m_executedTasksNumber = m_totalTasksNumber;
  }
}

void SemanticAnalysisWidget::onTimerTimeOut()
{
  std::unique_lock<std::mutex> lock( m_mutex );

  if( m_executedTasksNumber < m_totalTasksNumber )
  {
    m_uiPtr->m_statusLineEdit->setText( "ACTIVE" );
    m_uiPtr->m_progressBar->setMaximum( m_totalTasksNumber );
    m_uiPtr->m_progressBar->setValue( m_executedTasksNumber );

    lock.unlock();

    updateAllAnalysisDraw( false );
  }
  else
  {
    m_timer.stop();

    m_uiPtr->m_statusLineEdit->setText( "INACTIVE" );
    m_uiPtr->m_progressBar->setMaximum( 100 );
    m_uiPtr->m_progressBar->setValue( 0 );
    setEnabled( true );

    lock.unlock();

    m_executer.stopExecution();

    updateAllAnalysisDraw( false );
  }
}

void SemanticAnalysisWidget::updateAllAnalysisDraw(
  bool recreateTabs )
{
  std::unique_lock<std::mutex> lock( m_mutex );

  if( recreateTabs )
  {
    while( m_uiPtr->m_analysisTabWidget->count() )
    {
      delete m_uiPtr->m_analysisTabWidget->widget(0);
    }
  }

  // creating tabs if empty

  if( m_uiPtr->m_analysisTabWidget->count() == 0 )
  {
    std::vector< std::string > analysisNames;
    m_project.getAnalysisNames( analysisNames );

    for( std::size_t analysisNamesIdx = 0 ; analysisNamesIdx <
      analysisNames.size() ; ++analysisNamesIdx )
    {

    }
  }

  std::cout << std::endl << "updateAnalysisDraw" << std::endl;
}

}  // namespace guicmp
}  // namespace gdma
