/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/semanticanalysis/analysis.hpp
*/

#ifndef GEODMA_LIB_SEMANTICANALYSIS_ANALYSIS_HPP_
#define GEODMA_LIB_SEMANTICANALYSIS_ANALYSIS_HPP_

#include <string>
#include <mutex>

// GeoDMA Includes
#include "config.hpp"
#include "graph.hpp"
#include "context.hpp"

namespace gdma {
namespace sa {

class Project;

/*!
  \class Analysis

  \brief A basic concept of a Analysis.
*/
class GDMASEMANTICANALYSISEXPORT Analysis {
 public:
    Analysis();

    Analysis( const Analysis& other );

    ~Analysis();

    Analysis& operator=( const Analysis& other );

    /*!
      \brief Return the Analysis name.
      \return Return the Analysis name.
      \note This is a thread safe method.
     */
    const std::string& getName() const;

    /*!
      \brief Set the Analysis name.
      \param newName The Analysis name.
      \note This is a thread safe method.
     */
    void setName( const std::string& newName );

    /*!
      \brief Return the Analysis graph.
      \return Return the Analysis graph.
      \note This is a thread safe method.
     */
    const Graph& getGraph() const;

    /*!
      \brief Return the Analysis graph.
      \return Return the Analysis graph.
      \note This is a thread safe method.
     */
    Graph& getGraph();

    /*!
      \brief Set the Analysis graph.
      \param newGraph The Analysis graph.
      \return true if ok, false on errors.
      \note This is a thread safe method.
     */
    bool setGraph( const Graph& newGraph );

    /*!
      \brief Return the Analysis Context.
      \return Return the Analysis Context.
      \note This is a thread safe method.
     */
    const Context& getContext() const;

    /*!
      \brief Set the Analysis Context.
      \param newContext The Analysis Context.
      \note This is a thread safe method.
     */
    void setContext( const Context& newContext );

    /*!
      \brief Set a output directory to this Analysis and its graph.
      \param outDir Output directory to this Analysis.
      \return true if OK, false on errors.
      \note This is a thread safe method.
      \note The provided path must be canonical and system complete.
      */
    bool setOutputDirectoryName( const std::string& outDirName );

    /*!
      \brief Returns the output directory to this Analysis.
      \return Returns the output directory to this Analysis.
      \note This is a thread safe method.
      */
    const std::string& getOutputDirectoryName() const;

    /*!
      \brief Enable (true) or disable (false) the use of result cache.
      \param enabled Enable (true) or disable (false) the use of result cache.
      \note This is a thread safe method.
      */
    void enableCache( const bool enabled );

    /*!
      \brief Returns true if cache is enabled, false otherwise.
      \return Returns true if cache is enabled, false otherwise.
      \note This is a thread safe method.
      */
    bool isCacheEnabled() const;

    /*!
      \brief Reset this instance to its initial state.
      \note This is a thread safe method.
      */
    void reset();

    /*!
      \brief Set this Analysis's related Project.
      \param project Set this Analysis's related Project.
      \note This is a thread safe method.
      */
    void setProject( const Project& project );

    /*!
      \brief Returns a pointer to the related Project or null if none is related.
      \return Returns a pointer to the related Project or null if none is related.
      \note This is a thread safe method.
      */
    Project const* getProject() const;

 private:
    /*! \brief Tells if result cache must be used. */
    bool m_enableCache;

    /*! \brief A pointer to this analysis's related Project. */
    Project const* m_projectPtr;

    /*! \brief Analisis name. */
    std::string m_name;

    /*! \brief Analysis output directory. */
    std::string m_outputDirectoryName;

    /*! \brief Analisis graph. */
    Graph m_graph;

    /*! \brief Analisis context. */
    Context m_context;

    /*! \brief Internal vertice mutex. */
    mutable std::mutex m_mutex;
};

}  // end namespace sa
}  // end namespace gdma

#endif  // GEODMA_LIB_SEMANTICANALYSIS_ANALYSIS_HPP_
