/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/semanticanalysis/context.hpp
*/

#ifndef GEODMA_LIB_SEMANTICANALYSIS_CONTEXT_HPP_
#define GEODMA_LIB_SEMANTICANALYSIS_CONTEXT_HPP_

#include <terralib/core/uri/URI.h>

#include <string>
#include <map>
#include <mutex>

// GeoDMA Includes
#include "config.hpp"

namespace gdma {
namespace sa {

/*!
  \class Context

  \brief A basic concept of a Context.
*/
class GDMASEMANTICANALYSISEXPORT Context {
 public:
    Context();

    Context( const Context& other );

    Context& operator=( const Context& other );

    ~Context();

    /*!
      \brief Return the Context name.
      \return Return the Context name.
      \note This is a thread safe method.
     */
    const std::string& getName() const;

    /*!
      \brief Set the Context name.
      \param contextName The Context name.
      \note This is a thread safe method.
     */
    void setName( const std::string& contextName );

    /*!
      \brief Return the a Context resource.
      \param name Resource name.
      \param uri Resouce URI.
      \return Return true if the resource was found, false otherwise.
      \note This is a thread safe method.
     */
    bool getResource( const std::string& name, te::core::URI& uri ) const;

    /*!
      \brief Add a new resource.
      \param name The resource name.
      \param uri The resource URI.
      \return true if ok, false on errors.
      \note This is a thread safe method.
     */
    bool addResource( const std::string& name, const te::core::URI& uri );

    /*!
      \brief Reset this instance to its initial state.
      \note This is a thread safe method.
      */
    void reset();

     /*!
      \brief Return the a Context resources count.
      \return Return the a Context resources count.
      \note This is a thread safe method.
     */
    std::size_t getResourcesCount() const;

    /*!
      \brief Set a reference directory to this Context (where the Context file is located).
      \param outDir Reference directory to this Context (where the Context file is located)
      \note This is a thread safe method.
      */
    void setContextDirectoryName( const std::string& dirName );

    /*!
      \brief Returns the reference directory to this Context (where the Context file is located).
      \return Returns reference directory to this Context (where the Context file is located).
      \note This is a thread safe method.
      */
    const std::string& getContextDirectoryName() const;

 protected:
    /*!
      \brief Context name.
     */
    std::string m_contextName;

    /*! \brief Context directory (where the Context file is located). */
    std::string m_contextDirectoryName;

    /*!
      \brief Context resources.
     */
    std::map< std::string, te::core::URI > m_resources;

    /*! \brief Internal vertice mutex. */
    mutable std::mutex m_mutex;
};

}  // end namespace sa
}  // end namespace gdma

#endif  // GEODMA_LIB_SEMANTICANALYSIS_CONTEXT_HPP_
