/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/semanticanalysis/executer.hpp
*/

#ifndef GEODMA_LIB_SEMANTICANALYSIS_EXECUTER_HPP_
#define GEODMA_LIB_SEMANTICANALYSIS_EXECUTER_HPP_

// GeoDMA Includes
#include "executerCallBack.hpp"
#include "config.hpp"
#include "project.hpp"

namespace gdma {
namespace sa {

/*!
  \class Executer

  \brief A basic concept of a project executer.
*/
class GDMASEMANTICANALYSISEXPORT Executer {
 public:
    /*!
      \enum ExecMode Execution mode.
    */
    enum ExecMode {
      SINGLE,  //!< Just one task is executed on each call to the startExecution method.
      SEQUENTIAL,  //!< All tasks will be executed sequentially (no threads).
      CONCURRENT  //!< All tasks will be executed concurrently.
    };

    /*!
      \enum ExecuterStatus Executer status.
    */
    enum ExecuterStatus {
      INACTIVE_EXECUTER,  //!< Inactive executer.
      ACTIVE_EXECUTER  //!< Executer active.
    };

    virtual ~Executer();

    /*!
      \brief Initialize this executer instance.
      \param project A reference to the project to execute (it must remain valid while this instance is alive).
      \param execMode Execution mode.
      \return true if ok, false on errors.
      */
    virtual bool initialize( Project& project, const ExecMode execMode ) = 0;

    /*!
      \brief Finalize this executer instance.
      \return true if ok, false on errors.
      */
    virtual bool finalize() = 0;

    /*!
      \brief Start tasks execution.
    */
    virtual void startExecution() = 0;

    /*!
      \brief Stop tasks execution.
    */
    virtual void stopExecution() = 0;

    /*!
      \brief Hold the calling process and wait for the execution to end.
    */
    virtual void waitExecution() = 0;

    /*!
      \brief Configures the executer to call the informed call back when necessary.
      \param newExecCallBack The call back object.
      \note The callback maybe called asynchronously within a thread.
      \note This is a thread-safe method.
    */
    void setCallBack(ExecuterCallBack& newExecCallBack);

    /*!
      \brief Returns a pointer to the current call back or null if none exists.
      \return Returns a pointer to the current call back or null if none exists.
    */
    ExecuterCallBack* getCallBack();

    /*!
      \brief Returns the current executer status.
      \return Returns the current executer status.
      \note This is a thread-safe method.
    */
    virtual ExecuterStatus getExecuterStatus() const = 0;

    /*!
      \brief Returns true if this instance is initialized.
      \return Returns true if this instance is initialized.
      \note This is a thread-safe method.
    */
    virtual bool isInitialized() const = 0;

    /*!
      \brief Returns current stored tasks count.
      \return Returns current stored tasks count.
      \note This is a thread-safe method.
    */
    virtual unsigned int getTasksCount() const = 0;

 protected:
    Executer();

    /*!
      \brief Reset this instance to its initial settings.
    */
    void reset();

 private:
    ExecuterCallBack* m_execCallBackPtr;  //!< A pointer to a callback called when necessary.

    Executer(const Executer& other);
};

}  // end namespace sa
}  // end namespace gdma

#endif  // GEODMA_LIB_SEMANTICANALYSIS_EXECUTER_HPP_
