/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/semanticanalysis/geometryfunctions.hpp
*/

#ifndef GEODMA_LIB_SEMANTICANALYSIS_GEOMETRYFUNCTIONS_HPP_
#define GEODMA_LIB_SEMANTICANALYSIS_GEOMETRYFUNCTIONS_HPP_

#include <boost/ptr_container/ptr_vector.hpp>
//
#include <terralib/geometry.h>
#include <terralib/raster.h>
// #include <terralib/core.h>
//
// #include <string>
#include <memory>
#include <vector>
// #include <map>

// GeoDMA Includes
#include "config.hpp"
#include "typedefs.hpp"

namespace gdma {
namespace sa {

  /*!
    \brief Rasterize a set of vectors.
    \param inputVectors Input vectors of geometries.
    \param inputVectorsIDs Input vectors IDs.
    \param outXRes Output raster X resolution.
    \param outYRes Output raster Y resolution.
    \param outSRID Output raster SRID.
    \param outRasterizedResultPtr Output rasterized result.
    \return true if ok, false on errors.
    \note Only Polygons are processed.
    \note All geometries must have the same SRID.
    */
  GDMASEMANTICANALYSISEXPORT bool rasterizeVectors(
    const std::vector< boost::shared_ptr< boost::ptr_vector< te::gm::Geometry > > >& inputVectors,
    const std::vector< GraphVerticeIDT >& inputVectorsIDs,
    const double outXRes,
    const double outYRes,
    const int outSRID,
    std::unique_ptr< te::rst::Raster >& outRasterizedResultPtr );

  /*!
    \brief Do spatial resolution over the input vector sets.
    \param geomGroups Input vectors groups.
    \param nonIntersectedGroups Ouput non-intersected vector groups.
    \param outIntersectedVectors Output intersected vectors (union of all overlapped areas).
    \return true if ok, false on errors.
    \note nonIntersectedGroups.size() == geomGroups.size()
    \note Elements within the same group are supposed to not have spatial overlap.
    */
  GDMASEMANTICANALYSISEXPORT bool geometriesSetsSpatialResolve(
    const std::vector< boost::shared_ptr< boost::ptr_vector< te::gm::Geometry > > >& geomGroups,
    std::vector< boost::shared_ptr< boost::ptr_vector< te::gm::Geometry > > >& nonIntersectedGroups,
    boost::ptr_vector< te::gm::Geometry >& outIntersectedVectors );

  /*!
    \brief Create a geometries vector from another.
    \param inputGeomsVec Input geometries vector.
    \param outputGeomsVec Output geometries vector( it will take the ownership of all input vector geometries.
  */
  GDMASEMANTICANALYSISEXPORT void convertGeomsVectors( const std::vector< te::gm::Geometry* >& inputGeomsVec,
    boost::ptr_vector< te::gm::Geometry >& outputGeomsVec );

  /*!
    \brief Create a geometries vector from another.
    \param inputGeomsVec Input geometries vector.
    \param outputGeomsVec Output geometries vector
    \param release if true, the input vector will give up the ownership of all geometries and the caller of this method must take the ownership of those geometries.
  */
  GDMASEMANTICANALYSISEXPORT void convertGeomsVectors(
    boost::ptr_vector< te::gm::Geometry >& inputGeomsVec,
    std::vector< te::gm::Geometry* >& outputGeomsVec,
    bool release );

  /*!
    \brief Performs intersection between two geometries.
    \param geom1 Input geometry 1.
    \param geom1 Input geometry 2.
    \param result Result geometry (or a null pointer if there is none).
    \note Multi-geometries accepted.
    \note Geometries with distinct SRID accepted.
    \note geom1 SRID is taken as reference (it will not be changed).
  */
  GDMASEMANTICANALYSISEXPORT void geometryIntersection( const te::gm::Geometry& geom1,
     const te::gm::Geometry& geom2,
     std::unique_ptr< te::gm::Geometry >& result );

  /*!
    \brief Performs difference between two geometries.
    \param geom1 Input geometry 1.
    \param geom1 Input geometry 2.
    \param result Result geometry (or a null pointer if there is none).
    \note Multi-geometries accepted.
    \note Geometries with distinct SRID accepted.
    \note geom1 SRID is taken as reference (it will not be changed).
  */
  GDMASEMANTICANALYSISEXPORT void geometryDifference( const te::gm::Geometry& geom1,
     const te::gm::Geometry& geom2,
     std::unique_ptr< te::gm::Geometry >& result );

  /*!
    \brief Performs union between two geometries.
    \param geom1 Input geometry 1.
    \param geom1 Input geometry 2.
    \param result Result geometry (or a null pointer if there is none).
    \note Multi-geometries accepted.
    \note Geometries with distinct SRID accepted.
    \note geom1 SRID is taken as reference (it will not be changed).
  */
  GDMASEMANTICANALYSISEXPORT void geometryUnion( const te::gm::Geometry& geom1,
     const te::gm::Geometry& geom2,
     std::unique_ptr< te::gm::Geometry >& result );

  /*!
    \brief Returns true if both geometries are disjoint.
    \param geom1 Input geometry 1.
    \param geom1 Input geometry 2.
    \return Returns true if both geometries are disjoint.
    \note Multi-geometries accepted.
    \note Geometries with distinct SRID accepted.
    \note geom1 SRID is taken as reference (it will not be changed).
  */
  GDMASEMANTICANALYSISEXPORT bool areDisjoint( const te::gm::Geometry& geom1,
     const te::gm::Geometry& geom2 );

  /*!
    \brief Performs intersection between two geometries sets.
    \param inputGeomsVec1 Input geometries vector 1.
    \param inputGeomsVec2 Input geometries vector 2.
    \param outputGeomsVec Intersection output result (or an empty vector if there is no interesection).
    \param outputGeomsOriginIdxPtr A pointer to a vector where to store indexes from inputGeomsVec1 elements from where each output geometry was created, or a null pointer.
    \note Multi-geometries accepted.
    \note Geometries with distinct SRID accepted.
    \note inputGeomsVec1 SRID is taken as reference (it will not be changed).
    \note Elements within the same group are supposed to not have spatial overlap.
    \note The output geometries are sub-geomtries of each element from inputGeomsVec1.
  */
  GDMASEMANTICANALYSISEXPORT void geometrySetsIntersection(
     const boost::ptr_vector< te::gm::Geometry >& inputGeomsVec1,
     const boost::ptr_vector< te::gm::Geometry >& inputGeomsVec2,
     boost::ptr_vector< te::gm::Geometry >& outputGeomsVec,
     std::vector< std::size_t > * const outputGeomsOriginIdxPtr );

  /*!
    \brief Performs union of two geometries sets.
    \param inputGeomsVec1 Input geometries vector 1.
    \param inputGeomsVec2 Input geometries vector 2.
    \param outputGeomsVec Intersection output result (or an empty vector if there is no interesection).
    \param outputGeomsOriginIdxPtr A pointer to a vector where to store indexes from inputGeomsVec1 elements from where each output geometry was created, or a null pointer.
    \note Multi-geometries accepted.
    \note Geometries with distinct SRID accepted.
    \note inputGeomsVec1 SRID is taken as reference (it will not be changed).
    \note Elements within the same group are supposed to not have spatial overlap.
    \note The output geometries are sub-geomtries of each element from inputGeomsVec1.
  */
  GDMASEMANTICANALYSISEXPORT void geometrySetsUnion(
     const boost::ptr_vector< te::gm::Geometry >& inputGeomsVec1,
     const boost::ptr_vector< te::gm::Geometry >& inputGeomsVec2,
     boost::ptr_vector< te::gm::Geometry >& outputGeomsVec );

    /*!
      \brief Convert multi-geometries to single geometries (if applied).
      \param inputGeomsVec Input geometries vector (single geometries also accepted).
      \param outputGeomsVec  Output geometry pointers vectors( pointer to geometries from inputGeomsVec ).
      \note The returned pointers are owned by inputGeomsVec.
    */
  GDMASEMANTICANALYSISEXPORT void multi2Single(
    const boost::ptr_vector< te::gm::Geometry >& inputGeomsVec,
    std::vector< te::gm::Geometry* >& outputGeomsVec );

    /*!
      \brief Convert multi-geometries to single geometries (if applied).
      \param inputGeomsVec Input geometries vector (single geometries also accepted).
      \param outputGeomsVec  Output geometry vector.
    */
  GDMASEMANTICANALYSISEXPORT void multi2Single(
    const boost::ptr_vector< te::gm::Geometry >& inputGeomsVec,
    boost::ptr_vector< te::gm::Geometry >& outputGeomsVec );

    /*!
      \brief Convert multi-geometries to single geometries (if applied).
      \param targetSRID Target SRID.
      \param geomsVec Geometries vector.
      \note The returned pointers are owned by inputGeomsVec.
    */
  GDMASEMANTICANALYSISEXPORT void transformGeometries( const int targetSRID,
    boost::ptr_vector< te::gm::Geometry >& geomsVec );

    /*!
      \brief Get the envelope from a geometries set.
      \param inputGeomsVec Input geometries vector.
      \param envelope The calculated envelope.
      \param envelopeSRID The envelope SRID.
      \return true if ok, false on errors
    */
  GDMASEMANTICANALYSISEXPORT bool getGeomSetEnvelope(
    const boost::ptr_vector< te::gm::Geometry >& inputGeomsVec,
    te::gm::Envelope& envelope, int& envelopeSRID );

}  // end namespace sa
}  // end namespace gdma

#endif  // GEODMA_LIB_SEMANTICANALYSIS_GEOMETRYFUNCTIONS_HPP_
