/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/semanticanalysis/graph.hpp
*/

#ifndef GEODMA_LIB_SEMANTICANALYSIS_GRAPH_HPP_
#define GEODMA_LIB_SEMANTICANALYSIS_GRAPH_HPP_

#include <string>
#include <map>
#include <memory>
#include <vector>
#include <utility>
#include <mutex>

// GeoDMA Includes
#include "config.hpp"
#include "graphVertice.hpp"

namespace gdma {
namespace sa {

// Analysis forward
class Analysis;

/*!
  \class Graph

  \brief A basic concept of a Graph.
*/
class GDMASEMANTICANALYSISEXPORT Graph {
 public:
    Graph();

    Graph( const Graph& other );

    ~Graph();

    Graph& operator=( const Graph& other );

    /*!
      \brief Reset the graph to an initial state.
      \note This is a thread safe method.
     */
    void reset();

    /*!
      \brief Returns the Graph name.
      \return Returns the Graph name.
      \note This is a thread safe method.
     */
    const std::string& getName() const;

    /*!
      \brief Set the Graph name.
      \param newName The Graph name.
      \note This is a thread safe method.
     */
    void setName( const std::string& newName );

    /*!
      \brief Add a new vertice into the graph.
      \param verticePtr A pointer to the vertice to be added (the graph will take the ownership)
      \param inputConnections The given vertice input connections to other vertices already added to this graph (std::pair< other_vertice_ID, connection_alias >).
      \return true if ok, false on errors.
      \note This is a thread safe method.
     */
    bool addVertice( GraphVertice* verticePtr,
      const std::vector< std::pair< GraphVerticeIDT, std::string > >& inputConnections );

    /*!
      \brief Return the required vertice of null if it was not found.
      \param verticeID Vertice ID.
      \return Return the required vertice of null if it was not found.
      \note This is a thread safe method.
     */
    GraphVertice* getVertice( const GraphVerticeIDT& verticeID );

    /*!
      \brief Return the required vertice of null if it was not found.
      \param verticeName Vertice name.
      \return Return the required vertice of null if it was not found.
      \note This is a thread safe method.
     */
    GraphVertice* getVertice( const std::string& verticeName );

    /*!
      \brief Return all vertices names.
      \param verticesNames Return all vertices names.
      \note This is a thread safe method.
     */
    void getVerticesNames( std::vector< std::string >& verticesNames ) const;

    /*!
      \brief Return all vertices IDs.
      \param verticesNames Return all vertices Ids.
      \note This is a thread safe method.
     */
    void getVerticesIDs( std::vector< GraphVerticeIDT >& verticesIDs ) const;

    /*!
      \brief Returns the number of current vertices number.
      \return Returns the number of current vertices number.
      \note This is a thread safe method.
     */
    std::size_t getVerticesNumber() const;

    /*!
      \brief Set a output directory to this graph and its vertices.
      \param outDir Output directory to this graph.
      \return true if OK, false on errors.
      \note This is a thread safe method.
      \note The provided path must be canonical and system complete.
      */
    bool setOutputDirectoryName( const std::string& outDirName );

    /*!
      \brief Returns the output directory to this graph.
      \return Returns the output directory to this graph.
      \note This is a thread safe method.
      */
    const std::string& getOutputDirectoryName() const;

    /*!
      \brief Enable (true) or disable (false) the use of result cache.
      \param enabled Enable (true) or disable (false) the use of result cache.
      \note This is a thread safe method.
      */
    void enableCache( const bool enabled );

    /*!
      \brief Returns true if cache is enabled, false otherwise.
      \return Returns true if cache is enabled, false otherwise.
      \note This is a thread safe method.
      */
    bool isCacheEnabled() const;

    /*!
      \brief Set this vertice's related Analysis.
      \param graph This vertice's related Analysis.
      \note This is a thread safe method.
      */
    void setAnalysis( const Analysis& analysis );

    /*!
      \brief Returns a pointer to the related Analysis or null if none is related.
      \return Returns a pointer to the related Analysis or null if none is related.
      \note This is a thread safe method.
      */
    Analysis const* getAnalysis() const;

 private:
    /*! \brief Vertices container type definition. */
    typedef std::map< GraphVerticeIDT, std::unique_ptr< GraphVertice > >
      VertContT;

    /*! \brief Tells if result cache must be used. */
    bool m_enableCache;

    /*! \brief Next valid ID to use when adding a new vertice into this graph. */
    GraphVerticeIDT m_nextValidAddedVerticeIDCounter;

    /*! \brief A pointer to this vertice's related Analysis. */
    Analysis const* m_analysisPtr;

    /*! \brief Graph name. */
    std::string m_graphName;

    /*! \brief graph output directory. */
    std::string m_outputDirectoryName;

    /*! \brief Graph vertices (key:vertice name). */
    VertContT m_verticesPtrs;

    /*! \brief Internal vertice mutex. */
    mutable std::mutex m_mutex;
};

}  // end namespace sa
}  // end namespace gdma

#endif  // GEODMA_LIB_SEMANTICANALYSIS_GRAPH_HPP_
