/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/semanticanalysis/graphVertice.hpp
*/

#ifndef GEODMA_LIB_SEMANTICANALYSIS_GRAPHVERTICE_HPP_
#define GEODMA_LIB_SEMANTICANALYSIS_GRAPHVERTICE_HPP_

#include <string>
#include <vector>
#include <utility>
#include <map>
#include <mutex>
#include <set>
#include <cstdint>

// GeoDMA Includes
#include "config.hpp"
#include "defines.hpp"
#include "typedefs.hpp"
#include "executionResult.hpp"
#include "verticeTask.hpp"
#include "../taskprocessing/task.hpp"

namespace gdma {
namespace sa {

// Graph forward

class Graph;

/*!
  \class GraphVertice

  \brief A basic concept of a Graph Vertice.
*/
class GDMASEMANTICANALYSISEXPORT GraphVertice {
 public:
    /*!
      \enum The vertice output type.
    */
    enum OutputType {
      INVALID_OUT_TYPE,  //!< Invalid output type.
      RASTER_URI_OUT_TYPE,  //!< Raster output type.
      VECTOR_URI_OUT_TYPE  //!< Vector output type.
    };

    virtual ~GraphVertice();

    /*!
      \brief Set the vertice related name.
      \param newName New vertice related name.
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    void setName( const std::string& newName );

    /*!
      \brief Returns the vertice related name.
      \return Returns the vertice related name.
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    std::string getName() const;

    /*!
      \brief Returns the vertice type.
      \return Returns the vertice type.
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    virtual std::string getType() const = 0;

    /*!
      \brief Returns a vertice user-friendly description.
      \param descriptionStrReturns Vertice user-friendly description.
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    virtual void getDescription( std::string& descriptionStr ) const = 0;

    /*!
      \brief Set the vertice ID.
      \param newID New vertice ID.
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    void setID( const GraphVerticeIDT& newID );

    /*!
      \brief Returns the vertice ID.
      \return Returns the vertice ID.
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    GraphVerticeIDT getID() const;

    /*!
      \brief Add a connection from other vertice output to this one, and gives it a name alias.
      \param other The other vertice (it must always be valid).
      \param alias Connection name alias.
      \return true if ok, false on errors.
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    bool connectInputVertice( GraphVertice& other, const std::string& alias );

    /*!
      \brief Returns all current input connetions.
      \param connections A vector of pairs [ connection alias, connected vertice pointer ].
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    void getInputConnections(
      std::vector< std::pair< std::string, GraphVertice* > >& connections ) const;

    /*!
      \brief Returns all current output connetions.
      \param connections A vector of pointers to the connected vertices.
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    void getOutputConnections(
      std::vector< GraphVertice* >& connections ) const;

    /*!
      \brief Returns all first pass task dependencies.
      \param firstPassTaskDeps Pointers to all tasks that are expected to be finished before this vertice's first pass task execution.
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    virtual void get1stPassTaskDeps( std::vector< VerticeTask* >& firstPassTaskDeps ) const;

    /*!
      \brief Returns all second pass task dependencies.
      \param firstPassTaskDeps Pointers to all tasks that are expected to be finished before this vertice's second pass task execution.
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    virtual void get2stPassTaskDeps( std::vector< VerticeTask* >& firstPassTaskDeps ) const;

    /*!
      \brief Returns true if all input connected vertices have output (output file names not empty) or if there is no input connections.
      \return Returns true if all input connected vertices have output (output file names not empty) or if there is no input connections.
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    bool isInputSatisfied() const;

    /*!
      \brief Disconnects this vertice from all others.
      \note This is a thread safe method (using this vertice's internal mutex).
     */
    void disconnect();

    /*!
      \brief Set a vertice parameter.
      \param parameterName Parameter name.
      \param parameterValue Parameter value.
      \return true if ok, false on errors.
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    virtual bool setParameter( const std::string& parameterName,
      const std::string& parameterValue ) = 0;

    /*!
      \brief Get all current vertice parameters.
      \param parameters A vector of tuples [ parameter name, parameter value ].
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    virtual void getParameters( std::vector< std::pair< std::string, std::string > >& parameters ) const = 0;

    /*!
      \brief Set a vertice input resouce (from the current context)
      \param contextURIAlias The context URI alias.
      \param verticeURIAlias This vertice's URI alias.
      \return true if ok, false on errors.
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    virtual bool setResource( const std::string& contextURIAlias,
      const std::string& verticeURIAlias ) = 0;

    /*!
      \brief Returns a description for each input parameter accepted by this vertice.
      \param parasDesc A description for each input parameter accepted by this vertice (index:parameter name).
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    virtual void getInParsDescription( std::map< std::string, std::string >& inParsDesc ) const = 0;

    /*!
      \brief Returns a description for each input resource accepted by this vertice.
      \param parasDesc A description for each input resource accepted by this vertice (index:resource alias).
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    virtual void getInResDescription( std::map< std::string, std::string >& inResDesc ) const = 0;

    /*!
      \brief Returns this vertice 1st pass output file name (without spatial resolution ).
      \return Returns this vertice 1st pass output file name (without spatial resolution ).
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    virtual const std::string get1stPassOutputFileName() const;

    /*!
      \brief Returns this vertice 2nd pass output file name (with spatial resolution ).
      \return Returns this vertice 2nd pass output file name (with spatial resolution ).
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    virtual const std::string get2ndPassOutputFileName() const;

    /*!
      \brief Set a output directory to this vertice.
      \param outDir Output directory to this vertice.
      \return true if OK, false on errors.
      \note This is a thread safe method (using this vertice's internal mutex).
      \note The provided path must be canonical and system complete.
      */
    bool setOutputDirectoryName( const std::string& outDirName );

    /*!
      \brief Returns the output directory to this vertice.
      \return Returns the output directory to this vertice.
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    const std::string getOutputDirectoryName() const;

    /*!
      \brief Returns the vertice output type.
      \return Returns the vertice output type.
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    virtual OutputType getOutputType() const = 0;

    /*!
      \brief Returns a clone of this vertice (the caller must take the ownership of the returned object).
      \return Returns a clone of this vertice (the caller must take the ownership of the returned object).
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    virtual GraphVertice* clone() const = 0;

    /*!
      \brief Enable (true) or disable (false) the use of result cache.
      \param enabled Enable (true) or disable (false) the use of result cache.
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    void enableCache( const bool enabled );

    /*!
      \brief Returns true if cache is enabled, false otherwise.
      \return Returns true if cache is enabled, false otherwise.
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    bool isCacheEnabled() const;

    /*!
      \brief Returns true if a vertice is connectable to this one.
      \param other The other vertice (it must always be valid).
      \param alias Connection name alias.
      \return Returns true if a vertice is connectable to this one.
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    virtual bool isConnectable( const GraphVertice& other, const std::string& alias ) const = 0;

    /*!
      \brief Set this vertice's related graph.
      \param graph This vertice's related graph.
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    void setGraph( const Graph& graph );

    /*!
      \brief Returns a pointer to the related graph or null if none is related.
      \return Returns a pointer to the related graph or null if none is related.
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    Graph const* getGraph() const;

    /*!
      \brief Returns a reference to the 1st pass (without spatial resolution) vertice task.
      \brief Returns a reference to the 1st pass (without spatial resolution) vertice task.
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    VerticeTask& get1stPassTask();

    /*!
      \brief Returns a reference to the 2nd pass (with spatial resolution) vertice task.
      \brief Returns a reference to the 2nd pass (with spatial resolution) vertice task.
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    VerticeTask& get2ndPassTask();

    /*!
      \brief Enable/disable the second pass vertices tasks execution.
      \param enable If true, the second pass vertices tasks will be executed.
      \return true if ok, false on errors.
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    bool enable2ndPassTasks( const bool enable );

    /*!
      \brief Returns true if the second pass vertices tasks execution.
      \return Returns true if the second pass vertices tasks execution.
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    bool is2ndPassTasksEnabled() const;

    /*!
      \brief Returns true if this vertice is capable of executing a second pass task.
      \return Returns true if this vertice is capable of executing a second pass task.
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    virtual bool is2ndPassCapable() const = 0;

    /*!
      \brief Returns true if this vertice can produce an output with multiple class labels.
      \return Returns true if this vertice can produce an output with multiple class labels.
      \note This is a thread safe method (using this vertice's internal mutex).
      */
    virtual bool isMultiClass() const = 0;

 protected:
    GraphVertice();

    GraphVertice( const GraphVertice& other );

    GraphVertice& operator=( const GraphVertice& other );

    /*!
      \brief Reset this vertice to its initial state.
      \note THIS IS NOT A THREAD SAFE METHOD
    */
    void reset();

    /*!
      \brief Returns a reference to this vertice's internal mutex.
      \return Returns a reference to this vertice's internal mutex.
      */
    std::mutex& getMutex() const;

    /*!
      \brief Method called when the first pass task is executed.
      \param errorMessage Return error messega, if there is one.
      \return Returns true if ok, false on errors.
      \note THIS IS NOT A THREAD SAFE METHOD
      */
    virtual bool executeFirstPass( std::string& errorMessage ) = 0;

    /*!
      \brief Method called when the second pass task is executed.
      \param errorMessage Return error messega, if there is one.
      \return Returns true if ok, false on errors.
      \note THIS IS NOT A THREAD SAFE METHOD
      */
    virtual bool executeSecondPass( std::string& errorMessage );

 private:
    class Generic1stPassVerticeTask : public VerticeTask {
     public :
        /*! \brief A reference to this taks related vertice. */
        GraphVertice& m_vertice;

        /*!
         \brief A reference to this task related vertice.
         \param verticeRef reference to this task related vertice.
         */
        explicit Generic1stPassVerticeTask( GraphVertice& verticeRef );

        ~Generic1stPassVerticeTask();

        //  overload
        void execute();
    };

    class Generic2ndPassVerticeTask : public VerticeTask {
     public :
        /*! \brief A reference to this taks related vertice. */
        GraphVertice& m_vertice;

        /*!
         \brief A reference to this task related vertice.
         \param verticeRef reference to this task related vertice.
         */
        explicit Generic2ndPassVerticeTask( GraphVertice& verticeRef );

        ~Generic2ndPassVerticeTask();

        //  overload
        void execute();
    };

    /*! \brief Tells if result cache must be used. */
    bool m_enableCache;

    /*! \brief Enable/disable the second pass vertices tasks execution. */
    bool m_enable2ndPassVerticesTasksExecution;

    /*! \brief A pointer to this vertice's related graph. */
    Graph const* m_graphPtr;

    /*! \brief vertice ID. */
    GraphVerticeIDT m_ID;

    /*! \brief vertice name. */
    std::string m_name;

    /*! \brief vertice output directory. */
    std::string m_outputDirectoryName;

    /*! \brief Input connections (key: connection alias). */
    std::map< std::string, GraphVertice* > m_inputConnections;

    /*! \brief Output connections. */
    std::set< GraphVertice* > m_outputConnections;

    /*! \brief Vertice parameters. */
    std::map< std::string, std::string > m_parameters;

    /*! \brief Internal vertice mutex. */
    mutable std::mutex m_mutex;

    /*! \brief A generic 2nd pass internal vertice task. */
    Generic1stPassVerticeTask m_generic1stPassTask;

    /*! \brief A generic 2nd pass internal vertice task. */
    Generic2ndPassVerticeTask m_generic2ndPassTask;
};

}  // end namespace sa
}  // end namespace gdma

#endif  // GEODMA_LIB_SEMANTICANALYSIS_GRAPHVERTICE_HPP_
