/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

#include <terralib/dataaccess.h>
#include <terralib/memory.h>
#include <terralib/rp.h>
#include <terralib/common.h>

#include <boost/filesystem.hpp>

#include <limits>
#include <map>
#include <string>
#include <memory>
#include <utility>
#include <vector>
#include <tuple>

// GeoDMA Includes
#include "ioFunctions.hpp"
#include "generalFunctions.hpp"
#include "rasterFunctions.hpp"
#include "defines.hpp"
#include "../common/utils.hpp"

namespace gdma {
namespace sa {

  bool loadFileTokens( const std::string& fileName,
    std::vector< std::tuple< std::string, unsigned int, std::vector< std::string > > >& fileTokens )
  {
    std::fstream file;

    file.open( fileName.c_str(), std::fstream::in );

    if( !file.good() )
    {
      return false;
    }

    std::string lineStr;
    std::tuple< std::string, unsigned int, std::vector< std::string > > auxTuple;
    unsigned int lineNumber = 0;

    while( ! file.eof() )
    {
      lineStr.clear();
      std::getline( file, lineStr );

      if( ! lineStr.empty() )
      {
        // Finding tokens

        std::get<2>( auxTuple ).clear();
        findTokens( lineStr, std::get<2>( auxTuple ) );

        if( ! std::get<2>( auxTuple ).empty() )
        {
          if(
              ( std::get<2>( auxTuple ).size() == 2 )
              &&
              ( std::get<2>( auxTuple )[ 0 ] == "INCLUDE_FILE" )
            )
          {
            boost::filesystem::path auxPath;

            try
            {
              auxPath =
                boost::filesystem::system_complete(
                  boost::filesystem::path( fileName ).parent_path()
                  /
                  boost::filesystem::path( std::get<2>( auxTuple )[ 1 ] )
                );
            }
            catch(...)
            {
              return false;
            }

            if( ! loadFileTokens( auxPath.string(), fileTokens )  )
            {
              return false;
            }
          }
          else
          {
            std::get<0>( auxTuple ) = fileName;
            std::get<1>( auxTuple ) = lineNumber;
            fileTokens.push_back( auxTuple );
          }
        }
      }

      ++lineNumber;
    }

    return true;
  }

  bool loadFileTokens( const std::string& fileName,
    std::map< unsigned int, std::vector< std::string > > & fileTokens )
  {
    fileTokens.clear();

    std::fstream file;

    file.open( fileName.c_str(), std::fstream::in );

    if( !file.good() )
    {
      return false;
    }

    std::string lineStr;
    std::vector< std::string > lineTokens;
    unsigned int lineNumber = 0;

    while( ! file.eof() )
    {
      lineStr.clear();
      std::getline( file, lineStr );

      if( ! lineStr.empty() )
      {
        // Finding tokens

        lineTokens.clear();
        findTokens( lineStr, lineTokens );

        if( ! lineTokens.empty() )
        {
          fileTokens[ lineNumber ] = lineTokens;
        }
      }

      ++lineNumber;
    }

    return true;
  }

  bool loadFileTokens( const std::string& fileName,
    std::vector< std::pair< unsigned int, std::vector< std::string > > >& fileTokens )
  {
    fileTokens.clear();

    std::fstream file;

    file.open( fileName.c_str(), std::fstream::in );

    if( !file.good() )
    {
      return false;
    }

    std::string lineStr;
    unsigned int lineNumber = 0;
    std::pair< unsigned int, std::vector< std::string > > auxPair;

    while( ! file.eof() )
    {
      lineStr.clear();
      std::getline( file, lineStr );

      if( ! lineStr.empty() )
      {
        // Finding tokens

        auxPair.second.clear();
        findTokens( lineStr, auxPair.second );

        if( ! auxPair.second.empty() )
        {
          auxPair.first = lineNumber;
          fileTokens.push_back( auxPair );
        }
      }

      ++lineNumber;
    }

    return true;
  }

  bool openDiskRaster( const std::string& rasterFileName,
    std::unique_ptr< te::rst::Raster >& rasterPtr )
  {
    std::map<std::string, std::string> inputRasterInfo;
    inputRasterInfo["URI"] = rasterFileName;

    try
    {
      rasterPtr.reset( te::rst::RasterFactory::open(  inputRasterInfo ) );
    }
    catch(...)
    {
      rasterPtr.reset();
    }

    if( rasterPtr.get() )
    {
      return true;
    }
    else
    {
      return false;
    }
  }

  bool openRaster( const te::core::URI& inputRasterURI,
    std::unique_ptr< te::rst::Raster >& rasterPtr )
  {
    rasterPtr.reset();

    if( ! inputRasterURI.isValid() )
    {
      return false;
    }

    std::string resourceStr = inputRasterURI.uri();

    if( te::common::Convert2UCase( resourceStr.substr( 0, 7 ) ) == "FILE://" )
    {
      resourceStr = resourceStr.substr( 7, resourceStr.size() - 7 );

      return openDiskRaster( resourceStr, rasterPtr );
    }
    else
    {
      std::map<std::string, std::string> inputRasterInfo;
      inputRasterInfo["URI"] = resourceStr;

      try
      {
        rasterPtr.reset( te::rst::RasterFactory::open(  inputRasterInfo ) );
      }
      catch(...)
      {
        rasterPtr.reset();
      }
    }

    if( rasterPtr.get() == 0 )
    {
      return false;
    }
    else
    {
      return true;
    }
  }

  bool diskRasterFileCopy( const std::string& inputFileName,
    const std::string& outputFileName )
  {
    std::map<std::string, std::string> inputRasterInfo;
    inputRasterInfo["URI"] = inputFileName;

    std::unique_ptr< te::rst::Raster > inputRasterPointer;
    if( ! openDiskRaster( inputFileName, inputRasterPointer ) )
    {
      return false;
    }
    else
    {
      return te::rp::Copy2DiskRaster( *inputRasterPointer, outputFileName );
    }
  }

  bool createEmptyResultShapeFile( const std::string& shapeFileName )
  {
    boost::ptr_vector< te::gm::Geometry > inputGeometries;
    return save2ResultShapeFile( inputGeometries, shapeFileName, 0, 0, 0, 0 );
  }

  bool save2ResultShapeFile(
    const boost::ptr_vector< te::gm::Geometry >& inputGeometries,
    const std::string& shapeFileName,
    std::vector< std::string > const * const geomLabelsPtr,
    std::vector< GraphVerticeIDT > const * const geomsIDsPtr,
    std::vector< std::string > * const extraPropertiesNamesPtr,
    std::vector< std::vector< double > > * const extraPropertiesValuesPtr )
  {
    if( ( geomLabelsPtr != 0 ) && ( ! geomLabelsPtr->empty() ) )
    {
      if( inputGeometries.size() != geomLabelsPtr->size() ) return false;
    }

    if( ( geomsIDsPtr != 0 ) && ( ! geomsIDsPtr->empty() ) )
    {
      if( inputGeometries.size() != geomsIDsPtr->size() ) return false;
    }

    if(
        ( extraPropertiesValuesPtr != 0 )
        &&
        ( ! extraPropertiesValuesPtr->empty() )
        &&
        (
          ( extraPropertiesValuesPtr->size() != inputGeometries.size() )
          ||
          ( extraPropertiesNamesPtr == 0 )
          ||
          ( extraPropertiesNamesPtr->empty() )
        )
      )
    {
      return false;
    }

    // shape file canonica path

    boost::filesystem::path shapeFileCompletePath;
    try
    {
      shapeFileCompletePath = boost::filesystem::system_complete( shapeFileName );
    }
    catch(...)
    {
      return false;
    }

    // defining labels and IDs

    std::vector< std::string > internalLabels;
    std::vector< GraphVerticeIDT > internalIDs;
    std::vector< std::string > const * currentLabelsPtr = 0;
    std::vector< GraphVerticeIDT > const * currentIDsPtr = 0;

    {
      const std::size_t inputGeometriesSize = inputGeometries.size();

      if( ( geomsIDsPtr == 0 ) || ( geomsIDsPtr->empty() ) )
      {
        internalIDs.resize( inputGeometriesSize );
        currentIDsPtr = &internalIDs;

        if( ( geomLabelsPtr == 0 ) || ( geomLabelsPtr->empty() ) )
        {
          for( std::size_t inputGeometriesIdx = 0 ; inputGeometriesIdx <
            inputGeometriesSize ; ++inputGeometriesIdx )
          {
            internalIDs[ inputGeometriesIdx ] = (GraphVerticeIDT)(
              inputGeometriesIdx + 1 );
          }
        }
        else
        {
          std::map< std::string, GraphVerticeIDT > labels2IdsMap;
          GraphVerticeIDT nextValidID = 1;
          std::map< std::string, GraphVerticeIDT >::const_iterator labels2IdsMapIt;

          for( std::size_t inputGeometriesIdx = 0 ; inputGeometriesIdx <
            inputGeometriesSize ; ++inputGeometriesIdx )
          {
            labels2IdsMapIt = labels2IdsMap.find( geomLabelsPtr->operator[](
              inputGeometriesIdx ) );

            if( labels2IdsMapIt == labels2IdsMap.end() )
            {
              labels2IdsMap[ geomLabelsPtr->operator[]( inputGeometriesIdx ) ] = nextValidID;
              internalIDs[ inputGeometriesIdx ] = nextValidID;
              ++nextValidID;
            }
            else
            {
              internalIDs[ inputGeometriesIdx ] = labels2IdsMapIt->second;
            }
          }
        }
      }
      else
      {
        currentIDsPtr = geomsIDsPtr;
      }

      if( ( geomLabelsPtr == 0 ) || ( geomLabelsPtr->empty() ) )
      {
        currentLabelsPtr = &internalLabels;
        internalLabels.resize( inputGeometriesSize );

        for( std::size_t inputGeometriesIdx = 0 ; inputGeometriesIdx <
          inputGeometriesSize ; ++inputGeometriesIdx )
        {
          internalLabels[ inputGeometriesIdx ] = boost::lexical_cast< std::string>(
            currentIDsPtr->operator[]( inputGeometriesIdx ) );
        }
      }
      else
      {
        currentLabelsPtr = geomLabelsPtr;
      }
    }

    // Create a memory dataset

    const std::string dataSetName = boost::filesystem::path( shapeFileName ).stem().string();

    int dataSetSRID = 0;
    if( ! inputGeometries.empty() )
    {
      dataSetSRID = inputGeometries[ 0 ].getSRID();
    }

    std::auto_ptr<te::da::DataSetType> dataSetType1Ptr(new te::da::DataSetType(
      dataSetName ) );
    dataSetType1Ptr->add( new te::dt::SimpleProperty( GDMA_SA_DEFAULT_CLASSID_PROP_NAME,
      GDMA_SA_DEF_TL_RASTER_DATA_TYPE, true) );
    dataSetType1Ptr->add( new te::dt::SimpleProperty(GDMA_SA_DEFAULT_CLASSNAME_PROP_NAME,
      te::dt::STRING_TYPE, true) );
    dataSetType1Ptr->add( new te::gm::GeometryProperty(GDMA_SA_DEFAULT_GEOM_PROP_NAME,
      dataSetSRID, te::gm::PolygonType, true, 0, 0 ) );

    if( extraPropertiesNamesPtr && ( ! extraPropertiesNamesPtr->empty() ) )
    {
      for( std::size_t extraPropertiesNamesIdx = 0 ; extraPropertiesNamesIdx <
        extraPropertiesNamesPtr->size() ; ++extraPropertiesNamesIdx )
      {
        dataSetType1Ptr->add( new te::dt::SimpleProperty(
          extraPropertiesNamesPtr->operator[]( extraPropertiesNamesIdx ),
          te::dt::DOUBLE_TYPE, true) );
      }
    }

    std::auto_ptr<te::da::DataSetType> dataSetType2Ptr(new te::da::DataSetType( *dataSetType1Ptr ) );
    std::auto_ptr<te::mem::DataSet> memDataSetPtr( new te::mem::DataSet(dataSetType1Ptr.get() ) );
    const std::size_t inputGeometriesSize = inputGeometries.size();
    std::vector< te::gm::Geometry* > singleGeoms;
    std::size_t singleGeomsSize = 0;
    std::size_t singleGeomsIdx = 0;
    std::auto_ptr< te::mem::DataSetItem > dsItemPtr;
    te::gm::Geometry* singleGeomPtr = 0;
    std::unique_ptr< te::gm::Polygon > auxPolygonPtr;
    std::unique_ptr< te::gm::LinearRing > auxLinearRingPtr;
    te::gm::LineString* lineStringNPtr = 0;
    std::size_t lineStringSize = 0;
    std::size_t lineStringIdx = 0;
    const std::size_t extraPropertiesNamesSize = extraPropertiesNamesPtr ?
      extraPropertiesNamesPtr->size() : 0;
    std::size_t extraPropertiesIdx = 0;;

    for( std::size_t inputGeometriesIdx = 0 ; inputGeometriesIdx < inputGeometriesSize ;
      ++inputGeometriesIdx )
    {
      if( extraPropertiesNamesSize )
      {
        if( extraPropertiesValuesPtr->operator[]( inputGeometriesIdx ).size() !=
          extraPropertiesNamesSize )
        {
          return false;
        }
      }

      singleGeoms.clear();
      te::gm::Multi2Single( (te::gm::Geometry*)&( inputGeometries[ inputGeometriesIdx ]), singleGeoms );
      singleGeomsSize = singleGeoms.size();

      for( singleGeomsIdx = 0 ; singleGeomsIdx < singleGeomsSize ; ++singleGeomsIdx )
      {
        singleGeomPtr = singleGeoms[ singleGeomsIdx ];

        switch( singleGeomPtr->getGeomTypeId() )
        {
          case te::gm::PolygonType :
          {
            dsItemPtr.reset( new te::mem::DataSetItem( memDataSetPtr.get() ) );
            dsItemPtr->setInt32( 0, currentIDsPtr->operator[]( inputGeometriesIdx ) );
            dsItemPtr->setString( 1, currentLabelsPtr->operator[]( inputGeometriesIdx ) );

            if( singleGeomPtr->getSRID() == dataSetSRID )
            {
              dsItemPtr->setGeometry( 2, (te::gm::Geometry*)singleGeomPtr->clone() );
            }
            else
            {
              singleGeomPtr = (te::gm::Geometry*)singleGeomPtr->clone();
              singleGeomPtr->transform( dataSetSRID );
              dsItemPtr->setGeometry( 2, singleGeomPtr );
            }

            for( extraPropertiesIdx = 0 ; extraPropertiesIdx < extraPropertiesNamesSize ;
              ++extraPropertiesIdx )
            {
              dsItemPtr->setDouble( 3 + extraPropertiesIdx,
                extraPropertiesValuesPtr->operator[]( inputGeometriesIdx )[
                extraPropertiesIdx ] );
            }

            memDataSetPtr->add( dsItemPtr.release() );

            break;
          }
          case te::gm::LineStringType :
          {
            lineStringNPtr = (te::gm::LineString*)singleGeomPtr;

            if( lineStringNPtr->isClosed() )
            {
              dsItemPtr.reset( new te::mem::DataSetItem( memDataSetPtr.get() ) );
              dsItemPtr->setInt32( 0, currentIDsPtr->operator[]( inputGeometriesIdx ) );
              dsItemPtr->setString( 1, currentLabelsPtr->operator[]( inputGeometriesIdx ) );

              lineStringSize = lineStringNPtr->size();
              auxLinearRingPtr.reset( new te::gm::LinearRing( lineStringSize,
                te::gm::LineStringType, lineStringNPtr->getSRID(),
                (te::gm::Envelope*)lineStringNPtr->getEnvelope() ) );

              for( lineStringIdx = 0 ; lineStringIdx < lineStringSize ;
                ++lineStringIdx )
              {
                auxLinearRingPtr->setPoint( lineStringIdx, lineStringNPtr->getX( lineStringIdx ),
                  lineStringNPtr->getY( lineStringIdx ) );
              }

              auxPolygonPtr.reset( new te::gm::Polygon( 0, te::gm::PolygonType, auxLinearRingPtr->getSRID(), 0 ) );
              auxPolygonPtr->add( auxLinearRingPtr.release() );

              if( auxPolygonPtr->getSRID() != dataSetSRID )
              {
                auxPolygonPtr->transform( dataSetSRID );
              }

              dsItemPtr->setGeometry( 2, auxPolygonPtr.release() );

              for( extraPropertiesIdx = 0 ; extraPropertiesIdx < extraPropertiesNamesSize ;
                ++extraPropertiesIdx )
              {
                dsItemPtr->setDouble( 3 + extraPropertiesIdx,
                  extraPropertiesValuesPtr->operator[]( inputGeometriesIdx )[
                  extraPropertiesIdx ] );
              }

              memDataSetPtr->add( dsItemPtr.release() );
            }

            break;
          }
          default :
          {
            break;
          }
        }
      }
    }

    // creating the disk dataset



    std::unique_ptr<te::da::DataSource> dsOGR = te::da::DataSourceFactory::make("OGR",
      "file://" + shapeFileCompletePath.string() );
    if( dsOGR.get() == 0 )
    {
      return false;
    }
    dsOGR->open();

    memDataSetPtr->moveBeforeFirst();

    try
    {
      dsOGR->createDataSet(dataSetType2Ptr.get(), std::map<std::string, std::string>());
      dsOGR->add( dataSetName, memDataSetPtr.get(), std::map<std::string, std::string>());
      dsOGR->execute("CREATE SPATIAL INDEX ON " + dataSetName);
    }
    catch(...)
    {
      return false;
    }

    dsOGR->close();

    return true;
  }

  bool loadResultShapeFileGeoms( const std::string& pathShp,
      boost::ptr_vector< te::gm::Geometry >& gmVec )
  {
    gmVec.clear();

    return loadResultShapeFile( pathShp, 0, &gmVec, 0, 0, 0 );
  }

  bool loadResultShapeFileExtraPropNames(
    const std::string& shapeFileName,
    std::vector< std::string >& extraPropNames )
  {
    extraPropNames.clear();

    te::core::URI dataSourceUri( "file://" + shapeFileName );
    std::vector< std::string > allPropNames;

    if( ! loadDataSetPropNames( dataSourceUri, "", "OGR", allPropNames ) )
    {
      return false;
    }

    for( std::size_t allPropNamesIdx = 0 ; allPropNamesIdx < allPropNames.size() ; ++allPropNamesIdx )
    {
      if(
          ( allPropNames[ allPropNamesIdx ] != GDMA_SA_DEFAULT_CLASSNAME_PROP_NAME )
          &&
          ( allPropNames[ allPropNamesIdx ] != GDMA_SA_DEFAULT_CLASSID_PROP_NAME )
          &&
          ( allPropNames[ allPropNamesIdx ] != GDMA_SA_DEFAULT_GEOM_PROP_NAME )
        )
      {
        extraPropNames.push_back( allPropNames[ allPropNamesIdx ] );
      }
    }

    return true;
  }

  bool loadResultShapeFile(
    const std::string& shapeFileName,
    std::vector< std::string > const * const  extraFeaturesLabelsPtr,
    boost::ptr_vector< te::gm::Geometry >* const geometriesPtr,
    std::vector< std::string >* const classNamesPtr,
    std::vector< GraphVerticeIDT >* const classIDsPtr,
    std::vector< std::vector< double > >* const extraFeaturesPtr )
  {
    if( extraFeaturesPtr )
    {
      if( extraFeaturesLabelsPtr == 0 ) return false;
    }

    if( geometriesPtr ) geometriesPtr->clear();
    if( classNamesPtr ) classNamesPtr->clear();
    if( classIDsPtr ) classIDsPtr->clear();
    if( extraFeaturesPtr ) extraFeaturesPtr->clear();

    std::unique_ptr< te::da::DataSource > ogrDataSourcePtr( te::da::DataSourceFactory::make(
      "OGR", "file://" + shapeFileName ) );
    if( ogrDataSourcePtr.get() == 0 )
    {
      return false;
    }

    ogrDataSourcePtr->open();

    if( ! ogrDataSourcePtr->isOpened() )
    {
      return false;
    }

    const std::size_t extraFeaturesLabelsSize = extraFeaturesLabelsPtr ?
      extraFeaturesLabelsPtr->size() : 0;

    std::vector<std::string> dataSetNames = ogrDataSourcePtr->getDataSetNames();
    te::da::DataSourceTransactorPtr transactorPtr = ogrDataSourcePtr->getTransactor();
    std::size_t idPropIdx = std::numeric_limits< std::size_t >::max();
    std::size_t namePropIdx = std::numeric_limits< std::size_t >::max();
    std::size_t geomPropIdx = std::numeric_limits< std::size_t >::max();
    std::vector< std::size_t > extraFeaturesIndexes( extraFeaturesLabelsSize,
      std::numeric_limits< std::size_t >::max() );
    std::size_t extraFeaturesLabelsIdx = 0;
    std::vector< double > auxFeatures( extraFeaturesLabelsSize );

    for( unsigned int dataSetNamesIdx = 0 ; dataSetNamesIdx < dataSetNames.size() ;
      ++dataSetNamesIdx )
    {
      te::da::DataSetPtr datasetPtr = transactorPtr->getDataSet(dataSetNames[ dataSetNamesIdx ] );

      // Find properties indexes

      idPropIdx = std::numeric_limits< std::size_t >::max();
      namePropIdx = std::numeric_limits< std::size_t >::max();
      geomPropIdx = std::numeric_limits< std::size_t >::max();
      extraFeaturesLabelsIdx = 0;

      std::fill( extraFeaturesIndexes.begin(), extraFeaturesIndexes.end(),
        std::numeric_limits< std::size_t >::max() );

      {
        const std::size_t nProperties = datasetPtr->getNumProperties();

        for( std::size_t propIdx = 0 ; propIdx < nProperties ; ++propIdx )
        {
          if( datasetPtr->getPropertyName( propIdx ) == GDMA_SA_DEFAULT_CLASSNAME_PROP_NAME )
          {
            namePropIdx = propIdx;
          }

          if( datasetPtr->getPropertyName( propIdx ) == GDMA_SA_DEFAULT_CLASSID_PROP_NAME )
          {
            idPropIdx = propIdx;
          }

          if( datasetPtr->getPropertyName( propIdx ) == GDMA_SA_DEFAULT_GEOM_PROP_NAME )
          {
            geomPropIdx = propIdx;
          }

          for( extraFeaturesLabelsIdx = 0; extraFeaturesLabelsIdx < extraFeaturesLabelsSize ;
            ++extraFeaturesLabelsIdx )
          {
            if( datasetPtr->getPropertyName( propIdx ) ==
              extraFeaturesLabelsPtr->operator[]( extraFeaturesLabelsIdx ) )
            {
              extraFeaturesIndexes[ extraFeaturesLabelsIdx ] = propIdx;
            }
          }
        }

        if(
            ( idPropIdx == std::numeric_limits< std::size_t >::max() )
            ||
            ( namePropIdx == std::numeric_limits< std::size_t >::max() )
            ||
            ( geomPropIdx == std::numeric_limits< std::size_t >::max() )
          )
        {
          return false;
        }

        for( extraFeaturesLabelsIdx = 0; extraFeaturesLabelsIdx < extraFeaturesLabelsSize ;
          ++extraFeaturesLabelsIdx )
        {
          if( extraFeaturesIndexes[ extraFeaturesLabelsIdx ] ==
            std::numeric_limits< std::size_t >::max() )
          {
            return false;
          }
        }
      }

      // Load


      if( ! datasetPtr->moveBeforeFirst() )
      {
        return false;
      }

      while( datasetPtr->moveNext() )
      {
        if( geometriesPtr ) geometriesPtr->push_back(
          datasetPtr->getGeometry( geomPropIdx ).release() );

        if( classNamesPtr ) classNamesPtr->push_back(
          datasetPtr->getAsString( namePropIdx ) );

        if( classIDsPtr ) classIDsPtr->push_back(
          (GraphVerticeIDT)datasetPtr->getInt32( idPropIdx ) );

        if( extraFeaturesPtr )
        {
          for( extraFeaturesLabelsIdx = 0; extraFeaturesLabelsIdx < extraFeaturesLabelsSize ;
            ++extraFeaturesLabelsIdx )
          {
            auxFeatures[ extraFeaturesLabelsIdx ] = datasetPtr->getDouble(
              extraFeaturesIndexes[ extraFeaturesLabelsIdx ] );
          }

          extraFeaturesPtr->push_back( auxFeatures );
        }
      }
    }

    return true;
  }

  bool createDirectory( const std::string& dirName )
  {
    boost::filesystem::path dirPath( dirName );

    if( boost::filesystem::exists( dirPath ) )
    {
      if( ! boost::filesystem::is_directory( dirPath ) )
      {
        return false;
      }
    }
    else
    {
      try
      {
        if( ! boost::filesystem::create_directories( dirPath ) )
        {
          return false;
        }
      }
      catch(...)
      {
        return false;
      }
    }

    return true;
  }

  bool loadDataSetPropNames(
    const te::core::URI& dataSourceUri,
    const std::string& dataSetName,
    const std::string& dataSourceDriver,
    std::vector< std::string >& propNames )
  {
    propNames.clear();

    if( !dataSourceUri.isValid() )
    {
      return false;
    }

    if( dataSourceDriver.empty() )
    {
      return false;
    }

    try
    {
      std::unique_ptr< te::da::DataSource > dataSourcePtr( te::da::DataSourceFactory::make(
        dataSourceDriver, dataSourceUri ) );
      if( dataSourcePtr.get() == 0 )
      {
        return false;
      }

      dataSourcePtr->open();

      if( ! dataSourcePtr->isOpened() )
      {
        return false;
      }

      std::string internalDataSetName;

      if( dataSetName.empty() )
      {
        std::vector< std::string > dataSetNames = dataSourcePtr->getDataSetNames();

        if( dataSetNames.empty() )
        {
          return false;
        }

        internalDataSetName = dataSetNames[ 0 ];
      }
      else
      {
        internalDataSetName = dataSetName;
      }

      te::da::DataSourceTransactorPtr transactorPtr = dataSourcePtr->getTransactor();

      te::da::DataSetPtr dataSetPtr = transactorPtr->getDataSet( internalDataSetName );

      if( dataSetPtr.get() == 0 )
      {
        return false;
      }

      const std::size_t nProperties = dataSetPtr->getNumProperties();

      for( std::size_t propIdx = 0 ; propIdx < nProperties ; ++propIdx )
      {
        propNames.push_back( dataSetPtr->getPropertyName( propIdx ) );
      }
    }
    catch(...)
    {
      return false;
    }

    return true;
  }

  bool loadSamplesTextFile(
    const std::string& fileName,
    const std::string& separators,
    const std::string& classLabelsPropertyName,
    std::vector< unsigned int > const * const targetAttributesColumnsIdxPtr,
    std::vector< std::string >& attributeNames,
    std::vector< std::vector< double > >& samples,
    std::vector< std::string >& sampleLabels )
  {
    attributeNames.clear();
    samples.clear();
    sampleLabels.clear();

    std::fstream file;

    file.open( fileName.c_str(), std::fstream::in );

    if( !file.good() )
    {
      return false;
    }

    std::string lineStr;
    std::vector< std::string > lineTokens;
    unsigned int fileColumnsNumber = 0;
    std::size_t attributesColumnsIdxIdx = 0;
    std::size_t attributesColumnsIdxSize = 0;
    unsigned int classLabelsPropertyIdx = std::numeric_limits< unsigned int >::max();

    std::vector< unsigned int > attributesColumnsIdx;
    if( targetAttributesColumnsIdxPtr ) attributesColumnsIdx = *targetAttributesColumnsIdxPtr;

    try
    {
      // read header

      if( file.eof() )
      {
        return false;
      }
      else
      {
        std::getline( file, lineStr );

        // Finding tokens

        lineTokens.clear();
        te::common::Tokenize( lineStr, lineTokens, separators );

        if( lineTokens.empty() )
        {
          return false;
        }

        fileColumnsNumber = lineTokens.size();

        // locating the class label property index

        for( attributesColumnsIdxIdx = 0 ; attributesColumnsIdxIdx <
          fileColumnsNumber ; ++attributesColumnsIdxIdx )
        {
          if( lineTokens[ attributesColumnsIdxIdx ] == classLabelsPropertyName )
          {
            classLabelsPropertyIdx = attributesColumnsIdxIdx;
            break;
          }
        }
        if( classLabelsPropertyIdx == std::numeric_limits< unsigned int >::max() )
        {
          return false;
        }

        // other attribute Names

        if( attributesColumnsIdx.empty() )
        {
          for( attributesColumnsIdxIdx = 0 ; attributesColumnsIdxIdx <
            fileColumnsNumber ; ++attributesColumnsIdxIdx )
          {
            if( attributesColumnsIdxIdx != classLabelsPropertyIdx )
            {
              attributeNames.push_back( lineTokens[ attributesColumnsIdxIdx ] );
              attributesColumnsIdx.push_back( attributesColumnsIdxIdx );
            }
          }

          attributesColumnsIdxSize = attributesColumnsIdx.size();
        }
        else
        {
          attributesColumnsIdxSize = attributesColumnsIdx.size();

          for( attributesColumnsIdxIdx = 0 ; attributesColumnsIdxIdx <
            attributesColumnsIdxSize ; ++attributesColumnsIdxIdx )
          {
            if( attributesColumnsIdx[ attributesColumnsIdxIdx ] >= fileColumnsNumber )
            {
              return false;
            }

            if( attributesColumnsIdx[ attributesColumnsIdxIdx ] == classLabelsPropertyIdx )
            {
              return false;
            }

            attributeNames.push_back( lineTokens[ attributesColumnsIdx[ attributesColumnsIdxIdx ] ] );
          }
        }
      }

      // read data

      while( ! file.eof() )
      {
        lineStr.clear();
        std::getline( file, lineStr );

        if( ! lineStr.empty() )
        {
          lineTokens.clear();
          te::common::Tokenize( lineStr, lineTokens, separators );

          if( ! lineTokens.empty() )
          {
            if( lineTokens.size() != fileColumnsNumber ) return false;

            sampleLabels.push_back( lineTokens[ classLabelsPropertyIdx ] );

            samples.push_back( std::vector< double >() );
            std::vector< double >& sample = samples.back();
            sample.resize( attributesColumnsIdxSize );

            for( attributesColumnsIdxIdx = 0 ; attributesColumnsIdxIdx <
              attributesColumnsIdxSize ; ++attributesColumnsIdxIdx )
            {
              sample[ attributesColumnsIdxIdx ] = boost::lexical_cast< double >(
                lineTokens[ attributesColumnsIdx[ attributesColumnsIdxIdx ] ] );
            }
          }
        }
      }
    }
    catch(...)
    {
      return false;
    }

    return true;
  }

}  // end namespace sa
}  // end namespace gdma
