/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/semanticanalysis/ioFunctions.hpp
*/

#ifndef GEODMA_LIB_SEMANTICANALYSIS_IOFUNCTIONS_HPP_
#define GEODMA_LIB_SEMANTICANALYSIS_IOFUNCTIONS_HPP_

#include <boost/ptr_container/ptr_vector.hpp>
//
#include <terralib/geometry.h>
#include <terralib/raster.h>
#include <terralib/core.h>
#include <terralib/dataaccess.h>
//
#include <limits>
#include <map>
#include <memory>
#include <string>
#include <utility>
#include <vector>
#include <tuple>

// GeoDMA Includes
#include "config.hpp"
#include "typedefs.hpp"

namespace gdma {
namespace sa {

  /*!
    \brief Load and append all file tokens using space as separator.
    \param fileName Input file name.
    \param fileTokens Loaded tokens will be appended here, organized by filename + line number + tokens.
    \note Lines starting with # are ignored
    \note Lines sub-file recursion performed when a line with INCLUDE_FILE "filename.ext" is found.
    */
  GDMASEMANTICANALYSISEXPORT bool loadFileTokens( const std::string& fileName,
    std::vector< std::tuple< std::string, unsigned int, std::vector< std::string > > >& fileTokens );

  /*!
    \brief Load all file tokens using space as separator and ".
    \param fileName Input file name.
    \param fileTokens Output tokens organized by line number.
    \note Lines starting with # are ignored
    */
  GDMASEMANTICANALYSISEXPORT bool loadFileTokens( const std::string& fileName, std::map< unsigned int, std::vector< std::string > > & fileTokens );

  /*!
    \brief Load all file tokens using space as separator and ".
    \param fileName Input file name.
    \param fileTokens Output tokens organized by line number ( std::pair.first = line number ).
    \note Lines starting with # are ignored
    */
  GDMASEMANTICANALYSISEXPORT bool loadFileTokens( const std::string& fileName, std::vector< std::pair< unsigned int, std::vector< std::string > > >& fileTokens );

  /*!
    \brief Open a disk raster.
    \param rasterFileName Full path of an existent existent disk raster.
    \param rasterPtr Raster pointer.
    \return true if ok, false on errors.
    */
  GDMASEMANTICANALYSISEXPORT bool openDiskRaster( const std::string& rasterFileName,
    std::unique_ptr< te::rst::Raster >& rasterPtr );

  /*!
    \brief Open a raster from a given URI.
    \param inputRasterURI Input raster URI.
    \param rasterPtr Raster pointer.
    \return true if ok, false on errors.
    */
  GDMASEMANTICANALYSISEXPORT bool openRaster( const te::core::URI& inputRasterURI,
    std::unique_ptr< te::rst::Raster >& rasterPtr );

  /*!
    \brief Create a disk raster file copy.
    \param inputFileName Full path of an existent existent disk raster file.
    \param outputFileName Full path of the new disk raster file copy.
    \return true if ok, false on errors.
    */
  GDMASEMANTICANALYSISEXPORT bool diskRasterFileCopy( const std::string& inputFileName,
    const std::string& outputFileName );

  /*!
    \brief Create an empty result shape file.
    \param shapeFileName Output shape file full name.
    \return true if ok, false on errors.
    \note The geometries property name is assumed to follow the define GDMA_SA_DEFAULT_GEOM_PROP_NAME.
    \note The class name property name is assumed to follow the define GDMA_SA_DEFAULT_CLASSNAME_PROP_NAME.
    \note The class ID property name is assumed to follow the define GDMA_SA_DEFAULT_CLASSID_PROP_NAME.
  */
  GDMASEMANTICANALYSISEXPORT bool createEmptyResultShapeFile( const std::string& shapeFileName );

  /*!
    \brief Create a result shape file from the input geometries.
    \param inputGeometries Input geometries.
    \param shapeFileName Output shape file full name.
    \param geomLabelsPtr A pointer to geometries lables (must have the same size as inputGeometries) or a null pointer.
    \param geomsIDsPtr A pointer to Geometries IDs (must have the same size as inputGeometries) or a null pointer.
    \param extraPropertiesNamesPtr A pointer to extra properties names to save (or an null pointervector if none).
    \param extraPropertiesValuesPtr a pointer to extra properties values to save (points to a vector having the same size as inputGeometries or is null pointer);
    \return true if ok, false on errors.
    \note This function does not take the ownership of the geometries objects.
    \note The geometries property name is assumed to follow the define GDMA_SA_DEFAULT_GEOM_PROP_NAME.
    \note The class name property name is assumed to follow the define GDMA_SA_DEFAULT_CLASSNAME_PROP_NAME.
    \note The class ID property name is assumed to follow the define GDMA_SA_DEFAULT_CLASSID_PROP_NAME.
    \note Only polygons ( or multi-polygons ), closed line strings (and closed multi-line strings)  are supported
    \note extraPropertiesValues[ group index ][ geometry index ][ property index ].
  */
  GDMASEMANTICANALYSISEXPORT bool save2ResultShapeFile(
    const boost::ptr_vector< te::gm::Geometry >& inputGeometries,
    const std::string& shapeFileName,
    std::vector< std::string > const * const geomLabelsPtr,
    std::vector< GraphVerticeIDT > const * const geomsIDsPtr,
    std::vector< std::string > * const extraPropertiesNamesPtr,
    std::vector< std::vector< double > > * const extraPropertiesValuesPtr );

  /*!
    \brief Load a result shape file extra property names.
    \param shapeFileName Shape file full name.
    \param extraPropNames Extra features labels to load or an empty vector if none.
    \return true if ok, false on errors.
    \note The geometries property name will be ignored (defined by GDMA_SA_DEFAULT_GEOM_PROP_NAME ).
    \note The class name property name wil be ignored (defined by GDMA_SA_DEFAULT_CLASSNAME_PROP_NAME).
    \note The class ID property name will be ignored (defined by GDMA_SA_DEFAULT_CLASSID_PROP_NAME).
  */
  GDMASEMANTICANALYSISEXPORT bool loadResultShapeFileExtraPropNames(
    const std::string& shapeFileName,
    std::vector< std::string >& extraPropNames );

  /*!
    \brief Load a result shape file.
    \param shapeFileName Shape file full name.
    \param extraFeaturesLabelsPtr A optional pointer to extra features labels (property names) to load.
    \param geometriesPtr An optional pointer to a vector where geometries will be loaded.
    \param classNamesPtr An optional pointer to a vector where class names will be loaded.
    \param classIDsPtr An optional pointer to a vector where class Ids will be loaded.
    \param extraFeaturesPtr An optional pointer to a vector where extra features will be loaded (extraFeaturesLabelsPtr required).
    \return true if ok, false on errors.
    \note The geometries property name is assumed to follow the define GDMA_SA_DEFAULT_GEOM_PROP_NAME.
    \note The class name property name is assumed to follow the define GDMA_SA_DEFAULT_CLASSNAME_PROP_NAME.
    \note The class ID property name is assumed to follow the define GDMA_SA_DEFAULT_CLASSID_PROP_NAME.
  */
  GDMASEMANTICANALYSISEXPORT bool loadResultShapeFile(
    const std::string& shapeFileName,
    std::vector< std::string > const * const  extraFeaturesLabelsPtr,
    boost::ptr_vector< te::gm::Geometry >* const geometriesPtr,
    std::vector< std::string >* const classNamesPtr,
    std::vector< GraphVerticeIDT >* const classIDsPtr,
    std::vector< std::vector< double > >* const extraFeaturesPtr );

  /*!
    \brief Load shape file geometries.
    \param pathShp Shape file full path.
    \param gmVec The loaded geometries.
    \return true if ok, false on errors.
    \note The geometries property name is assumed to follow the define GDMA_SA_DEFAULT_GEOM_PROP_NAME.
    \note The class name property name is assumed to follow the define GDMA_SA_DEFAULT_CLASSNAME_PROP_NAME.
    \note The class ID property name is assumed to follow the define GDMA_SA_DEFAULT_CLASSID_PROP_NAME.
  */
  GDMASEMANTICANALYSISEXPORT bool loadResultShapeFileGeoms( const std::string& pathShp,
    boost::ptr_vector< te::gm::Geometry >& gmVec );

  /*!
    \brief Create a directory.
    \param dirName Full directory name.
    \return true if ok( or the requested directory already exists), false on errors.
    \note All parent directories will also be created if they do not exist.
  */
  GDMASEMANTICANALYSISEXPORT bool createDirectory( const std::string& dirName );

  /*!
    \brief Load data set data.
    \param dataSourceUri Data source connection info.
    \param dataSourceDriver Data source driver (i.e. OGR).
    \param dataSetName A data set name or an empty string to load properties from the first dataset found.
    \param classNamesPropertyName Class names property name or an empty string if this property must not be loeaded in any ordinary property.
    \param classIDsPropertyName Class IDs property name or an empty string if this property must not be loeaded in any ordinay property.
    \param otherPropertiesNames Other properties names to load or an empty vector..
    \param geometriesPtr A pointer to a vector where to store the loaded geometries or a null pointer.
    \param classNamesPtr A pointer to a vector where to store each class name.
    \param classIdsPtr A pointer to a vector where to store each class ID.
    \param otherPropertiesPtr A pointer to a vector where to store each loaded set of properties.
    \return true if ok, false on errors.
  */
  template< typename PropertyDataType >
  bool loadDataSetData(
    const te::core::URI& dataSourceUri,
    const std::string& dataSourceDriver,
    const std::string& dataSetName,
    const std::string& classNamesPropertyName,
    const std::string& classIDsPropertyName,
    const std::vector< std::string >& otherPropertiesNames,
    boost::ptr_vector< te::gm::Geometry > * const geometriesPtr,
    std::vector< std::string > * const classNamesPtr,
    std::vector< GraphVerticeIDT > * const classIdsPtr,
    std::vector< std::vector< PropertyDataType > > * const otherPropertiesPtr )
  {
    if( !dataSourceUri.isValid() )
    {
      return false;
    }

    if( dataSourceDriver.empty() )
    {
      return false;
    }

    if( geometriesPtr )
    {
      geometriesPtr->clear();
    }

    if( classNamesPtr )
    {
      classNamesPtr->clear();
    }

    if( classIdsPtr )
    {
      classIdsPtr->clear();
    }

    if( otherPropertiesPtr )
    {
      otherPropertiesPtr->clear();
    }

    const bool loadClassNames = ( ( ! classNamesPropertyName.empty() )
        && ( classNamesPtr!= 0 ) );
    const bool loadClassIDs = ( ( ! classIDsPropertyName.empty() )
      && ( classIdsPtr != 0 ) );
    const bool loadOtherProps = ( ( ! otherPropertiesNames.empty() ) &&
      ( otherPropertiesPtr != 0 ) );

    try
    {
      std::unique_ptr< te::da::DataSource > dataSourcePtr( te::da::DataSourceFactory::make(
        dataSourceDriver, dataSourceUri ) );
      if( dataSourcePtr.get() == 0 )
      {
        return false;
      }

      dataSourcePtr->open();

      if( ! dataSourcePtr->isOpened() )
      {
        return false;
      }

      std::string internalDataSetName;

      if( dataSetName.empty() )
      {
        std::vector< std::string > dataSetNames = dataSourcePtr->getDataSetNames();

        if( dataSetNames.empty() )
        {
          return false;
        }

        internalDataSetName = dataSetNames[ 0 ];
      }
      else
      {
        internalDataSetName = dataSetName;
      }

      std::unique_ptr< te::da::DataSourceTransactor > transactorPtr =
        dataSourcePtr->getTransactor();

      std::unique_ptr<te::da::DataSet> dataSetPtr = transactorPtr->getDataSet(
        internalDataSetName );

      if( dataSetPtr.get() == 0 )
      {
        return false;
      }

      // Find properties indexes

      const std::size_t invalidPropIdx = std::numeric_limits< std::size_t >::max();
      std::size_t geomPropIdx = invalidPropIdx;
      std::size_t classNamePropIdx = invalidPropIdx;
      std::size_t classIDPropIdx = invalidPropIdx;
      std::vector< std::size_t > otherPropsIndexes( otherPropertiesNames.size(),
        invalidPropIdx );

      {
        const std::size_t nProperties = dataSetPtr->getNumProperties();
        const std::size_t otherPropertiesNamesSize = otherPropertiesNames.size();
        std::size_t otherPropertiesNamesIdx = 0;

        for( std::size_t propIdx = 0 ; propIdx < nProperties ; ++propIdx )
        {
          if( dataSetPtr->getPropertyDataType( propIdx ) == te::dt::GEOMETRY_TYPE )
          {
            geomPropIdx = propIdx;
          }

          if(
              loadClassNames
              &&
              ( dataSetPtr->getPropertyName( propIdx ) == classNamesPropertyName )
            )
          {
            classNamePropIdx = propIdx;
          }

          if(
              loadClassIDs
              &
              ( dataSetPtr->getPropertyName( propIdx ) == classIDsPropertyName )
            )
          {
            classIDPropIdx = propIdx;
          }

          if( loadOtherProps )
          {
            for( otherPropertiesNamesIdx = 0; otherPropertiesNamesIdx < otherPropertiesNamesSize ;
              ++otherPropertiesNamesIdx )
            {
              if( dataSetPtr->getPropertyName( propIdx ) ==
                otherPropertiesNames[ otherPropertiesNamesIdx ] )
              {
                otherPropsIndexes[ otherPropertiesNamesIdx ] = propIdx;
                break;
              }
            }
          }
        }

        if(
            (
              loadClassNames
              &
              ( classNamePropIdx == invalidPropIdx )
            )
            ||
            (
              loadClassIDs
              &&
              ( classIDPropIdx == invalidPropIdx )
            )
            ||
            (
              ( geometriesPtr != 0 )
              &&
              ( geomPropIdx == invalidPropIdx )
            )
          )
        {
          return false;
        }

        if( loadOtherProps )
        {
          for( otherPropertiesNamesIdx = 0; otherPropertiesNamesIdx < otherPropertiesNamesSize ;
            ++otherPropertiesNamesIdx )
          {
            if( otherPropsIndexes[ otherPropertiesNamesIdx ] ==
              invalidPropIdx )
            {
              return false;
            }
          }
        }
      }

      // Load

      {
        const std::size_t otherPropertiesNamesSize = otherPropertiesNames.size();
        std::size_t otherPropertiesNamesIdx = 0;

        if( ! dataSetPtr->moveBeforeFirst() )
        {
          return false;
        }

        while( dataSetPtr->moveNext() )
        {
          if( geometriesPtr )
          {
            geometriesPtr->push_back( dataSetPtr->getGeometry( geomPropIdx ).release() );
          }

          if( loadClassNames )
          {
            classNamesPtr->push_back( dataSetPtr->getAsString( classNamePropIdx ) );
          }

          if( loadClassIDs )
          {
            classIdsPtr->push_back( boost::lexical_cast< GraphVerticeIDT >(
              dataSetPtr->getAsString( classNamePropIdx ) ) );
          }

          if( loadOtherProps )
          {
            otherPropertiesPtr->push_back( std::vector< PropertyDataType >() );
            std::vector< PropertyDataType >& outPropVec = otherPropertiesPtr->back();
            outPropVec.resize( otherPropertiesNamesSize );

            for( otherPropertiesNamesIdx = 0; otherPropertiesNamesIdx < otherPropertiesNamesSize ;
              ++otherPropertiesNamesIdx )
            {
              outPropVec[ otherPropertiesNamesIdx ] = boost::lexical_cast< PropertyDataType >(
                dataSetPtr->getAsString( otherPropsIndexes[ otherPropertiesNamesIdx ] ) );
            }
          }
        }
      }
    }
    catch(...)
    {
      return false;
    }

    return true;
  }

  /*!
    \brief Load a dataset properties names.
    \param dataSourceUri Data source URI.
    \param dataSetName A data set name or an empty string to load properties from the first dataset found.
    \param dataSourceDriver Data source driver (i.e. OGR).
    \param propNames Features (property names).
    \return true if ok, false on errors.
  */
  GDMASEMANTICANALYSISEXPORT bool loadDataSetPropNames(
    const te::core::URI& dataSourceUri,
    const std::string& dataSetName,
    const std::string& dataSourceDriver,
    std::vector< std::string >& propNames );

    /*!
      \brief Load a samples text file.
      \param fileName File name.
      \param separators The used separators.
      \param classLabelsPropertyName The class lables property (column) name .
      \param targetAttributesColumnsIdxPtr A pointer to a vector of target attributes columns indexes to load or a null pointer to load all attributes.
      \param attributeNames Loaded attribute (columns) names.
      \param samples Loaded samples.
      \param sampleLabels Loaded samples lables.
      \return true if ok, false on errors
    */
  GDMASEMANTICANALYSISEXPORT bool loadSamplesTextFile(
    const std::string& fileName,
    const std::string& separators,
    const std::string& classLabelsPropertyName,
    std::vector< unsigned int > const * const targetAttributesColumnsIdxPtr,
    std::vector< std::string >& attributeNames,
    std::vector< std::vector< double > >& samples,
    std::vector< std::string >& sampleLabels );


}  // end namespace sa
}  // end namespace gdma

#endif  // GEODMA_LIB_SEMANTICANALYSIS_IOFUNCTIONS_HPP_
