/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/semanticanalysis/localExecuter.hpp
*/

#ifndef GEODMA_LIB_SEMANTICANALYSIS_LOCALEXECUTER_HPP_
#define GEODMA_LIB_SEMANTICANALYSIS_LOCALEXECUTER_HPP_

#include <map>
#include <mutex>

// GeoDMA Includes
#include "config.hpp"
#include "executer.hpp"
#include "executerFactory.hpp"
#include "../taskprocessing/executerCallBack.hpp"
#include "../taskprocessing/defines.hpp"
#include "../taskprocessing/graphExecuter.hpp"

namespace gdma {
namespace sa {

/*!
  \class LocalExecuter

  \brief A basic concept of a local project executer.
*/
class GDMASEMANTICANALYSISEXPORT LocalExecuter : public Executer, protected gdma::tp::ExecuterCallBack {
 public:
    LocalExecuter();

    ~LocalExecuter();

    //  overload
    bool initialize( Project& project, const ExecMode execMode );

    //  overload
    bool finalize();

    //  overload
    void startExecution();

    //  overload
    void stopExecution();

    //  overload
    void waitExecution();

    //  overload
    ExecuterStatus getExecuterStatus() const;

    //  overload
    bool isInitialized() const;

    //  overload
    unsigned int getTasksCount() const;

 protected:
    /*! \brief True if this instance is initialized. */
    bool m_isInitialized;

    /*! \brief A pointer to the project to be executed. */
    Project* m_projectPtr;

    /*! \brief Current execution mode. */
    ExecMode m_execMode;

    /*! \brief Internal tasks executer. */
    gdma::tp::GraphExecuter m_executer;

    /*! \brief A map from vertice internal executer tasks IDs to the related vertices. */
    std::map< gdma::tp::TaskIdType, GraphVertice* > m_taskID2VerticeMap;

    /*! \brief Internal mutex. */
    mutable std::mutex m_mutex;

    /*! \brief Default constructor. */
    LocalExecuter(const LocalExecuter& other);

    /*!
     \brief Reset this instance to it's initial state.
     \note THIS IS NOT A THREAD SAFE METHOD.
     */
    void reset();

    //  overload
    void taskStarted(const gdma::tp::TaskIdType& taskId);

    //  overload
    void taskExecuted(const gdma::tp::TaskIdType& taskId,
      const gdma::tp::TaskExecutionResult::BasicResult& basicResut );
};

class LocalExecuterFactory : public ExecuterFactory {
 public:
    LocalExecuterFactory();
    ~LocalExecuterFactory();

 protected:
    // overload
    Executer* build();
};

}  // end namespace sa
}  // end namespace gdma

#endif  // GEODMA_LIB_SEMANTICANALYSIS_LOCALEXECUTER_HPP_
