/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/semanticanalysis/project.hpp
*/

#ifndef GEODMA_LIB_SEMANTICANALYSIS_PROJECT_HPP_
#define GEODMA_LIB_SEMANTICANALYSIS_PROJECT_HPP_

#include <string>
#include <vector>
#include <map>
#include <mutex>

// GeoDMA Includes
#include "config.hpp"
#include "analysis.hpp"
#include "executionResult.hpp"

namespace gdma {
namespace sa {

/*!
  \class Project

  \brief A basic concept of a Project.
*/
class GDMASEMANTICANALYSISEXPORT Project {
 public:
    Project();

    ~Project();

    Project(const Project& other);

    Project& operator=( const Project& other );

    /*!
      \brief Returns the project name.
      \return Returns the project name.
      \note This is a thread safe method.
     */
    const std::string& getName() const;

    /*!
      \brief Set the Project name.
      \param newName The Project name.
      \note This is a thread safe method.
     */
    void setName( const std::string& newName );

    /*!
      \brief Add a new analysis to this project.
      \param analysis The analysis to be added.
      \return true if ok, false on errors.
      \note This is a thread safe method.
     */
    bool addAnalysis( const Analysis& analysis );

    /*!
      \brief Remove a analysis from this project.
      \param analysisName The analysis name.
      \return true if ok, false on errors.
      \note This is a thread safe method.
     */
    bool removeAnalysis( const std::string& analysisName );

    /*!
      \brief Returns a pointer the required Analysis.
      \param analysisName Analysis name.
      \return A pointer to the required analysis or a NULL pointer on errors.
      \note This is a thread safe method.
     */
    Analysis* getAnalysis( const std::string& analysisName );

    /*!
      \brief Returns all current analysis names.
      \param analysisNames Returns all current analysis names.
      \note This is a thread safe method.
     */
    void getAnalysisNames( std::vector< std::string >& analysisNames ) const;

    /*!
      \brief Set a output directory to this Project.
      \param outDir Output directory to this Project.
      \return true if OK, false on errors.
      \note This is a thread safe method.
      \note The provided path must be canonical and system complete.
      */
    bool setOutputDirectoryName( const std::string& outDirName );

    /*!
      \brief Returns the output directory to this Project.
      \return Returns the output directory to this Project.
      \note This is a thread safe method.
      */
    const std::string& getOutputDirectoryName() const;

    /*!
      \brief Reset the project.
      \note This is a thread safe method.
      */
    void reset();

    /*!
      \brief Enable (true) or disable (false) the use of result cache.
      \param enabled Enable (true) or disable (false) the use of result cache.
      \note This is a thread safe method.
      */
    void enableCache( const bool enabled );

    /*!
      \brief Returns true if cache is enabled, false otherwise.
      \return Returns true if cache is enabled, false otherwise.
      \note This is a thread safe method.
      */
    bool isCacheEnabled() const;

    /*!
      \brief Set a reference directory to this Project (where the project file is located).
      \param outDir Reference directory to this Project (where the project file is located)
      \return true if OK, false on errors.
      \note This is a thread safe method.
      \note The provided path must be canonical and system complete.
      */
    bool setProjectDirectoryName( const std::string& dirName );

    /*!
      \brief Returns the reference directory to this Project (where the project file is located).
      \return Returns reference directory to this Project (where the project file is located).
      \note This is a thread safe method.
      */
    const std::string& getProjectDirectoryName() const;

    /*!
      \brief Returns the total number of vertices (all analysis included).
      \return Returns the total number of vertices (all analysis included).
      \note This is a thread safe method.
      */
    unsigned int getTotalVerticesCount() const;

 private:
    /*! \brief Tells if result cache must be used. */
    bool m_enableCache;

    /*! \brief Project name. */
    std::string m_projectName;

    /*! \brief Project output directory. */
    std::string m_outputDirectoryName;

    /*! \brief Project directory (where the project file is located). */
    std::string m_projectDirectoryName;

    /*! \brief Project analises. */
    std::map< std::string, Analysis > m_analises;

    /*! \brief Internal vertice mutex. */
    mutable std::mutex m_mutex;
};

}  // end namespace sa
}  // end namespace gdma

#endif  // GEODMA_LIB_SEMANTICANALYSIS_PROJECT_HPP_
