/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

// #include <terralib/dataaccess.h>
// #include <terralib/memory.h>
// #include <terralib/rp.h>
// #include <terralib/datatype.h>
// #include <terralib/sam/rtree.h>
// #include <terralib/common.h>
//
// #include <boost/filesystem.hpp>
//
#include <string>
#include <memory>
#include <vector>
#include <map>
// #include <fstream>
// #include <limits>
// #include <algorithm>
// #include <set>

// GeoDMA Includes
#include "rasterFunctions.hpp"
#include "generalFunctions.hpp"
#include "defines.hpp"
// #include "../common/utils.hpp"

namespace gdma {
namespace sa {

  bool vectorizeResultRaster(
    const te::rst::Raster& inputRaster,
    boost::ptr_vector< te::gm::Geometry >& vectors,
    std::vector< GraphVerticeIDT >* vectorsIDsPtr )
  {
    vectors.clear();
    if( vectorsIDsPtr ) vectorsIDsPtr->clear();

    std::vector< te::gm::Geometry * > geomsPtrs;
    std::vector< double > geomsPixelValues;
    ((te::rst::Raster*)(&inputRaster))->vectorize( geomsPtrs, 0, 0, &geomsPixelValues );

    const std::size_t geomsPtrsSize = geomsPtrs.size();
    const double rasterNoDataValue = inputRaster.getBand( 0 )->getProperty()->m_noDataValue;

    for( std::size_t geomsPtrsIdx = 0 ; geomsPtrsIdx < geomsPtrsSize ; ++geomsPtrsIdx )
    {
      if( geomsPixelValues[ geomsPtrsIdx ] != rasterNoDataValue )
      {
        vectors.push_back( geomsPtrs[ geomsPtrsIdx ] );

        if( vectorsIDsPtr ) vectorsIDsPtr->push_back( (GraphVerticeIDT)
          geomsPixelValues[ geomsPtrsIdx ] );
      }
      else
      {
        delete geomsPtrs[ geomsPtrsIdx ];
      }
    }

    return true;
  }

  bool vectorizeResultRaster(
    const std::string& inputRasterFileName,
    boost::ptr_vector< te::gm::Geometry >& vectors,
    std::vector< GraphVerticeIDT >* vectorsIDsPtr )
  {
    std::map<std::string, std::string> inputRasterInfo;
    inputRasterInfo["URI"] = inputRasterFileName;

    std::unique_ptr< te::rst::Raster > rasterPtr;

    try
    {
      rasterPtr.reset( te::rst::RasterFactory::open(  inputRasterInfo ) );
    }
    catch(...)
    {
      rasterPtr.reset();
    }

    if( rasterPtr.get() == 0 )
    {
      return false;
    }

    return vectorizeResultRaster( *rasterPtr, vectors, vectorsIDsPtr );
  }

  bool clipAndReprojectRaster(
    const te::rst::Raster& inputRaster, te::gm::Envelope const * const targetAreaPtr,
    int& targetSRID, std::unique_ptr< te::rst::Raster >& outputRasterPtr )
  {
    if( targetSRID < 1 ) return false;

    outputRasterPtr.reset();

    // target area

    te::gm::Envelope internalTargetArea;
    if( targetAreaPtr )
    {
      internalTargetArea = *targetAreaPtr;
    }
    else
    {
      internalTargetArea = *( inputRaster.getGrid()->getExtent() );
      internalTargetArea.transform( inputRaster.getSRID(), targetSRID );
    }

    te::gm::Envelope rasterEnvelopeReprojected = *( inputRaster.getGrid()->getExtent() );
    rasterEnvelopeReprojected.transform( inputRaster.getSRID(), targetSRID );

    // checking if there is overlap between the input geometries and the given raster

    if(
        rasterEnvelopeReprojected.touches( internalTargetArea )
        ||
        rasterEnvelopeReprojected.disjoint( internalTargetArea )
      )
    {
      return false;
    }
    else
    {
      te::gm::Envelope intersectionEnvelope = rasterEnvelopeReprojected.intersection(
        internalTargetArea );

      te::gm::Envelope intersectionEnvelopeReprojected = intersectionEnvelope;
      intersectionEnvelopeReprojected.transform( targetSRID,
        inputRaster.getGrid()->getSRID() );

      if( inputRaster.getGrid()->getSRID() == targetSRID )
      {
        // raster clipping

        std::map< std::string, std::string >  rinfo;
        rinfo[ "RTYPE" ] = "EXPANSIBLE";

        try
        {
          outputRasterPtr.reset( inputRaster.trim( &intersectionEnvelopeReprojected,
            rinfo ) );
        }
        catch(...)
        {
          outputRasterPtr.reset();
        };

        if( outputRasterPtr.get() == 0 )
        {
          return false;
        }
      }
      else
      {
        // raster reprojection

        std::map< std::string, std::string >  rinfo;
        rinfo[ "RTYPE" ] = "EXPANSIBLE";
        rinfo[ "USE_TERRALIB_REPROJECTION" ] = "TRUE";

        try
        {
          outputRasterPtr.reset(
            inputRaster.transform(
              targetSRID,
              intersectionEnvelopeReprojected.m_llx,
              intersectionEnvelopeReprojected.m_lly,
              intersectionEnvelopeReprojected.m_urx,
              intersectionEnvelopeReprojected.m_ury,
              rasterEnvelopeReprojected.getWidth() /
                static_cast< double >( inputRaster.getNumberOfColumns() ),
              rasterEnvelopeReprojected.getHeight() /
                static_cast< double >( inputRaster.getNumberOfRows() ),
              rinfo,
              te::rst::NearestNeighbor
            )
          );
        }
        catch(...)
        {
          outputRasterPtr.reset();
        }

        if( outputRasterPtr.get() == 0 )
        {
          return false;
        }
      }
    }

    return true;
  }

}  // end namespace sa
}  // end namespace gdma
