/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

#include <boost/filesystem.hpp>
#include <boost/lexical_cast.hpp>

#include <terralib/common.h>

#include <string>
#include <vector>
#include <utility>
#include <map>
#include <memory>
#include <limits>
#include <tuple>

// GeoDMA Includes
#include "textFileProjectIO.hpp"
#include "context.hpp"
#include "graph.hpp"
#include "graphVertice.hpp"
#include "graphVerticeFactory.hpp"
#include "generalFunctions.hpp"
#include "ioFunctions.hpp"

namespace
{
  static gdma::sa::TextFileProjectIOFactory TextFileProjectIOFactoryInstance;
}

namespace gdma {
namespace sa {

  TextFileProjectIO::TextFileProjectIO()
  {
  }

  TextFileProjectIO::TextFileProjectIO( const TextFileProjectIO& )
  {
  }

  TextFileProjectIO::~TextFileProjectIO()
  {
  }

  TextFileProjectIO& TextFileProjectIO::operator=( const TextFileProjectIO& )
  {
    return *this;
  }

  bool TextFileProjectIO::load( const std::string& fileName, Project& proj,
    std::string& errorMessage ) const
  {
    proj.reset();
    errorMessage.clear();

    // Checking projectFileName

    boost::filesystem::path projectFilePath = boost::filesystem::system_complete( fileName );

    if( ! projectFilePath.is_absolute() )
    {
      errorMessage = "Project file loading error - Project file path must be absolute";
      return false;
    }

    if( ! proj.setProjectDirectoryName( projectFilePath.parent_path().string() ) )
    {
      errorMessage = "Project file loading error - Invalid Project file path";
      return false;
    }

    if( ! proj.setOutputDirectoryName( projectFilePath.parent_path().string() ) )
    {
      errorMessage = "Project file loading error - Invalid Project file path";
      return false;
    }

    std::vector< std::tuple< std::string, unsigned int, std::vector< std::string > > > fileTokens;
    if( ! loadFileTokens( fileName, fileTokens ) )
    {
      errorMessage = "Project file loading error";
      return false;
    }

    std::map< std::string, Context > contexts;
    if( !createContexts( fileTokens, contexts, errorMessage ) )
    {
      errorMessage = "Project file loading error - " + errorMessage;
      return false;
    }

    std::vector< std::tuple< std::string, std::string, std::string > > analysisData;
    if( !createProject( fileTokens, proj, analysisData, errorMessage ) )
    {
      errorMessage = "Project creation error - " + errorMessage;
      return false;
    }

    if( !createAnalysis( fileTokens, contexts, analysisData, proj, errorMessage ) )
    {
      errorMessage = "Project creation error error - " + errorMessage;
      return false;
    }

//     // Loading the project file
//
//     std::vector< boost::filesystem::path > contextFilePaths;
//     std::vector< boost::filesystem::path > graphFilePaths;
//
//     // key: "Analysys Name" first="Context Name" second="Graph Name"
//     std::map< std::string, std::pair< std::string, std::string > > analysisSet;
//
//     {
//       std::map< unsigned int, std::vector< std::string > > fileTokens;
//       if( ! loadFileTokens( fileName, fileTokens ) )
//       {
//         errorMessage = "Project file loading error";
//         return false;
//       }
//
//       std::map< unsigned int, std::vector< std::string > >::const_iterator fileTokensIt =
//         fileTokens.begin();
//       std::map< unsigned int, std::vector< std::string > >::const_iterator fileTokensItEnd =
//         fileTokens.end();
//
//       while( fileTokensIt != fileTokensItEnd )
//       {
//         if(
//             ( fileTokensIt->second[ 0 ] == "PROJECT_NAME" )
//             &&
//             ( fileTokensIt->second.size() == 2 )
//           )
//         {
//           proj.setName( fileTokensIt->second[ 1 ] );
//         }
//         else if(
//                   (fileTokensIt->second[ 0 ] == "CONTEXT_FILE" )
//                   &&
//                   ( fileTokensIt->second.size() == 2 )
//                )
//         {
//           boost::filesystem::path auxPath;
//
//           try
//           {
//             auxPath =
//               boost::filesystem::system_complete(
//                 projectFilePath.parent_path()
//                 /
//                 boost::filesystem::path( fileTokensIt->second[ 1 ] )
//               );
//           }
//           catch(...)
//           {
//             errorMessage = "Project file loading error - Invalid project file path"
//               + fileName + ":" +
//               boost::lexical_cast< std::string >( fileTokensIt->first );
//             return false;
//           }
//
//           contextFilePaths.push_back( auxPath );
//         }
//         else if(
//                   ( fileTokensIt->second[ 0 ] == "GRAPH_FILE" )
//                   &&
//                   ( fileTokensIt->second.size() == 2 )
//                )
//         {
//           boost::filesystem::path auxPath;
//
//           try
//           {
//             auxPath =
//               boost::filesystem::system_complete(
//                   projectFilePath.parent_path()
//                   /
//                   boost::filesystem::path( fileTokensIt->second[ 1 ] )
//                 );
//           }
//           catch(...)
//           {
//             errorMessage = "Project file loading error - Invalid graph file path"
//               + fileName + ":" +
//               boost::lexical_cast< std::string >( fileTokensIt->first );
//             return false;
//           }
//
//           graphFilePaths.push_back( auxPath );
//         }
//         else if(
//                   ( fileTokensIt->second[ 0 ] == "ANALYSIS" )
//                   &&
//                   ( fileTokensIt->second.size() == 4 )
//                )
//         {
//           if( analysisSet.find( fileTokensIt->second[ 1 ] ) == analysisSet.end() )
//           {
//             std::pair< std::string, std::string > auxPair;
//             auxPair.first = fileTokensIt->second[ 2 ];
//             auxPair.second = fileTokensIt->second[ 3 ];
//
//             analysisSet[ fileTokensIt->second[ 1 ] ] = auxPair;
//           }
//           else
//           {
//             errorMessage = "Project file loading error - "
//               + fileName + ":" +
//               boost::lexical_cast< std::string >( fileTokensIt->first );
//
//             return false;
//           }
//         }
//         else
//         {
//           errorMessage = "Project file loading error - "
//             + fileName + ":" +
//             boost::lexical_cast< std::string >( fileTokensIt->first );
//
//           return false;
//         }
//
//         ++fileTokensIt;
//       }
//     }
//
//     // loading contexts
//
//     std::map< std::string, Context > contexts;
//
//     {
//       for( std::size_t contextFilePathsIdx = 0 ; contextFilePathsIdx <
//         contextFilePaths.size() ; ++contextFilePathsIdx )
//       {
//         const boost::filesystem::path& contextFilePath = contextFilePaths[
//           contextFilePathsIdx ];
//         if( ! contextFilePath.is_absolute() )
//         {
//           throw std::runtime_error( "Context file path must be absolute" );
//         }
//
//         std::map< unsigned int, std::vector< std::string > > fileTokens;
//         if( ! loadFileTokens( contextFilePath.string(),
//           fileTokens ) )
//         {
//           errorMessage = "Context file loading error:" +
//             contextFilePaths[ contextFilePathsIdx ].string();
//           return false;
//         }
//
//         Context newContext;
//         newContext.setContextDirectoryName( contextFilePath.parent_path().string() );
//
//         std::map< unsigned int, std::vector< std::string > >::const_iterator fileTokensIt =
//           fileTokens.begin();
//         std::map< unsigned int, std::vector< std::string > >::const_iterator fileTokensItEnd =
//           fileTokens.end();
//
//         while( fileTokensIt != fileTokensItEnd )
//         {
//           if(
//               ( fileTokensIt->second[ 0 ] == "CONTEXT_NAME" )
//               &&
//               ( fileTokensIt->second.size() == 2 )
//             )
//           {
//             newContext.setName( fileTokensIt->second[ 1 ] );
//           }
//           else if(
//                     ( fileTokensIt->second[ 0 ] == "RESOURCE_URI" )
//                     &&
//                     ( fileTokensIt->second.size() == 3 )
//                  )
//           {
//             std::string resourceSTR = fileTokensIt->second[ 2 ];
//
//             if( te::common::Convert2UCase( resourceSTR.substr( 0, 7 ) ) == "FILE://" )
//             {
//               resourceSTR = resourceSTR.substr( 7, resourceSTR.size() - 7 );
//
//               boost::filesystem::path resoucePath;
//               try
//               {
//                 resoucePath =
//                   boost::filesystem::canonical(
//                     boost::filesystem::system_complete(
//                       contextFilePath.parent_path()
//                       /
//                       boost::filesystem::path( resourceSTR )
//                     )
//                   );
//               }
//               catch(...)
//               {
//                 errorMessage = "Context file loading error - Invalid resource path - " +
//                   contextFilePaths[ contextFilePathsIdx ].string() + ":" +
//                   boost::lexical_cast< std::string >( fileTokensIt->first );
//
//                 return false;
//               }
//
//               te::core::URI uri( "file://" + resoucePath.string() );
//
//               if( !newContext.addResource( fileTokensIt->second[ 1 ], uri ) )
//               {
//                 errorMessage = "Context file loading error - " +
//                   contextFilePaths[ contextFilePathsIdx ].string() + ":" +
//                   boost::lexical_cast< std::string >( fileTokensIt->first );
//
//                 return false;
//               }
//             }
//             else
//             {
//               te::core::URI uri( resourceSTR );
//
//               if( !newContext.addResource( fileTokensIt->second[ 1 ], uri ) )
//               {
//                 errorMessage = "Context file loading error - " +
//                   contextFilePaths[ contextFilePathsIdx ].string() + ":" +
//                   boost::lexical_cast< std::string >( fileTokensIt->first );
//
//                 return false;
//               }
//             }
//           }
//           else
//           {
//             errorMessage = "Context file loading error - " +
//               contextFilePaths[ contextFilePathsIdx ].string() + ":" +
//               boost::lexical_cast< std::string >( fileTokensIt->first );
//
//             return false;
//           }
//
//           ++fileTokensIt;
//         }
//
//         if(
//             ( ! newContext.getName().empty() )
//             &&
//             ( newContext.getResourcesCount() != 0 )
//           )
//         {
//           contexts[ newContext.getName() ] = newContext;
//         }
//       }
//     }
//
//     // Creating empty analysis objects (without graphs or contexts)
//
//     {
//       std::map< std::string, std::pair< std::string, std::string > >::const_iterator anaIt =
//         analysisSet.begin();
//       std::map< std::string, std::pair< std::string, std::string > >::const_iterator anaItEnd =
//         analysisSet.end();
//
//       while( anaIt != anaItEnd )
//       {
//         if( contexts.find( anaIt->second.first ) == contexts.end() )
//         {
//           errorMessage = "Project file loading error - Invalid analysis context: " +
//             anaIt->first;
//
//           return false;
//         }
//
//         Analysis newAnalysis;
//         newAnalysis.setName( anaIt->first );
//         newAnalysis.setContext( contexts[ anaIt->second.first ] );
//
//         if( ! proj.addAnalysis( newAnalysis ) )
//         {
//           errorMessage = "Project file loading error - Invalid analysis: " +
//             anaIt->first;
//
//           return false;
//         }
//
//         ++anaIt;
//       }
//     }
//
//     // loading graphs tokens
//     // [ graph name : graph tokens ]
//     std::map< std::string , std::vector< std::pair< unsigned int, std::vector< std::string > > > > graphsTokens;
//
//     {
//       for( std::size_t graphFilePathsIdx = 0 ; graphFilePathsIdx <
//         graphFilePaths.size() ; ++graphFilePathsIdx )
//       {
//         std::vector< std::pair< unsigned int, std::vector< std::string > > > fileTokens;
//         if( ! loadFileTokens( graphFilePaths[ graphFilePathsIdx ].string(),
//           fileTokens ) )
//         {
//           errorMessage = "Graph file loading error:" +
//             graphFilePaths[ graphFilePathsIdx ].string();
//
//           return false;
//         }
//
//         // Looking for the graph name
//
//         std::vector< std::pair< unsigned int, std::vector< std::string > > >::const_iterator fileTokensIt =
//           fileTokens.begin();
//         std::vector< std::pair< unsigned int, std::vector< std::string > > >::const_iterator fileTokensItEnd =
//           fileTokens.end();
//         std::string graphName;
//
//         while( fileTokensIt != fileTokensItEnd )
//         {
//           if(
//               ( fileTokensIt->second.size() == 2 )
//               &&
//               ( fileTokensIt->second[ 0 ] == "GRAPH_NAME" ) )
//           {
//             graphName = fileTokensIt->second[ 1 ];
//             break;
//           }
//
//           ++fileTokensIt;
//         }
//
//         if( graphName.empty() )
//         {
//           errorMessage = "Unable to find the graph name - Graph file:" +
//             graphFilePaths[ graphFilePathsIdx ].string();
//
//           return false;
//         }
//
//         if( graphsTokens.find( graphName ) != graphsTokens.end() )
//         {
//           errorMessage = "Duplicated graph name:" +
//             graphFilePaths[ graphFilePathsIdx ].string();
//
//           return false;
//         }
//
//         graphsTokens[ graphName ] = fileTokens;
//       }
//     }
//
//     // Creating empty graphs for each analysis
//
//     {
//       std::map< std::string, std::pair< std::string, std::string > >::const_iterator anaIt =
//         analysisSet.begin();
//       std::map< std::string, std::pair< std::string, std::string > >::const_iterator anaItEnd =
//         analysisSet.end();
//
//       while( anaIt != anaItEnd )
//       {
//         const std::string& anaName = anaIt->first;
//
//         Analysis* anaPtr = proj.getAnalysis( anaName );
//         if( anaPtr == 0 )
//         {
//           errorMessage = "Graph creation error - Invalid analysis: " +
//             anaName;
//
//           return false;
//         }
//
//         const std::string& graphName = anaIt->second.second;
//
//         if( graphsTokens.find( graphName ) == graphsTokens.end() )
//         {
//           errorMessage = "Analsys (" + anaName + ") creation error - Invalid graph name: " +
//             graphName;
//
//           return false;
//         }
//
//         const std::vector< std::pair< unsigned int, std::vector< std::string > > >& graphTokens =
//           graphsTokens[ graphName ];
//         std::vector< std::pair< unsigned int, std::vector< std::string > > >::const_iterator fileTokensIt =
//           graphTokens.begin();
//         const std::vector< std::pair< unsigned int, std::vector< std::string > > >::const_iterator fileTokensItEnd =
//           graphTokens.end();
//
//         while( fileTokensIt != fileTokensItEnd )
//         {
//           if(
//               ( fileTokensIt->second.size() == 2 )
//               &&
//               ( fileTokensIt->second[ 0 ] == "GRAPH_NAME" ) )
//           {
//             Graph newGraph;
//             newGraph.setName( fileTokensIt->second[ 1 ] );
//
//             if( ! anaPtr->setGraph( newGraph ) )
//             {
//               errorMessage = "Analsys (" + anaName + ") creation error - Invalid graph: " +
//                 graphName;
//
//               return false;
//             }
//
//             break;
//           }
//
//           ++fileTokensIt;
//         }
//
//         ++anaIt;
//       }
//     }
//
//     // Creating graphs vertices
//
//     {
//       std::map< std::string, std::pair< std::string, std::string > >::const_iterator anaIt =
//         analysisSet.begin();
//       std::map< std::string, std::pair< std::string, std::string > >::const_iterator anaItEnd =
//         analysisSet.end();
//
//       while( anaIt != anaItEnd )
//       {
//         const std::string& anaName = anaIt->first;
//
//         Analysis* anaPtr = proj.getAnalysis( anaName );
//         if( anaPtr == 0 )
//         {
//           errorMessage = "Graph creation error - Invalid analysis: " +
//             anaName;
//
//           return false;
//         }
//
//         Graph& graph = anaPtr->getGraph();
//
//         const std::string& graphName = graph.getName();
//
//         const std::vector< std::pair< unsigned int, std::vector< std::string > > >& graphTokens =
//           graphsTokens[ graphName ];
//         if( graphTokens.empty() )
//         {
//           errorMessage = "Graph creation error - Invalid graph: " +
//             graphName;
//
//           return false;
//         }
//
//         bool verticeIsBeingBuild = false;
//         std::string verticeName;
//         std::string verticeType;
//         bool vertice2ndPassEnabled = false;
//         bool verticeCacheEnabled = false;
//         std::vector< std::pair< std::string, std::string > > verticeParams;
//         std::vector< std::pair< std::string, std::string > > verticeResources;
//         std::vector< std::pair< std::string, std::string > > verticeConnections;
//         const std::vector< std::pair< GraphVerticeIDT, std::string > > emptyInputConnectionsData;
//
//         std::vector< std::pair< unsigned int, std::vector< std::string > > >::const_iterator fileTokensIt =
//           graphTokens.begin();
//         const std::vector< std::pair< unsigned int, std::vector< std::string > > >::const_iterator fileTokensItEnd =
//           graphTokens.end();
//
//         while( fileTokensIt != fileTokensItEnd )
//         {
//           if(
//               ( !verticeIsBeingBuild )
//               &&
//               ( fileTokensIt->second.size() == 2 )
//               &&
//               ( fileTokensIt->second[ 0 ] == "GRAPH_NAME" )
//             )
//           {
//             // do nothing
//           }
//           else if(
//               ( !verticeIsBeingBuild )
//               &&
//               ( fileTokensIt->second.size() == 1 )
//               &&
//               ( fileTokensIt->second[ 0 ] == "VERTICE_START" )
//              )
//           {
//             verticeIsBeingBuild = true;
//             verticeName.clear();
//             verticeType.clear();
//             vertice2ndPassEnabled = false;
//             verticeCacheEnabled = false;
//             verticeParams.clear();
//             verticeResources.clear();
//             verticeConnections.clear();
//           }
//           else if(
//                     ( verticeIsBeingBuild )
//                     &&
//                     ( fileTokensIt->second.size() == 1 )
//                     &&
//                     ( fileTokensIt->second[ 0 ] == "VERTICE_END" )
//                  )
//           {
//             verticeIsBeingBuild = false;
//
//             if( verticeName.empty() || verticeType.empty() )
//             {
//               errorMessage = "Graph creation error - " +
//                 graphName + ":" +
//               boost::lexical_cast< std::string >( fileTokensIt->first );
//
//               return false;
//             }
//             else
//             {
//               GraphVertice* newVerticePtr = 0;
//
//               {
//                 std::unique_ptr< GraphVertice > newVerticeAuxPtr;
//
//                 try
//                 {
//                   newVerticeAuxPtr.reset( GraphVerticeFactory::make( verticeType ) );
//                 }
//                 catch(...)
//                 {
//                   newVerticeAuxPtr.reset();
//                 }
//
//                 newVerticeAuxPtr->setName( verticeName );
//                 newVerticeAuxPtr->enable2ndPassTasks( vertice2ndPassEnabled );
//                 newVerticeAuxPtr->enableCache( verticeCacheEnabled );
//
//                 if( !graph.addVertice( newVerticeAuxPtr.release(), emptyInputConnectionsData ) )
//                 {
//                   errorMessage = "Graph creation error - " +
//                     graphName + ":" +
//                     boost::lexical_cast< std::string >( fileTokensIt->first );
//
//                   return false;
//                 }
//
//                 newVerticePtr = graph.getVertice( verticeName );
//               }
//
//               if( newVerticePtr == 0 )
//               {
//                 errorMessage = "Graph creation error - " +
//                   graphName + ":" +
//                   boost::lexical_cast< std::string >( fileTokensIt->first );
//
//                 return false;
//               }
//               else
//               {
//                 // Vertice params
//
//                 for( std::size_t verticeParamsIdx = 0 ; verticeParamsIdx <
//                   verticeParams.size() ; ++verticeParamsIdx )
//                 {
//                   if( ! newVerticePtr->setParameter( verticeParams[ verticeParamsIdx ].first,
//                     verticeParams[ verticeParamsIdx ].second ) )
//                   {
//                     errorMessage = "Graph creation error - Invalid parameter : Graph:" +
//                       graphName + ":" +
//                       boost::lexical_cast< std::string >( fileTokensIt->first );
//
//                     return false;
//                   }
//                 }
//
//                 // Vertice resources
//
//                 for( std::size_t verticeResourcesIdx = 0 ; verticeResourcesIdx <
//                   verticeResources.size() ; ++verticeResourcesIdx )
//                 {
//                   if( ! newVerticePtr->setResource( verticeResources[ verticeResourcesIdx ].first,
//                     verticeResources[ verticeResourcesIdx ].second ) )
//                   {
//                     errorMessage = "Graph creation error - Invalid Resource (" +
//                       verticeResources[ verticeResourcesIdx ].first + ":" +
//                       verticeResources[ verticeResourcesIdx ].second + ") - Graph:" +
//                       graphName + ":" +
//                       boost::lexical_cast< std::string >( fileTokensIt->first );
//
//                     return false;
//                   }
//                 }
//
//                 // Vertice input connections
//
//                 for( std::size_t verticeConnectionsIdx = 0 ; verticeConnectionsIdx <
//                   verticeConnections.size() ; ++verticeConnectionsIdx )
//                 {
//                   GraphVertice* otherVerticePtr = graph.getVertice(
//                     verticeConnections[ verticeConnectionsIdx ].first );
//
//                   if( otherVerticePtr == 0 )
//                   {
//                     errorMessage = "Graph creation error - " +
//                       graphName + ":" +
//                       boost::lexical_cast< std::string >( fileTokensIt->first );
//
//                     return false;
//                   }
//                   else
//                   {
//                     if( ! newVerticePtr->connectInputVertice(
//                       *otherVerticePtr,
//                       verticeConnections[ verticeConnectionsIdx ].second ) )
//                     {
//                       errorMessage = "Graph creation error - " +
//                         graphName + ":" +
//                         boost::lexical_cast< std::string >( fileTokensIt->first );
//
//                       return false;
//                     }
//                   }
//                 }
//               }
//             }
//           }
//           else if(
//                     ( verticeIsBeingBuild )
//                     &&
//                     ( fileTokensIt->second.size() == 2 )
//                     &&
//                     ( fileTokensIt->second[ 0 ] == "VERTICE_NAME" )
//                  )
//           {
//             verticeName = fileTokensIt->second[ 1 ];
//           }
//           else if(
//                     ( verticeIsBeingBuild )
//                     &&
//                     ( fileTokensIt->second.size() == 2 )
//                     &&
//                     ( fileTokensIt->second[ 0 ] == "VERTICE_TYPE" )
//                  )
//           {
//             verticeType = fileTokensIt->second[ 1 ];
//           }
//           else if(
//                     ( verticeIsBeingBuild )
//                     &&
//                     ( fileTokensIt->second.size() == 3 )
//                     &&
//                     ( fileTokensIt->second[ 0 ] == "VERTICE_PARAM" )
//                  )
//           {
//             std::pair< std::string, std::string > param;
//             param.first = fileTokensIt->second[ 1 ];
//             param.second = fileTokensIt->second[ 2 ];
//
//             verticeParams.push_back( param );
//           }
//           else if(
//                     ( verticeIsBeingBuild )
//                     &&
//                     ( fileTokensIt->second.size() == 3 )
//                     &&
//                     ( fileTokensIt->second[ 0 ] == "VERTICE_RESOURCE" )
//                  )
//           {
//             std::pair< std::string, std::string > resource;
//             resource.first = fileTokensIt->second[ 1 ];
//             resource.second = fileTokensIt->second[ 2 ];
//
//             verticeResources.push_back( resource );
//           }
//           else if(
//                     ( verticeIsBeingBuild )
//                     &&
//                     ( fileTokensIt->second.size() == 3 )
//                     &&
//                     ( fileTokensIt->second[ 0 ] == "VERTICE_CONNECTION" )
//                  )
//           {
//             std::pair< std::string, std::string > param;
//             param.first = fileTokensIt->second[ 1 ];
//             param.second = fileTokensIt->second[ 2 ];
//
//             verticeConnections.push_back( param );
//           }
//           else if(
//                     ( verticeIsBeingBuild )
//                     &&
//                     ( fileTokensIt->second.size() == 2 )
//                     &&
//                     ( fileTokensIt->second[ 0 ] == "VERTICE_2ND_PASS" )
//                  )
//           {
//             if( te::common::Convert2UCase( fileTokensIt->second[ 1 ] ) == "YES" )
//             {
//               vertice2ndPassEnabled = true;
//             }
//             else if( te::common::Convert2UCase( fileTokensIt->second[ 1 ] ) == "NO" )
//             {
//               vertice2ndPassEnabled = false;
//             }
//             else
//             {
//               errorMessage = "Graph creation error - " +
//                 graphName + ":" +
//                 boost::lexical_cast< std::string >( fileTokensIt->first );
//
//               return false;
//             }
//           }
//           else if(
//                     ( verticeIsBeingBuild )
//                     &&
//                     ( fileTokensIt->second.size() == 2 )
//                     &&
//                     ( fileTokensIt->second[ 0 ] == "VERTICE_OUT_CACHE" )
//                  )
//           {
//             if( te::common::Convert2UCase( fileTokensIt->second[ 1 ] ) == "YES" )
//             {
//               verticeCacheEnabled = true;
//             }
//             else if( te::common::Convert2UCase( fileTokensIt->second[ 1 ] ) == "NO" )
//             {
//               verticeCacheEnabled = false;
//             }
//             else
//             {
//               errorMessage = "Graph creation error - " +
//                 graphName + ":" +
//                 boost::lexical_cast< std::string >( fileTokensIt->first );
//
//               return false;
//             }
//           }
//           else
//           {
//             errorMessage = "Graph creation error - " +
//               graphName + ":" +
//               boost::lexical_cast< std::string >( fileTokensIt->first );
//
//             return false;
//           }
//
//           ++fileTokensIt;
//         }
//
//         ++anaIt;
//       }
//     }

    return true;
  }

  bool TextFileProjectIO::save( const Project& proj, const std::string& fileName,
    std::string& errorMessage ) const
  {
    return false;
  }

  bool TextFileProjectIO::createContexts(
    const std::vector< std::tuple< std::string,
    unsigned int, std::vector< std::string > > >& fileTokens,
    std::map< std::string, Context >& contexts,
    std::string& errorMessage ) const
  {
    contexts.clear();

    const std::size_t fileTokensSize = fileTokens.size();
    bool objectCreationStarted = false;
    Context newContext;

    for( std::size_t fileTokensIdx = 0 ; fileTokensIdx < fileTokensSize ;
      ++fileTokensIdx )
    {
      if( objectCreationStarted )
      {
        if(
            ( std::get<2>( fileTokens[ fileTokensIdx ] ).size() == 1 )
            &&
            ( std::get<2>( fileTokens[ fileTokensIdx ] )[ 0 ] == "CONTEXT_END" )
          )
        {
          objectCreationStarted = false;

          if( newContext.getName().empty() )
          {
            errorMessage = "Error loading context file :" +
              std::get<0>( fileTokens[ fileTokensIdx ] ) +
              boost::lexical_cast< std::string >( std::get<1>(
              fileTokens[ fileTokensIdx ] ) );
            return false;
          }
          else
          {
            if( contexts.find( newContext.getName() ) == contexts.end() )
            {
              contexts[ newContext.getName() ] = newContext;
            }
            else
            {
              errorMessage = "Error loading context file :" +
                std::get<0>( fileTokens[ fileTokensIdx ] ) +
                boost::lexical_cast< std::string >( std::get<1>(
                fileTokens[ fileTokensIdx ] ) );
              return false;
            }
          }
        }
        else if(
                  ( std::get<2>( fileTokens[ fileTokensIdx ] ).size() == 2 )
                  &&
                  (
                    std::get<2>( fileTokens[ fileTokensIdx ] )[ 0 ]
                    ==
                    "CONTEXT_NAME"
                  )
                  &&
                 (
                   !std::get<2>( fileTokens[ fileTokensIdx ] )[ 1 ].empty()
                 )
                )
        {
          newContext.setName( std::get<2>( fileTokens[ fileTokensIdx ] )[ 1 ] );
        }
        else if(
                  ( std::get<2>( fileTokens[ fileTokensIdx ] ).size() == 3 )
                  &&
                  (
                    std::get<2>( fileTokens[ fileTokensIdx ] )[ 0 ]
                    ==
                    "RESOURCE_URI"
                  )
                  &&
                  (
                    !std::get<2>( fileTokens[ fileTokensIdx ] )[ 1 ].empty()
                  )
                  &&
                  (
                    !std::get<2>( fileTokens[ fileTokensIdx ] )[ 2 ].empty()
                  )
                )
        {
          const std::string& resourceAlias = std::get<2>( fileTokens[
            fileTokensIdx ] )[ 1 ];
          std::string resourceURIStr = std::get<2>( fileTokens[
            fileTokensIdx ] )[ 2 ];

          if(
              te::common::Convert2UCase( resourceURIStr.substr( 0, 7 ) )
              ==
              "FILE://"
            )
          {
            resourceURIStr = resourceURIStr.substr( 7, resourceURIStr.size() - 7 );
            const boost::filesystem::path contextFilePath(
              std::get<0>( fileTokens[ fileTokensIdx ] ) );

            boost::filesystem::path resoucePath;
            try
            {
              resoucePath =
                boost::filesystem::canonical(
                  boost::filesystem::system_complete(
                    contextFilePath.parent_path()
                    /
                    boost::filesystem::path( resourceURIStr )
                  )
                );
            }
            catch(...)
            {
              errorMessage = "Error loading context file :" +
                std::get<0>( fileTokens[ fileTokensIdx ] ) +
                boost::lexical_cast< std::string >( std::get<1>(
                fileTokens[ fileTokensIdx ] ) );
              return false;
            }

            te::core::URI uri( "file://" + resoucePath.string() );

            if( !newContext.addResource( resourceAlias, uri ) )
            {
              errorMessage = "Error loading context file :" +
                std::get<0>( fileTokens[ fileTokensIdx ] ) +
                boost::lexical_cast< std::string >( std::get<1>(
                fileTokens[ fileTokensIdx ] ) );
              return false;
            }
          }
          else
          {
            te::core::URI uri( resourceURIStr );

            if( !newContext.addResource( resourceAlias, uri ) )
            {
              errorMessage = "Error loading context file :" +
                std::get<0>( fileTokens[ fileTokensIdx ] ) +
                boost::lexical_cast< std::string >( std::get<1>(
                fileTokens[ fileTokensIdx ] ) );
              return false;
            }
          }
        }
        else
        {
          errorMessage = "Error loading context file :" +
            std::get<0>( fileTokens[ fileTokensIdx ] ) +
            boost::lexical_cast< std::string >( std::get<1>(
            fileTokens[ fileTokensIdx ] ) );
          return false;
        }
      }
      else
      {  // !objectCreationStarted
        if(
            ( !std::get<2>( fileTokens[ fileTokensIdx ] ).empty() )
            &&
            ( std::get<2>( fileTokens[ fileTokensIdx ] )[ 0 ] == "CONTEXT_START" )
          )
        {
          if(
              ( std::get<2>( fileTokens[ fileTokensIdx ] ).size() == 1 )
              &&
              (!objectCreationStarted)
            )
          {
            objectCreationStarted = true;
            newContext.reset();
          }
          else
          {
            errorMessage = "Error loading context file :" +
              std::get<0>( fileTokens[ fileTokensIdx ] ) +
              boost::lexical_cast< std::string >( std::get<1>(
              fileTokens[ fileTokensIdx ] ) );
            return false;
          }
        }
      }
    }

    return true;
  }

  bool TextFileProjectIO::createProject( const std::vector< std::tuple< std::string,
    unsigned int, std::vector< std::string > > >& fileTokens,
    Project& project,
    std::vector< std::tuple< std::string, std::string, std::string > >& analysisData,
    std::string& errorMessage ) const
  {
    project.reset();
    analysisData.clear();

    const std::size_t fileTokensSize = fileTokens.size();
    bool objectCreationStarted = false;

    for( std::size_t fileTokensIdx = 0 ; fileTokensIdx < fileTokensSize ;
      ++fileTokensIdx )
    {
      if( objectCreationStarted )
      {
        if(
            ( std::get<2>( fileTokens[ fileTokensIdx ] ).size() == 1 )
            &&
            ( std::get<2>( fileTokens[ fileTokensIdx ] )[ 0 ] == "PROJECT_END" )
          )
        {
          objectCreationStarted = false;

          if( project.getName().empty() )
          {
            errorMessage = "Error loading project file :" +
              std::get<0>( fileTokens[ fileTokensIdx ] ) +
              boost::lexical_cast< std::string >( std::get<1>(
              fileTokens[ fileTokensIdx ] ) );
            return false;
          }
          else
          {
            break;
          }
        }
        else if(
                  ( std::get<2>( fileTokens[ fileTokensIdx ] ).size() == 2 )
                  &&
                  (
                    std::get<2>( fileTokens[ fileTokensIdx ] )[ 0 ]
                    ==
                    "PROJECT_NAME"
                  )
                  &&
                  (
                    !std::get<2>( fileTokens[ fileTokensIdx ] )[ 1 ].empty()
                  )
                )
        {
          project.setName( std::get<2>( fileTokens[ fileTokensIdx ] )[ 1 ] );
        }
        else if(
                  ( std::get<2>( fileTokens[ fileTokensIdx ] ).size() == 4 )
                  &&
                  (
                    std::get<2>( fileTokens[ fileTokensIdx ] )[ 0 ]
                    ==
                    "ANALYSIS"
                  )
                  &&
                  ( !std::get<2>( fileTokens[ fileTokensIdx ] )[ 1 ].empty() )
                  &&
                  ( !std::get<2>( fileTokens[ fileTokensIdx ] )[ 2 ].empty() )
                  &&
                  ( !std::get<2>( fileTokens[ fileTokensIdx ] )[ 3 ].empty() )
                )
        {
          std::tuple< std::string, std::string, std::string > auxTuple;
          std::get<0>( auxTuple ) = std::get<2>( fileTokens[
            fileTokensIdx ] )[ 1 ];
          std::get<1>( auxTuple ) = std::get<2>( fileTokens[
            fileTokensIdx ] )[ 2 ];
          std::get<2>( auxTuple ) = std::get<2>( fileTokens[
            fileTokensIdx ] )[ 3 ];

          analysisData.push_back( auxTuple );
        }
        else
        {
          errorMessage = "Error loading project file :" +
            std::get<0>( fileTokens[ fileTokensIdx ] ) +
            boost::lexical_cast< std::string >( std::get<1>(
            fileTokens[ fileTokensIdx ] ) );
          return false;
        }
      }
      else
      {  // !objectCreationStarted
        if(
            ( !std::get<2>( fileTokens[ fileTokensIdx ] ).empty() )
            &&
            ( std::get<2>( fileTokens[ fileTokensIdx ] )[ 0 ] == "PROJECT_START" )
          )
        {
          if( std::get<2>( fileTokens[ fileTokensIdx ] ).size() == 1 )
          {
            objectCreationStarted = true;
          }
          else
          {
            errorMessage = "Error loading project file :" +
              std::get<0>( fileTokens[ fileTokensIdx ] ) +
              boost::lexical_cast< std::string >( std::get<1>(
              fileTokens[ fileTokensIdx ] ) );
            return false;
          }
        }
      }
    }

    return true;
  }

  bool TextFileProjectIO::createAnalysis( const std::vector< std::tuple< std::string,
    unsigned int, std::vector< std::string > > >& fileTokens,
    const std::map< std::string, Context >& contexts,
    const std::vector< std::tuple< std::string, std::string, std::string > >& analysisData,
    Project& project,
    std::string& errorMessage ) const
  {
    for( std::size_t analysisDataIdx = 0 ; analysisDataIdx < analysisData.size() ;
      ++analysisDataIdx )
    {
      const std::tuple< std::string, std::string, std::string >& anaData =
        analysisData[ analysisDataIdx ];

      Analysis newAnalysis;
      newAnalysis.setName( std::get< 0 >( anaData ) );

      const std::map< std::string, Context >::const_iterator contextsIt =
        contexts.find( std::get< 1 >( anaData ) );
      if( contextsIt == contexts.end() )
      {
        errorMessage = "Invalid context - " + std::get< 1 >( anaData );
        return false;
      }

      newAnalysis.setContext( contextsIt->second );

      if( ! project.addAnalysis( newAnalysis ) )
      {
        errorMessage = "Analysis creation error - Invalid analysis: " +
          newAnalysis.getName();

        return false;
      }

      if( ! createGraph( std::get< 0 >( anaData ), std::get< 2 >( anaData ),
        fileTokens, project, errorMessage ) )
      {
        project.removeAnalysis( std::get< 0 >( anaData ) );

        errorMessage = "Graph creation error - " + std::get< 2 >( anaData ) +
          " - " + errorMessage;
        return false;
      }
    }

    return true;
  }

  bool TextFileProjectIO::createGraph(
    const std::string& anaName, const std::string& graphName,
    const std::vector< std::tuple< std::string,
      unsigned int, std::vector< std::string > > >& fileTokens,
    Project& project,
    std::string& errorMessage ) const
  {
    if( graphName.empty() )
    {
      errorMessage = "Analysis creation error - Invalid graph name";
      return false;
    }

    Analysis* analysisPtr = project.getAnalysis( anaName );
    if( analysisPtr == 0 )
    {
      errorMessage = "Analysis creation error - Invalid analysis: " +
        anaName;
      return false;
    }

    // looking for the requested graph

    std::size_t fileTokensGraphStartIndex = fileTokens.size();

    {
      const std::size_t fileTokensSize = fileTokens.size();
      std::size_t currentGraphStartIndex = fileTokensSize;

      for( std::size_t fileTokensIdx = 0 ; fileTokensIdx < fileTokensSize ;
        ++fileTokensIdx )
      {
        if(
            ( std::get<2>( fileTokens[ fileTokensIdx ] ).size() == 1 )
            &&
            ( std::get<2>( fileTokens[ fileTokensIdx ] )[ 0 ] == "GRAPH_START" )
          )
        {
          currentGraphStartIndex = fileTokensIdx;
        }
        else if(
            ( std::get<2>( fileTokens[ fileTokensIdx ] ).size() == 2 )
            &&
            ( std::get<2>( fileTokens[ fileTokensIdx ] )[ 0 ] == "GRAPH_NAME" )
            &&
            ( std::get<2>( fileTokens[ fileTokensIdx ] )[ 1 ] == graphName )
            &&
            ( currentGraphStartIndex != fileTokensSize )
          )
        {
          fileTokensGraphStartIndex = currentGraphStartIndex;
          break;
        }
      }

      if( fileTokensGraphStartIndex == fileTokens.size() )
      {
        errorMessage = "Graph creation error - graph not found - " +
          graphName;
        return false;
      }
    }

    // creating the graph

    {
      {
        Graph newGraph;
        newGraph.setName( graphName );

        analysisPtr->setGraph( newGraph );
      }

      Graph& newGraph = analysisPtr->getGraph();
      std::string newGraphReadName;
      const std::size_t fileTokensSize = fileTokens.size();
      bool graphCreationStarted = false;
      bool vecticeCreationStarted = false;
      std::string verticeName;
      std::string verticeType;
      bool vertice2ndPassEnabled = false;
      bool verticeCacheEnabled = false;
      std::vector< std::pair< std::string, std::string > > verticeParams;
      std::vector< std::pair< std::string, std::string > > verticeResources;
      std::vector< std::pair< std::string, std::string > > verticeConnections;
      const std::vector< std::pair< GraphVerticeIDT, std::string > > emptyInputConnectionsData;

      for( std::size_t fileTokensIdx = fileTokensGraphStartIndex ;
          fileTokensIdx < fileTokensSize ; ++fileTokensIdx )
      {
        if( graphCreationStarted )
        {
          if( vecticeCreationStarted )
          {
            if(
                ( std::get<2>( fileTokens[ fileTokensIdx ] ).size() == 1 )
                &&
                (
                  std::get<2>( fileTokens[ fileTokensIdx ] )[ 0 ]
                  ==
                  "VERTICE_END"
                )
                &&
                ( !verticeName.empty() )
                &&
                ( !verticeType.empty() )
              )
            {
              vecticeCreationStarted = false;

              GraphVertice* newVerticePtr = 0;
              {
                std::unique_ptr< GraphVertice > newVerticeAuxPtr;

                try
                {
                  newVerticeAuxPtr.reset( GraphVerticeFactory::make(
                    verticeType ) );
                }
                catch(...)
                {
                  newVerticeAuxPtr.reset();
                }

                if( newVerticeAuxPtr.get() == 0 )
                {
                  errorMessage = "Graph creation error - Invalid vertice type -"
                    " Graph:" + graphName +
                    " Vertice:" + verticeName;
                  return false;
                }

                newVerticeAuxPtr->setName( verticeName );
                newVerticeAuxPtr->enable2ndPassTasks( vertice2ndPassEnabled );
                newVerticeAuxPtr->enableCache( verticeCacheEnabled );

                if( !newGraph.addVertice( newVerticeAuxPtr.release(),
                  emptyInputConnectionsData ) )
                {
                  errorMessage = "Graph creation error - Invalid vertice -"
                    " Graph:" + graphName +
                    " Vertice:" + verticeName;
                  return false;
                }

                newVerticePtr = newGraph.getVertice( verticeName );
              }

              // Vertice params

              for( std::size_t verticeParamsIdx = 0 ; verticeParamsIdx <
                verticeParams.size() ; ++verticeParamsIdx )
              {
                if( ! newVerticePtr->setParameter(
                  verticeParams[ verticeParamsIdx ].first,
                  verticeParams[ verticeParamsIdx ].second ) )
                {
                  errorMessage = "Graph creation error - Invalid parameter -"
                    " Graph:" + graphName +
                    " Vertice:" + verticeName +
                    " Parameter:" + verticeParams[ verticeParamsIdx ].first;
                  return false;
                }
              }

              // Vertice resources

              for( std::size_t verticeResourcesIdx = 0 ; verticeResourcesIdx <
                verticeResources.size() ; ++verticeResourcesIdx )
              {
                if( ! newVerticePtr->setResource(
                  verticeResources[ verticeResourcesIdx ].first,
                  verticeResources[ verticeResourcesIdx ].second ) )
                {
                  errorMessage = "Graph creation error - Invalid Resource -"
                    " Graph:" + graphName +
                    " Vertice:" + verticeName +
                    " Resource:" +
                    verticeResources[ verticeResourcesIdx ].first;
                  return false;
                }
              }

              // Vertice input connections

              for( std::size_t verticeConnectionsIdx = 0 ; verticeConnectionsIdx <
                verticeConnections.size() ; ++verticeConnectionsIdx )
              {
                GraphVertice* otherVerticePtr = newGraph.getVertice(
                  verticeConnections[ verticeConnectionsIdx ].first );

                if( otherVerticePtr == 0 )
                {
                  errorMessage = "Graph creation error -"
                    " Invalid input vertice"
                    " Graph:" + graphName +
                    " Vertice:" + verticeName +
                    " InputConnection:" +
                    verticeConnections[ verticeConnectionsIdx ].first;
                  return false;
                }
                else
                {
                  if( ! newVerticePtr->connectInputVertice(
                    *otherVerticePtr,
                    verticeConnections[ verticeConnectionsIdx ].second ) )
                  {
                    errorMessage = "Graph creation error -"
                      " Invalid input vertice"
                      " Graph:" + graphName +
                      " Vertice:" + verticeName +
                      " InputConnection:" +
                      verticeConnections[ verticeConnectionsIdx ].first;
                    return false;
                  }
                }
              }
            }
            else if(
                      ( std::get<2>( fileTokens[ fileTokensIdx ] ).size() == 2 )
                      &&
                      (
                        std::get<2>( fileTokens[ fileTokensIdx ] )[ 0 ]
                        ==
                        "VERTICE_NAME"
                      )
                      &&
                      (
                        !std::get<2>( fileTokens[ fileTokensIdx ] )[ 1 ].empty()
                      )
                    )
            {
              verticeName = std::get<2>( fileTokens[ fileTokensIdx ] )[ 1 ];
            }
            else if(
                      ( std::get<2>( fileTokens[ fileTokensIdx ] ).size() == 2 )
                      &&
                      (
                        std::get<2>( fileTokens[ fileTokensIdx ] )[ 0 ]
                        ==
                        "VERTICE_TYPE"
                      )
                      &&
                      (
                        !std::get<2>( fileTokens[ fileTokensIdx ] )[ 1 ].empty()
                      )
                    )
            {
              verticeType = std::get<2>( fileTokens[ fileTokensIdx ] )[ 1 ];
            }
            else if(
                      ( std::get<2>( fileTokens[ fileTokensIdx ] ).size() == 3 )
                      &&
                      (
                        std::get<2>( fileTokens[ fileTokensIdx ] )[ 0 ]
                        ==
                        "VERTICE_PARAM"
                      )
                      &&
                      (
                        !std::get<2>( fileTokens[ fileTokensIdx ] )[ 1 ].empty()
                      )
                    )
            {
              std::pair< std::string, std::string > param;
              param.first = std::get<2>( fileTokens[ fileTokensIdx ] )[ 1 ];
              param.second = std::get<2>( fileTokens[ fileTokensIdx ] )[ 2 ];

              verticeParams.push_back( param );
            }
            else if(
                      ( std::get<2>( fileTokens[ fileTokensIdx ] ).size() == 3 )
                      &&
                      (
                        std::get<2>( fileTokens[ fileTokensIdx ] )[ 0 ]
                        ==
                        "VERTICE_RESOURCE"
                      )
                      &&
                      (
                        !std::get<2>( fileTokens[ fileTokensIdx ] )[ 1 ].empty()
                      )
                      &&
                      (
                        !std::get<2>( fileTokens[ fileTokensIdx ] )[ 2 ].empty()
                      )
                    )
            {
              std::pair< std::string, std::string > resource;
              resource.first = std::get<2>( fileTokens[ fileTokensIdx ] )[ 1 ];
              resource.second = std::get<2>( fileTokens[ fileTokensIdx ] )[ 2 ];

              verticeResources.push_back( resource );
            }
            else if(
                      ( std::get<2>( fileTokens[ fileTokensIdx ] ).size() == 3 )
                      &&
                      (
                        std::get<2>( fileTokens[ fileTokensIdx ] )[ 0 ]
                        ==
                        "VERTICE_CONNECTION"
                      )
                      &&
                      (
                        !std::get<2>( fileTokens[ fileTokensIdx ] )[ 1 ].empty()
                      )
                      &&
                      (
                        !std::get<2>( fileTokens[ fileTokensIdx ] )[ 2 ].empty()
                      )
                    )
            {
              std::pair< std::string, std::string > param;
              param.first = std::get<2>( fileTokens[ fileTokensIdx ] )[ 1 ];
              param.second = std::get<2>( fileTokens[ fileTokensIdx ] )[ 2 ];

              verticeConnections.push_back( param );
            }
            else if(
                      ( std::get<2>( fileTokens[ fileTokensIdx ] ).size() == 2 )
                      &&
                      (
                        std::get<2>( fileTokens[ fileTokensIdx ] )[ 0 ]
                        ==
                        "VERTICE_2ND_PASS"
                      )
                      &&
                      (
                        !std::get<2>( fileTokens[ fileTokensIdx ] )[ 1 ].empty()
                      )
                    )
            {
              if( te::common::Convert2UCase( std::get<2>( fileTokens[
                fileTokensIdx ] )[ 1 ] ) == "YES" )
              {
                vertice2ndPassEnabled = true;
              }
              else
              {
                vertice2ndPassEnabled = false;
              }
            }
            else if(
                      ( std::get<2>( fileTokens[ fileTokensIdx ] ).size() == 2 )
                      &&
                      (
                        std::get<2>( fileTokens[ fileTokensIdx ] )[ 0 ]
                        ==
                        "VERTICE_OUT_CACHE"
                      )
                      &&
                      (
                        !std::get<2>( fileTokens[ fileTokensIdx ] )[ 1 ].empty()
                      )
                    )
            {
              if( te::common::Convert2UCase( std::get<2>( fileTokens[
                fileTokensIdx ] )[ 1 ] ) == "YES" )
              {
                verticeCacheEnabled = true;
              }
              else
              {
                verticeCacheEnabled = false;
              }
            }
            else
            {
              errorMessage = "Error creating graph :" +
                std::get<0>( fileTokens[ fileTokensIdx ] ) +
                boost::lexical_cast< std::string >( std::get<1>(
                fileTokens[ fileTokensIdx ] ) );
              return false;
            }
          }
          else
          {  // graphCreationStarted && !vecticeCreationStarted
            if(
                ( std::get<2>( fileTokens[ fileTokensIdx ] ).size() == 1 )
                &&
                ( std::get<2>( fileTokens[ fileTokensIdx ] )[ 0 ] == "GRAPH_END" )
                &&
                ( newGraphReadName == graphName )
              )
            {
              graphCreationStarted = false;
              return true;
            }
            else if(
                     ( std::get<2>( fileTokens[ fileTokensIdx ] ).size() == 2 )
                     &&
                     (
                       std::get<2>( fileTokens[ fileTokensIdx ] )[ 0 ]
                       ==
                       "GRAPH_NAME"
                     )
                     &&
                     (
                       !std::get<2>( fileTokens[ fileTokensIdx ] )[ 1 ].empty()
                     )
                   )
            {
              newGraphReadName = std::get<2>( fileTokens[ fileTokensIdx ] )[ 1 ];
            }
            else if(
                     ( std::get<2>( fileTokens[ fileTokensIdx ] ).size() == 1 )
                     &&
                     (
                       std::get<2>( fileTokens[ fileTokensIdx ] )[ 0 ]
                       ==
                       "VERTICE_START"
                     )
                   )
            {
              vecticeCreationStarted = true;
              verticeName.clear();
              verticeType.clear();
              vertice2ndPassEnabled = false;
              verticeCacheEnabled = false;
              verticeParams.clear();
              verticeResources.clear();
              verticeConnections.clear();
            }
            else
            {
              errorMessage = "Error creating graph :" +
                std::get<0>( fileTokens[ fileTokensIdx ] ) +
                boost::lexical_cast< std::string >( std::get<1>(
                fileTokens[ fileTokensIdx ] ) );
              return false;
            }
          }
        }
        else
        {  // !graphCreationStarted
          if( vecticeCreationStarted )
          {
            errorMessage = "Error creating graph - " +
              std::get<0>( fileTokens[ fileTokensIdx ] ) +
              boost::lexical_cast< std::string >( std::get<1>(
              fileTokens[ fileTokensIdx ] ) );
            return false;
          }
          else
          {  // !graphCreationStarted && !vecticeCreationStarted
            if(
                ( std::get<2>( fileTokens[ fileTokensIdx ] ).size() == 1 )
                &&
                ( std::get<2>( fileTokens[ fileTokensIdx ] )[ 0 ] == "GRAPH_START" )
              )
            {
              graphCreationStarted = true;
              newGraphReadName.clear();
            }
            else
            {
              errorMessage = "Error creating graph - " +
                std::get<0>( fileTokens[ fileTokensIdx ] ) +
                boost::lexical_cast< std::string >( std::get<1>(
                fileTokens[ fileTokensIdx ] ) );
              return false;
            }
          }
        }
      }
    }

    return true;
  }

  // -------------------------------------------------------------------------------------

  TextFileProjectIOFactory::TextFileProjectIOFactory()
  : ProjectIOFactory( "TXT" )
  {}

  TextFileProjectIOFactory::~TextFileProjectIOFactory()
  {}

  ProjectIO* TextFileProjectIOFactory::build()
  {
    return new TextFileProjectIO();
  }

}  // end namespace sa
}  // end namespace gdma
