/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/semanticanalysis/textFileProjectFactory.hpp
*/

#ifndef GEODMA_LIB_SEMANTICANALYSIS_TEXTFILEPROJECTIO_HPP_
#define GEODMA_LIB_SEMANTICANALYSIS_TEXTFILEPROJECTIO_HPP_

#include <string>
#include <vector>
#include <tuple>
#include <map>

// GeoDMA Includes
#include "projectIOFactory.hpp"
#include "projectIO.hpp"

namespace gdma {
namespace sa {

/*!
  \class ProjectIO

  \brief A basic concept of a Project IO.
*/
class GDMASEMANTICANALYSISEXPORT TextFileProjectIO : public ProjectIO {
    friend class TextFileProjectIOFactory;
 public:
    ~TextFileProjectIO();

    // overload
    bool load( const std::string& fileName, Project& proj,
      std::string& errorMessage ) const;

    // overload
    bool save( const Project& proj, const std::string& fileName,
      std::string& errorMessage ) const;

 protected:
    TextFileProjectIO();

 private:
    TextFileProjectIO( const TextFileProjectIO& );

    TextFileProjectIO& operator=( const TextFileProjectIO& );

    /*!
      \brief Create context objects from the given file tokens.
      \param fileTokens File tokens.
      \param contexts Created Contexts.
      \param errorMessage Return error message (if applicable).
      \return true if ok, false on errors.
     */
    bool createContexts( const std::vector< std::tuple< std::string,
      unsigned int, std::vector< std::string > > >& fileTokens,
      std::map< std::string, Context >& contexts,
      std::string& errorMessage ) const;

    /*!
      \brief Update a project object from the given file tokens.
      \param fileTokens File tokens.
      \param project Created Project.
      \param analysisData Project analysis data [ analysis name, context name, graph name ].
      \param errorMessage Return error message (if applicable).
      \return true if ok, false on errors.
      \note analysis objects will not be created
     */
    bool createProject( const std::vector< std::tuple< std::string,
      unsigned int, std::vector< std::string > > >& fileTokens,
      Project& project,
      std::vector< std::tuple< std::string, std::string, std::string > >& analysisData,
      std::string& errorMessage ) const;

    /*!
      \brief Create all analysis objets within the given project..
      \param fileTokens File tokens.
      \param contexts Created Contexts.
      \param analysisData Project analysis data [ analysis name, context name, graph name ].
      \param project Created Project.
      \param errorMessage Return error message (if applicable).
      \return true if ok, false on errors.
     */
    bool createAnalysis( const std::vector< std::tuple< std::string,
      unsigned int, std::vector< std::string > > >& fileTokens,
      const std::map< std::string, Context >& contexts,
      const std::vector< std::tuple< std::string, std::string, std::string > >& analysisData,
      Project& project,
      std::string& errorMessage ) const;

    /*!
      \brief Create a graph objet within the given project analysis.
      \param anaName Analysis name.
      \param graphName Graph name.
      \param fileTokens File tokens.
      \param project Created Project.
      \param errorMessage Return error message (if applicable).
      \return true if ok, false on errors.
     */
    bool createGraph( const std::string& anaName, const std::string& graphName,
      const std::vector< std::tuple< std::string,
        unsigned int, std::vector< std::string > > >& fileTokens,
      Project& project,
      std::string& errorMessage ) const;
};

/*!
  \class TextFileProjectIOFactory

  \brief A factory to deal with text file projects.
*/
class TextFileProjectIOFactory : public ProjectIOFactory {
 public:
    TextFileProjectIOFactory();
    ~TextFileProjectIOFactory();

 protected:
    // overload
    ProjectIO* build();
};

}  // end namespace sa
}  // end namespace gdma

#endif  // GEODMA_LIB_SEMANTICANALYSIS_TEXTFILEPROJECTIO_HPP_
