/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/taskprocessing/executer.hpp
  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

#ifndef GEODMA_LIB_TASKPROCESSING_EXECUTER_HPP_
#define GEODMA_LIB_TASKPROCESSING_EXECUTER_HPP_

// STL
#include <list>
#include <memory>
#include <mutex>
#include <thread>
#include <condition_variable>
#include <vector>

// GeoDMA Includes
#include "config.hpp"
#include "task.hpp"
#include "executerCallBack.hpp"

namespace gdma {
namespace tp {

/*!
  \class Executer
  \brief A task executer.
  \note Initial execution state: DISABLED
  \note Execution will finish if any thread returns TaskExecutionResult::FAILED or throws an excpetion.
*/
class GDMATASKPROCESSINGEXPORT Executer {
 public:
    /*!
      \enum Execution mode.
    */
    enum ExecMode {
      SINGLE,  //!< Just one task is executed on each call to the startExecution method.
      SEQUENTIAL,  //!< All tasks will be executed sequentially (no threads).
      CONCURRENT  //!< All tasks will be executed concurrently.
    };

    /*!
      \enum TaskStatus Task status.
    */
    enum TaskStatus {
      INACTIVE,  //!< Inactive task.
      EXECUTING,  //!< Task being executed.
      FINISHED  //!< Task alread was executed.
    };

    /*!
      \enum ExecuterStatus Executer status.
    */
    enum ExecuterStatus {
      INACTIVE_EXECUTER,  //!< Inactive executer.
      ACTIVE_EXECUTER  //!< Executer active.
    };

    /*! \brief Default constructor. */
    Executer();

    /*! \brief Default destructor. */
    ~Executer();

    /*!
      \brief Set de execution parameters.
      \param newExecMode New execution mode.
      \param threadsNumber The maximum number of threads to create when applicable (0-Automatic)
      \param takeTasksOwnership If true, the executer will take the ownership of added task objects.
      \return Returns true if OK, false on errors.
      \note This is a thread-safe method.
    */
    bool setExecutionParams(const ExecMode newExecMode,
      const unsigned int threadsNumber, const bool takeTasksOwnership);

    /*!
      \brief Returns the current execution mode.
      \return Returns the current execution mode.
      \note This is a thread-safe method.
    */
    ExecMode getExecutionMode() const;

    /*!
      \brief Returns the current executer status.
      \return Returns the current executer status.
      \note This is a thread-safe method.
    */
    ExecuterStatus getExecuterStatus() const;

    /*!
      \brief Start tasks execution.
    */
    void startExecution();

    /*!
      \brief Stop tasks execution.
    */
    void stopExecution();

    /*!
      \brief Hold the calling process and wait for the execution to end.
      \note This is a thread-safe method.
    */
    void waitExecution();

    /*!
      \brief Add a new task for execution.
      \param taskPtr Task pointer.
      \param taskIdTypePtr An optional pointer to a variable where the task Id will be stored.
      \return true if ok, false on errors.
      \note This is a thread-safe method.
    */
    bool addTask( Task* taskPtr, TaskIdType* taskIdTypePtr = 0);

    /*!
      \brief Remove a task.
      \param taskId The task Id.
      \param taskPtr An optional return pointer to the removed task (the returned object pointer must be acquired by the caller of this method).
      \return Returns true if OK, false on errors.
      \note This is a thread-safe method.
    */
    bool removeTask(TaskIdType taskId, Task** taskPtr = 0);

    /*!
      \brief Configures the executer to call the informed call back when necessary.
      \param newExecCallBack The call back object.
      \note The callback maybe called asynchronously within a thread.
      \note This is a thread-safe method.
    */
    void setCallBack(ExecuterCallBack& newExecCallBack);

    /*!
      \brief Returns true if a task with the given Id is present, false otherwise.
      \param taskId The task Id.
      \return Returns true if a task with the given Id is present, false otherwise.
      \note This is a thread-safe method.
    */
    bool isTaskPresent(const TaskIdType taskId) const;

    /*!
      \brief Get the required task status.
      \param taskId The task Id.
      \param taskStatus The returned task status.
      \return Returns true if a task with the given Id is present, false otherwise.
      \note This is a thread-safe method.
    */
    bool getTaskStatus(TaskIdType taskId, TaskStatus& taskStatus) const;

 protected:
    /*!
      \enum Executer status.
    */
    enum ThreadsExecCommand {
      THREAD_EXIT_EXECUTION,  //!< As soon as awaken, the thread must return (exit).
      THREAD_PROCESS_NEXT_AND_EXIT,  //!< As soon as awaken, the thread must process the next active task and return (exit).
      THREAD_PROCESS_ALL_AND_SLEEP,  //!< As soon as awaken, the thread must process all data and sleep again waiting for new data to process.
      THREAD_PROCESS_ALL_AND_EXIT  //!< As soon as awaken, the thread must process all data and return (exit).
    };

    /*!
      \class TasksContainerKeyType

      \brief Tasks container key type.
    */
    class TasksContainerKeyType {
     public:
      TasksContainerKeyType();
      TasksContainerKeyType(const TasksContainerKeyType& other);
      ~TasksContainerKeyType();
      TasksContainerKeyType& operator=(const TasksContainerKeyType& other);

      bool m_takeTasksOwnership;  //!< Take task objects ownership.
      TaskIdType m_id;  //!< Task ID.
      TaskStatus m_status;  //!< Task status.
      mutable Task* m_taskPtr;  //!< Task pointer (if not null and m_takeTasksOwnership is true it will be deleted when the destructor is called).
    };

    /*!
      \class ThreadsVector

      \brief A vector container for threads.
    */
    class ThreadsVector {
     public:
      ThreadsVector();

      ~ThreadsVector();

      /*!
        \brief Add a new active thread to the container (it will take its ownership ).
        \param threadPtr The thread pointer.
      */
      void add( std::thread* threadPtr );

      /*!
        \brief Returns true if this container is empty.
        \return Returns true if this container is empty.
      */
      bool empty();

      /*!
        \brief Returns the current container size.
        \return Returns the current container size.
      */
      std::size_t size();

      /*!
        \brief Wait for all threads to finish and clear.
      */
      void clear();

     private:
       std::vector< std::shared_ptr< std::thread > > m_threads;  //!< Threads vector;

       ThreadsVector(const ThreadsVector& other);

       ThreadsVector& operator=(const ThreadsVector& other);
    };

    /*!
      \typedef std::list< TasksContainerKeyType > TasksContainerType;
      \brief Tasks container type.
    */
    typedef std::list< TasksContainerKeyType > TasksContainerType;

    /*!
      \class ThreadParameters

      \brief Task thread parameters.
    */
    class ThreadParameters {
     public:
      ThreadsExecCommand const * m_threadsExecutionCommandPtr;  //!< A pointer to a threads execution control variable;
      ExecuterCallBack* m_execCallBackPtr;  //!< A pointer to a callback called when necessary.
      TasksContainerType* m_activeTasksPtr;  //!< Active tasks.
      TasksContainerType* m_finishedTasksPtr;  //!< Finished tasks.
      std::mutex* m_mutexPtr;  //!< Internal mutex to control threads general sync.
      std::mutex* m_awakeAllThreadsExecutionMutexPtr;  //!< A pointer to a internal mutex to control the nternal condition variable to awake all threads.
      std::condition_variable* m_awakeAllThreadsExecutionCondVarPtr;  //!< A pointer to a condition variable to awake all threads.
      unsigned int* m_activeThreadsNumberCounterPtr;  //!< A pointer to the active threads number counter.
      ExecuterStatus* m_executerStatusPtr;  //!< A pointer to the executer status variable.
    };

    /*!
      \class ExecuterStatusHandler

      \brief A class to control the current executer status based on the cound of active threads using the supplied mutex.
      \note When the counter points to a zero value the status will be changed to INACTIVE;
    */
    class ExecuterStatusHandler {
     public:
      /*!
        \brief Default constructor.
        \param activeThreadsNumberCounterPtr A pointer the the global active threads counter.
        \param executerStatusPtr A pointer the the global executer status.
        \param mutexPtr A pointer to the global mutex;
      */
       ExecuterStatusHandler( unsigned int* activeThreadsNumberCounterPtr,
         ExecuterStatus* executerStatusPtr,  std::mutex* mutexPtr );
       ~ExecuterStatusHandler();

     protected:
       unsigned int* m_activeThreadsNumberCounterPtr;
       ExecuterStatus* m_executerStatusPtr;
       std::mutex* m_mutexPtr;
    };

    ExecuterStatus m_executerStatus;  //!< Executer status;
    ExecMode m_executionMode;  //!< Execution mode.
    ThreadsExecCommand m_threadsExecutionCommand;  //!< Threads execution control variable;
    bool m_takeTasksOwnership;  //!< Take task objects ownership.
    unsigned int m_maxThreadsNumber;  //!< The number of threads to create.
    TaskIdType m_nextValidFreeTaskId;  //!< The next valid free task Id.
    mutable TasksContainerType m_activeTasks;  //!< Active tasks.
    mutable TasksContainerType m_finishedTasks;  //!< Finished tasks.
    mutable std::mutex m_mutex;  //!< Internal mutex to control external acess to this instance and threads sync.
    mutable std::mutex m_awakeAllThreadsExecutionMutex;  //!< Internal mutex to control the nternal condition variable to awake all threads.
    std::condition_variable m_awakeAllThreadsExecutionCondVar;  //!< Internal condition variable to awake all threads.
    ThreadsVector m_threadsVector;  //!< Threads vector;
    std::list< TaskIdType > m_freeTasksIds;  //!< Free avaliable tasks ids.
    unsigned int m_activeThreadsNumberCounter;  //!< Active threads number counter.
    ThreadParameters m_threadParameters;  //!< Global thread parameters;

    /*! \brief Default constructor. */
    Executer(const Executer&);

    /*!
      \brief The threads entry function.
      \param A pointer the the thread parameters.
    */
    static void threadFunction(ThreadParameters* threadParametersPtr);

    /*!
      \brief Returns an unique anused task Id.
      \param newId The new ID.
      \return true if ok, false on errors
      \note This is NOT a thread-safe method.
      \note The returned ID is valid only till the next call of removeTask().
    */
    bool getUniqueTaskId(TaskIdType& newId);

    /*!
     \brief Reset this instance to its initial state.
     \note THIS IS NOT A THREAD SAFE FUNCTION.
     */
    void reset();
};

}  // end namespace tp
}  // end namespace gdma

#endif  // GEODMA_LIB_TASKPROCESSING_EXECUTER_HPP_
