/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/taskprocessing/graphexecuter.hpp
  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

#ifndef GEODMA_LIB_TASKPROCESSING_GRAPHEXECUTER_HPP_
#define GEODMA_LIB_TASKPROCESSING_GRAPHEXECUTER_HPP_

// STL Includes
#include <map>
#include <vector>
#include <list>

// GeoDMA Includes
#include "config.hpp"
#include "executer.hpp"

namespace gdma {
namespace tp {

/*!
  \class GraphExecuter
  \brief A task executer witch takes into accound the depencence between tasks.
  \note Initial execution state: DISABLED
  \note Execution will finish if any thread returns TaskExecutionResult::FAILED or throws an excpetion.
*/
class GDMATASKPROCESSINGEXPORT GraphExecuter : protected ExecuterCallBack{
 public:
    /*!
      \enum ExecMode Execution mode.
    */
    enum ExecMode {
      SINGLE,  //!< Just one task is executed on each call to the startExecution method.
      SEQUENTIAL,  //!< All tasks will be executed sequentially (no threads).
      CONCURRENT  //!< All tasks will be executed concurrently.
    };

    /*!
      \enum TaskStatus Task status.
    */
    enum TaskStatus {
      INACTIVE,  //!< Inactive task.
      QUEUED,  //!< Task queued for execution.
      EXECUTING,  //!< Task being executed.
      FINISHED  //!< Task alread was executed.
    };

    /*!
      \enum Executer status.
    */
    enum ExecuterStatus {
      INACTIVE_EXECUTER,  //!< Inactive initial state.
      ACTIVE_EXECUTER  //!< Only one task is executed on startExecution() call.
    };

    /*! \brief Default constructor. */
    GraphExecuter();

    /*! \brief Default destructor. */
    ~GraphExecuter();

    /*!
      \brief Set de execution mode.
      \param newExecMode New execution mode.
      \param threadsNumber The maximum number of threads to create when applicable (0-Automatic)
      \param takeTasksOwnership If true, the executer will take the ownership of added task objects.
      \return Returns true if OK, false on errors.
      \note This is a thread-safe method.
    */
    bool setExecutionParams(const ExecMode newExecMode,
      const unsigned int threadsNumber, const bool takeTasksOwnership);

    /*!
      \brief Returns the current execution mode.
      \return Returns the current execution mode.
      \note This is a thread-safe method.
    */
    ExecMode getExecutionMode() const;

    /*!
      \brief Returns the current executer status.
      \return Returns the current executer status.
      \note This is a thread-safe method.
    */
    ExecuterStatus getExecuterStatus() const;

    /*!
      \brief Start tasks execution.
    */
    void startExecution();

    /*!
      \brief Stop tasks execution.
    */
    void stopExecution();

    /*!
      \brief Hold the calling process and wait for the execution to end.
    */
    void waitExecution();

    /*!
      \brief Add a new task for execution.
      \param taskPtr Task pointer (will be acquired by the executer)
      \param dependencies An Ids vector of tasks the new one depends on or an empty vector if there is no dependency.
      \param taskIdTypePtr An optional pointer to a variable where the task Id will be stored.
      \return Returns true if OK, false on errors.
      \note This is a thread-safe method.
    */
    bool addTask(Task* taskPtr, const std::vector< TaskIdType >& dependencies =
      std::vector< TaskIdType >(), TaskIdType* taskIdTypePtr = 0);

    /*!
      \brief Remove a task.
      \param taskId The task Id.
      \param taskPtr An optional return pointer to the removed task (the returned object pointer must be acquired by the caller of this method).
      \return Returns true if OK, false on errors.
      \note This is a thread-safe method.
    */
    bool removeTask(TaskIdType taskId, Task** taskPtr = 0);

    /*!
      \brief Configures the executer to call the informed call back when necessary.
      \param newExecCallBack The call back object.
      \note The callback maybe called asynchronously within a thread.
      \return Returns true if OK, false on errors.
      \note This is a thread-safe method.
    */
    void setCallBack(ExecuterCallBack& newExecCallBack);

    /*!
      \brief Get the required task status.
      \param taskId The task Id.
      \param taskStatus The returned task status.
      \return Returns true if a task with the given Id is present, false otherwise.
      \note This is a thread-safe method.
    */
    bool getTaskStatus(TaskIdType taskId, TaskStatus& taskStatus) const;

    /*!
      \brief Returns true if a task with the given Id is present, false otherwise.
      \param taskId The task Id.
      \return Returns true if a task with the given Id is present, false otherwise.
      \note This is a thread-safe method.
    */
    bool isTaskPresent(const TaskIdType taskId) const;

    /*!
      \brief Returns current stored tasks count.
      \return Returns current stored tasks count.
      \note This is a thread-safe method.
    */
    unsigned int getTasksCount() const;

 protected:
    /*!
      \class TasksContainerElementType

      \brief Tasks container element type.
    */
    class TasksContainerElementType {
     public:
        TasksContainerElementType();
        TasksContainerElementType(const TasksContainerElementType& other);
        ~TasksContainerElementType();
        TasksContainerElementType& operator=(const TasksContainerElementType& other);

        /*!
          \brief Pointer to an inactive not executed task (if not null and m_takeTasksOwnership is true it will be deleted when the destructor is called).
        */
        mutable Task* m_taskPtr;

        bool m_takeTasksOwnership;  //!< Take task objects ownership.
        TaskStatus m_status;  //!< Task status.
        TaskIdType m_executerTaskId;  //!< Internal executer task ID.

        /*!
          /brief Task dependencies (Ids of tasks it depends on).
        */
        std::list< TaskIdType > m_dependencies;
    };

    /*!
      \typedef std::map< TaskIdType, TasksContainerElementType > TasksContainerType
      \brief Tasks container type.
      \note The is an GraphExecuter ID.
    */
    typedef std::map< TaskIdType, TasksContainerElementType > TasksContainerType;

    ExecMode m_executionMode;  //!< Execution mode.
    bool m_takeTasksOwnership;  //!< Take task objects ownership.
    ExecuterCallBack* m_execCallBackPtr;  //!< A pointer to a callback called when necessary.
    TaskIdType m_nextValidFreeTaskId;  //!< The next valid free task Id.
    Executer m_executer;  //!< Executer instance.
    TasksContainerType m_tasksContainer;  //!< Tasks container.
    mutable std::mutex m_mutex;  //!< Internal mutex to control threads general sync.
    std::list< TaskIdType > m_freeTasksIds;  //!< Free avaliable tasks ids.

    /*! \brief Default constructor. */
    GraphExecuter(const GraphExecuter&);

    // Overload
    void taskStarted(const TaskIdType& internalExecuterTaskId);

    // Overload
    void taskExecuted(const TaskIdType& internalExecuterTaskId,
      const TaskExecutionResult::BasicResult& basicResut );

    /*!
      \brief Returns an unique anused task Id.
      \param newId The new ID.
      \return true if ok, false on errors
      \note This is NOT a thread-safe method.
      \note The returned ID is valid only till the next call of removeTask().
    */
    bool getUniqueTaskId(TaskIdType& newId);

    /*!
      \brief Add all inactive tasks with all dependencies solved into the executer
      \note This is NOT a thread-safe method.
    */
    void addITasks2Exec();

    /*!
     \brief Reset this instance to its initial state.
     \note THIS IS NOT A THREAD SAFE FUNCTION.
     */
    void reset();
};

}  // end namespace tp
}  // end namespace gdma

#endif  // GEODMA_LIB_TASKPROCESSING_GRAPHEXECUTER_HPP_
