/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/tests/unittest/classification/sampleSet_unittest.cpp

  \brief This file contains unit tests for the SampleSet class

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// TerraLib includes
#include <terralib/common/TerraLib.h>
#include <terralib/dataaccess.h>

// GTest includes
#include <gtest/gtest.h>

// Boost Includes
#include <boost/filesystem.hpp>

// STL Includes
#include <string>
#include <vector>

// GeoDMA Includes
#include "geodmalib/common/fileSystem.hpp"
#include "geodmalib/common/utils.hpp"
#include "geodmalib/datamanager/dataManager.hpp"
#include "geodmalib/classification/classDescription.hpp"
#include "geodmalib/classification/classDescriptionSet.hpp"

class ClassDescriptionSetTest : public ::testing::Test {
 protected:
    boost::filesystem::path m_tmpFolder;
    // std::unique_ptr<MockObserver> m_obs;
    // int m_obsId;
    std::string m_dsId;

    virtual void SetUp() {
        try {
//             TerraLib::getInstance().initialize();
//             gdma::common::loadTerraLibDrivers();

            m_tmpFolder = "./tmpTests";

            // m_obs.reset(new MockObserver());
            // m_obsId = gdma::common::MessageMediator::getInstance().addObserver(m_obs.get());

            boost::system::error_code error;
            if (boost::filesystem::exists(m_tmpFolder)) {
                if (!boost::filesystem::remove_all(m_tmpFolder, error)) {
                    std::string errorMsg = "Temporary folder can not be deleted: " +
                        error.message();
                    FAIL() << errorMsg;
                }
            } else {
                if (!boost::filesystem::create_directory(m_tmpFolder, error)) {
                    std::string errorMsg = "Temporary folder not created: " + error.message();
                    FAIL() << errorMsg;
                }
            }

            std::string inputFile = gdma::common::FindInGeoDMAPath(
                "data/tests/clipped_pols_tests-bigger.shp");
            std::string copiedFile = m_tmpFolder.string() + "/clipped_test.shp";

            if (!gdma::dm::diskVectorFileCopy(inputFile, copiedFile)) {
                FAIL() << "Could not create file copy!";
            }

            m_dsId = gdma::dm::openShapefile(copiedFile);
        }
        catch (te::common::Exception &e) {
            std::cerr << "ERROR: " << e.what() << std::endl;
        }
    }

    virtual void TearDown() {
//         gdma::common::unloadTerraLibDrivers();
//         TerraLib::getInstance().finalize();

        // gdma::common::MessageMediator::getInstance().removeObserver(m_obsId);

        boost::system::error_code error;
        if (boost::filesystem::exists(m_tmpFolder)) {
            if (!boost::filesystem::remove_all(m_tmpFolder, error)) {
                std::string errorMsg = "Temporary folder not deleted: " + error.message();
                FAIL() << errorMsg;
            }
        }
    }
};

TEST(ClassDescriptionSet, addOneClass) {
    te::da::DataSetTypePtr schema(new te::da::DataSetType("Teste"));
    gdma::cl::ClassDescription clazz("Teste", schema.get());
    clazz.setColor("#fffafa");

    gdma::cl::ClassDescriptionSet classSet;
    ASSERT_TRUE(classSet.addClass(clazz));

    EXPECT_EQ(1, classSet.getNumberOfClasses());
    EXPECT_STREQ("Teste", classSet.getClass("Teste")->getLabel().c_str());
    EXPECT_STREQ("#fffafa", classSet.getClass("Teste")->getColor().c_str());
}

TEST(ClassDescriptionSet, addTwoClasses) {
    te::da::DataSetTypePtr schema(new te::da::DataSetType("Teste"));
    gdma::cl::ClassDescription clazz("Teste", schema.get());
    clazz.setColor("#fffafa");

    gdma::cl::ClassDescription clazz2("Teste2", schema.get());

    gdma::cl::ClassDescriptionSet classSet;
    ASSERT_TRUE(classSet.addClass(clazz));
    ASSERT_TRUE(classSet.addClass(clazz2));

    EXPECT_EQ(2, classSet.getNumberOfClasses());
    EXPECT_STREQ("Teste", classSet.getClass("Teste")->getLabel().c_str());
    EXPECT_STREQ("#fffafa", classSet.getClass("Teste")->getColor().c_str());

    EXPECT_STREQ("Teste2", classSet.getClass("Teste2")->getLabel().c_str());
    EXPECT_STREQ("#ffffff", classSet.getClass("Teste2")->getColor().c_str());
}

TEST(ClassDescriptionSet, removeOneClass) {
    te::da::DataSetTypePtr schema(new te::da::DataSetType("Teste"));
    gdma::cl::ClassDescription clazz("Teste", schema.get());
    clazz.setColor("#fffafa");

    gdma::cl::ClassDescription clazz2("Teste2", schema.get());

    gdma::cl::ClassDescriptionSet classSet;
    ASSERT_TRUE(classSet.addClass(clazz));
    ASSERT_TRUE(classSet.addClass(clazz2));

    EXPECT_EQ(2, classSet.getNumberOfClasses());
    EXPECT_STREQ("Teste", classSet.getClass("Teste")->getLabel().c_str());
    EXPECT_STREQ("#fffafa", classSet.getClass("Teste")->getColor().c_str());

    EXPECT_STREQ("Teste2", classSet.getClass("Teste2")->getLabel().c_str());
    EXPECT_STREQ("#ffffff", classSet.getClass("Teste2")->getColor().c_str());

    ASSERT_TRUE(classSet.removeClass("Teste2"));
    EXPECT_EQ(1, classSet.getNumberOfClasses());
    EXPECT_EQ(0, classSet.getClass("Teste2"));
}

TEST(ClassDescriptionSet, changeOneClassLabel) {
    te::da::DataSetTypePtr schema(new te::da::DataSetType("Teste"));
    gdma::cl::ClassDescription clazz("Teste", schema.get());
    clazz.setColor("#fffafa");

    gdma::cl::ClassDescription clazz2("Teste2", schema.get());

    gdma::cl::ClassDescriptionSet classSet;
    ASSERT_TRUE(classSet.addClass(clazz));
    ASSERT_TRUE(classSet.addClass(clazz2));

    EXPECT_EQ(2, classSet.getNumberOfClasses());
    EXPECT_STREQ("Teste", classSet.getClass("Teste")->getLabel().c_str());
    EXPECT_STREQ("#fffafa", classSet.getClass("Teste")->getColor().c_str());

    EXPECT_STREQ("Teste2", classSet.getClass("Teste2")->getLabel().c_str());
    EXPECT_STREQ("#ffffff", classSet.getClass("Teste2")->getColor().c_str());

    ASSERT_TRUE(classSet.setClassLabel(clazz2, "Teste3"));
    EXPECT_EQ(2, classSet.getNumberOfClasses());
    EXPECT_EQ(0, classSet.getClass("Teste2"));
    EXPECT_EQ("Teste3", classSet.getClass("Teste3")->getLabel());
    EXPECT_EQ("#ffffff", classSet.getClass("Teste3")->getColor());
}

TEST(ClassDescriptionSet, createACopyByAssignment) {
    te::da::DataSetTypePtr schema(new te::da::DataSetType("Teste"));
    gdma::cl::ClassDescription clazz("Teste", schema.get());
    clazz.setColor("#fffafa");

    gdma::cl::ClassDescriptionSet classSet;
    ASSERT_TRUE(classSet.addClass(clazz));

    gdma::cl::ClassDescriptionSet newClassSet = classSet;

    EXPECT_EQ(1, newClassSet.getNumberOfClasses());
    EXPECT_STREQ("Teste", newClassSet.getClass("Teste")->getLabel().c_str());
    EXPECT_STREQ("#fffafa", newClassSet.getClass("Teste")->getColor().c_str());
}

TEST(ClassDescriptionSet, createACopy) {
    te::da::DataSetTypePtr schema(new te::da::DataSetType("Teste"));
    gdma::cl::ClassDescription clazz("Teste", schema.get());
    clazz.setColor("#fffafa");

    gdma::cl::ClassDescriptionSet classSet;
    ASSERT_TRUE(classSet.addClass(clazz));

    gdma::cl::ClassDescriptionSet newClassSet(classSet);

    EXPECT_EQ(1, newClassSet.getNumberOfClasses());
    EXPECT_STREQ("Teste", newClassSet.getClass("Teste")->getLabel().c_str());
    EXPECT_STREQ("#fffafa", newClassSet.getClass("Teste")->getColor().c_str());
}

TEST(ClassDescriptionSet, iterateOnTheStructure) {
    te::da::DataSetTypePtr schema(new te::da::DataSetType("Teste"));
    gdma::cl::ClassDescription clazz("Teste", schema.get());
    clazz.setColor("#fffafa");

    gdma::cl::ClassDescription clazz2("Teste2", schema.get());
    gdma::cl::ClassDescription clazz3("Teste3", schema.get());

    gdma::cl::ClassDescriptionSet classSet;
    ASSERT_TRUE(classSet.addClass(clazz));
    ASSERT_TRUE(classSet.addClass(clazz2));
    ASSERT_TRUE(classSet.addClass(clazz3));

    EXPECT_EQ(3, classSet.getNumberOfClasses());

    gdma::cl::ClassDescriptionMap::iterator it = classSet.begin();

    ASSERT_TRUE(it != classSet.end());

    EXPECT_STREQ("Teste", it->first.c_str());
    EXPECT_STREQ("Teste", it->second.getLabel().c_str());
    EXPECT_STREQ("#fffafa", it->second.getColor().c_str());

    it++;

    EXPECT_STREQ("Teste2", it->first.c_str());
    EXPECT_STREQ("Teste2", it->second.getLabel().c_str());
    EXPECT_STREQ("#ffffff", it->second.getColor().c_str());

    it++;

    EXPECT_STREQ("Teste3", it->first.c_str());
    EXPECT_STREQ("Teste3", it->second.getLabel().c_str());
    EXPECT_STREQ("#ffffff", it->second.getColor().c_str());

    it++;

    EXPECT_TRUE(it == classSet.end());
}

TEST_F(ClassDescriptionSetTest, saveSamplesInShapeFile) {
    te::da::DataSourcePtr dataSource = te::da::GetDataSource(m_dsId);
    te::da::DataSourceTransactorPtr transactor = dataSource->getTransactor();
    std::vector<std::string> dsNames = transactor->getDataSetNames();
    te::da::DataSetTypePtr schema = dataSource->getDataSetType(dsNames[0]);
    te::da::DataSetPtr dataset = transactor->getDataSet(dsNames[0]);

    gdma::cl::ClassDescription clazz("teste", schema.get());
    clazz.setColor("#ffffff");

    te::da::ObjectId id1;
    te::da::ObjectId id2;
    te::da::ObjectId id3;

    std::vector<std::string> pnames;
    te::da::GetOIDPropertyNames(schema.get(), pnames);

    dataset->moveFirst();
    id1 = *te::da::GenerateOID(dataset.get(), pnames);
    dataset->moveNext();
    dataset->moveNext();

    id2 = *te::da::GenerateOID(dataset.get(), pnames);
    dataset->moveNext();
    id3 = *te::da::GenerateOID(dataset.get(), pnames);

    clazz.addSample(id1.clone());
    clazz.addSample(id2.clone());
    clazz.addSample(id3.clone());

    gdma::cl::ClassDescriptionSet classSet;
    ASSERT_TRUE(classSet.addClass(clazz));
    EXPECT_EQ(3, classSet.getTotalNumberOfSamples());
    EXPECT_EQ(3, classSet.getNumberOfSamples("teste"));

    ASSERT_TRUE(classSet.saveSamplesInDataSource(m_dsId, "tclass"));

    dataSource->close();
    dataSource->open();
    transactor = dataSource->getTransactor();
    dsNames = transactor->getDataSetNames();
    schema = dataSource->getDataSetType(dsNames[0]);
    dataset = transactor->getDataSet(dsNames[0]);

    ASSERT_TRUE(dataset->moveFirst());
    EXPECT_STREQ("teste", dataset->getAsString("tclass").c_str());
    ASSERT_TRUE(dataset->moveNext());
    EXPECT_STREQ("", dataset->getAsString("tclass").c_str());
    ASSERT_TRUE(dataset->moveNext());
    EXPECT_STREQ("teste", dataset->getAsString("tclass").c_str());
    ASSERT_TRUE(dataset->moveNext());
    EXPECT_STREQ("teste", dataset->getAsString("tclass").c_str());
    ASSERT_TRUE(dataset->moveNext());
    EXPECT_STREQ("", dataset->getAsString("tclass").c_str());
    ASSERT_TRUE(dataset->moveNext());
    EXPECT_STREQ("", dataset->getAsString("tclass").c_str());
    ASSERT_FALSE(dataset->moveNext());
}

TEST_F(ClassDescriptionSetTest, saveSamplesInShapeFileEmptyClass) {
    te::da::DataSourcePtr dataSource;
    ASSERT_NO_THROW({ dataSource = te::da::GetDataSource(m_dsId); });
    te::da::DataSourceTransactorPtr transactor = dataSource->getTransactor();
    std::vector<std::string> dsNames = transactor->getDataSetNames();
    te::da::DataSetTypePtr schema = dataSource->getDataSetType(dsNames[0]);
    te::da::DataSetPtr dataset = transactor->getDataSet(dsNames[0]);

    gdma::cl::ClassDescription clazz("teste", schema.get());
    clazz.setColor("#ffffff");
    gdma::cl::ClassDescription clazz2("teste2", schema.get());
    clazz2.setColor("#fffffa");

    te::da::ObjectId id1;
    te::da::ObjectId id2;
    te::da::ObjectId id3;

    std::vector<std::string> pnames;
    te::da::GetOIDPropertyNames(schema.get(), pnames);

    dataset->moveFirst();
    id1 = *te::da::GenerateOID(dataset.get(), pnames);
    dataset->moveNext();
    dataset->moveNext();

    id2 = *te::da::GenerateOID(dataset.get(), pnames);
    dataset->moveNext();
    id3 = *te::da::GenerateOID(dataset.get(), pnames);

    clazz.addSample(id1.clone());
    clazz.addSample(id2.clone());
    clazz.addSample(id3.clone());

    gdma::cl::ClassDescriptionSet classSet;
    ASSERT_TRUE(classSet.addClass(clazz));
    ASSERT_TRUE(classSet.addClass(clazz2));
    EXPECT_EQ(3, classSet.getTotalNumberOfSamples());
    EXPECT_EQ(3, classSet.getNumberOfSamples("teste"));

    ASSERT_TRUE(classSet.saveSamplesInDataSource(m_dsId, "tclass"));

    dataSource->close();
    dataSource->open();
    transactor = dataSource->getTransactor();
    dsNames = transactor->getDataSetNames();
    dataset = transactor->getDataSet(dsNames[0]);

    ASSERT_TRUE(dataset->moveFirst());
    EXPECT_STREQ("teste", dataset->getAsString("tclass").c_str());
    ASSERT_TRUE(dataset->moveNext());
    EXPECT_STREQ("", dataset->getAsString("tclass").c_str());
    ASSERT_TRUE(dataset->moveNext());
    EXPECT_STREQ("teste", dataset->getAsString("tclass").c_str());
    ASSERT_TRUE(dataset->moveNext());
    EXPECT_STREQ("teste", dataset->getAsString("tclass").c_str());
    ASSERT_TRUE(dataset->moveNext());
    EXPECT_STREQ("", dataset->getAsString("tclass").c_str());
    ASSERT_TRUE(dataset->moveNext());
    EXPECT_STREQ("", dataset->getAsString("tclass").c_str());
    ASSERT_FALSE(dataset->moveNext());
}

TEST_F(ClassDescriptionSetTest, loadSamplesFromDataSource) {
    std::string inputFile = gdma::common::FindInGeoDMAPath(
        "data/tests/clipped_pols_tests-Samples.shp");

    m_dsId = gdma::dm::openShapefile(inputFile);

    gdma::cl::ClassDescriptionSet classSet;
    ASSERT_TRUE(classSet.loadSamplesFromDataSource(m_dsId, "trainClass"));

    EXPECT_EQ(2, classSet.getNumberOfClasses());
    EXPECT_EQ(5, classSet.getTotalNumberOfSamples());

    gdma::cl::ClassDescriptionMap::iterator it = classSet.begin();

    EXPECT_EQ("teste", it->first);
    EXPECT_EQ("teste", it->second.getLabel());
    EXPECT_EQ(2, it->second.getNumberOfSamples());

    EXPECT_TRUE(it->second.containsId("2"));
    EXPECT_TRUE(it->second.containsId("3"));

    it++;
    EXPECT_EQ("teste2", it->first);
    EXPECT_EQ("teste2", it->second.getLabel());
    EXPECT_EQ(3, it->second.getNumberOfSamples());

    EXPECT_TRUE(it->second.containsId("0"));
    EXPECT_TRUE(it->second.containsId("1"));
    EXPECT_TRUE(it->second.containsId("5"));

    it++;
    EXPECT_TRUE(it == classSet.end());
}
