/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/classification/classDescription.hpp

  \brief This file contains the ClassDescription class definition.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// GTest includes
#include <gtest/gtest.h>

#include <terralib/dataaccess/dataset/ObjectId.h>
#include <terralib/datatype/SimpleData.h>

// STL Includes
#include <map>
#include <set>
#include <string>

// GeoDMA Includes
#include "geodmalib/classification/classDescription.hpp"

TEST(ClassDescriptionTest, createOneClassDefaultColor) {
    te::da::DataSetTypePtr schema(new te::da::DataSetType("Teste"));
    gdma::cl::ClassDescription clazz("Teste", schema.get());

    EXPECT_STREQ("Teste", clazz.getLabel().c_str());
    EXPECT_STREQ("#ffffff", clazz.getColor().c_str());
    EXPECT_EQ(0, clazz.getMetadataSize());
}

TEST(ClassDescriptionTest, createOneClassDifferentColor) {
    te::da::DataSetTypePtr schema(new te::da::DataSetType("Teste"));
    gdma::cl::ClassDescription clazz("Teste", schema.get());
    clazz.setColor("#ffff00");

    EXPECT_STREQ("Teste", clazz.getLabel().c_str());
    EXPECT_STREQ("#ffff00", clazz.getColor().c_str());
    EXPECT_EQ(0, clazz.getMetadataSize());
}

TEST(ClassDescriptionTest, insertMetaData) {
    te::da::DataSetTypePtr schema(new te::da::DataSetType("Teste"));
    gdma::cl::ClassDescription clazz("Teste", schema.get(), "#ffff00");

    clazz.addMetadata("Metadata1", "Description of Metadata 1");

    EXPECT_STREQ("Teste", clazz.getLabel().c_str());
    EXPECT_STREQ("#ffff00", clazz.getColor().c_str());
    EXPECT_EQ(1, clazz.getMetadataSize());
    EXPECT_STREQ("Description of Metadata 1", clazz.getMetadataValue("Metadata1").c_str());
}

TEST(ClassDescriptionTest, insertMetaDataExistingId) {
    te::da::DataSetTypePtr schema(new te::da::DataSetType("Teste"));
    gdma::cl::ClassDescription clazz("Teste", schema.get(), "#ffff00");

    clazz.addMetadata("Metadata1", "Description of Metadata 1");

    EXPECT_STREQ("Teste", clazz.getLabel().c_str());
    EXPECT_STREQ("#ffff00", clazz.getColor().c_str());
    EXPECT_EQ(1, clazz.getMetadataSize());
    EXPECT_STREQ("Description of Metadata 1", clazz.getMetadataValue("Metadata1").c_str());

    clazz.addMetadata("Metadata1", "New Description of Metadata 1");
    EXPECT_EQ(1, clazz.getMetadataSize());
    EXPECT_STREQ("New Description of Metadata 1", clazz.getMetadataValue("Metadata1").c_str());
}

TEST(ClassDescriptionTest, removeExistingMetaData) {
    te::da::DataSetTypePtr schema(new te::da::DataSetType("Teste"));
    gdma::cl::ClassDescription clazz("Teste", schema.get(), "#ffff00");

    clazz.addMetadata("Metadata1", "Description of Metadata 1");
    clazz.addMetadata("Metadata2", "Description of Metadata 2");

    EXPECT_STREQ("Teste", clazz.getLabel().c_str());
    EXPECT_STREQ("#ffff00", clazz.getColor().c_str());
    EXPECT_EQ(2, clazz.getMetadataSize());
    EXPECT_STREQ("Description of Metadata 1", clazz.getMetadataValue("Metadata1").c_str());
    EXPECT_STREQ("Description of Metadata 2", clazz.getMetadataValue("Metadata2").c_str());

    bool removed = clazz.removeMetadata("Metadata2");

    EXPECT_TRUE(removed);
    EXPECT_STREQ("Teste", clazz.getLabel().c_str());
    EXPECT_STREQ("#ffff00", clazz.getColor().c_str());
    EXPECT_EQ(1, clazz.getMetadataSize());
    EXPECT_STREQ("Description of Metadata 1", clazz.getMetadataValue("Metadata1").c_str());
    EXPECT_STREQ("", clazz.getMetadataValue("Metadata2").c_str());
}

TEST(ClassDescriptionTest, removeNonexistentMetaData) {
    te::da::DataSetTypePtr schema(new te::da::DataSetType("Teste"));
    gdma::cl::ClassDescription clazz("Teste", schema.get(), "#ffff00");

    clazz.addMetadata("Metadata1", "Description of Metadata 1");

    EXPECT_STREQ("Teste", clazz.getLabel().c_str());
    EXPECT_STREQ("#ffff00", clazz.getColor().c_str());
    EXPECT_EQ(1, clazz.getMetadataSize());
    EXPECT_STREQ("Description of Metadata 1", clazz.getMetadataValue("Metadata1").c_str());

    bool removed = clazz.removeMetadata("Metadata2");

    EXPECT_FALSE(removed);
    EXPECT_STREQ("Teste", clazz.getLabel().c_str());
    EXPECT_STREQ("#ffff00", clazz.getColor().c_str());
    EXPECT_EQ(1, clazz.getMetadataSize());
    EXPECT_STREQ("Description of Metadata 1", clazz.getMetadataValue("Metadata1").c_str());
    EXPECT_STREQ("", clazz.getMetadataValue("Metadata2").c_str());
}

TEST(ClassDescriptionTest, retrieveAllMetadataAvailable) {
    te::da::DataSetTypePtr schema(new te::da::DataSetType("Teste"));
    gdma::cl::ClassDescription clazz("Teste", schema.get(), "#ffff00");

    clazz.addMetadata("Metadata1", "Description of Metadata 1");
    clazz.addMetadata("Metadata2", "Description of Metadata 2");

    EXPECT_STREQ("Teste", clazz.getLabel().c_str());
    EXPECT_STREQ("#ffff00", clazz.getColor().c_str());

    std::map<std::string, std::string> metadata = clazz.getAllMetadata();
    EXPECT_EQ(2, metadata.size());
    EXPECT_STREQ("Description of Metadata 1", metadata["Metadata1"].c_str());
    EXPECT_STREQ("Description of Metadata 2", metadata["Metadata2"].c_str());
}

TEST(ClassDescriptionTest, insertSample) {
    te::da::DataSetTypePtr schema(new te::da::DataSetType("Teste"));
    gdma::cl::ClassDescription clazz("Teste", schema.get(), "#ffff00");

    te::da::ObjectId id;
    id.addValue(new te::dt::SimpleData<std::string, te::dt::STRING_TYPE>("25"));

    clazz.addSample(&id);

    EXPECT_STREQ("Teste", clazz.getLabel().c_str());
    EXPECT_STREQ("#ffff00", clazz.getColor().c_str());
    EXPECT_TRUE(clazz.containsId(&id));
    EXPECT_EQ(1, clazz.getNumberOfSamples());
}

TEST(ClassDescriptionTest, removeSample) {
    te::da::DataSetTypePtr schema(new te::da::DataSetType("Teste"));
    gdma::cl::ClassDescription clazz("Teste", schema.get(), "#ffff00");

    te::da::ObjectId id;
    id.addValue(new te::dt::SimpleData<std::string, te::dt::STRING_TYPE>("25"));
    te::da::ObjectId id2;
    id2.addValue(new te::dt::SimpleData<std::string, te::dt::STRING_TYPE>("27"));

    clazz.addSample(&id);
    clazz.addSample(&id2);

    EXPECT_STREQ("Teste", clazz.getLabel().c_str());
    EXPECT_STREQ("#ffff00", clazz.getColor().c_str());
    EXPECT_TRUE(clazz.containsId(&id));
    EXPECT_TRUE(clazz.containsId(&id2));
    EXPECT_EQ(2, clazz.getNumberOfSamples());

    clazz.removeSample(&id2);
    EXPECT_TRUE(clazz.containsId(&id));
    EXPECT_FALSE(clazz.containsId(&id2));
    EXPECT_EQ(1, clazz.getNumberOfSamples());
}

TEST(ClassDescriptionTest, removeSampleByStringId) {
    te::da::DataSetTypePtr schema(new te::da::DataSetType("Teste"));
    gdma::cl::ClassDescription clazz("Teste", schema.get(), "#ffff00");

    te::da::ObjectId id;
    id.addValue(new te::dt::SimpleData<std::string, te::dt::STRING_TYPE>("25"));
    te::da::ObjectId id2;
    id2.addValue(new te::dt::SimpleData<std::string, te::dt::STRING_TYPE>("27"));

    clazz.addSample(&id);
    clazz.addSample(&id2);

    EXPECT_STREQ("Teste", clazz.getLabel().c_str());
    EXPECT_STREQ("#ffff00", clazz.getColor().c_str());
    EXPECT_TRUE(clazz.containsId("25"));
    EXPECT_TRUE(clazz.containsId("27"));
    EXPECT_EQ(2, clazz.getNumberOfSamples());

    clazz.removeSample("27");
    EXPECT_TRUE(clazz.containsId("25"));
    EXPECT_FALSE(clazz.containsId("27"));
    EXPECT_EQ(1, clazz.getNumberOfSamples());
}


TEST(ClassDescriptionTest, addSampleDuplicatedId) {
    te::da::DataSetTypePtr schema(new te::da::DataSetType("Teste"));
    gdma::cl::ClassDescription clazz("Teste", schema.get(), "#ffff00");

    te::da::ObjectId id;
    id.addValue(new te::dt::SimpleData<std::string, te::dt::STRING_TYPE>("25"));

    clazz.addSample(&id);

    EXPECT_STREQ("Teste", clazz.getLabel().c_str());
    EXPECT_STREQ("#ffff00", clazz.getColor().c_str());
    EXPECT_TRUE(clazz.containsId(&id));
    EXPECT_EQ(1, clazz.getNumberOfSamples());

    te::da::ObjectId id2;
    id2.addValue(new te::dt::SimpleData<std::string, te::dt::STRING_TYPE>("25"));
    clazz.addSample(&id2);

    EXPECT_TRUE(clazz.containsId(&id));
    EXPECT_TRUE(clazz.containsId(&id2));
    EXPECT_EQ(1, clazz.getNumberOfSamples());
}

TEST(ClassDescriptionTest, interactWithTheObjectIdSet) {
    te::da::DataSetTypePtr schema(new te::da::DataSetType("Teste"));
    gdma::cl::ClassDescription clazz("Teste", schema.get(), "#ffff00");

    te::da::ObjectId id;
    id.addValue(new te::dt::SimpleData<std::string, te::dt::STRING_TYPE>("25"));
    clazz.addSample(&id);

    te::da::ObjectId id2;
    id2.addValue(new te::dt::SimpleData<std::string, te::dt::STRING_TYPE>("26"));
    clazz.addSample(&id2);

    const te::da::ObjectIdSet* classSet = clazz.getSampleSet();

    EXPECT_TRUE(classSet->contains(&id));
    EXPECT_TRUE(classSet->contains(&id2));
    EXPECT_EQ(2, classSet->size());

    std::set<te::da::ObjectId*>::iterator it = classSet->begin();
    EXPECT_TRUE(it != classSet->end());

    int count = 0;
    while (it != classSet->end()) {
        count++;
        it++;
    }

    EXPECT_EQ(2, count);
}
