/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/unittest/classification/decision_trees/c50DecisionTree_unittest.cpp

  \brief This file contains the unit tests for the C5.0 algorithm

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// GTest includes
#include <gtest/gtest.h>

// TerraLib includes
#include <terralib/common/TerraLib.h>
#include <terralib/dataaccess.h>

// Boost Includes
#include <boost/filesystem.hpp>

// STL Includes
#include <string>
#include <vector>

// GeoDMA Includes
#include "geodmalib/classification/decision_trees/c50DecisionTree.hpp"
#include "geodmalib/classification/decision_trees/c50Parameters.hpp"
#include "geodmalib/common/utils.hpp"
#include "geodmalib/common/fileSystem.hpp"
#include "geodmalib/datamanager/dataManager.hpp"

class C50DecisionTreeTest : public ::testing::Test {
 protected:
    boost::filesystem::path m_tmpFolder;

    std::string m_pathShp = gdma::common::FindInGeoDMAPath(
        "data/tests/clipped_pols_tests-Samples_Feat.shp");
    te::da::DataSourcePtr m_dataSource;
    te::da::DataSetPtr m_inputDataSet;
    te::da::DataSetTypePtr m_inputDSType;

    virtual void SetUp() {
        try {
//             TerraLib::getInstance().initialize();
//             gdma::common::loadTerraLibDrivers();

            m_tmpFolder = "./tmpTests";

            te::core::URI dataSourceUri( "file://" + m_pathShp );
            m_dataSource = te::da::DataSourceFactory::make( "OGR", dataSourceUri );
            if( m_dataSource.get() == 0 )
            {
              std::string errorMsg = "Datasource open error";
              FAIL() << errorMsg;
            }
            m_dataSource->open();
            if( ! m_dataSource->isOpened() )
            {
              std::string errorMsg = "Datasource open error";
              FAIL() << errorMsg;
            }

            std::string dataSetName = m_dataSource->getDataSetNames()[0];
            m_inputDataSet = m_dataSource->getDataSet(dataSetName);
            m_inputDSType = m_dataSource->getDataSetType(dataSetName);

            boost::system::error_code error;
            if (boost::filesystem::exists(m_tmpFolder)) {
                if (!boost::filesystem::remove_all(m_tmpFolder, error)) {
                    std::string errorMsg = "Temporary folder can not be deleted: " +
                        error.message();
                    FAIL() << errorMsg;
                }
            } else {
                if (!boost::filesystem::create_directory(m_tmpFolder, error)) {
                    std::string errorMsg = "Temporary folder not created: " + error.message();
                    FAIL() << errorMsg;
                }
            }
        } catch (te::common::Exception &e) {
            std::cerr << "ERROR: " << e.what() << std::endl;
        }
    }

    virtual void TearDown() {
        m_inputDSType.reset();
        m_inputDataSet.reset();
        m_dataSource->close();
        m_dataSource.reset();

//         gdma::common::unloadTerraLibDrivers();
//         TerraLib::getInstance().finalize();

        boost::system::error_code error;
        if (boost::filesystem::exists(m_tmpFolder)) {
            if (!boost::filesystem::remove_all(m_tmpFolder, error)) {
                std::string errorMsg = "Temporary folder not deleted: " + error.message();
                FAIL() << errorMsg;
            }
        }
    }
};

TEST_F(C50DecisionTreeTest, parseNamesStream) {
    gdma::cl::C50DecisionTree tree;

    std::string path;
    ASSERT_NO_THROW({
        path = gdma::common::FindInGeoDMAPath("/data/tests/decision_trees_tests/test.names");
    });

    ASSERT_TRUE(tree.loadExternalFiles(path));
}

TEST_F(C50DecisionTreeTest, newTest) {
    std::vector<std::string> features;
    features.push_back("DISSE_B0");
    features.push_back("ENERGSE_B0");
    features.push_back("ENTRSE_B0");
    features.push_back("HOMOGSE_B0");
    features.push_back("DISSE_B1");
    features.push_back("ENERGSE_B1");
    gdma::cl::C50Parameters parameters;
    gdma::cl::C50DecisionTree tree(m_inputDataSet, m_inputDSType, features, parameters);

    tree.train();
    tree.predict();
}
