/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/classification/vectorDatasetInputAdaptor_unittest.hpp

  \brief This file contains the ClassDescription class definition.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// GTest includes
#include <gtest/gtest.h>

// TerraLib Includes
#include <terralib/common/TerraLib.h>
#include <terralib/dataaccess/dataset/DataSet.h>
#include <terralib/dataaccess/dataset/DataSetType.h>
#include <terralib/dataaccess/utils/Utils.h>
#include <terralib/memory/DataSet.h>

// STL Includes
#include <string>

// GeoDMA includes
#include "geodmalib/classification/vectorDatasetInputAdaptor.hpp"
#include "geodmalib/common/fileSystem.hpp"
#include "geodmalib/common/utils.hpp"
#include "geodmalib/datamanager/dataManager.hpp"

class VectorDatasetInputAdaptorTest : public ::testing::Test {
 protected:
    std::string m_pathShp = gdma::common::FindInGeoDMAPath(
        "data/tests/clipped_pols_tests_1.shp");
    te::da::DataSourcePtr m_dataSource;
    te::da::DataSetPtr m_inputDataSet;

    virtual void SetUp() {
        try {
//             TerraLib::getInstance().initialize();
//             gdma::common::loadTerraLibDrivers();

            std::string dsId = gdma::dm::openShapefile(m_pathShp);
            m_dataSource = te::da::GetDataSource(dsId, true);

            std::string dataSetName = m_dataSource->getDataSetNames()[0];
            m_inputDataSet = m_dataSource->getDataSet(dataSetName);
        }
        catch (te::common::Exception &e) {
            std::cerr << "ERROR: " << e.what() << std::endl;
        }
    }

    virtual void TearDown() {
        m_dataSource->close();
        m_dataSource.reset();
        m_inputDataSet.reset();
//         gdma::common::unloadTerraLibDrivers();
//         TerraLib::getInstance().finalize();
    }
};

TEST_F(VectorDatasetInputAdaptorTest, createAdaptorWithEmptyDataset) {
    te::da::DataSetTypePtr dataSetTypePtr(new te::da::DataSetType("teste"));
    te::mem::DataSetPtr memDataSetPtr(new te::mem::DataSet(dataSetTypePtr.get()));

    try {
        gdma::cl::VectorDatasetInputAdaptor<double> adaptor(memDataSetPtr);
        FAIL() << "Expected te::common::Exception Input Data Set is Empty!";
    } catch(te::common::Exception& e) {
        EXPECT_STREQ(e.what(), "Input Data Set is Empty!");
    } catch(...) {
        FAIL() << "Expected te::common::Exception";
    }
}

TEST_F(VectorDatasetInputAdaptorTest, createAdaptor) {
    ASSERT_NO_THROW({
        gdma::cl::VectorDatasetInputAdaptor<double> adaptor(m_inputDataSet);
        EXPECT_EQ(1, adaptor.getElementsCount());
        EXPECT_EQ(3, adaptor.getFeaturesCount());
    });
}

TEST_F(VectorDatasetInputAdaptorTest, getFeatureInexistentElement) {
    gdma::cl::VectorDatasetInputAdaptor<double> adaptor(m_inputDataSet);

    try {
        double featureValue;
        adaptor.getFeature(5, 0, featureValue);
        FAIL() << "Expected te::common::Exception Element index \"5\" does not exist in Data Set!";
    } catch (te::common::Exception& e) {
        EXPECT_STREQ(e.what(), "Element index \"5\" does not exist in Data Set!");
    } catch (...) {
        FAIL() << "Expected te::common::Exception";
    }
}

TEST_F(VectorDatasetInputAdaptorTest, getInexistentFeature) {
    gdma::cl::VectorDatasetInputAdaptor<double> adaptor(m_inputDataSet);

    try {
        double featureValue;
        adaptor.getFeature(0, 5, featureValue);
        FAIL() << "Expected te::common::Exception Feature index \"5\" does not exist in Data Set!";
    }
    catch (te::common::Exception& e) {
        EXPECT_STREQ(e.what(), "Feature index \"5\" does not exist in Data Set!");
    }
    catch (...) {
        FAIL() << "Expected te::common::Exception";
    }
}

TEST_F(VectorDatasetInputAdaptorTest, getFeature) {
    gdma::cl::VectorDatasetInputAdaptor<double> adaptor(m_inputDataSet);

    double featureValue;
    adaptor.getFeature(0, 0, featureValue);
    EXPECT_DOUBLE_EQ(0, featureValue);
    adaptor.getFeature(0, 1, featureValue);
    EXPECT_DOUBLE_EQ(259, featureValue);
}
