/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/tests/unittest/common/fileSystem_unittest.cpp

  \brief This file contains unit tests for the File System functions

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// GTest includes
#include <gtest/gtest.h>

// Boost includes
#include <boost/filesystem.hpp>

// STL includes
#include <string>

// GeoDMA includes
#include "BuildConfig.hpp"
#include "geodmalib/common/fileSystem.hpp"

class FileSystemTest : public ::testing::Test {
 protected :
    boost::filesystem::path m_tmpFolder;

    virtual void SetUp() {
        GEODMA_HOME = std::getenv(GEODMA_DIR_VAR_NAME);

        m_tmpFolder = "./tmpTests";

        boost::system::error_code error;
        if (boost::filesystem::exists(m_tmpFolder)) {
            if (!boost::filesystem::remove_all(m_tmpFolder, error)) {
                std::string errorMsg = "Temporary folder can not be deleted: " +
                    error.message();
                FAIL() << errorMsg;
            }
        }

        if (!boost::filesystem::create_directory(m_tmpFolder, error)) {
            std::string errorMsg = "Temporary folder not created: " + error.message();
            FAIL() << errorMsg;
        }
    }

    virtual void TearDown() {
        boost::system::error_code error;
        if (boost::filesystem::exists(m_tmpFolder)) {
            if (!boost::filesystem::remove_all(m_tmpFolder, error)) {
                std::string errorMsg = "Temporary folder not deleted: " + error.message();
                FAIL() << errorMsg;
            }
        }
    }
};

TEST_F(FileSystemTest, findExistingDataPath) {
    std::string path;
    ASSERT_NO_THROW({ path = gdma::common::FindInGeoDMAPath("/data/segmentation_example.shp"); });
    char expected[150] = GEODMA_DATA_DIR;
    strcat(expected, "/segmentation_example.shp");  // NOLINT()
    EXPECT_STREQ(expected, path.c_str());
}

TEST_F(FileSystemTest, findExistingCMakePath) {
    std::string path = gdma::common::FindInGeoDMAPath("/build/cmake/CMakeLists.txt");
    ASSERT_TRUE(boost::filesystem::exists(path));
}

TEST_F(FileSystemTest, findInGeoDMAHomePath) {
    gdma::common::SetEnvVariable(GEODMA_DIR_VAR_NAME, GEODMA_DEVEL_PATH);
    std::string path = gdma::common::FindInGeoDMAPath("/data/segmentation_example.shp");
    ASSERT_TRUE(boost::filesystem::exists(path));
}

TEST_F(FileSystemTest, findInexistentPath) {
    ASSERT_THROW(gdma::common::FindInGeoDMAPath("/data/test.tst"), std::exception);
}

TEST_F(FileSystemTest, loadTextFileToString) {
    std::string path;
    ASSERT_NO_THROW({
        path = gdma::common::FindInGeoDMAPath("/data/tests/decision_trees_tests/test.names");
    });

    std::string fileContents;
    ASSERT_TRUE(gdma::common::loadTextFileToString(path, fileContents));

    std::string templateText =
        "| Generated using GeoDMA 2.0.0-alpha_3\n"
        "| on 03/12/18 16:29:33\n"
        "|\n\n"
        "trainClass.\n\n"
        "trainClass: teste,teste2.\n"
        "DISSE_B0: continuous.\n"
        "ENERGSE_B0: continuous.\n"
        "ENTRSE_B0: continuous.\n"
        "HOMOGSE_B0: continuous.\n"
        "DISSE_B1: continuous.\n"
        "ENERGSE_B1: continuous.\n"
        "ENTRSE_B1: continuous.\n"
        "HOMOGSE_B1: continuous.\n"
        "DISSE_B2: continuous.\n"
        "ENERGSE_B2: continuous.\n"
        "ENTRSE_B2: continuous.\n"
        "HOMOGSE_B2: continuous.\n"
        "PGYRATIUS: continuous.\n"
        "POL_ANGLE: continuous.\n";

    ASSERT_STREQ(templateText.c_str(), fileContents.c_str());
}

TEST_F(FileSystemTest, loadTextFileToStringInexistentFile) {
    std::string pathFile(m_tmpFolder.string() + std::string("/unexistent.txt"));

    std::string fileContents;
    ASSERT_FALSE(gdma::common::loadTextFileToString(pathFile, fileContents));
}

TEST_F(FileSystemTest, saveStringToTextFile) {
    std::string pathOutputFile(m_tmpFolder.string() + std::string("/test.names"));

    std::string originalNamesStr =
        "| Generated using GeoDMA 2.0.0-alpha_3\n"
        "| on 03/12/18 16:29:33\n"
        "|\n\n"
        "trainClass.\n\n"
        "trainClass: teste,teste2.\n"
        "DISSE_B0: continuous.\n"
        "ENERGSE_B0: continuous.\n"
        "ENTRSE_B0: continuous.\n"
        "HOMOGSE_B0: continuous.\n"
        "DISSE_B1: continuous.\n"
        "ENERGSE_B1: continuous.\n"
        "ENTRSE_B1: continuous.\n"
        "HOMOGSE_B1: continuous.\n"
        "DISSE_B2: continuous.\n"
        "ENERGSE_B2: continuous.\n"
        "ENTRSE_B2: continuous.\n"
        "HOMOGSE_B2: continuous.\n"
        "PGYRATIUS: continuous.\n"
        "POL_ANGLE: continuous.\n";

    ASSERT_TRUE(gdma::common::saveStringToTextFile(pathOutputFile, originalNamesStr));

    std::string stringFile;
    ASSERT_TRUE(gdma::common::loadTextFileToString(pathOutputFile, stringFile));

    ASSERT_STREQ(originalNamesStr.c_str(), stringFile.c_str());
}

TEST(FileSystem, setEnvVariable) {
    gdma::common::SetEnvVariable("TEST", "Test");
    ASSERT_STREQ(std::getenv("TEST"), "Test");
}
