/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/tests/unittest/common/logMessageObserver_unittest.cpp

  \brief This file contains unit tests for the LogMessageObserver class

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// GTest includes
#include <gtest/gtest.h>

// TerraLib includes
#include <terralib/core/logger/Logger.h>

// Boost Includes
#include <boost/filesystem.hpp>

// STL Includes
#include <string>
#include <fstream>

// GeoDMA Includes
#include "geodmalib/common/logMessageObserver.hpp"

class LogMessageObserverTest : public ::testing::Test {
 protected:
    boost::filesystem::path m_tmpFolder;
    boost::filesystem::path m_logFile;
    std::string m_logFormat;
    std::string m_logName;

    virtual void SetUp() {
        m_tmpFolder = "./tmpTests";
        std::string file(m_tmpFolder.string() + std::string("/logTest.log"));
        m_logFile = file;

        m_logFormat = "<%Severity%>: %Message%";

        boost::system::error_code error;
        if (!boost::filesystem::exists(m_tmpFolder)) {
            if (!boost::filesystem::create_directory(m_tmpFolder, error)) {
                std::string errorMsg = "Temporary folder not created: " + error.message();
                FAIL() << errorMsg;
            }
        }
    }

    virtual void TearDown() {
        te::core::Logger::instance().removeAllLoggers();

        boost::system::error_code error;
        if (boost::filesystem::exists(m_logFile)) {
            if (!boost::filesystem::remove(m_logFile, error)) {
                std::string errorMsg = "Log file not deleted: " + error.message();
                FAIL() << errorMsg;
            }
        }

        if (boost::filesystem::exists(m_tmpFolder)) {
            if (!boost::filesystem::remove(m_tmpFolder, error)) {
                std::string errorMsg = "Temporary folder not deleted: " + error.message();
                FAIL() << errorMsg;
            }
        }
    }
};

TEST_F(LogMessageObserverTest, displayInformationMsgDefault) {
    gdma::common::LogMessageObserver logObs(m_logFile.string());

    logObs.displayInformationMsg("Information Message");

    ASSERT_TRUE(boost::filesystem::exists(m_logFile));

    std::ifstream file(m_logFile.string());
    std::stringstream output;

    output << file.rdbuf();

    std::string retMsg = output.str().substr(output.str().find("]"), output.str().length());

    EXPECT_STREQ("] <info>: Information Message\n", retMsg.c_str());
}

TEST_F(LogMessageObserverTest, displayInformationMsgWithFormat) {
    gdma::common::LogMessageObserver logObs("teste", m_logFile.string(), m_logFormat);

    logObs.displayInformationMsg("Information Message");

    ASSERT_TRUE(boost::filesystem::exists(m_logFile));

    std::ifstream file(m_logFile.string());
    std::stringstream output;

    output << file.rdbuf();

    EXPECT_STREQ("<info>: Information Message\n", output.str().c_str());
}

TEST_F(LogMessageObserverTest, displayWarningMsgWithoutTimeStamp) {
    gdma::common::LogMessageObserver logObs("teste", m_logFile.string(), false, true);

    logObs.displayWarningMsg("Warning Message");

    ASSERT_TRUE(boost::filesystem::exists(m_logFile));

    std::ifstream file(m_logFile.string());
    std::stringstream output;

    output << file.rdbuf();

    EXPECT_STREQ("<warning>: Warning Message\n", output.str().c_str());
}

TEST_F(LogMessageObserverTest, displayWarningMsgWithTimeStamp) {
    gdma::common::LogMessageObserver logObs("teste", m_logFile.string(), true, true);

    logObs.displayWarningMsg("Warning Message");

    ASSERT_TRUE(boost::filesystem::exists(m_logFile));

    std::ifstream file(m_logFile.string());
    std::stringstream output;

    output << file.rdbuf();

    std::string retMsg = output.str().substr(output.str().find("]"), output.str().length());

    EXPECT_STREQ("] <warning>: Warning Message\n", retMsg.c_str());
}

TEST_F(LogMessageObserverTest, displayErrorMsg) {
    gdma::common::LogMessageObserver logObs("teste", m_logFile.string(), false, true);

    logObs.displayErrorMsg("Error Message");

    ASSERT_TRUE(boost::filesystem::exists(m_logFile));

    std::ifstream file(m_logFile.string());
    std::stringstream output;

    output << file.rdbuf();

    EXPECT_STREQ("<error>: Error Message\n", output.str().c_str());
}

TEST_F(LogMessageObserverTest, displayDebugMsg) {
    gdma::common::LogMessageObserver logObs(m_logFile.string());

    logObs.displayDebugMsg("Debug Message");

    ASSERT_TRUE(boost::filesystem::exists(m_logFile));

    std::ifstream file(m_logFile.string());
    std::stringstream output;

    output << file.rdbuf();

    std::string retMsg = output.str().substr(output.str().find("]"), output.str().length());

    EXPECT_STREQ("] <debug>: Debug Message\n", retMsg.c_str());
}

TEST_F(LogMessageObserverTest, displayTraceMsg) {
    gdma::common::LogMessageObserver logObs("teste", m_logFile.string(), m_logFormat);

    logObs.displayTraceMsg("Trace Message");

    ASSERT_TRUE(boost::filesystem::exists(m_logFile));

    std::ifstream file(m_logFile.string());
    std::stringstream output;

    output << file.rdbuf();

    EXPECT_STREQ("<trace>: Trace Message\n", output.str().c_str());
}

TEST_F(LogMessageObserverTest, displayFatalMsg) {
    gdma::common::LogMessageObserver logObs("teste", m_logFile.string(), true, true);

    logObs.displayFatalMsg("Fatal Message");

    ASSERT_TRUE(boost::filesystem::exists(m_logFile));

    std::ifstream file(m_logFile.string());
    std::stringstream output;

    output << file.rdbuf();

    std::string retMsg = output.str().substr(output.str().find("]"), output.str().length());

    EXPECT_STREQ("] <fatal>: Fatal Message\n", retMsg.c_str());
}
