/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/tests/unittest/common/messageMediator_unittest.cpp

  \brief This file contains unit tests for the MessageMediator class

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// GTest includes
#include <gtest/gtest.h>
#include <gmock/gmock.h>

// STL includes
#include <memory>
#include <string>

// GeoDMA Includes
#include "geodmalib/common/abstractMessageObserver.hpp"
#include "geodmalib/common/messageMediator.hpp"

// using ::testing::AtLeast;

// Mock class to simulate Observers
class MockObserver : public gdma::common::AbstractMessageObserver {
 public:
    MockObserver() : AbstractMessageObserver() {}
    MOCK_METHOD1(displayInformationMsg, void(std::string msg));
    MOCK_METHOD1(displayWarningMsg, void(std::string msg));
    MOCK_METHOD1(displayErrorMsg, void(std::string msg));
    MOCK_METHOD1(displayDebugMsg, void(std::string msg));
    MOCK_METHOD1(displayTraceMsg, void(std::string msg));
    MOCK_METHOD1(displayFatalMsg, void(std::string msg));
};

TEST(MessageMediator, addOneObserver) {
    std::unique_ptr<MockObserver> obs(new MockObserver());
    int obsId = gdma::common::MessageMediator::getInstance().addObserver(obs.get());
    EXPECT_EQ(0, obsId);
    EXPECT_EQ(1, gdma::common::MessageMediator::getInstance().getNumberOfObservers());

    // @TODO(Raian): Verify how to destroy the singleton after each test and remove it.
    gdma::common::MessageMediator::getInstance().clearAll();
}

TEST(MessageMediator, addTwoObservers) {
    std::unique_ptr<MockObserver> obs1(new MockObserver());
    int obsId1 = gdma::common::MessageMediator::getInstance().addObserver(obs1.get());

    std::unique_ptr<MockObserver> obs2(new MockObserver());
    int obsId2 = gdma::common::MessageMediator::getInstance().addObserver(obs2.get());

    EXPECT_EQ(0, obsId1);
    EXPECT_EQ(1, obsId2);
    EXPECT_EQ(2, gdma::common::MessageMediator::getInstance().getNumberOfObservers());

    // @TODO(Raian): Verify how to destroy the singleton after each test and remove it.
    gdma::common::MessageMediator::getInstance().clearAll();
}

TEST(MessageMediator, removeObserver) {
    std::unique_ptr<MockObserver> obs1(new MockObserver());
    int obsId1 = gdma::common::MessageMediator::getInstance().addObserver(obs1.get());

    EXPECT_EQ(0, obsId1);
    EXPECT_EQ(1, gdma::common::MessageMediator::getInstance().getNumberOfObservers());

    gdma::common::MessageMediator::getInstance().removeObserver(obsId1);

    EXPECT_EQ(0, gdma::common::MessageMediator::getInstance().getNumberOfObservers());

    // @TODO(Raian): Verify how to destroy the singleton after each test and remove it.
    gdma::common::MessageMediator::getInstance().clearAll();
}

TEST(MessageMediator, sendInformationMsg) {
    std::unique_ptr<MockObserver> obs1(new MockObserver());
    int obsId1 = gdma::common::MessageMediator::getInstance().addObserver(obs1.get());

    EXPECT_CALL(*obs1, displayInformationMsg("Information Message")).Times(1);

    gdma::common::MessageMediator::getInstance().sendInformationMsg("Information Message");

    // @TODO(Raian): Verify how to destroy the singleton after each test and remove it.
    gdma::common::MessageMediator::getInstance().clearAll();
}

TEST(MessageMediator, sendInformationMsgWithoutObserver) {
    EXPECT_NO_THROW({
        gdma::common::MessageMediator::getInstance().sendInformationMsg("Information Message");
    });
}

TEST(MessageMediator, sendWarningMsg) {
    std::unique_ptr<MockObserver> obs1(new MockObserver());
    int obsId1 = gdma::common::MessageMediator::getInstance().addObserver(obs1.get());

    EXPECT_CALL(*obs1, displayWarningMsg("Warning Message")).Times(1);

    gdma::common::MessageMediator::getInstance().sendWarningMsg("Warning Message");

    // @TODO(Raian): Verify how to destroy the singleton after each test and remove it.
    gdma::common::MessageMediator::getInstance().clearAll();
}

TEST(MessageMediator, sendWarningMsgWithoutObserver) {
    EXPECT_NO_THROW({
        gdma::common::MessageMediator::getInstance().sendWarningMsg("Warning Message");
    });
}

TEST(MessageMediator, sendErrorMsg) {
    std::unique_ptr<MockObserver> obs1(new MockObserver());
    int obsId1 = gdma::common::MessageMediator::getInstance().addObserver(obs1.get());

    EXPECT_CALL(*obs1, displayErrorMsg("Error Message")).Times(1);

    gdma::common::MessageMediator::getInstance().sendErrorMsg("Error Message");

    // @TODO(Raian): Verify how to destroy the singleton after each test and remove it.
    gdma::common::MessageMediator::getInstance().clearAll();
}

TEST(MessageMediator, sendErrorMsgWithoutObserver) {
    EXPECT_NO_THROW({
        gdma::common::MessageMediator::getInstance().sendWarningMsg("Error Message");
    });
}

TEST(MessageMediator, sendDebugMsg) {
    std::unique_ptr<MockObserver> obs1(new MockObserver());
    int obsId1 = gdma::common::MessageMediator::getInstance().addObserver(obs1.get());

    EXPECT_CALL(*obs1, displayDebugMsg("Debug Message")).Times(1);

    gdma::common::MessageMediator::getInstance().sendDebugMsg("Debug Message");

    // @TODO(Raian): Verify how to destroy the singleton after each test and remove it.
    gdma::common::MessageMediator::getInstance().clearAll();
}

TEST(MessageMediator, sendDebugMsgWithoutObserver) {
    EXPECT_NO_THROW({
        gdma::common::MessageMediator::getInstance().sendDebugMsg("Debug Message");
    });
}

TEST(MessageMediator, sendTraceMsg) {
    std::unique_ptr<MockObserver> obs1(new MockObserver());
    int obsId1 = gdma::common::MessageMediator::getInstance().addObserver(obs1.get());

    EXPECT_CALL(*obs1, displayTraceMsg("Trace Message")).Times(1);

    gdma::common::MessageMediator::getInstance().sendTraceMsg("Trace Message");

    // @TODO(Raian): Verify how to destroy the singleton after each test and remove it.
    gdma::common::MessageMediator::getInstance().clearAll();
}

TEST(MessageMediator, sendTraceMsgWithoutObserver) {
    EXPECT_NO_THROW({
        gdma::common::MessageMediator::getInstance().sendTraceMsg("Trace Message");
    });
}

TEST(MessageMediator, sendFatalMsg) {
    std::unique_ptr<MockObserver> obs1(new MockObserver());
    int obsId1 = gdma::common::MessageMediator::getInstance().addObserver(obs1.get());

    EXPECT_CALL(*obs1, displayFatalMsg("Fatal Message")).Times(1);

    gdma::common::MessageMediator::getInstance().sendFatalMsg("Fatal Message");

    // @TODO(Raian): Verify how to destroy the singleton after each test and remove it.
    gdma::common::MessageMediator::getInstance().clearAll();
}

TEST(MessageMediator, sendFatalMsgWithoutObserver) {
    EXPECT_NO_THROW({
        gdma::common::MessageMediator::getInstance().sendFatalMsg("Fatal Message");
    });
}
