/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/tests/unittest/datamanager/dataManager_unittest.cpp

  \brief

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// TerraLib includes
#include <terralib/common/TerraLib.h>
#include <terralib/common/Module.h>
#include <terralib/common/PlatformUtils.h>
#include <terralib/common/STLUtils.h>
#include <terralib/dataaccess.h>
#include <terralib/raster/Band.h>
#include <terralib/raster/Grid.h>
#include <terralib/raster/Raster.h>
#include <terralib/raster/BandProperty.h>

// Boost Includes
#include <boost/bimap.hpp>
#include <boost/filesystem.hpp>

// Gtest includes
#include <gtest/gtest.h>

// STL includes
#include <map>
#include <set>
#include <string>
#include <vector>

// GeoDMA includes
#include "geodmalib/datamanager/dataManager.hpp"
#include "geodmalib/common/fileSystem.hpp"
#include "geodmalib/common/utils.hpp"

class DataManagerTest : public ::testing::Test {
 protected:
    boost::filesystem::path m_tmpFolder;

    virtual void SetUp() {
        try {
//             TerraLib::getInstance().initialize();
//             gdma::common::loadTerraLibDrivers();

            m_tmpFolder = "./tmpTests";

            boost::system::error_code error;
            if (boost::filesystem::exists(m_tmpFolder)) {
                if (!boost::filesystem::remove_all(m_tmpFolder, error)) {
                    std::string errorMsg = "Temporary folder can not be deleted: " +
                        error.message();
                    FAIL() << errorMsg;
                }
            } else {
                if (!boost::filesystem::create_directory(m_tmpFolder, error)) {
                    std::string errorMsg = "Temporary folder not created: " + error.message();
                    FAIL() << errorMsg;
                }
            }
        }
        catch (te::common::Exception &e) {
            std::cerr << "ERROR: " << e.what() << std::endl;
        }
    }

    virtual void TearDown() {
//         gdma::common::unloadTerraLibDrivers();
//         TerraLib::getInstance().finalize();

        boost::system::error_code error;
        if (boost::filesystem::exists(m_tmpFolder)) {
            if (!boost::filesystem::remove_all(m_tmpFolder, error)) {
                std::string errorMsg = "Temporary folder not deleted: " + error.message();
                FAIL() << errorMsg;
            }
        }
    }
};

/*!
 \test Test for gdma::dm::openRaster method. This test opens a raster file and verify all of its 
       features, to ensure that it has been correctly opened.
*/
TEST_F(DataManagerTest, openRaster) {
    std::string pathRaster = gdma::common::FindInGeoDMAPath(
        "data/GeoReferencedImagesRegisterExample.tif");
    te::rst::Raster* inputRaster = gdma::dm::openRaster(pathRaster);

    ASSERT_TRUE(inputRaster != NULL);

    // Raster Name
    EXPECT_STREQ(pathRaster.c_str(), inputRaster->getName().c_str());
    // Number of Columns
    EXPECT_EQ(875, inputRaster->getNumberOfColumns());
    // Number of Rows
    EXPECT_EQ(1009, inputRaster->getNumberOfRows());
    // Number of Bands
    EXPECT_EQ(3, inputRaster->getNumberOfBands());
    // SRID
    EXPECT_EQ(29192, inputRaster->getSRID());
    // Resolution in X
    EXPECT_NEAR(20, inputRaster->getResolutionX(), 0.01);
    // Resolution in Y
    EXPECT_NEAR(20, inputRaster->getResolutionY(), 0.01);
    // Upper Right coordinate in X
    EXPECT_NEAR(546480, inputRaster->getGrid()->getExtent()->getUpperRightX(), 0.01);
    // Upper Right Coordinate in Y
    EXPECT_NEAR(7945280, inputRaster->getGrid()->getExtent()->getUpperRightY(), 0.01);
    // Lower Left Coordinate in X
    EXPECT_NEAR(528980, inputRaster->getGrid()->getExtent()->getLowerLeftX(), 0.01);
    // Lower Left Coordinate in Y
    EXPECT_NEAR(7925099.9999, inputRaster->getGrid()->getExtent()->getLowerLeftY(), 0.01);

    // Minimum Values
    // Band 0
    EXPECT_EQ(24, inputRaster->getBand(0)->getMinValue(true).real());
    EXPECT_EQ(0, inputRaster->getBand(0)->getMinValue(true).imag());
    // Band 1
    EXPECT_EQ(34, inputRaster->getBand(1)->getMinValue(true).real());
    EXPECT_EQ(0, inputRaster->getBand(1)->getMinValue(true).imag());
    // Band 2
    EXPECT_EQ(42, inputRaster->getBand(2)->getMinValue(true).real());
    EXPECT_EQ(0, inputRaster->getBand(2)->getMinValue(true).imag());

    // Maximum Values
    // Band 0
    EXPECT_EQ(184, inputRaster->getBand(0)->getMaxValue(true).real());
    EXPECT_DOUBLE_EQ(2.2250738585072014e-308, inputRaster->getBand(0)->getMaxValue(true).imag());
    // Band 1
    EXPECT_EQ(255, inputRaster->getBand(1)->getMaxValue(true).real());
    EXPECT_DOUBLE_EQ(2.2250738585072014e-308, inputRaster->getBand(1)->getMaxValue(true).imag());
    // Band 2
    EXPECT_EQ(255, inputRaster->getBand(2)->getMaxValue(true).real());
    EXPECT_DOUBLE_EQ(2.2250738585072014e-308, inputRaster->getBand(2)->getMaxValue(true).imag());

    // Mean Values
    // Band 0
    EXPECT_DOUBLE_EQ(42.261863797253291, inputRaster->getBand(0)->getMeanValue().real());
    EXPECT_EQ(0, inputRaster->getBand(0)->getMeanValue().imag());
    // Band 1
    EXPECT_DOUBLE_EQ(154.53674076171598, inputRaster->getBand(1)->getMeanValue().real());
    EXPECT_EQ(0, inputRaster->getBand(1)->getMeanValue().imag());
    // Band 2
    EXPECT_DOUBLE_EQ(76.698673085091315, inputRaster->getBand(2)->getMeanValue().real());
    EXPECT_EQ(0, inputRaster->getBand(2)->getMeanValue().imag());

    // Standard Deviation Values
    // Band 0
    EXPECT_DOUBLE_EQ(7.0711116956937623, inputRaster->getBand(0)->getStdValue().real());
    EXPECT_EQ(0, inputRaster->getBand(0)->getStdValue().imag());
    // Band 1
    EXPECT_DOUBLE_EQ(38.284861758924997, inputRaster->getBand(1)->getStdValue().real());
    EXPECT_EQ(0, inputRaster->getBand(1)->getStdValue().imag());
    // Band 2
    EXPECT_DOUBLE_EQ(11.167684590675636, inputRaster->getBand(2)->getStdValue().real());
    EXPECT_EQ(0, inputRaster->getBand(2)->getStdValue().imag());

    // Gain Values
    // Band 0
    EXPECT_EQ(1, inputRaster->getBand(0)->getScaleValue().real());
    EXPECT_EQ(0, inputRaster->getBand(0)->getScaleValue().imag());
    // Band 1
    EXPECT_EQ(1, inputRaster->getBand(1)->getScaleValue().real());
    EXPECT_EQ(0, inputRaster->getBand(1)->getScaleValue().imag());
    // Band 2
    EXPECT_EQ(1, inputRaster->getBand(2)->getScaleValue().real());
    EXPECT_EQ(0, inputRaster->getBand(2)->getScaleValue().imag());

    // Standard Deviation Values
    // Band 0
    EXPECT_EQ(0, inputRaster->getBand(0)->getOffsetValue().real());
    EXPECT_EQ(0, inputRaster->getBand(0)->getOffsetValue().imag());
    // Band 1
    EXPECT_EQ(0, inputRaster->getBand(1)->getOffsetValue().real());
    EXPECT_EQ(0, inputRaster->getBand(1)->getOffsetValue().imag());
    // Band 2
    EXPECT_EQ(0, inputRaster->getBand(2)->getOffsetValue().real());
    EXPECT_EQ(0, inputRaster->getBand(2)->getOffsetValue().imag());
}

/*!
 \test Test for gdma::dm::getGeometriesFromShp method. This test retrieves all the geometries of a 
       shape file and verify if it was correctly executed.
*/
TEST_F(DataManagerTest, getGeometriesFromShpWithProps) {
    std::string pathShp = gdma::common::FindInGeoDMAPath(
        "data/tests/clipped_cells-2x2km-tests-bigger.shp");

    std::vector<std::string> prop_vec;
    std::vector<te::gm::Geometry*> geom_vec;
    boost::bimap<std::string, std::size_t> positions;
    std::set<std::string> exist_classes;

    std::string prop_class("id");

    std::string dsId = gdma::dm::openShapefile(pathShp);
    ASSERT_NO_THROW({
        gdma::dm::getGeometriesFromDataSource(dsId, positions, geom_vec, prop_class, &prop_vec,
            &exist_classes);
    });

    EXPECT_EQ(3, geom_vec.size());
    EXPECT_EQ(3, prop_vec.size());
    EXPECT_EQ(3, positions.size());
    EXPECT_EQ(3, exist_classes.size());

    EXPECT_STREQ("C04L00", prop_vec[positions.left.find("0")->second].c_str());
    EXPECT_STREQ("C05L00", prop_vec[positions.left.find("1")->second].c_str());
    EXPECT_STREQ("C06L00", prop_vec[positions.left.find("2")->second].c_str());

    for (std::size_t i = 0; i < geom_vec.size(); i++) {
        EXPECT_EQ(te::gm::MultiPolygonType, geom_vec[i]->getGeomTypeId());
    }
}

TEST_F(DataManagerTest, getGeometriesUnexistentProperty) {
    std::string pathShp = gdma::common::FindInGeoDMAPath(
        "data/tests/clipped_cells-2x2km-tests-bigger.shp");

    std::vector<std::string> prop_vec;
    std::vector<te::gm::Geometry*> geom_vec;
    boost::bimap<std::string, std::size_t> positions;
    std::set<std::string> exist_classes;

    std::string prop_class("teste");

    std::string dsId = gdma::dm::openShapefile(pathShp);

    try {
        gdma::dm::getGeometriesFromDataSource(dsId, positions, geom_vec, prop_class, &prop_vec,
            &exist_classes);
        FAIL() << "Expected te::dt::Exception column does not exists.";
    } catch (te::common::Exception& e) {
        EXPECT_STREQ(e.what(),
            "Column \"teste\" does not exists in the data set!");
    } catch (...) {
        FAIL() << "Expected te::dt::Exception";
    }

    EXPECT_EQ(0, geom_vec.size());
    EXPECT_EQ(0, prop_vec.size());
    EXPECT_EQ(0, positions.size());
    EXPECT_EQ(0, exist_classes.size());
}

TEST_F(DataManagerTest, getGeometriesFromShpNoProps) {
    std::string pathShp = gdma::common::FindInGeoDMAPath(
        "data/tests/clipped_cells-2x2km-tests-bigger.shp");

    std::vector<te::gm::Geometry*> geom_vec;
    boost::bimap<std::string, std::size_t> positions;

    std::string dsId = gdma::dm::openShapefile(pathShp);
    ASSERT_NO_THROW({
        gdma::dm::getGeometriesFromDataSource(dsId, positions, geom_vec);
    });

    EXPECT_EQ(3, geom_vec.size());
    EXPECT_EQ(3, positions.size());

    for (std::size_t i = 0; i < geom_vec.size(); i++) {
        EXPECT_EQ(te::gm::MultiPolygonType, geom_vec[i]->getGeomTypeId());
    }
}

TEST_F(DataManagerTest, getGeometriesFromShpUnemptyContainers) {
    std::string pathShp = gdma::common::FindInGeoDMAPath(
        "data/tests/clipped_cells-2x2km-tests-bigger.shp");

    std::vector<std::string> prop_vec;
    std::vector<te::gm::Geometry*> geom_vec;
    boost::bimap<std::string, std::size_t> positions;
    std::set<std::string> exist_classes;

    std::string prop_class("id");

    std::string dsId = gdma::dm::openShapefile(pathShp);
    ASSERT_NO_THROW({
        gdma::dm::getGeometriesFromDataSource(dsId, positions, geom_vec, prop_class, &prop_vec,
            &exist_classes);
    });

    EXPECT_EQ(3, geom_vec.size());
    EXPECT_EQ(3, prop_vec.size());
    EXPECT_EQ(3, positions.size());
    EXPECT_EQ(3, exist_classes.size());

    EXPECT_STREQ("C04L00", prop_vec[positions.left.find("0")->second].c_str());
    EXPECT_STREQ("C05L00", prop_vec[positions.left.find("1")->second].c_str());
    EXPECT_STREQ("C06L00", prop_vec[positions.left.find("2")->second].c_str());

    for (std::size_t i = 0; i < geom_vec.size(); i++) {
        EXPECT_EQ(te::gm::MultiPolygonType, geom_vec[i]->getGeomTypeId());
    }

    ASSERT_NO_THROW({
        gdma::dm::getGeometriesFromDataSource(dsId, positions, geom_vec, prop_class, &prop_vec,
            &exist_classes);
    });

    EXPECT_EQ(3, geom_vec.size());
    EXPECT_EQ(3, prop_vec.size());
    EXPECT_EQ(3, positions.size());
    EXPECT_EQ(3, exist_classes.size());

    EXPECT_STREQ("C04L00", prop_vec[positions.left.find("0")->second].c_str());
    EXPECT_STREQ("C05L00", prop_vec[positions.left.find("1")->second].c_str());
    EXPECT_STREQ("C06L00", prop_vec[positions.left.find("2")->second].c_str());
}

TEST_F(DataManagerTest, getGometriesFromShpDuplicateClasses) {
    std::string pathShp = gdma::common::FindInGeoDMAPath(
        "data/tests/clipped_cells-2x2km-tests-bigger.shp");

    std::vector<std::string> prop_vec;
    std::vector<te::gm::Geometry*> geom_vec;
    boost::bimap<std::string, std::size_t> positions;
    std::set<std::string> exist_classes;

    std::string prop_class("row");

    std::string dsId = gdma::dm::openShapefile(pathShp);
    ASSERT_NO_THROW({
        gdma::dm::getGeometriesFromDataSource(dsId, positions, geom_vec, prop_class, &prop_vec,
        &exist_classes);
    });

    EXPECT_EQ(3, geom_vec.size());
    EXPECT_EQ(3, prop_vec.size());
    EXPECT_EQ(3, positions.size());
    EXPECT_EQ(1, exist_classes.size());

    EXPECT_STREQ("0", prop_vec[positions.left.find("0")->second].c_str());
    EXPECT_STREQ("0", prop_vec[positions.left.find("1")->second].c_str());
    EXPECT_STREQ("0", prop_vec[positions.left.find("2")->second].c_str());

    EXPECT_EQ("0", *exist_classes.begin());

    for (std::size_t i = 0; i < geom_vec.size(); i++) {
        EXPECT_EQ(te::gm::MultiPolygonType, geom_vec[i]->getGeomTypeId());
    }
}

TEST_F(DataManagerTest, updateOGRDataSourceString) {
    std::string inputFile = gdma::common::FindInGeoDMAPath(
        "data/tests/clipped_pols_tests-bigger.shp");
    std::string copiedFile = m_tmpFolder.string() + "/clipped_test.shp";

    EXPECT_TRUE(gdma::dm::diskVectorFileCopy(inputFile, copiedFile));

    std::string dsId = gdma::dm::openShapefile(copiedFile);

    EXPECT_STRNE("", dsId.c_str());

    std::vector<te::gm::Geometry*> geom_vec;
    boost::bimap<std::string, std::size_t> positions;

    ASSERT_NO_THROW({
        gdma::dm::getGeometriesFromDataSource(dsId, positions, geom_vec);
    });

    std::map<std::string, gdma::fte::FeaturesSummary> attr_ctner;
    for (auto &elem : positions) {
        gdma::fte::FeaturesSummary fteSumm;
        fteSumm.addFeature("class", "class", "teste");
        attr_ctner[elem.left] = fteSumm;
    }

    ASSERT_NO_THROW({
        EXPECT_TRUE(gdma::dm::updateDataSource(dsId, attr_ctner));
    });

    std::vector<std::string> prop_vec;
    geom_vec.clear();
    positions.clear();
    std::set<std::string> exist_classes;
    std::string prop_class("class");

    ASSERT_NO_THROW({
        gdma::dm::getGeometriesFromDataSource(dsId, positions, geom_vec, prop_class, &prop_vec,
        &exist_classes);
    });

    for (auto &clazz : prop_vec)
        EXPECT_STREQ("teste", clazz.c_str());
}

TEST_F(DataManagerTest, DISABLED_updateOGRDataSourceInt16) {
    std::string inputFile = gdma::common::FindInGeoDMAPath(
        "data/tests/clipped_pols_tests-bigger.shp");
    std::string copiedFile = m_tmpFolder.string() + "/clipped_test.shp";

    EXPECT_TRUE(gdma::dm::diskVectorFileCopy(inputFile, copiedFile));

    std::string dsId = gdma::dm::openShapefile(copiedFile);

    EXPECT_STRNE("", dsId.c_str());

    std::vector<te::gm::Geometry*> geom_vec;
    boost::bimap<std::string, std::size_t> positions;

    ASSERT_NO_THROW({
        gdma::dm::getGeometriesFromDataSource(dsId, positions, geom_vec);
    });

    std::map<std::string, gdma::fte::FeaturesSummary> attr_ctner;
    for (auto &elem : positions) {
        gdma::fte::FeaturesSummary fteSumm;
        fteSumm.addFeature("class", "class", (int16_t)5);
        attr_ctner[elem.left] = fteSumm;
    }

    ASSERT_NO_THROW({
        EXPECT_TRUE(gdma::dm::updateDataSource(dsId, attr_ctner));
    });

    std::vector<std::string> prop_vec;
    geom_vec.clear();
    positions.clear();
    std::set<std::string> exist_classes;
    std::string prop_class("class");

    ASSERT_NO_THROW({
        gdma::dm::getGeometriesFromDataSource(dsId, positions, geom_vec, prop_class, &prop_vec,
        &exist_classes);
    });

    for (auto &clazz : prop_vec)
        EXPECT_STREQ("5", clazz.c_str());
}

TEST_F(DataManagerTest, DISABLED_updateOGRDataSourceInt32) {
    std::string inputFile = gdma::common::FindInGeoDMAPath(
        "data/tests/clipped_pols_tests-bigger.shp");
    std::string copiedFile = m_tmpFolder.string() + "/clipped_test.shp";

    EXPECT_TRUE(gdma::dm::diskVectorFileCopy(inputFile, copiedFile));

    std::string dsId = gdma::dm::openShapefile(copiedFile);

    EXPECT_STRNE("", dsId.c_str());

    std::vector<te::gm::Geometry*> geom_vec;
    boost::bimap<std::string, std::size_t> positions;

    ASSERT_NO_THROW({
        gdma::dm::getGeometriesFromDataSource(dsId, positions, geom_vec);
    });

    std::map<std::string, gdma::fte::FeaturesSummary> attr_ctner;
    for (auto &elem : positions) {
        gdma::fte::FeaturesSummary fteSumm;
        fteSumm.addFeature("class", "class", (int32_t)5);
        attr_ctner[elem.left] = fteSumm;
    }

    ASSERT_NO_THROW({
        EXPECT_TRUE(gdma::dm::updateDataSource(dsId, attr_ctner));
    });

    std::vector<std::string> prop_vec;
    geom_vec.clear();
    positions.clear();
    std::set<std::string> exist_classes;
    std::string prop_class("class");

    ASSERT_NO_THROW({
        gdma::dm::getGeometriesFromDataSource(dsId, positions, geom_vec, prop_class, &prop_vec,
        &exist_classes);
    });

    for (auto &clazz : prop_vec)
        EXPECT_STREQ("5", clazz.c_str());
}

TEST_F(DataManagerTest, DISABLED_updateOGRDataSourceInt64) {
    std::string inputFile = gdma::common::FindInGeoDMAPath(
        "data/tests/clipped_pols_tests-bigger.shp");
    std::string copiedFile = m_tmpFolder.string() + "/clipped_test.shp";

    EXPECT_TRUE(gdma::dm::diskVectorFileCopy(inputFile, copiedFile));

    std::string dsId = gdma::dm::openShapefile(copiedFile);

    EXPECT_STRNE("", dsId.c_str());

    std::vector<te::gm::Geometry*> geom_vec;
    boost::bimap<std::string, std::size_t> positions;

    ASSERT_NO_THROW({
        gdma::dm::getGeometriesFromDataSource(dsId, positions, geom_vec);
    });

    std::map<std::string, gdma::fte::FeaturesSummary> attr_ctner;
    for (auto &elem : positions) {
        gdma::fte::FeaturesSummary fteSumm;
        fteSumm.addFeature("class", "class", (int64_t)5);
        attr_ctner[elem.left] = fteSumm;
    }

    ASSERT_NO_THROW({
        EXPECT_TRUE(gdma::dm::updateDataSource(dsId, attr_ctner));
    });

    std::vector<std::string> prop_vec;
    geom_vec.clear();
    positions.clear();
    std::set<std::string> exist_classes;
    std::string prop_class("class");

    ASSERT_NO_THROW({
        gdma::dm::getGeometriesFromDataSource(dsId, positions, geom_vec, prop_class, &prop_vec,
        &exist_classes);
    });

    for (auto &clazz : prop_vec)
        EXPECT_STREQ("5", clazz.c_str());
}

TEST_F(DataManagerTest, DISABLED_updateOGRDataSourceFloat) {
    std::string inputFile = gdma::common::FindInGeoDMAPath(
        "data/tests/clipped_pols_tests-bigger.shp");
    std::string copiedFile = m_tmpFolder.string() + "/clipped_test.shp";

    EXPECT_TRUE(gdma::dm::diskVectorFileCopy(inputFile, copiedFile));

    std::string dsId = gdma::dm::openShapefile(copiedFile);

    EXPECT_STRNE("", dsId.c_str());

    std::vector<te::gm::Geometry*> geom_vec;
    boost::bimap<std::string, std::size_t> positions;

    ASSERT_NO_THROW({
        gdma::dm::getGeometriesFromDataSource(dsId, positions, geom_vec);
    });

    std::map<std::string, gdma::fte::FeaturesSummary> attr_ctner;
    for (auto &elem : positions) {
        gdma::fte::FeaturesSummary fteSumm;
        fteSumm.addFeature("class", "class", static_cast<float>(2.2));
        attr_ctner[elem.left] = fteSumm;
    }

    ASSERT_NO_THROW({
        EXPECT_TRUE(gdma::dm::updateDataSource(dsId, attr_ctner));
    });

    std::vector<std::string> prop_vec;
    geom_vec.clear();
    positions.clear();
    std::set<std::string> exist_classes;
    std::string prop_class("class");

    ASSERT_NO_THROW({
        gdma::dm::getGeometriesFromDataSource(dsId, positions, geom_vec, prop_class, &prop_vec,
        &exist_classes);
    });

    for (auto &clazz : prop_vec)
        EXPECT_STREQ("2.2", clazz.c_str());
}

TEST_F(DataManagerTest, updateOGRDataSourceDouble) {
    std::string inputFile = gdma::common::FindInGeoDMAPath(
        "data/tests/clipped_pols_tests-bigger.shp");
    std::string copiedFile = m_tmpFolder.string() + "/clipped_test.shp";

    EXPECT_TRUE(gdma::dm::diskVectorFileCopy(inputFile, copiedFile));

    std::string dsId = gdma::dm::openShapefile(copiedFile);

    EXPECT_STRNE("", dsId.c_str());

    std::vector<te::gm::Geometry*> geom_vec;
    boost::bimap<std::string, std::size_t> positions;

    ASSERT_NO_THROW({
        gdma::dm::getGeometriesFromDataSource(dsId, positions, geom_vec);
    });

    std::map<std::string, gdma::fte::FeaturesSummary> attr_ctner;
    for (auto &elem : positions) {
        gdma::fte::FeaturesSummary fteSumm;
        fteSumm.addFeature("class", "class", static_cast<double>(2.2));
        attr_ctner[elem.left] = fteSumm;
    }

    ASSERT_NO_THROW({
        EXPECT_TRUE(gdma::dm::updateDataSource(dsId, attr_ctner));
    });

    std::vector<std::string> prop_vec;
    geom_vec.clear();
    positions.clear();
    std::set<std::string> exist_classes;
    std::string prop_class("class");

    ASSERT_NO_THROW({
        gdma::dm::getGeometriesFromDataSource(dsId, positions, geom_vec, prop_class, &prop_vec,
        &exist_classes);
    });

    // @TODO(Raian): Verify why is it breaking here.
    for (auto &clazz : prop_vec)
        EXPECT_STREQ("2.2000000000000002", clazz.c_str());
}

TEST_F(DataManagerTest, createNewColumnWithString) {
    std::string inputFile = gdma::common::FindInGeoDMAPath(
        "data/tests/clipped_pols_tests-bigger.shp");
    std::string copiedFile = m_tmpFolder.string() + "/clipped_test.shp";

    EXPECT_TRUE(gdma::dm::diskVectorFileCopy(inputFile, copiedFile));

    std::string dsId = gdma::dm::openShapefile(copiedFile);

    EXPECT_STRNE("", dsId.c_str());

    ASSERT_TRUE(gdma::dm::createColumnOGRDataSource(dsId, "teste", te::dt::STRING_TYPE));

    std::vector<std::string> prop_vec;
    std::vector<te::gm::Geometry*> geom_vec;
    boost::bimap<std::string, std::size_t> positions;
    std::set<std::string> exist_classes;

    std::string prop_class("teste");

    ASSERT_NO_THROW({
        gdma::dm::getGeometriesFromDataSource(dsId, positions, geom_vec, prop_class, &prop_vec,
        &exist_classes);
    });

    for (auto &clazz : prop_vec)
        EXPECT_STREQ("", clazz.c_str());
}

TEST_F(DataManagerTest, createNewColumnWithInt16) {
    std::string inputFile = gdma::common::FindInGeoDMAPath(
        "data/tests/clipped_pols_tests-bigger.shp");
    std::string copiedFile = m_tmpFolder.string() + "/clipped_test.shp";

    EXPECT_TRUE(gdma::dm::diskVectorFileCopy(inputFile, copiedFile));

    std::string dsId = gdma::dm::openShapefile(copiedFile);

    EXPECT_STRNE("", dsId.c_str());

    ASSERT_TRUE(gdma::dm::createColumnOGRDataSource(dsId, "teste", te::dt::INT16_TYPE));

    std::vector<std::string> prop_vec;
    std::vector<te::gm::Geometry*> geom_vec;
    boost::bimap<std::string, std::size_t> positions;
    std::set<std::string> exist_classes;

    std::string prop_class("teste");

    ASSERT_NO_THROW({
        gdma::dm::getGeometriesFromDataSource(dsId, positions, geom_vec, prop_class, &prop_vec,
        &exist_classes);
    });

    for (auto &clazz : prop_vec)
        EXPECT_STREQ("0", clazz.c_str());
}

TEST_F(DataManagerTest, createNewColumnWithInt32) {
    std::string inputFile = gdma::common::FindInGeoDMAPath(
        "data/tests/clipped_pols_tests-bigger.shp");
    std::string copiedFile = m_tmpFolder.string() + "/clipped_test.shp";

    EXPECT_TRUE(gdma::dm::diskVectorFileCopy(inputFile, copiedFile));

    std::string dsId = gdma::dm::openShapefile(copiedFile);

    EXPECT_STRNE("", dsId.c_str());

    ASSERT_TRUE(gdma::dm::createColumnOGRDataSource(dsId, "teste", te::dt::INT32_TYPE));

    std::vector<std::string> prop_vec;
    std::vector<te::gm::Geometry*> geom_vec;
    boost::bimap<std::string, std::size_t> positions;
    std::set<std::string> exist_classes;

    std::string prop_class("teste");

    ASSERT_NO_THROW({
        gdma::dm::getGeometriesFromDataSource(dsId, positions, geom_vec, prop_class, &prop_vec,
        &exist_classes);
    });

    for (auto &clazz : prop_vec)
        EXPECT_STREQ("0", clazz.c_str());
}

TEST_F(DataManagerTest, createNewColumnWithInt64) {
    std::string inputFile = gdma::common::FindInGeoDMAPath(
        "data/tests/clipped_pols_tests-bigger.shp");
    std::string copiedFile = m_tmpFolder.string() + "/clipped_test.shp";

    EXPECT_TRUE(gdma::dm::diskVectorFileCopy(inputFile, copiedFile));

    std::string dsId = gdma::dm::openShapefile(copiedFile);

    EXPECT_STRNE("", dsId.c_str());

    ASSERT_TRUE(gdma::dm::createColumnOGRDataSource(dsId, "teste", te::dt::INT64_TYPE));

    std::vector<std::string> prop_vec;
    std::vector<te::gm::Geometry*> geom_vec;
    boost::bimap<std::string, std::size_t> positions;
    std::set<std::string> exist_classes;

    std::string prop_class("teste");

    ASSERT_NO_THROW({
        gdma::dm::getGeometriesFromDataSource(dsId, positions, geom_vec, prop_class, &prop_vec,
        &exist_classes);
    });

    for (auto &clazz : prop_vec)
        EXPECT_STREQ("0", clazz.c_str());
}

TEST_F(DataManagerTest, createNewColumnWithFloat) {
    std::string inputFile = gdma::common::FindInGeoDMAPath(
        "data/tests/clipped_pols_tests-bigger.shp");
    std::string copiedFile = m_tmpFolder.string() + "/clipped_test.shp";

    EXPECT_TRUE(gdma::dm::diskVectorFileCopy(inputFile, copiedFile));

    std::string dsId = gdma::dm::openShapefile(copiedFile);

    EXPECT_STRNE("", dsId.c_str());

    ASSERT_TRUE(gdma::dm::createColumnOGRDataSource(dsId, "teste", te::dt::FLOAT_TYPE));

    std::vector<std::string> prop_vec;
    std::vector<te::gm::Geometry*> geom_vec;
    boost::bimap<std::string, std::size_t> positions;
    std::set<std::string> exist_classes;

    std::string prop_class("teste");

    ASSERT_NO_THROW({
        gdma::dm::getGeometriesFromDataSource(dsId, positions, geom_vec, prop_class, &prop_vec,
        &exist_classes);
    });

    for (auto &clazz : prop_vec)
        EXPECT_STREQ("0", clazz.c_str());
}

TEST_F(DataManagerTest, createNewColumnWithDouble) {
    std::string inputFile = gdma::common::FindInGeoDMAPath(
        "data/tests/clipped_pols_tests-bigger.shp");
    std::string copiedFile = m_tmpFolder.string() + "/clipped_test.shp";

    EXPECT_TRUE(gdma::dm::diskVectorFileCopy(inputFile, copiedFile));

    std::string dsId = gdma::dm::openShapefile(copiedFile);

    EXPECT_STRNE("", dsId.c_str());

    ASSERT_TRUE(gdma::dm::createColumnOGRDataSource(dsId, "teste", te::dt::DOUBLE_TYPE));

    std::vector<std::string> prop_vec;
    std::vector<te::gm::Geometry*> geom_vec;
    boost::bimap<std::string, std::size_t> positions;
    std::set<std::string> exist_classes;

    std::string prop_class("teste");

    ASSERT_NO_THROW({
        gdma::dm::getGeometriesFromDataSource(dsId, positions, geom_vec, prop_class, &prop_vec,
        &exist_classes);
    });

    for (auto &clazz : prop_vec)
        EXPECT_STREQ("0", clazz.c_str());
}

TEST_F(DataManagerTest, createRepeatedColumn) {
    std::string inputFile = gdma::common::FindInGeoDMAPath(
        "data/tests/clipped_pols_tests-bigger.shp");
    std::string copiedFile = m_tmpFolder.string() + "/clipped_test.shp";

    EXPECT_TRUE(gdma::dm::diskVectorFileCopy(inputFile, copiedFile));
    std::string dsId = gdma::dm::openShapefile(copiedFile);
    EXPECT_STRNE("", dsId.c_str());

    ASSERT_NO_THROW({
        ASSERT_TRUE(gdma::dm::createColumnOGRDataSource(dsId, "teste", te::dt::DOUBLE_TYPE));
    });

    std::vector<std::string> prop_vec;
    std::vector<te::gm::Geometry*> geom_vec;
    boost::bimap<std::string, std::size_t> positions;
    std::set<std::string> exist_classes;

    std::string prop_class("teste");

    ASSERT_NO_THROW({
        gdma::dm::getGeometriesFromDataSource(dsId, positions, geom_vec, prop_class, &prop_vec,
        &exist_classes);
    });

    for (auto &clazz : prop_vec)
        EXPECT_STREQ("0", clazz.c_str());

    try {
        gdma::dm::createColumnOGRDataSource(dsId, "teste", te::dt::DOUBLE_TYPE);
        FAIL() << "Expected te::dt::Exception column not created.";
    } catch (te::common::Exception& e) {
        EXPECT_STREQ(e.what(),
            "Column \"teste\" can not be created. It already exists in the data set!");
    } catch (...) {
        FAIL() << "Expected te::dt::Exception";
    }
}

TEST_F(DataManagerTest, createColumnInvalidCharacters) {
    std::string inputFile = gdma::common::FindInGeoDMAPath(
        "data/tests/clipped_pols_tests-bigger.shp");
    std::string copiedFile = m_tmpFolder.string() + "/clipped_test.shp";

    EXPECT_TRUE(gdma::dm::diskVectorFileCopy(inputFile, copiedFile));
    std::string dsId = gdma::dm::openShapefile(copiedFile);
    EXPECT_STRNE("", dsId.c_str());

    try {
        gdma::dm::createColumnOGRDataSource(dsId, "ab", te::dt::DOUBLE_TYPE);  //NOLINT
        FAIL() << "Expected te::dt::Exception column not created.";
    } catch (te::common::Exception& e) {
        EXPECT_STREQ(e.what(),
            "Column Name \"ab\" contains special characters!");  //NOLINT
    } catch (...) {
        FAIL() << "Expected te::dt::Exception";
    }

    try {
        gdma::dm::createColumnOGRDataSource(dsId, "a b", te::dt::DOUBLE_TYPE);  //NOLINT
        FAIL() << "Expected te::dt::Exception column not created.";
    } catch (te::common::Exception& e) {
        EXPECT_STREQ(e.what(),
            "Column Name \"a b\" contains special characters!");  //NOLINT
    } catch (...) {
        FAIL() << "Expected te::dt::Exception";
    }

    try {
        gdma::dm::createColumnOGRDataSource(dsId, "a\b", te::dt::DOUBLE_TYPE);  //NOLINT
        FAIL() << "Expected te::dt::Exception column not created.";
    } catch (te::common::Exception& e) {
        EXPECT_STREQ(e.what(),
            "Column Name \"a\b\" contains special characters!");  //NOLINT
    } catch (...) {
        FAIL() << "Expected te::dt::Exception";
    }

    std::vector<std::string> prop_vec;
    std::vector<te::gm::Geometry*> geom_vec;
    boost::bimap<std::string, std::size_t> positions;
    std::set<std::string> exist_classes;

    std::string prop_class("ab");  //NOLINT

    try {
        gdma::dm::getGeometriesFromDataSource(dsId, positions, geom_vec, prop_class, &prop_vec,
            &exist_classes);
        FAIL() << "Expected te::dt::Exception column not created.";
    } catch (te::common::Exception& e) {
        EXPECT_STREQ(e.what(),
            "Column \"ab\" does not exists in the data set!");  //NOLINT
    } catch (...) {
        FAIL() << "Expected te::dt::Exception";
    }

    prop_class = "a b";  //NOLINT

    try {
        gdma::dm::getGeometriesFromDataSource(dsId, positions, geom_vec, prop_class, &prop_vec,
            &exist_classes);
        FAIL() << "Expected te::dt::Exception column not created.";
    } catch (te::common::Exception& e) {
        EXPECT_STREQ(e.what(),
            "Column \"a b\" does not exists in the data set!");  //NOLINT
    } catch (...) {
        FAIL() << "Expected te::dt::Exception";
    }

    prop_class = "a\b";  //NOLINT

    try {
        gdma::dm::getGeometriesFromDataSource(dsId, positions, geom_vec, prop_class, &prop_vec,
            &exist_classes);
        FAIL() << "Expected te::dt::Exception column not created.";
    } catch (te::common::Exception& e) {
        EXPECT_STREQ(e.what(),
            "Column \"a\b\" does not exists in the data set!");  //NOLINT
    } catch (...) {
        FAIL() << "Expected te::dt::Exception";
    }
}

TEST_F(DataManagerTest, updateOGRDataSourceInvalidColumnName) {
    std::string inputFile = gdma::common::FindInGeoDMAPath(
        "data/tests/clipped_pols_tests-bigger.shp");
    std::string copiedFile = m_tmpFolder.string() + "/clipped_test.shp";

    EXPECT_TRUE(gdma::dm::diskVectorFileCopy(inputFile, copiedFile));

    std::string dsId = gdma::dm::openShapefile(copiedFile);

    EXPECT_STRNE("", dsId.c_str());

    std::vector<te::gm::Geometry*> geom_vec;
    boost::bimap<std::string, std::size_t> positions;

    ASSERT_NO_THROW({
        gdma::dm::getGeometriesFromDataSource(dsId, positions, geom_vec);
    });

    std::map<std::string, gdma::fte::FeaturesSummary> attr_ctner;
    for (auto &elem : positions) {
        gdma::fte::FeaturesSummary fteSumm;
        fteSumm.addFeature("c", "c", "teste");  //NOLINT
        attr_ctner[elem.left] = fteSumm;
    }

    try {
        EXPECT_TRUE(gdma::dm::updateDataSource(dsId, attr_ctner));
        FAIL() << "Expected te::dt::Exception column not created.";
    } catch (te::common::Exception& e) {
        EXPECT_STREQ(e.what(), "Column Name \"c\" contains special characters!");  //NOLINT
    } catch (...) {
        FAIL() << "Expected te::dt::Exception";
    }

    std::vector<std::string> prop_vec;
    geom_vec.clear();
    positions.clear();
    std::set<std::string> exist_classes;
    std::string prop_class("c");  //NOLINT

    try {
        gdma::dm::getGeometriesFromDataSource(dsId, positions, geom_vec, prop_class, &prop_vec,
            &exist_classes);
        FAIL() << "Expected te::dt::Exception column not created.";
    } catch (te::common::Exception& e) {
        EXPECT_STREQ(e.what(),
            "Column \"c\" does not exists in the data set!");  //NOLINT
    } catch (...) {
        FAIL() << "Expected te::dt::Exception";
    }
}

TEST_F(DataManagerTest, saveDataSourceToCSV) {
    std::string inputSHP = gdma::common::FindInGeoDMAPath(
        "data/tests/clipped_pols_tests-Samples.shp");
    std::string dsId = gdma::dm::openShapefile(inputSHP);
    std::string pathCSV = m_tmpFolder.string() + std::string("/csvSamples.csv");

    ASSERT_TRUE(gdma::dm::saveDataSourceToCSV(dsId, pathCSV));

    std::string pathTemplateFile =
        gdma::common::FindInGeoDMAPath("data/tests/csvSamplesTemplate2.csv");
    std::ifstream templateFile;
    templateFile.open(pathTemplateFile, std::ios::in);

    std::ifstream generatedFile;
    generatedFile.open(pathCSV.c_str(), std::ios::in);

    std::string templateStr;
    std::string generatedStr;

    std::getline(templateFile, templateStr);
    std::getline(generatedFile, generatedStr);
    EXPECT_STREQ(templateStr.c_str(), generatedStr.c_str());

    while (!templateFile.eof()) {
        std::getline(templateFile, templateStr);
        std::getline(generatedFile, generatedStr);
        EXPECT_STREQ(templateStr.c_str(), generatedStr.c_str());
    }

    EXPECT_TRUE(templateFile.eof());
    EXPECT_TRUE(generatedFile.eof());
}

TEST_F(DataManagerTest, saveDataSourceToCSVVerifyTruncate) {
    std::string inputSHP = gdma::common::FindInGeoDMAPath(
        "data/tests/clipped_pols_tests-Samples_Feat.shp");
    std::string dsId = gdma::dm::openShapefile(inputSHP);
    std::string pathCSV = m_tmpFolder.string() + std::string("/csvSamples.csv");

    ASSERT_TRUE(gdma::dm::saveDataSourceToCSV(dsId, pathCSV));

    std::string pathTemplateFile =
        gdma::common::FindInGeoDMAPath("data/tests/csv_feat_template.csv");
    std::ifstream templateFile;
    templateFile.open(pathTemplateFile, std::ios::in);

    std::ifstream generatedFile;
    generatedFile.open(pathCSV.c_str(), std::ios::in);

    std::string templateStr;
    std::string generatedStr;

    std::getline(templateFile, templateStr);
    std::getline(generatedFile, generatedStr);
    EXPECT_STREQ(templateStr.c_str(), generatedStr.c_str());

    while (!templateFile.eof()) {
        std::getline(templateFile, templateStr);
        std::getline(generatedFile, generatedStr);
        EXPECT_STREQ(templateStr.c_str(), generatedStr.c_str());
    }

    EXPECT_TRUE(templateFile.eof());
    EXPECT_TRUE(generatedFile.eof());
}

TEST_F(DataManagerTest, saveShpFromVector) {
    std::string inputSHP = gdma::common::FindInGeoDMAPath(
        "data/tests/clipped_pols_tests-Samples_Feat.shp");
    std::string path_output_SHP = m_tmpFolder.string() + std::string("/outputSHP.shp");

    te::da::DataSourceInfoPtr dsInfo = gdma::dm::openDataSource(inputSHP);
    te::da::DataSourcePtr input_data_source = te::da::GetDataSource(dsInfo->getId());
    std::string dataset_name = input_data_source->getDataSetNames()[0];
    // Retrieve the datasetType
    te::da::DataSetTypePtr dataSetTypePtr = input_data_source->getDataSetType(dataset_name);

    te::da::DataSetPtr dataset = input_data_source->getDataSet(dataset_name);


    std::vector<te::gm::Geometry*> vec_gm = gdma::dm::getGeometriesFromDataset(
        dataset.get(), dataset->size(), 0);

    ASSERT_TRUE(gdma::dm::saveShpFromVector(vec_gm, path_output_SHP));

    ASSERT_TRUE(boost::filesystem::exists(boost::filesystem::path(path_output_SHP)));
}

TEST_F(DataManagerTest, saveToOGRDataSource) {
    std::string inputSHP = gdma::common::FindInGeoDMAPath(
        "data/tests/clipped_pols_tests-Samples_Feat.shp");
    std::string path_output_SHP = m_tmpFolder.string() + std::string("/outputSHP.shp");

    te::da::DataSourceInfoPtr dsInfo = gdma::dm::openDataSource(inputSHP);
    te::da::DataSourcePtr input_data_source = te::da::GetDataSource(dsInfo->getId());
    std::string dataset_name = input_data_source->getDataSetNames()[0];
    // Retrieve the datasetType
    te::da::DataSetTypePtr dataSetTypePtr = input_data_source->getDataSetType(dataset_name);

    std::map<std::string, gdma::fte::FeaturesSummary> attr_container;

    for (int i = 0; i < 5; i++) {
        gdma::fte::FeaturesSummary feat;
        feat.addFeature("test_sv", "test_sv", 2.0);
        attr_container[std::to_string(i)] = feat;
    }

    ASSERT_TRUE(gdma::dm::saveToOGRDataSource(path_output_SHP, dataset_name, dsInfo->getId(),
        attr_container));

    ASSERT_TRUE(boost::filesystem::exists(boost::filesystem::path(path_output_SHP)));
}
