/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/tests/unittest/datamanager/datasetIterator_unittest.cpp

  \brief This file contains the implementation of the unit tests for the DatasetIterator class.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// TerraLib Includes
#include <terralib/common/TerraLib.h>
#include <terralib/dataaccess/dataset/DataSet.h>
#include <terralib/dataaccess/utils/Utils.h>

// Gtest includes
#include <gtest/gtest.h>

// STL Includes
#include <string>
#include <vector>

// GeoDMA Includes
#include "geodmalib/common/utils.hpp"
#include "geodmalib/common/fileSystem.hpp"
#include "geodmalib/datamanager/dataManager.hpp"
#include "geodmalib/datamanager/datasetIterator.hpp"

class DatasetIteratorTest : public ::testing::Test {
 protected:
    te::da::DataSetPtr m_dataset;

    virtual void SetUp() {
        try {
//             TerraLib::getInstance().initialize();
//             gdma::common::loadTerraLibDrivers();

            std::string pathShp = gdma::common::FindInGeoDMAPath(
                "data/tests/clipped_pols_tests-Samples_Feat.shp");
            std::string dsId = gdma::dm::openShapefile(pathShp);
            te::da::DataSourcePtr datasource = te::da::GetDataSource(dsId);
            m_dataset.reset(te::da::GetDataSet(datasource->getDataSetNames()[0], dsId));
        } catch (te::common::Exception &e) {
            std::cerr << "ERROR: " << e.what() << std::endl;
        }
    }

    virtual void TearDown() {
        m_dataset.reset();

//         gdma::common::unloadTerraLibDrivers();
//         TerraLib::getInstance().finalize();
    }
};

TEST_F(DatasetIteratorTest, iteratorCreation) {
    int current;

    ASSERT_NO_THROW({
        gdma::dm::DatasetIterator it = gdma::dm::DatasetIterator::begin(m_dataset);
        current = it.currentPosition();
    });

    EXPECT_EQ(0, current);
}

TEST_F(DatasetIteratorTest, iteratorEnd) {
    int current;

    ASSERT_NO_THROW({
        gdma::dm::DatasetIterator it = gdma::dm::DatasetIterator::end(m_dataset);
        current = it.currentPosition();
    });

    EXPECT_EQ(6, current);
}

TEST_F(DatasetIteratorTest, iteratorNotEqual) {
    gdma::dm::DatasetIterator it_begin = gdma::dm::DatasetIterator::begin(m_dataset);
    gdma::dm::DatasetIterator it_end = gdma::dm::DatasetIterator::end(m_dataset);

    EXPECT_FALSE(it_begin == it_end);
    EXPECT_TRUE(it_begin != it_end);
}

TEST_F(DatasetIteratorTest, iterationOnPositions) {
    gdma::dm::DatasetIterator it_begin = gdma::dm::DatasetIterator::begin(m_dataset);
    gdma::dm::DatasetIterator it_end = gdma::dm::DatasetIterator::end(m_dataset);

    int count = 0;
    while (it_begin != it_end) {
        count++;
        ++it_begin;
    }

    EXPECT_EQ(6, count);
}

TEST_F(DatasetIteratorTest, iterationOnPositionsGettingValues) {
    gdma::dm::DatasetIterator it = gdma::dm::DatasetIterator::begin(m_dataset);
    gdma::dm::DatasetIterator it_end = gdma::dm::DatasetIterator::end(m_dataset);

    std::vector<double> values;
    while (it != it_end) {
        values.push_back(it[4]);
        ++it;
    }

    EXPECT_EQ(6, values.size());
    EXPECT_EQ(8847, values[0]);
    EXPECT_EQ(3344, values[1]);
    EXPECT_EQ(523, values[2]);
    EXPECT_EQ(1288, values[3]);
    EXPECT_EQ(639, values[4]);
    EXPECT_EQ(1055, values[5]);
}

TEST_F(DatasetIteratorTest, iterationOnPositionsGettingValuesByString) {
    gdma::dm::DatasetIterator it = gdma::dm::DatasetIterator::begin(m_dataset);
    gdma::dm::DatasetIterator it_end = gdma::dm::DatasetIterator::end(m_dataset);

    std::vector<double> values;
    while (it != it_end) {
        values.push_back(it["KURT_B0"]);
        ++it;
    }

    EXPECT_EQ(6, values.size());
    EXPECT_DOUBLE_EQ(0.0052163704656000001, values[0]);
    EXPECT_DOUBLE_EQ(-0.32835944741712297, values[1]);
    EXPECT_DOUBLE_EQ(6.1806382237532436, values[2]);
    EXPECT_DOUBLE_EQ(8.6046974973289707, values[3]);
    EXPECT_DOUBLE_EQ(5.6349595404761388, values[4]);
    EXPECT_DOUBLE_EQ(6.238792187156343, values[5]);
}

TEST_F(DatasetIteratorTest, assignmentOperator) {
    gdma::dm::DatasetIterator it = gdma::dm::DatasetIterator::begin(m_dataset);
    EXPECT_EQ(0, it.currentPosition());
    ++it;
    gdma::dm::DatasetIterator itCp = it;
    EXPECT_EQ(1, itCp.currentPosition());
    EXPECT_EQ(1, it.currentPosition());

    EXPECT_TRUE(it == itCp);
    ++it;
    EXPECT_TRUE(it != itCp);
    EXPECT_EQ(1, itCp.currentPosition());
    EXPECT_EQ(2, it.currentPosition());
}

TEST_F(DatasetIteratorTest, iterationOnPositionsGettingStringValues) {
    gdma::dm::DatasetIterator it = gdma::dm::DatasetIterator::begin(m_dataset);
    gdma::dm::DatasetIterator it_end = gdma::dm::DatasetIterator::end(m_dataset);

    std::vector<std::string> values;
    while (it != it_end) {
        values.push_back(it(2));
        ++it;
    }

    EXPECT_EQ(6, values.size());
    EXPECT_STREQ("teste2", values[0].c_str());
    EXPECT_STREQ("teste2", values[1].c_str());
    EXPECT_STREQ("teste", values[2].c_str());
    EXPECT_STREQ("teste", values[3].c_str());
    EXPECT_STREQ("", values[4].c_str());
    EXPECT_STREQ("teste2", values[5].c_str());
}

TEST_F(DatasetIteratorTest, iterationOnPositionsGettingStringValuesByString) {
    gdma::dm::DatasetIterator it = gdma::dm::DatasetIterator::begin(m_dataset);
    gdma::dm::DatasetIterator it_end = gdma::dm::DatasetIterator::end(m_dataset);

    std::vector<std::string> values;
    while (it != it_end) {
        values.push_back(it("trainClass"));
        ++it;
    }

    EXPECT_EQ(6, values.size());
    EXPECT_STREQ("teste2", values[0].c_str());
    EXPECT_STREQ("teste2", values[1].c_str());
    EXPECT_STREQ("teste", values[2].c_str());
    EXPECT_STREQ("teste", values[3].c_str());
    EXPECT_STREQ("", values[4].c_str());
    EXPECT_STREQ("teste2", values[5].c_str());
}
