/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/tests/unittest/datamanager/datasetParcialIterator_unittest.cpp

  \brief This file contains the implementation of the unit tests for the
         DatasetRandomParcialIterator class.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// TerraLib Includes
#include <terralib/common/Exception.h>
#include <terralib/common/TerraLib.h>
#include <terralib/dataaccess/datasource/DataSource.h>
#include <terralib/dataaccess/utils/Utils.h>

// Gtest includes
#include <gtest/gtest.h>

// Boost Includes
#include <boost/filesystem.hpp>

// STL Includes
#include <memory>
#include <string>
#include <vector>

// GeoDMA Includes
#include "geodmalib/common/utils.hpp"
#include "geodmalib/common/fileSystem.hpp"
#include "geodmalib/datamanager/dataManager.hpp"
#include "geodmalib/datamanager/datasetParcialIterator.hpp"

class DatasetParcialIteratorTest : public ::testing::Test {
 protected:
    te::da::DataSetPtr m_dataset;
    std::shared_ptr<std::vector<unsigned int>> m_positions;

    virtual void SetUp() {
        try {
//             TerraLib::getInstance().initialize();
//             gdma::common::loadTerraLibDrivers();

            std::string pathShp = gdma::common::FindInGeoDMAPath(
                "data/tests/clipped_pols_tests-Samples_Feat.shp");
            std::string dsId = gdma::dm::openShapefile(pathShp);
            te::da::DataSourcePtr datasource = te::da::GetDataSource(dsId);
            m_dataset.reset(te::da::GetDataSet(datasource->getDataSetNames()[0], dsId));

            m_positions.reset(new std::vector<unsigned int>());
            m_positions->push_back(1);
            m_positions->push_back(2);
            m_positions->push_back(4);
            m_positions->push_back(5);
        } catch (te::common::Exception &e) {
            std::cerr << "ERROR: " << e.what() << std::endl;
        }
    }

    virtual void TearDown() {
        m_dataset.reset();

//         gdma::common::unloadTerraLibDrivers();
//         TerraLib::getInstance().finalize();
    }
};

TEST_F(DatasetParcialIteratorTest, iteratorCreation) {
    int current;

    ASSERT_NO_THROW({
        gdma::dm::DatasetParcialIterator it =
            gdma::dm::DatasetParcialIterator::begin(m_dataset, m_positions);
        current = it.currentPosition();
    });

    EXPECT_EQ(1, current);
}

TEST_F(DatasetParcialIteratorTest, iteratorEnd) {
    int current;

    ASSERT_NO_THROW({
        gdma::dm::DatasetParcialIterator it =
            gdma::dm::DatasetParcialIterator::end(m_dataset, m_positions);
        current = it.currentPosition();
    });

    EXPECT_EQ(6, current);
}

TEST_F(DatasetParcialIteratorTest, iteratorMorePositionsThanDataSet) {
    m_positions->push_back(6);
    m_positions->push_back(7);
    m_positions->push_back(8);
    m_positions->push_back(9);

    ASSERT_THROW({
        gdma::dm::DatasetParcialIterator it =
        gdma::dm::DatasetParcialIterator::begin(m_dataset, m_positions);
    }, te::common::Exception);
}

TEST_F(DatasetParcialIteratorTest, iteratorNotEqual) {
    gdma::dm::DatasetParcialIterator it_begin =
        gdma::dm::DatasetParcialIterator::begin(m_dataset, m_positions);
    gdma::dm::DatasetParcialIterator it_end =
        gdma::dm::DatasetParcialIterator::end(m_dataset, m_positions);

    EXPECT_FALSE(it_begin == it_end);
    EXPECT_TRUE(it_begin != it_end);
}

TEST_F(DatasetParcialIteratorTest, iterationOnPositions) {
    gdma::dm::DatasetParcialIterator it_begin =
        gdma::dm::DatasetParcialIterator::begin(m_dataset, m_positions);
    gdma::dm::DatasetParcialIterator it_end =
        gdma::dm::DatasetParcialIterator::end(m_dataset, m_positions);

    int count = 0;
    while (it_begin != it_end) {
        count++;
        ++it_begin;
    }

    EXPECT_EQ(4, count);
}

TEST_F(DatasetParcialIteratorTest, iterationOnPositionsGettingValues) {
    gdma::dm::DatasetParcialIterator it =
        gdma::dm::DatasetParcialIterator::begin(m_dataset, m_positions);
    gdma::dm::DatasetParcialIterator it_end =
        gdma::dm::DatasetParcialIterator::end(m_dataset, m_positions);

    std::vector<double> values;
    while (it != it_end) {
        values.push_back(it[4]);
        ++it;
    }

    EXPECT_EQ(4, values.size());
    EXPECT_EQ(3344, values[0]);
    EXPECT_EQ(523, values[1]);
    EXPECT_EQ(639, values[2]);
    EXPECT_EQ(1055, values[3]);
}

TEST_F(DatasetParcialIteratorTest, iterationOnPositionsGettingValuesByString) {
    gdma::dm::DatasetParcialIterator it =
        gdma::dm::DatasetParcialIterator::begin(m_dataset, m_positions);
    gdma::dm::DatasetParcialIterator it_end =
        gdma::dm::DatasetParcialIterator::end(m_dataset, m_positions);

    std::vector<double> values;
    while (it != it_end) {
        values.push_back(it["KURT_B0"]);
        ++it;
    }

    EXPECT_EQ(4, values.size());
    EXPECT_DOUBLE_EQ(-0.32835944741712297, values[0]);
    EXPECT_DOUBLE_EQ(6.1806382237532436, values[1]);
    EXPECT_DOUBLE_EQ(5.6349595404761388, values[2]);
    EXPECT_DOUBLE_EQ(6.238792187156343, values[3]);
}

TEST_F(DatasetParcialIteratorTest, assignmentOperator) {
    gdma::dm::DatasetParcialIterator it =
        gdma::dm::DatasetParcialIterator::begin(m_dataset, m_positions);
    EXPECT_EQ(1, it.currentPosition());
    ++it;
    gdma::dm::DatasetParcialIterator itCp = it;
    EXPECT_TRUE(it == itCp);
    ++it;
    EXPECT_TRUE(it != itCp);
}

TEST_F(DatasetParcialIteratorTest, iterationOnPositionsGettingStringValues) {
    gdma::dm::DatasetParcialIterator it =
        gdma::dm::DatasetParcialIterator::begin(m_dataset, m_positions);
    gdma::dm::DatasetParcialIterator it_end =
        gdma::dm::DatasetParcialIterator::end(m_dataset, m_positions);

    std::vector<std::string> values;
    while (it != it_end) {
        values.push_back(it(2));
        ++it;
    }

    EXPECT_EQ(4, values.size());
    EXPECT_STREQ("teste2", values[0].c_str());
    EXPECT_STREQ("teste", values[1].c_str());
    EXPECT_STREQ("", values[2].c_str());
    EXPECT_STREQ("teste2", values[3].c_str());
}

TEST_F(DatasetParcialIteratorTest, iterationOnPositionsGettingStringValuesByString) {
    gdma::dm::DatasetParcialIterator it =
        gdma::dm::DatasetParcialIterator::begin(m_dataset, m_positions);
    gdma::dm::DatasetParcialIterator it_end =
        gdma::dm::DatasetParcialIterator::end(m_dataset, m_positions);

    std::vector<std::string> values;
    while (it != it_end) {
        values.push_back(it("trainClass"));
        ++it;
    }

    EXPECT_EQ(4, values.size());
    EXPECT_STREQ("teste2", values[0].c_str());
    EXPECT_STREQ("teste", values[1].c_str());
    EXPECT_STREQ("", values[2].c_str());
    EXPECT_STREQ("teste2", values[3].c_str());
}
