/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/featureextraction/attribute_unittest.hpp

  \brief Unit tests for the class gdma::fte::Attribute

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// GTest Includes
#include <gtest/gtest.h>
#include <gmock/gmock.h>

#include <string>

// GeoDMA Includes
#include "geodmalib/common/abstractMessageObserver.hpp"
#include "geodmalib/common/messageMediator.hpp"
#include "geodmalib/featureextraction/attribute.hpp"

// Mock class to simulate Observers
class MockObserver : public gdma::common::AbstractMessageObserver {
 public:
    MockObserver() : AbstractMessageObserver() {}
    MOCK_METHOD1(displayInformationMsg, void(std::string msg));
    MOCK_METHOD1(displayWarningMsg, void(std::string msg));
    MOCK_METHOD1(displayErrorMsg, void(std::string msg));
    MOCK_METHOD1(displayDebugMsg, void(std::string msg));
    MOCK_METHOD1(displayTraceMsg, void(std::string msg));
    MOCK_METHOD1(displayFatalMsg, void(std::string msg));
};

TEST(AttributeTest, stringAttribute) {
    gdma::fte::Attribute a("name", "name", "value");

    EXPECT_STREQ("name", a.getName().c_str());
    EXPECT_STREQ("name", a.getShortName().c_str());
    EXPECT_STREQ("value", a.getString().c_str());
    EXPECT_EQ(te::dt::STRING_TYPE, a.getValueType());
}

TEST(AttributeTest, stringAttributeLargeShortName) {
    MockObserver obs;
    gdma::common::MessageMediator::getInstance().addObserver(&obs);

    std::string msg =
        "Feature short name 'nameTooLarge' is too long! It will be truncated to 'nameTooLar'!";

    EXPECT_CALL(obs, displayWarningMsg(msg)).Times(1);
    gdma::fte::Attribute a("name", "nameTooLarge", "value");

    EXPECT_STREQ("name", a.getName().c_str());
    EXPECT_STREQ("nameTooLar", a.getShortName().c_str());
    EXPECT_STREQ("value", a.getString().c_str());
    EXPECT_EQ(te::dt::STRING_TYPE, a.getValueType());

    // @TODO(Raian): Verify how to destroy the singleton after each test and remove it.
    gdma::common::MessageMediator::getInstance().clearAll();
}

TEST(AttributeTest, int16Attribute) {
    int16_t val = 10;
    gdma::fte::Attribute a("name", "name", val);

    EXPECT_STREQ("name", a.getName().c_str());
    EXPECT_STREQ("name", a.getShortName().c_str());
    EXPECT_EQ(10, a.getInt16());
    EXPECT_EQ(te::dt::INT16_TYPE, a.getValueType());
}

TEST(AttributeTest, int16AttributeLargeShortName) {
    MockObserver obs;
    gdma::common::MessageMediator::getInstance().addObserver(&obs);

    std::string msg =
        "Feature short name 'nameTooLarge' is too long! It will be truncated to 'nameTooLar'!";
    EXPECT_CALL(obs, displayWarningMsg(msg)).Times(1);

    int16_t val = 10;
    gdma::fte::Attribute a("name", "nameTooLarge", val);

    EXPECT_STREQ("name", a.getName().c_str());
    EXPECT_STREQ("nameTooLar", a.getShortName().c_str());
    EXPECT_EQ(10, a.getInt16());
    EXPECT_EQ(te::dt::INT16_TYPE, a.getValueType());

    // @TODO(Raian): Verify how to destroy the singleton after each test and remove it.
    gdma::common::MessageMediator::getInstance().clearAll();
}

TEST(AttributeTest, int32Attribute) {
    int32_t val = 10;
    gdma::fte::Attribute a("name", "name", val);

    EXPECT_STREQ("name", a.getName().c_str());
    EXPECT_STREQ("name", a.getShortName().c_str());
    EXPECT_EQ(10, a.getInt32());
    EXPECT_EQ(te::dt::INT32_TYPE, a.getValueType());
}

TEST(AttributeTest, int32AttributeLargeShortName) {
    MockObserver obs;
    gdma::common::MessageMediator::getInstance().addObserver(&obs);

    std::string msg =
        "Feature short name 'nameTooLarge' is too long! It will be truncated to 'nameTooLar'!";
    EXPECT_CALL(obs, displayWarningMsg(msg)).Times(1);

    int32_t val = 10;
    gdma::fte::Attribute a("name", "nameTooLarge", val);

    EXPECT_STREQ("name", a.getName().c_str());
    EXPECT_STREQ("nameTooLar", a.getShortName().c_str());
    EXPECT_EQ(10, a.getInt32());
    EXPECT_EQ(te::dt::INT32_TYPE, a.getValueType());

    // @TODO(Raian): Verify how to destroy the singleton after each test and remove it.
    gdma::common::MessageMediator::getInstance().clearAll();
}

TEST(AttributeTest, int64Attribute) {
    int64_t val = 10;
    gdma::fte::Attribute a("name", "name", val);

    EXPECT_STREQ("name", a.getName().c_str());
    EXPECT_STREQ("name", a.getShortName().c_str());
    EXPECT_EQ(10, a.getInt64());
    EXPECT_EQ(te::dt::INT64_TYPE, a.getValueType());
}

TEST(AttributeTest, int64AttributeLargeShortName) {
    MockObserver obs;
    gdma::common::MessageMediator::getInstance().addObserver(&obs);

    std::string msg =
        "Feature short name 'nameTooLarge' is too long! It will be truncated to 'nameTooLar'!";
    EXPECT_CALL(obs, displayWarningMsg(msg)).Times(1);

    int64_t val = 10;
    gdma::fte::Attribute a("name", "nameTooLarge", val);

    EXPECT_STREQ("name", a.getName().c_str());
    EXPECT_STREQ("nameTooLar", a.getShortName().c_str());
    EXPECT_EQ(10, a.getInt64());
    EXPECT_EQ(te::dt::INT64_TYPE, a.getValueType());

    // @TODO(Raian): Verify how to destroy the singleton after each test and remove it.
    gdma::common::MessageMediator::getInstance().clearAll();
}

TEST(AttributeTest, floatAttribute) {
    float val = 10.54;
    gdma::fte::Attribute a("name", "name", val);

    EXPECT_STREQ("name", a.getName().c_str());
    EXPECT_STREQ("name", a.getShortName().c_str());
    EXPECT_FLOAT_EQ(10.54, a.getFloat());
    EXPECT_EQ(te::dt::FLOAT_TYPE, a.getValueType());
}

TEST(AttributeTest, floatAttributeLargeShortName) {
    MockObserver obs;
    gdma::common::MessageMediator::getInstance().addObserver(&obs);

    std::string msg =
        "Feature short name 'nameTooLarge' is too long! It will be truncated to 'nameTooLar'!";
    EXPECT_CALL(obs, displayWarningMsg(msg)).Times(1);

    float val = 10.54;
    gdma::fte::Attribute a("name", "nameTooLarge", val);

    EXPECT_STREQ("name", a.getName().c_str());
    EXPECT_STREQ("nameTooLar", a.getShortName().c_str());
    EXPECT_FLOAT_EQ(10.54, a.getFloat());
    EXPECT_EQ(te::dt::FLOAT_TYPE, a.getValueType());

    // @TODO(Raian): Verify how to destroy the singleton after each test and remove it.
    gdma::common::MessageMediator::getInstance().clearAll();
}

TEST(AttributeTest, doubleAttribute) {
    double val = 10.54;
    gdma::fte::Attribute a("name", "name", val);

    EXPECT_STREQ("name", a.getName().c_str());
    EXPECT_STREQ("name", a.getShortName().c_str());
    EXPECT_FLOAT_EQ(10.54, a.getDouble());
    EXPECT_EQ(te::dt::DOUBLE_TYPE, a.getValueType());
}

TEST(AttributeTest, doubleAttributeLargeShortName) {
    MockObserver obs;
    gdma::common::MessageMediator::getInstance().addObserver(&obs);

    std::string msg =
        "Feature short name 'nameTooLarge' is too long! It will be truncated to 'nameTooLar'!";
    EXPECT_CALL(obs, displayWarningMsg(msg)).Times(1);

    double val = 10.54;
    gdma::fte::Attribute a("name", "nameTooLarge", val);

    EXPECT_STREQ("name", a.getName().c_str());
    EXPECT_STREQ("nameTooLar", a.getShortName().c_str());
    EXPECT_FLOAT_EQ(10.54, a.getDouble());
    EXPECT_EQ(te::dt::DOUBLE_TYPE, a.getValueType());

    // @TODO(Raian): Verify how to destroy the singleton after each test and remove it.
    gdma::common::MessageMediator::getInstance().clearAll();
}

TEST(AttributeTest, int16AsDoubleAttribute) {
    int16_t val = 1;
    gdma::fte::Attribute a("name", "name", val);

    EXPECT_STREQ("name", a.getName().c_str());
    EXPECT_STREQ("name", a.getShortName().c_str());
    EXPECT_FLOAT_EQ(1.0, a.getAsDouble());
}

TEST(AttributeTest, int32AsDoubleAttribute) {
    int32_t val = 1;
    gdma::fte::Attribute a("name", "name", val);

    EXPECT_STREQ("name", a.getName().c_str());
    EXPECT_STREQ("name", a.getShortName().c_str());
    EXPECT_FLOAT_EQ(1.0, a.getAsDouble());
}

TEST(AttributeTest, int64AsDoubleAttribute) {
    int64_t val = 1;
    gdma::fte::Attribute a("name", "name", val);

    EXPECT_STREQ("name", a.getName().c_str());
    EXPECT_STREQ("name", a.getShortName().c_str());
    EXPECT_FLOAT_EQ(1.0, a.getAsDouble());
}

TEST(AttributeTest, floatAsDoubleAttribute) {
    float val = 1;
    gdma::fte::Attribute a("name", "name", val);

    EXPECT_STREQ("name", a.getName().c_str());
    EXPECT_STREQ("name", a.getShortName().c_str());
    EXPECT_FLOAT_EQ(1.0, a.getAsDouble());
}

TEST(AttributeTest, doubleAsDoubleAttribute) {
    double val = 1;
    gdma::fte::Attribute a("name", "name", val);

    EXPECT_STREQ("name", a.getName().c_str());
    EXPECT_STREQ("name", a.getShortName().c_str());
    EXPECT_FLOAT_EQ(1.0, a.getAsDouble());
}
