/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/featureextraction/intersectionCache_unittest.hpp

  \brief Functionalities to compute Landscape Ecology features.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// TerraLib includes
#include <terralib/geometry/Geometry.h>
#include <terralib/common/TerraLib.h>

// GTest includes
#include <gtest/gtest.h>

// Boost Includes
#include <boost/bimap.hpp>

// STL Includes
#include <map>
#include <memory>
#include <vector>
#include <string>

// GeoDMA includes
#include "BuildConfig.hpp"
#include "geodmalib/datamanager/dataManager.hpp"
#include "geodmalib/common/fileSystem.hpp"
#include "geodmalib/featureextraction/intersectionCache.hpp"
#include "geodmalib/common/utils.hpp"

class IntersectionCacheTest : public ::testing::Test {
 protected:
    // Polygons Layer
    boost::bimap<std::string, std::size_t> pols_positions;
    std::vector<te::gm::Geometry*> pols_container;
    // std::vector<std::string> pol_classes;
    // std::set<std::string> class_values;

    // cells Layer
    boost::bimap<std::string, std::size_t> cells_positions;
    std::vector<te::gm::Geometry*> cells_container;

    // Intersection cache
    gdma::fte::IntersectionCache *intersectionCache;

    virtual void SetUp() {
        try {
//             TerraLib::getInstance().initialize();
//             gdma::common::loadTerraLibDrivers();
        }
        catch (te::common::Exception &e) {
            std::cerr << "ERROR: " << e.what() << std::endl;
        }

        std::string pathShpPols = gdma::common::FindInGeoDMAPath(
            "data/tests/terraclass_tests/terraclass_tests.shp");
        std::string pathShpCells = gdma::common::FindInGeoDMAPath(
            "data/tests/terraclass_tests/terraclass_cells_tests.shp");

        std::string polsDataSource = gdma::dm::openShapefile(pathShpPols);
        std::string cellsDataSource = gdma::dm::openShapefile(pathShpCells);

        gdma::dm::getGeometriesFromDataSource(polsDataSource, pols_positions, pols_container);
        gdma::dm::getGeometriesFromDataSource(cellsDataSource, cells_positions, cells_container);

        intersectionCache = new gdma::fte::IntersectionCache(cells_container, cells_positions,
            pols_container, pols_positions);
    }

    virtual void TearDown() {
        delete intersectionCache;

//         gdma::common::unloadTerraLibDrivers();
//         TerraLib::getInstance().finalize();
    }

    std::size_t GetCellPosition(std::string cellId) {
        return cells_positions.left.find(cellId)->second;
    }
};


TEST_F(IntersectionCacheTest, searchForEmptyIntersection) {
    std::size_t cellPos = GetCellPosition("12");

    gdma::fte::MatrixLine intersections;
    EXPECT_TRUE(intersectionCache->searchForIntersections(cellPos, intersections));

    EXPECT_EQ(0, intersections.nnz());
}

TEST_F(IntersectionCacheTest, searchForUnemptyIntersections) {
    std::size_t cellPos = GetCellPosition("0");

    gdma::fte::MatrixLine intersections;
    EXPECT_TRUE(intersectionCache->searchForIntersections(cellPos, intersections));

    EXPECT_EQ(4, intersections.nnz());
    EXPECT_EQ(pols_container.size(), intersections.size());

    std::size_t cellPos_1 = GetCellPosition("1");
    EXPECT_TRUE(intersectionCache->searchForIntersections(cellPos_1, intersections));
    EXPECT_EQ(3, intersections.nnz());
    EXPECT_EQ(pols_container.size(), intersections.size());

    std::size_t cellPos_2 = GetCellPosition("2");
    EXPECT_TRUE(intersectionCache->searchForIntersections(cellPos_2, intersections));

    EXPECT_EQ(3, intersections.nnz());
    EXPECT_EQ(pols_container.size(), intersections.size());

    std::size_t cellPos_3 = GetCellPosition("3");
    EXPECT_TRUE(intersectionCache->searchForIntersections(cellPos_3, intersections));

    EXPECT_EQ(5, intersections.nnz());
    EXPECT_EQ(pols_container.size(), intersections.size());
}

TEST_F(IntersectionCacheTest, searchIntersectionsAccessGeometries) {
    std::size_t cellPos = GetCellPosition("1");

    gdma::fte::MatrixLine intersections;
    EXPECT_TRUE(intersectionCache->searchForIntersections(cellPos, intersections));

    boost::numeric::ublas::compressed_vector<te::gm::GeometryShrPtr>::const_iterator it
        = intersections.begin();

    while (it != intersections.end()) {
        EXPECT_TRUE((*it)->isValid());
        EXPECT_FALSE((*it)->isEmpty());
        it++;
    }
}

TEST_F(IntersectionCacheTest, intersectionInvalidPolygon) {
    std::string pathShpPols = gdma::common::FindInGeoDMAPath(
        "data/tests/terraclass_clip_invalid_pols.shp");
    std::string pathShpCells = gdma::common::FindInGeoDMAPath(
        "data/tests/terraclass_clip_invalid_pols_cells.shp");

    // Polygons Layer
    boost::bimap<std::string, std::size_t> pols_positions_1;
    std::vector<te::gm::Geometry*> pols_container_1;

    // cells Layer
    boost::bimap<std::string, std::size_t> cells_positions_1;
    std::vector<te::gm::Geometry*> cells_container_1;

    std::string dsPols = gdma::dm::openShapefile(pathShpPols);
    std::string dsCells = gdma::dm::openShapefile(pathShpCells);

    gdma::dm::getGeometriesFromDataSource(dsPols, pols_positions_1, pols_container_1);
    gdma::dm::getGeometriesFromDataSource(dsCells, cells_positions_1, cells_container_1);

    intersectionCache = new gdma::fte::IntersectionCache(cells_container_1, cells_positions_1,
        pols_container_1, pols_positions_1);

    gdma::fte::MatrixLine intersections;

    for (auto cellPos : cells_positions_1) {
        EXPECT_FALSE(intersectionCache->searchForIntersections(cellPos.right, intersections));
        EXPECT_EQ(0, intersections.nnz());
    }
}
