/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/featureextraction/landscapeFeatures_unittest.hpp

  \brief Functionalities to compute Landscape Ecology features.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// GTest includes
#include <gtest/gtest.h>
#include <gmock/gmock.h>

// TerraLib includes
#include <terralib/common/TerraLib.h>
#include <terralib/common/Exception.h>
#include <terralib/common/progress/ProgressManager.h>
#include <terralib/geometry.h>

// Boost Includes
#include <boost/bimap.hpp>

// STL Includes
#include <map>
#include <memory>
#include <set>
#include <string>
#include <utility>  // pair
#include <vector>

// GeoDMA includes
#include "BuildConfig.hpp"
// #include "geodmalib/common/commandLineProgViewer.hpp"
#include "geodmalib/common/fileSystem.hpp"
#include "geodmalib/common/observerProgressBar.hpp"
#include "geodmalib/common/utils.hpp"
#include "geodmalib/datamanager/dataManager.hpp"
#include "geodmalib/featureextraction/landscapeFeatures.hpp"

// using ::testing::AtLeast;

// @TODO(Raian): Verify how to make it work (problem- parameters taskId)
// class MockProgressViewer : public gdma::common::CommandLineProgViewer {
// public:
//    MockProgressViewer() : CommandLineProgViewer(100) {}
//    MOCK_METHOD1(updateValue, void(int taskId));
// };

template <typename T>
void GetMapFromSet(std::set<std::string> setVals, std::map<std::string, T> &map_values) {
    for (auto &val : setVals) {
        map_values.insert(std::pair<std::string, T>(val, (T)0.0));
    }
}

te::gm::Polygon* GetCell(te::gm::Geometry* multiGeom) {
    te::gm::MultiPolygon* multipol = static_cast<te::gm::MultiPolygon*>(multiGeom);
    te::gm::Polygon* cell = static_cast<te::gm::Polygon*>(multipol->getGeometryN(0));
    return cell;
}

class LandscapeFeaturesTest : public ::testing::Test {
 protected:
    // Polygons Layer
    boost::bimap<std::string, std::size_t> pols_positions;
    std::vector<te::gm::Geometry*> pols_container;
    std::vector<std::string> pol_classes;
    std::set<std::string> class_values;

    // cells Layer
    boost::bimap<std::string, std::size_t> cells_positions;
    std::vector<te::gm::Geometry*> cells_container;

    gdma::fte::LandscapeFeatures* ldscFeat;

    // std::unique_ptr<MockProgressViewer> progViewer;

    virtual void SetUp() {
        try {
//             TerraLib::getInstance().initialize();
//             gdma::common::loadTerraLibDrivers();
        }
        catch (te::common::Exception &e) {
            std::cerr << "ERROR: " << e.what() << std::endl;
        }

        std::string pathShpPols = gdma::common::FindInGeoDMAPath(
            "data/tests/terraclass_tests/terraclass_tests.shp");
        std::string pathShpCells = gdma::common::FindInGeoDMAPath(
            "data/tests/terraclass_tests/terraclass_cells_tests.shp");

        // progViewer.reset(new MockProgressViewer());
        // int id = te::common::ProgressManager::getInstance().addViewer(progViewer.get());

        std::string dsPols = gdma::dm::openShapefile(pathShpPols);
        std::string dsCells = gdma::dm::openShapefile(pathShpCells);

        gdma::dm::getGeometriesFromDataSource(dsPols, pols_positions, pols_container, "tcclasse",
            &pol_classes, &class_values);
        // EXPECT_CALL(*progViewer.get(), updateValue(5)).Times(pols_positions.size());

        gdma::dm::getGeometriesFromDataSource(dsCells, cells_positions, cells_container);

        ldscFeat = new gdma::fte::LandscapeFeatures(cells_container, cells_positions,
            pols_container, pols_positions, pol_classes, class_values);
    }

    virtual void TearDown() {
        delete ldscFeat;
//         gdma::common::unloadTerraLibDrivers();
//         TerraLib::getInstance().finalize();
    }

    std::size_t GetCellPosition(std::string cellId) {
        return cells_positions.left.find(cellId)->second;
    }
};

/*!
  \class gdma::fte::LandscapeFeatures
  \test Test for gdma::fte::LandscapeFeatures::getClassArea method. This test tests the most
        basic case, when there is no intersections between the finnest layer of geometries and the
        given geometry of the coarser layer.
*/
TEST_F(LandscapeFeaturesTest, getClassAreaEmptyIntersection) {
    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);

    std::size_t cellPos = GetCellPosition("12");
    ldscFeat->getClassArea(cellPos, class_areas);

    EXPECT_EQ(5, class_areas.size());
    std::map<std::string, double>::iterator itMap = class_areas.begin();
    while (itMap != class_areas.end()) {
        EXPECT_DOUBLE_EQ(0.0, itMap->second);
        itMap++;
    }
}

TEST_F(LandscapeFeaturesTest, getClassAreaIntersection) {
    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);

    std::size_t cellPos = GetCellPosition("0");
    ldscFeat->getClassArea(cellPos, class_areas);

    EXPECT_EQ(5, class_areas.size());
    EXPECT_DOUBLE_EQ(740.20873937191857, class_areas["FLORESTA"]);
    EXPECT_DOUBLE_EQ(858.58123024330064, class_areas["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, class_areas["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, class_areas["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(0.0, class_areas["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getClassAreaIntersectionMultiPolygon) {
    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);

    std::size_t cellPos = GetCellPosition("10");
    ldscFeat->getClassArea(cellPos, class_areas);

    EXPECT_EQ(5, class_areas.size());
    EXPECT_DOUBLE_EQ(0.0, class_areas["FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, class_areas["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, class_areas["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, class_areas["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(84.376107492069778, class_areas["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getPercentLandEmptyIntersection) {
    std::size_t cellPos = GetCellPosition("12");

    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);

    ldscFeat->getClassArea(cellPos, class_areas);

    std::map<std::string, double> percent_lands;
    ldscFeat->getPercentLand(cellPos, class_areas, percent_lands);

    EXPECT_EQ(5, percent_lands.size());
    std::map<std::string, double>::iterator itMap = percent_lands.begin();
    while (itMap != percent_lands.end()) {
        EXPECT_DOUBLE_EQ(0.0, itMap->second);
        itMap++;
    }
}

TEST_F(LandscapeFeaturesTest, getPercentLandIntersection) {
    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);

    std::size_t cellPos = GetCellPosition("0");

    ldscFeat->getClassArea(cellPos, class_areas);

    std::map<std::string, double> percent_lands;
    ldscFeat->getPercentLand(cellPos, class_areas, percent_lands);

    EXPECT_EQ(5, percent_lands.size());
    EXPECT_DOUBLE_EQ(46.263046210744903, percent_lands["FLORESTA"]);
    EXPECT_DOUBLE_EQ(53.661326890206297, percent_lands["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, percent_lands["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, percent_lands["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(0.0, percent_lands["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getPercentLandIntersectionMultiPolygon) {
    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);

    std::size_t cellPos = GetCellPosition("10");
    ldscFeat->getClassArea(cellPos, class_areas);

    std::map<std::string, double> percent_lands;
    ldscFeat->getPercentLand(cellPos, class_areas, percent_lands);

    EXPECT_EQ(5, percent_lands.size());
    EXPECT_DOUBLE_EQ(0.0, percent_lands["FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, percent_lands["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, percent_lands["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, percent_lands["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(5.2735067182543611, percent_lands["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getPatchDensityEmptyIntersection) {
    std::map<std::string, double> patch_densities;
    GetMapFromSet(class_values, patch_densities);

    std::size_t cellPos = GetCellPosition("12");
    ldscFeat->getPatchDensities(cellPos, patch_densities);

    EXPECT_EQ(5, patch_densities.size());
    std::map<std::string, double>::iterator itMap = patch_densities.begin();
    while (itMap != patch_densities.end()) {
        EXPECT_DOUBLE_EQ(0.0, itMap->second);
        itMap++;
    }
}

TEST_F(LandscapeFeaturesTest, getPatchDensityIntersection) {
    std::map<std::string, double> patch_densities;
    GetMapFromSet(class_values, patch_densities);

    std::size_t cellPos = GetCellPosition("0");
    ldscFeat->getPatchDensities(cellPos, patch_densities);

    EXPECT_EQ(5, patch_densities.size());
    EXPECT_DOUBLE_EQ(1.25, patch_densities["FLORESTA"]);
    EXPECT_DOUBLE_EQ(1.25, patch_densities["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, patch_densities["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, patch_densities["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(0.0, patch_densities["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getPatchDensityIntersectionMultiPolygon) {
    std::map<std::string, double> patch_densities;
    GetMapFromSet(class_values, patch_densities);

    std::size_t cellPos = GetCellPosition("10");
    ldscFeat->getPatchDensities(cellPos, patch_densities);

    EXPECT_EQ(5, patch_densities.size());
    EXPECT_DOUBLE_EQ(0.0, patch_densities["FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, patch_densities["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, patch_densities["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, patch_densities["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(0.625, patch_densities["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getMeanPatchSizeEmptyIntersection) {
    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);

    std::size_t cellPos = GetCellPosition("12");
    ldscFeat->getNumberOfPatches(cellPos, numberOfPatches);

    std::map<std::string, double> mean_patch_sizes;
    GetMapFromSet(class_values, mean_patch_sizes);

    ldscFeat->getMeanPatchSizes(cellPos, numberOfPatches,
        mean_patch_sizes);

    EXPECT_EQ(5, mean_patch_sizes.size());
    std::map<std::string, double>::iterator itMap = mean_patch_sizes.begin();
    while (itMap != mean_patch_sizes.end()) {
        EXPECT_DOUBLE_EQ(0.0, itMap->second);
        itMap++;
    }
}

TEST_F(LandscapeFeaturesTest, getMeanPatchSizeIntersection) {
    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);

    std::size_t cellPos = GetCellPosition("0");

    ldscFeat->getNumberOfPatches(cellPos, numberOfPatches);

    std::map<std::string, double> mean_patch_sizes;
    GetMapFromSet(class_values, mean_patch_sizes);

    ldscFeat->getMeanPatchSizes(cellPos, numberOfPatches,
        mean_patch_sizes);

    EXPECT_EQ(5, mean_patch_sizes.size());
    EXPECT_DOUBLE_EQ(370.10436968595928, mean_patch_sizes["FLORESTA"]);
    EXPECT_DOUBLE_EQ(429.29061512165032, mean_patch_sizes["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, mean_patch_sizes["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, mean_patch_sizes["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(0.0, mean_patch_sizes["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getMeanPatchSizeIntersectionMultiPolygon) {
    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);

    std::size_t cellPos = GetCellPosition("10");

    ldscFeat->getNumberOfPatches(cellPos, numberOfPatches);

    std::map<std::string, double> mean_patch_sizes;
    GetMapFromSet(class_values, mean_patch_sizes);

    ldscFeat->getMeanPatchSizes(cellPos, numberOfPatches,
        mean_patch_sizes);

    EXPECT_EQ(5, mean_patch_sizes.size());
    EXPECT_DOUBLE_EQ(0.0, mean_patch_sizes["FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, mean_patch_sizes["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, mean_patch_sizes["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, mean_patch_sizes["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(84.376107492069778, mean_patch_sizes["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getPatchSizesStdEmptyIntersection) {
    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);

    std::size_t cellPos = GetCellPosition("12");

    ldscFeat->getNumberOfPatches(cellPos, numberOfPatches);

    std::map<std::string, double> mean_patch_sizes;
    GetMapFromSet(class_values, mean_patch_sizes);

    ldscFeat->getMeanPatchSizes(cellPos, numberOfPatches,
        mean_patch_sizes);

    std::map<std::string, double> patch_sizes_std;
    GetMapFromSet(class_values, patch_sizes_std);

    ldscFeat->getPatchSizesStd(cellPos, mean_patch_sizes,
        numberOfPatches, patch_sizes_std);

    EXPECT_EQ(5, patch_sizes_std.size());
    std::map<std::string, double>::iterator itMap = patch_sizes_std.begin();
    while (itMap != patch_sizes_std.end()) {
        EXPECT_DOUBLE_EQ(0.0, itMap->second);
        itMap++;
    }
}

TEST_F(LandscapeFeaturesTest, getPatchSizesStdIntersection) {
    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);

    std::size_t cellPos = GetCellPosition("0");

    ldscFeat->getNumberOfPatches(cellPos, numberOfPatches);

    std::map<std::string, double> mean_patch_sizes;
    GetMapFromSet(class_values, mean_patch_sizes);

    ldscFeat->getMeanPatchSizes(cellPos, numberOfPatches,
        mean_patch_sizes);

    std::map<std::string, double> patch_sizes_std;
    GetMapFromSet(class_values, patch_sizes_std);

    ldscFeat->getPatchSizesStd(cellPos, mean_patch_sizes,
        numberOfPatches, patch_sizes_std);

    EXPECT_EQ(5, patch_sizes_std.size());
    EXPECT_DOUBLE_EQ(136.01894341986178, patch_sizes_std["FLORESTA"]);
    EXPECT_DOUBLE_EQ(409.38178540977771, patch_sizes_std["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, patch_sizes_std["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, patch_sizes_std["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(0.0, patch_sizes_std["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getPatchSizesStdIntersectionMultiPolygon) {
    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);

    std::size_t cellPos = GetCellPosition("10");

    ldscFeat->getNumberOfPatches(cellPos, numberOfPatches);

    std::map<std::string, double> mean_patch_sizes;
    GetMapFromSet(class_values, mean_patch_sizes);

    ldscFeat->getMeanPatchSizes(cellPos, numberOfPatches,
        mean_patch_sizes);

    std::map<std::string, double> patch_sizes_std;
    GetMapFromSet(class_values, patch_sizes_std);

    ldscFeat->getPatchSizesStd(cellPos, mean_patch_sizes,
        numberOfPatches, patch_sizes_std);

    EXPECT_EQ(5, patch_sizes_std.size());
    EXPECT_DOUBLE_EQ(0.0, patch_sizes_std["FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, patch_sizes_std["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, patch_sizes_std["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, patch_sizes_std["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(0.0, patch_sizes_std["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getLandscapeShapeIndexesEmptyIntersection) {
    std::map<std::string, double> landscapeShapeIndexes;
    GetMapFromSet(class_values, landscapeShapeIndexes);

    std::size_t cellPos = GetCellPosition("12");

    ldscFeat->getLandscapeShapeIndexes(cellPos, landscapeShapeIndexes);

    EXPECT_EQ(5, landscapeShapeIndexes.size());
    std::map<std::string, double>::iterator itMap = landscapeShapeIndexes.begin();
    while (itMap != landscapeShapeIndexes.end()) {
        EXPECT_DOUBLE_EQ(0.0, itMap->second);
        itMap++;
    }
}

TEST_F(LandscapeFeaturesTest, getLandscapeShapeIndexesIntersection) {
    std::map<std::string, double> landscapeShapeIndexes;
    GetMapFromSet(class_values, landscapeShapeIndexes);

    std::size_t cellPos = GetCellPosition("0");

    ldscFeat->getLandscapeShapeIndexes(cellPos, landscapeShapeIndexes);

    EXPECT_EQ(5, landscapeShapeIndexes.size());
    EXPECT_DOUBLE_EQ(1.7115831339902852, landscapeShapeIndexes["FLORESTA"]);
    EXPECT_DOUBLE_EQ(1.8720488764905925, landscapeShapeIndexes["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, landscapeShapeIndexes["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, landscapeShapeIndexes["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(0.0, landscapeShapeIndexes["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getLandscapeShapeIndexesIntersectionMultipolygon) {
    std::map<std::string, double> landscapeShapeIndexes;
    GetMapFromSet(class_values, landscapeShapeIndexes);

    std::size_t cellPos = GetCellPosition("10");

    ldscFeat->getLandscapeShapeIndexes(cellPos, landscapeShapeIndexes);

    EXPECT_EQ(5, landscapeShapeIndexes.size());
    EXPECT_DOUBLE_EQ(0.0, landscapeShapeIndexes["FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, landscapeShapeIndexes["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, landscapeShapeIndexes["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, landscapeShapeIndexes["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(0.49044228157823333, landscapeShapeIndexes["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getMeanShapeIndexesEmptyIntersection) {
    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);

    std::size_t cellPos = GetCellPosition("12");

    ldscFeat->getNumberOfPatches(cellPos, numberOfPatches);

    std::map<std::string, double> meanShapeIndexes;
    GetMapFromSet(class_values, meanShapeIndexes);

    ldscFeat->getMeanShapeIndexes(cellPos, numberOfPatches, meanShapeIndexes);

    EXPECT_EQ(5, meanShapeIndexes.size());
    std::map<std::string, double>::iterator itMap = meanShapeIndexes.begin();
    while (itMap != meanShapeIndexes.end()) {
        EXPECT_DOUBLE_EQ(0.0, itMap->second);
        itMap++;
    }
}

TEST_F(LandscapeFeaturesTest, getMeanShapeIndexesIntersection) {
    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);

    std::size_t cellPos = GetCellPosition("0");

    ldscFeat->getNumberOfPatches(cellPos, numberOfPatches);

    std::map<std::string, double> meanShapeIndexes;
    GetMapFromSet(class_values, meanShapeIndexes);

    ldscFeat->getMeanShapeIndexes(cellPos, numberOfPatches,
        meanShapeIndexes);

    EXPECT_EQ(5, meanShapeIndexes.size());
    EXPECT_DOUBLE_EQ(1.8309635497322037, meanShapeIndexes["FLORESTA"]);
    EXPECT_DOUBLE_EQ(1.9944150183041602, meanShapeIndexes["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, meanShapeIndexes["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, meanShapeIndexes["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(0.0, meanShapeIndexes["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getMeanShapeIndexesIntersectionMultipolygon) {
    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);

    std::size_t cellPos = GetCellPosition("10");

    ldscFeat->getNumberOfPatches(cellPos, numberOfPatches);

    std::map<std::string, double> meanShapeIndexes;
    GetMapFromSet(class_values, meanShapeIndexes);

    ldscFeat->getMeanShapeIndexes(cellPos, numberOfPatches,
        meanShapeIndexes);

    EXPECT_EQ(5, meanShapeIndexes.size());
    EXPECT_DOUBLE_EQ(0.0, meanShapeIndexes["FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, meanShapeIndexes["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, meanShapeIndexes["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, meanShapeIndexes["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(2.1356896959660929, meanShapeIndexes["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getAreaWeightedMeanShapeIndexesEmptyIntersection) {
    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);

    std::size_t cellPos = GetCellPosition("12");

    ldscFeat->getClassArea(cellPos, class_areas);

    std::map<std::string, double> awMeanShapeIndexes;
    GetMapFromSet(class_values, awMeanShapeIndexes);

    ldscFeat->getAWMeanShapeIndexes(cellPos, class_areas, awMeanShapeIndexes);

    EXPECT_EQ(5, awMeanShapeIndexes.size());
    std::map<std::string, double>::iterator itMap = awMeanShapeIndexes.begin();
    while (itMap != awMeanShapeIndexes.end()) {
        EXPECT_DOUBLE_EQ(0.0, itMap->second);
        itMap++;
    }
}

TEST_F(LandscapeFeaturesTest, getAreaWeightedMeanShapeIndexesIntersection) {
    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);

    std::size_t cellPos = GetCellPosition("0");

    ldscFeat->getClassArea(cellPos, class_areas);

    std::map<std::string, double> awMeanShapeIndexes;
    GetMapFromSet(class_values, awMeanShapeIndexes);

    ldscFeat->getAWMeanShapeIndexes(cellPos, class_areas, awMeanShapeIndexes);

    EXPECT_EQ(5, awMeanShapeIndexes.size());
    EXPECT_DOUBLE_EQ(1.7930919917659638, awMeanShapeIndexes["FLORESTA"]);
    EXPECT_DOUBLE_EQ(2.3145902329450134, awMeanShapeIndexes["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, awMeanShapeIndexes["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, awMeanShapeIndexes["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(0.0, awMeanShapeIndexes["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getAreaWeightedMeanShapeIndexesIntersectionMultipolygon) {
    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);

    std::size_t cellPos = GetCellPosition("10");

    ldscFeat->getClassArea(cellPos, class_areas);

    std::map<std::string, double> awMeanShapeIndexes;
    GetMapFromSet(class_values, awMeanShapeIndexes);

    ldscFeat->getAWMeanShapeIndexes(cellPos, class_areas, awMeanShapeIndexes);

    EXPECT_EQ(5, awMeanShapeIndexes.size());
    EXPECT_DOUBLE_EQ(0.0, awMeanShapeIndexes["FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, awMeanShapeIndexes["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, awMeanShapeIndexes["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, awMeanShapeIndexes["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(2.1356896959660929, awMeanShapeIndexes["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getMeanPatchFractalDimensionsEmptyIntersection) {
    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);

    std::size_t cellPos = GetCellPosition("12");

    ldscFeat->getNumberOfPatches(cellPos, numberOfPatches);

    std::map<std::string, double> mpfds;
    GetMapFromSet(class_values, mpfds);

    ldscFeat->getMeanPatchFractalDimensions(cellPos, numberOfPatches, mpfds);

    EXPECT_EQ(5, mpfds.size());
    std::map<std::string, double>::iterator itMap = mpfds.begin();
    while (itMap != mpfds.end()) {
        EXPECT_DOUBLE_EQ(0.0, itMap->second);
        itMap++;
    }
}

TEST_F(LandscapeFeaturesTest, getMeanPatchFractalDimensionsIntersection) {
    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);

    std::size_t cellPos = GetCellPosition("0");

    ldscFeat->getNumberOfPatches(cellPos, numberOfPatches);

    std::map<std::string, double> mpfds;
    GetMapFromSet(class_values, mpfds);

    ldscFeat->getMeanPatchFractalDimensions(cellPos, numberOfPatches, mpfds);

    EXPECT_EQ(5, mpfds.size());
    EXPECT_DOUBLE_EQ(1.2486702498668161, mpfds["FLORESTA"]);
    EXPECT_DOUBLE_EQ(1.2776336656266576, mpfds["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, mpfds["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, mpfds["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(0.0, mpfds["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getMeanPatchFractalDimensionsIntersectionMultipolygon) {
    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);

    std::size_t cellPos = GetCellPosition("10");

    ldscFeat->getNumberOfPatches(cellPos, numberOfPatches);

    std::map<std::string, double> mpfds;
    GetMapFromSet(class_values, mpfds);

    ldscFeat->getMeanPatchFractalDimensions(cellPos, numberOfPatches, mpfds);

    EXPECT_EQ(5, mpfds.size());
    EXPECT_DOUBLE_EQ(0.0, mpfds["FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, mpfds["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, mpfds["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, mpfds["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(1.2966960931411782, mpfds["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getAWMPFDEmptyIntersection) {
    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);

    std::size_t cellPos = GetCellPosition("12");

    ldscFeat->getClassArea(cellPos, class_areas);

    std::map<std::string, double> awmpfds;
    GetMapFromSet(class_values, awmpfds);
    ldscFeat->getAreaWeightedMeanPatchFractalDimensions(cellPos, class_areas, awmpfds);

    EXPECT_EQ(5, awmpfds.size());
    std::map<std::string, double>::iterator itMap = awmpfds.begin();
    while (itMap != awmpfds.end()) {
        EXPECT_DOUBLE_EQ(0.0, itMap->second);
        itMap++;
    }
}

TEST_F(LandscapeFeaturesTest, getAWMPFDIntersection) {
    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);

    std::size_t cellPos = GetCellPosition("0");

    ldscFeat->getClassArea(cellPos, class_areas);

    std::map<std::string, double> awmpfds;
    GetMapFromSet(class_values, awmpfds);
    ldscFeat->getAreaWeightedMeanPatchFractalDimensions(cellPos, class_areas, awmpfds);

    EXPECT_EQ(5, awmpfds.size());
    EXPECT_DOUBLE_EQ(1.2435779691422959, awmpfds["FLORESTA"]);
    EXPECT_DOUBLE_EQ(1.2654798001362555, awmpfds["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, awmpfds["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, awmpfds["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(0.0, awmpfds["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getAWMPFDIntersectionMultipolygon) {
    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);

    std::size_t cellPos = GetCellPosition("10");

    ldscFeat->getClassArea(cellPos, class_areas);

    std::map<std::string, double> awmpfds;
    GetMapFromSet(class_values, awmpfds);

    ldscFeat->getAreaWeightedMeanPatchFractalDimensions(cellPos, class_areas, awmpfds);

    EXPECT_EQ(5, awmpfds.size());
    EXPECT_DOUBLE_EQ(0.0, awmpfds["FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, awmpfds["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, awmpfds["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, awmpfds["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(1.2966960931411782, awmpfds["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getEdgeDensitiesEmptyIntersection) {
    std::map<std::string, double> class_perimeters;
    GetMapFromSet(class_values, class_perimeters);

    std::size_t cellPos = GetCellPosition("12");

    ldscFeat->getTotalEdges(cellPos, class_perimeters);

    std::map<std::string, double> edgeDensities;
    GetMapFromSet(class_values, edgeDensities);
    ldscFeat->getEdgeDensities(cellPos, class_perimeters, edgeDensities);

    EXPECT_EQ(5, edgeDensities.size());
    std::map<std::string, double>::iterator itMap = edgeDensities.begin();
    while (itMap != edgeDensities.end()) {
        EXPECT_DOUBLE_EQ(0.0, itMap->second);
        itMap++;
    }
}

TEST_F(LandscapeFeaturesTest, getEdgeDensitiesIntersection) {
    std::map<std::string, double> class_perimeters;
    GetMapFromSet(class_values, class_perimeters);

    std::size_t cellPos = GetCellPosition("0");

    ldscFeat->getTotalEdges(cellPos, class_perimeters);

    std::map<std::string, double> edgeDensities;
    GetMapFromSet(class_values, edgeDensities);
    ldscFeat->getEdgeDensities(cellPos, class_perimeters, edgeDensities);

    EXPECT_EQ(5, edgeDensities.size());
    EXPECT_DOUBLE_EQ(15.168510584930065, edgeDensities["FLORESTA"]);
    EXPECT_DOUBLE_EQ(16.590601201095478, edgeDensities["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, edgeDensities["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, edgeDensities["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(0.0, edgeDensities["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getEdgeDensitiesIntersectionMultipolygon) {
    std::map<std::string, double> class_perimeters;
    GetMapFromSet(class_values, class_perimeters);

    std::size_t cellPos = GetCellPosition("10");

    ldscFeat->getTotalEdges(cellPos, class_perimeters);

    std::map<std::string, double> edgeDensities;
    GetMapFromSet(class_values, edgeDensities);
    ldscFeat->getEdgeDensities(cellPos, class_perimeters, edgeDensities);

    EXPECT_EQ(5, edgeDensities.size());
    EXPECT_DOUBLE_EQ(0.0, edgeDensities["FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, edgeDensities["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, edgeDensities["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, edgeDensities["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(4.3464315531511355, edgeDensities["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getTotalEdgesEmptyIntersection) {
    std::map<std::string, double> class_perimeters;
    GetMapFromSet(class_values, class_perimeters);

    std::size_t cellPos = GetCellPosition("12");

    ldscFeat->getTotalEdges(cellPos, class_perimeters);

    EXPECT_EQ(5, class_perimeters.size());
    std::map<std::string, double>::iterator itMap = class_perimeters.begin();
    while (itMap != class_perimeters.end()) {
        EXPECT_DOUBLE_EQ(0.0, itMap->second);
        itMap++;
    }
}

TEST_F(LandscapeFeaturesTest, getTotalEdgesIntersection) {
    std::map<std::string, double> class_perimeters;
    GetMapFromSet(class_values, class_perimeters);

    std::size_t cellPos = GetCellPosition("0");

    ldscFeat->getTotalEdges(cellPos, class_perimeters);

    EXPECT_EQ(5, class_perimeters.size());
    EXPECT_DOUBLE_EQ(24269.616935888102, class_perimeters["FLORESTA"]);
    EXPECT_DOUBLE_EQ(26544.961921752762, class_perimeters["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, class_perimeters["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, class_perimeters["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(0.0, class_perimeters["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getTotalEdgesIntersectionMultiPolygon) {
    std::map<std::string, double> class_perimeters;
    GetMapFromSet(class_values, class_perimeters);

    std::size_t cellPos = GetCellPosition("10");

    ldscFeat->getTotalEdges(cellPos, class_perimeters);

    EXPECT_EQ(5, class_perimeters.size());
    EXPECT_DOUBLE_EQ(0.0, class_perimeters["FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, class_perimeters["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, class_perimeters["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, class_perimeters["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(6954.2904850418163, class_perimeters["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getMeanPerimeterAreaRatiosEmptyIntersection) {
    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);

    std::size_t cellPos = GetCellPosition("12");

    ldscFeat->getNumberOfPatches(cellPos, numberOfPatches);

    std::map<std::string, double> meanPerimAreaRatio;
    GetMapFromSet(class_values, meanPerimAreaRatio);
    ldscFeat->getMeanPerimeterAreaRatios(cellPos, numberOfPatches, meanPerimAreaRatio);

    EXPECT_EQ(5, meanPerimAreaRatio.size());
    std::map<std::string, double>::iterator itMap = meanPerimAreaRatio.begin();
    while (itMap != meanPerimAreaRatio.end()) {
        EXPECT_DOUBLE_EQ(0.0, itMap->second);
        itMap++;
    }
}

TEST_F(LandscapeFeaturesTest, getMeanPerimeterAreaRatiosIntersection) {
    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);

    std::size_t cellPos = GetCellPosition("0");

    ldscFeat->getNumberOfPatches(cellPos, numberOfPatches);

    std::map<std::string, double> meanPerimAreaRatio;
    GetMapFromSet(class_values, meanPerimAreaRatio);
    ldscFeat->getMeanPerimeterAreaRatios(cellPos, numberOfPatches, meanPerimAreaRatio);

    EXPECT_EQ(5, meanPerimAreaRatio.size());
    EXPECT_DOUBLE_EQ(0.0036018578515944998, meanPerimAreaRatio["FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0080150321003293439, meanPerimAreaRatio["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, meanPerimAreaRatio["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, meanPerimAreaRatio["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(0.0, meanPerimAreaRatio["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getMeanPerimeterAreaRatiosIntersectionMultipolygon) {
    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);

    std::size_t cellPos = GetCellPosition("10");

    ldscFeat->getNumberOfPatches(cellPos, numberOfPatches);

    std::map<std::string, double> meanPerimAreaRatio;
    GetMapFromSet(class_values, meanPerimAreaRatio);
    ldscFeat->getMeanPerimeterAreaRatios(cellPos, numberOfPatches, meanPerimAreaRatio);

    EXPECT_EQ(5, meanPerimAreaRatio.size());
    EXPECT_DOUBLE_EQ(0.0, meanPerimAreaRatio["FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, meanPerimAreaRatio["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, meanPerimAreaRatio["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, meanPerimAreaRatio["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(0.0082420138730569283, meanPerimAreaRatio["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getNumberOfPatchesEmptyIntersection) {
    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);

    std::size_t cellPos = GetCellPosition("12");

    ldscFeat->getNumberOfPatches(cellPos, numberOfPatches);

    EXPECT_EQ(5, numberOfPatches.size());
    std::map<std::string, int>::iterator itMap = numberOfPatches.begin();
    while (itMap != numberOfPatches.end()) {
        EXPECT_DOUBLE_EQ(0.0, itMap->second);
        itMap++;
    }
}

TEST_F(LandscapeFeaturesTest, getNumberOfPatchesIntersection) {
    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);

    std::size_t cellPos = GetCellPosition("0");

    ldscFeat->getNumberOfPatches(cellPos, numberOfPatches);

    EXPECT_EQ(5, numberOfPatches.size());
    EXPECT_EQ(2, numberOfPatches["FLORESTA"]);
    EXPECT_EQ(2, numberOfPatches["NAO_FLORESTA"]);
    EXPECT_EQ(0, numberOfPatches["AREA_NAO_OBSERVADA"]);
    EXPECT_EQ(0, numberOfPatches["FLORESTA_SOB_NUVEM"]);
    EXPECT_EQ(0, numberOfPatches["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getNumberOfPatchesIntersectionMultipolygon) {
    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);

    std::size_t cellPos = GetCellPosition("10");

    ldscFeat->getNumberOfPatches(cellPos, numberOfPatches);

    EXPECT_EQ(5, numberOfPatches.size());
    EXPECT_EQ(0, numberOfPatches["FLORESTA"]);
    EXPECT_EQ(0, numberOfPatches["NAO_FLORESTA"]);
    EXPECT_EQ(0, numberOfPatches["AREA_NAO_OBSERVADA"]);
    EXPECT_EQ(0, numberOfPatches["FLORESTA_SOB_NUVEM"]);
    EXPECT_EQ(1, numberOfPatches["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getPatchSizeCoeffVariationEmptyIntersection) {
    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);

    std::size_t cellPos = GetCellPosition("12");

    ldscFeat->getNumberOfPatches(cellPos, numberOfPatches);

    std::map<std::string, double> mean_patch_sizes;
    GetMapFromSet(class_values, mean_patch_sizes);
    ldscFeat->getMeanPatchSizes(cellPos, numberOfPatches, mean_patch_sizes);

    std::map<std::string, double> patch_sizes_std;
    GetMapFromSet(class_values, patch_sizes_std);
    ldscFeat->getPatchSizesStd(cellPos, mean_patch_sizes, numberOfPatches, patch_sizes_std);

    std::map<std::string, double> psCoeffVariation;
    GetMapFromSet(class_values, psCoeffVariation);
    ldscFeat->getPatchSizeCoeffVariation(mean_patch_sizes, patch_sizes_std, psCoeffVariation);

    EXPECT_EQ(5, psCoeffVariation.size());
    std::map<std::string, double>::iterator itMap = psCoeffVariation.begin();
    while (itMap != psCoeffVariation.end()) {
        EXPECT_DOUBLE_EQ(0.0, itMap->second);
        itMap++;
    }
}

TEST_F(LandscapeFeaturesTest, getPatchSizeCoeffVariationIntersection) {
    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);

    std::size_t cellPos = GetCellPosition("0");

    ldscFeat->getNumberOfPatches(cellPos, numberOfPatches);

    std::map<std::string, double> mean_patch_sizes;
    GetMapFromSet(class_values, mean_patch_sizes);
    ldscFeat->getMeanPatchSizes(cellPos, numberOfPatches,
        mean_patch_sizes);

    std::map<std::string, double> patch_sizes_std;
    GetMapFromSet(class_values, patch_sizes_std);
    ldscFeat->getPatchSizesStd(cellPos, mean_patch_sizes, numberOfPatches, patch_sizes_std);

    std::map<std::string, double> psCoeffVariation;
    GetMapFromSet(class_values, psCoeffVariation);
    ldscFeat->getPatchSizeCoeffVariation(mean_patch_sizes, patch_sizes_std, psCoeffVariation);

    EXPECT_EQ(5, psCoeffVariation.size());
    EXPECT_DOUBLE_EQ(36.751509725561064, psCoeffVariation["FLORESTA"]);
    EXPECT_DOUBLE_EQ(95.36238878498871, psCoeffVariation["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, psCoeffVariation["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, psCoeffVariation["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(0.0, psCoeffVariation["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getPatchSizeCoeffVariationIntersectionMultipolygon) {
    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);

    std::size_t cellPos = GetCellPosition("10");

    ldscFeat->getNumberOfPatches(cellPos, numberOfPatches);

    std::map<std::string, double> mean_patch_sizes;
    GetMapFromSet(class_values, mean_patch_sizes);
    ldscFeat->getMeanPatchSizes(cellPos, numberOfPatches, mean_patch_sizes);

    std::map<std::string, double> patch_sizes_std;
    GetMapFromSet(class_values, patch_sizes_std);
    ldscFeat->getPatchSizesStd(cellPos, mean_patch_sizes, numberOfPatches, patch_sizes_std);

    std::map<std::string, double> psCoeffVariation;
    GetMapFromSet(class_values, psCoeffVariation);
    ldscFeat->getPatchSizeCoeffVariation(mean_patch_sizes, patch_sizes_std, psCoeffVariation);

    EXPECT_EQ(5, psCoeffVariation.size());
    EXPECT_DOUBLE_EQ(0.0, psCoeffVariation["FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, psCoeffVariation["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, psCoeffVariation["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, psCoeffVariation["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(0.0, psCoeffVariation["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getInterspersionAndJuxtapositionIndexEmptyIntersection) {
    std::size_t cellPos = GetCellPosition("12");

    std::map<std::string, double> total_edges;
    GetMapFromSet(class_values, total_edges);
    ldscFeat->getTotalEdges(cellPos, total_edges);

    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);
    ldscFeat->getNumberOfPatches(cellPos, numberOfPatches);

    int patchRichness = ldscFeat->getPatchRichness(cellPos);

    std::map<std::string, double> ijIndexes;
    GetMapFromSet(class_values, ijIndexes);
    ldscFeat->getInterspersionAndJuxtapositionIndex(cellPos, total_edges, patchRichness, ijIndexes);

    EXPECT_EQ(5, ijIndexes.size());
    std::map<std::string, double>::iterator itMap = ijIndexes.begin();
    while (itMap != ijIndexes.end()) {
        EXPECT_DOUBLE_EQ(0.0, itMap->second);
        itMap++;
    }
}

TEST_F(LandscapeFeaturesTest, getInterspersionAndJuxtapositionIndexIntersection) {
    std::size_t cellPos = GetCellPosition("0");

    std::map<std::string, double> total_edges;
    GetMapFromSet(class_values, total_edges);
    ldscFeat->getTotalEdges(cellPos, total_edges);

    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);
    ldscFeat->getNumberOfPatches(cellPos, numberOfPatches);

    int patchRichness = ldscFeat->getPatchRichness(cellPos);

    std::map<std::string, double> ijIndexes;
    GetMapFromSet(class_values, ijIndexes);
    ldscFeat->getInterspersionAndJuxtapositionIndex(cellPos, total_edges, patchRichness, ijIndexes);

    EXPECT_EQ(5, ijIndexes.size());
    EXPECT_DOUBLE_EQ(0.0, ijIndexes["FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, ijIndexes["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, ijIndexes["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, ijIndexes["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(0.0, ijIndexes["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getInterspersionAndJuxtapositionIndexIntersectionNonZeros) {
    std::map<std::string, double> total_edges;
    GetMapFromSet(class_values, total_edges);
    std::size_t cellPos = GetCellPosition("7");

    ldscFeat->getTotalEdges(cellPos, total_edges);

    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);
    ldscFeat->getNumberOfPatches(cellPos, numberOfPatches);

    int patchRichness = ldscFeat->getPatchRichness(cellPos);

    std::map<std::string, double> ijIndexes;
    GetMapFromSet(class_values, ijIndexes);
    ldscFeat->getInterspersionAndJuxtapositionIndex(cellPos, total_edges, patchRichness, ijIndexes);

    EXPECT_EQ(5, ijIndexes.size());
    EXPECT_DOUBLE_EQ(63.284522109733032, ijIndexes["FLORESTA"]);
    EXPECT_DOUBLE_EQ(43.619156455122948, ijIndexes["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, ijIndexes["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, ijIndexes["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(42.014259834790757, ijIndexes["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getInterspersionAndJuxtapositionIndexMultipolygon) {
    std::map<std::string, double> total_edges;
    GetMapFromSet(class_values, total_edges);
    std::size_t cellPos = GetCellPosition("10");

    ldscFeat->getTotalEdges(cellPos, total_edges);

    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);
    ldscFeat->getNumberOfPatches(cellPos, numberOfPatches);

    int patchRichness = ldscFeat->getPatchRichness(cellPos);

    std::map<std::string, double> ijIndexes;
    GetMapFromSet(class_values, ijIndexes);
    ldscFeat->getInterspersionAndJuxtapositionIndex(cellPos, total_edges, patchRichness, ijIndexes);

    EXPECT_EQ(5, ijIndexes.size());
    EXPECT_DOUBLE_EQ(0.0, ijIndexes["FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, ijIndexes["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, ijIndexes["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, ijIndexes["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(0.0, ijIndexes["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getTotalAreaBiggestObjectEmptyIntersection) {
    std::size_t cellPos = GetCellPosition("12");

    std::map<std::string, double> tabo;
    GetMapFromSet(class_values, tabo);
    ldscFeat->getTotalAreaBiggestObject(cellPos, tabo);

    EXPECT_EQ(5, tabo.size());
    std::map<std::string, double>::iterator itMap = tabo.begin();
    while (itMap != tabo.end()) {
        EXPECT_DOUBLE_EQ(0.0, itMap->second);
        itMap++;
    }
}

TEST_F(LandscapeFeaturesTest, getTotalAreaBiggestObjectIntersection) {
    std::size_t cellPos = GetCellPosition("7");

    std::map<std::string, double> tabo;
    GetMapFromSet(class_values, tabo);
    ldscFeat->getTotalAreaBiggestObject(cellPos, tabo);

    EXPECT_EQ(5, tabo.size());
    EXPECT_DOUBLE_EQ(1759.8582351239788, tabo["FLORESTA"]);
    EXPECT_DOUBLE_EQ(5300.9113647956137, tabo["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, tabo["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, tabo["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(4143.9358800413038, tabo["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getTotalAreaBiggestObjectIntersectionMultipolygon) {
    std::size_t cellPos = GetCellPosition("10");

    std::map<std::string, double> tabo;
    GetMapFromSet(class_values, tabo);
    ldscFeat->getTotalAreaBiggestObject(cellPos, tabo);

    EXPECT_EQ(5, tabo.size());
    EXPECT_DOUBLE_EQ(0.0, tabo["FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, tabo["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, tabo["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, tabo["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(4143.9358800413038, tabo["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getBiggestIntersectionAreasEmptyIntersection) {
    std::size_t cellPos = GetCellPosition("12");

    std::map<std::string, double> bia;
    GetMapFromSet(class_values, bia);
    std::map<std::string, double> taobia;
    GetMapFromSet(class_values, taobia);

    ldscFeat->getBiggestIntersectionAreas(cellPos, bia, taobia);

    EXPECT_EQ(5, bia.size());
    for (auto &area : bia) {
        EXPECT_DOUBLE_EQ(0.0, area.second);
        EXPECT_DOUBLE_EQ(0.0, taobia[area.first]);
    }
}

TEST_F(LandscapeFeaturesTest, getBiggestIntersectionAreasIntersection) {
    std::size_t cellPos = GetCellPosition("7");

    std::map<std::string, double> bia;
    GetMapFromSet(class_values, bia);
    std::map<std::string, double> taobia;
    GetMapFromSet(class_values, taobia);

    ldscFeat->getBiggestIntersectionAreas(cellPos, bia, taobia);

    EXPECT_EQ(5, bia.size());
    EXPECT_EQ(5, taobia.size());

    EXPECT_DOUBLE_EQ(29.74145948975433, bia["FLORESTA"]);
    EXPECT_DOUBLE_EQ(45.537516931651943, bia["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, bia["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, bia["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(1139.3522722995215, bia["HIDROGRAFIA"]);

    EXPECT_DOUBLE_EQ(1759.8582351239788, taobia["FLORESTA"]);
    EXPECT_DOUBLE_EQ(5300.9113647956137, taobia["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, taobia["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, taobia["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(4143.9358800413038, taobia["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getBiggestIntersectionAreasMultipolygon) {
    std::size_t cellPos = GetCellPosition("10");

    std::map<std::string, double> bia;
    GetMapFromSet(class_values, bia);
    std::map<std::string, double> taobia;
    GetMapFromSet(class_values, taobia);

    ldscFeat->getBiggestIntersectionAreas(cellPos, bia, taobia);

    EXPECT_EQ(5, bia.size());
    EXPECT_EQ(5, taobia.size());

    EXPECT_DOUBLE_EQ(0.0, bia["FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, bia["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, bia["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, bia["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(84.376107492069778, bia["HIDROGRAFIA"]);

    EXPECT_DOUBLE_EQ(0.0, taobia["FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, taobia["NAO_FLORESTA"]);
    EXPECT_DOUBLE_EQ(0.0, taobia["AREA_NAO_OBSERVADA"]);
    EXPECT_DOUBLE_EQ(0.0, taobia["FLORESTA_SOB_NUVEM"]);
    EXPECT_DOUBLE_EQ(4143.9358800413038, taobia["HIDROGRAFIA"]);
}

TEST_F(LandscapeFeaturesTest, getPatchRichnessEmptyIntersection) {
    std::size_t cellPos = GetCellPosition("12");
    int patchRichness = ldscFeat->getPatchRichness(cellPos);

    EXPECT_EQ(0, patchRichness);
}

TEST_F(LandscapeFeaturesTest, getPatchRichnessIntersection) {
    std::size_t cellPos = GetCellPosition("0");
    int patchRichness = ldscFeat->getPatchRichness(cellPos);

    EXPECT_EQ(2, patchRichness);
}

TEST_F(LandscapeFeaturesTest, getPatchRichnessIntersectionMultipolygon) {
    std::size_t cellPos = GetCellPosition("10");
    int patchRichness = ldscFeat->getPatchRichness(cellPos);

    EXPECT_EQ(1, patchRichness);
}

TEST_F(LandscapeFeaturesTest, getPatchRichnessDensityEmptyIntersection) {
    std::size_t cellPos = GetCellPosition("12");

    int patchRichness = ldscFeat->getPatchRichness(cellPos);
    double patchRichnessDensity = ldscFeat->getPatchRichnessDensity(cellPos, patchRichness);

    EXPECT_DOUBLE_EQ(0.0, patchRichnessDensity);
}

TEST_F(LandscapeFeaturesTest, getPatchRichnessDensityIntersection) {
    std::size_t cellPos = GetCellPosition("0");

    int patchRichness = ldscFeat->getPatchRichness(cellPos);
    double patchRichnessDensity = ldscFeat->getPatchRichnessDensity(cellPos, patchRichness);

    EXPECT_DOUBLE_EQ(0.125, patchRichnessDensity);
}

TEST_F(LandscapeFeaturesTest, getPatchRichnessDensityIntersectionMultipolygon) {
    std::size_t cellPos = GetCellPosition("10");

    int patchRichness = ldscFeat->getPatchRichness(cellPos);
    double patchRichnessDensity = ldscFeat->getPatchRichnessDensity(cellPos, patchRichness);

    EXPECT_DOUBLE_EQ(0.0625, patchRichnessDensity);
}

TEST_F(LandscapeFeaturesTest, getShannonsDiversityIndexEmptyIntersection) {
    std::size_t cellPos = GetCellPosition("12");

    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);
    ldscFeat->getClassArea(cellPos, class_areas);

    std::map<std::string, double> percent_lands;
    ldscFeat->getPercentLand(cellPos, class_areas, percent_lands);

    double shannonDivIndex = ldscFeat->getShannonsDiversityIndex(cellPos, percent_lands);

    EXPECT_DOUBLE_EQ(0.0, shannonDivIndex);
}

TEST_F(LandscapeFeaturesTest, getShannonsDiversityIndexIntersection) {
    std::size_t cellPos = GetCellPosition("0");

    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);
    ldscFeat->getClassArea(cellPos, class_areas);

    std::map<std::string, double> percent_lands;
    ldscFeat->getPercentLand(cellPos, class_areas, percent_lands);

    double shannonDivIndex = ldscFeat->getShannonsDiversityIndex(cellPos, percent_lands);

    EXPECT_DOUBLE_EQ(0.69063765093359786, shannonDivIndex);
}

TEST_F(LandscapeFeaturesTest, getShannonsDiversityIndexIntersectionMultipolygon) {
    std::size_t cellPos = GetCellPosition("10");

    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);
    ldscFeat->getClassArea(cellPos, class_areas);

    std::map<std::string, double> percent_lands;
    ldscFeat->getPercentLand(cellPos, class_areas, percent_lands);

    double shannonDivIndex = ldscFeat->getShannonsDiversityIndex(cellPos, percent_lands);

    EXPECT_DOUBLE_EQ(0.155171597469446, shannonDivIndex);
}

TEST_F(LandscapeFeaturesTest, getSimpsonsDiversityIndexEmptyIntersection) {
    std::size_t cellPos = GetCellPosition("12");

    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);
    ldscFeat->getClassArea(cellPos, class_areas);

    std::map<std::string, double> percent_lands;
    ldscFeat->getPercentLand(cellPos, class_areas, percent_lands);

    double simpsonDivIndex = ldscFeat->getSimpsonsDiversityIndex(cellPos, percent_lands);

    EXPECT_DOUBLE_EQ(1.0, simpsonDivIndex);
}

TEST_F(LandscapeFeaturesTest, getSimpsonsDiversityIndexIntersection) {
    std::size_t cellPos = GetCellPosition("0");

    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);
    ldscFeat->getClassArea(cellPos, class_areas);

    std::map<std::string, double> percent_lands;
    ldscFeat->getPercentLand(cellPos, class_areas, percent_lands);

    double simpsonDivIndex = ldscFeat->getSimpsonsDiversityIndex(cellPos, percent_lands);

    EXPECT_DOUBLE_EQ(0.49801925516849044, simpsonDivIndex);
}

TEST_F(LandscapeFeaturesTest, getSimpsonsDiversityIndexIntersectionMultipolygon) {
    std::size_t cellPos = GetCellPosition("10");

    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);
    ldscFeat->getClassArea(cellPos, class_areas);

    std::map<std::string, double> percent_lands;
    ldscFeat->getPercentLand(cellPos, class_areas, percent_lands);

    double simpsonDivIndex = ldscFeat->getSimpsonsDiversityIndex(cellPos, percent_lands);

    EXPECT_DOUBLE_EQ(0.99721901268925262, simpsonDivIndex);
}

TEST_F(LandscapeFeaturesTest, getShannonsEvennessIndexEmptyIntersection) {
    std::size_t cellPos = GetCellPosition("12");

    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);
    ldscFeat->getClassArea(cellPos, class_areas);

    std::map<std::string, double> percent_land;
    ldscFeat->getPercentLand(cellPos, class_areas, percent_land);

    int patchRichness = ldscFeat->getPatchRichness(cellPos);

    double shannonDivIndex = ldscFeat->getShannonsDiversityIndex(cellPos, percent_land);

    double shannonEvenIndex = ldscFeat->getShannonsEvennessIndex(shannonDivIndex, patchRichness);

    EXPECT_DOUBLE_EQ(0.0, shannonEvenIndex);
}

TEST_F(LandscapeFeaturesTest, getShannonsEvennessIndexIntersection) {
    std::size_t cellPos = GetCellPosition("0");

    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);
    ldscFeat->getClassArea(cellPos, class_areas);

    std::map<std::string, double> percent_land;
    ldscFeat->getPercentLand(cellPos, class_areas, percent_land);

    int patchRichness = ldscFeat->getPatchRichness(cellPos);

    double shannonDivIndex = ldscFeat->getShannonsDiversityIndex(cellPos, percent_land);

    double shannonEvenIndex = ldscFeat->getShannonsEvennessIndex(shannonDivIndex, patchRichness);

    EXPECT_DOUBLE_EQ(0.99637951405310465, shannonEvenIndex);
}

TEST_F(LandscapeFeaturesTest, getShannonsEvennessIndexIntersectionMultipolygon) {
    std::size_t cellPos = GetCellPosition("10");

    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);
    ldscFeat->getClassArea(cellPos, class_areas);

    std::map<std::string, double> percent_land;
    ldscFeat->getPercentLand(cellPos, class_areas, percent_land);

    int patchRichness = ldscFeat->getPatchRichness(cellPos);

    double shannonDivIndex = ldscFeat->getShannonsDiversityIndex(cellPos, percent_land);

    double shannonEvenIndex = ldscFeat->getShannonsEvennessIndex(shannonDivIndex, patchRichness);

    EXPECT_DOUBLE_EQ(0.0, shannonEvenIndex);
}

TEST_F(LandscapeFeaturesTest, getSimpsonsEvennessIndexEmptyIntersection) {
    std::size_t cellPos = GetCellPosition("12");

    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);
    ldscFeat->getClassArea(cellPos, class_areas);

    std::map<std::string, double> percent_land;
    ldscFeat->getPercentLand(cellPos, class_areas, percent_land);

    int patchRichness = ldscFeat->getPatchRichness(cellPos);

    double simpsonDivIndex = ldscFeat->getSimpsonsDiversityIndex(cellPos, percent_land);

    double simpsonEvenIndex = ldscFeat->getSimpsonsEvennessIndex(simpsonDivIndex, patchRichness);

    EXPECT_DOUBLE_EQ(0.0, simpsonEvenIndex);
}

TEST_F(LandscapeFeaturesTest, getSimpsonsEvennessIndexIntersection) {
    std::size_t cellPos = GetCellPosition("0");

    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);
    ldscFeat->getClassArea(cellPos, class_areas);

    std::map<std::string, double> percent_land;
    ldscFeat->getPercentLand(cellPos, class_areas, percent_land);

    int patchRichness = ldscFeat->getPatchRichness(cellPos);

    double simpsonDivIndex = ldscFeat->getSimpsonsDiversityIndex(cellPos, percent_land);

    double simpsonEvenIndex = ldscFeat->getSimpsonsEvennessIndex(simpsonDivIndex, patchRichness);

    EXPECT_DOUBLE_EQ(0.99603851033698088, simpsonEvenIndex);
}

TEST_F(LandscapeFeaturesTest, getSimpsonsEvennessIndexIntersectionMultipolygon) {
    std::size_t cellPos = GetCellPosition("10");

    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);
    ldscFeat->getClassArea(cellPos, class_areas);

    std::map<std::string, double> percent_land;
    ldscFeat->getPercentLand(cellPos, class_areas, percent_land);

    int patchRichness = ldscFeat->getPatchRichness(cellPos);

    double simpsonDivIndex = ldscFeat->getSimpsonsDiversityIndex(cellPos, percent_land);

    double simpsonEvenIndex = ldscFeat->getSimpsonsEvennessIndex(simpsonDivIndex, patchRichness);

    EXPECT_DOUBLE_EQ(0.0, simpsonEvenIndex);
}

TEST(LandscapeFeaturesTestNoSetup, getFeaturesDifferentSRIDs) {
//     ASSERT_NO_THROW({
//         TerraLib::getInstance().initialize();
//         gdma::common::loadTerraLibDrivers();
//     });

    std::string pathPolsDegree = gdma::common::FindInGeoDMAPath(
        "data/tests/terraclass_tests/terraclass_tests.shp");
    std::string pathCellDegree = gdma::common::FindInGeoDMAPath(
        "data/tests/terraclass_tests/terraclass_cells_tests_DEGREE.shp");

    boost::bimap<std::string, std::size_t> positionsPolsDegree;
    std::vector<te::gm::Geometry*> containerPolsDegree;
    std::vector<std::string> classesPolsDegree;
    std::set<std::string> classesValuesPolsDegree;

    std::string dsPolsDegree = gdma::dm::openShapefile(pathPolsDegree);
    gdma::dm::getGeometriesFromDataSource(dsPolsDegree, positionsPolsDegree, containerPolsDegree,
        "tcclasse", &classesPolsDegree, &classesValuesPolsDegree);

    boost::bimap<std::string, std::size_t> positionsCellsDegree;
    std::vector<te::gm::Geometry*> containerCellsDegree;

    std::string dsCellsDegree = gdma::dm::openShapefile(pathCellDegree);
    gdma::dm::getGeometriesFromDataSource(dsCellsDegree, positionsCellsDegree,
        containerCellsDegree);

    ASSERT_THROW({
        gdma::fte::LandscapeFeatures landscapeFeatures(containerCellsDegree, positionsCellsDegree,
            containerPolsDegree, positionsPolsDegree, classesPolsDegree, classesValuesPolsDegree);
    }, te::common::Exception);

//     ASSERT_NO_THROW({
//         gdma::common::unloadTerraLibDrivers();
//         TerraLib::getInstance().finalize();
//     });
}

TEST(LandscapeFeaturesTestNoSetup, getFeaturesDegreesSRIDs) {
//     ASSERT_NO_THROW({
//         TerraLib::getInstance().initialize();
//         gdma::common::loadTerraLibDrivers();
//     });

    std::string pathPolsDegree = gdma::common::FindInGeoDMAPath(
        "data/tests/terraclass_tests/terraclass_tests_DEGREE.shp");
    std::string pathCellDegree = gdma::common::FindInGeoDMAPath(
        "data/tests/terraclass_tests/terraclass_cells_tests_DEGREE.shp");

    boost::bimap<std::string, std::size_t> positionsPolsDegree;
    std::vector<te::gm::Geometry*> containerPolsDegree;
    std::vector<std::string> classesPolsDegree;
    std::set<std::string> classesValuesPolsDegree;

    std::string dsPolsDegree = gdma::dm::openShapefile(pathPolsDegree);
    gdma::dm::getGeometriesFromDataSource(dsPolsDegree, positionsPolsDegree, containerPolsDegree,
        "tcclasse", &classesPolsDegree, &classesValuesPolsDegree);

    boost::bimap<std::string, std::size_t> positionsCellsDegree;
    std::vector<te::gm::Geometry*> containerCellsDegree;

    std::string dsCellsDegree = gdma::dm::openShapefile(pathCellDegree);
    gdma::dm::getGeometriesFromDataSource(dsCellsDegree, positionsCellsDegree,
        containerCellsDegree);

    ASSERT_THROW({
        gdma::fte::LandscapeFeatures landscapeFeatures(containerCellsDegree, positionsCellsDegree,
            containerPolsDegree, positionsPolsDegree, classesPolsDegree, classesValuesPolsDegree);
    }, te::common::Exception);

//     ASSERT_NO_THROW({
//         gdma::common::unloadTerraLibDrivers();
//         TerraLib::getInstance().finalize();
//     });
}

TEST_F(LandscapeFeaturesTest, getFeatures) {
    std::map<std::string, gdma::fte::FeaturesSummary> attr_container;
    int size = cells_container.size();
    ASSERT_NO_THROW({
        // MockObsProgressBar obsProgBar;
        bool success = ldscFeat->getFeatures(attr_container);
        // EXPECT_CALL(obsProgBar, notifyProgress()).Times(size);
        ASSERT_TRUE(success);
    });

    EXPECT_EQ(15, attr_container.size());
    for (auto &element : attr_container) {
        EXPECT_EQ(101, element.second.getFeatures().size());
    }
}

//-----------------------------------------------------------------------------------------------
// ------------------ Testing execution with shapefiles having invalid polygons -----------------
//-----------------------------------------------------------------------------------------------

class LandscapeFeaturesTestInvalidPols : public ::testing::Test {
 protected:
    // Polygons Layer
    boost::bimap<std::string, std::size_t> pols_positions;
    std::vector<te::gm::Geometry*> pols_container;
    std::vector<std::string> pol_classes;
    std::set<std::string> class_values;

    // cells Layer
    boost::bimap<std::string, std::size_t> cells_positions;
    std::vector<te::gm::Geometry*> cells_container;

    gdma::fte::LandscapeFeatures* ldscFeat;

    std::size_t testCellPos;

    virtual void SetUp() {
//         try {
//             TerraLib::getInstance().initialize();
//             gdma::common::loadTerraLibDrivers();
//         }
//         catch (te::common::Exception &e) {
//             std::cerr << "ERROR: " << e.what() << std::endl;
//         }

        // Initializing Data sources with invalid polygons
        std::string pathShpInvalidPols = gdma::common::FindInGeoDMAPath(
            "data/tests/terraclass_clip_invalid_pols.shp");
        std::string pathShpInvalidCells = gdma::common::FindInGeoDMAPath(
            "data/tests/terraclass_clip_invalid_pols_cells.shp");

        std::string dsPolsInv = gdma::dm::openShapefile(pathShpInvalidPols);
        std::string dsCellsInv = gdma::dm::openShapefile(pathShpInvalidCells);

        gdma::dm::getGeometriesFromDataSource(dsPolsInv, pols_positions, pols_container, "tcclasse",
            &pol_classes, &class_values);
        gdma::dm::getGeometriesFromDataSource(dsCellsInv, cells_positions, cells_container);

        ldscFeat = new gdma::fte::LandscapeFeatures(cells_container, cells_positions,
            pols_container, pols_positions, pol_classes, class_values);

        testCellPos = GetCellPosition("12");
    }

    virtual void TearDown() {
        delete ldscFeat;
//         gdma::common::unloadTerraLibDrivers();
//         TerraLib::getInstance().finalize();
    }

    std::size_t GetCellPosition(std::string cellId) {
        return cells_positions.left.find(cellId)->second;
    }
};


TEST_F(LandscapeFeaturesTestInvalidPols, getClassAreaInvalidPolygons) {
    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);
    ldscFeat->getClassArea(testCellPos, class_areas);

    EXPECT_EQ(7, class_areas.size());

    for (auto &ca : class_areas)
        EXPECT_DOUBLE_EQ(-1.0, ca.second);
}

TEST_F(LandscapeFeaturesTestInvalidPols, getPercentLandInvalidPolygons) {
    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);
    ldscFeat->getClassArea(testCellPos, class_areas);

    std::map<std::string, double> percent_lands;
    ldscFeat->getPercentLand(testCellPos, class_areas, percent_lands);

    EXPECT_EQ(7, percent_lands.size());

    for (auto &pl : percent_lands)
        EXPECT_DOUBLE_EQ(-1.0, pl.second);
}

TEST_F(LandscapeFeaturesTestInvalidPols, getPatchDensityInvalidPolygons) {
    std::map<std::string, double> patch_densities;
    GetMapFromSet(class_values, patch_densities);
    ldscFeat->getPatchDensities(testCellPos, patch_densities);

    EXPECT_EQ(7, patch_densities.size());

    for (auto &pd : patch_densities)
        EXPECT_DOUBLE_EQ(-1.0, pd.second);
}

TEST_F(LandscapeFeaturesTestInvalidPols, getMeanPatchSizeInvalidPolygons) {
    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);

    ldscFeat->getNumberOfPatches(testCellPos, numberOfPatches);

    std::map<std::string, double> mean_patch_sizes;
    GetMapFromSet(class_values, mean_patch_sizes);

    ldscFeat->getMeanPatchSizes(testCellPos, numberOfPatches, mean_patch_sizes);

    EXPECT_EQ(7, mean_patch_sizes.size());

    for (auto &mps : mean_patch_sizes)
        EXPECT_DOUBLE_EQ(-1.0, mps.second);
}

TEST_F(LandscapeFeaturesTestInvalidPols, getPatchSizesStdInvalidPolygons) {
    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);
    ldscFeat->getNumberOfPatches(testCellPos, numberOfPatches);

    std::map<std::string, double> mean_patch_sizes;
    GetMapFromSet(class_values, mean_patch_sizes);
    ldscFeat->getMeanPatchSizes(testCellPos, numberOfPatches, mean_patch_sizes);

    std::map<std::string, double> patch_sizes_std;
    GetMapFromSet(class_values, patch_sizes_std);
    ldscFeat->getPatchSizesStd(testCellPos, mean_patch_sizes, numberOfPatches, patch_sizes_std);

    EXPECT_EQ(7, patch_sizes_std.size());

    for (auto &psstd : patch_sizes_std)
        EXPECT_DOUBLE_EQ(-1.0, psstd.second);
}

TEST_F(LandscapeFeaturesTestInvalidPols, getLandscapeShapeIndexesInvalidPolygons) {
    std::map<std::string, double> landscapeShapeIndexes;
    GetMapFromSet(class_values, landscapeShapeIndexes);
    ldscFeat->getLandscapeShapeIndexes(testCellPos, landscapeShapeIndexes);

    EXPECT_EQ(7, landscapeShapeIndexes.size());

    for (auto &lsi : landscapeShapeIndexes)
        EXPECT_DOUBLE_EQ(-1.0, lsi.second);
}

TEST_F(LandscapeFeaturesTestInvalidPols, getMeanShapeIndexesInvalidPolygons) {
    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);
    ldscFeat->getNumberOfPatches(testCellPos, numberOfPatches);

    std::map<std::string, double> meanShapeIndexes;
    GetMapFromSet(class_values, meanShapeIndexes);
    ldscFeat->getMeanShapeIndexes(testCellPos, numberOfPatches, meanShapeIndexes);

    EXPECT_EQ(7, meanShapeIndexes.size());

    for (auto &msi : meanShapeIndexes)
        EXPECT_DOUBLE_EQ(-1.0, msi.second);
}

TEST_F(LandscapeFeaturesTestInvalidPols, getAreaWeightedMeanShapeIndexesInvalidPolygons) {
    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);
    ldscFeat->getClassArea(testCellPos, class_areas);

    std::map<std::string, double> awMeanShapeIndexes;
    GetMapFromSet(class_values, awMeanShapeIndexes);
    ldscFeat->getAWMeanShapeIndexes(testCellPos, class_areas, awMeanShapeIndexes);

    EXPECT_EQ(7, awMeanShapeIndexes.size());

    for (auto &awmsi : awMeanShapeIndexes)
        EXPECT_DOUBLE_EQ(-1.0, awmsi.second);
}

TEST_F(LandscapeFeaturesTestInvalidPols, getMeanPatchFractalDimensionsInvalidPolygons) {
    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);
    ldscFeat->getNumberOfPatches(testCellPos, numberOfPatches);

    std::map<std::string, double> mpfds;
    GetMapFromSet(class_values, mpfds);
    ldscFeat->getMeanPatchFractalDimensions(testCellPos, numberOfPatches, mpfds);

    EXPECT_EQ(7, mpfds.size());

    for (auto &mpfd : mpfds)
        EXPECT_DOUBLE_EQ(-1.0, mpfd.second);
}

TEST_F(LandscapeFeaturesTestInvalidPols, getAWMPFDInvalidPolygons) {
    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);
    ldscFeat->getClassArea(testCellPos, class_areas);

    std::map<std::string, double> awmpfds;
    GetMapFromSet(class_values, awmpfds);
    ldscFeat->getAreaWeightedMeanPatchFractalDimensions(testCellPos, class_areas, awmpfds);

    EXPECT_EQ(7, awmpfds.size());

    for (auto &awmpfd : awmpfds)
        EXPECT_DOUBLE_EQ(-1.0, awmpfd.second);
}

TEST_F(LandscapeFeaturesTestInvalidPols, getEdgeDensitiesInvalidPolygons) {
    std::map<std::string, double> class_perimeters;
    GetMapFromSet(class_values, class_perimeters);
    ldscFeat->getTotalEdges(testCellPos, class_perimeters);

    std::map<std::string, double> edgeDensities;
    GetMapFromSet(class_values, edgeDensities);
    ldscFeat->getEdgeDensities(testCellPos, class_perimeters, edgeDensities);

    EXPECT_EQ(7, edgeDensities.size());

    for (auto &ed : edgeDensities)
        EXPECT_DOUBLE_EQ(-1.0, ed.second);
}

TEST_F(LandscapeFeaturesTestInvalidPols, getTotalEdgesInvalidPolygons) {
    std::map<std::string, double> class_perimeters;
    GetMapFromSet(class_values, class_perimeters);
    ldscFeat->getTotalEdges(testCellPos, class_perimeters);

    EXPECT_EQ(7, class_perimeters.size());

    for (auto &te : class_perimeters)
        EXPECT_DOUBLE_EQ(-1.0, te.second);
}

TEST_F(LandscapeFeaturesTestInvalidPols, getMeanPerimeterAreaRatiosInvalidPolygons) {
    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);
    ldscFeat->getNumberOfPatches(testCellPos, numberOfPatches);

    std::map<std::string, double> meanPerimAreaRatio;
    GetMapFromSet(class_values, meanPerimAreaRatio);
    ldscFeat->getMeanPerimeterAreaRatios(testCellPos, numberOfPatches, meanPerimAreaRatio);

    EXPECT_EQ(7, meanPerimAreaRatio.size());

    for (auto &mpar : meanPerimAreaRatio)
        EXPECT_DOUBLE_EQ(-1.0, mpar.second);
}

TEST_F(LandscapeFeaturesTestInvalidPols, getNumberOfPatchesInvalidPolygons) {
    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);
    ldscFeat->getNumberOfPatches(testCellPos, numberOfPatches);

    EXPECT_EQ(7, numberOfPatches.size());

    for (auto &np : numberOfPatches)
        EXPECT_EQ(-1, np.second);
}

TEST_F(LandscapeFeaturesTestInvalidPols, getPatchSizeCoeffVariationInvalidPolygons) {
    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);
    ldscFeat->getNumberOfPatches(testCellPos, numberOfPatches);

    std::map<std::string, double> mean_patch_sizes;
    GetMapFromSet(class_values, mean_patch_sizes);
    ldscFeat->getMeanPatchSizes(testCellPos, numberOfPatches, mean_patch_sizes);

    std::map<std::string, double> patch_sizes_std;
    GetMapFromSet(class_values, patch_sizes_std);
    ldscFeat->getPatchSizesStd(testCellPos, mean_patch_sizes, numberOfPatches, patch_sizes_std);

    std::map<std::string, double> psCoeffVariation;
    GetMapFromSet(class_values, psCoeffVariation);
    ldscFeat->getPatchSizeCoeffVariation(mean_patch_sizes, patch_sizes_std, psCoeffVariation);

    EXPECT_EQ(7, psCoeffVariation.size());

    for (auto &pscv : psCoeffVariation)
        EXPECT_DOUBLE_EQ(-1.0, pscv.second);
}

TEST_F(LandscapeFeaturesTestInvalidPols, getInterspersionAndJuxtapositionIndexInvalidPolygons) {
    std::map<std::string, double> total_edges;
    GetMapFromSet(class_values, total_edges);
    ldscFeat->getTotalEdges(testCellPos, total_edges);

    std::map<std::string, int> numberOfPatches;
    GetMapFromSet(class_values, numberOfPatches);
    ldscFeat->getNumberOfPatches(testCellPos, numberOfPatches);

    int patchRichness = ldscFeat->getPatchRichness(testCellPos);

    std::map<std::string, double> ijIndexes;
    GetMapFromSet(class_values, ijIndexes);
    ldscFeat->getInterspersionAndJuxtapositionIndex(testCellPos, total_edges, patchRichness,
        ijIndexes);

    EXPECT_EQ(7, ijIndexes.size());

    for (auto &iji : ijIndexes)
        EXPECT_DOUBLE_EQ(-1.0, iji.second);
}

TEST_F(LandscapeFeaturesTestInvalidPols, getTotalAreaBiggestObjectInvalidPolygons) {
    std::map<std::string, double> tabo;
    GetMapFromSet(class_values, tabo);
    ldscFeat->getTotalAreaBiggestObject(testCellPos, tabo);

    EXPECT_EQ(7, tabo.size());

    for (auto &elem : tabo)
        EXPECT_DOUBLE_EQ(-1.0, elem.second);
}

TEST_F(LandscapeFeaturesTestInvalidPols, getBiggestIntersectionAreasInvalidPolygons) {
    std::map<std::string, double> bia;
    GetMapFromSet(class_values, bia);
    std::map<std::string, double> taobia;
    GetMapFromSet(class_values, taobia);

    ldscFeat->getBiggestIntersectionAreas(testCellPos, bia, taobia);

    EXPECT_EQ(7, bia.size());
    for (auto &area : bia) {
        EXPECT_DOUBLE_EQ(-1, area.second);
        EXPECT_DOUBLE_EQ(-1, taobia[area.first]);
    }
}

TEST_F(LandscapeFeaturesTestInvalidPols, getPatchRichnessInvalidPolygons) {
    int patchRichness = ldscFeat->getPatchRichness(testCellPos);
    EXPECT_EQ(-1, patchRichness);
}

TEST_F(LandscapeFeaturesTestInvalidPols, getPatchRichnessDensityInvalidPolygons) {
    int patchRichness = ldscFeat->getPatchRichness(testCellPos);
    double patchRichnessDensity = ldscFeat->getPatchRichnessDensity(testCellPos, patchRichness);

    EXPECT_DOUBLE_EQ(-1.0, patchRichnessDensity);
}

TEST_F(LandscapeFeaturesTestInvalidPols, getShannonsDiversityIndexInvalidPolygons) {
    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);
    ldscFeat->getClassArea(testCellPos, class_areas);

    std::map<std::string, double> percent_lands;
    ldscFeat->getPercentLand(testCellPos, class_areas, percent_lands);

    double shannonDivIndex = ldscFeat->getShannonsDiversityIndex(testCellPos, percent_lands);

    EXPECT_DOUBLE_EQ(-1.0, shannonDivIndex);
}

TEST_F(LandscapeFeaturesTestInvalidPols, getSimpsonsDiversityInvalidPolygons) {
    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);
    ldscFeat->getClassArea(testCellPos, class_areas);

    std::map<std::string, double> percent_lands;
    ldscFeat->getPercentLand(testCellPos, class_areas, percent_lands);

    double simpsonDivIndex = ldscFeat->getSimpsonsDiversityIndex(testCellPos, percent_lands);

    EXPECT_DOUBLE_EQ(-1.0, simpsonDivIndex);
}

TEST_F(LandscapeFeaturesTestInvalidPols, getShannonsEvennessIndexInvalidPolygons) {
    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);
    ldscFeat->getClassArea(testCellPos, class_areas);

    std::map<std::string, double> percent_land;
    ldscFeat->getPercentLand(testCellPos, class_areas, percent_land);

    int patchRichness = ldscFeat->getPatchRichness(testCellPos);

    double shannonDivIndex = ldscFeat->getShannonsDiversityIndex(testCellPos, percent_land);

    double shannonEvenIndex = ldscFeat->getShannonsEvennessIndex(shannonDivIndex, patchRichness);

    EXPECT_DOUBLE_EQ(-1.0, shannonEvenIndex);
}

TEST_F(LandscapeFeaturesTestInvalidPols, getSimpsonsEvennessIndexInvalidPolygons) {
    std::map<std::string, double> class_areas;
    GetMapFromSet(class_values, class_areas);
    ldscFeat->getClassArea(testCellPos, class_areas);

    std::map<std::string, double> percent_land;
    ldscFeat->getPercentLand(testCellPos, class_areas, percent_land);

    int patchRichness = ldscFeat->getPatchRichness(testCellPos);

    double simpsonDivIndex = ldscFeat->getSimpsonsDiversityIndex(testCellPos, percent_land);

    double simpsonEvenIndex = ldscFeat->getSimpsonsEvennessIndex(simpsonDivIndex, patchRichness);

    EXPECT_DOUBLE_EQ(-1.0, simpsonEvenIndex);
}

TEST_F(LandscapeFeaturesTestInvalidPols, getFeatures) {
    std::map<std::string, gdma::fte::FeaturesSummary> attr_container;
    int size = cells_container.size();
    ASSERT_NO_THROW({
        bool success = ldscFeat->getFeatures(attr_container);
        ASSERT_TRUE(success);
    });

    EXPECT_EQ(23, attr_container.size());
    for (auto &element : attr_container) {
        EXPECT_EQ(139, element.second.getFeatures().size());
        std::vector<gdma::fte::Attribute> features = element.second.getFeatures();
        for (auto &attr : features) {
            switch (attr.getValueType()) {
            case te::dt::DOUBLE_TYPE:
                EXPECT_DOUBLE_EQ(-1.0, attr.getDouble());
                break;
            case te::dt::INT32_TYPE:
                EXPECT_EQ(-1, attr.getInt32());
                break;
            default:
                FAIL();
            }
        }
    }
}
