/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/featureextraction/multiGeometryAlgorithms.hpp

  \brief Testing functionalities to deal with multi geometry types.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// GTest includes
#include <gtest/gtest.h>

// TerraLib includes
#include <terralib/common/TerraLib.h>
#include <terralib/common/Exception.h>
#include <terralib/geometry.h>

// STL includes
#include <map>
#include <string>
#include <vector>

// GeoDMA includes
#include "geodmalib/common/fileSystem.hpp"
#include "geodmalib/common/utils.hpp"
#include "geodmalib/datamanager/dataManager.hpp"
#include "geodmalib/featureextraction/multiGeometryAlgorithms.hpp"

class MultiGeometryAlgorithmsTest : public ::testing::Test {
 protected:
    std::map<std::string, te::gm::Geometry*> m_geom_container_1;
    std::map<std::string, te::gm::Geometry*> m_geom_container_2;
    std::map<std::string, te::gm::Geometry*> m_geom_container_3;

    virtual void SetUp() {
        try {
//             TerraLib::getInstance().initialize();
//             gdma::common::loadTerraLibDrivers();
        }
        catch (te::common::Exception &e) {
            std::cerr << "ERROR: " << e.what() << std::endl;
        }

        std::string pathShp_1 = gdma::common::FindInGeoDMAPath(
            "data/tests/clipped_pols_tests_1.shp");
        std::string pathShp_2 = gdma::common::FindInGeoDMAPath(
            "data/tests/multipolygon_data.shp");
        std::string pathShp_3 = gdma::common::FindInGeoDMAPath(
            "data/tests/multiline_data.shp");

        std::string dsShp_1 = gdma::dm::openShapefile(pathShp_1);
        std::string dsShp_2 = gdma::dm::openShapefile(pathShp_2);
        std::string dsShp_3 = gdma::dm::openShapefile(pathShp_3);

        gdma::dm::getGeometriesFromDataSource(dsShp_1, m_geom_container_1);
        gdma::dm::getGeometriesFromDataSource(dsShp_2, m_geom_container_2);
        gdma::dm::getGeometriesFromDataSource(dsShp_3, m_geom_container_3);
    }

    virtual void TearDown() {
//         gdma::common::unloadTerraLibDrivers();
//         TerraLib::getInstance().finalize();
    }
};

TEST_F(MultiGeometryAlgorithmsTest, Multi2SilgleFilterPolygonsOnlyPolygons) {
    std::vector<te::gm::Geometry*> vec_geom;

    ASSERT_NO_THROW({
        gdma::fte::Multi2SilgleFilterPolygons(m_geom_container_1["0"], vec_geom);
    });

    EXPECT_EQ(1, vec_geom.size());

    for (auto geom : vec_geom) {
        EXPECT_EQ(te::gm::PolygonType, geom->getGeomTypeId());
    }
}

TEST_F(MultiGeometryAlgorithmsTest, Multi2SilgleFilterPolygonsNoPolygons) {
    std::vector<te::gm::Geometry*> vec_geom;

    ASSERT_NO_THROW({
        gdma::fte::Multi2SilgleFilterPolygons(m_geom_container_3["0"], vec_geom);
    });

    EXPECT_EQ(0, vec_geom.size());
}

TEST_F(MultiGeometryAlgorithmsTest, Multi2SilgleFilterPolygonsMoreThanOnePolygon) {
    std::vector<te::gm::Geometry*> vec_geom;

    ASSERT_NO_THROW({
        gdma::fte::Multi2SilgleFilterPolygons(m_geom_container_2["0"], vec_geom);
    });

    EXPECT_EQ(2, vec_geom.size());
    for (auto geom : vec_geom) {
        EXPECT_EQ(te::gm::PolygonType, geom->getGeomTypeId());
    }
}

TEST_F(MultiGeometryAlgorithmsTest, Multi2SilgleFilterLineStringsOnlyLines) {
    std::vector<te::gm::Geometry*> vec_geom;

    ASSERT_NO_THROW({
        gdma::fte::Multi2SilgleFilterLineStrings(m_geom_container_3["0"], vec_geom);
    });

    EXPECT_EQ(3, vec_geom.size());

    for (auto geom : vec_geom) {
        EXPECT_EQ(te::gm::LineStringType, geom->getGeomTypeId());
    }
}

TEST_F(MultiGeometryAlgorithmsTest, Multi2SilgleFilterLineStringsNoLines) {
    std::vector<te::gm::Geometry*> vec_geom;

    ASSERT_NO_THROW({
        gdma::fte::Multi2SilgleFilterLineStrings(m_geom_container_2["0"], vec_geom);
    });

    EXPECT_EQ(0, vec_geom.size());
}

TEST_F(MultiGeometryAlgorithmsTest, Multi2SilgleFilterPolygonsRestrictNoPolygons) {
    std::vector<te::gm::Geometry*> vec_geom;

    ASSERT_THROW({
        gdma::fte::Multi2SilgleFilterPolygonsRestrict(m_geom_container_3["0"], vec_geom);
    }, te::common::Exception);

    EXPECT_EQ(0, vec_geom.size());
}

TEST_F(MultiGeometryAlgorithmsTest, Multi2SilgleFilterPolygonsRestrictOnlyPolygons) {
    std::vector<te::gm::Geometry*> vec_geom;

    ASSERT_NO_THROW({
        gdma::fte::Multi2SilgleFilterPolygonsRestrict(m_geom_container_2["0"], vec_geom);
    });

    EXPECT_EQ(2, vec_geom.size());
}

TEST_F(MultiGeometryAlgorithmsTest, Multi2SilgleFilterLineStringRestrictNoLines) {
    std::vector<te::gm::Geometry*> vec_geom;

    ASSERT_THROW({
        gdma::fte::Multi2SilgleFilterLineStringRestrict(m_geom_container_2["0"], vec_geom);
    }, te::common::Exception);

    EXPECT_EQ(0, vec_geom.size());
}

TEST_F(MultiGeometryAlgorithmsTest, Multi2SilgleFilterLineStringRestrictOnlyLines) {
    std::vector<te::gm::Geometry*> vec_geom;

    ASSERT_NO_THROW({
        gdma::fte::Multi2SilgleFilterLineStringRestrict(m_geom_container_3["0"], vec_geom);
    });

    EXPECT_EQ(3, vec_geom.size());
}
