/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/featureextraction/spatialFeatures_unittest.hpp

  \brief Functionalities to compute Spatial features.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// GTest includes
#include <gtest/gtest.h>

// TerraLib includes
#include <terralib/common/TerraLib.h>
#include <terralib/common/Exception.h>
#include <terralib/geometry.h>

// Boost Includes
#include <boost/math/constants/constants.hpp>
#include <boost/filesystem.hpp>

// STL includes
#include <map>
#include <memory>
#include <string>
#include <vector>

// GeoDMA includes
#include "geodmalib/common/fileSystem.hpp"
#include "geodmalib/common/utils.hpp"
#include "geodmalib/datamanager/dataManager.hpp"
#include "geodmalib/featureextraction/multiGeometryAlgorithms.hpp"
#include "geodmalib/featureextraction/spatialFeatures.hpp"

class SpatialFeaturesTest : public ::testing::Test {
 protected:
     std::map<std::string, te::gm::Geometry*> m_geom_container;
     te::da::DataSourceInfoPtr m_dsShp;
     boost::filesystem::path m_tmpFolder;

    virtual void SetUp() {
        try {
//             TerraLib::getInstance().initialize();
//             gdma::common::loadTerraLibDrivers();
        }
        catch (te::common::Exception &e) {
            std::cerr << "ERROR: " << e.what() << std::endl;
        }

        std::string pathShp = gdma::common::FindInGeoDMAPath(
            "data/tests/clipped_pols_tests-bigger.shp");

        m_tmpFolder = "./tmpTests";

        boost::system::error_code error;
        if (boost::filesystem::exists(m_tmpFolder)) {
            if (!boost::filesystem::remove_all(m_tmpFolder, error)) {
                std::string errorMsg = "Temporary folder can not be deleted: " +
                    error.message();
                FAIL() << errorMsg;
            }
        } else {
            if (!boost::filesystem::create_directory(m_tmpFolder, error)) {
                std::string errorMsg = "Temporary folder not created: " + error.message();
                FAIL() << errorMsg;
            }
        }

        m_dsShp = gdma::dm::openDataSource(pathShp);
        gdma::dm::getGeometriesFromDataSource(m_dsShp->getId(), m_geom_container);
    }

    virtual void TearDown() {
//         gdma::common::unloadTerraLibDrivers();
//         TerraLib::getInstance().finalize();

        boost::system::error_code error;
        if (boost::filesystem::exists(m_tmpFolder)) {
            if (!boost::filesystem::remove_all(m_tmpFolder, error)) {
                std::string errorMsg = "Temporary folder not deleted: " + error.message();
                FAIL() << errorMsg;
            }
        }

        for (auto &elem : m_geom_container) {
            if (elem.second) {
                EXPECT_NO_FATAL_FAILURE({
                    delete elem.second;
                    elem.second = 0;
                });
            }
        }
    }
};

TEST_F(SpatialFeaturesTest, computeArea) {
    std::vector<te::gm::Geometry*> pols;
    gdma::fte::Multi2SilgleFilterPolygons(m_geom_container["0"], pols);

    double area = 0.0;
    for (auto polygon : pols)
        area += gdma::fte::SpatialFeatures::computeArea((te::gm::Polygon*)polygon);

    EXPECT_DOUBLE_EQ(3538800.0231033117, area);
}

TEST_F(SpatialFeaturesTest, computePBoxArea) {
    std::vector<te::gm::Geometry*> pols;
    gdma::fte::Multi2SilgleFilterPolygons(m_geom_container["0"], pols);

    double area = 0.0;
    for (auto polygon : pols) {
        std::unique_ptr<te::gm::Envelope> boundingBox((te::gm::Envelope*)polygon->getMBR());
        area += gdma::fte::SpatialFeatures::computePBoxArea(boundingBox.get());
    }

    EXPECT_DOUBLE_EQ(6432000.0419906899, area);
}

TEST_F(SpatialFeaturesTest, computePerimeter) {
    std::vector<te::gm::Geometry*> pols;
    gdma::fte::Multi2SilgleFilterPolygons(m_geom_container["0"], pols);

    double perimeter = 0.0;
    for (auto polygon : pols)
        perimeter += gdma::fte::SpatialFeatures::computePerimeter((te::gm::Polygon*)polygon);

    EXPECT_DOUBLE_EQ(15360.000050136819, perimeter);
}

TEST_F(SpatialFeaturesTest, computeFractalDim) {
    std::vector<te::gm::Geometry*> pols;
    gdma::fte::Multi2SilgleFilterPolygons(m_geom_container["0"], pols);

    double area = 0.0;
    for (auto polygon : pols)
        area += gdma::fte::SpatialFeatures::computeArea((te::gm::Polygon*)polygon);

    double perimeter = 0.0;
    for (auto polygon : pols)
        perimeter += gdma::fte::SpatialFeatures::computePerimeter((te::gm::Polygon*)polygon);

    double fractalDim = 0.0;
    for (auto polygon : pols)
        fractalDim += gdma::fte::SpatialFeatures::computeFractalDim(area, perimeter);

    EXPECT_DOUBLE_EQ(0.31962767263555686, fractalDim);
}

TEST_F(SpatialFeaturesTest, computePBoxLength) {
    std::vector<te::gm::Geometry*> pols;
    gdma::fte::Multi2SilgleFilterPolygons(m_geom_container["0"], pols);

    double pboxLength = 0.0;
    for (auto polygon : pols) {
        std::unique_ptr<te::gm::Envelope> boundingBox((te::gm::Envelope*)polygon->getMBR());
        pboxLength += gdma::fte::SpatialFeatures::computePBoxLength(boundingBox.get());
    }

    EXPECT_DOUBLE_EQ(1600.000005222857, pboxLength);
}

TEST_F(SpatialFeaturesTest, computePBoxWidth) {
    std::vector<te::gm::Geometry*> pols;
    gdma::fte::Multi2SilgleFilterPolygons(m_geom_container["0"], pols);

    double pboxWidth = 0.0;
    for (auto polygon : pols) {
        std::unique_ptr<te::gm::Envelope> boundingBox((te::gm::Envelope*)polygon->getMBR());
        pboxWidth += gdma::fte::SpatialFeatures::computePBoxWidth(boundingBox.get());
    }

    EXPECT_DOUBLE_EQ(4020.000013121753, pboxWidth);
}

TEST_F(SpatialFeaturesTest, computePerimAreaRatio) {
    std::vector<te::gm::Geometry*> pols;
    gdma::fte::Multi2SilgleFilterPolygons(m_geom_container["0"], pols);

    double area = 0.0;
    for (auto polygon : pols)
        area += gdma::fte::SpatialFeatures::computeArea((te::gm::Polygon*)polygon);

    double perimeter = 0.0;
    for (auto polygon : pols)
        perimeter += gdma::fte::SpatialFeatures::computePerimeter((te::gm::Polygon*)polygon);

    double perimAreaRatio = 0.0;
    for (auto polygon : pols)
        perimAreaRatio += gdma::fte::SpatialFeatures::computePerimAreaRatio(area, perimeter);

    EXPECT_DOUBLE_EQ(0.0043404543771498666, perimAreaRatio);
}

TEST_F(SpatialFeaturesTest, computeCompacity) {
    std::vector<te::gm::Geometry*> pols;
    gdma::fte::Multi2SilgleFilterPolygons(m_geom_container["0"], pols);

    double area = 0.0;
    for (auto polygon : pols)
        area += gdma::fte::SpatialFeatures::computeArea((te::gm::Polygon*)polygon);

    double perimeter = 0.0;
    for (auto polygon : pols)
        perimeter += gdma::fte::SpatialFeatures::computePerimeter((te::gm::Polygon*)polygon);

    double perimAreaRatio = 0.0;
    for (auto polygon : pols)
        perimAreaRatio += gdma::fte::SpatialFeatures::computePerimAreaRatio(area, perimeter);

    double compacity = 0.0;
    for (auto polygon : pols)
        compacity += gdma::fte::SpatialFeatures::computeCompacity(area, perimAreaRatio);

    EXPECT_DOUBLE_EQ(2.3073165699371295e-006, compacity);
}

TEST_F(SpatialFeaturesTest, fitEllipse) {
    std::vector<te::gm::Geometry*> pols;
    gdma::fte::Multi2SilgleFilterPolygons(m_geom_container["0"], pols);

    std::vector<double> coefs;
    double offsetX, offsetY;
    gdma::fte::SpatialFeatures::fitEllipse((te::gm::Polygon*)pols[0], coefs, offsetX, offsetY);

    EXPECT_DOUBLE_EQ(537959.00012271351, offsetX);
    EXPECT_DOUBLE_EQ(7925118.9976836992, offsetY);

    EXPECT_EQ(6, coefs.size());
    EXPECT_FALSE(coefs.empty());
}

TEST_F(SpatialFeaturesTest, solveEllipse) {
    std::vector<te::gm::Geometry*> pols;
    gdma::fte::Multi2SilgleFilterPolygons(m_geom_container["0"], pols);

    std::vector<double> coefs;
    double offsetX, offsetY;
    gdma::fte::SpatialFeatures::fitEllipse((te::gm::Polygon*)pols[0], coefs, offsetX, offsetY);

    double minor_radius, major_radius, centerX, centerY, angle;
    gdma::fte::SpatialFeatures::solveEllipse(coefs, minor_radius, major_radius,
        centerX, centerY, angle);

    EXPECT_NEAR(817.0034, minor_radius, 0.0001);
    EXPECT_NEAR(2245.7175, major_radius, 0.0001);
    EXPECT_NEAR(2026.3865, centerX, 0.0001);
    EXPECT_NEAR(844.2376, centerY, 0.0001);
    EXPECT_NEAR(-0.1966, angle, 0.0001);
}

TEST_F(SpatialFeaturesTest, createEllipsePolygon) {
    std::vector<te::gm::Geometry*> pols;
    gdma::fte::Multi2SilgleFilterPolygons(m_geom_container["0"], pols);

    std::vector<double> coefs;
    double offsetX, offsetY;
    gdma::fte::SpatialFeatures::fitEllipse((te::gm::Polygon*)pols[0], coefs, offsetX, offsetY);

    double minor_radius, major_radius, centerX, centerY, angle;
    gdma::fte::SpatialFeatures::solveEllipse(coefs, minor_radius, major_radius,
        centerX, centerY, angle);

    std::unique_ptr<te::gm::Polygon> ellipse(gdma::fte::SpatialFeatures::createEllipsePolygon(
        minor_radius, major_radius, centerX, centerY, angle, pols[0]->getSRID(), offsetX, offsetY));

    EXPECT_EQ(te::gm::PolygonType, ellipse->getGeomTypeId());
    EXPECT_NEAR(5760273.8736415505, ellipse->getArea(), 0.0001);
    EXPECT_NEAR(10151.142913746111, ellipse->getPerimeter(), 0.0001);
    EXPECT_EQ(101, ellipse->getNPoints());
}

TEST_F(SpatialFeaturesTest, computeEllipticFit) {
    std::vector<te::gm::Geometry*> pols;
    gdma::fte::Multi2SilgleFilterPolygons(m_geom_container["0"], pols);

    std::vector<double> coefs;
    double offsetX, offsetY;
    gdma::fte::SpatialFeatures::fitEllipse((te::gm::Polygon*)pols[0], coefs, offsetX, offsetY);

    double minor_radius, major_radius, centerX, centerY, angle;
    gdma::fte::SpatialFeatures::solveEllipse(coefs, minor_radius, major_radius,
        centerX, centerY, angle);

    // Computes Elliptic Fit
    double ellipseArea = boost::math::constants::pi<double>() * major_radius * minor_radius;
    double polArea = ((te::gm::Polygon*)pols[0])->getArea();
    double ellipticFit = gdma::fte::SpatialFeatures::computeEllipticFit(ellipseArea, polArea);

    EXPECT_NEAR(0.61394164985320177, ellipticFit, 0.0001);
}

TEST_F(SpatialFeaturesTest, computeGyrationRadius) {
    std::vector<te::gm::Geometry*> pols;
    gdma::fte::Multi2SilgleFilterPolygons(m_geom_container["0"], pols);

    double gyrRadius, maxDistance;
    gdma::fte::SpatialFeatures::computeGyrationRadius((te::gm::Polygon*)pols[0], gyrRadius,
        maxDistance);

    EXPECT_DOUBLE_EQ(1282.2651407943829, gyrRadius);
    EXPECT_DOUBLE_EQ(2289.1487944049113, maxDistance);
}

TEST_F(SpatialFeaturesTest, computePolygonCircle) {
    std::vector<te::gm::Geometry*> pols;
    gdma::fte::Multi2SilgleFilterPolygons(m_geom_container["0"], pols);

    double polArea = ((te::gm::Polygon*)pols[0])->getArea();

    double gyrRadius, polRadius;
    gdma::fte::SpatialFeatures::computeGyrationRadius((te::gm::Polygon*)pols[0], gyrRadius,
        polRadius);

    // Computes Polygon Circle
    double polCircle = gdma::fte::SpatialFeatures::computePolygonCircle(polRadius, polArea);

    EXPECT_DOUBLE_EQ(0.78503977729096908, polCircle);
}

TEST_F(SpatialFeaturesTest, computePolShapeIndex) {
    std::vector<te::gm::Geometry*> pols;
    gdma::fte::Multi2SilgleFilterPolygons(m_geom_container["0"], pols);

    double polArea = ((te::gm::Polygon*)pols[0])->getArea();
    double polPerimeter = ((te::gm::Polygon*)pols[0])->getPerimeter();

    double polShpIdx = gdma::fte::SpatialFeatures::computePolShapeIndex(polPerimeter, polArea);

    EXPECT_DOUBLE_EQ(2.0412829827500421, polShpIdx);
}

TEST_F(SpatialFeaturesTest, computePolygonDensity) {
    std::vector<te::gm::Geometry*> pols;
    gdma::fte::Multi2SilgleFilterPolygons(m_geom_container["0"], pols);

    double polArea = ((te::gm::Polygon*)pols[0])->getArea();
    double gyrRadius, polRadius;
    gdma::fte::SpatialFeatures::computeGyrationRadius((te::gm::Polygon*)pols[0], gyrRadius,
        polRadius);

    double polDensity = gdma::fte::SpatialFeatures::computePolygonDensity(polRadius, polArea);

    EXPECT_DOUBLE_EQ(1545.9021413342684, polDensity);
}

TEST_F(SpatialFeaturesTest, fitRectangle) {
    std::vector<te::gm::Geometry*> pols;
    gdma::fte::Multi2SilgleFilterPolygons(m_geom_container["0"], pols);

    std::unique_ptr<te::gm::Polygon> rectangle(
        gdma::fte::SpatialFeatures::fitRectangle((te::gm::Polygon*)pols[0]));

    EXPECT_EQ(te::gm::PolygonType, rectangle->getGeomTypeId());
    EXPECT_NEAR(5752606.0902542835, rectangle->getArea(), 0.0001);
    EXPECT_NEAR(10986.37828193472, rectangle->getPerimeter(), 0.0001);
    EXPECT_EQ(5, rectangle->getNPoints());
}

TEST_F(SpatialFeaturesTest, computeRectangularFit) {
    std::vector<te::gm::Geometry*> pols;
    gdma::fte::Multi2SilgleFilterPolygons(m_geom_container["0"], pols);

    double polArea = ((te::gm::Polygon*)pols[0])->getArea();

    std::unique_ptr<te::gm::Polygon> rectangle(
        gdma::fte::SpatialFeatures::fitRectangle((te::gm::Polygon*)pols[0]));
    double rectArea = rectangle->getArea();

    double rectFit = gdma::fte::SpatialFeatures::computeRectangularFit(polArea, rectArea);

    EXPECT_DOUBLE_EQ(0.61516466929632674, rectFit);
}

// @TODO(Raian): Verify a way to assert the values of the features in the container. Save a file?
TEST_F(SpatialFeaturesTest, getFeatures) {
    std::map<std::string, gdma::fte::FeaturesSummary> attr_container;

    ASSERT_NO_THROW({
        bool success = gdma::fte::SpatialFeatures::getFeatures(m_geom_container, attr_container, false, -1000 );
        ASSERT_TRUE(success);
    });

    EXPECT_EQ(6, attr_container.size());
    for (auto &element : attr_container) {
        EXPECT_EQ(17, element.second.getFeatures().size());
    }

    /*std::string pathTempFile = gdma::common::FindInGeoDMAPath("data/test/attributes_test.csv");
    std::string outputCSV = m_tmpFolder.string() + std::string("/csv_test.csv");

    gdma::dm::saveAttributesToCSV(attr_container, m_dsShp->getId(),
        boost::filesystem::basename(m_dsShp->getTitle()), outputCSV);

    std::ifstream outFile(outputCSV);
    std::ifstream templateFile(pathTempFile);*/
}

TEST_F(SpatialFeaturesTest, getFeaturesWithMultiThread) {
    std::map<std::string, gdma::fte::FeaturesSummary> attr_container;

    ASSERT_NO_THROW({
        bool success = gdma::fte::SpatialFeatures::getFeatures(m_geom_container,
          attr_container, true, -1000 );
        ASSERT_TRUE(success);
    });

    EXPECT_EQ(6, attr_container.size());
    for (auto &element : attr_container) {
        EXPECT_EQ(17, element.second.getFeatures().size());
    }

    /*std::string pathTempFile = gdma::common::FindInGeoDMAPath("data/test/attributes_test.csv");
    std::string outputCSV = m_tmpFolder.string() + std::string("/csv_test.csv");

    gdma::dm::saveAttributesToCSV(attr_container, m_dsShp->getId(),
        boost::filesystem::basename(m_dsShp->getTitle()), outputCSV);

    std::ifstream outFile(outputCSV);
    std::ifstream templateFile(pathTempFile);*/
}

TEST_F(SpatialFeaturesTest, getFeaturesInvalidPolygon) {
    std::map<std::string, gdma::fte::FeaturesSummary> attr_container;

    std::string pathShp = gdma::common::FindInGeoDMAPath("data/tests/terraclass_clip_invalid_pols.shp");

    std::map<std::string, te::gm::Geometry*> geom_container;

    te::da::DataSourceInfoPtr dsShp = gdma::dm::openDataSource(pathShp);
    gdma::dm::getGeometriesFromDataSource(dsShp->getId(), geom_container);

    ASSERT_NO_THROW({
        bool success = gdma::fte::SpatialFeatures::getFeatures(geom_container, attr_container, false, -1000 );
        ASSERT_TRUE(success);
    });

    EXPECT_EQ(89, attr_container.size());
    for (auto &element : attr_container) {
        /*if (element.first.compare("72") == 0) {
            std::vector<gdma::fte::Attribute> features = element.second.getFeatures();
            for (auto &feat : features) {
                EXPECT_DOUBLE_EQ(-1.0, feat.getAsDouble());
            }
        }*/
        EXPECT_EQ(17, element.second.getFeatures().size());
    }
}

TEST_F(SpatialFeaturesTest, pointersDeletionBugTest) {
    std::map<std::string, gdma::fte::FeaturesSummary> attr_container;

    ASSERT_NO_THROW({
        bool success = gdma::fte::SpatialFeatures::getFeatures(m_geom_container, attr_container, false, -1000 );
        ASSERT_TRUE(success);
    });

    for (auto &elem : m_geom_container) {
        EXPECT_STRNE("", elem.first.c_str());
        EXPECT_EQ(te::gm::MultiPolygonType, elem.second->getGeomTypeId());
        EXPECT_NO_FATAL_FAILURE({
            delete elem.second;
            elem.second = 0;
        });
    }
}
