/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/tests/unittest/featureextraction/spectralFeatures.cpp

  \brief

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
  */

// GTest Includes
#include <gtest/gtest.h>

// TerraLib Includes
#include <terralib/common.h>

// STL Includes
#include <map>
#include <memory>
#include <string>
#include <vector>

// GeoDMA Includes
#include "geodmalib/common/fileSystem.hpp"
#include "geodmalib/common/utils.hpp"
#include "geodmalib/datamanager/dataManager.hpp"
#include "geodmalib/featureextraction/featuresSummary.hpp"
#include "geodmalib/featureextraction/spectralFeatures.hpp"

class SpectralFeaturesTest : public ::testing::Test {
 protected:
    virtual void SetUp() {
        try {
//             TerraLib::getInstance().initialize();
//             gdma::common::loadTerraLibDrivers();
        } catch (te::common::Exception &e) {
            FAIL() << e.what();
        }
    }

    virtual void TearDown() {
//         gdma::common::unloadTerraLibDrivers();
//         TerraLib::getInstance().finalize();
    }
};

TEST_F(SpectralFeaturesTest, getFeaturesOnePolygonAllBands) {
    std::string pathRaster = gdma::common::FindInGeoDMAPath("/data/tests/clipped_rst_test1.tif");
    std::string pathShp = gdma::common::FindInGeoDMAPath("/data/tests/clipped_pols_tests_1.shp");

    std::unique_ptr<te::rst::Raster> inputRaster(gdma::dm::openRaster(pathRaster));
    std::vector<unsigned int> bands;

    for (int i = 0; i < inputRaster->getNumberOfBands(); i++) {
        bands.push_back(i);
    }

    std::map<std::string, te::gm::Geometry*> geom_container;
    std::string dsShp = gdma::dm::openShapefile(pathShp);
    gdma::dm::getGeometriesFromDataSource(dsShp, geom_container);

    std::map<std::string, gdma::fte::FeaturesSummary> attr_container;

    bool result;
    ASSERT_NO_THROW({
        result = gdma::fte::SpectralFeatures::getFeatures(inputRaster.get(), geom_container,
            bands, attr_container, true);
    });

    ASSERT_TRUE(result);

    EXPECT_EQ(1, attr_container.size());

    for (auto &elem : attr_container)
        EXPECT_EQ(63, elem.second.getFeatures().size());

    std::vector<gdma::fte::Attribute> features = attr_container["0"].getFeatures();

    // Features Band 0
    EXPECT_STREQ("AMPLITUDE_BAND_0", features[0].getName().c_str());
    EXPECT_DOUBLE_EQ(25, features[0].getDouble());
    EXPECT_STREQ("COUNT_BAND_0", features[1].getName().c_str());
    EXPECT_EQ(8847, features[1].getInt32());
    EXPECT_STREQ("KURTOSIS_BAND_0", features[2].getName().c_str());
    EXPECT_NEAR(0.00521637, features[2].getDouble(), 0.000001);
    EXPECT_STREQ("MAXIMUM_VAL_BAND_0", features[3].getName().c_str());
    EXPECT_DOUBLE_EQ(56, features[3].getDouble());
    EXPECT_STREQ("MEAN_BAND_0", features[4].getName().c_str());
    EXPECT_NEAR(43.3469, features[4].getDouble(), 0.0001);
    EXPECT_STREQ("MEDIAN_BAND_0", features[5].getName().c_str());
    EXPECT_DOUBLE_EQ(44, features[5].getDouble());
    EXPECT_STREQ("MINIMUM_VAL_BAND_0", features[6].getName().c_str());
    EXPECT_DOUBLE_EQ(31, features[6].getDouble());
    EXPECT_STREQ("MODE_BAND_0", features[7].getName().c_str());
    EXPECT_DOUBLE_EQ(44, features[7].getDouble());
    EXPECT_STREQ("NUM_MODES_BAND_0", features[8].getName().c_str());
    EXPECT_EQ(1, features[8].getInt32());
    EXPECT_STREQ("SKEWNESS_BAND_0", features[9].getName().c_str());
    EXPECT_NEAR(-0.671505, features[9].getDouble(), 0.0001);
    EXPECT_STREQ("STD_DEVIATION_BAND_0", features[10].getName().c_str());
    EXPECT_NEAR(3.71875, features[10].getDouble(), 0.0001);
    EXPECT_STREQ("SUM_BAND_0", features[11].getName().c_str());
    EXPECT_DOUBLE_EQ(383490, features[11].getDouble());
    EXPECT_STREQ("VALID_COUNT_BAND_0", features[12].getName().c_str());
    EXPECT_EQ(8847, features[12].getInt32());
    EXPECT_STREQ("VAR_COEF_BAND_0", features[13].getName().c_str());
    EXPECT_NEAR(8.57905, features[13].getDouble(), 0.0001);
    EXPECT_STREQ("VARIANCE_BAND_0", features[14].getName().c_str());
    EXPECT_NEAR(13.8291, features[14].getDouble(), 0.0001);
    EXPECT_STREQ("CONTRAST_SE_BAND_0", features[15].getName().c_str());
    // EXPECT_NEAR(1313.1199, features[15].getDouble(), 0.0001);
    EXPECT_STREQ("DISSIMILARITY_SE_BAND_0", features[16].getName().c_str());
    // EXPECT_NEAR(14.6453, features[16].getDouble(), 0.0001);
    EXPECT_STREQ("ENERGY_SE_BAND_0", features[17].getName().c_str());
    EXPECT_NEAR(0.122897, features[17].getDouble(), 0.01);
    EXPECT_STREQ("ENTROPY_SE_BAND_0", features[18].getName().c_str());
    EXPECT_NEAR(5.37921, features[18].getDouble(), 0.2);
    EXPECT_STREQ("HOMOGENEITY_SE_BAND_0", features[19].getName().c_str());
    EXPECT_NEAR(5.37921, features[19].getDouble(), 0.2);
    EXPECT_STREQ("BAND_RATIO_BAND_0", features[60].getName().c_str());
    EXPECT_NEAR(0.150461, features[60].getDouble(), 0.0001);

    // Features Band 1
    EXPECT_STREQ("AMPLITUDE_BAND_1", features[20].getName().c_str());
    EXPECT_DOUBLE_EQ(110, features[20].getDouble());
    EXPECT_STREQ("COUNT_BAND_1", features[21].getName().c_str());
    EXPECT_EQ(8847, features[21].getInt32());
    EXPECT_STREQ("KURTOSIS_BAND_1", features[22].getName().c_str());
    EXPECT_NEAR(0.00447189, features[22].getDouble(), 0.000001);
    EXPECT_STREQ("MAXIMUM_VAL_BAND_1", features[23].getName().c_str());
    EXPECT_DOUBLE_EQ(206, features[23].getDouble());
    EXPECT_STREQ("MEAN_BAND_1", features[24].getName().c_str());
    EXPECT_NEAR(164.381, features[24].getDouble(), 0.0005);
    EXPECT_STREQ("MEDIAN_BAND_1", features[25].getName().c_str());
    EXPECT_DOUBLE_EQ(165, features[25].getDouble());
    EXPECT_STREQ("MINIMUM_VAL_BAND_1", features[26].getName().c_str());
    EXPECT_DOUBLE_EQ(96, features[26].getDouble());
    EXPECT_STREQ("MODE_BAND_1", features[27].getName().c_str());
    EXPECT_DOUBLE_EQ(162, features[27].getDouble());
    EXPECT_STREQ("NUM_MODES_BAND_1", features[28].getName().c_str());
    EXPECT_EQ(1, features[28].getInt32());
    EXPECT_STREQ("SKEWNESS_BAND_1", features[29].getName().c_str());
    EXPECT_NEAR(-0.237124, features[29].getDouble(), 0.0001);
    EXPECT_STREQ("STD_DEVIATION_BAND_1", features[30].getName().c_str());
    EXPECT_NEAR(15.075, features[30].getDouble(), 0.0001);
    EXPECT_STREQ("SUM_BAND_1", features[31].getName().c_str());
    EXPECT_DOUBLE_EQ(1454283, features[31].getDouble());
    EXPECT_STREQ("VALID_COUNT_BAND_1", features[32].getName().c_str());
    EXPECT_EQ(8847, features[32].getInt32());
    EXPECT_STREQ("VAR_COEF_BAND_1", features[33].getName().c_str());
    EXPECT_NEAR(9.17076, features[33].getDouble(), 0.0001);
    EXPECT_STREQ("VARIANCE_BAND_1", features[34].getName().c_str());
    EXPECT_NEAR(227.257, features[34].getDouble(), 0.0003);
    EXPECT_STREQ("CONTRAST_SE_BAND_1", features[35].getName().c_str());
    // EXPECT_NEAR(1313.1199, features[35].getDouble(), 0.0001);
    EXPECT_STREQ("DISSIMILARITY_SE_BAND_1", features[36].getName().c_str());
    // EXPECT_NEAR(14.6453, features[36].getDouble(), 0.0001);
    EXPECT_STREQ("ENERGY_SE_BAND_1", features[37].getName().c_str());
    EXPECT_NEAR(0.152148, features[37].getDouble(), 0.1);
    EXPECT_STREQ("ENTROPY_SE_BAND_1", features[38].getName().c_str());
    EXPECT_NEAR(6.43066, features[38].getDouble(), 0.6);
    EXPECT_STREQ("HOMOGENEITY_SE_BAND_1", features[39].getName().c_str());
    EXPECT_NEAR(6.43066, features[39].getDouble(), 0.6);
    EXPECT_STREQ("BAND_RATIO_BAND_1", features[61].getName().c_str());
    EXPECT_NEAR(0.570581, features[61].getDouble(), 0.0001);

    // Features Band 0
    EXPECT_STREQ("AMPLITUDE_BAND_2", features[40].getName().c_str());
    EXPECT_DOUBLE_EQ(31, features[40].getDouble());
    EXPECT_STREQ("COUNT_BAND_2", features[41].getName().c_str());
    EXPECT_EQ(8847, features[41].getInt32());
    EXPECT_STREQ("KURTOSIS_BAND_2", features[42].getName().c_str());
    EXPECT_NEAR(0.00681356, features[42].getDouble(), 0.000001);
    EXPECT_STREQ("MAXIMUM_VAL_BAND_2", features[43].getName().c_str());
    EXPECT_DOUBLE_EQ(93, features[43].getDouble());
    EXPECT_STREQ("MEAN_BAND_2", features[44].getName().c_str());
    EXPECT_NEAR(80.3663, features[44].getDouble(), 0.0001);
    EXPECT_STREQ("MEDIAN_BAND_2", features[45].getName().c_str());
    EXPECT_DOUBLE_EQ(81, features[45].getDouble());
    EXPECT_STREQ("MINIMUM_VAL_BAND_2", features[46].getName().c_str());
    EXPECT_DOUBLE_EQ(62, features[46].getDouble());
    EXPECT_STREQ("MODE_BAND_2", features[47].getName().c_str());
    EXPECT_DOUBLE_EQ(81, features[47].getDouble());
    EXPECT_STREQ("NUM_MODES_BAND_2", features[48].getName().c_str());
    EXPECT_EQ(1, features[48].getInt32());
    EXPECT_STREQ("SKEWNESS_BAND_2", features[49].getName().c_str());
    EXPECT_NEAR(-1.30591, features[49].getDouble(), 0.0001);
    EXPECT_STREQ("STD_DEVIATION_BAND_2", features[50].getName().c_str());
    EXPECT_NEAR(4.8304, features[50].getDouble(), 0.0001);
    EXPECT_STREQ("SUM_BAND_2", features[51].getName().c_str());
    EXPECT_DOUBLE_EQ(711001, features[51].getDouble());
    EXPECT_STREQ("VALID_COUNT_BAND_2", features[52].getName().c_str());
    EXPECT_EQ(8847, features[52].getInt32());
    EXPECT_STREQ("VAR_COEF_BAND_2", features[53].getName().c_str());
    EXPECT_NEAR(6.01048, features[53].getDouble(), 0.0001);
    EXPECT_STREQ("VARIANCE_BAND_2", features[54].getName().c_str());
    EXPECT_NEAR(23.3328, features[54].getDouble(), 0.0001);
    EXPECT_STREQ("CONTRAST_SE_BAND_2", features[55].getName().c_str());
    // EXPECT_NEAR(1313.1199, features[55].getDouble(), 0.0001);
    EXPECT_STREQ("DISSIMILARITY_SE_BAND_2", features[56].getName().c_str());
    // EXPECT_NEAR(14.6453, features[56].getDouble(), 0.0001);
    EXPECT_STREQ("ENERGY_SE_BAND_2", features[57].getName().c_str());
    EXPECT_NEAR(0.237413, features[57].getDouble(), 0.05);
    EXPECT_STREQ("ENTROPY_SE_BAND_2", features[58].getName().c_str());
    EXPECT_NEAR(4.66498, features[58].getDouble(), 0.2);
    EXPECT_STREQ("HOMOGENEITY_SE_BAND_2", features[59].getName().c_str());
    EXPECT_NEAR(4.66498, features[59].getDouble(), 0.2);
    EXPECT_STREQ("BAND_RATIO_BAND_2", features[62].getName().c_str());
    EXPECT_NEAR(0.278958, features[62].getDouble(), 0.0001);

    ASSERT_NO_FATAL_FAILURE({ te::common::FreeContents(geom_container); });
}

TEST_F(SpectralFeaturesTest, getFeaturesDifferentSRIDs) {
    std::string pathRaster =
        gdma::common::FindInGeoDMAPath("/data/tests/clipped_rst_tests-bigger.tif");
    std::string pathShp =
        gdma::common::FindInGeoDMAPath("/data/tests/clipped_pols_tests-bigger-REP.shp");

    std::unique_ptr<te::rst::Raster> inputRaster(gdma::dm::openRaster(pathRaster));
    std::vector<unsigned int> bands;

    for (int i = 0; i < inputRaster->getNumberOfBands(); i++) {
        bands.push_back(i);
    }

    std::map<std::string, te::gm::Geometry*> geom_container;
    std::string dsShp = gdma::dm::openShapefile(pathShp);
    gdma::dm::getGeometriesFromDataSource(dsShp, geom_container);

    std::map<std::string, gdma::fte::FeaturesSummary> attr_container;

    std::ostringstream excptMessage;
    excptMessage <<  "Geometry received have an Spatial Reference different from the "
        "Input Raster received!" << std::endl;
    excptMessage << "\tSRID Raster: 29192" << std::endl;
    excptMessage << "\tSRID Geometry: 5880" << std::endl;

    bool result;
    ASSERT_THROW({
        try {
            result = gdma::fte::SpectralFeatures::getFeatures(inputRaster.get(), geom_container,
                bands, attr_container, true);
        } catch (te::common::Exception &e) {
            EXPECT_STREQ(TE_TR(excptMessage.str()), e.what());
            throw te::common::Exception();
        }
    }, te::common::Exception);

    ASSERT_NO_FATAL_FAILURE({ te::common::FreeContents(geom_container); });
}

TEST_F(SpectralFeaturesTest, getFeaturesEmptyContainer) {
    std::string pathRaster =
        gdma::common::FindInGeoDMAPath("/data/tests/clipped_rst_tests-bigger.tif");

    std::unique_ptr<te::rst::Raster> inputRaster(gdma::dm::openRaster(pathRaster));
    std::vector<unsigned int> bands;

    for (int i = 0; i < inputRaster->getNumberOfBands(); i++) {
        bands.push_back(i);
    }

    std::map<std::string, te::gm::Geometry*> geom_container;
    std::map<std::string, gdma::fte::FeaturesSummary> attr_container;

    bool result;
    ASSERT_THROW({
        try {
            result = gdma::fte::SpectralFeatures::getFeatures(inputRaster.get(), geom_container,
                bands, attr_container, true);
        }
        catch (te::common::Exception &e) {
            EXPECT_STREQ(TE_TR("Geometry container received is empty."), e.what());
            throw te::common::Exception();
        }
    }, te::common::Exception);

    ASSERT_NO_FATAL_FAILURE({ te::common::FreeContents(geom_container); });
}

TEST_F(SpectralFeaturesTest, getFeaturesInvalidGeomType) {
    std::string pathRaster =
        gdma::common::FindInGeoDMAPath("/data/tests/clipped_rst_tests-bigger.tif");
    std::string pathShp =
        gdma::common::FindInGeoDMAPath("/data/tests/multiline_data.shp");

    std::unique_ptr<te::rst::Raster> inputRaster(gdma::dm::openRaster(pathRaster));
    std::vector<unsigned int> bands;

    for (int i = 0; i < inputRaster->getNumberOfBands(); i++) {
        bands.push_back(i);
    }

    std::map<std::string, te::gm::Geometry*> geom_container;
    std::string dsShp = gdma::dm::openShapefile(pathShp);
    gdma::dm::getGeometriesFromDataSource(dsShp, geom_container);

    std::map<std::string, gdma::fte::FeaturesSummary> attr_container;

    std::ostringstream excptMessage;
    excptMessage << "Invalid Geometry Type! Expected MultiPolygon or Polygon, got MultiLineString.";
    excptMessage << std::endl;

    bool result;
    ASSERT_THROW({
        try {
            result = gdma::fte::SpectralFeatures::getFeatures(inputRaster.get(), geom_container,
                bands, attr_container, true);
        }
        catch (te::common::Exception &e) {
            EXPECT_STREQ(TE_TR(excptMessage.str()), e.what());
            throw te::common::Exception();
        }
    }, te::common::Exception);

    ASSERT_NO_FATAL_FAILURE({ te::common::FreeContents(geom_container); });
}


TEST_F(SpectralFeaturesTest, getFeaturesRasterSmallerThanPolygons) {
    std::string pathRaster =
        gdma::common::FindInGeoDMAPath("/data/tests/clipped_rst_extent-small.tif");
    std::string pathShp =
        gdma::common::FindInGeoDMAPath("/data/tests/clipped_pols_tests-bigger.shp");

    std::unique_ptr<te::rst::Raster> inputRaster(gdma::dm::openRaster(pathRaster));
    std::vector<unsigned int> bands;

    for (int i = 0; i < inputRaster->getNumberOfBands(); i++) {
        bands.push_back(i);
    }

    std::map<std::string, te::gm::Geometry*> geom_container;
    std::string dsShp = gdma::dm::openShapefile(pathShp);
    gdma::dm::getGeometriesFromDataSource(dsShp, geom_container);

    std::map<std::string, gdma::fte::FeaturesSummary> attr_container;

    ASSERT_TRUE(gdma::fte::SpectralFeatures::getFeatures(inputRaster.get(), geom_container,
            bands, attr_container, true));

    for (auto &elem : attr_container) {
        std::vector<gdma::fte::Attribute> features = elem.second.getFeatures();

        if (elem.first.compare("2") == 0) {
            for (std::size_t i = 0; i < features.size(); i++)
                EXPECT_DOUBLE_EQ(-1, features[i].getAsDouble()) << features[i].getName();
        } else {
            for (std::size_t i = 0; i < features.size(); i++) {
                EXPECT_NE(-1, features[i].getAsDouble());
            }
        }
    }

    ASSERT_NO_FATAL_FAILURE({ te::common::FreeContents(geom_container); });
}
