/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/classification/classDescription.hpp

  \brief This file contains the ClassDescription class definition.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// GTest includes
#include <gtest/gtest.h>

// TerraLib includes
#include <terralib/common/TerraLib.h>
#include <terralib/dataaccess/datasource/DataSource.h>
#include <terralib/dataaccess/datasource/DataSourceInfo.h>
#include <terralib/dataaccess/utils/Utils.h>

// STL Includes
#include <map>
#include <string>

// GeoDMA Includes
#include "geodmalib/common/fileSystem.hpp"
#include "geodmalib/common/utils.hpp"
#include "geodmalib/datamanager/dataManager.hpp"
#include "geodmalib/guicomponents/decision_trees/c50ClassificationController.hpp"

class C50ClassificationControllerTest : public ::testing::Test {
 protected:
    boost::filesystem::path m_tmpFolder;

    virtual void SetUp() {
        try {
//             TerraLib::getInstance().initialize();
//             gdma::common::loadTerraLibDrivers();

            m_tmpFolder = "./tmpTests";

            boost::system::error_code error;
            if (boost::filesystem::exists(m_tmpFolder)) {
                if (!boost::filesystem::remove_all(m_tmpFolder, error)) {
                    std::string errorMsg = "Temporary folder can not be deleted: " +
                        error.message();
                    FAIL() << errorMsg;
                }
            } else {
                if (!boost::filesystem::create_directory(m_tmpFolder, error)) {
                    std::string errorMsg = "Temporary folder not created: " + error.message();
                    FAIL() << errorMsg;
                }
            }
        }
        catch (te::common::Exception &e) {
            std::cerr << "ERROR: " << e.what() << std::endl;
        }
    }

    virtual void TearDown() {
//         gdma::common::unloadTerraLibDrivers();
//         TerraLib::getInstance().finalize();

        boost::system::error_code error;
        if (boost::filesystem::exists(m_tmpFolder)) {
            if (!boost::filesystem::remove_all(m_tmpFolder, error)) {
                std::string errorMsg = "Temporary folder not deleted: " + error.message();
                FAIL() << errorMsg;
            }
        }
    }
};

// TEST_F(C50ClassificationControllerTest, testCreateLayer) {
//    gdma::guicmp::C50ClassificationController controller;
//
//    std::string inputSHP = gdma::common::FindInGeoDMAPath(
//        "data/tests/clipped_pols_tests-Samples_Feat.shp");
//
//    ASSERT_NO_THROW({ controller.createLayer(inputSHP, "teste"); });
// }

TEST_F(C50ClassificationControllerTest, saveNewSHPAndCreateLayer) {
    gdma::guicmp::C50ClassificationController controller;

    std::string inputSHP = gdma::common::FindInGeoDMAPath(
        "data/tests/clipped_pols_tests-Samples_Feat.shp");
    std::string path_output_SHP = m_tmpFolder.string() + std::string("/outputSHP.shp");

    te::da::DataSourceInfoPtr dsInfo = gdma::dm::openDataSource(inputSHP);
    te::da::DataSourcePtr input_data_source = te::da::GetDataSource(dsInfo->getId());
    std::string dataset_name = input_data_source->getDataSetNames()[0];
    // Retrieve the datasetType
    te::da::DataSetTypePtr dataSetTypePtr = input_data_source->getDataSetType(dataset_name);

    std::map<std::string, gdma::fte::FeaturesSummary> attr_container;

    for (int i = 0; i < 5; i++) {
        gdma::fte::FeaturesSummary feat;
        feat.addFeature("test_sv", "test_sv", 2.0);
        attr_container[std::to_string(i)] = feat;
    }

    ASSERT_TRUE(gdma::dm::saveToOGRDataSource(path_output_SHP, dataset_name, dsInfo->getId(),
        attr_container));

    ASSERT_TRUE(boost::filesystem::exists(boost::filesystem::path(path_output_SHP)));

    // ASSERT_NO_THROW({ controller.createLayer(path_output_SHP, "teste"); });
}
