/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/guicomponents/kmeans/kmeansClassificationController.cpp

  \brief This file contains the KmeansClassificationController class definition.

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// GTest Includes
#include <gtest/gtest.h>

// TerraLib Includes
#include <terralib/common/TerraLib.h>

// Boost Includes
#include <boost/filesystem.hpp>

// STL Includes
#include <string>
#include <vector>

// GeoDMA Includes
#include "geodmalib/common/utils.hpp"
#include "geodmalib/common/fileSystem.hpp"
#include "geodmalib/guicomponents/layerUtils.hpp"
#include "geodmalib/guicomponents/kmeans/kmeansClassificationController.hpp"

class KmeansClassificationControllerTest : public ::testing::Test {
 protected:
    boost::filesystem::path m_tmpFolder;

    virtual void SetUp() {
        try {
//             TerraLib::getInstance().initialize();
//             gdma::common::loadTerraLibDrivers();

            m_tmpFolder = "./tmpTests";

            boost::system::error_code error;
            if (boost::filesystem::exists(m_tmpFolder)) {
                if (!boost::filesystem::remove_all(m_tmpFolder, error)) {
                    std::string errorMsg = "Temporary folder can not be deleted: " +
                        error.message();
                    FAIL() << errorMsg;
                }
            } else {
                if (!boost::filesystem::create_directory(m_tmpFolder, error)) {
                    std::string errorMsg = "Temporary folder not created: " + error.message();
                    FAIL() << errorMsg;
                }
            }
        } catch (te::common::Exception &e) {
            std::cerr << "ERROR: " << e.what() << std::endl;
        }
    }

    virtual void TearDown() {
//         gdma::common::unloadTerraLibDrivers();
//         TerraLib::getInstance().finalize();

        boost::system::error_code error;
        if (boost::filesystem::exists(m_tmpFolder)) {
            if (!boost::filesystem::remove_all(m_tmpFolder, error)) {
                std::string errorMsg = "Temporary folder not deleted: " + error.message();
                FAIL() << errorMsg;
            }
        }
    }
};

TEST_F(KmeansClassificationControllerTest, setCurrentLayer) {
    std::string inputSHP = gdma::common::FindInGeoDMAPath(
        "data/tests/clipped_pols_tests-Samples_Feat.shp");

    te::map::AbstractLayerPtr testLayer;

    ASSERT_NO_THROW({
        testLayer = gdma::guicmp::createTerraLibLayer(inputSHP, "teste");
    });

    gdma::guicmp::KmeansClassificationController controller;
    controller.setCurrentLayer(testLayer);
    EXPECT_EQ("teste", controller.getCurrentLayer()->getTitle());
    EXPECT_EQ(0, controller.getNumberOfClusters());
    EXPECT_EQ(0, controller.getMaxIterations());
    EXPECT_DOUBLE_EQ(0.0, controller.getEpsilon());
}

TEST_F(KmeansClassificationControllerTest, setParameters) {
    std::string inputSHP = gdma::common::FindInGeoDMAPath(
        "data/tests/clipped_pols_tests-Samples_Feat.shp");

    te::map::AbstractLayerPtr testLayer;

    ASSERT_NO_THROW({
        testLayer = gdma::guicmp::createTerraLibLayer(inputSHP, "teste");
    });

    gdma::guicmp::KmeansClassificationController controller;
    controller.setCurrentLayer(testLayer);

    controller.setNumberOfClusters(5);
    controller.setMaxIterations(200);
    controller.setEpsilon(15.0);
    controller.setPercentTrainEntries(60.0);

    EXPECT_EQ(5, controller.getNumberOfClusters());
    EXPECT_EQ(200, controller.getMaxIterations());
    EXPECT_DOUBLE_EQ(15, controller.getEpsilon());
    EXPECT_EQ(4, controller.getNumberOfTrainEntries());

    std::vector<std::string> featToUse;
    featToUse.push_back("AMPL_B0");
    featToUse.push_back("MEAN_B0");
    featToUse.push_back("AMPL_B1");
    featToUse.push_back("MEAN_B1");
    featToUse.push_back("MEAN_B2");

    controller.setFeaturesToUse(featToUse);

    std::vector<unsigned int> featUsed = controller.getFeaturesToUse();
    EXPECT_EQ(0, featUsed[0]);
    EXPECT_EQ(0, featUsed[1]);
    EXPECT_EQ(0, featUsed[2]);
    EXPECT_EQ(0, featUsed[3]);
    EXPECT_EQ(0, featUsed[4]);
}

