/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/tests/unittest/guicomponents/sampleSelectionController_unittest.cpp

  \brief

  \author Raian Vargas Maretto
  \author Thales Sehn Korting
  \author Emiliano Ferreira Castejon
*/

// TerraLib includes
#include <terralib/common/TerraLib.h>
#include <terralib/dataaccess.h>
#include <terralib/dataaccess/utils/Utils.h>
#include <terralib/maptools/AbstractLayer.h>
#include <terralib/qt/widgets/layer/utils/DataSet2Layer.h>

// GTest Includes
#include <gtest/gtest.h>
#include <gmock/gmock.h>

// Boost Includes
#include <boost/filesystem.hpp>

// STL Includes
#include <map>
#include <memory>
#include <set>
#include <string>
#include <vector>

// GeoDMA Includes
#include "geodmalib/common/abstractMessageObserver.hpp"
#include "geodmalib/common/fileSystem.hpp"
#include "geodmalib/common/messageMediator.hpp"
#include "geodmalib/common/utils.hpp"
#include "geodmalib/datamanager/dataManager.hpp"
#include "geodmalib/guicomponents/sampleSelectionController.hpp"

// Mock class to simulate Observers
class MockObserver : public gdma::common::AbstractMessageObserver {
 public:
    MockObserver() : AbstractMessageObserver() {}
    MOCK_METHOD1(displayInformationMsg, void(std::string msg));
    MOCK_METHOD1(displayWarningMsg, void(std::string msg));
    MOCK_METHOD1(displayErrorMsg, void(std::string msg));
    MOCK_METHOD1(displayDebugMsg, void(std::string msg));
    MOCK_METHOD1(displayTraceMsg, void(std::string msg));
    MOCK_METHOD1(displayFatalMsg, void(std::string msg));
};


class SampleSelectionControllerTest : public ::testing::Test {
 protected:
    boost::filesystem::path m_tmpFolder;
    std::unique_ptr<MockObserver> m_obs;
    int m_obsId;
    gdma::guicmp::SampleSelectionController m_controller;
    te::map::AbstractLayerPtr m_layer;
    std::string m_dsId;

    virtual void SetUp() {
        try {
//             TerraLib::getInstance().initialize();
//             gdma::common::loadTerraLibDrivers();

            m_tmpFolder = "./tmpTests";

            m_obs.reset(new MockObserver());
            m_obsId = gdma::common::MessageMediator::getInstance().addObserver(m_obs.get());

            boost::system::error_code error;
            if (boost::filesystem::exists(m_tmpFolder)) {
                if (!boost::filesystem::remove_all(m_tmpFolder, error)) {
                    std::string errorMsg = "Temporary folder can not be deleted: " +
                        error.message();
                    FAIL() << errorMsg;
                }
            } else {
                if (!boost::filesystem::create_directory(m_tmpFolder, error)) {
                    std::string errorMsg = "Temporary folder not created: " + error.message();
                    FAIL() << errorMsg;
                }
            }

            std::string inputFile = gdma::common::FindInGeoDMAPath(
                "data/tests/clipped_pols_tests-bigger.shp");
            std::string copiedFile = m_tmpFolder.string() + "/clipped_test.shp";

            if (!gdma::dm::diskVectorFileCopy(inputFile, copiedFile)) {
                FAIL() << "Could not create file copy!";
            }

            m_dsId = gdma::dm::openShapefile(copiedFile);
            te::da::DataSourcePtr outDataSource = te::da::GetDataSource(m_dsId);
            te::qt::widgets::DataSet2Layer converter(m_dsId);
            te::da::DataSetTypePtr dt =
                outDataSource->getDataSetType(boost::filesystem::basename(copiedFile.c_str()));
            m_layer = converter(dt);

            m_controller.setCurrentLayer(m_layer);
        } catch (te::common::Exception &e) {
            std::cerr << "ERROR: " << e.what() << std::endl;
        }
    }

    virtual void TearDown() {
//         gdma::common::unloadTerraLibDrivers();
//         TerraLib::getInstance().finalize();

        gdma::common::MessageMediator::getInstance().removeObserver(m_obsId);

        boost::system::error_code error;
        if (boost::filesystem::exists(m_tmpFolder)) {
            if (!boost::filesystem::remove_all(m_tmpFolder, error)) {
                std::string errorMsg = "Temporary folder not deleted: " + error.message();
                FAIL() << errorMsg;
            }
        }
    }

    void selectSamplesInLayer(std::vector<std::size_t> dsPositions) {
        // Let's generate the oids
        std::unique_ptr<const te::map::LayerSchema> schema(m_layer->getSchema());
        te::da::ObjectIdSet* oids = 0;
        te::da::GetEmptyOIDSet(schema.get(), oids);
        std::vector<std::string> pnames;
        te::da::GetOIDPropertyNames(schema.get(), pnames);

        std::string dataSetName = m_layer->getDataSetName();
        te::da::DataSourcePtr dataSource = te::da::GetDataSource(m_layer->getDataSourceId());
        te::da::DataSourceTransactorPtr transactor = dataSource->getTransactor();
        te::da::DataSetPtr dataSet = transactor->getDataSet(dataSetName);

        for (auto & pos : dsPositions) {
            dataSet->move(pos);
            oids->add(te::da::GenerateOID(dataSet.get(), pnames));
        }

        m_layer->clearSelected();
        m_layer->select(oids);
    }
};

TEST_F(SampleSelectionControllerTest, changeLayerSelection) {
    m_controller.addNewClass("teste", "#fffafa");
    EXPECT_EQ(1, m_controller.getNumberOfClasses());

    std::vector<std::size_t> selectedObjs = { 0, 2, 3 };
    selectSamplesInLayer(selectedObjs);
    m_controller.addSelectedSamples("teste");
    EXPECT_EQ(3, m_controller.getNumberOfSamples("teste"));

    std::multimap<std::string, std::string> selectedIDs = {
        { "teste", "0" },
        { "teste", "2" },
        { "teste", "3" }};

    m_controller.selectSamplesInLayer(selectedIDs);

    const te::da::ObjectIdSet* oidSet = m_layer->getSelected();

    EXPECT_EQ(3, oidSet->size());
    std::set<te::da::ObjectId*>::const_iterator it = oidSet->begin();

    EXPECT_STREQ("0", (*it)->getValueAsString().c_str());
    it++;
    EXPECT_STREQ("2", (*it)->getValueAsString().c_str());
    it++;
    EXPECT_STREQ("3", (*it)->getValueAsString().c_str());

    m_controller.clearLayerSelection();
    oidSet = m_layer->getSelected();
    EXPECT_EQ(0, oidSet->size());
}

TEST_F(SampleSelectionControllerTest, createSampleColumn) {
    EXPECT_STREQ("clipped_test", m_layer->getTitle().c_str());

    std::vector<te::gm::Geometry*> geom_vec;
    boost::bimap<std::string, std::size_t> positions;

    ASSERT_NO_THROW({
        gdma::dm::getGeometriesFromDataSource(m_dsId, positions, geom_vec);
    });

    ASSERT_TRUE(m_controller.createSampleColumn("trainClass"));

    std::vector<std::string> prop_vec;
    geom_vec.clear();
    positions.clear();
    std::set<std::string> exist_classes;
    std::string prop_class("trainClass");

    ASSERT_NO_THROW({
        gdma::dm::getGeometriesFromDataSource(m_dsId, positions, geom_vec, prop_class, &prop_vec,
        &exist_classes);
    });

    for (auto &clazz : prop_vec)
        EXPECT_STREQ("", clazz.c_str());
}

TEST_F(SampleSelectionControllerTest, createSampleColumnBiggerThanLimit) {
    EXPECT_STREQ("clipped_test", m_layer->getTitle().c_str());

    std::vector<te::gm::Geometry*> geom_vec;
    boost::bimap<std::string, std::size_t> positions;

    ASSERT_NO_THROW({
        gdma::dm::getGeometriesFromDataSource(m_dsId, positions, geom_vec);
    });

    std::stringstream msg;
    msg << "Column name must have maximum 10 characters." << std::endl;
    msg << "Enter another name!" << std::endl;
    EXPECT_CALL(*m_obs, displayErrorMsg(msg.str())).Times(1);

    ASSERT_FALSE(m_controller.createSampleColumn("training_class"));

    std::vector<std::string> prop_vec;
    geom_vec.clear();
    positions.clear();
    std::set<std::string> exist_classes;
    std::string prop_class("training_c");

    try {
        gdma::dm::getGeometriesFromDataSource(m_dsId, positions, geom_vec, prop_class, &prop_vec,
            &exist_classes);
        FAIL() << "Expected te::dt::Exception column does not exists.";
    } catch (te::common::Exception& e) {
        EXPECT_STREQ(e.what(),
            "Column \"training_c\" does not exists in the data set!");
    } catch (...) {
        FAIL() << "Expected te::dt::Exception";
    }

    EXPECT_EQ(0, prop_vec.size());
    EXPECT_EQ(0, geom_vec.size());
    EXPECT_EQ(0, positions.size());
    EXPECT_EQ(0, exist_classes.size());
}

TEST_F(SampleSelectionControllerTest, changeSamplesColumnName) {
    m_controller.setSamplesColumnName("teste");

    EXPECT_STREQ("teste", m_controller.getSamplesColumnName().c_str());
}

TEST_F(SampleSelectionControllerTest, addOneClass) {
    m_controller.addNewClass("teste", "#fffafa");
    EXPECT_EQ(1, m_controller.getNumberOfClasses());
}

TEST_F(SampleSelectionControllerTest, removeExistentClass) {
    m_controller.addNewClass("teste1", "#fffafa");
    m_controller.addNewClass("teste2", "#fffafa");
    EXPECT_EQ(2, m_controller.getNumberOfClasses());
    m_controller.removeClass("teste1");
    EXPECT_EQ(1, m_controller.getNumberOfClasses());
}

TEST_F(SampleSelectionControllerTest, removeInexistentClass) {
    m_controller.addNewClass("teste2", "#fffafa");
    EXPECT_EQ(1, m_controller.getNumberOfClasses());
    EXPECT_CALL(*m_obs, displayErrorMsg("Class \"teste1\" does not exist!")).Times(1);
    m_controller.removeClass("teste1");
    EXPECT_EQ(1, m_controller.getNumberOfClasses());
}

TEST_F(SampleSelectionControllerTest, setClassColor) {
    m_controller.addNewClass("teste", "#fffafa");
    EXPECT_EQ(1, m_controller.getNumberOfClasses());
    EXPECT_STREQ("#fffafa", m_controller.getClassColor("teste").c_str());
    EXPECT_TRUE(m_controller.setClassColor("teste", "#ffffff"));
    EXPECT_STREQ("#ffffff", m_controller.getClassColor("teste").c_str());
}

TEST_F(SampleSelectionControllerTest, setColorInexistentClass) {
    m_controller.addNewClass("teste", "#fffafa");
    EXPECT_EQ(1, m_controller.getNumberOfClasses());

    EXPECT_CALL(*m_obs, displayErrorMsg("Class \"teste2\" does not exist!")).Times(2);

    EXPECT_STREQ("", m_controller.getClassColor("teste2").c_str());
    EXPECT_FALSE(m_controller.setClassColor("teste2", "#ffffff"));
}

TEST_F(SampleSelectionControllerTest, setClassLabel) {
    m_controller.addNewClass("teste", "#fffafa");
    EXPECT_EQ(1, m_controller.getNumberOfClasses());

    EXPECT_TRUE(m_controller.setClassLabel("teste", "teste2"));
    EXPECT_STREQ("#fffafa", m_controller.getClassColor("teste2").c_str());
}

TEST_F(SampleSelectionControllerTest, setLabelInexistentClass) {
    m_controller.addNewClass("teste", "#fffafa");
    EXPECT_EQ(1, m_controller.getNumberOfClasses());

    EXPECT_CALL(*m_obs, displayErrorMsg("Class \"teste2\" does not exist!")).Times(2);

    EXPECT_FALSE(m_controller.setClassLabel("teste2", "teste3"));
    EXPECT_STREQ("", m_controller.getClassColor("teste2").c_str());
    EXPECT_STREQ("#fffafa", m_controller.getClassColor("teste").c_str());
}

TEST_F(SampleSelectionControllerTest, addSamples) {
    m_controller.addNewClass("teste", "#fffafa");
    EXPECT_EQ(1, m_controller.getNumberOfClasses());

    std::vector<std::size_t> selectedObjs = {0, 2, 3};
    selectSamplesInLayer(selectedObjs);
    std::set<std::string> addedIds = m_controller.addSelectedSamples("teste");
    EXPECT_EQ(3, addedIds.size());
    std::set<std::string>::const_iterator itAdded = addedIds.begin();
    EXPECT_STREQ("0", itAdded->c_str());
    itAdded++;
    EXPECT_STREQ("2", itAdded->c_str());
    itAdded++;
    EXPECT_STREQ("3", itAdded->c_str());

    EXPECT_EQ(3, m_controller.getNumberOfSamples("teste"));

    std::set<std::string> selectedIDs = m_controller.getSampleIds("teste");

    std::set<std::string>::const_iterator it = selectedIDs.begin();
    EXPECT_STREQ("0", it->c_str());
    it++;
    EXPECT_STREQ("2", it->c_str());
    it++;
    EXPECT_STREQ("3", it->c_str());
}

TEST_F(SampleSelectionControllerTest, addSamplesChangingClass) {
    m_controller.addNewClass("teste", "#fffafa");
    m_controller.addNewClass("teste2", "#fffffa");
    EXPECT_EQ(2, m_controller.getNumberOfClasses());

    std::vector<std::size_t> selectedObjs = {0, 2, 3};
    selectSamplesInLayer(selectedObjs);
    std::set<std::string> addedIds = m_controller.addSelectedSamples("teste");
    EXPECT_EQ(3, addedIds.size());
    std::set<std::string>::const_iterator itAdded = addedIds.begin();
    EXPECT_STREQ("0", itAdded->c_str());
    itAdded++;
    EXPECT_STREQ("2", itAdded->c_str());
    itAdded++;
    EXPECT_STREQ("3", itAdded->c_str());

    EXPECT_EQ(3, m_controller.getNumberOfSamples("teste"));

    std::set<std::string> selectedIDs_class1 = m_controller.getSampleIds("teste");
    std::set<std::string>::const_iterator itClass1 = selectedIDs_class1.begin();
    EXPECT_STREQ("0", itClass1->c_str());
    itClass1++;
    EXPECT_STREQ("2", itClass1->c_str());
    itClass1++;
    EXPECT_STREQ("3", itClass1->c_str());

    selectedObjs.clear();
    selectedObjs = {1, 2, 4};
    selectSamplesInLayer(selectedObjs);
    addedIds = m_controller.addSelectedSamples("teste2");
    EXPECT_EQ(3, addedIds.size());
    itAdded = addedIds.begin();
    EXPECT_STREQ("1", itAdded->c_str());
    itAdded++;
    EXPECT_STREQ("2", itAdded->c_str());
    itAdded++;
    EXPECT_STREQ("4", itAdded->c_str());

    EXPECT_EQ(2, m_controller.getNumberOfSamples("teste"));
    EXPECT_EQ(3, m_controller.getNumberOfSamples("teste2"));

    selectedIDs_class1 = m_controller.getSampleIds("teste");
    itClass1 = selectedIDs_class1.begin();
    EXPECT_STREQ("0", itClass1->c_str());
    itClass1++;
    EXPECT_STREQ("3", itClass1->c_str());
    itClass1++;
    EXPECT_TRUE(itClass1 == selectedIDs_class1.end());

    std::set<std::string> selectedIDs_class2 = m_controller.getSampleIds("teste2");
    std::set<std::string>::const_iterator itClass2 = selectedIDs_class2.begin();
    EXPECT_STREQ("1", itClass2->c_str());
    itClass2++;
    EXPECT_STREQ("2", itClass2->c_str());
    itClass2++;
    EXPECT_STREQ("4", itClass2->c_str());
}

TEST_F(SampleSelectionControllerTest, addSamplesInexistentClass) {
    m_controller.addNewClass("teste", "#fffafa");
    EXPECT_EQ(1, m_controller.getNumberOfClasses());

    std::vector<std::size_t> selectedObjs = { 0, 2, 3 };
    selectSamplesInLayer(selectedObjs);
    EXPECT_CALL(*m_obs, displayErrorMsg("Class \"teste1\" does not exist!")).Times(2);
    m_controller.addSelectedSamples("teste1");
    EXPECT_EQ(0, m_controller.getNumberOfSamples("teste1"));
    EXPECT_EQ(0, m_controller.getNumberOfSamples("teste"));

    std::set<std::string> selectedIDs = m_controller.getSampleIds("teste");

    EXPECT_EQ(0, selectedIDs.size());
}

TEST_F(SampleSelectionControllerTest, removeSamples) {
    m_controller.addNewClass("teste", "#fffafa");
    EXPECT_EQ(1, m_controller.getNumberOfClasses());

    std::vector<std::size_t> selectedObjs = { 0, 2, 3 };
    selectSamplesInLayer(selectedObjs);
    m_controller.addSelectedSamples("teste");
    EXPECT_EQ(3, m_controller.getNumberOfSamples("teste"));

    std::set<std::string> selectedIDs = m_controller.getSampleIds("teste");
    std::set<std::string>::const_iterator it = selectedIDs.begin();
    EXPECT_STREQ("0", it->c_str());
    it++;
    EXPECT_STREQ("2", it->c_str());
    it++;
    EXPECT_STREQ("3", it->c_str());

    m_controller.removeSample("teste", "2");
    EXPECT_EQ(2, m_controller.getNumberOfSamples("teste"));

    selectedIDs = m_controller.getSampleIds("teste");
    it = selectedIDs.begin();
    EXPECT_STREQ("0", it->c_str());
    it++;
    EXPECT_STREQ("3", it->c_str());
}

TEST_F(SampleSelectionControllerTest, getPolygonSamples) {
    m_controller.addNewClass("teste", "#fffafa");
    m_controller.addNewClass("teste2", "#fffffa");
    EXPECT_EQ(2, m_controller.getNumberOfClasses());

    std::vector<std::size_t> selectedObjs = { 0, 2, 3 };
    selectSamplesInLayer(selectedObjs);
    m_controller.addSelectedSamples("teste");
    EXPECT_EQ(3, m_controller.getNumberOfSamples("teste"));

    selectedObjs.clear();
    selectedObjs = { 1, 2, 4 };
    selectSamplesInLayer(selectedObjs);
    m_controller.addSelectedSamples("teste2");
    EXPECT_EQ(2, m_controller.getNumberOfSamples("teste"));
    EXPECT_EQ(3, m_controller.getNumberOfSamples("teste2"));

    std::multimap<std::string, te::gm::GeometryShrPtr> geometrySamples =
        m_controller.getSampleGeometries();
    EXPECT_EQ(5, geometrySamples.size());
    std::multimap<std::string, te::gm::GeometryShrPtr>::iterator it = geometrySamples.begin();
    EXPECT_STREQ("#fffafa", m_controller.getClassColor(it->first).c_str());
    EXPECT_EQ(te::gm::MultiPolygonType, it->second->getGeomTypeId());
    it++;
    EXPECT_STREQ("#fffafa", m_controller.getClassColor(it->first).c_str());
    EXPECT_EQ(te::gm::MultiPolygonType, it->second->getGeomTypeId());
    it++;
    EXPECT_STREQ("#fffffa", m_controller.getClassColor(it->first).c_str());
    EXPECT_EQ(te::gm::MultiPolygonType, it->second->getGeomTypeId());
    it++;
    EXPECT_STREQ("#fffffa", m_controller.getClassColor(it->first).c_str());
    EXPECT_EQ(te::gm::MultiPolygonType, it->second->getGeomTypeId());
    it++;
    EXPECT_STREQ("#fffffa", m_controller.getClassColor(it->first).c_str());
    EXPECT_EQ(te::gm::MultiPolygonType, it->second->getGeomTypeId());
    it++;
    EXPECT_TRUE(it == geometrySamples.end());
}

TEST_F(SampleSelectionControllerTest, getPolygonSamplesEmptyClass) {
    m_controller.addNewClass("teste", "#fffafa");
    m_controller.addNewClass("teste1", "#ffffff");
    m_controller.addNewClass("teste2", "#fffffa");
    EXPECT_EQ(3, m_controller.getNumberOfClasses());

    std::vector<std::size_t> selectedObjs = { 0, 2, 3 };
    selectSamplesInLayer(selectedObjs);
    m_controller.addSelectedSamples("teste");
    EXPECT_EQ(3, m_controller.getNumberOfSamples("teste"));

    selectedObjs.clear();
    selectedObjs = { 1, 2, 4 };
    selectSamplesInLayer(selectedObjs);
    m_controller.addSelectedSamples("teste2");
    EXPECT_EQ(2, m_controller.getNumberOfSamples("teste"));
    EXPECT_EQ(0, m_controller.getNumberOfSamples("teste1"));
    EXPECT_EQ(3, m_controller.getNumberOfSamples("teste2"));

    std::multimap<std::string, te::gm::GeometryShrPtr> geometrySamples =
        m_controller.getSampleGeometries();
    EXPECT_EQ(5, geometrySamples.size());
    std::multimap<std::string, te::gm::GeometryShrPtr>::iterator it = geometrySamples.begin();
    EXPECT_STREQ("#fffafa", m_controller.getClassColor(it->first).c_str());
    EXPECT_EQ(te::gm::MultiPolygonType, it->second->getGeomTypeId());
    it++;
    EXPECT_STREQ("#fffafa", m_controller.getClassColor(it->first).c_str());
    EXPECT_EQ(te::gm::MultiPolygonType, it->second->getGeomTypeId());
    it++;
    EXPECT_STREQ("#fffffa", m_controller.getClassColor(it->first).c_str());
    EXPECT_EQ(te::gm::MultiPolygonType, it->second->getGeomTypeId());
    it++;
    EXPECT_STREQ("#fffffa", m_controller.getClassColor(it->first).c_str());
    EXPECT_EQ(te::gm::MultiPolygonType, it->second->getGeomTypeId());
    it++;
    EXPECT_STREQ("#fffffa", m_controller.getClassColor(it->first).c_str());
    EXPECT_EQ(te::gm::MultiPolygonType, it->second->getGeomTypeId());
    it++;
    EXPECT_TRUE(it == geometrySamples.end());
}

TEST_F(SampleSelectionControllerTest, getLayerSelectedGeometries) {
    m_controller.addNewClass("teste", "#fffafa");
    EXPECT_EQ(1, m_controller.getNumberOfClasses());

    std::vector<std::size_t> selectedObjs = { 0, 2, 3 };
    selectSamplesInLayer(selectedObjs);
    std::vector<te::gm::GeometryShrPtr> vecGeoms = m_controller.getLayerSelectedGeometries();
    EXPECT_EQ(3, vecGeoms.size());
    for (auto &geom : vecGeoms)
        EXPECT_EQ(te::gm::MultiPolygonType, geom->getGeomTypeId());

    m_controller.addSelectedSamples("teste");
    EXPECT_EQ(3, m_controller.getNumberOfSamples("teste"));

    vecGeoms = m_controller.getLayerSelectedGeometries();
    EXPECT_EQ(0, vecGeoms.size());
}

TEST_F(SampleSelectionControllerTest, getLayerSelectedIds) {
    m_controller.addNewClass("teste", "#fffafa");
    EXPECT_EQ(1, m_controller.getNumberOfClasses());

    std::vector<std::size_t> selectedObjs = { 0, 2, 3 };
    selectSamplesInLayer(selectedObjs);
    std::vector<std::string> selectedIds = m_controller.getLayerSelectedIds();
    EXPECT_EQ(3, selectedIds.size());

    EXPECT_STREQ("0", selectedIds[0].c_str());
    EXPECT_STREQ("2", selectedIds[1].c_str());
    EXPECT_STREQ("3", selectedIds[2].c_str());

    m_controller.addSelectedSamples("teste");
    EXPECT_EQ(3, m_controller.getNumberOfSamples("teste"));

    selectedIds = m_controller.getLayerSelectedIds();
    EXPECT_EQ(0, selectedIds.size());
}

TEST_F(SampleSelectionControllerTest, saveSamplesInDataSource) {
    m_controller.addNewClass("teste", "#fffafa");
    m_controller.addNewClass("teste2", "#fffffa");
    EXPECT_EQ(2, m_controller.getNumberOfClasses());

    std::vector<std::size_t> selectedObjs = { 0, 2, 3 };
    selectSamplesInLayer(selectedObjs);
    std::set<std::string> addedIds = m_controller.addSelectedSamples("teste");
    EXPECT_EQ(3, addedIds.size());

    EXPECT_EQ(3, m_controller.getNumberOfSamples("teste"));

    selectedObjs.clear();
    selectedObjs = { 1, 2, 4 };
    selectSamplesInLayer(selectedObjs);
    addedIds = m_controller.addSelectedSamples("teste2");
    EXPECT_EQ(3, addedIds.size());

    EXPECT_EQ(2, m_controller.getNumberOfSamples("teste"));
    EXPECT_EQ(3, m_controller.getNumberOfSamples("teste2"));

    m_controller.setSamplesColumnName("tclass");

    EXPECT_TRUE(m_controller.saveSamplesInDataSource());

    te::da::DataSourcePtr dataSource = te::da::GetDataSource(m_layer->getDataSourceId());
    te::da::DataSourceTransactorPtr transactor = dataSource->getTransactor();
    te::da::DataSetPtr dataset = transactor->getDataSet(m_layer->getDataSetName());

    ASSERT_TRUE(dataset->moveFirst());
    EXPECT_STREQ("teste", dataset->getAsString("tclass").c_str());
    ASSERT_TRUE(dataset->moveNext());
    EXPECT_STREQ("teste2", dataset->getAsString("tclass").c_str());
    ASSERT_TRUE(dataset->moveNext());
    EXPECT_STREQ("teste2", dataset->getAsString("tclass").c_str());
    ASSERT_TRUE(dataset->moveNext());
    EXPECT_STREQ("teste", dataset->getAsString("tclass").c_str());
    ASSERT_TRUE(dataset->moveNext());
    EXPECT_STREQ("teste2", dataset->getAsString("tclass").c_str());
    ASSERT_TRUE(dataset->moveNext());
    EXPECT_STREQ("", dataset->getAsString("tclass").c_str());
    ASSERT_FALSE(dataset->moveNext());
}

TEST_F(SampleSelectionControllerTest, saveSamplesInCSV) {
    m_controller.addNewClass("teste", "#fffafa");
    m_controller.addNewClass("teste2", "#fffffa");
    EXPECT_EQ(2, m_controller.getNumberOfClasses());

    std::vector<std::size_t> selectedObjs = { 0, 2, 3 };
    selectSamplesInLayer(selectedObjs);
    std::set<std::string> addedIds = m_controller.addSelectedSamples("teste");
    EXPECT_EQ(3, addedIds.size());

    EXPECT_EQ(3, m_controller.getNumberOfSamples("teste"));

    selectedObjs.clear();
    selectedObjs = { 1, 2, 4 };
    selectSamplesInLayer(selectedObjs);
    addedIds = m_controller.addSelectedSamples("teste2");
    EXPECT_EQ(3, addedIds.size());

    EXPECT_EQ(2, m_controller.getNumberOfSamples("teste"));
    EXPECT_EQ(3, m_controller.getNumberOfSamples("teste2"));

    std::string pathCSV = m_tmpFolder.string() + std::string("/csvSamples.csv");

    m_controller.setSamplesColumnName("tclass");

    ASSERT_TRUE(m_controller.saveSamplesInCSV(pathCSV));

    std::string pathTemplateFile =
        gdma::common::FindInGeoDMAPath("data/tests/csvSamplesTemplate.csv");
    std::ifstream templateFile;
    templateFile.open(pathTemplateFile, std::ios::in);

    std::ifstream generatedFile;
    generatedFile.open(pathCSV.c_str(), std::ios::in);

    std::string templateStr;
    std::string generatedStr;

    std::getline(templateFile, templateStr);
    std::getline(generatedFile, generatedStr);
    EXPECT_STREQ(templateStr.c_str(), generatedStr.c_str());

    while (!templateFile.eof()) {
        std::getline(templateFile, templateStr);
        std::getline(generatedFile, generatedStr);
        EXPECT_STREQ(templateStr.c_str(), generatedStr.c_str());
    }

    EXPECT_TRUE(templateFile.eof());
    EXPECT_TRUE(generatedFile.eof());
}

TEST_F(SampleSelectionControllerTest, loadSamplesFromDataSource) {
    m_controller.setSamplesColumnName("trainClass");
    EXPECT_CALL(*m_obs,
        displayErrorMsg("Column \"trainClass\" does not exist in the data set!")).Times(1);
    ASSERT_FALSE(m_controller.loadSamplesFromDataSource());

    EXPECT_EQ(0, m_controller.getNumberOfClasses());
    EXPECT_EQ(0, m_controller.getTotalNumberOfSamples());
}
