/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/semanticanalysis/semanticanalysis.hpp
*/

// GTest includes
#include <gtest/gtest.h>

// TerraLib includes
#include <terralib/common.h>

// STL Includes
#include <memory>
#include <vector>
#include <string>
#include <mutex>
#include <algorithm>
#include <thread>

// GeoDMA includes
#include "BuildConfig.hpp"
#include "geodmalib/semanticanalysis/executerCallBack.hpp"
#include "geodmalib/semanticanalysis/executerFactory.hpp"
#include "geodmalib/semanticanalysis/projectIOFactory.hpp"
#include "geodmalib/semanticanalysis/project.hpp"
#include "geodmalib/common/fileSystem.hpp"
#include "geodmalib/common/utils.hpp"

class SemanticAnalysisTest : public ::testing::Test {
 protected:
    virtual void SetUp() {
//         try {
//             TerraLib::getInstance().initialize();
//             gdma::common::loadTerraLibDrivers();
//         }
//         catch (te::common::Exception &e) {
//             std::cerr << "ERROR: " << e.what() << std::endl;
//         }
    }

    virtual void TearDown() {
//         gdma::common::unloadTerraLibDrivers();
//         TerraLib::getInstance().finalize();
    }
};

class ExecuterCallBackTest : public gdma::sa::ExecuterCallBack {
 public :
    ExecuterCallBackTest() {}

    ~ExecuterCallBackTest() {}

    void verticeTaskStarted( const std::string& analysisName,
      const std::string& graphName, const std::string& verticeName )
    {
      std::lock_guard<std::mutex> lock( m_mutex );

      std::cout << std::endl << "Execution started: Analysis [" +
        analysisName + "] - Graph [" + graphName + "] - Vertice [" +
        verticeName + "]" << std::endl;

      m_startedAnalysisNames.push_back( analysisName );
      m_startedGraphNames.push_back( graphName );
      m_startedVerticeNames.push_back( verticeName );
    };

    void verticeTaskExecuted( const std::string& analysisName,
      const std::string& graphName, const std::string& verticeName,
      const gdma::tp::TaskExecutionResult::BasicResult& basicResut )
    {
      std::lock_guard<std::mutex> lock( m_mutex );

      std::cout << std::endl << "Execution finished: Analysis [" +
        analysisName + "] - Graph [" + graphName + "] - Vertice [" +
        verticeName + "] BasicResult:[]" <<
        basicResut << "]" << std::endl;

      m_executedAnalysisNames.push_back( analysisName );
      m_executedGraphNames.push_back( graphName );
      m_executedVerticeNames.push_back( verticeName );
    };

    std::mutex m_mutex;
    std::vector< std::string > m_startedAnalysisNames;
    std::vector< std::string > m_startedGraphNames;
    std::vector< std::string > m_startedVerticeNames;
    std::vector< std::string > m_executedAnalysisNames;
    std::vector< std::string > m_executedGraphNames;
    std::vector< std::string > m_executedVerticeNames;
};

TEST_F(SemanticAnalysisTest, textfileprojectload)
{
  gdma::sa::Project proj;
  std::string errorMessage;
  ASSERT_TRUE( gdma::sa::ProjectIOFactory::load( gdma::common::FindInGeoDMAPath(
    "data/semanticanalysis/project_example.txt" ), proj, errorMessage ) );
}

TEST_F(SemanticAnalysisTest, singleSegmenterVerticeTest)
{
  gdma::sa::Project proj;
  std::string errorMessage;
  ASSERT_TRUE( gdma::sa::ProjectIOFactory::load( gdma::common::FindInGeoDMAPath(
    "data/semanticanalysis/project_segmenter_test1.txt" ), proj, errorMessage ) );

  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" ), (gdma::sa::Analysis*)0 );
  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name" ), (gdma::sa::GraphVertice*)0 );

  proj.enableCache( false );
  proj.setOutputDirectoryName( "./singleSegmenterVerticeTest" );

  ExecuterCallBackTest callBack;

  std::unique_ptr< gdma::sa::Executer > execPtr( gdma::sa::ExecuterFactory::make(
    "LOCAL" ) );
  ASSERT_TRUE( execPtr.get() != NULL );
  ASSERT_TRUE( execPtr->initialize( proj, gdma::sa::Executer::SINGLE ) );
  execPtr->setCallBack( callBack );

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );

  execPtr->startExecution();  // 1st pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );

  ASSERT_EQ( callBack.m_executedAnalysisNames.size(), (std::size_t)1 );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 0 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames.size(), (std::size_t)1 );
  ASSERT_EQ( callBack.m_executedGraphNames[ 0 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedVerticeNames.size(), (std::size_t)1 );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 0 ], "Regions Vertice Name" );

  execPtr->startExecution();  // 2nd pass task (due gdma::sa::Executer::SINGLE)

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );

  ASSERT_EQ( callBack.m_executedAnalysisNames.size(), (std::size_t)2 );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 0 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 1 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames.size(), (std::size_t)2 );
  ASSERT_EQ( callBack.m_executedGraphNames[ 0 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 1 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedVerticeNames.size(), (std::size_t)2 );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 0 ], "Regions Vertice Name" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 1 ], "Regions Vertice Name" );
}

TEST_F(SemanticAnalysisTest, multipleSegmenterVerticesTest)
{
  gdma::sa::Project proj;
  std::string errorMessage;
  ASSERT_TRUE( gdma::sa::ProjectIOFactory::load( gdma::common::FindInGeoDMAPath(
    "data/semanticanalysis/project_segmenter_test2.txt" ), proj, errorMessage ) );

  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" ), (gdma::sa::Analysis*)0 );
  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" ), (gdma::sa::GraphVertice*)0 );
  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" ), (gdma::sa::GraphVertice*)0 );
  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 3" ), (gdma::sa::GraphVertice*)0 );

  proj.enableCache( false );
  proj.setOutputDirectoryName( "./multipleSegmenterVerticesTest" );

  ExecuterCallBackTest callBack;

  std::unique_ptr< gdma::sa::Executer > execPtr( gdma::sa::ExecuterFactory::make(
    "LOCAL" ) );
  ASSERT_TRUE( execPtr.get() != NULL );
  ASSERT_TRUE( execPtr->initialize( proj, gdma::sa::Executer::SINGLE ) );
  execPtr->setCallBack( callBack );

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 3" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 3" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );

  execPtr->startExecution();  // vertice 1 - 1st pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 3" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 3" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );

  execPtr->startExecution();  // vertice 2 - 1st pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 3" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 3" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );

  execPtr->startExecution();  // vertice 3 - 1st pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 3" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 3" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );

  execPtr->startExecution();  // vertice 2 - 2nd pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 3" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 3" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );

  execPtr->startExecution();  // vertice 3 - 2nd pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 3" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 3" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );

  execPtr->startExecution();  // vertice 1 - 2nd pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 3" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 3" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );

  ASSERT_EQ( callBack.m_startedAnalysisNames.size(), (std::size_t)6 );
  ASSERT_EQ( callBack.m_startedGraphNames.size(), (std::size_t)6 );
  ASSERT_EQ( callBack.m_startedVerticeNames.size(), (std::size_t)6 );
  ASSERT_EQ( callBack.m_executedAnalysisNames.size(), (std::size_t)6 );
  ASSERT_EQ( callBack.m_executedGraphNames.size(), (std::size_t)6 );
  ASSERT_EQ( callBack.m_executedVerticeNames.size(), (std::size_t)6 );

  ASSERT_EQ( callBack.m_executedAnalysisNames[ 0 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 1 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 2 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 3 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 4 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 5 ], "Analysys Name 1" );

  ASSERT_EQ( callBack.m_startedAnalysisNames[ 0 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 1 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 2 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 3 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 4 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 5 ], "Analysys Name 1" );

  ASSERT_EQ( callBack.m_executedGraphNames[ 0 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 1 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 2 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 3 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 4 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 5 ], "Graph Name 1" );

  ASSERT_EQ( callBack.m_startedGraphNames[ 0 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 1 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 2 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 3 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 4 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 5 ], "Graph Name 1" );

  ASSERT_EQ( callBack.m_executedVerticeNames[ 0 ], "Regions Vertice Name 1" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 1 ], "Regions Vertice Name 2" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 2 ], "Regions Vertice Name 3" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 3 ], "Regions Vertice Name 2" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 4 ], "Regions Vertice Name 3" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 5 ], "Regions Vertice Name 1" );

  ASSERT_EQ( callBack.m_startedVerticeNames[ 0 ], "Regions Vertice Name 1" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 1 ], "Regions Vertice Name 2" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 2 ], "Regions Vertice Name 3" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 3 ], "Regions Vertice Name 2" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 4 ], "Regions Vertice Name 3" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 5 ], "Regions Vertice Name 1" );

  return;
}

TEST_F(SemanticAnalysisTest, multipleSegmenterVerticesMultiThreadTest)
{
  gdma::sa::Project proj;
  std::string errorMessage;
  ASSERT_TRUE( gdma::sa::ProjectIOFactory::load( gdma::common::FindInGeoDMAPath(
    "data/semanticanalysis/project_segmenter_test2.txt" ), proj, errorMessage ) );

  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" ), (gdma::sa::Analysis*)0 );
  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" ), (gdma::sa::GraphVertice*)0 );
  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" ), (gdma::sa::GraphVertice*)0 );
  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 3" ), (gdma::sa::GraphVertice*)0 );

  proj.enableCache( false );
  proj.setOutputDirectoryName( "./multipleSegmenterVerticesMultiThreadTest" );

  ExecuterCallBackTest callBack;

  std::unique_ptr< gdma::sa::Executer > execPtr( gdma::sa::ExecuterFactory::make(
    "LOCAL" ) );
  ASSERT_TRUE( execPtr.get() != NULL );
  ASSERT_TRUE( execPtr->initialize( proj, gdma::sa::Executer::CONCURRENT ) );
  execPtr->setCallBack( callBack );

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 3" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 3" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );

  execPtr->startExecution();

  std::unique_lock<std::mutex> callBackLock( callBack.m_mutex, std::defer_lock );

  while( true )
  {
    std::cout << std::endl << "Waiting execution" << std::endl;

    callBackLock.lock();

    if( callBack.m_executedVerticeNames.size() == 6 )
    {
      std::cout << std::endl << "All tasks executed" << std::endl;

      callBackLock.unlock();

      execPtr->stopExecution();
      ASSERT_EQ( execPtr->getExecuterStatus(), gdma::sa::Executer::INACTIVE_EXECUTER );

      std::cout << std::endl << "Executer stopped" << std::endl;

      break;
    }
    else
    {
      callBackLock.unlock();
    }

    std::this_thread::sleep_for(std::chrono::milliseconds(1000));
  }

  ASSERT_EQ( execPtr->getExecuterStatus(), gdma::sa::Executer::INACTIVE_EXECUTER );

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 3" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 3" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );

  ASSERT_EQ( callBack.m_startedAnalysisNames.size(), (std::size_t)6 );
  ASSERT_EQ( callBack.m_startedGraphNames.size(), (std::size_t)6 );
  ASSERT_EQ( callBack.m_startedVerticeNames.size(), (std::size_t)6 );
  ASSERT_EQ( callBack.m_executedAnalysisNames.size(), (std::size_t)6 );
  ASSERT_EQ( callBack.m_executedGraphNames.size(), (std::size_t)6 );
  ASSERT_EQ( callBack.m_executedVerticeNames.size(), (std::size_t)6 );

  ASSERT_EQ( callBack.m_executedAnalysisNames[ 0 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 1 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 2 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 3 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 4 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 5 ], "Analysys Name 1" );

  ASSERT_EQ( callBack.m_startedAnalysisNames[ 0 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 1 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 2 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 3 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 4 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 5 ], "Analysys Name 1" );

  ASSERT_EQ( callBack.m_executedGraphNames[ 0 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 1 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 2 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 3 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 4 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 5 ], "Graph Name 1" );

  ASSERT_EQ( callBack.m_startedGraphNames[ 0 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 1 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 2 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 3 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 4 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 5 ], "Graph Name 1" );

  return;
}

TEST_F(SemanticAnalysisTest, dummyVerticeTest)
{
  gdma::sa::Project proj;
  std::string errorMessage;
  ASSERT_TRUE( gdma::sa::ProjectIOFactory::load( gdma::common::FindInGeoDMAPath(
    "data/semanticanalysis/project_dummy_test1.txt" ), proj, errorMessage ) );

  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" ), (gdma::sa::Analysis*)0 );
  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" ), (gdma::sa::GraphVertice*)0 );
  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" ), (gdma::sa::GraphVertice*)0 );

  proj.enableCache( false );
  proj.setOutputDirectoryName( "./dummyVerticeTest" );

  ExecuterCallBackTest callBack;

  std::unique_ptr< gdma::sa::Executer > execPtr( gdma::sa::ExecuterFactory::make(
    "LOCAL" ) );
  ASSERT_TRUE( execPtr.get() != NULL );
  ASSERT_TRUE( execPtr->initialize( proj, gdma::sa::Executer::SINGLE ) );
  execPtr->setCallBack( callBack );

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );

  execPtr->startExecution();  // vertice 1 - 1st pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );


  execPtr->startExecution();  // vertice 2 - 1st pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );

  execPtr->startExecution();  // vertice 2 - 2nd pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );

  execPtr->startExecution();  // vertice 1 - 2nd pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );

  ASSERT_EQ( callBack.m_executedAnalysisNames.size(), (std::size_t)4 );
  ASSERT_EQ( callBack.m_executedGraphNames.size(), (std::size_t)4 );
  ASSERT_EQ( callBack.m_executedVerticeNames.size(), (std::size_t)4 );

  ASSERT_EQ( callBack.m_startedAnalysisNames.size(), (std::size_t)4 );
  ASSERT_EQ( callBack.m_startedGraphNames.size(), (std::size_t)4 );
  ASSERT_EQ( callBack.m_startedVerticeNames.size(), (std::size_t)4 );

  ASSERT_EQ( callBack.m_startedAnalysisNames[ 0 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 1 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 2 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 3 ], "Analysys Name 1" );

  ASSERT_EQ( callBack.m_startedAnalysisNames[ 0 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 1 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 2 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 3 ], "Analysys Name 1" );

  ASSERT_EQ( callBack.m_startedGraphNames[ 0 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 1 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 2 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 3 ], "Graph Name 1" );

  ASSERT_EQ( callBack.m_executedGraphNames[ 0 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 1 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 2 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 3 ], "Graph Name 1" );

  ASSERT_EQ( callBack.m_startedVerticeNames[ 0 ], "Regions Vertice Name 1" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 1 ], "Regions Vertice Name 2" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 2 ], "Regions Vertice Name 2" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 3 ], "Regions Vertice Name 1" );

  ASSERT_EQ( callBack.m_executedVerticeNames[ 0 ], "Regions Vertice Name 1" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 1 ], "Regions Vertice Name 2" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 2 ], "Regions Vertice Name 2" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 3 ], "Regions Vertice Name 1" );

  return;
}

TEST_F(SemanticAnalysisTest, featuresVerticeTest)
{
  gdma::sa::Project proj;
  std::string errorMessage;
  ASSERT_TRUE( gdma::sa::ProjectIOFactory::load( gdma::common::FindInGeoDMAPath(
    "data/semanticanalysis/project_features_test1.txt" ), proj, errorMessage ) );

  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" ), (gdma::sa::Analysis*)0 );
  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" ), (gdma::sa::GraphVertice*)0 );
  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" ), (gdma::sa::GraphVertice*)0 );

  proj.enableCache( false );
  proj.setOutputDirectoryName( "./featuresVerticeTest" );

  ExecuterCallBackTest callBack;

  std::unique_ptr< gdma::sa::Executer > execPtr( gdma::sa::ExecuterFactory::make(
    "LOCAL" ) );
  ASSERT_TRUE( execPtr.get() != NULL );
  ASSERT_TRUE( execPtr->initialize( proj, gdma::sa::Executer::SINGLE ) );
  execPtr->setCallBack( callBack );

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );

  execPtr->startExecution();  // vertice 1 - 1st pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );


  execPtr->startExecution();  // vertice 2 - 1st pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );

  execPtr->startExecution();  // vertice 2 - 2nd pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );

  execPtr->startExecution();  // vertice 1 - 2nd pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );

  ASSERT_EQ( callBack.m_startedAnalysisNames.size(), (std::size_t)4 );
  ASSERT_EQ( callBack.m_startedGraphNames.size(), (std::size_t)4 );
  ASSERT_EQ( callBack.m_startedVerticeNames.size(), (std::size_t)4 );

  ASSERT_EQ( callBack.m_executedAnalysisNames.size(), (std::size_t)4 );
  ASSERT_EQ( callBack.m_executedGraphNames.size(), (std::size_t)4 );
  ASSERT_EQ( callBack.m_executedVerticeNames.size(), (std::size_t)4 );

  ASSERT_EQ( callBack.m_startedAnalysisNames[ 0 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 1 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 2 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 3 ], "Analysys Name 1" );

  ASSERT_EQ( callBack.m_executedAnalysisNames[ 0 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 1 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 2 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 3 ], "Analysys Name 1" );

  ASSERT_EQ( callBack.m_startedGraphNames[ 0 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 1 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 2 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 3 ], "Graph Name 1" );

  ASSERT_EQ( callBack.m_executedGraphNames[ 0 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 1 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 2 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 3 ], "Graph Name 1" );

  ASSERT_EQ( callBack.m_startedVerticeNames[ 0 ], "Regions Vertice Name 1" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 1 ], "Regions Vertice Name 2" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 2 ], "Regions Vertice Name 2" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 3 ], "Regions Vertice Name 1" );

  ASSERT_EQ( callBack.m_executedVerticeNames[ 0 ], "Regions Vertice Name 1" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 1 ], "Regions Vertice Name 2" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 2 ], "Regions Vertice Name 2" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 3 ], "Regions Vertice Name 1" );

  return;
}

TEST_F(SemanticAnalysisTest, mapVerticeTest1)
{
  gdma::sa::Project proj;
  std::string errorMessage;
  ASSERT_TRUE( gdma::sa::ProjectIOFactory::load( gdma::common::FindInGeoDMAPath(
    "data/semanticanalysis/project_map_test1.txt" ), proj, errorMessage ) );

  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" ), (gdma::sa::Analysis*)0 );
  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" ), (gdma::sa::GraphVertice*)0 );
  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Classified Regions Vertice Name 1" ), (gdma::sa::GraphVertice*)0 );

  proj.enableCache( false );
  proj.setOutputDirectoryName( "./mapVerticeTest1" );

  ExecuterCallBackTest callBack;

  std::unique_ptr< gdma::sa::Executer > execPtr( gdma::sa::ExecuterFactory::make(
    "LOCAL" ) );
  ASSERT_TRUE( execPtr.get() != NULL );
  ASSERT_TRUE( execPtr->initialize( proj, gdma::sa::Executer::SINGLE ) );
  execPtr->setCallBack( callBack );

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Classified Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Classified Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );

  execPtr->startExecution();  // vertice 1 - 1st pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Classified Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Classified Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );


  execPtr->startExecution();  // vertice 2 - 1st pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Classified Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Classified Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );

  execPtr->startExecution();  // vertice 2 - 2nd pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Classified Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Classified Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );

  execPtr->startExecution();  // vertice 1 - 2nd pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Classified Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Classified Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );

  ASSERT_EQ( callBack.m_startedAnalysisNames.size(), (std::size_t)4 );
  ASSERT_EQ( callBack.m_startedGraphNames.size(), (std::size_t)4 );
  ASSERT_EQ( callBack.m_startedVerticeNames.size(), (std::size_t)4 );

  ASSERT_EQ( callBack.m_executedAnalysisNames.size(), (std::size_t)4 );
  ASSERT_EQ( callBack.m_executedGraphNames.size(), (std::size_t)4 );
  ASSERT_EQ( callBack.m_executedVerticeNames.size(), (std::size_t)4 );

  ASSERT_EQ( callBack.m_startedAnalysisNames[ 0 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 1 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 2 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 3 ], "Analysys Name 1" );

  ASSERT_EQ( callBack.m_executedAnalysisNames[ 0 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 1 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 2 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 3 ], "Analysys Name 1" );

  ASSERT_EQ( callBack.m_startedGraphNames[ 0 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 1 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 2 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 3 ], "Graph Name 1" );

  ASSERT_EQ( callBack.m_executedGraphNames[ 0 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 1 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 2 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 3 ], "Graph Name 1" );

  ASSERT_EQ( callBack.m_startedVerticeNames[ 0 ], "Regions Vertice Name 1" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 1 ], "Classified Regions Vertice Name 1" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 2 ], "Classified Regions Vertice Name 1" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 3 ], "Regions Vertice Name 1" );

  ASSERT_EQ( callBack.m_executedVerticeNames[ 0 ], "Regions Vertice Name 1" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 1 ], "Classified Regions Vertice Name 1" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 2 ], "Classified Regions Vertice Name 1" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 3 ], "Regions Vertice Name 1" );

  return;
}

TEST_F(SemanticAnalysisTest, mapVerticeTest2)
{
  gdma::sa::Project proj;
  std::string errorMessage;
  ASSERT_TRUE( gdma::sa::ProjectIOFactory::load( gdma::common::FindInGeoDMAPath(
    "data/semanticanalysis/project_map_test2.txt" ), proj, errorMessage ) );

  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" ), (gdma::sa::Analysis*)0 );
  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions" ), (gdma::sa::GraphVertice*)0 );
  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions With Features" ), (gdma::sa::GraphVertice*)0 );
  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Classified Regions" ), (gdma::sa::GraphVertice*)0 );

  proj.enableCache( false );
  proj.setOutputDirectoryName( "./mapVerticeTest2" );

  ExecuterCallBackTest callBack;

  std::unique_ptr< gdma::sa::Executer > execPtr( gdma::sa::ExecuterFactory::make(
    "LOCAL" ) );
  ASSERT_TRUE( execPtr.get() != NULL );
  ASSERT_TRUE( execPtr->initialize( proj, gdma::sa::Executer::SINGLE ) );
  execPtr->setCallBack( callBack );

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions With Features" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions With Features" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Classified Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Classified Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );


  execPtr->startExecution();  // vertice 1 - 1st pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions With Features" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions With Features" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Classified Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Classified Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );


  execPtr->startExecution();  // vertice 2 - 1st pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions With Features" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions With Features" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Classified Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Classified Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );

  execPtr->startExecution();  // vertice 3 - 1st pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions With Features" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions With Features" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Classified Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Classified Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );

  execPtr->startExecution();  // vertice 3 - 2nd pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions With Features" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions With Features" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Classified Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Classified Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );

  execPtr->startExecution();  // vertice 2 - 2nd pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions With Features" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions With Features" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Classified Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Classified Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );

  execPtr->startExecution();  // vertice 1 - 2nd pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions With Features" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions With Features" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Classified Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Classified Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );

  ASSERT_EQ( callBack.m_startedAnalysisNames.size(), (std::size_t)6 );
  ASSERT_EQ( callBack.m_startedGraphNames.size(), (std::size_t)6 );
  ASSERT_EQ( callBack.m_startedVerticeNames.size(), (std::size_t)6 );

  ASSERT_EQ( callBack.m_executedAnalysisNames.size(), (std::size_t)6 );
  ASSERT_EQ( callBack.m_executedGraphNames.size(), (std::size_t)6 );
  ASSERT_EQ( callBack.m_executedVerticeNames.size(), (std::size_t)6 );

  ASSERT_EQ( callBack.m_startedAnalysisNames[ 0 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 1 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 2 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 3 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 4 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 5 ], "Analysys Name 1" );

  ASSERT_EQ( callBack.m_executedAnalysisNames[ 0 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 1 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 2 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 3 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 4 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 5 ], "Analysys Name 1" );

  ASSERT_EQ( callBack.m_startedGraphNames[ 0 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 1 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 2 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 3 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 4 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 5 ], "Graph Name 1" );

  ASSERT_EQ( callBack.m_executedGraphNames[ 0 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 1 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 2 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 3 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 4 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 5 ], "Graph Name 1" );

  ASSERT_EQ( callBack.m_startedVerticeNames[ 0 ], "Regions" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 1 ], "Regions With Features" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 2 ], "Classified Regions" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 3 ], "Classified Regions" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 4 ], "Regions With Features" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 5 ], "Regions" );

  ASSERT_EQ( callBack.m_executedVerticeNames[ 0 ], "Regions" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 1 ], "Regions With Features" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 2 ], "Classified Regions" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 3 ], "Classified Regions" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 4 ], "Regions With Features" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 5 ], "Regions" );

  return;
}

TEST_F(SemanticAnalysisTest, importVerticeTest1)
{
  gdma::sa::Project proj;
  std::string errorMessage;
  ASSERT_TRUE( gdma::sa::ProjectIOFactory::load( gdma::common::FindInGeoDMAPath(
    "data/semanticanalysis/project_import_test1.txt" ), proj, errorMessage ) );

  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" ), (gdma::sa::Analysis*)0 );
  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" ), (gdma::sa::GraphVertice*)0 );
  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" ), (gdma::sa::GraphVertice*)0 );

  proj.enableCache( false );
  proj.setOutputDirectoryName( "./importVerticeTest" );

  ExecuterCallBackTest callBack;

  std::unique_ptr< gdma::sa::Executer > execPtr( gdma::sa::ExecuterFactory::make(
    "LOCAL" ) );
  ASSERT_TRUE( execPtr.get() != NULL );
  ASSERT_TRUE( execPtr->initialize( proj, gdma::sa::Executer::SINGLE ) );
  execPtr->setCallBack( callBack );

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );

  execPtr->startExecution();  // vertice 1 - 1st pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );


  execPtr->startExecution();  // vertice 2 - 1st pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );

  execPtr->startExecution();  // vertice 2 - 2nd pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );

  execPtr->startExecution();  // vertice 1 - 2nd pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );

  ASSERT_EQ( callBack.m_startedAnalysisNames.size(), (std::size_t)4 );
  ASSERT_EQ( callBack.m_startedGraphNames.size(), (std::size_t)4 );
  ASSERT_EQ( callBack.m_startedVerticeNames.size(), (std::size_t)4 );

  ASSERT_EQ( callBack.m_executedAnalysisNames.size(), (std::size_t)4 );
  ASSERT_EQ( callBack.m_executedGraphNames.size(), (std::size_t)4 );
  ASSERT_EQ( callBack.m_executedVerticeNames.size(), (std::size_t)4 );

  ASSERT_EQ( callBack.m_startedAnalysisNames[ 0 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 1 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 2 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 3 ], "Analysys Name 1" );

  ASSERT_EQ( callBack.m_executedAnalysisNames[ 0 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 1 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 2 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 3 ], "Analysys Name 1" );

  ASSERT_EQ( callBack.m_startedGraphNames[ 0 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 1 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 2 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 3 ], "Graph Name 1" );

  ASSERT_EQ( callBack.m_executedGraphNames[ 0 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 1 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 2 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 3 ], "Graph Name 1" );

  ASSERT_EQ( callBack.m_startedVerticeNames[ 0 ], "Regions Vertice Name 1" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 1 ], "Regions Vertice Name 2" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 2 ], "Regions Vertice Name 2" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 3 ], "Regions Vertice Name 1" );

  ASSERT_EQ( callBack.m_executedVerticeNames[ 0 ], "Regions Vertice Name 1" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 1 ], "Regions Vertice Name 2" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 2 ], "Regions Vertice Name 2" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 3 ], "Regions Vertice Name 1" );

  return;
}

TEST_F(SemanticAnalysisTest, scriptVerticeTest1)
{
  gdma::sa::Project proj;
  std::string errorMessage;
  ASSERT_TRUE( gdma::sa::ProjectIOFactory::load( gdma::common::FindInGeoDMAPath(
    "data/semanticanalysis/project_script_test1.txt" ), proj, errorMessage ) );

  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" ), (gdma::sa::Analysis*)0 );
  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Vertice1" ), (gdma::sa::GraphVertice*)0 );
  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Vertice2" ), (gdma::sa::GraphVertice*)0 );

  proj.enableCache( false );
  proj.setOutputDirectoryName( "./scriptVerticeTest1" );

  ExecuterCallBackTest callBack;

  std::unique_ptr< gdma::sa::Executer > execPtr( gdma::sa::ExecuterFactory::make(
    "LOCAL" ) );
  ASSERT_TRUE( execPtr.get() != NULL );
  ASSERT_TRUE( execPtr->initialize( proj, gdma::sa::Executer::SINGLE ) );
  execPtr->setCallBack( callBack );

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Vertice1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Vertice2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );

  execPtr->startExecution();  // vertice 1 - 1st pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Vertice1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Vertice2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );

  execPtr->startExecution();  // vertice 2 - 1st pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Vertice1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Vertice2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );

  ASSERT_EQ( callBack.m_startedAnalysisNames.size(), (std::size_t)2 );
  ASSERT_EQ( callBack.m_startedGraphNames.size(), (std::size_t)2 );
  ASSERT_EQ( callBack.m_startedVerticeNames.size(), (std::size_t)2 );

  ASSERT_EQ( callBack.m_executedAnalysisNames.size(), (std::size_t)2 );
  ASSERT_EQ( callBack.m_executedGraphNames.size(), (std::size_t)2 );
  ASSERT_EQ( callBack.m_executedVerticeNames.size(), (std::size_t)2 );

  ASSERT_EQ( callBack.m_startedAnalysisNames[ 0 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 1 ], "Analysys Name 1" );

  ASSERT_EQ( callBack.m_executedAnalysisNames[ 0 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 1 ], "Analysys Name 1" );

  ASSERT_EQ( callBack.m_startedGraphNames[ 0 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 1 ], "Graph Name 1" );

  ASSERT_EQ( callBack.m_executedGraphNames[ 0 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 1 ], "Graph Name 1" );

  ASSERT_EQ( callBack.m_startedVerticeNames[ 0 ], "Vertice1" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 1 ], "Vertice2" );

  ASSERT_EQ( callBack.m_executedVerticeNames[ 0 ], "Vertice1" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 1 ], "Vertice2" );

  return;
}

TEST_F(SemanticAnalysisTest, selectVerticeTest)
{
  gdma::sa::Project proj;
  std::string errorMessage;
  ASSERT_TRUE( gdma::sa::ProjectIOFactory::load( gdma::common::FindInGeoDMAPath(
    "data/semanticanalysis/project_select_test1.txt" ), proj, errorMessage ) );

  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" ), (gdma::sa::Analysis*)0 );
  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" ), (gdma::sa::GraphVertice*)0 );
  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" ), (gdma::sa::GraphVertice*)0 );
  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Selected Regions" ), (gdma::sa::GraphVertice*)0 );

  proj.enableCache( false );
  proj.setOutputDirectoryName( "./selectVerticeTest" );

  ExecuterCallBackTest callBack;

  std::unique_ptr< gdma::sa::Executer > execPtr( gdma::sa::ExecuterFactory::make(
    "LOCAL" ) );
  ASSERT_TRUE( execPtr.get() != NULL );
  ASSERT_TRUE( execPtr->initialize( proj, gdma::sa::Executer::SINGLE ) );
  execPtr->setCallBack( callBack );

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Selected Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Selected Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );

  execPtr->startExecution();  // vertice 1 - 1st pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Selected Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Selected Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );

  execPtr->startExecution();  // vertice 2 - 1st pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Selected Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Selected Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );

  execPtr->startExecution();  // vertice 3 - 1st pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Selected Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Selected Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );

  execPtr->startExecution();  // vertice 3 - 2nd pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Selected Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Selected Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );

  execPtr->startExecution();  // vertice 1 - 2nd pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Selected Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Selected Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );

  execPtr->startExecution();  // vertice 2 - 2nd pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 1" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions Vertice Name 2" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Selected Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Selected Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );

  ASSERT_EQ( callBack.m_executedAnalysisNames.size(), (std::size_t)6 );
  ASSERT_EQ( callBack.m_executedGraphNames.size(), (std::size_t)6 );
  ASSERT_EQ( callBack.m_executedVerticeNames.size(), (std::size_t)6 );

  ASSERT_EQ( callBack.m_startedAnalysisNames.size(), (std::size_t)6 );
  ASSERT_EQ( callBack.m_startedGraphNames.size(), (std::size_t)6 );
  ASSERT_EQ( callBack.m_startedVerticeNames.size(), (std::size_t)6 );

  ASSERT_EQ( callBack.m_startedAnalysisNames[ 0 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 1 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 2 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 3 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 4 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 5 ], "Analysys Name 1" );

  ASSERT_EQ( callBack.m_startedAnalysisNames[ 0 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 1 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 2 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 3 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 4 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 5 ], "Analysys Name 1" );

  ASSERT_EQ( callBack.m_startedGraphNames[ 0 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 1 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 2 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 3 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 4 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 5 ], "Graph Name 1" );

  ASSERT_EQ( callBack.m_executedGraphNames[ 0 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 1 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 2 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 3 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 4 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 5 ], "Graph Name 1" );

  ASSERT_EQ( callBack.m_startedVerticeNames[ 0 ], "Regions Vertice Name 1" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 1 ], "Regions Vertice Name 2" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 2 ], "Selected Regions" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 3 ], "Selected Regions" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 4 ], "Regions Vertice Name 1" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 5 ], "Regions Vertice Name 2" );

  ASSERT_EQ( callBack.m_executedVerticeNames[ 0 ], "Regions Vertice Name 1" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 1 ], "Regions Vertice Name 2" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 2 ], "Selected Regions" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 3 ], "Selected Regions" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 4 ], "Regions Vertice Name 1" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 5 ], "Regions Vertice Name 2" );

  return;
}

TEST_F(SemanticAnalysisTest, arithmeticVerticeTest1)
{
  gdma::sa::Project proj;
  std::string errorMessage;
  ASSERT_TRUE( gdma::sa::ProjectIOFactory::load( gdma::common::FindInGeoDMAPath(
    "data/semanticanalysis/project_arithmetic_test1.txt" ), proj, errorMessage ) );

  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" ), (gdma::sa::Analysis*)0 );
  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions" ), (gdma::sa::GraphVertice*)0 );
  ASSERT_NE( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Arithmetic Regions" ), (gdma::sa::GraphVertice*)0 );

  proj.enableCache( false );
  proj.setOutputDirectoryName( "arithmeticTest1" );

  ExecuterCallBackTest callBack;

  std::unique_ptr< gdma::sa::Executer > execPtr( gdma::sa::ExecuterFactory::make(
    "LOCAL" ) );
  ASSERT_TRUE( execPtr.get() != NULL );
  ASSERT_TRUE( execPtr->initialize( proj, gdma::sa::Executer::SINGLE ) );
  execPtr->setCallBack( callBack );

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Arithmetic Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Arithmetic Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );

  execPtr->startExecution();  // vertice 1 - 1st pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Arithmetic Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Arithmetic Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );


  execPtr->startExecution();  // vertice 2 - 1st pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Arithmetic Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Arithmetic Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );

  execPtr->startExecution();  // vertice 2 - 2nd pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::INVALID );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Arithmetic Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Arithmetic Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );

  execPtr->startExecution();  // vertice 1 - 2nd pass task

  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Arithmetic Regions" )->get1stPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );
  ASSERT_EQ( proj.getAnalysis( "Analysys Name 1" )->getGraph().getVertice( "Arithmetic Regions" )->get2ndPassTask().getVerticeTaskExecResult().getBasicResult(),
    gdma::tp::TaskExecutionResult::SUCCESS );

  ASSERT_EQ( callBack.m_startedAnalysisNames.size(), (std::size_t)4 );
  ASSERT_EQ( callBack.m_startedGraphNames.size(), (std::size_t)4 );
  ASSERT_EQ( callBack.m_startedVerticeNames.size(), (std::size_t)4 );

  ASSERT_EQ( callBack.m_executedAnalysisNames.size(), (std::size_t)4 );
  ASSERT_EQ( callBack.m_executedGraphNames.size(), (std::size_t)4 );
  ASSERT_EQ( callBack.m_executedVerticeNames.size(), (std::size_t)4 );

  ASSERT_EQ( callBack.m_startedAnalysisNames[ 0 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 1 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 2 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_startedAnalysisNames[ 3 ], "Analysys Name 1" );

  ASSERT_EQ( callBack.m_executedAnalysisNames[ 0 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 1 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 2 ], "Analysys Name 1" );
  ASSERT_EQ( callBack.m_executedAnalysisNames[ 3 ], "Analysys Name 1" );

  ASSERT_EQ( callBack.m_startedGraphNames[ 0 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 1 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 2 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_startedGraphNames[ 3 ], "Graph Name 1" );

  ASSERT_EQ( callBack.m_executedGraphNames[ 0 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 1 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 2 ], "Graph Name 1" );
  ASSERT_EQ( callBack.m_executedGraphNames[ 3 ], "Graph Name 1" );

  ASSERT_EQ( callBack.m_startedVerticeNames[ 0 ], "Regions" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 1 ], "Arithmetic Regions" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 2 ], "Arithmetic Regions" );
  ASSERT_EQ( callBack.m_startedVerticeNames[ 3 ], "Regions" );

  ASSERT_EQ( callBack.m_executedVerticeNames[ 0 ], "Regions" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 1 ], "Arithmetic Regions" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 2 ], "Arithmetic Regions" );
  ASSERT_EQ( callBack.m_executedVerticeNames[ 3 ], "Regions" );

  return;
}

