/*  Copyright (C) 2012 National Institute For Space Research (INPE) - Brazil.

This file is part of the GeoDMA - a Toolbox that integrates Data Mining Techniques with object-based
and multi-temporal analysis of satellite remotely sensed imagery.

GeoDMA is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

GeoDMA is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with GeoDMA. See COPYING. If not, write to
GeoDMA Team at <thales@dpi.inpe.br, raian@dpi.inpe.br, castejon@dpi.inpe.br>.
*/

/*!
  \file gdma/lib/taskprocessing/taskprocessing_unittest.hpp
*/

// GTest includes
#include <gtest/gtest.h>

// TerraLib includes
#include <terralib/common.h>

// STL Includes
#include <string>
#include <vector>
#include <mutex>
#include <memory>

// GeoDMA includes
#include "BuildConfig.hpp"
#include "geodmalib/taskprocessing/executer.hpp"
#include "geodmalib/taskprocessing/graphExecuter.hpp"
#include "geodmalib/taskprocessing/task.hpp"
#include "geodmalib/taskprocessing/defines.hpp"
#include "geodmalib/common/utils.hpp"
#include "geodmalib/taskprocessing/taskExecutionResult.hpp"

class CallBack1 : public gdma::tp::ExecuterCallBack {
 public:
    std::vector< gdma::tp::TaskIdType > m_startedTaskIds;
    std::vector< gdma::tp::TaskIdType > m_executedTaskIds;
    std::mutex m_mutex;

    CallBack1()
    {};

    ~CallBack1() {}

    void taskStarted(const gdma::tp::TaskIdType& taskId)
    {
      std::unique_lock<std::mutex> lock( m_mutex );
      m_startedTaskIds.push_back( taskId );
      std::cout << std::endl << "Task started - ID:" << taskId << std::endl;
    }

    void taskExecuted(const gdma::tp::TaskIdType& taskId,
      const gdma::tp::TaskExecutionResult::BasicResult& basicResut )
    {
      std::unique_lock<std::mutex> lock( m_mutex );
      m_executedTaskIds.push_back( taskId );
      std::cout << std::endl << "Task executed - ID:" << taskId <<
        " BasicResult:" << basicResut << std::endl;
    }
};

class ExecTask1 : public gdma::tp::Task {
 public:
    unsigned int& m_counter;
    std::mutex& m_mutex;
    std::unique_ptr< gdma::tp::TaskExecutionResult > m_resultPtr;

    ExecTask1( unsigned int& counter, std::mutex& mutex )
    : m_counter( counter ), m_mutex( mutex )
    {
      m_resultPtr.reset( new gdma::tp::TaskExecutionResult() );
      m_resultPtr->setBasicResult( gdma::tp::TaskExecutionResult::FAILED );
    };

    ~ExecTask1() {}

    void execute()
    {
      std::unique_lock<std::mutex> lock( m_mutex );
      ++m_counter;
      m_resultPtr->setBasicResult( gdma::tp::TaskExecutionResult::SUCCESS );
      std::cout << std::endl << "Increment done - new value:" << m_counter << std::endl;
    }

    const gdma::tp::TaskExecutionResult& getExecutionResult() const
    {
      return *m_resultPtr;
    }
};

class TaskProcessingTest : public ::testing::Test {
 protected:
    virtual void SetUp() {
//         try {
//             TerraLib::getInstance().initialize();
//             gdma::common::loadTerraLibDrivers();
//         }
//         catch (te::common::Exception &e) {
//             std::cerr << "ERROR: " << e.what() << std::endl;
//         }
    }

    virtual void TearDown() {
//         gdma::common::unloadTerraLibDrivers();
//         TerraLib::getInstance().finalize();
    }
};

TEST_F(TaskProcessingTest, executerSINGLETest) {
  std::mutex mutex;
  unsigned int counter = 0;
  CallBack1 callback;
  gdma::tp::TaskIdType taskId = 0;
  gdma::tp::Executer executer;
  gdma::tp::Executer::TaskStatus taskStatus = gdma::tp::Executer::INACTIVE;
  ExecTask1* taskPtr1 = 0;
  gdma::tp::Task* taskPtr2 = 0;

  taskPtr1 = new ExecTask1( counter, mutex );
  ASSERT_TRUE( taskPtr1->getExecutionResult().getBasicResult() == gdma::tp::TaskExecutionResult::FAILED );

  ASSERT_TRUE( executer.setExecutionParams( gdma::tp::Executer::SINGLE, 0, true ) );
  ASSERT_TRUE( executer.addTask( taskPtr1, &taskId ) );
  executer.setCallBack( callback );
  executer.startExecution();
  executer.waitExecution();
  ASSERT_TRUE( executer.getExecuterStatus() == gdma::tp::Executer::INACTIVE_EXECUTER );
  ASSERT_TRUE( executer.isTaskPresent( taskId ) == true );
  ASSERT_TRUE( executer.getTaskStatus( taskId, taskStatus ) );
  ASSERT_TRUE( taskStatus == gdma::tp::Executer::FINISHED );

  ASSERT_TRUE( executer.removeTask( taskId, &taskPtr2 ) );
  ASSERT_EQ( taskPtr1, taskPtr2 );

  ASSERT_TRUE( executer.isTaskPresent( taskId ) == false );

  ASSERT_EQ( counter, 1 );

  ASSERT_TRUE( taskPtr2->getExecutionResult().getBasicResult() == gdma::tp::TaskExecutionResult::SUCCESS );

  ASSERT_EQ( callback.m_startedTaskIds.size(), 1 );
  ASSERT_TRUE( callback.m_startedTaskIds[ 0 ] == taskId );

  ASSERT_EQ( callback.m_executedTaskIds.size(), 1 );
  ASSERT_TRUE( callback.m_executedTaskIds[ 0 ] == taskId );

  delete taskPtr2;
}

TEST_F(TaskProcessingTest, executerCONCURRENTWaitTest) {
  std::mutex mutex;
  unsigned int counter = 0;
  CallBack1 callback;
  gdma::tp::TaskIdType taskId1 = 0;
  gdma::tp::TaskIdType taskId2 = 0;
  gdma::tp::Executer executer;
  gdma::tp::Executer::TaskStatus taskStatus = gdma::tp::Executer::INACTIVE;
  ExecTask1* execTaskPtr1 = 0;
  ExecTask1* execTaskPtr2 = 0;
  gdma::tp::Task* taskPtr = 0;

  execTaskPtr1 = new ExecTask1( counter, mutex );
  ASSERT_TRUE( execTaskPtr1->getExecutionResult().getBasicResult() == gdma::tp::TaskExecutionResult::FAILED );

  execTaskPtr2 = new ExecTask1( counter, mutex );
  ASSERT_TRUE( execTaskPtr2->getExecutionResult().getBasicResult() == gdma::tp::TaskExecutionResult::FAILED );

  ASSERT_TRUE( executer.setExecutionParams( gdma::tp::Executer::CONCURRENT, 2, true ) );
  ASSERT_TRUE( executer.addTask( execTaskPtr1, &taskId1 ) );
  ASSERT_TRUE( executer.addTask( execTaskPtr2, &taskId2 ) );
  executer.setCallBack( callback );
  executer.startExecution();
  executer.waitExecution();
  ASSERT_TRUE( executer.getExecuterStatus() == gdma::tp::Executer::INACTIVE_EXECUTER );
  ASSERT_TRUE( executer.isTaskPresent( taskId1 ) == true );
  ASSERT_TRUE( executer.isTaskPresent( taskId2 ) == true );
  ASSERT_TRUE( executer.getTaskStatus( taskId1, taskStatus ) );
  ASSERT_TRUE( taskStatus == gdma::tp::Executer::FINISHED );
  ASSERT_TRUE( executer.getTaskStatus( taskId2, taskStatus ) );
  ASSERT_TRUE( taskStatus == gdma::tp::Executer::FINISHED );

  ASSERT_TRUE( executer.removeTask( taskId1, &taskPtr ) );
  ASSERT_EQ( taskPtr, execTaskPtr1 );
  ASSERT_TRUE( taskPtr->getExecutionResult().getBasicResult() == gdma::tp::TaskExecutionResult::SUCCESS );
  delete taskPtr;
  ASSERT_TRUE( executer.isTaskPresent( taskId1 ) == false );

  ASSERT_TRUE( executer.removeTask( taskId2, &taskPtr ) );
  ASSERT_EQ( taskPtr, execTaskPtr2 );
  ASSERT_TRUE( taskPtr->getExecutionResult().getBasicResult() == gdma::tp::TaskExecutionResult::SUCCESS );
  delete taskPtr;
  ASSERT_TRUE( executer.isTaskPresent( taskId2 ) == false );

  ASSERT_EQ( counter, 2 );

  ASSERT_EQ( callback.m_startedTaskIds.size(), 2 );
  ASSERT_TRUE( std::find( callback.m_startedTaskIds.begin(), callback.m_startedTaskIds.end(), taskId1 ) != callback.m_startedTaskIds.end() );
  ASSERT_TRUE( std::find( callback.m_startedTaskIds.begin(), callback.m_startedTaskIds.end(), taskId2 ) != callback.m_startedTaskIds.end() );

  ASSERT_EQ( callback.m_executedTaskIds.size(), 2 );
  ASSERT_TRUE( std::find( callback.m_executedTaskIds.begin(), callback.m_executedTaskIds.end(), taskId1 ) != callback.m_executedTaskIds.end() );
  ASSERT_TRUE( std::find( callback.m_executedTaskIds.begin(), callback.m_executedTaskIds.end(), taskId2 ) != callback.m_executedTaskIds.end() );
}

TEST_F(TaskProcessingTest, executerCONCURRENTStopTest) {
  std::mutex tasksMutex;
  unsigned int counter = 0;
  CallBack1 callback;
  gdma::tp::TaskIdType taskId1 = 0;
  gdma::tp::TaskIdType taskId2 = 0;
  gdma::tp::Executer executer;
  gdma::tp::Executer::TaskStatus taskStatus = gdma::tp::Executer::INACTIVE;
  ExecTask1* execTaskPtr1 = 0;
  ExecTask1* execTaskPtr2 = 0;
  gdma::tp::Task* taskPtr = 0;

  execTaskPtr1 = new ExecTask1( counter, tasksMutex );
  ASSERT_TRUE( execTaskPtr1->getExecutionResult().getBasicResult() == gdma::tp::TaskExecutionResult::FAILED );

  execTaskPtr2 = new ExecTask1( counter, tasksMutex );
  ASSERT_TRUE( execTaskPtr2->getExecutionResult().getBasicResult() == gdma::tp::TaskExecutionResult::FAILED );

  ASSERT_TRUE( executer.setExecutionParams( gdma::tp::Executer::CONCURRENT, 1, true ) );
  ASSERT_TRUE( executer.addTask( execTaskPtr1, &taskId1 ) );
  ASSERT_TRUE( executer.addTask( execTaskPtr2, &taskId2 ) );
  executer.setCallBack( callback );

  executer.startExecution();

  std::unique_lock<std::mutex> callBackLock( callback.m_mutex, std::defer_lock );

  while( true )
  {
    std::cout << std::endl << "Waiting execution" << std::endl;

    callBackLock.lock();

    if( callback.m_executedTaskIds.size() == 2 )
    {
      std::cout << std::endl << "All tasks executed" << std::endl;

      callBackLock.unlock();

      executer.stopExecution();
      ASSERT_EQ( executer.getExecuterStatus(), gdma::tp::Executer::INACTIVE_EXECUTER );

      std::cout << std::endl << "Executer stopped" << std::endl;

      break;
    }
    else
    {
      callBackLock.unlock();
    }

    std::this_thread::sleep_for(std::chrono::milliseconds(1000));
  }

  ASSERT_TRUE( executer.getExecuterStatus() == gdma::tp::Executer::INACTIVE_EXECUTER );
  ASSERT_TRUE( executer.isTaskPresent( taskId1 ) == true );
  ASSERT_TRUE( executer.isTaskPresent( taskId2 ) == true );
  ASSERT_TRUE( executer.getTaskStatus( taskId1, taskStatus ) );
  ASSERT_TRUE( taskStatus == gdma::tp::Executer::FINISHED );
  ASSERT_TRUE( executer.getTaskStatus( taskId2, taskStatus ) );
  ASSERT_TRUE( taskStatus == gdma::tp::Executer::FINISHED );

  ASSERT_TRUE( executer.removeTask( taskId1, &taskPtr ) );
  ASSERT_EQ( taskPtr, execTaskPtr1 );
  ASSERT_TRUE( taskPtr->getExecutionResult().getBasicResult() == gdma::tp::TaskExecutionResult::SUCCESS );
  delete taskPtr;
  ASSERT_TRUE( executer.isTaskPresent( taskId1 ) == false );

  ASSERT_TRUE( executer.removeTask( taskId2, &taskPtr ) );
  ASSERT_EQ( taskPtr, execTaskPtr2 );
  ASSERT_TRUE( taskPtr->getExecutionResult().getBasicResult() == gdma::tp::TaskExecutionResult::SUCCESS );
  delete taskPtr;
  ASSERT_TRUE( executer.isTaskPresent( taskId2 ) == false );

  ASSERT_EQ( counter, 2 );

  ASSERT_EQ( callback.m_startedTaskIds.size(), 2 );
  ASSERT_TRUE( std::find( callback.m_startedTaskIds.begin(), callback.m_startedTaskIds.end(), taskId1 ) != callback.m_startedTaskIds.end() );
  ASSERT_TRUE( std::find( callback.m_startedTaskIds.begin(), callback.m_startedTaskIds.end(), taskId2 ) != callback.m_startedTaskIds.end() );

  ASSERT_EQ( callback.m_executedTaskIds.size(), 2 );
  ASSERT_TRUE( std::find( callback.m_executedTaskIds.begin(), callback.m_executedTaskIds.end(), taskId1 ) != callback.m_executedTaskIds.end() );
  ASSERT_TRUE( std::find( callback.m_executedTaskIds.begin(), callback.m_executedTaskIds.end(), taskId2 ) != callback.m_executedTaskIds.end() );
}

TEST_F(TaskProcessingTest, graphExecuterSINGLETest) {
  std::mutex mutex;
  unsigned int counter = 0;
  CallBack1 callback;
  gdma::tp::TaskIdType taskId1 = 0;
  gdma::tp::TaskIdType taskId2 = 0;
  gdma::tp::GraphExecuter executer;
  gdma::tp::GraphExecuter::TaskStatus taskStatus = gdma::tp::GraphExecuter::INACTIVE;
  ExecTask1* execTaskPtr1 = 0;
  ExecTask1* execTaskPtr2 = 0;
  gdma::tp::Task* taskPtr = 0;
  std::vector< gdma::tp::TaskIdType > dependencies;

  execTaskPtr1 = new ExecTask1( counter, mutex );
  ASSERT_TRUE( execTaskPtr1->getExecutionResult().getBasicResult() == gdma::tp::TaskExecutionResult::FAILED );

  execTaskPtr2 = new ExecTask1( counter, mutex );
  ASSERT_TRUE( execTaskPtr2->getExecutionResult().getBasicResult() == gdma::tp::TaskExecutionResult::FAILED );

  ASSERT_TRUE( executer.setExecutionParams( gdma::tp::GraphExecuter::SINGLE, 0, true ) );
  ASSERT_TRUE( executer.addTask( execTaskPtr1, std::vector< gdma::tp::TaskIdType >(), &taskId1 ) );
  dependencies.push_back( taskId1 );
  ASSERT_TRUE( executer.addTask( execTaskPtr2, dependencies, &taskId2 ) );
  executer.setCallBack( callback );
  executer.startExecution();
  executer.startExecution();  // Executing again due to the second added task
  executer.waitExecution();
  ASSERT_TRUE( executer.getExecuterStatus() == gdma::tp::Executer::INACTIVE_EXECUTER );
  ASSERT_TRUE( executer.isTaskPresent( taskId1 ) == true );
  ASSERT_TRUE( executer.isTaskPresent( taskId2 ) == true );
  ASSERT_TRUE( executer.getTaskStatus( taskId1, taskStatus ) );
  ASSERT_TRUE( taskStatus == gdma::tp::GraphExecuter::FINISHED );
  ASSERT_TRUE( executer.getTaskStatus( taskId2, taskStatus ) );
  ASSERT_TRUE( taskStatus == gdma::tp::GraphExecuter::FINISHED );

  ASSERT_TRUE( executer.removeTask( taskId1, &taskPtr ) );
  ASSERT_EQ( taskPtr, execTaskPtr1 );
  ASSERT_TRUE( taskPtr->getExecutionResult().getBasicResult() == gdma::tp::TaskExecutionResult::SUCCESS );
  delete taskPtr;
  ASSERT_TRUE( executer.isTaskPresent( taskId1 ) == false );

  ASSERT_TRUE( executer.removeTask( taskId2, &taskPtr ) );
  ASSERT_EQ( taskPtr, execTaskPtr2 );
  ASSERT_TRUE( taskPtr->getExecutionResult().getBasicResult() == gdma::tp::TaskExecutionResult::SUCCESS );
  delete taskPtr;
  ASSERT_TRUE( executer.isTaskPresent( taskId2 ) == false );

  ASSERT_EQ( counter, 2 );

  ASSERT_EQ( callback.m_startedTaskIds.size(), 2 );
  ASSERT_EQ( callback.m_startedTaskIds[ 0 ], taskId1 );
  ASSERT_EQ( callback.m_startedTaskIds[ 1 ], taskId2 );

  ASSERT_EQ( callback.m_executedTaskIds.size(), 2 );
  ASSERT_EQ( callback.m_executedTaskIds[ 0 ], taskId1 );
  ASSERT_EQ( callback.m_executedTaskIds[ 1 ], taskId2 );
}

TEST_F(TaskProcessingTest, graphExecuterCONCURRENTWaitTest) {
  std::mutex mutex;
  unsigned int counter = 0;
  CallBack1 callback;
  gdma::tp::TaskIdType taskId1 = 0;
  gdma::tp::TaskIdType taskId2 = 0;
  gdma::tp::GraphExecuter executer;
  gdma::tp::GraphExecuter::TaskStatus taskStatus = gdma::tp::GraphExecuter::INACTIVE;
  ExecTask1* execTaskPtr1 = 0;
  ExecTask1* execTaskPtr2 = 0;
  gdma::tp::Task* taskPtr = 0;
  std::vector< gdma::tp::TaskIdType > dependencies;

  execTaskPtr1 = new ExecTask1( counter, mutex );
  ASSERT_TRUE( execTaskPtr1->getExecutionResult().getBasicResult() == gdma::tp::TaskExecutionResult::FAILED );

  execTaskPtr2 = new ExecTask1( counter, mutex );
  ASSERT_TRUE( execTaskPtr2->getExecutionResult().getBasicResult() == gdma::tp::TaskExecutionResult::FAILED );

  ASSERT_TRUE( executer.setExecutionParams( gdma::tp::GraphExecuter::CONCURRENT, 2, true ) );
  ASSERT_TRUE( executer.addTask( execTaskPtr1, std::vector< gdma::tp::TaskIdType >(), &taskId1 ) );
  dependencies.push_back( taskId1 );
  ASSERT_TRUE( executer.addTask( execTaskPtr2, dependencies, &taskId2 ) );
  executer.setCallBack( callback );
  executer.startExecution();
  executer.waitExecution();
  ASSERT_TRUE( executer.getExecuterStatus() == gdma::tp::Executer::INACTIVE_EXECUTER );
  ASSERT_EQ( executer.isTaskPresent( taskId1 ), true );
  ASSERT_EQ( executer.isTaskPresent( taskId2 ), true );
  ASSERT_TRUE( executer.getTaskStatus( taskId1, taskStatus ) );
  ASSERT_EQ( taskStatus, gdma::tp::GraphExecuter::FINISHED );
  ASSERT_TRUE( executer.getTaskStatus( taskId2, taskStatus ) );
  ASSERT_EQ( taskStatus, gdma::tp::GraphExecuter::FINISHED );

  ASSERT_TRUE( executer.removeTask( taskId1, &taskPtr ) );
  ASSERT_EQ( taskPtr, execTaskPtr1 );
  ASSERT_EQ( taskPtr->getExecutionResult().getBasicResult(), gdma::tp::TaskExecutionResult::SUCCESS );
  delete taskPtr;

  ASSERT_TRUE( executer.removeTask( taskId2, &taskPtr ) );
  ASSERT_EQ( taskPtr, execTaskPtr2 );
  ASSERT_EQ( taskPtr->getExecutionResult().getBasicResult(), gdma::tp::TaskExecutionResult::SUCCESS );
  delete taskPtr;

  ASSERT_EQ( counter, 2 );

  ASSERT_EQ( callback.m_startedTaskIds.size(), 2 );
  ASSERT_TRUE( std::find( callback.m_startedTaskIds.begin(), callback.m_startedTaskIds.end(), taskId1 ) != callback.m_startedTaskIds.end() );
  ASSERT_TRUE( std::find( callback.m_startedTaskIds.begin(), callback.m_startedTaskIds.end(), taskId2 ) != callback.m_startedTaskIds.end() );

  ASSERT_EQ( callback.m_executedTaskIds.size(), 2 );
  ASSERT_TRUE( std::find( callback.m_executedTaskIds.begin(), callback.m_executedTaskIds.end(), taskId1 ) != callback.m_executedTaskIds.end() );
  ASSERT_TRUE( std::find( callback.m_executedTaskIds.begin(), callback.m_executedTaskIds.end(), taskId2 ) != callback.m_executedTaskIds.end() );
}

TEST_F(TaskProcessingTest, graphExecuterCONCURRENTStopTest) {
  std::mutex tasksMutex;
  unsigned int counter = 0;
  CallBack1 callback;
  gdma::tp::TaskIdType taskId1 = 0;
  gdma::tp::TaskIdType taskId2 = 0;
  gdma::tp::GraphExecuter executer;
  gdma::tp::GraphExecuter::TaskStatus taskStatus = gdma::tp::GraphExecuter::INACTIVE;
  ExecTask1* execTaskPtr1 = 0;
  ExecTask1* execTaskPtr2 = 0;
  gdma::tp::Task* taskPtr = 0;
  std::vector< gdma::tp::TaskIdType > dependencies;

  execTaskPtr1 = new ExecTask1( counter, tasksMutex );
  ASSERT_TRUE( execTaskPtr1->getExecutionResult().getBasicResult() == gdma::tp::TaskExecutionResult::FAILED );

  execTaskPtr2 = new ExecTask1( counter, tasksMutex );
  ASSERT_TRUE( execTaskPtr2->getExecutionResult().getBasicResult() == gdma::tp::TaskExecutionResult::FAILED );

  ASSERT_TRUE( executer.setExecutionParams( gdma::tp::GraphExecuter::CONCURRENT, 2, true ) );
  ASSERT_TRUE( executer.addTask( execTaskPtr1, std::vector< gdma::tp::TaskIdType >(), &taskId1 ) );
  dependencies.push_back( taskId1 );
  ASSERT_TRUE( executer.addTask( execTaskPtr2, dependencies, &taskId2 ) );
  executer.setCallBack( callback );
  executer.startExecution();

  std::unique_lock<std::mutex> callBackLock( callback.m_mutex, std::defer_lock );

  while( true )
  {
    std::cout << std::endl << "Waiting execution" << std::endl;

    callBackLock.lock();

    if( callback.m_executedTaskIds.size() == 2 )
    {
      std::cout << std::endl << "All tasks executed" << std::endl;

      callBackLock.unlock();

      executer.stopExecution();
      ASSERT_EQ( executer.getExecuterStatus(), gdma::tp::Executer::INACTIVE_EXECUTER );

      std::cout << std::endl << "Executer stopped" << std::endl;

      break;
    }
    else
    {
      callBackLock.unlock();
    }

    std::this_thread::sleep_for(std::chrono::milliseconds(1000));
  }


  ASSERT_TRUE( executer.getExecuterStatus() == gdma::tp::Executer::INACTIVE_EXECUTER );
  ASSERT_EQ( executer.isTaskPresent( taskId1 ), true );
  ASSERT_EQ( executer.isTaskPresent( taskId2 ), true );
  ASSERT_TRUE( executer.getTaskStatus( taskId1, taskStatus ) );
  ASSERT_EQ( taskStatus, gdma::tp::GraphExecuter::FINISHED );
  ASSERT_TRUE( executer.getTaskStatus( taskId2, taskStatus ) );
  ASSERT_EQ( taskStatus, gdma::tp::GraphExecuter::FINISHED );

  ASSERT_TRUE( executer.removeTask( taskId1, &taskPtr ) );
  ASSERT_EQ( taskPtr, execTaskPtr1 );
  ASSERT_EQ( taskPtr->getExecutionResult().getBasicResult(), gdma::tp::TaskExecutionResult::SUCCESS );
  delete taskPtr;

  ASSERT_TRUE( executer.removeTask( taskId2, &taskPtr ) );
  ASSERT_EQ( taskPtr, execTaskPtr2 );
  ASSERT_EQ( taskPtr->getExecutionResult().getBasicResult(), gdma::tp::TaskExecutionResult::SUCCESS );
  delete taskPtr;

  ASSERT_EQ( counter, 2 );

  ASSERT_EQ( callback.m_startedTaskIds.size(), 2 );
  ASSERT_TRUE( std::find( callback.m_startedTaskIds.begin(), callback.m_startedTaskIds.end(), taskId1 ) != callback.m_startedTaskIds.end() );
  ASSERT_TRUE( std::find( callback.m_startedTaskIds.begin(), callback.m_startedTaskIds.end(), taskId2 ) != callback.m_startedTaskIds.end() );

  ASSERT_EQ( callback.m_executedTaskIds.size(), 2 );
  ASSERT_TRUE( std::find( callback.m_executedTaskIds.begin(), callback.m_executedTaskIds.end(), taskId1 ) != callback.m_executedTaskIds.end() );
  ASSERT_TRUE( std::find( callback.m_executedTaskIds.begin(), callback.m_executedTaskIds.end(), taskId2 ) != callback.m_executedTaskIds.end() );
}


